// File:  C:\cbproj\Remote_CW_Keyer\CwGen.c
// Date:  2023-11-22
// Author:  Wolfgang Buescher (DL4YHF)
// Purpose: State-machine-driven 'CW Generator' .
//          Generates Morse code from text (plain "C" strings).
//          Periodically called from the 'keyer thread' (or task,
//          or timer interrupt, or whatever on a microcontroller).
//          Depends on the 'Elbug' module, at least for the functions
//          to convert text ("ASCII") into the shift register values
//          with the dash- and dot pattern for each character.
//

#include <string.h> // no string functions used here, but memset()

#include "Elbug.h" // header for the basic 'Elbug' functions (plain C)
#include "CwGen.h" // header for THIS module (Morse code "generator")


//----------------------------------------------------------------------------
void CwGen_InitInstanceWithDefaults( T_CwGen *pCwGen )
  // Called from the main application (or "GUI") before loading an actual
  // configuration FROM A FILE. The results (in *pConfig) are the "defaults".
{
  memset( pCwGen, 0, sizeof(T_CwGen) );
  pCwGen->cfg.iDotTime_ms = Elbug_WordsPerMinuteToDotTimeInMilliseconds( 25 );

} // end CwGen_InitInstanceWithDefaults()

//---------------------------------------------------------------------------
static void CwGen_ReloadCountdownForDashOrDot( T_CwGen *pCwGen, int nDotTimes )
{
  // Due to the calling interval of CwGen_Handler() with approximately 2 ms,
  // simply setting
  //   pCwGen->i32CountdownForDashOrDot_us = pCwGen->cfg.iDotTime_ms * 1000 * nDotTimes
  // would cause slipping of timestamps. For example, when sent at extreme speeds,
  // the word "PARIS" took 44 .. 45 "dot times", which should have been 43:
  //     P  A  R  I  S
  //  .--. .- .-. .. ...
  //   11 + 5+ 7 + 3 + 5 (length of characters in dot times) \  total =
  //      3 +3  +3 +3 (gaps between characters)              /   43 dots
  // Fixed this by ADDING the new element duration to whatever remained
  // in pCwGen->i32CountdownForDashOrDot_us (up to 3000 microseconds) :
  if( (pCwGen->i32CountdownForDashOrDot_us < -3000) || (pCwGen->i32CountdownForDashOrDot_us > 3000) )
   { pCwGen->i32CountdownForDashOrDot_us = 0;
   }
  pCwGen->i32CountdownForDashOrDot_us += pCwGen->cfg.iDotTime_ms * 1000 * nDotTimes;

} // end CwGen_ReloadCountdownForDashOrDot()


//---------------------------------------------------------------------------
static void CwGen_StartDashOrDot( T_CwGen *pCwGen )
  // Does what the name implies. Only called from CwGen_Handler() .
{
  pCwGen->fMorseOutput = TRUE;
  // pCwGen->wShiftReg still contains <pCwGen->nDashesAndDotsInShiftReg>
  // elements to send, with the NEXT (or first) element to send in the
  // most significant bit (the "start bit", which is used to encode the
  // length in A SINGLE BIT of the Morse code pattern) has already been
  // shifted out *to the left*, when counting nDashesAndDotsInShiftReg.
  // Thus:
  if( pCwGen->wShiftReg & 0x8000 ) // next element to send is a DASH :
   { CwGen_ReloadCountdownForDashOrDot( pCwGen, 3 );
   }
  else // next element to send is a DOT :
   { CwGen_ReloadCountdownForDashOrDot( pCwGen, 1 );
   }
  pCwGen->wShiftReg <<= 1;
  --pCwGen->nDashesAndDotsInShiftReg;
  pCwGen->iState = CW_GEN_SEND_DASH_OR_DOT;
}

//---------------------------------------------------------------------------
static void CwGen_StartGap( // start a gap, with duration of 1..7 dot times
              T_CwGen *pCwGen,
              int nDotTimes )
  // Only called from CwGen_Handler() .
{
  pCwGen->fMorseOutput = FALSE;
  CwGen_ReloadCountdownForDashOrDot( pCwGen, nDotTimes );
  pCwGen->iState = CW_GEN_SEND_GAP;  // new state : "sending a gap"
}


//---------------------------------------------------------------------------
static void CwGen_OnEndOfChar(T_CwGen *pCwGen) // called from CwGen_Handler()
  // at the end of a character's last dash or dot -> "end of character" .
{
  pCwGen->fMorseOutput = FALSE;
  // TWO dot-times have already expired, so if TWO or THREE MORE dot times expire,
  // it's time to emit a SPACE character. Someone specified (or at least explained):
  //   > The space between letters is 3 time units.
  //   > The space between words is 7 time units.
  // In this implementation, a "time unit" is pCwGen->cfg.iDotTime_ms [N milliseconds] .
  // To allow "tolerance" for operators using a slightly shorter inter-WORD-space,
  // a SPACE CHARACTER will be emitted by the keyer (when sending) after FIVE time-units.
  // An advanced keyer firmware may extend the 'word space' to those SEVEN time-units.
  CwGen_ReloadCountdownForDashOrDot( pCwGen, 3 ); // two time-units already expired, so wait for at least THREE additional ones
  pCwGen->iState = ES_WAIT_FOR_SPACE;  // new state : "wait for the minimum gap to emit a SPACE character"
} // end CwGen_OnEndOfChar()


//---------------------------------------------------------------------------
static BOOL CwGen_PrepareSendingNextChar(T_CwGen *pCwGen)
  // NO API FUNCTION ! Only called internally, from CwGen_Handler() .
  // [in]  pCwGen->szTxMemory[pCwGen->iTxMemoryCharIndex++]
  // [out] pCwGen->wShiftReg, wCurrentCwChr, nDashesAndDotsInShiftReg, iState .
  // [return] TRUE = "ok, there's more to send"
  //          FALSE= "nothing more to send in the TX buffer"
{
  char *cpSrc;
  WORD wCwChar;
  if(  (pCwGen->iTxMemoryCharIndex >= 0 )
    && (pCwGen->iTxMemoryCharIndex < CW_GEN_TX_MEMORY_SIZE )
    && (pCwGen->szTxMemory[ pCwGen->iTxMemoryCharIndex ] != '\0') )
   { // At least one character to send .. convert from ASCII to Morse code,
     // count the number of dashes and dots to send,
     // then turn the generator on:
     cpSrc = pCwGen->szTxMemory + pCwGen->iTxMemoryCharIndex;
     wCwChar = Elbug_ParseMorseCodePatternFromASCII( &cpSrc );
     if( wCwChar != 0 )  // successfully retrieved the next 'Morse code pattern'
      { // Elbug_ParseMorseCodePatternFromASCII() has incremented the "char pointer"
        // by the number of characters actually parsed (usually only ONE character,
        // but it may have been more than that for 'prosigns' like "^AR ^SK" .
        //   ,------------------------------------------------------|_|
        //   '--> cpSrc would have been incremented by 3 in this example.
        pCwGen->iTxMemoryCharIndex = (int)(cpSrc - pCwGen->szTxMemory);
        pCwGen->wCurrentCwChr = wCwChar; // remember this for the timing scope
        // Next state: CW_GEN_SEND_DASH_OR_DOT (for non-space characters)
        //             or CW_GEN_SEND_GAP (for space characters ~ "inter-word gap")
        if( wCwChar & CW_CHR_SPACE )  // bit 15 set -> it's a SPACE ...
         { // most likely, an "inter-word space", SEVEN dots long,
           // but since the previously sent character already included
           // a THREE-dot "inter-character space", the additional time
           // to wait (for the gap) are only FOUR dot-times :
           CwGen_StartGap( pCwGen, 4/*nDotTimes*/ );
         }
        else // not a SPACE but a character, beginning with a dash or dot.
         { // First determine the number of dashes and dots to send,
           // by bitwise shifting wCwChar to THE LEFT, until the "start bit"
           // arrives in bit 15. If the startbit ALREADY WAS in bit 15,
           // there would be FIFTEEN dashes or dots to send.. but that's
           // impossible, because bit 15 in wCwChar is CW_CHR_SPACE (0x8000).
           // Example: CW_CHR_V = 0x0011 = 0b00000000 00010001' ( "V" = ...- )
           // ,------------------------------------------'||||
           // '--> "Start bit"            three dots -----'''|
           //                               one dash --------'
           pCwGen->nDashesAndDotsInShiftReg = 15;
           while( pCwGen->nDashesAndDotsInShiftReg > 0 )
            { if( wCwChar & 0x8000 ) // "start bit" arrived in bit 15 ?
               { break;
               }
              wCwChar <<= 1;
              --pCwGen->nDashesAndDotsInShiftReg;
            }
           pCwGen->wShiftReg = wCwChar << 1; // startbit "falls out"
             // (from bit 15 into the non-existing bit 16 in a 16-bit "WORD")
           // Example ("V") : pCwGen->wShiftReg = 4096 = 0b0001 0000 0000 0000;
           //                                              ||||
           //                                             "...-"
           //                 pCwGen->nDashesAndDotsInShiftReg = 4 .
           CwGen_StartDashOrDot( pCwGen );
         }  // end else < not a SPACE >
        return TRUE;
      }    // end if( wCwChar != 0 )
     else // Elbug_ParseMorseCodePatternFromASCII() failed to parse
      {   // pCwGen->szTxMemory[] at pCwGen->iTxMemoryCharIndex, so STOP here ?
        pCwGen->iState = CW_GEN_OFF;
        return FALSE;
      }
   }
  else
   { pCwGen->iState = CW_GEN_OFF;
     return FALSE;
   }

} // end CwGen_PrepareSendingNextChar()


//---------------------------------------------------------------------------
int CwGen_Handler( T_CwGen *pCwGen, // API for the *worker thread*
       long nMicrosecondsSinceLastCall ) // [in] ideally 2000 us, if windoze permits !
  // Values returned by CwGen_Handler() : One of the CW_CHR-codes (More code
  // pattern emitted on each new 'character' (or prosign) emitted from here,
  // or one of the ELBUG_RESULT_.. values defined in Elbug.h, for example
  //               ELBUG_RESULT_NOTHING_DECODED  ( == 0 ) .
  // Called periodically from a time-critical worker thread / task,
  //        or even from a timer interrupt handler .
{
  char c;
  int  iResult = ELBUG_RESULT_NOTHING_DECODED;

  if( pCwGen->i32CountdownForDashOrDot_us > 0 )  // software timer still running ?
   {  pCwGen->i32CountdownForDashOrDot_us -= nMicrosecondsSinceLastCall; // let it COUNT DOWN,
      // until it stops at zero or with a "slightly negative value".
   }
  switch( pCwGen->iState )
   {
     case CW_GEN_OFF     : // CwGen_Handler() won't do anything
        pCwGen->fMorseOutput = FALSE;
        break;
     case CW_GEN_START   : // Let CwGen_Handler() START if there's somthing to send
        pCwGen->fMorseOutput = FALSE;
        CwGen_PrepareSendingNextChar( pCwGen ); // -> wShiftReg, nDashesAndDotsInShiftReg
           // '--> next state: CW_GEN_SEND_DASH_OR_DOT (for non-space characters)
           //               or CW_GEN_SEND_GAP (for space characters ~ "inter-word gap")
        break; // end case CW_GEN_WAITING

     case CW_GEN_SEND_DASH_OR_DOT:  // sending a dot, waiting for the end, register "dash" for sequeezing
        if( pCwGen->i32CountdownForDashOrDot_us <= 0 ) // finished sending the element .. what next ?
         { pCwGen->fMorseOutput = FALSE;
           if( pCwGen->nDashesAndDotsInShiftReg > 0 ) // if more elements (dashes and/or dots) to send,
            { CwGen_StartGap( pCwGen, 1/*nDotTimes*/ );
            }
           else // just completed sending a character, so what next ?
            { CwGen_StartGap( pCwGen, 3/*nDotTimes*/ );
              if( ! (pCwGen->wCurrentCwChr & CW_CHR_SPACE ) )
               { iResult = pCwGen->wCurrentCwChr; // emit this "info" for the timing scope
               }
            }
         }
        break; // end case CW_GEN_SEND_DASH_OR_DOT

     case CW_GEN_SEND_GAP: // sending a GAP with the duration of N dots
        if( pCwGen->i32CountdownForDashOrDot_us <= 0 ) // finished sending a GAP .. what next ?
         {
           if( pCwGen->nDashesAndDotsInShiftReg > 0 ) // if more elements (dashes and/or dots) to send,
            { CwGen_StartDashOrDot( pCwGen );         // start sending the next ..
            }
           else
            { CwGen_PrepareSendingNextChar( pCwGen );
              // '--> next state: CW_GEN_SEND_DASH_OR_DOT (for non-space characters)
              //               or CW_GEN_SEND_GAP (for space characters ~ "inter-word gap")
              //          or even CW_GEN_OFF (when there's nothing more to send)
              iResult |= ELBUG_RESULT_BEGIN_NEW_CHAR;
            }
          }
         break; // end case CW_GEN_SEND_GAP

     default:  // oops...
        pCwGen->iState = CW_GEN_OFF;
        break;

        // ;======= end of CW GENERATOR states ===============================
   } // end switch( pCwGen->iState )

  return iResult;

} // end CwGen_Handler()


//---------------------------------------------------------------------------
void CwGen_StartReplay( T_CwGen *pCwGen, char* pszTextToSend )
  // Called from the GUI when the operator wants to send text
  //   from one of the "Memories".
  // [in] pszTextToSend : text from the to-be-played 'keyer memory'.
  // The 'real work' happens later, when the keyer thread (or task on a uC)
  // calls CwGen_Handler() every two milliseconds or so...
{
  strncpy( pCwGen->szTxMemory, pszTextToSend, CW_GEN_TX_MEMORY_SIZE );
  // ToDo : Replace certain "macros" in the message text, like the
  //        "contest serial number" ( ^NNN in DL4YHF's QRP Keyer with a PIC ).
  pCwGen->iTxMemoryCharIndex = 0;
  pCwGen->iState = CW_GEN_START;
  // Note: Sending "from memory" automatically stops on the first
  //       zero-byte in pCwGen->szTxMemory, or when
  //       pCwGen->iTxMemoryCharIndex exceeds CW_GEN_TX_MEMORY_SIZE .
  //       Dont't call CwGen_PrepareSendingNextChar() from here,
  //       because due to the complete (and intended) lack of thread synchro-
  //       nisation ('critital sections', etc), functions that are called from the
  //       orker thread (like CwGen_Handler() -> CwGen_PrepareSendingNextChar())
  //       shall not be called from API functions (intended to be called from
  //       the GUI), like CwGen_StartReplay() .
} // end CwGen_StartReplay()

//---------------------------------------------------------------------------
void CwGen_StopReplay( T_CwGen *pCwGen )
{
  pCwGen->iState = CW_GEN_OFF; // too trivial for an extra subroutine,
                               // but this may change in future versions
} // end CwGen_StopReplay()

/* EOF < CwGen.c > */

