//---------------------------------------------------------------------------
// File: C:\cbproj\Remote_CW_Keyer\Keyer_GUI.cpp
// Date: 2024-11-29
// Author: Wolfgang Buescher (DL4YHF)
// Contains a part(!) of the Remote CW Keyer's graphic user interface,
//          with as few dependencies on the VCL-based 'main form' itself.
// Many variables and functions were moved from Keyer_Main.cpp / *.h
// into Keyer_GUI.cpp / *.h, to avoid duplicating them in the *.cpp modules
//  of the 'main form' (TKeyerMainForm *KeyerMainForm), with one variant for
// Borland C++Builder V6, and another one for Embarcadero C++Builder V12 !
//
// Most recent modifications: See C:\cbproj\Remote_CW_Keyer\Keyer_GUI.cpp .
//  2024-11-25: First steps to migrate from Borland C++Builder V6
//              to Embarcadero C++Builder V12 "Athens", free Community Edition,
//              using the "Clang enhanced" compiler "bcc32c" (with "B"="Borland"?).
//              Details about migrating from C++Builder V6 to V12 are in file
//               Migrating_from_CppBuilder6_to_CppBuilder12.htm (blog style).
//    * Replaced HUNDREDS of "char pointers" by "const char *"
//      because 'modern compilers' (e.g. "ISO C++11") are
//      very pedantic about char pointers, and refuse to
//      use STRING LITERALS as a "char *" !
//      (details on that in C:\cbproj\YHF_Tools\StringLib.h, which, as a
//       low-footprint string processing library for embedded systems, had an
//       AWFUL LOT of "char *" arguments that are now "const char pointers"
//       whereever a function can promise that it NEVER modifies the string)
//    * Added a AnsiString()-constructor-call in DOZENS of places where the
//      text in a visual control needs to be tried as a nice old-fashioned
//      C string. Example: Instead of
//       > CwKeyer_Config.iComPortNumber_IN = ParseSerialPortNumberFromText( CB_MorseKeyPort->Text.c_str() );
//      now use
//       > CwKeyer_Config.iComPortNumber_IN = ParseSerialPortNumberFromText( AnsiString(CB_MorseKeyPort->Text).c_str() );
//      because almost any 'Visual Component' uses UnicodeString instead of the
//      old, Borland-proprietary AnsiString. Both string type have a method
//      "c_str()" which sounds as if they would retrieve the string as an
//      old-fashined "C"-string... but THEY DON'T ! Only AnsiString's "c_str()"
//      returns an 8-bit characters string, while UnicodeString's "c_str()" ...
//      > Returns a pointer to the underlying string data as const wchar_t*.
//      Ouch.
//    * Had to duplicate the main form (Keyer_Main.dfm) along with its C++
//      module (Keyer_Main.cpp and Keyer_Main.h) because when loading it
//      in Embarcadero C++Builder V12, the new IDE and form designer
//      immediately trashed (automatically converted) the *.dfm so it could
//      not be used in the old, trusted, and never-expiring Borland C++Builder V6
//      IDE. Thus, the main form / GUI module source for *Embarcadero BCB V12"
//      are in Remote_CW_Keyer/BCB12_Project/*.dfm, *.cpp, *.h and *.cbproj now,
//      while "all the rest" (the REALLY interesting stuff for other developers)
//      remains in Remote_CW_Keyer/*.c, *.h, and a homeopathic dose of *.cpp .
//    * Because 'Embarcadero C++Builder V12' cannot import an
//      old C++Builder project- or make file (here: Remote_CW_Keyer.bpr),
//      the entire project had to be 'clicked together from scratch'.
//      To simplify that, here are the C / C++  modules in the old *.bpr file,
//      with paths relative to the Remote_CW_Keyer directory as PLAIN ASCII:
//        * Remote_CW_Keyer.cpp  (that's the small, IDE-generated "main module")
//        * switches.h           (just a header file, but very important)
//        * Keyer_Main.cpp       (that's the file tightly coupled with the 'main form')
//        * Keyer_GUI.cpp        (part of the graphic user interface, NOT managed by the IDE)
//        * Utilities.c
//        * Timers.c
//        * KeyerThread.c        (you guessed it, THIS is the interesting part, no VCL dependencies, hooray)
//        * Debouncer.c
//        * Elbug.c
//        * StraightKeyDecoder.c
//        * CwStreamEnc.c
//        * CwGen.c
//        * SampleRateConv.c
//        * ALawCompression.c
//        * CwDSP.c
//        * CwNet.c
//        * CircularFifo.c
//        * HamlibServer.c
//        * HttpServer.c
//        * dsound_wrapper.c
//        * Inet_Tools.c
//        * LinradColors.c
//        * HelpMap.c
//        * VorbisStream.c  (possibly not used anymore due to the extreme audio latency)
//        * RigControl.c
//        * JS_WebAudio.c
//        * OpenWebRX_Server.c
//        * TIconToFavicon.cpp
//        * SpecDisp.cpp
//        * Translator.cpp
//        * Translations.c
//        * SunriseColors.c
//        * ..\YHF_Tools\SL_Append.c  ( part of DL4YHF's "String Library" for classic "C" string manipulation... )
//        * ..\YHF_Tools\SL_ftoa.c
//        * ..\YHF_Tools\SL_hash.c
//        * ..\YHF_Tools\SL_JSON.c
//        * ..\YHF_Tools\SL_memclr.c
//        * ..\YHF_Tools\SL_ParseIP.c
//        * ..\YHF_Tools\str2int.c
//        * ..\YHF_Tools\stricmp.c
//        * ..\YHF_Tools\StrEncod.c
//        * ..\YHF_Tools\strnlen.c
//        * ..\YHF_Tools\strnstr.c
//        * ..\YHF_Tools\strsep.c
//        * ..\YHF_Tools\QFile.c
//        * ..\YHF_Tools\YHF_Dialogs.cpp
//        * ..\YHF_Tools\YHF_Help.cpp
//        * C:\CBuilder6\Lib\Psdk\dsound.lib (or wherever they have placed the "direct sound" library now)
//        * Not listed here because probably removed soon: "Libvorbis" for COMPRESSED audio streams

#include "switches.h"   // project specific compiler switches ("options"),
                        // must be included before anything else !
#include "yhf_type.h"   // classic types like BYTE, WORD, DWORD, BOOL, ..
#include <windows.h>    // Must be included BEFORE vcl.h for some strange reason.
                        // Contains stuff like WAVEINCAPS, WAVEOUTCAPS, etc^255 .
#include <vcl.h>        // Borland's stoneage Visual Component Library
#include <buttons.hpp>  // TBitBtn       a la Borland VCL (but why in an extra header?)
#include <ComCtrls.hpp> // TTabSheet     a la Borland VCL ...
#include <CheckLst.hpp> // TCheckListBox a la Borland VCL ....
#include <Grids.hpp>    // TStringGrid   a la Borland VCL .....
#include <string.h>
#include <stdio.h>      // not using "standard I/O" here, but e.g. sprintf()
#include <math.h>
#pragma hdrstop

#include "Utilities.h"   // stuff like UTL_iWindowsVersion, UTL_iAppInstance, ShowError(), etc
#include "YHF_Dialogs.h" // API for a few common dialogs, e.g. YHF_RunStringEditDialog()
#include "YHF_Help.h"    // DL4YHF's HTML-based replacement for the defunct *.hlp system
#include "HelpIDs.h"     // application specific help topic identifiers
#include "Translator.h"  // APPL_TranslateAllForms() [only works with Borland VCL]
#include "Translations.h" // tranlation table (static, built-in string table w/o whistles and bells)

#if( SWI_USE_DSOUND ) // use "Direct Sound" / dsound_wrapper.c ?
# include "dsound_wrapper.h"
#endif // SWI_USE_DSOUND ?

#if( SWI_USE_WAVE_AUDIO || SWI_USE_MIDI )
# include <mmsystem.h> // maybe MIDI is fast enough to start / stope playing notes ?
#endif // SWI_USE_WAVE_AUDIO or SWI_USE_MIDI ?

#include "StringLib.h"
#include "Elbug.h"   // old 'Elbug' functions, converted from PIC-assembler to "C"
#include "CwGen.h"   // CW generator (converts text to Morse code)
#include "CwDSP.h"   // CW-'Digital Signal Processor' / sidetone generator
#include "CwNet.h"   // Socket-based 'Client or Server' for the Remote CW Keyer
#include "CwKeyer.h" // prototypes for the 'worker thread' that stitches all together
#include "Inet_Tools.h" // Base64 encoding/decoding, SHA1 calculation, etc..
#if(SWI_USE_HTTP_SERVER) // build an application with integrated HTTP server ?
# include "HttpServer.h" // formerly simple HTTP server (turned into a monster)
# include "TIconToFavicon.h" // function to convert the application's icon
                             // into "Favicon.ico" for the web server .
#endif // SWI_USE_HTTP_SERVER ?

#include "SpecDisp.h"  // SpecDisp_UpdateSpectrum(), ..Waterfall(), FreqScale(),..
#include "Keyer_GUI.h" // subroutines formerly in Keyer_Main.cpp (now only CALLED from there)

HINSTANCE APPL_hInstance= 0; // handle to current instance, a Win32 API thing
BOOL APPL_fLaunchKeyerOnStart = FALSE;
BOOL APPL_fPausedDuringSleep  = FALSE;
BOOL APPL_fSaveSettingsOnExit = FALSE;

char g_sz255CommandLine[256]     = "";
char g_sz255PathToDataFiles[256] = "";

// Colours for various GUI components that we'd like to control via SetColourScheme():
// (but convincing Windows and the VCL to actually USE these colours turned into
//  a nightmare; especially for stuff in the 'Non-Client Area', e.g. title and main menu)
TColor g_clMenuBackground, g_clMenuSelBackgnd, g_clMenuForeground;
TColor g_clDecoderOutputForeground, g_clDecoderOutputBackground;
int    g_iNewLanguage = TRANSLATOR_LANGUAGE_ENGLISH;
int    g_iUserInterfaceMode = UI_MODE_COMPLETE;

volatile DWORD g_dwDebugMessage,  // kluge for debugging / dummy assignments for breakpoints..
               g_dwDebugWParam, g_dwDebugLParam;
volatile int   g_iDebugDummyIndex;
int g_iTimingScopeBitmapWidth = 0;  // <- only for diagnostic / trouble with "destroyed forms"

int KeyerGUI_iBandStackingRegIndices[KEYER_GUI_MAX_BANDS][KEYER_GUI_MAX_STACKING_REGS_PER_BAND];
           // '--> filled in FillComboWithBandList(), used in CB_BandClick()

T_CwNet MyCwNet; // <- not a shiny C++ class, but a single instance representing the "CW Network"
DSoundWrapper MyDirectSound; // a "DirectSound wrapper" instance, first used for the SIDETONE OUTPUT

#if( SWI_USE_HAMLIB_SERVER ) // build an application with integrated "Hamlib Net rigctld"-compatible server ?
T_HLSrv MyHamlibServer; // Hamlib-'rigctld'-compatible server (guess the 'd' means 'daemon',
       // but neither demons nor daemons lurking in here. There's NOTHING running 'in the background'.)
#endif // SWI_USE_HAMLIB_SERVER ?


T_ErrorFifoEntry ErrorHistoryFifo[ C_ERROR_FIFO_SIZE ]; // <- classic lock-free circular FIFO between ShowError() and the GUI
int DEBUG_iErrorHistoryHead = 0;
int DEBUG_iErrorHistoryTail = 0;
T_TimingScopeOverlay TimingScopeOverlay[N_SCOPE_OVERLAYS];
int TimingScope_iCurrentOverlay = 0;
int TimingScope_iUpdateCountOnTestTab, TimingScope_iUpdateCountOnKeyerTab;

T_SpecDispControl g_SpecDispControl = { 0 };

//----------------------------------------------------------------------------
// Global variables for hardcore debugging / post-mortem crash analysis.
//        Useful when the debugger's call stack shows nothing but garbage,
//        e.g. after an exception at the infamous address 0xFEEEFEEE .
//        ( C++Builder's debugger could still inspect SIMPLE GLOBAL
//          VARIABLES after most kinds of exceptions, but of course nothing
//          stack-based because the CPU- or task-stack was usually trashed.)
//----------------------------------------------------------------------------
#if(SWI_HARDCORE_DEBUGGING) // (1) = hardcore-debugging, (0)=normal compilation
 int GUI_iLastSourceLine = 0; // WATCH THIS after crashing with e.g. "0xFEEEFEEE"  ...
# define HERE_I_AM__GUI()  GUI_iLastSourceLine=__LINE__
# if( 0 )
    If certain worker threads of the application don't seem to terminate
    "politely", watch the following (complete list only HERE, in Keyer_Main.cpp):
     * GUI_iLastSourceLine
     * DSW_iLastSourceLine
     * DSP_iLastSourceLine
     * Keyer_iLastSourceLine
     * CwNet_iLastSourceLine
     * HttpSrv_iLastSourceLine
     * RigCtrl_iLastSourceLine
#  endif // (0) .. kludge to inspect the above variables in Borland via mouse-over ...
  extern int DSW_iLastSourceLine, DSP_iLastSourceLine, Keyer_iLastSourceLine,
             CwNet_iLastSourceLine, HttpSrv_iLastSourceLine;
  CPROT void CheckSystemHealth(const char *pszModuleName, int iSourceLine); // <- periodically called via macro CHECK_SYSTEM_HEALTH()
  static  const char *s_ModuleName = "KeyerMain";
# define CHECK_SYSTEM_HEALTH() CheckSystemHealth(s_ModuleName,__LINE__)
#else
# define HERE_I_AM__GUI()      /* rien */
# define CHECK_SYSTEM_HEALTH() /* nada */
#endif // SWI_HARDCORE_DEBUGGING ?

const T_SL_TokenList C_TL_None_Zero[] =
{ { "NONE",  0 },
  { NULL, 0 } // "all zeros" mark the end of the list
};

const T_SL_TokenList SerialPortOutputSignals_Key[] =
{ { "DTR (DE-9 pin 4) on KEYER port",            KEYER_SIGNAL_INDEX_MORSE_KEY_DTR },
  { "DTR (DE-9 pin 4) on KEYER port, INVERTED", -KEYER_SIGNAL_INDEX_MORSE_KEY_DTR },
  { "RTS (DE-9 pin 7) on KEYER port",            KEYER_SIGNAL_INDEX_MORSE_KEY_RTS },
  { "RTS (DE-9 pin 7) on KEYER port, INVERTED", -KEYER_SIGNAL_INDEX_MORSE_KEY_RTS },
  { NULL, 0 } // "all zeros" mark the end of the list
};

const T_SL_TokenList SerialPortOutputSignals_Radio[] =
{ { "DTR (DE-9 pin 4) on RADIO port",            KEYER_SIGNAL_INDEX_RADIO_KEYING_DTR },
  { "DTR (DE-9 pin 4) on RADIO port, INVERTED", -KEYER_SIGNAL_INDEX_RADIO_KEYING_DTR },
  { "RTS (DE-9 pin 7) on RADIO port",            KEYER_SIGNAL_INDEX_RADIO_KEYING_RTS },
  { "RTS (DE-9 pin 7) on RADIO port, INVERTED", -KEYER_SIGNAL_INDEX_RADIO_KEYING_RTS },
  { NULL, 0 } // "all zeros" mark the end of the list
};

const T_SL_TokenList SerialPortInputSignals_Key[] =
{ { "DCD (DE-9 pin 1) on KEYER port",            KEYER_SIGNAL_INDEX_MORSE_KEY_DCD },
  { "DCD (DE-9 pin 1) on KEYER port, INVERTED", -KEYER_SIGNAL_INDEX_MORSE_KEY_DCD },
  { "DSR (DE-9 pin 6) on KEYER port",            KEYER_SIGNAL_INDEX_MORSE_KEY_DSR },
  { "DSR (DE-9 pin 6) on KEYER port, INVERTED", -KEYER_SIGNAL_INDEX_MORSE_KEY_DSR },
  { "CTS (DE-9 pin 8) on KEYER port",            KEYER_SIGNAL_INDEX_MORSE_KEY_CTS },
  { "CTS (DE-9 pin 8) on KEYER port, INVERTED", -KEYER_SIGNAL_INDEX_MORSE_KEY_CTS },
  { "RI  (DE-9 pin 9) on KEYER port",            KEYER_SIGNAL_INDEX_MORSE_KEY_RI  },
  { "RI  (DE-9 pin 9) on KEYER port, INVERTED", -KEYER_SIGNAL_INDEX_MORSE_KEY_RI  },
  { NULL, 0 } // "all zeros" mark the end of the list
};

const T_SL_TokenList SerialPortInputSignals_Radio[] =
{ { "DCD (DE-9 pin 1) on RADIO port",            KEYER_SIGNAL_INDEX_RADIO_KEYING_DCD },
  { "DCD (DE-9 pin 1) on RADIO port, INVERTED", -KEYER_SIGNAL_INDEX_RADIO_KEYING_DCD },
  { "DSR (DE-9 pin 6) on RADIO port",            KEYER_SIGNAL_INDEX_RADIO_KEYING_DSR },
  { "DSR (DE-9 pin 6) on RADIO port, INVERTED", -KEYER_SIGNAL_INDEX_RADIO_KEYING_DSR },
  { "CTS (DE-9 pin 8) on RADIO port",            KEYER_SIGNAL_INDEX_RADIO_KEYING_CTS },
  { "CTS (DE-9 pin 8) on RADIO port, INVERTED", -KEYER_SIGNAL_INDEX_RADIO_KEYING_CTS },
  { "RI  (DE-9 pin 9) on RADIO port",            KEYER_SIGNAL_INDEX_RADIO_KEYING_RI  },
  { "RI  (DE-9 pin 9) on RADIO port, INVERTED", -KEYER_SIGNAL_INDEX_RADIO_KEYING_RI  },
  { NULL, 0 } // "all zeros" mark the end of the list
};

const T_SL_TokenList RigControlMethods[] =
{ { "No rig control, only CW keying (or client mode)", RIGCTRL_PROTOCOL_NONE },
  { "Icom CI-V, using the 'COM port for CW keying'", RIGCTRL_PROTOCOL_ICOM_CI_V },
  { "Hamlib Net rigctl via remote TCP/IP server", RIGCTRL_PROTOCOL_HAMLIB_RIGCTLD  },
  { NULL, 0 } // "all zeros" mark the end of the list
};

const T_SL_TokenList KeyTypes[] =
{ { "None (no CW key)", KEY_TYPE_PASSIVE  },
  { "Straight",         KEY_TYPE_STRAIGHT },
  { "Paddle, IAMBIC MODE B", KEY_TYPE_IAMBIC_B },
  { "Paddle, IAMBIC MODE A", KEY_TYPE_IAMBIC_A },
  { "Basic iambic w/o dash/dot memory", KEY_TYPE_IAMBIC_NO_MEM},
  { NULL, 0 } // "all zeros" mark the end of the list
};

const T_SL_TokenList SidetonesOnTXD[] =  // Sidetone tapped from a serial port's TXD(!) output:
{ { "None",   KEYER_SIDETONE_NONE       },
  { "480 Hz", KEYER_SIDETONE_TXD_480_HZ }, // 4800 "baud" divided by 10 bits (standard baudrate)
  { "960 Hz", KEYER_SIDETONE_TXD_960_HZ }, // also 4800 "baud" but with TWO pulses in the TXD pattern
  { NULL, 0 } // "all zeros" mark the end of the list
};

const T_SL_TokenList SidetonesOnAudioOut[] =  // Sidetones sent to an audio device (with higher latency than "TXD sidetone"):
{ { "None",   0   },
  { "400 Hz", 400 },
  { "450 Hz", 450 }, // ~~ Icom's name "CW Pitch" (but here, it's only the LOCAL KEYER SIDETONE)
  { "500 Hz", 450 },
  { "550 Hz", 550 },
  { "600 Hz", 600 },
  { "650 Hz", 650 },
  { "700 Hz", 700 },
  { "750 Hz", 750 },
  { "800 Hz", 800 },
  { "1000 Hz",1000},
  { NULL, 0 } // "all zeros" mark the end of the list
};

const T_SL_TokenList SidetoneRiseTimes[] =  // "Rise"-, aka "Ramp"-times for sidetone VIA AUDIO OUTPUT
{ { "0 ms (hard-keyed)", 0 },
  { "2 ms",  2 },
  { "4 ms",  4 },
  { "6 ms",  6 },
  { "8 ms",  8 },
  { "20 ms (very soft)",20 }, // huuh-huh, thought Mr Bubo Bubo, and flew away when he heard this
  { NULL, 0 } // "all zeros" mark the end of the list
};


const char *CwKeyer_psz6MemoryDefaults[6] =
{ "Remote CW Keyer. Press F1 to send this message,",
  "Testing remote CW keying. Press F2 to send this message.",
  "Message Nr 3. Click into these fields to edit.",
  "For details, please read the friendly manual.",
  "",
  ""
};


//---------------------------------------------------------------------------
// Helper functions ( no shiny VCL class methods ! )
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
void ParseCommandLine(const char *pszCmdLine)
  // Some examples (copy & paste to CBuilder IDE: Run..Parameters)
{
  const char *cp = pszCmdLine;
  const char *cp2;
  char *cp3;
  char c;
  char sz255Temp[256], *pszDest=sz255Temp, *pszEndstop=sz255Temp+255;

  if( *cp=='"' ) // guess this is the name of the executable -> grab 'our path' from it
   { cp2 = cp+1;
     SL_SkipDoubleQuotedString( &cp );  // <- returns the NUMBER OF CHARS SKIPPED in cp, *including* the double quotes
   }
  else  // Name of the executable NOT in double quotes (thanks for the snag..)
   { cp2 = cp;
     SL_SkipCharsUntilDelimiter( &cp, " ", SL_SKIP_NORMAL );
   }

  // One of the 'nice surprises' from Windoze is that the first parameter
  // on the command line sometimes does NOT contain a full path,
  // To avoid having the configuration files, the "debug run log",
  // and who-knows-what somewhere in the dark where Microsoft
  // wants to have them ( e.g. C:\Users\WerHatsInstalliert\AppData\Local | LocalLow | Roaming | Weiss-der-Teufel ..),
  // extract the full path to the EXECUTABLE (which is where all other files
  // have been unzipped to, without a shiny installer at all) :
  //
  if( ! UTL_GetPathToMyDataFiles( g_sz255PathToDataFiles, 255 ) )
   { // SOS .. unable to find out where "our files" should be for this user,
     //        on this windows version ..
     cp3 = strrchr( g_sz255PathToDataFiles, '\\' );
     if( cp3==NULL )  // quite unlikely that Microsoft uses POSIX one day, but... :
      {  cp3 = strrchr( g_sz255PathToDataFiles, '/' );
      }
     if(cp3!=NULL)  // if there's any kind of slash, TRUNCATE AFTER THAT CHAR..
      { cp3[1] = '\0'; // -> result e.g. g_sz255PathToDataFiles = "C:\\cbproj\Remote_CW_Keyer\\"
      }
   }
  if( ((c=SL_GetLastChar(g_sz255PathToDataFiles,255)) != '\\') && (c!='/') )
   { cp3 = g_sz255PathToDataFiles;
     SL_AppendChar( &cp3, cp2+255, '\\' );
   }

  // After the above alternative for "the first command line argument", parse the rest..
  while( *cp != '\0' ) // repeat until end-of-string
   { SL_SkipSpaces( &cp );
     switch( *cp )
      { case '"' :  // another double-quoted string ? -> throw it away
           if( SL_SkipDoubleQuotedString( &cp ) <= 0 )
            { return;  // something wrong with the syntax -> bail out
            }
           break;
        case '\\' :
        case '/'  : // forward or backward slash : Guess this is an "option"..
           ++cp;
           if( SL_SkipToken( &cp, "debug" ) ) // run this program with a "debug log" :
            { pszDest=sz255Temp;
              SL_AppendString( &pszDest, pszEndstop, g_sz255PathToDataFiles );
              SL_AppendString( &pszDest, pszEndstop, "debug_run_log.txt" );
              UTL_OpenRunLogFile( sz255Temp );
            }
           else if( SL_SkipToken( &cp, "xlate" ) ) // run this program in "translation test mode" ..
            { APPL_iLanguageTestMode = 1; // details in Remote_CW_Keyer\Translator.cpp !
            }
           break;
        default:
           return;  // something wrong with the syntax -> bail out
      }
   } // end while < more characters or substrings in the command line >

} // end ParseCommandLine()

//---------------------------------------------------------------------------
void LoadSettings(void) // .. from an old-fashioned *.ini file, NOT from the bloody registry !
{ // [out] CwKeyer_Config , and maybe A FEW more things for the GUI itself.
  const char *pszIniFileName = SWI_APP_EXE_NAME".ini"; // all instances use THE SAME ini file..
        // -> The full path MAY be C:\Windows\RemoteCwKeyer.ini ,
        //    but also seen :
        //  C:\Users\Wolf\AppData\Local\VirtualStore\Windows\RemoteCwKeyer.ini
        //   - seems to depend on FROM WHERE, and with which privileges,
        //     and by which USER the application was launched. Omg.
  char szSec[84];  // "section" (in the ini file, that's the string in squared brackets)
  char szKey[84];  // "key name" (that's the string on the left side of the assignment operator)
  char sz255Value[256];
  TEdit *pEdit;
  int   i;

  CwKeyer_SetDefaultConfig( &CwKeyer_Config );
  CwKeyer_InitTimingScope( &CwKeyer_TimingScope  );
  Elbug_InitInstanceWithDefaults( &CwKeyer_Elbug );
  CwGen_InitInstanceWithDefaults( &CwKeyer_Gen   );
  CwDSP_InitInstanceWithDefaults( &CwKeyer_DSP   );

  CwNet_InitInstanceWithDefaults( &MyCwNet );
  CwKeyer_Config.pCwNet = &MyCwNet; // 'marry' the keyer to the 'CW Network' instance
  MyCwNet.pCwDSP = &CwKeyer_DSP; // 'marry' the audio DSP to the 'CW Network' instance (for audio streaming in both directions)
     // '--> Sometimes, after STEPPING OVER the above instruction, MyCwNet.pCwDSP was still NULL !
     //   Hard to believe but true : Problem fixed by doing a BUILD ALL !
     //   Obviously the ancient IDE (Borland C++ Builder) lost track of header dependencies.

#if( SWI_USE_DSOUND ) // use "DirectSound" / dsound_wrapper.c ?
  CwKeyer_DSP.pDSW = &MyDirectSound; // let the 'DSP' use our DirectSound-wrapper-instance
  // (contains a list of audio-in- and -output devices already 'enumerated')
  if( CwKeyer_DSP.cfg.sz255AudioInputDevice[0] == '\0' )
   { strcpy( CwKeyer_DSP.cfg.sz255AudioInputDevice, "None" );
   }
  if( CwKeyer_DSP.cfg.sz255AudioOutputDevice[0] == '\0' )
   { // Obviously not set yet -> pick an audio output device for 'DirectSound'
     // that usually exists: THE FIRST ONE enumerated by our wrapper !
     if( MyDirectSound.m_nOutputDevices > 0 )
      { strncpy( CwKeyer_DSP.cfg.sz255AudioOutputDevice,
                 MyDirectSound.m_sOutputDevices[0].sz255DevName, 255 );
        // ,-----|_____________________________________________|
        // '--> e.g. "Primary Sound Driver", but ymmv ..
      }
   }
#endif // SWI_USE_DSOUND ?

  sprintf( szSec, "Keyer%d", 1+UTL_iAppInstance ); // e.g. "Keyer1" for the FIRST running instance, etc
  CwKeyer_Config.iComPortNumber_IN = UTL_ReadIntFromIniFile(pszIniFileName,
      szSec, "ComPort_IN"/*pszKeyName*/, CwKeyer_Config.iComPortNumber_IN/*default*/ );
  CwKeyer_Config.iRadioKeyingAndControlPort = UTL_ReadIntFromIniFile(pszIniFileName,
      szSec, "RadioControlAndKeyingPort", CwKeyer_Config.iRadioKeyingAndControlPort );
  CwKeyer_Config.iRadioControlProtocol = UTL_ReadIntFromIniFile(pszIniFileName,
      szSec, "RadioControlProtocol", CwKeyer_Config.iRadioControlProtocol );
  CwKeyer_Config.iRadioControlBaudrate = UTL_ReadIntFromIniFile(pszIniFileName,
      szSec, "RadioControlBaudrate", CwKeyer_Config.iRadioControlBaudrate );
  CwKeyer_Config.iRadioCIVAddress = UTL_ReadIntFromIniFile(pszIniFileName,
      szSec, "RadioCIVAddress", CwKeyer_Config.iRadioCIVAddress );
  UTL_ReadStringFromIniFile( pszIniFileName, szSec, "RemoteRigCtrlServerAddress",
                             CwKeyer_Config.sz80RemoteRigCtrlServerAddress/*Default*/,
                             CwKeyer_Config.sz80RemoteRigCtrlServerAddress, 80 );
#if( SWI_USE_HAMLIB_SERVER ) // not sure if this 'feature' is going to stay...
  CwKeyer_Config.fEnableBuiltInHamlibServer = UTL_ReadIntFromIniFile(pszIniFileName,
      szSec, "EnableBuiltInHamlibServer",CwKeyer_Config.fEnableBuiltInHamlibServer);
  MyHamlibServer.cfg.iServerListeningPort = UTL_ReadIntFromIniFile(pszIniFileName,
      szSec, "BuiltInHamlibServerPort", MyHamlibServer.cfg.iServerListeningPort);
  MyHamlibServer.cfg.iServerOptions = UTL_ReadIntFromIniFile(pszIniFileName,
      szSec, "BuiltInHamlibServerOptions", MyHamlibServer.cfg.iServerOptions);
  MyHamlibServer.cfg.iDiagnosticFlags= UTL_ReadIntFromIniFile(pszIniFileName,
      szSec, "BuiltInHamlibServerDiagnostics", MyHamlibServer.cfg.iDiagnosticFlags);
#endif // SWI_USE_HAMLIB_SERVER ?

  CwKeyer_Config.iMorseKeyType=UTL_ReadIntFromIniFile(pszIniFileName,szSec, "MorseKeyType",
      CwKeyer_Config.iMorseKeyType/* KEY_TYPE_PASSIVE/STRAIGHT/PADDLE */ );
  CwKeyer_Config.iDotInput  = UTL_ReadIntFromIniFile(pszIniFileName,szSec,"DotInputSignal", CwKeyer_Config.iDotInput );
  CwKeyer_Config.iDashInput = UTL_ReadIntFromIniFile(pszIniFileName,szSec,"DashInputSignal",CwKeyer_Config.iDashInput );
  CwKeyer_Config.iManualPTTInput=UTL_ReadIntFromIniFile(pszIniFileName,szSec,"ManualPTTInput",CwKeyer_Config.iManualPTTInput);
  CwKeyer_Config.iTestInput = UTL_ReadIntFromIniFile(pszIniFileName,szSec,"TestInputSignal",CwKeyer_Config.iTestInput );
  CwKeyer_Config.iKeySupply = UTL_ReadIntFromIniFile(pszIniFileName,szSec,"KeySupplySignal",CwKeyer_Config.iKeySupply );
  CwKeyer_Config.iSidetoneOnTXD  = UTL_ReadIntFromIniFile(pszIniFileName,szSec,"SidetoneOnTXD", CwKeyer_Config.iSidetoneOnTXD );
  CwKeyer_Config.iRadioCWKeying  = UTL_ReadIntFromIniFile(pszIniFileName,szSec,"CwKeyingSignal",CwKeyer_Config.iRadioCWKeying );
  CwKeyer_Config.iRadioPTTControl= UTL_ReadIntFromIniFile(pszIniFileName,szSec,"PTTControlSignal",CwKeyer_Config.iRadioPTTControl);
  CwKeyer_Config.iTxDelayTime_ms = UTL_ReadIntFromIniFile(pszIniFileName,szSec,"TxDelayTime_ms",CwKeyer_Config.iTxDelayTime_ms);
  CwKeyer_Config.iTxHangTime_ms  = UTL_ReadIntFromIniFile(pszIniFileName,szSec,"TxHangTime_ms",CwKeyer_Config.iTxHangTime_ms);
  CwKeyer_Config.fDebouncePaddleInputs=UTL_ReadIntFromIniFile(pszIniFileName,szSec,"DebouncePaddleInputs",CwKeyer_Config.fDebouncePaddleInputs);
  CwKeyer_Config.fKeyInputWatchdog=UTL_ReadIntFromIniFile(pszIniFileName,szSec,"KeyInputWatchdog",CwKeyer_Config.fKeyInputWatchdog);
  CwKeyer_Config.fKeyViaShiftAndControl = UTL_ReadIntFromIniFile(pszIniFileName,szSec,"CwViaShiftAndControlKey",
    ( CwKeyer_Config.iComPortNumber_IN < 0) ); // <- per default, only enable this "gadget" if there's no serial port for the Morse Key
  UTL_ReadStringFromIniFile( pszIniFileName, szSec, "AudioInputDevice",
                             CwKeyer_DSP.cfg.sz255AudioInputDevice/*Default*/,
                             CwKeyer_DSP.cfg.sz255AudioInputDevice, 255 );
  UTL_ReadStringFromIniFile( pszIniFileName, szSec, "AudioOutputDevice",
                             CwKeyer_DSP.cfg.sz255AudioOutputDevice/*Default*/,
                             CwKeyer_DSP.cfg.sz255AudioOutputDevice, 255 );
  CwKeyer_DSP.cfg.iSidetoneFreq_Hz = UTL_ReadIntFromIniFile(pszIniFileName,szSec,
         "AudioSidetone_Hz", 650 + 50 * UTL_iAppInstance/*gadget: try running THREE instances..*/ );
  CwKeyer_DSP.cfg.iSidetoneRiseTime_ms = UTL_ReadIntFromIniFile(pszIniFileName,szSec,
         "SidetoneRiseTime_ms", CwKeyer_DSP.cfg.iSidetoneRiseTime_ms );
  CwKeyer_DSP.cfg.iSidetoneGain_dB = UTL_ReadIntFromIniFile(pszIniFileName,szSec,
         "SidetoneGain_dB", CwKeyer_DSP.cfg.iSidetoneGain_dB );
  CwKeyer_DSP.cfg.iAudioInGain_dB = UTL_ReadIntFromIniFile(pszIniFileName,szSec,
         "AudioInGain_dB", CwKeyer_DSP.cfg.iAudioInGain_dB );
  CwKeyer_DSP.cfg.iAudioOutGain_dB = UTL_ReadIntFromIniFile(pszIniFileName,szSec,
         "AudioOutGain_dB", CwKeyer_DSP.cfg.iAudioOutGain_dB );
  CwKeyer_DSP.cfg.iNetworkTonesGain_dB = UTL_ReadIntFromIniFile(pszIniFileName,szSec,
         "NetworkTonesGain_dB", CwKeyer_DSP.cfg.iNetworkTonesGain_dB );

  CwKeyer_DSP.cfg.iAudioFlags = UTL_ReadIntFromIniFile(pszIniFileName, szSec,
          "AudioFlags"/*pszKeyName*/, CwKeyer_DSP.cfg.iAudioFlags/*default*/ );

  CwKeyer_Elbug.cfg.iDotTime_ms = Elbug_WordsPerMinuteToDotTimeInMilliseconds(UTL_ReadIntFromIniFile(pszIniFileName,szSec,"WordsPerMinute",25) );
  CwKeyer_Gen.cfg.iDotTime_ms = CwKeyer_Elbug.cfg.iDotTime_ms; // let 'playback' use the same speed as direct keyed Morse
  APPL_fLaunchKeyerOnStart  = UTL_ReadIntFromIniFile(pszIniFileName,szSec,"LaunchOnStart", APPL_fLaunchKeyerOnStart );
  CwKeyer_TimingScope.cfg.iChannel4Source = UTL_ReadIntFromIniFile(pszIniFileName,szSec,"Channel4Source", CwKeyer_TimingScope.cfg.iChannel4Source );
  CwKeyer_TimingScope.cfg.iVisibleChannels= UTL_ReadIntFromIniFile(pszIniFileName,szSec,"TScopeChannels", CwKeyer_TimingScope.cfg.iVisibleChannels);
  CwKeyer_TimingScope.cfg.iTriggerOptions = UTL_ReadIntFromIniFile(pszIniFileName,szSec,"TScopeTrigger",  CwKeyer_TimingScope.cfg.iTriggerOptions );

  // Still in the ini-file-section "KeyerN", where "N" is the instance index...
  CwKeyer_Config.iTRXOptions = UTL_ReadIntFromIniFile(pszIniFileName,
       szSec, "TRXOptions"/*pszKeyName*/, CwKeyer_Config.iTRXOptions/*default*/ );
       // -> bitwise combineable flag like KEYER_TRX_OPT_KEEP_RUNNING, etc,
       //    all defined for struct T_CwKeyerConfig in Remote_CW_Keyer/CwKeyer.h .

  // Load options for the built-in "Rig Control" module :
  sprintf( szSec, "RigControl%d", 1+UTL_iAppInstance ); // e.g. "RigControl1" for the FIRST running instance, etc
  CwKeyer_Config.dwRejectMessages4Log = (DWORD)UTL_ReadIntFromIniFile(pszIniFileName,szSec,"RejectMessagesInLog",
                             CwKeyer_Config.dwRejectMessages4Log/*Default*/ );
  RigCtrl_iTrafficMonitorDisplayOptions = UTL_ReadIntFromIniFile(pszIniFileName,szSec,"TrafficMonitorOptions",
                             RigCtrl_iTrafficMonitorDisplayOptions );



  // Load the "Network" configuration (for module CwNet.c; simple TCP-client or server):
  sprintf( szSec, "Network%d", 1+UTL_iAppInstance ); // e.g. "Network1" for the FIRST running instance, etc
  MyCwNet.cfg.iFunctionality = UTL_ReadIntFromIniFile(pszIniFileName, szSec, "Functionality",  MyCwNet.cfg.iFunctionality );
  MyCwNet.cfg.iDiagnosticFlags= UTL_ReadIntFromIniFile(pszIniFileName,szSec,"DiagnosticFlags", 0 );
  sprintf( szSec, "Network%dClient", 1+UTL_iAppInstance ); // e.g. "Network1Client" for the FIRST running instance, etc
  UTL_ReadStringFromIniFile( pszIniFileName, szSec, "RemoteServerIP",
                             MyCwNet.cfg.sz80ClientRemoteIP/*Default*/,
                             MyCwNet.cfg.sz80ClientRemoteIP, 80 );
  UTL_ReadStringFromIniFile( pszIniFileName, szSec, "UserName",
                             MyCwNet.cfg.sz80ClientUserName/*Default*/,
                             MyCwNet.cfg.sz80ClientUserName, 80 );
  UTL_ReadStringFromIniFile( pszIniFileName, szSec, "Callsign",
                             MyCwNet.cfg.sz80ClientCallsign/*Default*/,
                             MyCwNet.cfg.sz80ClientCallsign, 80 );

  sprintf( szSec, "Network%dServer", 1+UTL_iAppInstance ); // e.g. "Network1Server" for the FIRST running instance, etc
  MyCwNet.cfg.iServerListeningPort = UTL_ReadIntFromIniFile(pszIniFileName,szSec,
                             "ListeningPort", MyCwNet.cfg.iServerListeningPort );
  UTL_ReadStringFromIniFile( pszIniFileName, szSec, "AcceptUsers",
                             MyCwNet.cfg.sz255AcceptUsers/*Default*/,
                             MyCwNet.cfg.sz255AcceptUsers, 255 );
  UTL_ReadStringFromIniFile( pszIniFileName, szSec, "Pwd",
                             MyCwNet.cfg.sz20ServerAdminPWD/*Default*/,
                             MyCwNet.cfg.sz20ServerAdminPWD, 20 );
  MyCwNet.cfg.iHttpServerOptions = UTL_ReadIntFromIniFile(pszIniFileName,szSec,
                             "HttpOptions", MyCwNet.cfg.iHttpServerOptions );
  MyCwNet.cfg.iNetworkLatency_ms = UTL_ReadIntFromIniFile(pszIniFileName,szSec,
                             "Latency_ms",  MyCwNet.cfg.iNetworkLatency_ms );

  sprintf( szSec, "Keyer%dMainWindow", 1+UTL_iAppInstance ); // e.g. "Keyer1MainWindow" for the FIRST running instance, etc
  CwKeyer_Config.iMainWindowLeft = UTL_ReadIntFromIniFile(pszIniFileName,
       szSec, "Left"/*pszKeyName*/, 50 + 100 * UTL_iAppInstance/*default*/ );
  CwKeyer_Config.iMainWindowTop = UTL_ReadIntFromIniFile(pszIniFileName,
       szSec, "Top"/*pszKeyName*/,  100 + 30 * UTL_iAppInstance/*default*/ );
  CwKeyer_Config.iMainWindowWidth = UTL_ReadIntFromIniFile(pszIniFileName, szSec, "Width", 600 );
  CwKeyer_Config.iMainWindowHeight= UTL_ReadIntFromIniFile(pszIniFileName, szSec, "Height",500 );
  g_SpecDispControl.iColourScheme = UTL_ReadIntFromIniFile(pszIniFileName, szSec, "ColourScheme", COLOUR_SCHEME_DARK );
  g_iNewLanguage       = UTL_ReadIntFromIniFile(pszIniFileName, szSec, "Language", APPL_iLanguage );
  g_iUserInterfaceMode = UTL_ReadIntFromIniFile(pszIniFileName, szSec, "UserInterfaceMode", g_iUserInterfaceMode );
  RigCtrl_ITU_Region   = UTL_ReadIntFromIniFile(pszIniFileName, szSec, "ITU_Region", RigCtrl_ITU_Region );
  sprintf( szSec, "Keyer%d", 1+UTL_iAppInstance ); // e.g. back to "Keyer1" for the FIRST running instance, etc
  for(i=0; i<KEYER_NUM_MESSAGE_MEMORIES; ++i)
   { sprintf( szKey, "Memory%d", (int)(i+1) );
     UTL_ReadStringFromIniFile( pszIniFileName,
        szSec, szKey, (char*)CwKeyer_psz6MemoryDefaults[i],
        CwKeyer_Config.szKeyerMemory[i]/*pszResult*/, KEYER_MAX_CHARS_PER_MEMORY );
   } // end for < all 'keyer message memories' >

} // end LoadSettings()

//---------------------------------------------------------------------------
void SaveSettings(void) // .. in an old-fashioned *.ini file, NOT the bloody registry !
{ // [in] CwKeyer_Config .
  const char *pszIniFileName = SWI_APP_EXE_NAME".ini";
  char szSec[84];  // "section" (in the ini file, that's the string in squared brackets)
  char szKey[84];  // "key name" (that's the string on the left side of the assignment operator)
  TEdit *pEdit;
  int   i;

  strcpy( szSec, "About" );
  UTL_WriteStringToIniFile(pszIniFileName,szSec,"1", "Configuration file for the 'Remote CW Keyer'.");
  UTL_WriteStringToIniFile(pszIniFileName,szSec,"2", "If you don't care about sending Morse code via PC,");
  UTL_WriteStringToIniFile(pszIniFileName,szSec,"3", "you can safely delete this file, wherever you found it.");
  UTL_WriteStringToIniFile(pszIniFileName,szSec,"4", " ");

  sprintf( szSec, "Keyer%d", 1+UTL_iAppInstance ); // e.g. "Keyer1" for the FIRST running instance, etc
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "ComPort_IN",  CwKeyer_Config.iComPortNumber_IN );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "RadioControlAndKeyingPort",CwKeyer_Config.iRadioKeyingAndControlPort );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "RadioControlProtocol", CwKeyer_Config.iRadioControlProtocol );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "RadioControlBaudrate", CwKeyer_Config.iRadioControlBaudrate );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "RadioCIVAddress", CwKeyer_Config.iRadioCIVAddress );
  UTL_WriteStringToIniFile(pszIniFileName,szSec,"RemoteRigCtrlServerAddress", CwKeyer_Config.sz80RemoteRigCtrlServerAddress);
#if( SWI_USE_HAMLIB_SERVER ) // not sure if this 'feature' is going to stay...
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "EnableBuiltInHamlibServer",CwKeyer_Config.fEnableBuiltInHamlibServer);
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "BuiltInHamlibServerPort", MyHamlibServer.cfg.iServerListeningPort);
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "BuiltInHamlibServerOptions", MyHamlibServer.cfg.iServerOptions);
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "BuiltInHamlibServerDiagnostics", MyHamlibServer.cfg.iDiagnosticFlags);
#endif // SWI_USE_HAMLIB_SERVER ?

  UTL_WriteIntToIniFile(pszIniFileName, szSec, "MorseKeyType",    CwKeyer_Config.iMorseKeyType );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "DotInputSignal",  CwKeyer_Config.iDotInput );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "DashInputSignal", CwKeyer_Config.iDashInput );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "ManualPTTInput",  CwKeyer_Config.iManualPTTInput);
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "TestInputSignal", CwKeyer_Config.iTestInput );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "KeySupplySignal", CwKeyer_Config.iKeySupply );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "SidetoneOnTXD",   CwKeyer_Config.iSidetoneOnTXD );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "CwKeyingSignal",  CwKeyer_Config.iRadioCWKeying );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "PTTControlSignal",CwKeyer_Config.iRadioPTTControl);
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "TxDelayTime_ms",  CwKeyer_Config.iTxDelayTime_ms);
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "TxHangTime_ms",   CwKeyer_Config.iTxHangTime_ms);
  UTL_WriteIntToIniFile(pszIniFileName,szSec,"DebouncePaddleInputs",CwKeyer_Config.fDebouncePaddleInputs);
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "KeyInputWatchdog",CwKeyer_Config.fKeyInputWatchdog);
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "CwViaShiftAndControlKey", CwKeyer_Config.fKeyViaShiftAndControl);
  UTL_WriteStringToIniFile(pszIniFileName,szSec,"AudioInputDevice", CwKeyer_DSP.cfg.sz255AudioInputDevice);
  UTL_WriteStringToIniFile(pszIniFileName,szSec,"AudioOutputDevice",CwKeyer_DSP.cfg.sz255AudioOutputDevice);
  UTL_WriteIntToIniFile(pszIniFileName, szSec,  "AudioSidetone_Hz", CwKeyer_DSP.cfg.iSidetoneFreq_Hz );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "SidetoneRiseTime_ms",CwKeyer_DSP.cfg.iSidetoneRiseTime_ms );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "SidetoneGain_dB",CwKeyer_DSP.cfg.iSidetoneGain_dB );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "AudioInGain_dB", CwKeyer_DSP.cfg.iAudioInGain_dB );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "AudioOutGain_dB",CwKeyer_DSP.cfg.iAudioOutGain_dB );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "NetworkTonesGain_dB",CwKeyer_DSP.cfg.iNetworkTonesGain_dB );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "AudioFlags",      CwKeyer_DSP.cfg.iAudioFlags );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "WordsPerMinute",Elbug_DotTimeInMillisecondsToWordsPerMinute(CwKeyer_Elbug.cfg.iDotTime_ms) );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "LaunchOnStart", APPL_fLaunchKeyerOnStart );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "Channel4Source",CwKeyer_TimingScope.cfg.iChannel4Source );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "TScopeChannels",CwKeyer_TimingScope.cfg.iVisibleChannels);
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "TScopeTrigger", CwKeyer_TimingScope.cfg.iTriggerOptions );

  // Still in the ini-file-section "KeyerN", where "N" is the instance index...
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "TRXOptions", CwKeyer_Config.iTRXOptions );

  // Save options for the built-in "Rig Control" module :
  sprintf( szSec, "RigControl%d", 1+UTL_iAppInstance ); // e.g. "RigControl1" for the FIRST running instance, etc
  UTL_WriteIntToIniFile(pszIniFileName,szSec,"RejectMessagesInLog",CwKeyer_Config.dwRejectMessages4Log ); // ex: MyCwNet.RigControl.dwRejectMessages4Log);
  UTL_WriteIntToIniFile(pszIniFileName,szSec,"TrafficMonitorOptions",RigCtrl_iTrafficMonitorDisplayOptions);

  // Save the "Network" configuration (from module CwNet.c; simple TCP-client or server):
  sprintf( szSec, "Network%d", 1+UTL_iAppInstance ); // e.g. "Network1" for the FIRST running instance, etc
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "Functionality",  MyCwNet.cfg.iFunctionality );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "DiagnosticFlags",MyCwNet.cfg.iDiagnosticFlags );
  sprintf( szSec, "Network%dClient", 1+UTL_iAppInstance );
  UTL_WriteStringToIniFile( pszIniFileName, szSec, "RemoteServerIP",MyCwNet.cfg.sz80ClientRemoteIP );
  UTL_WriteStringToIniFile( pszIniFileName, szSec, "UserName", MyCwNet.cfg.sz80ClientUserName );
  UTL_WriteStringToIniFile( pszIniFileName, szSec, "Callsign", MyCwNet.cfg.sz80ClientCallsign );

  sprintf( szSec, "Network%dServer", 1+UTL_iAppInstance );
  UTL_WriteIntToIniFile(    pszIniFileName, szSec, "ListeningPort", MyCwNet.cfg.iServerListeningPort );
  UTL_WriteStringToIniFile( pszIniFileName, szSec, "AcceptUsers",MyCwNet.cfg.sz255AcceptUsers );
  UTL_WriteStringToIniFile( pszIniFileName, szSec, "Pwd",MyCwNet.cfg.sz20ServerAdminPWD );
  UTL_WriteIntToIniFile(    pszIniFileName, szSec, "HttpOptions",MyCwNet.cfg.iHttpServerOptions );
  UTL_WriteIntToIniFile(    pszIniFileName, szSec, "Latency_ms", MyCwNet.cfg.iNetworkLatency_ms );

  sprintf( szSec, "Keyer%d", 1+UTL_iAppInstance ); // e.g. back to "Keyer1" for the FIRST running instance, etc
  for(i=0; i<KEYER_NUM_MESSAGE_MEMORIES; ++i)
   { sprintf( szKey, "Memory%d", (int)(i+1) );
     UTL_WriteStringToIniFile( pszIniFileName, szSec, szKey, CwKeyer_Config.szKeyerMemory[i] );
   } // end for < all 'keyer message memories' >
  sprintf( szSec, "Keyer%dMainWindow", 1+UTL_iAppInstance ); // e.g. "Keyer1MainWindow" for the FIRST running instance, etc
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "Left",  CwKeyer_Config.iMainWindowLeft );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "Top",   CwKeyer_Config.iMainWindowTop );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "Width", CwKeyer_Config.iMainWindowWidth );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "Height",CwKeyer_Config.iMainWindowHeight );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "ColourScheme", g_SpecDispControl.iColourScheme );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "Language", APPL_iLanguage );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "UserInterfaceMode", g_iUserInterfaceMode );
  UTL_WriteIntToIniFile(pszIniFileName, szSec, "ITU_Region", RigCtrl_ITU_Region );

  APPL_fSaveSettingsOnExit = FALSE;  // "done" (flag cleared in SaveSettings())


} // end SaveSettings()

//---------------------------------------------------------------------------
DWORD CalcHashForSettings(void) // .. to tell if we need to SAVE SETTINGS on exit
{
#ifndef  sizeof_member
# define sizeof_member(type,member) sizeof(((type*)0)->member) /* ugly.. but works */
#endif

  // Actually, the hash function is the 32-bit CRC implemented in Utilities.c :
  DWORD dwHash= UTL_CRC32( 0/*seed*/, (BYTE*)&CwKeyer_Config, sizeof(CwKeyer_Config) );
  dwHash = UTL_CRC32( dwHash/*seed*/, (BYTE*)&CwKeyer_Elbug.cfg, sizeof_member(T_ElbugInstance,cfg) );
  dwHash = UTL_CRC32( dwHash/*seed*/, (BYTE*)&CwKeyer_DSP.cfg, sizeof_member(T_CwDSP,cfg) );
  dwHash = UTL_CRC32( dwHash/*seed*/, (BYTE*)&MyCwNet.cfg, sizeof_member(T_CwNet,cfg) );
  dwHash = UTL_CRC32( dwHash/*seed*/, (BYTE*)&CwKeyer_TimingScope.cfg, sizeof_member(T_KeyerTimingScope,cfg) );
#if( SWI_USE_HAMLIB_SERVER ) // not sure if the 'Hamlib Net rigctld'-emulation is going to stay...
  dwHash = UTL_CRC32( dwHash/*seed*/, (BYTE*)&MyHamlibServer.cfg, sizeof_member(T_HLSrv,cfg) );
#endif // SWI_USE_HAMLIB_SERVER ?

  return dwHash;

} // end CalcHashForSettings()


//---------------------------------------------------------------------------
extern "C" void ShowError( int iErrorClass, const char * pszFormat, ... )
  // ,---------------------------'
  // '--> ERROR_CLASS_FATAL, ERROR_CLASS_ERROR, ERROR_CLASS_WARNING, ERROR_CLASS_INFO .
  //      If an optional 'debug run log' (file) is supported,
  //      iErrorClass bitmask SHOW_ERROR_IN_RUN_LOG can be used to emit
  //      the same message in "debug_run_log.txt", besides the GUI's "Debug" tab.
  //
{
  va_list parameter;
  T_ErrorFifoEntry *pErrorFifoEntry;
  char *pszDest, *pszEndstop;
  double dblUnixDateTime;

  va_start( parameter, pszFormat );
  // Because ShowError() *may* be called from a worker thread, don't "print"
  // directly into the non-thread-safe "RichEdit" control ! Instead, use a
  // classic lock-free circular FIFO between ShowError() and the GUI itself:
  pErrorFifoEntry = &ErrorHistoryFifo[ DEBUG_iErrorHistoryHead ];
  pszDest    = pErrorFifoEntry->sz511Text;
  pszEndstop = pErrorFifoEntry->sz511Text + 510/*!*/;
  if( iErrorClass & SHOW_ERROR_TIMESTAMP )
   { dblUnixDateTime = UTL_GetCurrentUnixDateAndTime();
     UTL_FormatDateAndTime( "hh:mm:ss.s ", dblUnixDateTime, pszDest );
     SL_SkipToEndOfString( (const char**)&pszDest );
   }
  vsnprintf( pszDest, (pszEndstop-pszDest)/*maxlen*/, pszFormat, parameter );
  pErrorFifoEntry->iErrorClass = iErrorClass & (~SHOW_ERROR_IN_RUN_LOG);
  va_end(parameter);
  // Make the new entry available for the "Debug" window by incrementing the FIFO HEAD index:
  DEBUG_iErrorHistoryHead = (DEBUG_iErrorHistoryHead + 1) % C_ERROR_FIFO_SIZE;

  if( iErrorClass & SHOW_ERROR_IN_RUN_LOG )
   { UTL_WriteRunLogEntry( "%s", pErrorFifoEntry->sz511Text );
     // '--> "Does nothing" if UTL_fRunLogOpened==FALSE (no need to check HERE)
   }

  // ex: SetStatusWindowText(MainWnd_hStatusWindow, STATUS_PART_ERR_MSG, 0, szText );
} // end ShowError()


//---------------------------------------------------------------------------
BOOL FillComboWithSerialPorts( TComboBox *pCombo,
        int iSelectedComPortNr ) // [in] "COM"-port number (1..N) or COMBO_BOX_COM_PORT_NONE
   // Returns TRUE if iSelectedComPortNr is currently valid,
   //         FALSE otherwise (i.e. not found in the devices "enumerated" by the OS)
{
  int  nEntriesFound = 0;
  BOOL fFoundSelectedPort = FALSE;
  int iEnumerator = 0;  // initialize 'enumerator'
  char sz80[84];
  static BOOL firstBug = TRUE;

  // On this occasion, allow the "TComboBox" to get WIDER when DROPPED DOWN :
  SendMessage( pCombo->Handle,CB_SETDROPPEDWIDTH, 280/*MinimumWidthInPixels*/, 0);
  // ,---------------------------------------------'
  // '--> This was even long enough for "COM3 (Silicon Labs CP210x USB to UART Bridge)",
  //      which was the funny "friendly" COM port name of an Icom IC-9700 !

  pCombo->Clear();
  pCombo->ItemIndex = -1; // nothing selected yet

  // First entry : "NONE", which is C_CFG_PORT_NONE = 0 in C:\cbproj\SpecLab\config.h
  pCombo->Items->Add( "NONE" );
  if( iSelectedComPortNr < 0 )
   {
     pCombo->ItemIndex = nEntriesFound; // index ZERO for the FIRST item in the list, etc
     fFoundSelectedPort = TRUE;
   }
  ++nEntriesFound;

  // Let Utilities.c : UTL_EnumerateComPorts() enumerate the COM ports.
  //     If possible (and Windows permitting), 'decorated' with parts of
  //     the 'friendly' name, but AFTER the significant part, in parentheses,
  //     e.g. "COM11 (USB Serial Port)" instead of just "COM11" :
  iEnumerator = 0;  // initialize the 'enumerator' for UTL_EnumerateComPorts()
  UTL_sz255LastError[0] = '\0';
  while( UTL_EnumerateComPorts( &iEnumerator, sz80, 80 ) )
   { // Add to the list of 'COM' ports, and possibly SELECT the item :
     pCombo->Items->Add( sz80 );  // hopefully something like "COM1", "COM2", ..
     // .. and if the "The Mimosa", SetupDiEnumDeviceInfo(), feels like cooperating,
     //    the list of COM ports even contains the 'friendly name' (every now and then).
     if( iSelectedComPortNr == UTL_ParseNumberFromComPortName(sz80) )
      { pCombo->ItemIndex = nEntriesFound; // index ZERO for the FIRST item in the list, etc
        fFoundSelectedPort = TRUE;
      }
     ++nEntriesFound;
   } // end while( UTL_EnumerateComPorts( &iEnumerator, sz80, 80 ) )
  if( (firstBug) && (UTL_sz255LastError[0] != '\0') ) // trouble extracting the "friendly names" again ?
   { ShowError( ERROR_CLASS_WARNING, "%s", UTL_sz255LastError );
     firstBug = FALSE;
     // 2024-01-05 : After being absent for several weeks, the "Bug" returned,
     // and UTL_EnumerateComPorts() diagnosed :
     // > EnumerateComPorts: SetupDiEnumDeviceInfo() failed at device #1 !
   }

  // If the configured COM port is NOT listed in the above loop
  //   (because stupid windows assigned a new COM port number to the device,
  //    or for some reason the COM port driver refuses to enumerate,
  //    as typically happens with COM ports from "VSPE"),
  // add a DUMMY ENTRY at the end of the list, so the pitiful windoze user
  //   can at least see that THERE IS SOMETHING WRONG WITH THE "COM PORT" .
  if( (!fFoundSelectedPort) && (iSelectedComPortNr>0) )
   {
     sprintf( sz80, "COM%d *NOT ENUMERATED*", iSelectedComPortNr );
     pCombo->Items->Add( sz80 );  // hopefully something like "COM1", "COM2", ..
     pCombo->ItemIndex = nEntriesFound; // index ZERO for the FIRST item in the list, etc
   } // end if( ! fFoundSelectedPort ) && ...

  return fFoundSelectedPort;  // -> TRUE=ok, FALSE=something wrong with iSelectedComPortNr
} // end FillComboWithSerialPorts()


//---------------------------------------------------------------------------
BOOL FillComboWithItemsFromTokenLists( TComboBox *pCombo,
        const T_SL_TokenList *pTokens1, // [in] e.g. C_TL_None_Zero (with string = "None", value = zero)
        const T_SL_TokenList *pTokens2, // [in] e.g. SerialPortInputSignals_Key
        const T_SL_TokenList *pTokens3, // [in] e.g. SerialPortInputSignals_Radio, or NULL for only *one* list
        int iSelectedTokenValue  )      // [in] e.g. [-]KEYER_SIGNAL_INDEX_MORSE_KEY_CTS
{
  int   iList;
  int   nEntries = 0;
  const T_SL_TokenList *pTokens;
  pCombo->Clear();
  pCombo->ItemIndex = -1;
  for( iList=1; iList<=3; ++iList)
   { switch(iList)
      { case 1: pTokens = pTokens1; break;
        case 2: pTokens = pTokens2; break;
        case 3: pTokens = pTokens3; break;
        default: pTokens = NULL; break;
      }
     if( pTokens != NULL ) // emit the strings from another list ?
      { while( pTokens->pszKeyword != NULL ) // not the end of the list yet ?
         {
           pCombo->Items->Add( pTokens->pszKeyword );  // hopefully something like "COM1", "COM2", ..
           if( iSelectedTokenValue == pTokens->iTokenValue )
            { pCombo->ItemIndex = nEntries; // index ZERO for the FIRST item in the list, etc
            }
           ++nEntries;
           ++pTokens;
         }
      } // if( pTokens != NULL )
   }
  if( pCombo->ItemIndex >= 0)
   { return TRUE;
   }
  pCombo->ItemIndex = 0; // instead of selecting garbage, select the first ...
  return FALSE;          // ... but give the caller the chance to mark this item in RED
} // end FillComboWithItemsFromTokenLists()


//---------------------------------------------------------------------------
BOOL FillComboWithBandList( TComboBox *pCombo, // [out] combo box with bands or even band-stacking registers
        DWORD dwSupportedBands, // [in] bitwise combination of RIGCTL_BAND_...
        DWORD dwSelectedBand )  // [in] a single bit representing the SELECTED band,
                                //      e.g. RIGCTL_BAND_40M .
{
  int   iBandIndex;
  DWORD dwBitmask, dwStackedBand;
  const char *pszBandName;
  char  sz80ItemText[84];
  char  *pszDest, *cpDestEndstop;
  BOOL  fFoundSelectedBand = FALSE;
  int   nEntriesFound = 0;
  int   nStackingEntriesInCurrentBand, iBandStackingRegIndex, iLen, iWantedPos;
  T_RigCtrlInstance *pRC = &MyCwNet.RigControl;
  T_RigCtrlFreqMemEntry *pBandStackingReg;

  // On this occasion, allow the "TComboBox" to get WIDER when DROPPED DOWN :
  SendMessage( pCombo->Handle,CB_SETDROPPEDWIDTH, 440/*MinimumWidthInPixels*/, 0);
  // ,---------------------------------------------'
  // '--> Should be wide enough for the BAND NAME (like "40 m")
  //      AND three "Band Stacking Registers" that modern Icoms radios like
  //      the IC-7300 has stored automatically, with the "most recent frequencies"
  //      that the operator has visited.
  //      Alternatively (without band-stacking registers), the combo list
  //      may show frequency limits or presets DEFINED BY THE SYSOP (future plan).
  //      With the combo's font set to "Courier New", size EIGHT,
  //      each character had a fixed width of NINE(!) pixels .. but ymmv .
  //  Icom's own 'Band Stacking Register' display on an IC-7300 looked
  //  like this (and this is what we're going to mimick with the TComboBox):
  //      " 50.085 CW    50.095 CW    50.135 USB"
  //      |<-12 chars->|<-12 chars->|<-12 chars->|
  //   In addition, for the NON-DROPPED-DOWN combo list,
  //   the begin of each line shall indicate the BAND NAME, e.g. "80 m: " (6 chars).
  //    ( 6 + 3 * 12 ) characters * 9 pixels_per_character = 396+x, say 440 pixels.

  pCombo->Clear();
  pCombo->ItemIndex = -1; // nothing selected yet

  // To know which rows and columns have been filled with which
  // 'band stacking registers' later in CB_BandClick(),
  // use an extra array :
  memset( KeyerGUI_iBandStackingRegIndices, -1, sizeof(KeyerGUI_iBandStackingRegIndices) );

  for( iBandIndex = 0; iBandIndex<=31; ++iBandIndex ) // try all 32 possible "band-bit" indices
   { dwBitmask = 1UL << iBandIndex; // -> RIGCTRL_BAND_160M (bit 0), etc etc ..
     pszBandName = SL_GetStringFromTokenList( RigCtrl_BandNames, dwBitmask );
     if( (pszBandName!=NULL) && ((dwBitmask & dwSupportedBands)!=0) )
      { // Regardless of what's in pRC->BandStackingRegs[ 0..RIGCTRL_NUM_BAND_STACKING_REGS-1 ],
        // will get here with pszBandName = "160 m", "80 m", "60 m", "40 m" .. .
        // The BAND NAME will always be added to the combo, even if there are
        // no band-stacking-register entries for this band *anywhere*
        // (This is important, *anywhere*, because for example an IC-7300
        //  modified for 'all band transmit' or a few extra bands like 60 m or 4 m
        //  will NOT add extra 'band codes' to the band stacking registers.
        //  Instead, the 'added bands' like 60 m, 4 m, etc will all share a
        //  band that Icom labelled 'GENE' in the IC-7300's band stacking
        //  register display.)
        // EXAMPLE: Content of pRC->BandStackingRegs[], as displayed via
        //   'Settings'..'Report Rig Control parameters on the Debug tab' :
        //      BandStack[00..02]:  1.915  CW,   1.910  CW,   1.915  CW
        //      BandStack[03..05]:  3.573  CW,   3.793 LSB,   3.557  CW
        //      BandStack[06..08]:  7.030  CW,   7.076 LSB,   7.201 LSB
        //      BandStack[09..11]: 10.120  CW,  10.130  CW,  10.140  CW
        //      BandStack[12..14]: 14.244 USB,  14.100  CW,  14.054  CW
        //      BandStack[15..17]: 18.150  CW,  18.150 USB,  18.070  CW
        //      BandStack[18..20]: 21.024  CW,  21.300 USB,  21.050  CW
        //      BandStack[21..23]: 24.898  CW,  24.980 USB,  24.900  CW
        //      BandStack[24..26]: 28.036  CW,  28.565 USB,  29.269  FM
        //      BandStack[27..29]: 50.085  CW,  50.097  CW,  50.095  CW
        //      BandStack[30..32]:  5.353  CW,   0.472  CW,  70.107  CW
        //                          /|\           /|\          /|\   ..
        //      "GENE" band entries--'-------------'------------'
        //
        //    DON'T ASSUME ANYTHING about which frequencies are stored
        //    in which part of the array ... things will be COMPLETELY DIFFERENT
        //    in other transceivers; e.g. IC-9700 and especially IC-705 !
        pszDest = sz80ItemText;
        cpDestEndstop = sz80ItemText + sizeof(sz80ItemText) - 1;
        SL_AppendString( &pszDest, cpDestEndstop, pszBandName );
        // Check all 'Band Stacking Registers' if they are in THIS band (dwBitmask) .
        // If a band stacking register belongs to this band, append it to
        // the string (limited to THREE stack entries per band, as in Icom radios)
        nStackingEntriesInCurrentBand = 0; // -> 0..2
        for( iBandStackingRegIndex=0;
             iBandStackingRegIndex < pRC->iNumBandStackingRegs;
             iBandStackingRegIndex++)
         { pBandStackingReg = &pRC->BandStackingRegs[ iBandStackingRegIndex ];
           dwStackedBand = RigCtrl_FrequencyToBand( pBandStackingReg->RxTx[0].dblOperatingFreq_Hz );
                // '--> RIGCTL_BAND_xyz (a BITMASK, not an index)
           // Only add this entry from pRC->BandStackingRegs[] if it belongs
           // to the current BAND (in a single line of the combo box):
           if( ( (dwStackedBand & dwBitmask) != 0 ) && (nStackingEntriesInCurrentBand<3) )
            { if( nStackingEntriesInCurrentBand == 0 )
               { SL_AppendString( &pszDest, cpDestEndstop, ":" );
               }
              iLen = strlen( sz80ItemText );
              iWantedPos = BANDLIST_BAND_COLUM_WIDTH + BANDLIST_FREQ_COLUMN_WIDTH * nStackingEntriesInCurrentBand;
              if( iLen < iWantedPos )
               { SL_PadSpaces( &pszDest, cpDestEndstop, iWantedPos - iLen );
               }
              // Like Icom in their "BAND STACKING REGISTER" display,
              // show the frequency in MEGAHERTZ (but without a unit),
              // with leading space for a fixed width, and only
              // three decimal places. Umm.. what was the printf format string ?
#            if( 1 ) // format string to have everything vertically aligned:
              SL_AppendPrintf( &pszDest, cpDestEndstop, "%9.4lf %3s",
                 // "width" (or something in that style)--' |
                 // "precision" (or whatever it was..)------'
#            else // format string to have the leftmost digits vertically aligned,
                  // but at least one space as separator (save even for e.g. "1296.200 USB")
              SL_AppendPrintf( &pszDest, cpDestEndstop, " %.3lf",
#            endif
                 (double)(pBandStackingReg->RxTx[0].dblOperatingFreq_Hz * 1e-6),
                  RigCtrl_OperatingModeToString( pBandStackingReg->RxTx[0].iOpMode
                           & RIGCTRL_OPMODE_MASK_TO_STRIP_FLAGS ) );
              KeyerGUI_iBandStackingRegIndices[iBandIndex][nStackingEntriesInCurrentBand]=iBandStackingRegIndex;
              //                        '--> 0..31  '--> 0..2                       |
              //                   ("band index",     "stack levels per band")      |
              //          not the COMBO ITEM INDEX !)                               |
              //    ,---------------------------------------------------------------'
              //    '--> Index into pRC->BandStackingRegs[ iBandStackingRegIndex ]
              // EXAMPLE with BandStackingRegs[3] = "3.5730 CW", in the SECOND LINE
              //         of the combo box:
              //    nEntriesFound = 1 (here used as the zero-based combo item index),
              //    nStackingEntriesInCurrentBand = 0 (i.e. the MOST RECENT entry, aka "top of stack")
              //    dwStackedBand = 2 = (1<<1) = RIGCTRL_BAND_80M
              //    iBandStackingRegIndex = 3
              //  When selecting that entry ("3.5730 CW") in TKeyerMainForm::CB_BandClick() :
              //    (1) from the BEGIN of the item text (sz80ItemText : "80 m: .."),
              //        i32NewBand will be parsed -> result = 2 = (1<<1) = RIGCTRL_BAND_80M .
              //        From that BITMASK, the new "band index" (a zero based array index)
              //        will be: iBandIndex = BitmaskToBitIndex( i32NewBand );
              //    (2) from the HORIZONTAL MOUSE POINTER POSITION,
              //        TKeyerMainForm::CB_BandClick() determines iBandStackIndex .
              //    (3) iBandIndex and iBandStackIndex are the indices into
              //         KeyerGUI_iBandStackingRegIndices[iBandIndex][iBandStackIndex].
              //        That two-dimensional array delivers the array index
              //        (iBandStackingRegIndex) into pRC->BandStackingRegs[iBandStackingRegIndex].
              //    (4) That band-stacking-register should contain everything
              //        module RigControl.c needs to know (frequency, mode, AND MANY MORE)
              //        to switch back to the "old" (stacked) frequency.
              ++nStackingEntriesInCurrentBand;
            } // end if < matching frequency for the current "band" >
         }   // end for( iBandStackingRegIndex .. )

        pCombo->Items->Add( sz80ItemText );
        if( dwBitmask & dwSelectedBand )
         { pCombo->ItemIndex = nEntriesFound; // index ZERO for the FIRST item in the list, etc
           fFoundSelectedBand = TRUE;
         }
        ++nEntriesFound;
      }
   }
  return fFoundSelectedBand;  // -> TRUE=ok, FALSE=something wrong
} // end FillComboWithBandList()


//---------------------------------------------------------------------------
int GetSelectedTokenValueFromComboBox( TComboBox *pCombo,
        const T_SL_TokenList *pTokens1, // [in] e.g. C_TL_None_Zero (with string = "None", value = zero)
        const T_SL_TokenList *pTokens2, // [in] e.g. SerialPortInputSignals_Key
        const T_SL_TokenList *pTokens3) // [in] e.g. SerialPortInputSignals_Radio, or NULL for only *one* list
  // If the selected text in the combo list doesn't match ANY of the
  // strings in the three token lists, this function returns
  // GUI_ERROR_NO_MATCHING_ENTRY to let THE CALLER decide what to use.
# define GUI_ERROR_NO_MATCHING_ENTRY -99999
{
  int   iList;
  const T_SL_TokenList *pTokens;

  for( iList=1; iList<=3; ++iList)
   { switch(iList)
      { case 1: pTokens = pTokens1; break;
        case 2: pTokens = pTokens2; break;
        case 3: pTokens = pTokens3; break;
        default: pTokens = NULL; break;
      }
     if( pTokens != NULL ) // check the strings from another list ?
      { while( pTokens->pszKeyword != NULL ) // not the end of the list yet ?
         {
           if( pCombo->Text == AnsiString( pTokens->pszKeyword ) )
            { return pTokens->iTokenValue;
            }
           ++pTokens;
         }
      }
   }
  return GUI_ERROR_NO_MATCHING_ENTRY;  // No valid selection ? -> let THE CALLER handle this !
} // end GetSelectedTokenValueFromComboBox()

//---------------------------------------------------------------------------
BOOL SelectComboItemByDecimalValue( TComboBox *pCombo, int iValue )
{ int i, n, iSelItem=-1;
  n = pCombo->Items->Count;
  for(i=0; i<n; ++i)
   { if( StrToIntDef( pCombo->Items->Strings[i], 0) == iValue )
      { iSelItem = i;
        break;
      }
   }
  if( iSelItem>=0 )
   { pCombo->ItemIndex = iSelItem;
     return TRUE;
   }
  else
   { return FALSE;
   }
}

//---------------------------------------------------------------------------
BOOL SelectComboItemByHexadecimalValue( TComboBox *pCombo, int iValue )
{ int i, n, iSelItem=-1;
  char sz40Text[44];
  const char *pszSrc;
  AnsiString s;  // in BCB V12, only an "AnsiString" returns a char * from c_str() !
                 // in BCV V6, there was no UnicodeString at all ..
  n = pCombo->Items->Count;
  for(i=0; i<n; ++i)
   { s = pCombo->Items->Strings[i];
     SL_strncpy( sz40Text, s.c_str(), 40 );
     // -> BCB V12 "Athens" : [bcc32c Error] no matching function for call to 'SL_strncpy'
     // >  StringLib.h(194): candidate function not viable:
     // > no known conversion from 'System::WideChar *' (aka 'wchar_t *')
     // > to 'const char *' for 2nd argument
     pszSrc = sz40Text;
     if( SL_SkipString( (const char**)&pszSrc, "0x" ) ) // looks like a HEX number ...
      { if( (int)SL_ParseHex( &pszSrc, 4/*nMaxDigits*/) == iValue )
         { iSelItem = i;
           break;
         }
      }
   }
  if( iSelItem>=0 )
   { pCombo->ItemIndex = iSelItem;
     return TRUE;
   }
  else
   { pCombo->ItemIndex = -1;
     sprintf( sz40Text, "0x%02X", (unsigned int) iValue );
     pCombo->Text = sz40Text; // works with TComboBox.Style=csDropDown, not with csDropDownList !
     return FALSE;
   }
}

//---------------------------------------------------------------------------
int  GetSerialPortSignalIndexFromComboBox( TComboBox *pCombo )
  // Return value : KEYER_SIGNAL_INDEX_... , negative when inverted, 0 = "none"
{
  int iSelectedTokenValue = GetSelectedTokenValueFromComboBox( pCombo,
        C_TL_None_Zero,                // [in] const T_SL_TokenList *pTokens1
        SerialPortInputSignals_Key,    // [in] const T_SL_TokenList *pTokens2
        SerialPortInputSignals_Radio); // [in] const T_SL_TokenList *pTokens3
  if( iSelectedTokenValue == GUI_ERROR_NO_MATCHING_ENTRY )
   { iSelectedTokenValue = GetSelectedTokenValueFromComboBox( pCombo,
        SerialPortOutputSignals_Key,   // [in] const T_SL_TokenList *pTokens1
        SerialPortOutputSignals_Radio, // [in] const T_SL_TokenList *pTokens2
        NULL );                        // [in] const T_SL_TokenList *pTokens3
   }
  return iSelectedTokenValue;
}

//---------------------------------------------------------------------------
int ParseSerialPortNumberFromText( const char *pszSource )
   // Returns a non-negative serial port number when successful,
   //         otherwise -1 (also for the entry named "NONE", in combo boxes)
{
  int i;
  if( strncmp( pszSource, "NONE", 4) == 0 )
   { return -1;
   }
  if( pszSource[0] != '\0' )
   { // Does the stupid sscanf() parse the '9' in "COM9 (invalid)" ? [2015-01: yes]
     i =  UTL_ParseNumberFromComPortName( pszSource );
     if( i >= 0 )
      { return i;
      }
   }
  return -1;  // invalid COM port number !

} // end ParseSerialPortNumberFromText()


//---------------------------------------------------------------------------
double GetFrequencyFromEditField( TEdit *pEd, // [in] edit field with frequency
                 double *pdblFreqIncrement ) // [out] 10^N, f(cursor_position)
  // Accepts a string from the VFO frequency input field (or similar fields)
  // in different formats, including 'technical' notations like 12.345 kHz
  //                                                         or 7.0374 MHz .
  // For incrementing / decrementing the frequency via cursor up/down keys
  //      or mouse wheel, the function may pass back the position of the
  //      text cursor (aka 'caret') with in the TEdit, and the value to add/
  //      subtract on each cursor up/down event. Examples

  //      char index:  0123456789ABCDEF
  //      pEd->Text = "   7.012 345 MHz"  (note the spaces, also "in between")
  //                            '------- blinking caret or underscore HERE
  //   -> [out] pdblFreqIncrement = 100 [Hz]
  //
  //
{
  char c, szA[44], *cp;
  int i, slen, nDigits, iExponent;
  int iExponentFromTechUnit=0, nDecimalPlaces=0;
  double dblFreqHz, dblUpperPart, dblLowerPartMult;
  BOOL   fNegative = FALSE;
  BOOL   fGotDecimalPlaces = FALSE;
  AnsiString s;

  // ex: SL_strcpy_trunc(szA, sizeof(szA), pEd->Text.c_str(), 39 );
  // When compiled with BCB V12 "Athens" :
  // > [bcc32c Error] no matching function for call to 'SL_strcpy_trunc'
  // > StringLib.h(196): candidate function not viable: no known conversion
  // > from 'System::WideChar *' (aka 'wchar_t *') to 'const char *' for 3rd argument
  s = pEd->Text;  // only an AnsiString's "c_str()" returns a "char *" !
  SL_strcpy_trunc(szA, sizeof(szA), s.c_str(), 39 );

  // Convert string to decimal, ignoring spaces:
  dblFreqHz = dblUpperPart = 0.0;
  dblLowerPartMult = 0.0;  // there's no "lower part" to be added later
  slen = 0;
  cp = szA;
  while( *cp == ' ' )
   { ++cp;
     ++slen;
   }
  if( *cp == '-' )
   { ++cp;
     ++slen;
     fNegative = TRUE;
   }
  while( *cp != '\0' )
   { if( ((c=*cp)>='0') && (c<='9') )
      { dblFreqHz = 10.0*dblFreqHz + (c-'0');
        dblLowerPartMult /= 10.0;
        if( fGotDecimalPlaces )
         { ++nDecimalPlaces;
         }
      }
     else  // a "technical" exponent like 'k', 'M', 'G' ?
      { if( c=='.' || c=='k' || c=='M' || c=='G' || c=='m' || c=='u'  )
         {  // Combine the previous 'upper' and 'lower' part into a single new 'upper' part,
            // and begin a new string->float conversion for the rest (fractional part):
            if( dblLowerPartMult > 0.0 )  // fraction from k(Hz), M(Hz), or G(Hz) ?
             { dblFreqHz *= dblLowerPartMult;
             }
            dblUpperPart += dblFreqHz; // the sum now completely in the "upper part" (incl. fraction)
            dblFreqHz    = 0;        // begin new conversion for the fractional part
            dblLowerPartMult = 1.0;  // factor for trailing digits (will be divided by ten with each new digit)
         }
        switch( c )
         {
           case '.':  // already parsed number was in HERTZ (or whatever), the rest is the fractional part:
              dblFreqHz         = 0;
              dblLowerPartMult  = 1.0;  // factor for trailing digits (will be divided by ten with each new digit)
              nDecimalPlaces    = 0;
              fGotDecimalPlaces = TRUE;
              break;
           case 'k':  // already parsed number was in KILOHERTZ, rest is a fraction:
              iExponentFromTechUnit = 3;  // required later for inc/dec
              dblUpperPart *= 1000.0;
              dblLowerPartMult = 1000.0;  // factor for trailing digits (will be divided by ten with each new digit)
              break;
           case 'M':  // already parsed number was in MEGAHERTZ, rest is a fraction:
              iExponentFromTechUnit = 6;  // required later for inc/dec
              dblUpperPart *= 1E6;
              dblLowerPartMult = 1E6;
              break;
           case 'G':  // already parsed number was in GIGAHERTZ, rest is a fraction:
              iExponentFromTechUnit = 9;  // required later for inc/dec
              dblUpperPart *= 1E9;
              dblLowerPartMult = 1E9;
              break;
           case 'm':  // already parsed number was in MILLIHERTZ, rest a tiny fraction:
              iExponentFromTechUnit = -3;
              dblUpperPart *= 1E-3;
              dblLowerPartMult = 1E-3;
              break;
           case 'u':  // already parsed number was in MICROHERTZ, rest a ridiculous fraction:
              iExponentFromTechUnit = -6;
              dblUpperPart *= 1E-6;
              dblLowerPartMult = 1E-6;
              break;
           default :  // e.g. c='-' if the frequency in Ed_VFO is invalid
              break;
         }
      }
     ++cp;
     ++slen;
   }
  // Examples:
  //  (a) "3M5" -> UpperPart=3000000, LowerPart=5, LowerPartMult=100000 .
  //  (b) "---.----- MHz" ->  fNegative=TRUE, UpperPart=0, dblFreqHz=0, LowerPartMult=0 .
  if( dblLowerPartMult > 0.0 )  // fraction from k(Hz), M(Hz), or G(Hz) ?
   { dblFreqHz *= dblLowerPartMult;
   }
  dblFreqHz += dblUpperPart;
  if( fNegative )
   { dblFreqHz = -dblFreqHz;
   }
  //
  // Convert the current cursor position into a power of ten for inc/dec-editing.
  // EXAMPLE: szA="123", slen=3, i=2 -> iFreqIncrement=1 .
  //          (in this case, there's NO digit to examine, because
  //           the RIGHTMOST digit is selected for inc/dec-editing)
  i = pEd->SelStart+1; // cursor position within the string; 0=first char
  // Because the cursor shall be RIGHT NEXT to the incremented/decremented digit,
  //    a cursor after the 'end' of the string is meaningless, i.e. :
  //    with sLen==1, the only valid cursor position (i) is ZERO, not ONE:
  if( i>slen )  i=slen;
  if( i<0 )     i=0;
  //                      Cursor (left of the digit to increment/decrement,
  //                       |  usually with "SelLength"=1 to highlight digit)
  //                       |
  //  Example:  szA = "   12 345 567"  slen=13, nDigits=8
  //                   |   |       |
  //                   |   |       |
  //               szA[0] i=4  szA[12]
  //  Note that there may be SPACES or other thousand-separators
  //  in the string, so COUNT THE DIGITS here to find the decimal place:
  nDigits = 0; // ... i.e. dblFreqIncrement = 10 ^ iCursorDigitPos
  while( i<slen )
   { if( szA[i]>='0' && szA[i]<='9' )  // only count the DIGITS, nothing else !
      { ++nDigits;
      }
     ++i;
   }
  iExponent = nDigits + iExponentFromTechUnit - nDecimalPlaces;
  if( pdblFreqIncrement != NULL ) // caller wants the power of ten (f(cursor-pos) )
   { *pdblFreqIncrement = pow10( iExponent ); // pow10 takes an INTEGER argument; HERE that's ok
   }

  return dblFreqHz;
} // end GetFrequencyFromEditField()


//---------------------------------------------------------------------------
void UpdateTimingScope( // Does what the name implies :) ..
              T_KeyerTimingScope *pScope,     // [in] timing scope data and configuration
              Graphics::TBitmap *pDestBitmap) // [out] graphic bitmap (unfortunately, a Borland-specific VCL thing)
      // [in]  TimingScopeOverlay[] : mouse-controlled 'overlay' to measure times
{
  int w = pDestBitmap->Width;  // width in pixels
  int h = pDestBitmap->Height; // height in pixels
  int iSample, x, y, h2, h3, yc, y_min, y_max, val, prev_x, prev_y, th, tw, tx, ty;
  int iSampleAtBeginOfChar;
  float flt;
  WORD  wCwChar;
  DWORD dwPenColour; // R-G-B colour mix (Red in bits 7..0, Green bits 15..8, Blue bits 23..16)
                     // Compatible with the VCL type "TColor". Possibly a 'Windows thing'.
  Graphics::TBitmap *osb;
  TPoint *pPoints;
  int     nPoints;
  // Test result:
  //   Plotting each curve in a single call of TCanvas.Polyline() : 2.5 % CPU load from the "Remote CW keyer"
  //   Plotting curves with MoveTo() / hundreds of LineTo() calls : 7.5 % CPU load  "    "     "     "  "
  //   Not plotting anything (only running the 'keyer thread')    : 1.3 % CPU load  "    "     "     "  "
  // -> Worth the effort (to reduce the CPU load) !
  const char *pszDecoded;
  AnsiString s;

  HERE_I_AM__GUI();

  pPoints = (TPoint*)malloc( TIMING_SCOPE_NUM_SAMPLE_POINTS * sizeof(POINT) );
  if(pPoints==NULL) // oops .. malloc() failed !
   { return;
   }

  osb = new Graphics::TBitmap;
  osb->Width = w;
  osb->Height= h;
  g_iTimingScopeBitmapWidth = w;  // <- only for diagnostic / trouble with "destroyed forms"

  // Erase the background :
  osb->Canvas->Brush->Style = bsSolid;
  osb->Canvas->Brush->Color = clBlack;
  osb->Canvas->FillRect( TRect(0,0,w,h) );  // <- this wasn't the CPU hog !

  // Prepare a few parameters for the "oscilloscope screen layout" :
  h2 = (h-16) / TIMING_SCOPE_N_DIGITAL_CHANNELS; // height for channel, including space between channels
  h3 = (3 * h2) / 4;           // height in pixels for the CURVES, without spaces

  // Draw optional info / warnings 'into the background' ?
  osb->Canvas->Font->Name  = "Arial";
  osb->Canvas->Font->Size  = 12;
  osb->Canvas->Font->Color = clGray;
  th = osb->Canvas->TextHeight( "Q" );
  tx = ty = 16;
  if( pScope->cfg.iVisibleChannels == 0 )   // oops... all scope channels turned off !
   { osb->Canvas->TextOut( tx, ty, "All scope channels are turned off - " );
     ty += th;
     osb->Canvas->TextOut( tx, ty, " Use the RIGHT MOUSE BUTTON" );
     ty += th;
     osb->Canvas->TextOut( tx, ty, " to open the scope's context menu." );
     ty += th;
   }

#if( TIMING_SCOPE_N_ANALOG_CHANNELS > 0 )
  // Draw the ANALOG channels (if any, to have them in the BACKGROUND) :
  for( int iChannel=0; iChannel<TIMING_SCOPE_N_ANALOG_CHANNELS; ++iChannel )
   { // (draw channel by channel, to avoid having to swap colours too often)
     switch( iChannel )  // select an analog-channel-specific "pen colour" :
      { case 0: dwPenColour = 0xFF0080;  // VIOLET
                break;
        case 1: dwPenColour = 0x007FFF;  // ORANGE
                break;
        case 2: dwPenColour = 0xC0C000;  // slightly dimmed CYAN
                break;
        case 3: dwPenColour = 0x007F7F;  // BROWN
                break;
        default: dwPenColour= 0xFF00FF;  // MAGENTA
                break;
      }
     if( ( dwPenColour != 0x000000 )  // only draw the curve if the channel isn't hidden ..
      && ( pScope->cfg.iVisibleChannels & ( 1<<(iChannel+TIMING_SCOPE_FIRST_ANALOG_CHANNEL) ) )  )
      { osb->Canvas->Pen->Color = (TColor)dwPenColour;
        yc = h/2 - 1 + 2*iChannel;
        nPoints = 0;
        for( iSample=0; iSample<TIMING_SCOPE_NUM_SAMPLE_POINTS; ++iSample)
         { g_iDebugDummyIndex = iSample;
           x = (w * iSample) / TIMING_SCOPE_NUM_SAMPLE_POINTS;
           flt = pScope->sSamplePoints[iSample].fltAnalog[iChannel];
           if( flt > 1.0f ) // gotcha..  clipping (no problem for a "float", but for PLOTTING further below)
            {  flt = 1.0f;  // <- good place for a breakpoint ..
               // WATCH the scope's sample-array as described further below.
            }
           if( flt < -1.0f )
            {  flt = -1.0f;
            }
           flt *= (float)(h/2);
           // 2024-01-14: Occasionally got a 'floating point overrun' here ?!
           //      After that, the antique debugger was totally clueless,
           //      and even though PAUSED AT THIS POINT,
           //      could not inspect any local variable or pointer anymore.
           //  To circumvent that (and see what was wrong in pScope->sSamplePoints[]),
           //  WATCH the following expression instead (works because there's only ONE INSTANCE yet,
           //  fortunately, an old-fashioned debugger-friendly global variable):
           //  CwKeyer_TimingScope.sSamplePoints[g_iDebugDummyIndex-2] (with a 'repeat count' of e.g. 4) .
           //
           y = yc - (int)flt;
           if(iSample==0)
            {
              pPoints[nPoints].x   = x;     // add the first point to the polyline
              pPoints[nPoints++].y = y;
            }
           else
            { if( (x != prev_x) || (y != prev_y) ) // avoid unnecessary GDI calls to save CPU time
               {
                 pPoints[nPoints].x   = x;     // append another point to the polyline
                 pPoints[nPoints++].y = y;
               } // end if < different x or y >
            }
           prev_x = x;
           prev_y = y;
         } // end for < all sample of a channel >
        if( nPoints >= 2 )
         { osb->Canvas->Polyline( pPoints, nPoints-1 );
           // ,----------------------------|_______|
           // '--> funny but true: Not the NUMBER OF POINTS,
           //                      but the NUMBER OF POINTS MINUS ONE !
         }
      }  // end if < ANALOG channel visible > ?
   }    // end for < all ANALOG channels >
#endif // TIMING_SCOPE_N_ANALOG_CHANNELS > 0 ?


  // Draw the DIGITAL channels (in the foreground, considered 'more important' than the audio waveforms).
  for( int iChannel=0; iChannel<TIMING_SCOPE_N_DIGITAL_CHANNELS; ++iChannel )
   { // (also here, draw channel by channel, to avoid having to swap colours too often)
     switch( iChannel )  // select a digital-channel-specific "pen colour" :
      { case TIMING_SCOPE_CHANNEL_DASH_INPUT:
             dwPenColour = 0x00FFFF; // YELLOW
             break;
        case TIMING_SCOPE_CHANNEL_DOT_INPUT :
             dwPenColour = 0x3FFF3F; // LIGHTGREEN
             break;
        case TIMING_SCOPE_CHANNEL_CW_OUTPUT :
             dwPenColour = 0x3F3FFF; // LIGHTRED ("we're most likely on air now")
             break;
        case TIMING_SCOPE_CHANNEL_FOUR: // ex: .. _PTT_OUTPUT
             dwPenColour = 0xFF7F7F; // LIGHTBLUE for the user-selectable source
             break;
        default:
             dwPenColour = 0x000000; // black -> HIDDEN !
             break;
      }
     if( (dwPenColour != 0x000000 )  // only draw the curve if the channel isn't hidden ..
       &&((pScope->cfg.iVisibleChannels & (1<<iChannel) ) != 0 ) )
      { osb->Canvas->Pen->Color = (TColor)dwPenColour;
        y_min = iChannel * h2;
        y_max = y_min + (h2-1);
        nPoints = 0;
        for( iSample=0; iSample<TIMING_SCOPE_NUM_SAMPLE_POINTS; ++iSample)
         { x = (w * iSample) / TIMING_SCOPE_NUM_SAMPLE_POINTS;
           val = pScope->sSamplePoints[iSample].iChannel[iChannel];
           y = y_max - (val * h3) / 100/*percent*/;
           if(iSample==0)
            {
              pPoints[nPoints].x   = x;     // add the first point to the polyline
              pPoints[nPoints++].y = y;
            }
           else
            { if( (x != prev_x) || (y != prev_y) ) // avoid unnecessary GDI calls to save CPU time
               {
                 pPoints[nPoints].x   = x;     // append another point to the polyline
                 pPoints[nPoints++].y = y;
               } // end if < different x or y >
            }
           prev_x = x;
           prev_y = y;
         } // end for < all sample of a channel >
         if( nPoints >= 2 )
          { osb->Canvas->Polyline( pPoints, nPoints-1 );
            // ,----------------------------|_______|
            // '--> funny but true: Not the NUMBER OF POINTS,
            //                      but the NUMBER OF POINTS MINUS ONE !
          }
      } // end if < DIGITAL channel visible > ?
   }   // end for < all DIGITAL channels >

  // Draw "decoded text" into the Timing Scope, too :
  osb->Canvas->Font->Name  = "Arial";
  osb->Canvas->Font->Size  = 16;
  osb->Canvas->Font->Color = clWhite;
  osb->Canvas->Pen->Color  = clWhite;
  th = osb->Canvas->TextHeight( "Q" );
  for( iSample=iSampleAtBeginOfChar=0; iSample<TIMING_SCOPE_NUM_SAMPLE_POINTS; ++iSample)
   {
     if( (wCwChar = pScope->sSamplePoints[iSample].wCwChar) != 0)
      { pszDecoded = Elbug_MorseCodePatternToASCII( wCwChar );
        if( wCwChar & ELBUG_RESULT_BEGIN_NEW_CHAR ) // nothing "decoded" at this sample index,
         { // but a marker for the BEGIN of a character (decoded and printed LATER):
           iSampleAtBeginOfChar = iSample;
         }
        else
         { // It's a decoded character, so print it, horizontally centered...
           x = (w * iSample) / TIMING_SCOPE_NUM_SAMPLE_POINTS;
           y = (TIMING_SCOPE_N_DIGITAL_CHANNELS-1) * h2 + th/4; // near the "CW output" channel
           prev_x = (w * iSampleAtBeginOfChar) / TIMING_SCOPE_NUM_SAMPLE_POINTS;
           // Horizontally center the text under the "CW Signal" that belongs to it:
           //      ___   _   ___     ___   _
           //  ___|   |_| |_|   |___|   |_| |___ <- "CW Signal" (keyer output)
           //
           //     '-------K -------''--- N ---'  <- CW decoder output
           //     |       |        |
           // prev_x      tx       x
           //  ,----------------|__|
           //  '--> This TWO-DOT-GAP belongs to the character,
           //       because without it, the decoder cannot decode it.
           //       If this 'Timing Scope' convinces anyone to stop "smearing
           //       characters together", this little gadget has already paid off :o)
           //
           tw = osb->Canvas->TextWidth( AnsiString(pszDecoded) );
           tx = prev_x + (x-prev_x-tw) / 2;
           SetBkMode( osb->Canvas->Handle, TRANSPARENT ); // <- old Windows 'GDI' function
           osb->Canvas->TextOut( tx, y, AnsiString(pszDecoded) );
           if( x != prev_x )
            { osb->Canvas->MoveTo( prev_x,  y      );
              osb->Canvas->LineTo( prev_x,  y+th/2 );
              osb->Canvas->LineTo( tx,      y+th/2 );
              osb->Canvas->MoveTo( tx+tw,   y+th/2 );
              osb->Canvas->LineTo( x,       y+th/2 );
              osb->Canvas->LineTo( x,       y      );
            }
         }
      } // end if < "decoded character" at index iSample > ?
   } // end for < all sample-indices, looking for DECODED CHARACTERS to print >

  // Draw a primitive, relative timescale; stepwidth = ONE DOT INTERVAL..
  osb->Canvas->Pen->Color = clWhite;
  iSample=0;
  while( iSample<TIMING_SCOPE_NUM_SAMPLE_POINTS )
   { x = (w * iSample) / TIMING_SCOPE_NUM_SAMPLE_POINTS;
     osb->Canvas->MoveTo( x,  0   );
     osb->Canvas->LineTo( x,  3   );
     osb->Canvas->MoveTo( x,  h-4 );
     osb->Canvas->LineTo( x,  h-1 );

     // Assume the CW-keyer's worker thread really samples the inputs
     //        at <pScope->iMillisecondsPerSample> ...
     if(  (pScope->iMillisecondsPerSample >= 1 )
        &&(CwKeyer_Elbug.cfg.iDotTime_ms >= pScope->iMillisecondsPerSample ) )
      { iSample += (CwKeyer_Elbug.cfg.iDotTime_ms / pScope->iMillisecondsPerSample);
      }
     else // something wrong with the Elbug or Timing Scope parameters..
      { // draw a tick every TEN samples (this may be something like 20 ms per tick)
        iSample += 10;
      }
   }

  // Draw the overlays for 'mouse cursor readout' (especially time differences)
  for( int i=0; i<N_SCOPE_OVERLAYS; ++i )
   { T_TimingScopeOverlay *pOverlay = &TimingScopeOverlay[i];
     if( pOverlay->visible )
      { osb->Canvas->Pen->Color = clWhite;
        osb->Canvas->Font->Size  = 10;
        osb->Canvas->MoveTo( pOverlay->x1, pOverlay->y1 );
        osb->Canvas->LineTo( pOverlay->x2, pOverlay->y2 );
        SetBkMode( osb->Canvas->Handle, OPAQUE ); // <- old Windows 'GDI' function
        s = "dt="+IntToStr( pOverlay->t2_ms - pOverlay->t1_ms )+" ms";
        th = osb->Canvas->TextHeight( s );
        tw = osb->Canvas->TextWidth( s );
        tx = (pOverlay->x1 + pOverlay->x2) / 2 - tw/2;
        ty = (pOverlay->y1 + pOverlay->y2) / 2 + th/4;
        osb->Canvas->TextOut( tx, ty, s );
        if( CwKeyer_Elbug.cfg.iDotTime_ms >= 1 )
         { s = "="+FormatFloat( "#0.0",
                     (double)(pOverlay->t2_ms - pOverlay->t1_ms)
                   / (double)CwKeyer_Elbug.cfg.iDotTime_ms) + " dots";
           osb->Canvas->TextOut( tx, ty+th, s );
         }

      }
   }


  // Draw the "windscreen wiper" to show the current position on the
  // non-scrolling, Radar-like sweeping display:
  x = (w * pScope->iLatestSampleIndex) / TIMING_SCOPE_NUM_SAMPLE_POINTS;
  if( x>=0 && x<w )
   { osb->Canvas->Pen->Color = clWhite;
     osb->Canvas->MoveTo( x, 0 );
     osb->Canvas->LineTo( x, h-15 );
   }

  pDestBitmap->Canvas->Draw( 0,0, osb );

  HERE_I_AM__GUI();

  delete osb; // clean up: delete the temporary "off-screen bitmap"

  free(pPoints);
  HERE_I_AM__GUI();

} // end UpdateTimingScope()

//---------------------------------------------------------------------------
void HandleMouseEventInTimingScope( Graphics::TBitmap *pBitmap,
           int iEvent, int x, int y )
#        define EVT_MOUSE_DOWN 1
#        define EVT_MOUSE_MOVE 2
#        define EVT_MOUSE_UP   3
      // [out]  TimingScopeOverlay[] : mouse-controlled 'overlay' to measure times,
      //                               displayed in UpdateTimingScope()
{
  int w = pBitmap->Width;  // width in pixels
  int h = pBitmap->Height; // height in pixels
  int iSample, t_ms;
  T_TimingScopeOverlay *pOverlay;

  if( iEvent==EVT_MOUSE_DOWN ) // switch to the next overlay ?
   { TimingScope_iCurrentOverlay = (TimingScope_iCurrentOverlay+1) % N_SCOPE_OVERLAYS;
   }
  pOverlay = &TimingScopeOverlay[ TimingScope_iCurrentOverlay ];

  // Convert 'x' into a sample index, inverse to the following :
  //   x = (w * iSample) / TIMING_SCOPE_NUM_SAMPLE_POINTS;
  if( (w > 0 ) && (CwKeyer_TimingScope.iMillisecondsPerSample > 0 ) )
   { iSample = (x * TIMING_SCOPE_NUM_SAMPLE_POINTS) / w;
     // Convert the (relative) SAMPLE INDEX into an offset in milliseconds :
     t_ms = iSample * CwKeyer_TimingScope.iMillisecondsPerSample;
     if( iEvent == EVT_MOUSE_DOWN )
      { pOverlay->x1 = x;
        pOverlay->y1 = y;
        pOverlay->t1_ms = t_ms;
      }
     else //  EVT_MOUSE_MOVE or EVT_MOUSE_UP ->
      { pOverlay->x2 = x;
        pOverlay->y2 = y;
        pOverlay->t2_ms = t_ms;
        pOverlay->visible = TRUE;
      }
     TimingScope_iUpdateCountOnTestTab = -1; // kludge to redraw the timing scope..
     TimingScope_iUpdateCountOnKeyerTab= -1; // ... even when 'paused' .
   }
  (void)h;
} // end HandleMouseEventInTimingScope()

//---------------------------------------------------------------------------
void ClearScopeOverlays(void)
{
  for(int i=0; i<N_SCOPE_OVERLAYS; ++i)
   { TimingScopeOverlay[i].visible = FALSE;
   }
  TimingScope_iUpdateCountOnTestTab = -1; // redraw a.s.a.p. ...
  TimingScope_iUpdateCountOnKeyerTab= -1; // ... even when 'paused' .
} // end ClearScopeOverlays()

//---------------------------------------------------------------------------
void SetColourScheme( int iColourScheme )
{
#if( ! defined __clang__ ) // guess it's "an older VCL" NOT supporting 'Styles' ..
  int iComp, iForm, nForms = Screen->FormCount;  // how many FORMS do we have ? [at least one]
  TForm      *pForm;
  TComponent *pComp;
#endif // old VCL without support for 'Styles' ?
  T_SpecDispControl *pDispCtrl = &g_SpecDispControl;
  DWORD dwErrorCode;
  char  sz255[256], *pszMsg;

  HERE_I_AM__GUI();
  pDispCtrl->clButtonFace = (DWORD)clBtnFace;    // form's BACKGROUND COLOUR - don't ask why
  pDispCtrl->clWindowText = (DWORD)clWindowText;
  pDispCtrl->clWindowBackground = (DWORD)clWindow;     // per default, WHITE background in e.g. combos and edit fields
  pDispCtrl->clActiveCaption  = (DWORD)clActiveCaption;   // > "color of the active window's title bar"
  pDispCtrl->clInactiveCaption= (DWORD)clInactiveCaption; // > "color of inactive windows' title bar"
  pDispCtrl->clButtonFont = pDispCtrl->clWindowText;

  switch( iColourScheme )
   { case COLOUR_SCHEME_DEFAULT : // back to e.g. "clBtnFace", etc etc etc etc
     default:
        // (most already set in the variable initialisers)
        g_clMenuBackground = TColor( 0xF0F0F0 ); // light-gray background
        g_clMenuSelBackgnd = TColor( 0xC08080 ); // light grayish-blue when selected
        g_clMenuForeground = TColor( 0x000000 ); // black text (regardless of selection)
        g_clDecoderOutputForeground = TColor( 0x000000 ); // black text
        g_clDecoderOutputBackground = TColor( 0xFFC0C0 ); // light-blue background
        break;
     case COLOUR_SCHEME_DARK : // for night owls and low-band CW ops
        pDispCtrl->clButtonFace = TColor( 0x202020 ); // dark gray
        pDispCtrl->clWindowText = clWhite;
        pDispCtrl->clWindowBackground     = clBlack;
        pDispCtrl->clActiveCaption   = TColor( 0xC04040 ); // dark grayish-blue
        pDispCtrl->clInactiveCaption = TColor( 0x404040 ); // even darker, almost dead
        // As long as we cannot change a button's BACKGROUND COLOUR (a light gray)
        // it's a bad idea to set the button's FONT COLOR to ALMOST WHITE. Thus:
        pDispCtrl->clButtonFont = TColor( 0x804040 ); // dark grayish-blue
        g_clMenuBackground = TColor( 0x404040 );   // dark background
        g_clMenuSelBackgnd = TColor( 0xC04040 );   // dark grayish-blue when selected
        g_clMenuForeground = TColor( 0xE0E0E0 );   // white text
        g_clDecoderOutputForeground = TColor( 0xE0E0E0 ); // white text
        g_clDecoderOutputBackground = TColor( 0x800000 ); // dark blue background

        break;
   } // end switch( iColourScheme )

#if( SWI_USE_VCL_STYLES ) // guess it's "a newer VCL" supporting 'Styles' ..
  // The C++Builder 12  "RAD Studio Topics", chapter "Application Appearance"
  // mentions a "TStyleManager.TrySetStyle("name"), where "name" seems to be
  // what can be selected in the 'Project' options under "Application"..
  // "Appearance".."Custom Styles", e.g.
  //  * "Charcoal Dark Slate" (resembles iColourScheme=COLOUR_SCHEME_DARK)
  //  * "Light"               (resembles iColourScheme=COLOUR_SCHEME_DEFAULT)
  // Surprisingly (and a bit scary), a VCL form's current "Style" is not
  // a property of the form itself, but set through the following :
  bool ok;
  switch( iColourScheme )
   { case COLOUR_SCHEME_DARK:
        ok = Vcl::Themes::TStyleManager::TrySetStyle("Charcoal Dark Slate", false/*ShowErrorDialog?NO!*/);
        // ,------------------------|_________|
        // '-> Sets the style specified by name as the active style,
        //   > without raising an exception.
        //   > TrySetStyle returns True if the style was successfully set
        //   > as the active style, and False otherwise.
        // When first tried, ok was TRUE, but the entire form still used the
        // default colours as shown in the FORM DESIGNER . Why ?
        // Possibly because of this (found in an outdated online resource):
        // > You Must Define Control Ownership
        // > In styled VCL applications, it is specially important that all
        // > controls are created with a valid owner, such as the form
        // > where they belong, so that the VCL can identify them as part
        // > of the styled application and apply the configured style
        // > to them accordingly.
        //
        break;
     default:
        ok = Vcl::Themes::TStyleManager::TrySetStyle("Light", false/*ShowErrorDialog?NO!*/);
        break;
   }
#else // ! SWI_USE_VCL_STYLES -> guess it's "the old VCL" without 'Styles'

  // The venerable old VCL (from Borland C++ Builder V6) did not support
  // "Schemes", "Themes", "Styles", "Skins" or whatever they call it now.
  // But for standard controls like menus, tabsheets, edit fields,
  // combo boxes and the like, Borland's form designer / "object inspector"
  // often uses colour names like e.g.
  //    "clBtnFace"    (funny name, often used as the entire form's
  //                    BACKGROUND COLOUR, for example in 'KeyerMainForm'),
  //    "clWindowText" (typically used for a TTabsheet's "Font.Color" property),
  //
  //    "clWindow"     (funny name for the WHITE background in combo boxes,
  //                    edit fields, and a couple of others).
  // The VCL documentation lists the above as ..
  //  > colors that are defined in the Windows Control panel.
  // Ok, but we don't want to fool around with the 'Windows Control panel'
  // (whatever / wherever that may be these days) - instead, THIS APPLICATION
  // shall of course only modify ITS OWN COLOURS .
  // There is no 'TStyleManager'. Didn't want to hook into Win32 "GetSysColor()".
  //   Maybe the Win32 "SetSysColors()" function will do the trick ?
  //   No, forget it, SetSysColors() affects the colours for ALL windows.

  // Thus: Iterate through all controls on all forms, find their TYPES,
  //       and depending on that, replace some of their colours:
  nForms = Screen->FormCount;  // how many FORMS do we have ? [at least one]

  for(iForm=0; iForm<nForms; ++iForm)
   { // 2024-11-30: Controlling BACKGROUND COLOURS this way was ok when compiled with BCB V6,
     //             but failed when compiled with Embarcadero C++Builder V12 "Athens".
     // For the "old VCL" (from C++Builder V6), TTabSheets like e.g. "TS_IO"
     //     were coloured
     pForm = Screen->Forms[iForm];  // glorious VCL allows accessing them all 'like an array'
     pForm->Color = (TColor)pDispCtrl->clButtonFace;
     // |--> With the 'old VCL' (from BCB V6), setting this TForm.Color also affected
     // |    the colour on all TABSHEETS of the form; regardless of nesting. Fine.
     // '--> With the 'new VCL' (in C++Builder V12) ...
     //         * all those TTabSheets did NOT change their background colour,
     //         * TLabels on those TTabSheets did NOT change their background colour,
     //           but their FOREGROUND colour,
     //         *
     //
     pForm->Font->Color = (TColor)pDispCtrl->clWindowText;
     for( iComp=0; (iComp<pForm->ComponentCount); iComp++ )
      { HERE_I_AM__GUI();
        pComp = pForm->Components[iComp];
        if (pComp->ClassNameIs("TMainMenu") )
         { // ex: TMainMenu *menu = (TMainMenu*)pComp;
           // Adapting the stupid main menu's back- and foreground colours
           // via e.g. "SetMenuInfo()" is absolutely hopeless. The crappy API
           // can modifiy the background of the DROPPED-DOWN items but not the
           // text colour, and it cannot modify the colours in the menu title.
         }
        else if (pComp->ClassNameIs("TMenuItem") )
         { // Unfortunately, the VCL designers at Borland forgot to add a
           // 'Color'/'Colour' property for this beast as well,
           // so without digging in AWFULLY DEEP into Win32 programming,
           // we cannot adapt the background of the title bar and main menu
           // to our 'Dark' colour scheme.
         }
        else if (pComp->ClassNameIs("TPopupMenu") )
         {
         }
        else if (pComp->ClassNameIs("TLabel") )
         { TLabel *label = (TLabel*)pComp;
           // Replace the TLabel's back- and foreground colour:
           label->Color = (TColor)pDispCtrl->clButtonFace;
           if( label->OnClick != NULL )
            { // "TLabel" acting like a 'link' (clickable) : BLUE
              if( iColourScheme == COLOUR_SCHEME_DARK )
               { label->Font->Color = clSkyBlue; // there's no clLightBlue :o(
               }
              else // white background -> dark blue text
               { label->Font->Color = clBlue;
               }
            }
           else  // not clickable -> no "link" but normal static text
            { label->Font->Color = (TColor)pDispCtrl->clWindowText;
            }
         } // end if <component is a TLabel>
        else if (pComp->ClassNameIs("TButton") )
         { TButton *btn = (TButton *)pComp;
           btn->Font->Color = (TColor)pDispCtrl->clWindowText;
         } // end if <component is a TButton>
        else if (pComp->ClassNameIs("TBitBtn") )
         { TBitBtn *btn = (TBitBtn *)pComp;
           // ex: btn->Font->Color = pDispCtrl->clWindowText;
           btn->Font->Color = (TColor)pDispCtrl->clButtonFont;
         } // end if <component is a TBitBtn>
        else if (pComp->ClassNameIs("TRadioButton") )
         { TRadioButton *btn = (TRadioButton *)pComp;
           btn->Font->Color = (TColor)pDispCtrl->clWindowText;
         } // end if <component is a TRadioButton>
        else if (pComp->ClassNameIs("TSpeedButton") )
         { // ex: TSpeedButton *btn = (TSpeedButton *)pComp;
           // A "speed button" neither has a foreground nor a background colour
           // that we can modify here. Modifying the FONT colour is nonsense
           // because a "speed button" is usually a square thing without any text,
           // but just a "glyph" (microscopic bitmap) that cannot be re-coloured.
         } // end if <component is a TSpeedButton>
        else if (pComp->ClassNameIs("TPanel") )
         { TPanel *pnl = (TPanel *)pComp;
           pnl->Color  = (TColor)pDispCtrl->clWindowBackground;  // !
           pnl->Font->Color = (TColor)pDispCtrl->clWindowText;
         }
        else if (pComp->ClassNameIs("TTabSheet") )
         { TTabSheet *ts = (TTabSheet *)pComp;
           ts->Font->Color = (TColor)pDispCtrl->clWindowText;
           // For obscure reasons, even the form's "Refresh()" didn't
           // cause the VCL TTabSheet to redraw its BACKGROUND . Nnngrrr.
           // ts->Refresh();  // <- tried this.. no effect
         } // end if <component is a TTabSheet>
        else if (pComp->ClassNameIs("TCheckBox") )
         { TCheckBox *chk = (TCheckBox *)pComp;
           chk->Color = (TColor)pDispCtrl->clButtonFace;  // !
           chk->Font->Color = (TColor)pDispCtrl->clWindowText;
         } // end if <component is a TCheckBox>
        else if (pComp->ClassNameIs("TGroupBox") )
         { TGroupBox *gb = (TGroupBox *)pComp;
           gb->Color       = (TColor)pDispCtrl->clWindowBackground;  // !
           gb->Font->Color = (TColor)pDispCtrl->clWindowText;
         }
        else if (pComp->ClassNameIs("TListBox") )
         { TListBox *lb = (TListBox *)pComp;
           lb->Color       = (TColor)pDispCtrl->clWindowBackground;
           lb->Font->Color = (TColor)pDispCtrl->clWindowText;
         } // end if <component is a TListBox>
        else if (pComp->ClassNameIs("TComboBox") )
         { TComboBox *cb = (TComboBox *)pComp;
           cb->Color       = (TColor)pDispCtrl->clWindowBackground;
           cb->Font->Color = (TColor)pDispCtrl->clWindowText;
         } // end if <component is a TComboBox>
        else if (pComp->ClassNameIs("TRadioGroup") )
         { TRadioGroup *rg = (TRadioGroup *)pComp;
           rg->Color       = (TColor)pDispCtrl->clWindowBackground;
           rg->Font->Color = (TColor)pDispCtrl->clWindowText;
         } // end if <component is a TRadioGroup>
        else if (pComp->ClassNameIs("TCheckListBox") )
         { TCheckListBox *cl = (TCheckListBox *)pComp;
           cl->Color       = (TColor)pDispCtrl->clWindowBackground;
           cl->Font->Color = (TColor)pDispCtrl->clWindowText;
         } // end if <component is a TCheckListBox>
        else if (pComp->ClassNameIs("TStringGrid") )
         { TStringGrid *sgr = (TStringGrid *)pComp;
           (void)sgr;
         } // end if <component is a TStringGrid>
        else if (pComp->ClassNameIs("TEdit") )
         { TEdit *ed = (TEdit *)pComp;
           ed->Color       = (TColor)pDispCtrl->clWindowBackground;
           ed->Font->Color = (TColor)pDispCtrl->clWindowText;
         } // end if <component is a TEdit>
        else if (pComp->ClassNameIs("TRichEdit") )
         { TRichEdit *ed = (TRichEdit *)pComp;
           ed->Color       = (TColor)pDispCtrl->clWindowBackground;
           ed->Font->Color = (TColor)pDispCtrl->clWindowText;
           ed->Clear();
         } // end if <component is a TRichEdit>
        else if (pComp->ClassNameIs("TScrollBar") )
         { TScrollBar *sb = (TScrollBar *)pComp;
           // Like many others, this poor VCL fellow also doen't have
           // an accessable 'Color' property...
           (void)sb;
         }
        else if (pComp->ClassNameIs("TTrackBar") )
         { TTrackBar *tb = (TTrackBar *)pComp;
           // This poor fellow doesn't even have an accessable 'Color' property !
           // C compiler said "Access to TControl::Color is impossible". Nnngrrr.
           // tb->Color = (TColor)pDispCtrl->clWindowBackground;
           (void)tb;
         } // end if <component is a TTrackBar>
        else if (pComp->ClassNameIs("TPageControl") )
         { // no attempt made to 'colourize' this yet ..
         }
        else if (pComp->ClassNameIs("TScrollBox") )
         { // no attempt made to 'colourize' this yet ..
         }
        else if (pComp->ClassNameIs("TImage") )
         { // no attempt made to 'colourize' this yet ..
         }
        else if (pComp->ClassNameIs("TBevel") )
         { // no attempt made to 'colourize' this yet ..
         }
        else if (pComp->ClassNameIs("TStaticText") )
         { // no attempt made to 'colourize' this yet ..
         }
        else if (pComp->ClassNameIs("TTimer") )
         { // no need to ever 'colourize' this, because a "TTimer" is never visible (except in the form designer)
         }
        else // kludge for development .. what's missing in the above list ?
         { AnsiString sClassName = pComp->ClassName();
           ShowError( ERROR_CLASS_ERROR, "SetColourScheme: Unknown class name %s",
                                         sClassName.c_str() );
         }
        HERE_I_AM__GUI();
      } // end for( iComp=0; (iComp<pForm->ComponentCount); iComp++ )
   } // end for<all forms of the application>
  HERE_I_AM__GUI();
  pDispCtrl->fUpdateSpectrum = pDispCtrl->fUpdateFreqScale = TRUE;
#endif // ! SWI_USE_VCL_STYLES ?
} // SetColourScheme()

//---------------------------------------------------------------------------
void UpdateSMeter(
        TImage *pImg,  // [in,out] Borland-VCL-style "graphic image", with a "canvas" to paint into
        double dblSMeterLevel_dB ) // [in] "decibel over S0"
        // Keep it simple, begin at "S zero" :
#define C_SMETER_LEVEL_S0   0 /* "S0" on HF:       -20 dBuV */
#define C_SMETER_LEVEL_S9  54 /* "S9" on HF:       +34 dBuV */
#define C_SMETER_LEVEL_MAX 99 /* "S9+45 dB"                 */
{
  int h, w, w2, th, tw, dB, x1, x, xt, x_S9, x_Lev;
  AnsiString sLabel;
  BOOL fShowLabel;

  // Borland's/Embarcadero's VCL-style "TImage" contains a "TPicture".
  //    A "TPicture" is a "TGraphic" container; here it's a "TBitmap".
  //    The "TBitmap" has a "TCanvas", which is the thing we can actually paint on.
  TCanvas *pCanvas;

  // Make sure the "TBitmap" has the same pixel dimensions as the "TImage":
  pImg->Picture->Bitmap->Height = h = pImg->Height;
  pImg->Picture->Bitmap->Width  = w = pImg->Width;
  pCanvas = pImg->Picture->Bitmap->Canvas;

  // Prepare a sufficiently SMALL font for the S-Meter labels:
  pCanvas->Font->Name  = "Arial";
  pCanvas->Font->Size  = 6;
  pCanvas->Font->Color = (TColor)g_SpecDispControl.clWindowText; // <- this may be WHITE or BLACK .. good contrast against the bargraph colours and clWindowBackground
  th = pCanvas->TextHeight( "S" );

  x1 = pCanvas->TextWidth( "S" ); // wide enough to draw e.g. "S0" for the leftmost tick
  w2 = w - x1 - 4;

  // From the width (w, in pixels), calculate the end of the coloured bargraph:
  x_Lev = x1 + (int)((double)(dblSMeterLevel_dB  - C_SMETER_LEVEL_S0) * (double)w2
                   / (double)(C_SMETER_LEVEL_MAX - C_SMETER_LEVEL_S0) );
  if( x_Lev<0 ) x_Lev=0;
  if( x_Lev>=w) x_Lev=w-1;
  // Calculate the pixel position for "S9"
  // (where Icom's S-meter changes from lightblue to red)
  x_S9 = x1 + (int)((double)(C_SMETER_LEVEL_S9  - C_SMETER_LEVEL_S0) * (double)w2
                  / (double)(C_SMETER_LEVEL_MAX - C_SMETER_LEVEL_S0) );

  // Draw the "bargraph" into the BACKGROUND ..
  x = x_Lev;
  if( x > x_S9 )
   {  x = x_S9;
   }
  pCanvas->Brush->Color = clSkyBlue;    // colours inspired by Icom IC-7300 ..
  pCanvas->FillRect( TRect(x1,0,x,h) ); // below S9 = light blue
  if( x_Lev > x_S9 )
   { pCanvas->Brush->Color = clRed;    // above S9 = red
     pCanvas->FillRect( TRect(x_S9,0, x_Lev,h) );
   }
  pCanvas->Brush->Color = (TColor)g_SpecDispControl.clWindowBackground; // <- this may be BLACK or WHITE !
  pCanvas->FillRect( TRect(0,0, x1,h) );
  pCanvas->FillRect( TRect(x_Lev,0, w,h) );


  // Draw a crude "S-meter scale" into the foreground
  SetBkMode( pCanvas->Handle, TRANSPARENT ); // <- old Windows 'GDI' function .. what's the VCL equivalent ?
  pCanvas->Pen->Color = pCanvas->Font->Color;
  for( dB=C_SMETER_LEVEL_S0; dB<=C_SMETER_LEVEL_S9; dB+=6 )
   { x = x1 + (int)((double)( dB - C_SMETER_LEVEL_S0) * (double)w2
                  / (double)(C_SMETER_LEVEL_MAX - C_SMETER_LEVEL_S0) );
     pCanvas->MoveTo( x, 0 );
     pCanvas->LineTo( x, h-th );
     if(dB==C_SMETER_LEVEL_S0)
      { sLabel = "S0";
      }
     else
      { sLabel = IntToStr( (dB-C_SMETER_LEVEL_S0) / 6 );
      }
     tw = pCanvas->TextWidth( sLabel );
     xt = x-tw/2;
     pCanvas->TextOut( xt, h-th, sLabel );
   }
  // Above "S9" : only show "+x" every TWENTY dB (like Icom)
  for( dB=C_SMETER_LEVEL_S9+10; dB<=C_SMETER_LEVEL_MAX; dB+=10 )
   { x = x1 + (int)((double)( dB - C_SMETER_LEVEL_S0) * (double)w2
                  / (double)(C_SMETER_LEVEL_MAX - C_SMETER_LEVEL_S0) );
     pCanvas->MoveTo( x, 0 );
     pCanvas->LineTo( x, h-th );
     sLabel = AnsiString("+") + IntToStr( dB-C_SMETER_LEVEL_S9 );
     tw = pCanvas->TextWidth( sLabel );
     xt = x-tw/2;
     pCanvas->TextOut( xt, h-th, sLabel );
   }


} // end UpdateSMeter()

//---------------------------------------------------------------------------
void UpdateMultiFunctionMeter( // renders any of the horizontal bargraphs on the "Mult-funtion meter" (panel)
        TImage *pImg,  // [in,out] Borland-VCL-style "graphic image", with a "canvas" to paint into
        int x1, int y1, int x2, int y2, // [in] graphic area for one of the six(?) "meters" within pImg
        double dblValue, // [in] value to be displayed (percent, "dB", SWR, current, voltage, power, temperature..)
        double dblMinValue, double dblMaxValue, double dblStepwidth, // [in] physical value range (also used for the labelled scale)
        const char *pszParamName, // [in] name like "Power", "ALC", "COMP", "SWR", "Id", "Vd", "Temp"
        const char *pszPhysUnit,  // [in] physical unit (shown at the end of the tick scale), e.g. "%", "dB", "V", "A", "W", "°C"
        double dblRedStartValue, double dblRedEndValue ) // e.g. 10.0 .. 12.0 [Volts for the "red line" indicating UNDERVOLTAGE]

  // Result (approximately..)
  //
  //    ,----------------------------------------,   -  y1
  //    | Power ##|##|##|##|##|##|##|##|##|##| % |
  //    |       0 10 20 30 40 50 60 70 80 90 100 |
  //    '----------------------------------------'   -  y2
  //    |       |                            |   |
  //    x1     x1i ("inner bargraph area"  x2i  x2

{
  int x1i, y1i, x2i, y2i, h, w, w2, th, tw, dB, x, xt, yt;
  int x_Lev, x_RedStart, x_RedEnd;
  double d;
  AnsiString sLabel;
  BOOL fShowLabel;
  BOOL fEraseAll = FALSE;

#define L_COLOUR_TEST 0 // (0)=normal compilation, (1)=TEST using fixed colours

  // Borland's/Embarcadero's VCL-style "TImage" contains a "TPicture".
  //    A "TPicture" is a "TGraphic" container; here it's a "TBitmap".
  //    The "TBitmap" has a "TCanvas", which is the thing we can actually paint on.
  TCanvas *pCanvas;


  // Make sure the "TBitmap" has the same pixel dimensions as the "TImage":
  h = pImg->Height;
  w = pImg->Width;
  if( (pImg->Picture->Bitmap->Height != h) || (pImg->Picture->Bitmap->Width != w) )
   { fEraseAll = TRUE;
     pImg->Picture->Bitmap->Height = h;
     pImg->Picture->Bitmap->Width  = w;
   }
  pCanvas = pImg->Picture->Bitmap->Canvas;

  if( fEraseAll )
   {
#   if( L_COLOUR_TEST ) // Fill the entire bitmap with a painful colour ?
     pCanvas->Brush->Color = clRed;
#   else  // no eye-cancer please ..
     pCanvas->Brush->Color = (TColor)g_SpecDispControl.clWindowBackground; // <- this may be BLACK or WHITE !
#   endif
     pCanvas->Brush->Style = bsSolid;
     pCanvas->FillRect( TRect(0,0,w,h) );
   }

  // return;  // what caused the WHITE-filled area ? [not the stuff below..]

  // Prepare a sufficiently SMALL font for the S-Meter labels:
  pCanvas->Font->Name  = "Arial";
  pCanvas->Font->Size  = 6;
#if( L_COLOUR_TEST )
  pCanvas->Font->Color = clYellow;
#else
  pCanvas->Font->Color = (TColor)g_SpecDispControl.clWindowText; // <- this may be WHITE or BLACK .. good contrast against the bargraph colours and clWindowBackground
#endif // L_COLOUR_TEST ?
  th = pCanvas->TextHeight( "S" );
  tw = pCanvas->TextWidth( "S" ); // wide enough to draw e.g. "S0" for the leftmost tick

  // "inner width" for the bargraph, without margins for the labels and :
  x1i = x1 + 6*tw;
  x2i = x2 - 2*tw;
  y1i = y1 + 2;
  y2i = y2 - 2;
  w2  = x2i - x1i;

  // From the "inner width" (w2 in pixels), calculate the end of the coloured bargraph:
  x_Lev = x1i + (int)((double)(dblValue    - dblMinValue) * (double)w2
                    / (double)(dblMaxValue - dblMinValue) );
  UTL_LimitInteger( &x_Lev, x1, x2 );
  // Calculate the pixel position for "S9"
  // (where Icom's S-meter changes from lightblue to red)
  x_RedStart = x1i + (int)((double)(dblRedStartValue - dblMinValue) * (double)w2
                         / (double)(dblMaxValue      - dblMinValue) );
  x_RedEnd = x1i + (int)((double)(dblRedEndValue - dblMinValue) * (double)w2
                       / (double)(dblMaxValue    - dblMinValue) );
  UTL_LimitInteger( &x_RedStart, x1i, x2i );
  UTL_LimitInteger( &x_RedEnd,   x1i, x2i );
  // Note: For SOME "meters", the RED AREA is at the start (e.g. undervoltage).
  //       For OTHER "meters", the RED AREA is at the end (e.g. PA temperature).
  //       The "red area" is indicated as a line below the bargraph.
  //       Unlike the S-meter, the colour of the bargraph itself doesn't seem
  //       to change depending on the displayed value (?)

  // Clear the entire background (no problem with flicker; we use "double buffering")
#if( L_COLOUR_TEST )
  pCanvas->Brush->Color = clBlue;
#else
  pCanvas->Brush->Color = (TColor)g_SpecDispControl.clWindowBackground; // <- this may be BLACK or WHITE !
#endif // L_COLOUR_TEST ?
  pCanvas->Brush->Style = bsSolid;
  pCanvas->FillRect( TRect(x1,y1,x2, y2) );
    // ,-------------'
    // '--> "TRect represents the dimensions of a rectangle. The coordinates
    //       are specified as either four separate integers representing the
    //       left, top, right, and bottom sides, or as two points (...) "

#if( L_COLOUR_TEST )
  pCanvas->Brush->Color = clGreen;
#else
  pCanvas->Brush->Color = (TColor)g_SpecDispControl.clWindowText; // <- this may be BLACK or WHITE !
#endif // L_COLOUR_TEST ?

  pCanvas->FrameRect( TRect(x1+2,y1+2,x2-2,y2-2) ); // quite counter-intuitive, "FrameRect()" doesn't use the "Pen" but the "Brush" to draw a 1-pixel-wide border..

#if( L_COLOUR_TEST )
  pCanvas->Font->Color = (TColor)0x00C0FF; // "clOrange", but that doesn't exist in the VCL
#else
  pCanvas->Font->Color = (TColor)g_SpecDispControl.clWindowText;
#endif // L_COLOUR_TEST ?
  SetBkMode( pCanvas->Handle, TRANSPARENT ); // <- Windows 'GDI' .. what's the VCL equivalent ?
  if( (pszPhysUnit != NULL ) && (pszPhysUnit[0] != '\0') )
   { yt = (y1+y2)/2-th;  // vertical position for the parameter name ..
   }
  else // no "physical unit" so vertically align the PARAMETER NAME:
   { yt = (y1+y2)/2-th/2; // <- for example, SWR and ALC were dimensionless
   }
  if( pszParamName != NULL )
   { pCanvas->TextOut( x1 + 4, yt, pszParamName );
     yt += th;
   }
  if( (pszPhysUnit != NULL ) && (pszPhysUnit[0] != '\0') )
   { pCanvas->TextOut( x1 + 8, yt, "/ " + AnsiString(pszPhysUnit) );
     yt += th;
   }
  SetBkMode( pCanvas->Handle, OPAQUE );

#if( L_COLOUR_TEST )
  pCanvas->Brush->Color = clGray;
#else
  pCanvas->Brush->Color = (TColor)g_SpecDispControl.clWindowBackground; // <- this may be BLACK or WHITE !
#endif // L_COLOUR_TEST ?


  // Draw the "bargraph" into the BACKGROUND ..
  x = x_Lev;
  if( x>x1i )
   { pCanvas->Brush->Color = clSkyBlue;         // colours inspired by Icom IC-7300 ..
     pCanvas->FillRect( TRect(x1i,y1i,x,y2i) ); // light blue (for MOST bargraphs)
   }

  // Draw a crude "labelled scale" into the foreground ?
  if( (dblStepwidth>0.0) && (dblMaxValue>dblMinValue) )
   {
#   if( L_COLOUR_TEST )
     pCanvas->Font->Color = (TColor)0x00C0FF; // "clOrange", but that doesn't exist in the VCL
#   else
     pCanvas->Font->Color = (TColor)g_SpecDispControl.clWindowText;
#   endif // L_COLOUR_TEST ?
     pCanvas->Pen->Color = pCanvas->Font->Color;
     SetBkMode( pCanvas->Handle, TRANSPARENT ); // <- old Windows 'GDI' function .. what's the VCL equivalent ?
     for( d=dblMinValue; d<=dblMaxValue; d += dblStepwidth )
      { x = x1i + (int)((double)( d          - dblMinValue) * (double)w2
                      / (double)(dblMaxValue - dblMinValue) );
        pCanvas->MoveTo( x, y1i );
        pCanvas->LineTo( x, y2i-th );
        if( dblStepwidth >= 1.0 )
         { sLabel = IntToStr( (int)d );
         }
        else
         { sLabel = FormatFloat( "#.0", d );
         }
        tw = pCanvas->TextWidth( sLabel );
        xt = x-tw/2;
        pCanvas->TextOut( xt, y2i-th, sLabel );
      }
     SetBkMode( pCanvas->Handle, OPAQUE );

   }

} // end UpdateMultiFunctionMeter()


#if( SWI_HARDCORE_DEBUGGING )
//---------------------------------------------------------------------------
void CheckSystemHealth(const char *pszModuleName, int iSourceLine) // checks for 'memory corruption'...
  // Periodically called from a few places via macro CHECK_SYSTEM_HEALTH() .
  // Added 2024-02-25 when suspecting a problem with the hyper-complicated,
  // memory-hoggig Ogg/Vorbis encoder (that later turned out to be utterly
  // non-suited for low-latency audio, at least when used for the <audio>
  // element in HTML 5 without megatons of Javascript..)
  //
  // Because CheckSystemHealth() may be called from WORKER THREADS, hundreds
  // of times per second, keep the "checks" performed in this function to the
  // ultimate minimum (e.g. check a few 'magic numbers' at the end of FIFOs,
  // etc).
{
  BOOL fGotcha = FALSE;
  static BOOL firstBug = TRUE;
  //  2024-02-25: Out of the blue, there was garbage in MyDirectSound.sz255OutputDeviceName :
  if(  (MyDirectSound.sz255OutputDeviceName[0] >  0x00)
     &&(MyDirectSound.sz255OutputDeviceName[0] <= 0x20) )
   { fGotcha = TRUE;
   }

  if( fGotcha && firstBug )
   { ShowError( ERROR_CLASS_ERROR, "System health check failed, caller = %s, line = %d",
                      pszModuleName, (int)iSourceLine );
   }

} // end CheckSystemHealth()
#endif // SWI_HARDCORE_DEBUGGING ?


//---------------------------------------------------------------------------
const char* RigCtrl_GetRadioControlPortAsString(void) // -> e.g. "COM5", if that's the port "talking to the radio" .
  // Called from within RigControl.c to emit 'info messages' like
  //     "RigControl: IC-9700 on COM5, 432250.0 kHz, CWN"  .
  // Since RighControl.c doesn't know anything about serial ports,
  // RigCtrl_GetRadioControlPortAsString() must be provided by the application.
{
  static char sz15Result[16];

  // If the "COM port to 'key' the radio" is ONLY used for keying,
  //             but not to CONTROL the radio, the Remote CW Keyer
  //             can alternatively control the radio via Hamlib's "rigctld"
  //             ('Rig Control Daemon' which is basically 'Hamlib control
  //               commands in ASCII sent over TCP/IP; details at
  //               github.com/Hamlib/Hamlib/wiki/Network-Device-Control ).
  //             Other applications (like wfview) can also EMULATE
  //             Hamlib's "rigctld", with a TCP server listening on port 4532.
  switch( CwKeyer_Config.iRadioControlProtocol )
   { case RIGCTRL_PROTOCOL_NONE : return "NoPort";
     case RIGCTRL_PROTOCOL_ICOM_CI_V: // supported since the early beginnings..
        // HERE, to keep things simple in the Remote CW Keyer,
        // Icom's CI-V protocol is only supported via "the same local COM port
        // that also KEYS the rig (provides the Morse code modulation):
        sprintf( sz15Result, "COM%d", (int)CwKeyer_Config.iRadioKeyingAndControlPort );
        return sz15Result;
     // case RIGCTRL_PROTOCOL_KENWOOD, RIGCTRL_PROTOCOL_YAESU_CAT, RIGCTRL_PROTOCOL_ELECRAFT, ...
     // NONE of these will ever be supported natively .. so instead,
     //   use use something Hamlib/Rigctld-compatible that runs a server on a
     //   TCP/IP port, e.g wfview's built-in "rigctld emulation" !
     //
     case RIGCTRL_PROTOCOL_HAMLIB_RIGCTLD : // added 2024-06 in the Remote CW Keyer...
     // ex: case RIGCTRL_PROTOCOL_FLRIG_XMLRPC   :
        // Doesn't use the built-in CI-V engine in DL4YHF's RigControl.c,
        // but the ASCII-based 'Hamlib' compatible commands via TCP/IP:
        return CwKeyer_Config.sz80RemoteRigCtrlServerAddress;

     default:
        sprintf( sz15Result, "??%d??", (int)CwKeyer_Config.iRadioControlProtocol );
        return sz15Result;

   } // end switch( CwKeyer_Config.iRadioControlProtocol )

} // end RigCtrl_GetRadioControlPortAsString()

