//---------------------------------------------------------------------------
// File: C:\cbproj\Remote_CW_Keyer\RigControl.h
// Authors: Wolfgang Buescher (DL4YHF)    .
//       (this stripped-down variant for DL4YHF's "Remote CW Keyer"
//        does NOT support the ugly "RS-BA1 protocol" and thus is MUCH
//        easier to port to a simple microcontroller firmware) .
// Date:   2021-09-16
// Purpose:
//  Common header for CAT ("computer aided tuning") protocol parsers.
//  Started 2018-12 with a parser for Icom'c CI-V protocol, implemented in
//          c:\cbproj\SpecLab\RigControl\CIV_parser.c .
//          Written in C, not C++, to MAYBE use parts of this in a micro-
//          controller firmare (a plan that didn't materialize yet...) .
//  First used in DL4YHF's Spectrum Lab .
//  Later (2024) stripped down for use in DL4YHF's "Remote CW Keyer",
//               without support for the poorly specified "RS-BA1" protocol
//               by Icom ..
//
// Details / Literature / Revsision History : See implementation (*.c) !
//---------------------------------------------------------------------------

#ifndef RigControlH
#define RigControlH

#ifndef _DATA_TYPES_H_
# include "yhf_type.h"  // old fashioned stuff like BYTE, WORD, DWORD, etc^10
#endif
#ifndef _YHF_TIMERS_H // C:\cbproj\SpecLab\Timers.h not included yet ? Include it here:
# include "Timers.h"  // TIM_GetCurrentUTC() for the 'traffic log', stopwatches, etc
#endif
#ifndef  CIRCULAR_FIFO_INCLUDED
# include "CircularFifo.h" // T_CFIFO is used as thread-safe buffer for serial ports
#endif
#ifndef  _STRING_LIB_H_
# include "StringLib.h"  // DL4YHF's string library (originally designed for Microcontrollers, thus pure "C")
#endif

/*-------------- Macros --------------------------------------------------*/

#ifndef  RIGCONTROL_SUPPORT_ICOM_UDP    // support communicating with Icom-radios via UDP (LAN,WLAN) ?
# define RIGCONTROL_SUPPORT_ICOM_UDP 0  // 1=yes (omg..); 0= HEAVENS, NO ! :o)
#endif // ndef  RIGCONTROL_SUPPORT_ICOM_UDP ?


/*-------------- Constants -----------------------------------------------*/


// Values for iRadioCtrlProtocol .
// To avoid including "all headers everywhere", these values must be
// compatible to those in SpecLab/Config.h (MConfig.iRadioCtrlProtocol) !
#define RIGCTRL_PROTOCOL_NONE         0
#define RIGCTRL_PROTOCOL_ICOM_CI_V    1 // supported since the early beginnings
#define RIGCTRL_PROTOCOL_KENWOOD      2 // not supported here yet
#define RIGCTRL_PROTOCOL_YAESU_CAT    3 // not supported here yet
#define RIGCTRL_PROTOCOL_ELECRAFT     4 // why did they have to CRAFT their own ? Not supported here.
#define RIGCTRL_PROTOCOL_AD9832_I2C   5 // supported somewhere else (in SpecLab) :)
#define RIGCTRL_PROTOCOL_HAMLIB_RIGCTLD 6 // first used in the Remote CW Keyer, 2024-06. Only via TCP/IP !
//#define RIGCTRL_PROTOCOL_FLRIG_XMLRPC 7 // FLRIG has an easy-to-use GUI, but XML-RPC is sooo wasteful
//  and completely unecessary for this purpose (why not just send commands via TCP/IP ?),
//  and FLRIG always wrecked the settings in an IC-9700 -> removed again.
// 


// Values for iRigControlFlags (bitwise combineable) :
#define RIGCTRL_FLAGS_NONE                    0x00
#define RIGCTRL_FLAG_SEND_SETTINGS_ON_STARTUP 0x01 // if not set, "read settings on startup" (but don't modify anything yet)
#define RIGCTRL_FLAG_WANT_SPECTRUM_DATA       0x02 // if set, convince the radio to periodically send SPECTRUM DATA (IC-7300)

// Values for iPort (to identify the "radio control" port, or any of the optional "duplicator ports"):
#define RIGCTRL_PORT_RADIO    0 // communication port that "talks to the radio"
#define RIGCTRL_PORT_CLIENT_1 1 // first "duplicator" port, allows sharing the radio with other apps
#define RIGCTRL_PORT_CLIENT_2 2 // second "duplicator" port, for another application
#define RIGCTRL_MAX_CLIENT_PORTS 2


// Return values from most parsing functions indicate the message type (iMsgType):
#define RIGCTRL_MSGTYPE_INFO      0 // not a MESSAGE but just an INFO, e.g. an important state transition. No "Packet Data".
#define RIGCTRL_MSGTYPE_ECHO      1 // "just an ECHO" (caused by the CI-V bus, originally a SINGLE-WIRE interface)
#define RIGCTRL_MSGTYPE_RADIO_ID  2 // "something to IDENTIFY the radio"
#define RIGCTRL_MSGTYPE_FREQUENCY_REPORT 3 // "frequency report" (from radio to user / application)
#define RIGCTRL_MSGTYPE_SET_FREQUENCY 4 // command to SET a new frequency (from user to radio)
#define RIGCTRL_MSGTYPE_OP_MODE   5 // "mode data" like USB/LSB/CW/FM/AM (often unsolicited, "on change")
#define RIGCTRL_MSGTYPE_RIT_XIT   6 // RIT or XIT data (often unsolicited, "on change")
#define RIGCTRL_MSGTYPE_RF_GAIN   7 // anything that affects the RF gain (preamp, attenuator, "AIP", ..)
#define RIGCTRL_MSGTYPE_C_SPECTRUM 14 // configure transmission of spectrum data
#define RIGCTRL_MSGTYPE_SPECTRUM   15 // "waveform data" (in fact, for Icom's spectrum scope)
#define RIGCTRL_MSGTYPE_SCOPE_MODE 16 // flag for pRC->iPostponedSetMessages to send pRC->iScopeMode a.s.a.p. (when not waiting for a response)
#define RIGCTRL_MSGTYPE_SCOPE_SPAN 17 // flag for pRC->iPostponedSetMessages to send pRC->dblScopeSpan_Hz a.s.a.p. (when not waiting for a response)
#define RIGCTRL_MSGTYPE_POWER_ON_OFF 18 // Packet to turn the remotely controlled rig on/off (or similar)
#define RIGCTRL_MSGTYPE_UDP_CONTROL 20 // Packet to/from Icom's UDP "Control" port (50001 by default)
#define RIGCTRL_MSGTYPE_UDP_SERIAL  21 // Packet to/from Icom's UDP "Serial" port (50002 by default), but none of the well-known CI-V message types like RIGCTRL_MSGTYPE_SPECTRUM
#define RIGCTRL_MSGTYPE_UDP_AUDIO   22 // Packet to/from Icom's UDP "Audio" port (50003 by default)
#define RIGCTRL_MSGTYPE_IDLE_ETC 23 // any kind of 'idle' message (extra flag to selectively REJECT them from logging)
#define RIGCTRL_MSGTYPE_UNKNOWN  30 // returned when a parser couldn't handle the data.
#define RIGCTRL_MSGTYPE_OTHER    31 // acknowledge for 'set', read "some other parameter", etc pp.
#define RIGCTRL_MSGTYPE_OK       32 // <- bitwise combineable. In CI-V : unspecific "OK", cmd 0xFB.  Without "Real data" (neither _TX nor _RX) : "info, all OK here."
#define RIGCTRL_MSGTYPE_NOT_OK   64 // <- bitwise combineable. In CI-V : unspecific "NG" / "Not Good", cmd 0xFA
// Add new message types also in RigCtrl_MessageTypeToString() !
#define RIGCTRL_MSGTYPE_STRIP_FLAGS 0x00FF // bitwise-AND mask to remove the following flags:
#define RIGCTRL_MSGTYPE_FLAG_TX       256 // For the log: Real "data", TRANSMITTED from controller to radio, bitwise OR-able to the above for the "CAT traffic log"
#define RIGCTRL_MSGTYPE_FLAG_RX       512 // For the log: Real "data", RECEIVED by the controller from the radio, "   "   "   "  "   ...
#define RIGCTRL_MSGTYPE_FLAG_EXPECTED 1024 // EXPECTED message (response matching earlier request, shown in GREEN colour in the traffic log)
#define RIGCTRL_MSGTYPE_FLAG_CIV     2048 // for the TRAFFIC LOG: pure CI-V payload, without overhead from any "transport layer" (and no "UDP-CIV-Header")
#define RIGCTRL_MSGTYPE_FLAG_UDP     4096 // not from/to a SERIAL PORT but UDP (datagram, usually Icom RS-BA compatible)
#define RIGCTRL_MSGTYPE_FLAG_PERIODIC_POLL 8192 // "periodic polled" message, e.g for the S-meter. Often rejected for the LOG
#define RIGCTRL_MSGTYPE_ERROR       16384 // <- set, for example, on response-reception-TIMEOUT.
#define RIGCTRL_MSGTYPE_FLAG_SUGGEST_NEW_HEADLINE_FOR_TRAFFIC_LOG 32768 // *suggestion* for the GUI to emit another line generated by RigCtrl_GetHeadlineForTrafficLog()


// Possible operating modes, not protocol-specific, and different from CI-V :
//   (a less ambiguous name would be "modulation", but "op mode" includes
//   "wide" or "narrow" settings or even "reverse", for historic reasons)
#define RIGCTRL_OPMODE_UNKNOWN 0  // possible values for iOpMode ...
#define RIGCTRL_OPMODE_CW      1
#define RIGCTRL_OPMODE_LSB     2
#define RIGCTRL_OPMODE_USB     3
#define RIGCTRL_OPMODE_AM      4
#define RIGCTRL_OPMODE_FM      5
#define RIGCTRL_OPMODE_FM_WIDE 6
#define RIGCTRL_OPMODE_RTTY    7
#define RIGCTRL_OPMODE_PSK     8
#define RIGCTRL_OPMODE_DIGITAL_VOICE 9 // "Digital Voice", presumably D-Star, not DMR
#define RIGCTRL_OPMODE_DIGITAL_DATA 10 // "DD" seen in the IC-9700 CI-V Reference Guide (no idea what "DD" really stands for.. 128 kbp "Digital Data" on 23 cm ?)
#define RIGCTRL_OPMODE_MASK_TO_STRIP_FLAGS 15
#define RIGCTRL_OPMODE_REVERSE     16 // bitwise combineable with some op-modes (CW, RTTY, PSK)
#define RIGCTRL_OPMODE_NARROW      32 // bitwise combineable with some op-modes ...
#define RIGCTRL_OPMODE_VERY_NARROW 64

// Possible values for T_RigCtrlInstance.iParameterPollingState (for CI-V sequencing):
#define RIGCTRL_POLLSTATE_PASSIVE             0 // initial state: have neither READ nor WRITTEN any setting from/to the remote rig
#define RIGCTRL_POLLSTATE_START_RD            1 // begin reading all we need to know about/from the remote rig
#define RIGCTRL_POLLSTATE_READ_RADIO_ID       2 // (try to) read the radio's "default" ID, as a connection test and to check its type
#define RIGCTRL_POLLSTATE_PARAMS_FOR_RIGCTL   3 // (try to) read the parameters required for RigControl.c itself
#define RIGCTRL_POLLSTATE_TX_BAND_EDGES       4 // (try to) retrieve all of the radio's TRANSMIT BANDS (with edge frequencies)
#define RIGCTRL_POLLSTATE_BAND_STACKING_REGS  5 // (try to) read out the 'band stacking registers' (special VFO memories)
#define RIGCTRL_POLLSTATE_PARAMS_FOR_HAMLIB   6 // (try to) read a few more params used by HamlibServer.c
#define RIGCTRL_POLLSTATE_CONFIGURE_SPECTRUM  7 // (try to) activate transmission of "spectrum scope waveform data"
#define RIGCTRL_POLLSTATE_DONE               10 // finished polling "all we need to know" about the radio
#define RIGCTRL_POLLSTATE_TURN_RIG_ON        11 // special service for modern rigs that can be turned on / off via CAT (e.g. IC-7300 via USB/CI-V)
#define RIGCTRL_POLLSTATE_WAIT_TURN_ON       12 // waiting for an "OK"-response after the command to "turn on"
#define RIGCTRL_POLLSTATE_TURN_RIG_OFF       13 // send command to turn the radio off (as soon as possible)
#define RIGCTRL_POLLSTATE_WAIT_TURN_OFF      14 // waiting for the rig to turn off
#define RIGCTRL_POLLSTATE_TURNED_OFF         15 // after RIGCTRL_POLLSTATE_TURN_RIG_OFF / WAIT_TURN_OFF, the only allowed command is to "turn on"
        // When adding new POLLSTATEs here, also add a string for them
        //  in RigCtrl_ParameterPollingStateToString() !

// CI-V default addresses, also used internally to identify a certain radio model .
//  Hint: Many of these are 'exposed' via RigCtrl_RadioInfo_CIV[] .
#define RIGCTRL_DEF_ADDR_NO_REMOTE_CTRL -1   // special dummy for .iDefaultAddress or CwKeyer_Config.iRadioCIVAddress if there shall be NO REMOTE CONTROL at all
#define RIGCTRL_DEF_ADDR_AUTO_DETECT    0    // special dummy for .iDefaultAddress or CwKeyer_Config.iRadioCIVAddress for 'auto-detect'
#define RIGCTRL_DEF_ADDR_IC_271         0x20 // VHF Allmode
#define RIGCTRL_DEF_ADDR_IC_275         0x10 // VHF Allmode
#define RIGCTRL_DEF_ADDR_IC_375         0x12 // 220MHz Allmode
#define RIGCTRL_DEF_ADDR_IC_471         0x22 // UHF Allmode
#define RIGCTRL_DEF_ADDR_IC_475         0x14 // UHF Allmode
#define RIGCTRL_DEF_ADDR_IC_575         0x16 // 10m & 6m Allmode
#define RIGCTRL_DEF_ADDR_IC_703         0x68 // HF Allmode QRP
#define RIGCTRL_DEF_ADDR_IC_705         0xA4 // HF/VHF/UHF portable QRP DSP transceiver, spectrum via CI-V, CI-V via USB & WLAN
#define RIGCTRL_DEF_ADDR_IC_905         0xAC // VHF/UHF/SHF portable QRP DSP transceiver, spectrum via CI-V, CI-V via USB & WLAN
#define RIGCTRL_DEF_ADDR_IC_706         0x48 // HF/VHF Allmode
#define RIGCTRL_DEF_ADDR_IC_706MkII     0x4E // HF/VHF Allmode
#define RIGCTRL_DEF_ADDR_IC_706MkIIG    0x58 // HF/VHF/UHF Allmode
#define RIGCTRL_DEF_ADDR_IC_707         0x3e // HF Allmode
#define RIGCTRL_DEF_ADDR_IC_718         0x5E // HF Allmode
#define RIGCTRL_DEF_ADDR_IC_725         0x28 // HF Allmode
#define RIGCTRL_DEF_ADDR_IC_726         0x30 // HF/6m Allmode
#define RIGCTRL_DEF_ADDR_IC_728         0x38 // HF Allmode
#define RIGCTRL_DEF_ADDR_IC_729         0x3A // HF/6m Allmode
#define RIGCTRL_DEF_ADDR_IC_735         0x04 // HF Allmode (possibly the first radio with CI-V ever made)
#define RIGCTRL_DEF_ADDR_IC_736         0x40 // HF/6m Allmode
#define RIGCTRL_DEF_ADDR_IC_737         0x3C // HF Allmode
#define RIGCTRL_DEF_ADDR_IC_738         0x44 // HF Allmode
#define RIGCTRL_DEF_ADDR_IC_746         0x56 // HF/VHF Allmode
#define RIGCTRL_DEF_ADDR_IC_751A        0x1C // HF Allmode
#define RIGCTRL_DEF_ADDR_IC_756         0x50 // HF/6m Allmode
#define RIGCTRL_DEF_ADDR_IC_756Pro      0x5C // HF/6m Allmode
#define RIGCTRL_DEF_ADDR_IC_756ProII    0x64 // HF/6m Allmode
#define RIGCTRL_DEF_ADDR_IC_756ProIII   0x6e // HF/6m Allmode
#define RIGCTRL_DEF_ADDR_IC_761         0x1E // HF Allmode
#define RIGCTRL_DEF_ADDR_IC_765         0x2C // HF Allmode
#define RIGCTRL_DEF_ADDR_IC_775         0x46 // HF Allmode
#define RIGCTRL_DEF_ADDR_IC_78          0x62 // HF Allmode
#define RIGCTRL_DEF_ADDR_IC_7000        0x70 // HF/VHF/UHF Allmode
#define RIGCTRL_DEF_ADDR_IC_7100        0x88 // HF/VHF/UHF
#define RIGCTRL_DEF_ADDR_IC_7200        0x76 // HF Allmode
#define RIGCTRL_DEF_ADDR_IC_7300        0x94 // HF/6m/4m, full DSP, spectrum data via CI-V)
#define RIGCTRL_DEF_ADDR_IC_9700        0xA2 // 2m/70cm/23cm, full DSP transceiver, spectrum via CI-V, CI-V via USB or LAN)
#define RIGCTRL_DEF_ADDR_IC_7400        0x66 // same ID as IC-746Pro ?
#define RIGCTRL_DEF_ADDR_IC_7600        0x7A // HF Allmode (non-DSP, no spectrum data via CI-V)
#define RIGCTRL_DEF_ADDR_IC_7610        0x98 // HF/6m Allmode, full DSP, spectrum data,..)
#define RIGCTRL_DEF_ADDR_IC_7700        0x74 // HF Allmode
#define RIGCTRL_DEF_ADDR_IC_7800        0x6A // HF Allmode
#define RIGCTRL_DEF_ADDR_IC_7851        0x8F // HF/6m Allmode, IF DSP, spectrum via CI-V
#define RIGCTRL_DEF_ADDR_IC_781         0x26 // HF Allmode
#define RIGCTRL_DEF_ADDR_IC_820         0x42 // VHF/UHF Allmode
#define RIGCTRL_DEF_ADDR_IC_821         0x4C // VHF/UHF Allmode
#define RIGCTRL_DEF_ADDR_IC_910         0x60 // VHF/UHF Allmode
#define RIGCTRL_DEF_ADDR_IC_970         0x2E // VHF/UHF Allmode
#define RIGCTRL_DEF_ADDR_IC_1271        0x24 // SHF
#define RIGCTRL_DEF_ADDR_IC_1275        0x18 // SHF
#define RIGCTRL_DEF_ADDR_IC_R10         0x52 // Handheld scanner
#define RIGCTRL_DEF_ADDR_IC_R20         0x6C // Handheld scanner
#define RIGCTRL_DEF_ADDR_IC_R71         0x1A // General Coverage Receiver
#define RIGCTRL_DEF_ADDR_IC_R72         0x32 // General Coverage Receiver
#define RIGCTRL_DEF_ADDR_IC_R75         0x5A // General Coverage Receiver
#define RIGCTRL_DEF_ADDR_IC_R7000       0x08 // General Coverage Receiver
#define RIGCTRL_DEF_ADDR_IC_R7100       0x34 // General Coverage Receiver
#define RIGCTRL_DEF_ADDR_IC_R8500       0x4A // Gen.Cov.RX
#define RIGCTRL_DEF_ADDR_IC_R8600       0x96 // DSP receiver, supports spectrum data via CI-V
#define RIGCTRL_DEF_ADDR_IC_R9000       0x2A // General Coverage Receiver
#define RIGCTRL_DEF_ADDR_IC_R9500       0x72 // General Coverage Receiver
#define RIGCTRL_DEF_ADDR_IC_RX7         0x78 // Handheld scanner

// "Unified" parameter numbers, used within RigControl.c but NOT (always) by Icom.
// Some of the parameters, transceiver settings, or current mode can be read
//   can be read via CI-V, e.g. command 0x1C for "transceiver status (RX/TX)".
// Some of the RIG-specific 4-digit subcodes for Icom's command 0x1A 0x05
//   can also be translated into one of these.
// Note: When adding new PNs here, also add them in RigCtrl_ParameterInfo[] !
#define RIGCTRL_PN_UNKNOWN                   0
#define RIGCTRL_PN_TRANSCEIVER_ID            1 // "Transceiver ID" alias "Default CI-V Address" : Icom: cmd 0x19, subcommand 0x00.
#define RIGCTRL_PN_FREQUENCY                 2 // current 'VFO' frequency in Hertz,  pRC->dblVfoFrequency
#define RIGCTRL_PN_OP_MODE                   3 // current 'operation mode' (USB/LSB/CW..) pRC->iOpMode
#define RIGCTRL_PN_FILTER_BANDWIDTH          4 // (IF-) filter bandwidth in Hz. Icom: cmd 0x1A, subcommand 0x03.
#define RIGCTRL_PN_DATA_MODE                 5 // 0=no data mode, 1=data mode.  Icom: cmd 0x1A, subcommand 0x06.


#define RIGCTRL_PN_TRANSMITTING             10 // 0=receiving, 1=transmitting. IC-9700: cmd 0x1C, subcommand 0x00 = "Send/read the transceiver's status".
#define RIGCTRL_PN_ATU_ENABLED              11 // 0=automatic antenna tuner disabled, 1=enabled. Icom: 0x017 0x01
#define RIGCTRL_PN_XFC_PRESSED              12 // 0="transmit frequency monitor" off, 1=on. Icom: 0x17 0x02.
                 // Guess this is what others call the "XFC button". Useful in SPLIT mode to monitor the own TX-frequency.
#define RIGCTRL_PN_TRANSMIT_FREQ            13 // Icom: 0x17 subcommand 0x03. They call it "Read transmit frequency".
                 // Guess the response contains the frequency displayed on the radio while pressing the 'XFC' button.
#define RIGCTRL_PN_REPEATER_TONE_FREQ       14 // Repeater Tone frequency in Hertz
#define RIGCTRL_PN_TONE_SQUELCH_FREQ        15 // Tone Squelch frequency in Hertz

#define RIGCTRL_PN_NUM_TX_BANDS             16 // Number of TRANSMIT bands supported by connected radio.
#define RIGCTRL_PN_TX_BAND_EDGES            17 // dummy represending T_RigCtrlFrequencyRange TxBandEdges[RIGCTRL_PN_TX_BAND_EDGES]
#define RIGCTRL_PN_BAND_STACKING_REGS       18 // dummy represending T_RigCtrlFreqMemEntry BandStackingRegs[RIGCTRL_NUM_BAND_STACKING_REGS]



        //--- SPECTRUM SCOPE related (with CI-V, exchanged via command 0x27) ---
#define RIGCTRL_PN_SCOPE_MODE               20 // pRC->iScopeMode
#define RIGCTRL_PN_SCOPE_SPAN               21 // pRC->dblScopeSpan_Hz
#define RIGCTRL_PN_SCOPE_ATTEN              22 // pRC->dblScopeAttenuator_dB
#define RIGCTRL_PN_SCOPE_REF_LEVEL          23 // pRC->dblScopeRefLevel_dB
#define RIGCTRL_PN_SCOPE_EDGE_NR            24 // pRC->iScopeEdgeNumber
#define RIGCTRL_PN_SCOPE_HOLD_FLAG          25 // pRC->iScopeHoldFlag
#define RIGCTRL_PN_SCOPE_SPEED              26 // pRC->iScopeSpeed (RIGCTRL_SCOPE_SPEED_FAST/MID/SLOW)

        //--- Various "poti" settings (with CI-V, exchanged via command 0x14) ---
#define RIGCTRL_PN_AUDIO_VOLUME_PERCENT     30 // pRC->iAudioVolume_Percent
#define RIGCTRL_PN_RF_GAIN_PERCENT          31 // pRC->iRFGain_Percent
#define RIGCTRL_PN_SQUELCH_LEVEL_PERCENT    32 // pRC->iSquelchLevel_Percent
#define RIGCTRL_PN_NOISE_REDUCTION_PERCENT  33 // pRC->iNoiseReduction_Percent
#define RIGCTRL_PN_PASSBAND_TUNING_POS1     34 // pRC->iPassbandTuningPos1; "inner ring" of "TWIN PBT", 50 % = center
#define RIGCTRL_PN_PASSBAND_TUNING_POS2     35 // pRC->iPassbandTuningPos2; "outer ring" of "TWIN PBT", 50 % = center
#define RIGCTRL_PN_CW_PITCH_HZ              36 // pRC->iCWPitch_Hz; here: in Hertz, not the strange CI-V internal unit
#define RIGCTRL_PN_RF_POWER_SETTING_PERCENT 37 // pRC->iRFPowerSetting_Percent;  "setting" to avoid confusion with the POWER METER MEASUREMENT !
#define RIGCTRL_PN_MIC_GAIN_PERCENT         38 // pRC->iMicGain_Percent
#define RIGCTRL_PN_KEYER_SPEED_PERCENT      39 // pRC->iKeyerSpeed_Percent;  0 = often 6 WPM, 100 % = 48 WPM (at least for Icom)
#define RIGCTRL_PN_NOTCH_POS_PERCENT        40 // pRC->iNotchPos_Percent;
#define RIGCTRL_PN_COMP_SETTING_PERCENT     41 // pRC->iCompSetting_Percent; audio compressor SETTING (not "measurement")
#define RIGCTRL_PN_BREAK_IN_DELAY_PERCENT   42 // pRC->iBreakInDelay_Percent;
#define RIGCTRL_PN_NBL_SETTING_PERCENT      43 // pRC->iNoiseBlankerSetting_Percent;
#define RIGCTRL_PN_MONITOR_GAIN_PERCENT     44 // pRC->iMonitorGain_Percent;
#define RIGCTRL_PN_VOX_GAIN_PERCENT         45 // pRC->iVoxGain_Percent;
#define RIGCTRL_PN_ANTI_VOX_GAIN_PERCENT    46 // pRC->iAntiVoxGain_Percent;
#define RIGCTRL_PN_BRIGHTNESS_PERCENT       47 // pRC->iBrightness_Percent;

        //------ Various "meter" readings (with CI-V, exchanged via command 0x15) -----
#define RIGCTRL_PN_SQUELCH_STATUS           50 // pRC->iSquelchStatus;  (0 or 1) from 0x1501
#define RIGCTRL_PN_S_METER_LEVEL_DB         51 // pRC->iSMeterLevel_dB; here: "decibel over S0",  not the strange CI-V internal unit
#define RIGCTRL_PN_POWER_METER_PERCENT      52 // pRC->iPowerMeterLevel_Percent; not the strange CI-V internal unit
#define RIGCTRL_PN_SWR_METER_VALUE          53 // pRC->dblSWRMeterValue;  dimensionless, usual range 1.0 (perfect) to 3.0 (poor)
#define RIGCTRL_PN_ALC_METER_LEVEL          54 // pRC->iALCMeterLevel;  here: 0 = min, 100 = max, not the strange CI-V internal unit
#define RIGCTRL_PN_COMP_METER_LEVEL_DB      55 // pRC->iCompMeterLevel_dB; here: in decibel, 0=no compression, 30 dB = maximum reading for IC-7300)
#define RIGCTRL_PN_SUPPLY_VOLTAGE_mV        56 // pRC->iSupplyVoltage_mV;  here: in millivolts, not the strange CI-V internal unit
#define RIGCTRL_PN_DRAIN_CURRENT_mA         57 // pRC->iDrainCurrent_mA;   here: in milliamps,  not the strange CI-V internal unit

        //-- extended Preamp, AGC speed, Noise blanker, Audio peak filter,
        // noise reduction, auto notch, etc. With CI-V, via command 0x16) -----
#define RIGCTRL_PN_PREAMP_SETTING           60 // CI-V : 0x16 0x02
#define RIGCTRL_PN_AGC_SPEED                61 // CI-V : 0x16 0x12
#define RIGCTRL_PN_NOISE_BLANKER_ON         62 // CI-V : 0x16 0x22
#define RIGCTRL_PN_NOISE_REDUCT_ON          63 // CI-V : 0x16 0x40
#define RIGCTRL_PN_AUTO_NOTCH_ON            64 // CI-V : 0x16 0x41
#define RIGCTRL_PN_RPTR_TONE_ON             65 // CI-V : 0x16 0x42
#define RIGCTRL_PN_TONE_SQUELCH_ON          66 // CI-V : 0x16 0x43
#define RIGCTRL_PN_COMP_ON                  67 // CI-V : 0x16 0x44
#define RIGCTRL_PN_MONITOR_ON               68 // CI-V : 0x16 0x45
#define RIGCTRL_PN_VOX_ON                   69 // CI-V : 0x16 0x46
#define RIGCTRL_PN_BK_IN_ON                 70 // CI-V : 0x16 0x47
#define RIGCTRL_PN_MANUAL_NOTCH_ON          71 // CI-V : 0x16 0x48
#define RIGCTRL_PN_TWIN_PEAK_FILTER_ON      72 // CI-V : 0x16 0x4F
#define RIGCTRL_PN_DIAL_LOCK_ON             73 // CI-V : 0x16 0x50
#define RIGCTRL_PN_DSP_SHARP_SOFT           74 // CI-V : 0x16 0x56
#define RIGCTRL_PN_MANUAL_NOTCH_WIDTH       75 // CI-V : 0x16 0x57
#define RIGCTRL_PN_SSB_TX_BANDWIDTH         76 // CI-V : 0x16 0x58



        //-- "Special" settings, audio equalizer, connector settings, etc -----
        // Many of these exchanged via command 0x1A with an awful number of subcodes.
#define RIGCTRL_PN_SSB_RX_HPF_LPF           80
#define RIGCTRL_PN_SSB_RX_BASS_LEVEL        81
#define RIGCTRL_PN_SSB_RX_TREBLE_LEVEL      82
#define RIGCTRL_PN_SSB_CW_SYNC_TUNING       83 // "SSB/CW synchronous tuning". In IC-7300, cmd 0x1A sub 5 PN #0052. Used by RS-BA1.
#define RIGCTRL_PN_CALIBRATION_MARKER       84 // "Calibration marker".  In IC-7300, cmd 0x1A sub 5 PN #0057. Used by RS-BA1.
#define RIGCTRL_PN_AF_OUTPUT_LEVEL_TO_ACC   85 // .. or "to ACC or USB". In IC-7300, cmd 0x1A sub 5 PN #0060
#define RIGCTRL_PN_SQUELCH_FOR_AF_ON_ACC    86 // "Squelch function for the AF signal output to ACC/USB"
#define RIGCTRL_PN_BEEP_AND_SPEECH_ON_ACC   87 // "Beep and speech output setting to ACC/USB"
#define RIGCTRL_PN_IF_OUTPUT_LEVEL_TO_ACC   88 // "IF signal output level to ACC/USB" (0..255 for 100%)
#define RIGCTRL_PN_MOD_INPUT_LEVEL_FROM_ACC 89 // "MOD input level from ACC" (0..255 for 100%)
#define RIGCTRL_PN_MOD_INPUT_LEVEL_FROM_USB 90 // "MOD input level from USB" (0..255 for 100%)
#define RIGCTRL_PN_MOD_IN_CONN_FOR_DATA_OFF 91 // "MOD input connector during DATA OFF" (0=MIC, 1=ACC, 2=MIC/ACC, 3=USB)
#define RIGCTRL_PN_MOD_IN_CONN_FOR_DATA     92 // "MOD input connector during DATA" (0=MIC, 1=ACC, 2=MIC/ACC, 3=USB)
#define RIGCTRL_PN_CIV_TRANSCEIVE           93
#define RIGCTRL_PN_CIV_ADDR_USB_TO_REMOTE   94
#define RIGCTRL_PN_CIV_OUTPUT_FOR_ANT       95 // Icom: 0x17 sub 4. They call it "CI-V output for ANT".
#define RIGCTRL_PN_CIV_UNLINK_FROM_REMOTE   96
#define RIGCTRL_PN_CIV_USB_ECHO             97 // "CI-V USB Echo Back" : 0075 in IC-7300, 0116 in IC-7610
#define RIGCTRL_PN_IP_PLUS                 100 // IC-7300 : 0x1A sub 0x07 (totally different in IC-7610)
#define RIGCTRL_PN_NTP_SERVER_ACCESS       101 // IC-7610 : 0x1A sub 0x07 (!)
#define RIGCTRL_PN_NTP_ACCESS_RESULT       102 // IC-7610 : 0x1A sub 0x08
#define RIGCTRL_PN_AF_MUTE                 103 // IC-7610 : 0x1A sub 0x09

        //-- Even more settings (in CI-V, controlled via command 0x0E..0x10, etc) --
#define RIGCTRL_PN_SCAN_MODE               110  // Icom: 0x0E sub 0
#define RIGCTRL_PN_SPLIT_MODE              111  // Icom: 0x0F (NO subcodes here)
#define RIGCTRL_PN_TUNING_STEP             112  // Icom: 0x10 (NO subcodes here)
#define RIGCTRL_PN_ATTENUATOR_DB           113  // Icom: 0x11 (NO subcodes here)

        // "RIT" / "XIT" related .. even WSJT-X wanted to play with these via Hamlib:
#define RIGCTRL_PN_RIT_FREQ    120 // 'Receiver Incremental Tuning' in Hertz.    IC-7300: Cmd 0x21, Sub 0x00
#define RIGCTRL_PN_RIT_ENABLED 121 // 'Receiver Incremental Tuning' ON/OFF.      IC-7300: Cmd 0x21, Sub 0x01
#define RIGCTRL_PN_XIT_ENABLED 122 // 'Transmitter Incremental Tuning' ON/OFF.   IC-7300: Cmd 0x21, Sub 0x02
#define RIGCTRL_PN_XIT_FREQ    123 // 'Transmitter Incremental Tuning' in Hertz. IC-7300: alias "delta TX", but:
        // (accidentally found in ?/Hamlib/SOURCE/rigs/icom/icom.c after wasting an hour on this):
        // > The Icom rigs have only one register for both RIT and Delta TX.
        // > you can turn one or both on -- but both end up just being in sync.
        // OMG. Who made this crazy decision ?



// Scaling units / physical units for the DISPLAY. Only defined HERE
//  (in DL4YHF's RigControl/RigControl.h) if not already defined
//   in FFT_FILTER_DLL.H or UTILITY1.H :
#ifndef  SCALE_UNIT_BLANK
# define SCALE_UNIT_BLANK   -2  // for UTL_FloatToTechNotation(): tech exponent but a BLANK unit (appended as string by caller)
# define SCALE_UNIT_NONE    -1
# define SCALE_UNIT_UNKNOWN  0  // in FiltCtrl.cpp : "amplitude unit as currently configured for the display" (dB or linear scale)
  // Voltages, Power units ...
# define SCALE_UNIT_NORM_V   1  // normalized ADC units (voltage), +/- 1.0  for full A/D converter input range
# define SCALE_UNIT_PERCENT  2  // linear, reference is "full scale" (100 % = ADC clipping)
# define SCALE_UNIT_VOLT     3  // voltage (V), 'calibrated' using UTL_dblAdcMaxInputVoltage
# define SCALE_UNIT_WATT     4  // power (W), 'calibrated' with UTL_dblAdcMaxInputVoltage + UTL_dblAdcInputImpedance
# define SCALE_UNIT_NORM_PWR 5  // power, normalized to 0...1 for full A/D converter input range
# define SCALE_UNIT_AMPERE   6  // current (A), just for completeness.. and CLI_ParseTechUnit()

  // Logarithmic units, dB, relative to "something" ...
# define SCALE_UNIT_MASK_dB 0x18  // mask to detect unit "dB" in all of its flavours..
# define SCALE_UNIT_dB      0x08  // dB , unknown reference level (including Icom's flavour for their 'Spectrum Scope')
# define SCALE_UNIT_dBfs    0x09  // dB , reference level is  "full scale" (0 dBfs = clipping point)
# define SCALE_UNIT_dBV     0x0A  // dB , reference level is 1 volt
# define SCALE_UNIT_dBuV    0x0B  // dB , reference level is 1 microvolt rms across any impedance (dBV)
# define SCALE_UNIT_dBm     0x0C  // dB , reference level is 1 milliwatt (often across 50 or 600 ohms)
  // Other UNIT CODES, which are NO amplitudes/magnitudes ...
# define SCALE_UNIT_DEGREE  0x10
# define SCALE_UNIT_HERTZ   0x11
# define SCALE_UNIT_SECONDS 0x12
# define SCALE_UNIT_TIME    0x13  // almost the same as SCALE_UNIT_SECONDS, but UTL_FloatToTechNotation() may decide to use MINUTES or HOURS for the display
# define SCALE_UNIT_SAMPLES_PER_SECOND 0x14 // technically the same as 'Hz' but used for SAMPLING RATES in combo lists
#endif // ndef SCALE_UNIT_BLANK ?


/*-------------- Data types ----------------------------------------------*/


typedef struct tRigCtrl_ParamInfo // -> const T_RigCtrl_ParamInfo RigCtrl_ParameterInfo[] in RigControl.c ...
{
  int iUnifiedPN;  // Unified parameter number (not Icom-specific), e.g. RIGCTRL_PN_TRANSMITTING.
                   // ZERO marks the end of the lookup table RigCtr_ParameterInfo[] .
  const char *pszToken;  // SHORT but descriptive, unique token like "transmitting" (used for access via interpreter)
  const char *pszUnit;   // physical unit where applicable, e.g. "Hz", "dB", "%", often empty ("") but not NULL.
  const T_SL_TokenList *pStringValueTable; // string/value lookup table for certain parameters,
                   // e.g. "CW","USB","LSB",.. for RIGCTRL_PN_OP_MODE . NULL for most other PNs.

  int  iByteOffsetIntoRigCtrlInstance; // hacky way to retrieve the address of e.g. T_RigCtrlInstance.iTransmitting,
                   // RELATIVE TO THE BEGIN of struct T_RigCtrlInstance, using offsetof().
  int  iRigCtrlDataType; // data type for acrobatic pointer arithmetics; may be one of the following:
#       define RIGCTRL_DT_UNKNOWN    0
#       define RIGCTRL_DT_INT        1
#       define RIGCTRL_DT_DOUBLE     2
#       define RIGCTRL_DT_STRING     3
#       define RIGCTRL_DT_FREQ_RANGE 4
#       define RIGCTRL_DT_OTHER    255
  // Note: When adding new data types above, consider supporting them
  //       in Remote_CW_Keyer\CwNet.c : CwNet_OnPoll(), too !
  // Parameters accessable via iByteOffsetIntoRigCtrlInstance
  //  will have the following values, as long as not set via CI-V parser or the application:
#       define RIGCTL_NOVALUE_INT    -2147483647  // == 0x80000001 = "almost the most negative integer" (ex: INVALID_INT_VALUE)
#       define RIGCTL_NOVALUE_DOUBLE -99e99       // ex: INVALID_DOUBLE_VALUE

  int  iAccessFlags;     // bitwise combination of the following:
#       define RIGCTRL_AF_READWRITE 0
#       define RIGCTRL_AF_READONLY  1  // only *readable* parameters, e.g. T_RigCtrlInstance.iSMeterLevel_dB



  // Info to read or write the parameter via Icom's CI-V protocol :
  const BYTE *pbCIVReadCmd; // byte-sequence for RigCtrl_SendAndLogMessage() to READ
              // this parameter via Icom CI-V protocol, beginning with
              // the "Command number", sometimes followed by a
              // "Sub command number".
              // The READ RESPONSE or WRITE COMMAND always has a
              // "Data area" . In some cases, even the READ REQUEST
              // has a "Data area", for example a "memory number".
              // NULL when the command format is "too special",
              //  and requires special treatment in an extra function.
              //  For example, cmd 0x03 READS the "operating frequency",
              //               cmd 0x05 WRITES the same parameter in CI-V.
  int  iCIVReadCmdLength; // length of the above sequence in bytes (w/o "Data Area") .
              // Note: strlen(pbCIVReadCmd) doen't work because
              //       the "sub command" (part of the byte sequence)
              //       may be 0x00 .
  int  iCIVDataType;      // Icom-specific data type :
#       define CIV_DTYPE_NONE  0   // dummy for "no data"
#       define CIV_DTYPE_BCD1  1   // BCD with only ONE digit (in the data byte bits 3..0)
#       define CIV_DTYPE_BCD2  2   // two-digit BCD (tens in the bytes bits 7..4, ones in bits 3..0)
#       define CIV_DTYPE_BCD3  3   // three-digit BCD (also least significant digit FIRST)
#       define CIV_DTYPE_BCD4  4   // FOUR-DIGIT BCD ..
#       define CIV_DTYPE_BCD5  5   // etc, etc, ...
#       define CIV_DTYPE_BCD6  6   // SIX-DIGIT BCD as used for "Duplex Offset frequency setting" in IC-9700
#       define CIV_DTYPE_BCD7  7
#       define CIV_DTYPE_BCD8  8
#       define CIV_DTYPE_BCD9  9
#       define CIV_DTYPE_BCD10 10  // TEN-digit BCD as used for VFO- and "Transmit"-frequencies in IC-7300, IC-9700, etc)
#       define CIV_DTYPE_BAND_EDGE_FREQUENCIES 11 // 2-digit "edge number" + 12..14-digit "Lower edge" + 0x2D as "Separator" + 12..14-digit "Higher edge"
#       define CIV_DTYPE_SUFFIX_BCD2_SIGN 0x80 // TWO-digit BCD suffic indicating "plus" (0x00) or "minus" (0x01)
   // Example for the above, from the "IC-7300 Full Manual", A-7292-4EX-6
   // (the -6 seems to be the document version): "Read transmit frequency" :
   // pbCIVReadCmd = { 0x1C, 0x03 },  iCIVReadCmdLength = 3,
   // iCIVDataType = CIV_DTYPE_BCD10


} T_RigCtrl_ParamInfo; // -> const T_RigCtrl_ParamInfo RigCtrl_ParameterInfo[] in RigControl.c

typedef struct tRigCtrlTrafficLogEntry // used by RigCtrl_AddToTrafficLog(), RigCtrl_ReadTrafficLog(), RigCtrl_ClearTrafficLog(), etc.
{
  long i32MessageNr;    // unique "message number", assigned in RigCtrl_AddToTrafficLog(), details there.
                        // Only incremented when logging PACKETS, for reasons explained in RigCtrl_AddToTrafficLog().
  double dblUnixTime;   // date and time of transmission or reception
  BYTE b80Message[80];  // first 80 bytes of the message data
  int  iMsgLength;      // full length of the message (including pre- and postamble). Zero if the T_RigCtrlTrafficLogEntry doesn't contain 'packet data'.
  int  iMsgType;        // e.g. RIGCTRL_MSGTYPE_FREQUENCY_REPORT (if the parser 'understood')
  int  iRigCtrlPort;    //  0=RIGCTRL_PORT_RADIO, 1=RIGCTRL_PORT_CLIENT_1, etc
# define RIGCTRL_MAX_COMMENT_LENGTH 120
  char szComment[ RIGCTRL_MAX_COMMENT_LENGTH+4 ]; // optional comment, possibly "human readable" message text
} T_RigCtrlTrafficLogEntry;
#define RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES 2048

typedef struct tRigCtrlFrequencyRange
{ double dblFmin_Hz;
  double dblFmax_Hz;
} T_RigCtrlFrequencyRange; // -> TxBandEdges[], ScopeFreqRange[], .. (?)
#define RIGCTRL_MAX_FIXED_EDGE_SCOPE_FREQ_RANGES 15 /* deliberately 2 more than used by IC-7300 */
#define RIGCTRL_MAX_FIXED_EDGES_PER_FREQ_RANGE    3


typedef struct tRigCtrlFreqMemEntry // .. for classic "memory channels" and Icom's "Band stacking register"
{ // What belongs to Icom's "Memory content" (for a single entry) can be seen
  // in the IC-9700 'CI-V Reference Guide', document #A7508-3EX-4, page 14:
  // Listed for "Command: 1A 00", but the same format is used for
  // the response to 0x1A 0x01 = "Read Band stacking register", too.
  // Except for the first few bytes (with two-digit BCD each), e.g.
  //  (1) = "Frequency band" [this first byte is completely missing in the IC-7300's "Memory content", see A7292-4EX-6 page "19-10"]
  //  (2)+(3) = "Memory channel number" [this also exists in an IC-7300, but there, in bytes (1)+(2).. bleah !
  //  (4) = "Select memory setting" (??),
  // everything is stored in a struct member that resembles the name by Icom:
  struct // sub-struct because bytes (5)..(51) "white" are duplicated for SPLIT MODE [(5)..(51) "black"]
   { double dblOperatingFreq_Hz;  // (5)..(9)  [IC-7300: (4)..(8) ]
     int    iOpMode;     // IC-9700: (10)..(11), but Icom's code it translated into RIGCTRL_OPMODE_CW/LSB/USB/etc
                         // IC-7300: "Operating mode setting" in bytes (9)+(10), but "(10") is the FILTER NUMBER
     BOOL   fDataMode;   // IC-9700: (12) "Data mode setting" : FALSE=Data mode OFF, TRUE=Data mode ON.
     int    iDuplexMode; // IC-9700: (13, upper BCD digit) : 0=not Duplex, 1=Duplex-, 2=Duplex+, 3="RPS"(?)
     int    iToneType;   // IC-9700: (13, lower BCD digit) : 0=off, 1=TONE, 2=TSQL, 3=DTCS
                         // IC-7300: (11) = "Data mode       and  tone type settings"
                         //                 |_upper digit_|       |___lower digit__|
#            define RIGCTRL_TONE_TYPE_OFF  0
#            define RIGCTRL_TONE_TYPE_TONE 1  // guess what they mean by this is the 1750 Hz repeater tone (?)
#            define RIGCTRL_TONE_TYPE_TSQL 2  // guess what they mean by this is a sub-audio "Tone Squelch" (?)
#            define RIGCTRL_TONE_TYPE_DTCS 3  // seen in an IC-9700. Guess this is what the rest of the world calls "DCS" for "Digital-Coded Squelch" (except for ICOM)

     int    iDigitalSquelch; // IC-9700: (14, only the UPPER BCD DIGIT in use): 0=off, 1=Digital Call Sign Squelch (DSQL), 2=Digital Code Squelch (CSQL)
     int    iRepeaterToneFreq_Hz; // IC-9700: (15)..(17), IC-7300: (12)..(14)
     int    iToneSquelchFreq_Hz;  // IC-9700: (18)..(20), IC-7300: (15)..(17)
        // The above ("Tone squelch frequency setting") is the last common item
        // that exists in an IC-9700 *and* an IC-7300 .. but at different byte-offsets. AAAARGH !
        // You will have a lot of fun taking care of all these annoying little differences
        // when writing a parser for ALL Icom radios, since there are no
        // "separator characters" that would help to identify different fields.
        // Cramming all this stuff into stupid, unformatted BCD (Binary-Coded Decimal)
        // is an anachronism, but WE cannot change this, and Icom won't switch
        // to an easily parsable format or even something JSON-ish, or at least
        // a "memory content" consisting of key=value pairs.
     int    iDTCS_Code;  // (21)..(23) "DTCS code settings" (never used, and most likely never really supported here)
     int    iDV_DigitalCodeSquelch; // (24) .. will share the same fate as the "DTCS", whatever that is
     int    iDuplexOffsetFreq_Hz;   // (25)..(27)
     char   c8DestCall[8+4];        // (28)..(35) "UR (Destination) call sign setting, 8 chars, fixed" (here: plus trailing zero for "C")
     char   c8AccessRptrCall[8+4];  // (36)..(43) "R1 (Access repeater) call sign setting, 8 chars, fixed" (here: plus trailing zero for "C")
     char   c8GatewayRptrCall[8+4]; // (44)..(51) "R2 (Gateway/Link repeater) call sign setting, 8 chars, fixed" (here: plus trailing zero for "C")
   } RxTx[2];  // [0] for the RECEIVER, [1] for the TRANSMITTER (TX settings "when the Split function is ON")
  char c16MemoryName[16+4]; // (52)..(67) "Memory name setting, 16 chars, fixed" (here: plus trailing zero for "C")
       // '--> the MEMORY NAME ("channel name") is only stored ONCE .
  // From A7508-3EX-4 page 14:
  // > NOTE:
  // >  * The same data as (5) ~ (51) "white" are stored in (5) ~ (51) "black".
  // >  * When the Split function is ON, the data of (5) ~ (51) "black"
  // >    is used for transmit.
  // >  * Even if the Split function is OFF, enter the data into (5) ~ (51) "black"
  // >    to match your transceiver. We recommend that you set the same data
  // >    as (5) ~ (51) "black".
  //    WB: "black" are the circled numbers with a BLACK BACKGROUND, aka "inverse".
  //        Very unfriendly for developers using plain text editors for code.
  //        For the BAND STACKING REGISTERS, "(5) ~ (51) black" seem to be
  //        non-existent, as Icom wrote in A7508-3EX-4 page 15:
  // > NOTE:  (WB: about the "Band stacking register", cmd 0x1A 0x01) :
  // >    When sending the contents, the codes, such as operating frequency
  // >    and operating mode*, should be added after the frequency band code
  // >    and the register code, as shown below.
  // >  * See (5) to (51) [WB: "white"] on 'Memory content setting.' (p. 14)

} T_RigCtrlFreqMemEntry;


typedef struct tRigCtrlSpectrum
{ // Simplified variant of T_SPECTRUM in C:\CBproj\SpecLab\TSpectrum.h,
  // to eliminate as many dependencies from non-standard headers as possible.
  // Since 2019-01, also used in Spectrum Lab's built-in OpenWebRX server,
  //        see c:\cbproj\SpecLab\http_server_sourcecode\OpenWebRX_Server.c .
  // Doesn't use dynamically allocated memory to keep this module suitable
  // for embedded systems (microcontrollers without an operating system) .
  //
  //     Because Icom radios (and most likely all similar radios with
  //                          with built-in waterfall or spectrum display)
  //     only spit out amplitude spectra, keep it simple, and keep everything
  //     in the same "decibel" unit (possibly relative to the receiver's
  //     noise floor.. ?) as used by the radio itself.
  //     Attenuator-, preamplifier- or reference level settings are considered
  //     when READING a T_RigCtrlSpectrum from the radio. There's no need
  //     to store such information in the header.
  int    nBinsUsed;      // may range from 1 to RIGCTRL_MAX_FREQ_BINS_PER_SPECTRUM
  double dblBinWidth_Hz; // 'width' of a single FFT bin in Hertz (=resolution
  double dblFmin_Hz;     // "radio frequency" stored in the first frequency bin.
     //  dblFmax_Hz would be dblFmin_Hz + (nBinsUsed - 1) * dblBinWidth_Hz !
  double dblCenterFreq_Hz;
      // THE VFO FREQUENCY IS NOT ALWAYS EQUAL TO the above "center" frequency !
      // That's why some applications need to know the current 'fixed/center' setting:
  BOOL   fFixedEdgeMode; // TRUE=fixed frequency edges, FALSE=centered on VFO-frequency
  double dblUnixTime;    // timestamp in UNIX format (second elapsed since 1970-01-01 00:00:00 UTC).
         // If the SOURCE doesn't support this, use the system time in UTC instead.
#define RIGCTRL_MAX_FREQ_BINS_PER_SPECTRUM 1024 // at least as many frequency bins
         // as the width of the spectrum display in the "large" Icom radio.
         // For example, an IC-7851 sends 704-byte-messages via LAN ->
         // a MAXIMUM of 1024 frequency bins per spectrum should be enough.
  float  fltMagnitudes_dB[ RIGCTRL_MAX_FREQ_BINS_PER_SPECTRUM ];
  int    iAmplitudeUnit; // e.g. SCALE_UNIT_dB = "Icom's dB scale where ZERO is near the NOISE FLOOR" (below "S0")
  float  fltAmplOffset_dBm; // number to add to fltMagnitudes_dB[] to convert into dBm (dB "over" one milliwatt)

} T_RigCtrl_Spectrum;


typedef struct tRigCtrl_MsgFilter // .. for client message forwarding
{
  int iCmd;      // [in] CI-V main command like 0x03 = "Read operating frequency" (response on request)
  int iSubCmd;   // [in] optional sub-command. -1 to omit or ignore.
  int iDataB1;   // [in] optional 1st parameter byte. -1 to omit or ignore.
  int iDataB2;   // [in] optional 2nd parameter byte. -1 to omit or ignore.
  int iMsgTypes; // [in] additional message types like RIGCTRL_MSGTYPE_OK | RIGCTRL_MSGTYPE_NOT_OK
} T_RigCtrl_MsgFilter;

typedef struct tRigCtrl_PortInstance // instance data for forwarding data to/from a radio for other applications
{
  // Multiple of these serial port 'duplicators' may be used to forward requests
  //    and responses from other applications (CLIENTS) that need to control the radio,
  //    e.g. WSJT-X .
  //    Usually connected via virtual "NULL-modem" cable (extra COM ports created by com0com).
  //
  // Only T_RigCtrlInstance.PortInstance[RIGCTRL_PORT_RADIO] is the instance "talking to the radio".
  //
#define RIGCTRL_MAX_MESSAGE_LENGTH 2048
  BYTE bRcvdMsg[RIGCTRL_MAX_MESSAGE_LENGTH];
  int  iRcvdMsgLength;    // index into bRcvdMsg[], with post-increment
  BYTE bTxMsg[RIGCTRL_MAX_MESSAGE_LENGTH];
  int  iTxMsgLength;    // length of transmitted message, as usual including pre- and postamble
  BOOL fLastSendFailed; // to avoid flooding log with "TX error" messages when a client (like RS-BA1) goes away
  BYTE bPrevRcvdByte;
  int  iClientOptions;  // bitwise combination of the following flags :
#define RIGCTRL_CLIENT_OPTION_REJECT_UNSOLICITED_MSGS 0x0001
#define RIGCTRL_CLIENT_OPTION_EMULATE_CIV_ECHO        0x0002
  // From a 3rd party client's point of view, the 'local CI-V echo'
  // can be modified via CI-V itself, to emulate the behaviour of modern rigs
  // where CI-V isn't a simple SINGLE-WIRE interface ("always with an echo")
  // but an USB port or LAN with (logically) separated "TxD" and "RxD" lines,
  // which eliminates the need for collision detection on the physical layer.
  // The following flag controls (and indicates) if the 'local CI-V echo'
  // is currently active:
  BOOL fEmulateCIVEcho;  // "initial setting" copied from iClientOptions !

  // Currently registered responses (expected by the client on a certain port,
  //   used by RigCtrl_RegisterReponseForClient) :
#define RIGCTRL_MAX_REGISTERED_MESSAGES_PER_CLIENT 32 /* who knows.. without buying "RS-BA1" V2 */
  T_RigCtrl_MsgFilter RegisteredMessages[RIGCTRL_MAX_REGISTERED_MESSAGES_PER_CLIENT];

  // For debugging / troubleshooting / statistics in SL's "Debug"-window :
  DWORD dwNumGarbageBytes;
  DWORD dwNumMessagesSent, dwNumMessagesRcvd;

} T_RigCtrl_PortInstance;


typedef struct tRigCtrlInstance
{
  // "Config data". Usually set once (in or shortly after RigCtrl_Init() ):
  int iRadioDeviceAddr;   // "to" transceiver's address (0x48 = IC706)
  int iRadioMasterAddr;   // "from" controller's address (often 0xE0)
  int iRadioCtrlProtocol; // RIGCTRL_PROTOCOL_ICOM_CI_V, possibly more one day
  int iRigControlFlags;   // RIGCTRL_FLAG_SEND_SETTINGS_ON_STARTUP,
                          // RIGCTRL_FLAG_WANT_SPECTRUM_DATA, etc (bit combination)
  int iCapabilities;      // RIGCTRL_CAPS_POWER_ON_OFF_VIA_CAT, RIGCTRL_CAPS_SPECTRUM_VIA_CAT, etc (bit combination)
  DWORD dwAvailableBands; // bitwise combination of RIGCTRL_BAND_...,
                          //  initialized in RigCtrl_Init() but restricted/revised later
  int iMaxResponseDelay_ms; // maximum delay for getting a response, in milliseconds
  int iMaxTimeToSendMsg_ms; // maximum time spent waiting for message-transmission, in milliseconds
  T_CFIFO *pRadioPortTxFifo; // thread-safe circular FIFO to TRANSMIT data to the radio
  T_CFIFO *pRadioPortRxFifo; // thread-safe circular FIFO to RECEIVE data from the radio

  uint32_t u32RxAudioSamplesPerSecond; // for Icom transceivers with AUDIO VIA UDP, possibly fixed to 48 kHz !
  uint32_t u32TxAudioSamplesPerSecond;
  BOOL fReceiveOnlyMode; // FALSE="active mode" (where THIS INSTANCE may send commands),
                         // TRUE ="receive-only" (this instance MUST NOT send anything).
  BOOL fLocalPTTFlag; // PTT flag controlled by the "local application" (e.g. the keyer thread).
                      // Not to be confused with T_RigCtrlInstance.iTransmitting,
                      // which is the last known 'PTT'- aka 'Transmit'-flag
                      // parsed from a report from a received CI-V message
                      // with cmd=0x1C, subCmd=0x00, "Send/read the transceiver's status".
                      // (the transmitter may transmit due to 'local operator control',
                      //  e.g. PTT on the microphone pressed,
                      //       Morse key plugged in and operated with Semi-BK on, etc.

  // Variables exchanged with the remote rig (and a few "update counters" for them):
  int iDefaultAddress;   // For Icom radios, the DEFAULT CI-V ADDRESS also identifies the model !
                         // RigControl.c tries to READ THIS FROM THE RADIO,
                         // because e.g. iRadioDeviceAddr(!) = 0x48 doesn't mean it's an IC-706.
                         // Only iDefaultAddress(!) = 0x48 means "it's REALLY an IC-706.
  double dblVfoFrequency; // current 'VFO' frequency (interpretation depends on the radio.
                         // Most use the (suppressed) carrier frequency for USB + LSB,
                         // some always use the center frequency (even for USB + LSB).
                         // Accessable as parameter # RIGCTRL_PN_FREQUENCY .
  double dblTxFrequency; // frequency (in Hz) that WOULD BE USED for transmission.
                         // Accessable as parameter # RIGCTRL_PN_TX_FREQUENCY .

  int    iFrequencyModifiedByRadio_cnt; // counter incremented whenever .dblFrequency was modified BY THE RADIO,
                                        // for example after turning the VFO knob. Polled in a GUI-thread somewhere.
  int    iOpMode; // RIGCTRL_OPMODE_CW/LSB/USB/.., combineable with _REVERSE, _NARROW, etc.
                       // Accessable as parameter # RIGCTRL_PN_OP_MODE .
  double dblDemodBasebandFreqOffset, dblDemodBasebandFreqFactor; // f(iOpMode), see RigCtrl_UpdateBasebandFrequencyConversionParams()
  int    iFilterBW_Hz; // may be readable via cmd 0x1A, subcmd 0x03 (polled by WSJT-X)
  int    iDataMode;    // 0=no data mode, 1=data mode. Icom: cmd 0x1A, subcommand 0x06 ("").
                       // Accessable as parameter # RIGCTRL_PN_DATA_MODE .
# define RIGCTRL_NUM_TX_BAND_EDGES 32
  T_RigCtrlFrequencyRange TxBandEdges[RIGCTRL_NUM_TX_BAND_EDGES]; // read from THE RIG ...
  int    iNumTxBands;  // number of valid entries in TxBandEdges[]; read via CI-V cmd 0x1E subcmd 0x00
  BOOL   fTxBandsModified; // <- application may POLL and CLEAR this flag after taking notice
  BOOL   fTxBandsModifiedForCwNet; // <- similar flag to be "consumed" by CwNet.c,
          // when sending the radio's transmit-band-edges from server to clients.

# define RIGCTRL_NUM_BAND_STACKING_REGS 64 // Consider an HF/VHF/UHF "all-mode": (10+3+2) bands * 3 "stack levels" per band plus a few entries for what Icom calls "GENE" (generic frequencies)
  T_RigCtrlFreqMemEntry BandStackingRegs[RIGCTRL_NUM_BAND_STACKING_REGS];
  int    iNumBandStackingRegs; // number of valid entries in BandStackingRegs[]; determined in state RIGCTRL_POLLSTATE_BAND_STACKING_REGS
  BOOL   fBandStackingRegsModified;         // <- application may POLL and CLEAR this flag after taking notice
  BOOL   fBandStackingRegsModifiedForCwNet; // <- similar flag to be "consumed" by CwNet.c,
          // when sending band stacking registers from server to clients.


  //------ SPECTRUM SCOPE related (with CI-V, exchanged via command 0x27) -----
  int    iScopeMode;      // HERE, independently of the protocol, one of the following:
#        define RIGCTRL_SCOPE_MODE_OFF    0  // spectrum scope currently off, or not available
#        define RIGCTRL_SCOPE_MODE_FIXED  1  // "fixed frequency range"
#        define RIGCTRL_SCOPE_MODE_CENTER 2  // "centered around the VFO frequency"
  double dblScopeSpan_Hz;       // last value reported for Icom's "Scope span settings" (cmd 0x27 0x15)
  double dblScopeAttenuator_dB; // "Scope Attenuator setting" (IC-7851, n/a on IC-7300)
  double dblScopeRefLevel_dB;   // "Scope Reference level setting" in dB
  double dblSpectrumCenterFreq_Hz;
  int    iScopeEdgeNumber; // "Edge number setting in the Fixed mode Scope", runs from 1 to 3 FOR EACH BAND
  int    iScopeHoldFlag;   // "Scope hold function status" : 0=hold OFF, 1=hold ON
  T_RigCtrlFrequencyRange ScopeFreqRange[RIGCTRL_MAX_FIXED_EDGE_SCOPE_FREQ_RANGES][RIGCTRL_MAX_FIXED_EDGES_PER_FREQ_RANGE];
  int    iScopeNumFreqRanges;
  int    iScopeNumEdgesPerFreqRange;
  int    iScopeDataCountdown_ms; // timer to monitor periodic 'scope data' reception
#        define RIGCTRL_SPECTRUM_TIMEOUT_MS 6000 /* large enough to tolerate "interference" from RS-BA1 */
  int    iScopeSpeed;      // may contain one of the following :
#        define RIGCTRL_SCOPE_SPEED_FAST 0x00
#        define RIGCTRL_SCOPE_SPEED_MID  0x01
#        define RIGCTRL_SCOPE_SPEED_SLOW 0x02
  int    iScopeCenterMode; // may contain one of the following:
#        define RIGCTRL_SCOPE_CM_FILTER_CENTER 0
#        define RIGCTRL_SCOPE_CM_CARRIER_POINT 1
#        define RIGCTRL_SCOPE_CM_CARRIER_ABS   2 // .. whatever that means !
  int    iScopeVideoBW;    // may contain one of the following:
#        define RIGCTRL_SCOPE_VBW_NARROW 0
#        define RIGCTRL_SCOPE_VBW_WIDE   1


  //--- "Transceiver status" (RX/TX). With CI-V, exchanged via command 0x1C ---
  int    iTransmitting;  // 0=receive, 1=transmit.  With CI-V, parsed from cmd 0x1C00.
  int    iATU_enabled;   // 0=automatic antenna tuner disabled, 1=enabled
  int    iXFC_pressed;   // 0='XFC'-button not pressed, 1=button pressed
  double dblRepeaterTone_Hz; // "repeater tone frequency". Icom: 0x1B 0x00 .
  double dblToneSquelch_Hz;  // "TSQL tone frequency".     Icom: 0x1B 0x01 .

  //------ Various "poti" settings (with CI-V, exchanged via command 0x14) ----
  int    iAudioVolume_Percent;
  int    iRFGain_Percent;
  int    iSquelchLevel_Percent;
  int    iNoiseReduction_Percent;
  int    iPassbandTuningPos1;  // "inner ring" of "TWIN PBT", 50 % = center
  int    iPassbandTuningPos2;  // "outer ring" of "TWIN PBT", 50 % = center
  int    iCWPitch_Hz; // here: in Hertz, not the strange CI-V internal unit
  int    iRFPowerSetting_Percent;
  int    iMicGain_Percent;
  int    iKeyerSpeed_Percent;  // 0 = often 6 WPM, 100 % = 48 WPM
  int    iNotchPos_Percent;
  int    iCompSetting_Percent; // audio compressor SETTING (not "measurement")
  int    iBreakInDelay_Percent;
  int    iNoiseBlankerSetting_Percent;
  int    iMonitorGain_Percent;
  int    iVoxGain_Percent;
  int    iAntiVoxGain_Percent;
  int    iBrightness_Percent;


  //------ Various "meter" readings (with CI-V, exchanged via command 0x15) -----
  int    iSquelchStatus;           // (0 or 1) from 0x1501
  int    iSMeterLevel_dB;          // here: "decibel over S0",  not the strange CI-V internal unit
  int    iPowerMeterLevel_pcnt;    // *measured* output power in PERCENT, not the strange CI-V internal unit
  double dblSWRMeterValue;         // dimensionless "SWR" value, usual range 1.0 (perfect) to 3.0 (poor)
  int    iALCMeterLevel_pcnt;      // here: 0 = min, 100 = max, not the strange CI-V internal unit
  int    iCompMeterLevel_dB;       // here: in decibel, 0=no compression, 30 dB = maximum reading for IC-7300)
  int    iSupplyVoltage_mV;        // here: in millivolts, not the strange CI-V internal unit (241 = 16 V)
  int    iDrainCurrent_mA;         // here: in milliamps,  not the strange CI-V internal unit (241 = 25 A)
  int    iPATemperature_degC;      // power amplifier temperature in C (not sure how to read this via CI-V yet...)


  //-- Preamp step, AGC speed, Noise blanker, Audio peak filter,
  //   noise reduction, auto notch, etc. With CI-V, via command 0x16) -----
  int    iPreampSet;        // RIGCTRL_PN_PREAMP_SETTING,       CI-V : 0x16 0x02
  int    iAGCSpeed;         // RIGCTRL_PN_AGC_SPEED,            CI-V : 0x16 0x12
  int    iNoiseBlankerON;   // RIGCTRL_PN_NOISE_BLANKER_ON,     CI-V : 0x16 0x22
  int    iNoiseReductON;    // RIGCTRL_PN_NOISE_REDUCT_ON,      CI-V : 0x16 0x40
  int    iAutoNotchON;      // RIGCTRL_PN_AUTO_NOTCH_ON,        CI-V : 0x16 0x41
  int    iRptrToneON;       // RIGCTRL_PN_RPTR_TONE_ON,         CI-V : 0x16 0x42
  int    iToneSquelchON;    // RIGCTRL_PN_TONE_SQUELCH_ON,      CI-V : 0x16 0x43
  int    iCompressON;       // RIGCTRL_PN_COMP_ON,              CI-V : 0x16 0x44
  int    iMonitorON;        // RIGCTRL_PN_MONITOR_ON,           CI-V : 0x16 0x45
  int    iVoxON;            // RIGCTRL_PN_VOX_ON,               CI-V : 0x16 0x46
  int    iBreakInON;        // RIGCTRL_PN_BK_IN_ON,             CI-V : 0x16 0x47
  int    iManualNotchON;    // RIGCTRL_PN_MANUAL_NOTCH_ON,      CI-V : 0x16 0x48
  int    iTwinPeakFilterON; // RIGCTRL_PN_TWIN_PEAK_FILTER_ON,  CI-V : 0x16 0x4F
  int    iDialLockON;       // RIGCTRL_PN_DIAL_LOCK_ON,         CI-V : 0x16 0x50
  int    iDSP_SharpSoft;    // RIGCTRL_PN_DSP_SHARP_SOFT,       CI-V : 0x16 0x56
  int    iManualNotchWidth; // RIGCTRL_PN_MANUAL_NOTCH_WIDTH,   CI-V : 0x16 0x57
  int    iSSB_Tx_BW;        // RIGCTRL_PN_SSB_TX_BANDWIDTH,     CI-V : 0x16 0x58



  //---- Various parameters accessed via RIG-SPECIFIC(!) 4-digit subcodes -----
  //    ...after Icom's CI-V command 0x1A 0x05. These 4-digit subcodes
  //       must be translated into PRIGCTRL_PN_xyz by
  //       RigCtrl_SubcodeFrom1A05_to_UnifiedParameterNumber() !
  int    iSSB_RX_HPF_LPF;
  int    iSSB_RX_BASS_LEVEL;
  int    iSSB_RX_TREBLE_LEVEL;
  int    iSSB_CW_sync_tuning;
  int    iCalibMarker;
  int    iAF2ACC_level;
  int    iSquelch_for_ACC;
  int    iBeep_on_ACC;
  int    iIF_level_on_ACC;
  int    iMOD_level_ACC;
  int    iMOD_level_USB;
  int    iMOD_conn_NoData;
  int    iMOD_conn_Data;
  int    iCIV_transceive;
  int    iCIV_addr_USB_Rem;
  int    iCIV_for_ANT;
  int    iCIV_unlink_Remote;
  int    iCIV_USB_echo;
  int    iIP_Plus;
  int    iNTP_Server_Access, iNTP_Access_Result;
  int    iAF_mute;

  //-- Even more settings (in CI-V, controlled via command 0x0E..0x10) --
  int    iScanMode;
  int    iSplitMode;     // one of the following :
#        define RIGCTRL_SPLIT_MODE_OFF      0 // neither "Split" not "Duplex for Repeater operation"
#        define RIGCTRL_SPLIT_MODE_ON       1 // classic "Split" mode, usually a few kHz
#        define RIGCTRL_SPLIT_MODE_DUP_NEG  2 // "DUP- operation" (IC-9700)
#        define RIGCTRL_SPLIT_MODE_DUP_POS  3 // "DUP+ operation" (IC-9700)
#        define RIGCTRL_SPLIT_MODE_REPEATER_SIMPLEX 4 // <- IC-9700 extravaganza ? "DD Repeater Simplex mode (RPS)"
  int    iTuningStep_Hz;
  int    iAttenuator_dB;
  int    iRIT_freq_Hz;   // RIGCTRL_PN_RIT_FREQ, 'Receiver Incremental Tuning' in Hertz.
  int    iXIT_freq_Hz;   // RIGCTRL_PN_XIT_FREQ, 'Transmitter Incremental Tuning' in Hertz.
  int    iRIT_ON;        // RIGCTRL_PN_RIT_ENABLED, 'Receiver Incremental Tuning' ON(1) / OFF(0).
  int    iXIT_ON;        // RIGCTRL_PN_XIT_ENABLED, 'Transmitter Incremental Tuning' ON(1) / OFF(0).

  // GPS data (from radios with built-in GPS receivers)
  double dblGpsLat_deg, dblGpsLon_deg;

  // Lock-free buffer (circular FIFO) for a few spectra stored by RigControl .
  // There may(!) be MULTIPLE READERS, each of them has his own "head index"
  // but there is only one "tail index". If head_index == tail_index,
  // there are no new spectra available for the display. The entry at <head_index>
  // should not be accessed (read) by the application because RigControl.c
  // may already be "filling" it in another thread or task.
# define RIGCTRL_SPECTRUM_FIFO_SIZE 4
  T_RigCtrl_Spectrum SpectrumFifo[ RIGCTRL_SPECTRUM_FIFO_SIZE ];
  int    nSpectrumBinsUsed;    // number of frequency bins acually used in the LASTEST entry in SpectrumFifo[]
  int    iScopeFifoHeadIndex;  // index for WRITING the next entry into SpectrumFifo[]
  BOOL   fScopeFifoFull;  // Set to TRUE when iScopeFifoHeadIndex has wrapped around.
         // Then all entries (except the one at iScopeFifoHeadIndex) are valid.
  int    iScopeBinCounter; // frequency bin counter for assembling 'fragments' of spectra
  T_RigCtrlFrequencyRange ScopeFreqRangeFromFirstFragment;

  // Internal states, used for reading a sequence of parameters from the "radio" port, etc..
  enum // tracks calls of RigCtrl_Init() and RigCtrl_Close().
   { initState_ColdStart = 0,    // didn't call RigCtrl_Init() yet
     initState_Opened,           // last call was RigCtrl_Init() or RigCtrl_ReConnect(); allow processing UDP packets or chunks of CI-V
     initState_Closed            // last call was RigCtrl_Close()
   } initState;
  int    iParameterPollingState; // RIGCTRL_POLLSTATE_PASSIVE, RIGCTRL_POLLSTATE_START_RD, etc. Used shortly after connecting, to poll the most important parameters
  int    iParameterPollingSubState;  // sub-index or similar, depends on the above
  int    iParameterPollingArrayIndex; // index for reading e.g. TWELVE "Fixed edge frequency settings",
                                      //  or THREE "Stack levels" per Band Stacking Register, etc, etc.
  long   i32ExpectedResponse_CmdAndSubcode; // e.g. set to 0x03 when asking for a VFO frequency via CI-V, -1 when NOT expecting a certain response.
                                 // Also set (>=0) when sending a command TO THE RADIO PORT on behalf of a client !
                                 // iExpectedResponseCmd is only "valid" as long as iResponseCountdown_ms is nonzero.
                                 // (but it doesn't hurt to set iExpectedResponseCmd to -1 when the response arrives,
                                 //  to avoid MISINTERPRETING the next unspecific "OK" or "NOT OK" from the radio)
         // Macros to assemble i32ExpectedResponse_CmdAndSubcode for CI-V commands
         // WITHOUT a "Sub cmd." (that's the column title in Icom's "Command table"),
         // with a SINGLE-BYTE "Sub cmd." (as in 0x1A 0x04),
         //   or a  THREE-BYTE "Sub cmd." (as in 0x1A 0x05 with HUNDREDS of four-digit indices)
# define RCTL_COMBINE_CMD_ONLY(cmd)                ((long)cmd<<24)
# define RCTL_COMBINE_CMD_AND_SUB(cmd,sub)         (((long)cmd<<24)|((long)sub<<16))
# define RCTL_COMBINE_CMD_SUB_AND_IDX(cmd,sub,idx) (((long)cmd<<24)|((long)sub<<16)|((long)idx))

  int    iExpectReadResponseForUnifiedPN; // more 'universal' replacement for i32ExpectedResponse_CmdAndSubcode .
                                 // Would contain e.g. RIGCTRL_PN_FREQUENCY after asking the radio for the VFO frequency,
                                 // regardless of the PROTOCOL in use .
  int    iResponseCountdown_ms;  // timer to skip parameters if the radio doesn't answer.
                                 // As long as iResponseCountdown_ms is positive,
                                 // periodic polling (also from RigCtrl_SendReadCommandFromFIFO() )
                                 // is disabled, but requests from e.g. RigCtrl_SendCmdToReadUnifiedPN()
                                 // can still be queued up in iReadPNFifo[] .
  int    numResponseTimeouts;    // Number of response timeouts in a row. Too many of these cause an attempt to turn the radio on again, etc.
  int    iPostponedSetMessages;  // bitwise combination of (1<<RIGCTRL_MSGTYPE_RADIO_ID) .. (1<<RIGCTRL_MSGTYPE_BAND_SCOPE).
                                 // Only applies to "Set"-commands for the LOCALLY CONNECTED RADIO, i.e. on the server side.
                                 // Setting a flag in iPostponedSetMessages does not cause transmissions via CwNet.c (TCP/IP).
  double dblPostponedFrequency;  // VFO frequency that RigCtrl_SetVFOFrequency() should set but couldn't
  int    iPostponedOpMode;       // .. etc, etc, (parameters pending to be SET)
# define RIGCTRL_READ_PN_FIFO_SIZE 4 // FIFO for RigCtrl_SendCmdToReadUnifiedPN()
  int    iReadPNFifo[RIGCTRL_READ_PN_FIFO_SIZE]; // FIFO to queue up parameter numbers to read
  int    iReadPNFifoHead, iReadPNFifoTail; // circular buffer indices into iReadPNFifo[]


  // Multiple *SERIAL* port instances, to allow 3rd party applications
  // to control the same radio over a serial port or 'RIGCTLD'-compatible TCP/IP.
  // Spectrum Lab (or the Remote CW Keyer) lets those clients think
  // they talk to the radio "directly" (via virtual NULL-modem like com0com):
  T_RigCtrl_PortInstance PortInstance[1+RIGCTRL_MAX_CLIENT_PORTS];
    // Index into the above array is iRigControlPort :
    //        0 = RIGCTRL_PORT_RADIO, 1 = RIGCTRL_PORT_CLIENT_1, ... !
    // Note : To keep those ugly details about accessing serial ports
    //        under windows out of this sourcecode module, the windows-specific
    //        "COM port settings" (and the COM port interface itself) are NOT
    //        part of RigControl.c / RigControl.h !


  // Transmit FIFO for the RADIO port. Used for buffering messages that
  // need to be forwarded from CLIENTS(!) to the radio, if the RADIO port
  // is currently occupied (radio still busy from a previously transmitted
  // command, for which the response is still pending).
  // Filled out  by RigCtrl_AppendMessageToRadioPortTxFifo(),
  // and drained by RigCtrl_SendMessageFromRadioPortTxFifo().
#define RIGCTL_RADIO_PORT_TX_FIFO_SIZE 2048 /* # bytes, must be a power of two */
  BYTE bRadioPortTxFifo[RIGCTL_RADIO_PORT_TX_FIFO_SIZE];
  int  iRadioPortTxFifoHead, iRadioPortTxFifoTail;

  // Debugging stuff, including a FIFO for spectrum lab's "CAT traffic monitor":
  char sz255LastError[256]; // buffer for 'sprintf'-generated error messages
  T_RigCtrlTrafficLogEntry TrafficLog[RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES];
  int    iTrafficLogHeadIndex; // circular FIFO head index (post-incremented, wraps from RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES-1 to zero)
  BOOL   fTrafficLogFull;      // TRUE when all <RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES> entries are IN USE.
  // (but in most cases, the circular head index 'iTrafficLogHeadIndex' keeps incrementing,
  //  unless RigCtrl_iTrafficMonitorDisplayOptions.RIGCTRL_TRAFFIC_DISPLAY_OPTION_PAUSE_ON_FULL_FIFO is set.)

                               // but the circular head index keeps incrementing
  DWORD  dwRejectMessages4Log;  // bit combination of (1<<RIGCTRL_MSGTYPE_..)
  char*  pszCommentForTxFunction; // <- added 2023-07 to avoid yet another parameter for the (UDP-) transmit function called from RigCtrl_SendAndLogMessage(),
                                  // in combination with RIGCTRL_TRAFFIC_DISPLAY_OPTION_SHOW_UDP_PAYLOAD  .
  int    iMsgTypeForTxFunction;   // <- added 2023-07 so RigCtrl_TxFunctionForCivViaUDP() can pass 'iMsgType' on to RigCtrl_AddToTrafficLog(),
                                  //    especially for flags like RIGCTRL_MSGTYPE_FLAG_SUGGEST_NEW_HEADLINE_FOR_TRAFFIC_LOG .

  long   nCompleteSpectraReceived; // counter for COMPLETELY received spectra (not fragments) for WF speed check
  double dblUnixTimeOfFirstSpectrum; // = TIM_GetCurrentUTC() on reception of the FIRST spectrum

#if( SWI_USE_HAMLIB_SERVER ) // build an application with integrated "Hamlib Net rigctld"-compatible server ?
  void *pvHamlibServer;      // 'link'  from the T_RigCtrlInstance to a Hamlib-Server (T_HLSrv*)
#endif // SWI_USE_HAMLIB_SERVER ?

  // Queue for all 'Unified parameter numbers' that still need to be sent from
  // CLIENT to SERVER using the TCP/IP based protocol implemented in e.g.
  // Remote_CW_Keyer/CwNet.c .   Details/Purpose in the implementation
  //                             of
  //     If there is sufficient space in the outbound network buffer, CwNet_OnPoll()
  //     will compare the queue's HEAD index (iTxParameterQueueHeadIndex) with its own,
  //     "per-client" TAIL index, and if they are different, append another "SET xyz"-command,
  //     along with the CURRENT VALUE retrieved via 'unified parameter number' in this queue:
# define RIGCTRL_TX_PARAMETER_QUEUE_LENGTH 8  
  int iTxParameterQueue[RIGCTRL_TX_PARAMETER_QUEUE_LENGTH];   // <- only contains PARAMETER NUMBER, no VALUES !
  int iTxParameterQueueHeadIndex, iTxParameterQueueTailIndex; // <- 0..RIGCTRL_TX_PARAMETER_QUEUE_LENGTH-1
      // '--> New items (unified parameter numbers) are added to this queue
      //      via RigCtrl_AddParameterNumberToTxQueue().

} T_RigCtrlInstance; // -> main instance = 'RigControlData'



typedef struct tRigCtrl_RadioInfo
{ int  iDefaultAddress; // for Icom radios, the DEFAULT CI-V ADDRESS also identifies the model !
                        // Possible values :  iDefaultAddress = -1 = "OFF (No RigControl)",
                        //                    iDefaultAddress = 0  = "Auto-Detect",
                        //  or RIGCTRL_DEF_ADDR_IC_7300, etc etc (

  const char *pszName;  // user-friendly name like "IC-7300"
  int  iCapabilities;   // <- don't take anything for granted. Almost nothing could be tested.
                        // Zero means (you guessed it) : "I don't know anything about this radio".
# define RIGCTRL_CAPS_NONE                 0x0000
# define RIGCTRL_CAPS_ALL                  0xFFFF
# define RIGCTRL_CAPS_POWER_ON_OFF_VIA_CAT 0x0010 /* set if the rig can be powered on/off via CAT command*/
# define RIGCTRL_CAPS_SPECTRUM_VIA_CAT     0x0100
# define RIGCTRL_CAPS_HAS_IQ_OUTPUT        0x0200 /* the IC-7300 definitely does NOT ! */
# define RIGCTRL_CAPS_HAS_NO_IQ_OUTPUT     0x0400 /* flag only set if we SURELY know.. */

   DWORD dwBands;  // bitwise combination of the following amateur radio bands.
         // For traditional reasons, most use APPROXIMATE WAVELENGTHS, not FREQUENCIES.
# define RIGCTRL_BAND_160M   (1UL<<0)
# define RIGCTRL_BAND_80M    (1UL<<1)
# define RIGCTRL_BAND_60M    (1UL<<2)
# define RIGCTRL_BAND_40M    (1UL<<3)
# define RIGCTRL_BAND_30M    (1UL<<4)
# define RIGCTRL_BAND_20M    (1UL<<5)
# define RIGCTRL_BAND_17M    (1UL<<6)
# define RIGCTRL_BAND_15M    (1UL<<7)
# define RIGCTRL_BAND_12M    (1UL<<8)
# define RIGCTRL_BAND_11M    (1UL<<9) // useful as 'band indicator' for Sporadic E
# define RIGCTRL_BAND_10M    (1UL<<10)
# define RIGCTRL_BAND_R_SW   (1UL<<11) // reserve for a future SHORTWAVE Ham radio band
# define RIGCTRL_BAND_6M     (1UL<<12) // low VHF bands: 50 MHz, ..
# define RIGCTRL_BAND_4M     (1UL<<13) // 70 MHz, only 'temporarily tolerated' in DL !
# define RIGCTRL_BAND_2M     (1UL<<14) // 144 .. 145 MHz, more in some ITU zones
# define RIGCTRL_BAND_R_VHF  (1UL<<15) // reserve for future VHF Ham radio band (no, not 220 MHz)
# define RIGCTRL_BAND_70CM   (1UL<<16) // UHF bands: 430 .. 440 MHz, more in some ITU zones, less in some countries
# define RIGCTRL_BAND_23CM   (1UL<<17) // circa 1.2 GHz but beware.. MANY national specialities!
# define RIGCTRL_BAND_13CM   (1UL<<18) // circa 2.3 GHz (satellite uplink a bit higher)
# define RIGCTRL_BAND_9CM    (1UL<<19) // circa 3.4 GHz
# define RIGCTRL_BAND_6CM    (1UL<<20) // circa 5.6 GHz (supported by IC-905 ?)
# define RIGCTRL_BAND_3CM    (1UL<<21) // circa 10 GHz  (satellite downlink for QO-100)
  // "Exotic bands" (not microwave) use even higher-valued bits:
# define RIGCTRL_BAND_2200M  (1UL<<24) // Longwave (136 kHz, once humming with activity, now deserted)
# define RIGCTRL_BAND_600M   (1UL<<25) // Medium wave (472 kHz, also deserted these days)
  // Bits 28..31 are special flags in 'dwBands' when used as function parameter in e.g. RigCtrl_BandToFrequencyRange():
# define RIGCTRL_BAND_ITU_RGN_1 0x40000000 // Europe, Africa, FSU, Mongolia, Middle East, Iraq
# define RIGCTRL_BAND_ITU_RGN_2 0x80000000 // N & S America, Greenland, some east pacific islands
# define RIGCTRL_BAND_ITU_RGN_3 0xC0000000 // Far East, non-FSU Asia, Iran, Australia, Oceania
# define RIGCTRL_BAND_ITU_MASK  0xC0000000 // mask to extract only the two "ITU Region Bits"

  // Frequenctly seen combinations of the above bits:
# define RIGCTRL_BAND_ALL_SHORTWAVE (RIGCTRL_BAND_160M|RIGCTRL_BAND_80M|RIGCTRL_BAND_40M|RIGCTRL_BAND_30M|RIGCTRL_BAND_20M|RIGCTRL_BAND_17M|RIGCTRL_BAND_15M|RIGCTRL_BAND_12M|RIGCTRL_BAND_10M)
         // '---> Note the absence of e.g. the 60-meter-band and 11 meter (CB) here !


} T_RigCtrl_RadioInfo;


//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Structures used in UDP packets compatible with Icom's "RS-BA" software.
//   Based on Wfview's "packettypes.h", but heavily modified to get rid of
//   the Qt-specific data types, for example "quint8/16/32".
//   Kudos to whoever spent days reverse-engineering Icom's "RS-BA" protocol
//   with Wireshark !  It's a shame that this information isn't officially
//   available from Icom.
// Anything that is really Icom-specific has the prefix "ICOM_"
//   or "ICOM_PKT_" for "Icom-specific UDP packet" .
//- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


#pragma pack(push, 1)

// Fixed Size Packets
#define ICOM_PKT_CONTROL_SIZE            0x10
#define ICOM_PKT_WATCHDOG_SIZE           0x14
#define ICOM_PKT_PING_SIZE               0x15
#define ICOM_PKT_CIV_HEADER_SIZE         ICOM_PKT_PING_SIZE // actually, almost the same HEADER(!) as for "Ping" !
#define ICOM_PKT_OPENCLOSE_SIZE          0x16
#define ICOM_PKT_RETRANSMIT_RANGE_SIZE   0x18
#define ICOM_PKT_TOKEN_SIZE              0x40
#define ICOM_PKT_STATUS_SIZE             0x50
#define ICOM_PKT_LOGIN_RESPONSE_SIZE     0x60
#define ICOM_PKT_LOGIN_SIZE              0x80
#define ICOM_PKT_CONNINFO_SIZE           0x90
#define ICOM_PKT_CAPABILITIES_SIZE       0xA8

// Variable size packets + payload (via Icom's UDP..)
#define ICOM_MIN_PACKET_SIZE_FOR_CIV     0x16 /* 0x15 = 21 would be an "empty CI-V packet", not distnguishable from "PING"! */
#define ICOM_PKT_AUDIO_SIZE              0x18
#define ICOM_PKT_DATA_SIZE               0x15



// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// UDP packet types ordered by their PAYLOAD LENGTH, converted from WFview,
//     with packet numbers in parentheses from a Wireshark dump shown in
//     RigControl.c : RigCtrl_Handler(), RIGCTRL_CONNSTATE_CONNECTING .
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// 0x10 length control packet (connect/disconnect/idle.)  (35)(36)(37)(38)
typedef union tIcomControlPacket
{
  struct             // Byte-Offset into UDP payload | [ : possible values]
   { uint32_t len;   // 0x00
     uint16_t type;  // 0x04 : 0x00="idle", 0x01="missing packet/please re-send", 0x03="connect",
                     //        0x04="connected", 0x05="disconnect", 0x06="are you ready / I am ready" ?
#      define ICOM_CONTROL_PACKET_TYPE_IDLE       0x00
#      define ICOM_CONTROL_PACKET_TYPE_RESEND     0x01
#      define ICOM_CONTROL_PACKET_TYPE_CONNECT    0x03 // possibly also means "Are You There ?" ..
#      define ICOM_CONTROL_PACKET_TYPE_CONNECTED  0x04 // possibly also means "I Am Here", but this is NOT a "ping" (connection speed test)
#      define ICOM_CONTROL_PACKET_TYPE_DISCONNECT 0x05
#      define ICOM_CONTROL_PACKET_TYPE_ARE_YOU_READY 0x06 // udphandler.cpp line 1068: 'sendControl(false, 0x06, 0x01); // Send Are you ready - untracked'
#      define ICOM_CONTROL_PACKET_TYPE_I_AM_READY    0x06 // udphandler.cpp line 160 : 'This is "I am ready" in response to "Are you ready" so send login'

     uint16_t seq;     // 0x06 "seq" = sequence number of RE-TRANSMISSION,
     // see IcomPkt_SendControl() -> IcomPkt_Send() with SEND_PKT_TRACKED !

     uint32_t sentid;  // 0x08
     uint32_t rcvdid;  // 0x0c
   } s;  // .s = "structured"
  BYTE packet[ICOM_PKT_CONTROL_SIZE];
} T_IcomControlPacket;


// 0x14 length watchdog packet        (39)(42)
typedef union tIcomWatchdogPacket {
    struct {                 // Byte-Offset into UDP payload:
        uint32_t len;        // 0x00
        uint16_t type;       // 0x04
        uint16_t seq;        // 0x06
        uint32_t sentid;     // 0x08
        uint32_t rcvdid;     // 0x0c
        uint16_t secondsa;   // 0x10
        uint16_t secondsb;   // 0x12
    } s;
    BYTE packet[ICOM_PKT_WATCHDOG_SIZE];
} T_IcomWatchdogPacket;


// 0x15 length ping packet
// Also used for the slightly different civ header packet.
typedef union tIcomPingPacket
{
  struct // Details ONLY in file:///C:/cbproj/SpecLab/RigControl/Icom_UDP_Info.htm#CIV_data_packet_header !
    {                       // Byte-Offset into the UDP payload for "Ping" and "CI-V over UDP":
        uint32_t len;       // 0x00
        uint16_t type;      // 0x04
#      define ICOM_PING_PACKET_TYPE_PING  0x07
#      define ICOM_PING_PACKET_TYPE_CIV   0x00
        uint16_t seq;       // 0x06
        uint32_t sentid;    // 0x08  // set to pUdpPortInst->myId somewhere, which in fact depends on "my IP address and my ephemeral UDP port number"
        uint32_t rcvdid;    // 0x0c  // set to pUdpPortInst->remoteId somewhere, which is what THE PEER has decided to use in certain 'Control' packets
        BYTE     reply;     // 0x10 .. one of the following:
#      define ICOM_PING_PACKET_REPLY_PING_REQUEST  0x00
#      define ICOM_PING_PACKET_REPLY_PING_RESPONSE 0x01
#      define ICOM_PING_PACKET_REPLY_CIV 0xC1   // CI-V-"header" (followed by payload) ? another of those former 'magic numbers' in *.cpp; used in udpCivData::send()
        union { // This contains differences between the send/receive packet
            struct { // 'Ping'
                uint32_t time;    // 0x11..0x14
            };
            struct { // 'Data' (CI-V-"header" for UDP; payload follows in the same datagram)
                uint16_t datalen; // 0x11..0x12  <- number of CI-V PAYLOAD bytes that follow. LITTLE ENDIAN.
                uint16_t sendseq; // 0x13..0x14  <- STUPID BIG ENDIAN .. WHY ?!
            };
        } u;
    } s;
    BYTE packet[ICOM_PKT_PING_SIZE]; // <- this does NOT include the CI-V payload - ONLY THE "CI-V Header" !
} T_IcomPingPacket, T_IcomCiVPacketHeader;

// 0x16 length open/close packet .  For speculations about the purpose, see IcomPkt_SendOpenClose().
typedef union tIcomOpenClosePacket
{
  struct
   {                      // Byte-Offset into UDP payload:
     uint32_t len;        // 0x00
     uint16_t type;       // 0x04
     uint16_t seq;        // 0x06
     uint32_t sentid;     // 0x08
     uint32_t rcvdid;     // 0x0c
     uint16_t data;       // 0x10
#      define ICOM_OPENCLOSE_PACKET_DATA_01C0 0x01C0 // <- no idea what this means - nothing else was seen yet
              // replace the above with more descriptive names as soon as we know the purpose
     BYTE     unused;     // 0x12 (not 0x11 as shown in WFview's "packettypes.h")
     uint16_t sendseq;    // 0x13
     char magic;          // 0x15
#      define ICOM_OPENCLOSE_PACKET_MAGIC_CLOSE 0x00
#      define ICOM_OPENCLOSE_PACKET_MAGIC_OPEN  0x04

    } s;
    BYTE packet[ICOM_PKT_OPENCLOSE_SIZE];
} T_IcomOpenClosePacket;


// 0x18 length audio packet
typedef union tIcomAudioPacket
{
  struct
   {                      // Byte-Offset into UDP payload:
     uint32_t len;        // 0x00
     uint16_t type;       // 0x04
     uint16_t seq;        // 0x06
     uint32_t sentid;     // 0x08
     uint32_t rcvdid;     // 0x0c
     uint16_t ident;      // 0x10
     uint16_t sendseq;    // 0x12
     uint16_t unused;     // 0x14
     uint16_t datalen;    // 0x16
   } s;
   BYTE packet[ICOM_PKT_AUDIO_SIZE];
} T_IcomAudioPacket;

// 0x18 length retransmit_range packet
typedef union tIcomRetransmitRangePacket
{
  struct
   {                      // Byte-Offset into UDP payload:
     uint32_t len;        // 0x00
     uint16_t type;       // 0x04
     uint16_t seq;        // 0x06
     uint32_t sentid;     // 0x08
     uint32_t rcvdid;     // 0x0c
     uint16_t first;      // 0x10
     uint16_t second;     // 0x12
     uint16_t third;      // 0x14
     uint16_t fourth;     // 0x16
   } s;
  BYTE packet[ICOM_PKT_RETRANSMIT_RANGE_SIZE];
} T_IcomRetransmitRangePacket;


// 0x40 length token packet
typedef union tIcomTokenPacket
{
  struct
   {                            // Byte-Offset into UDP payload:
     uint32_t len;              // 0x00  // value always ICOM_PKT_TOKEN_SIZE = 64 ?
     uint16_t type;             // 0x04  // leave this ZERO when SENDING this from client to server ?
#      define ICOM_TOKEN_PACKET_TYPE_IGNORE_RESPONSE 0x01 // in udpHandler::dataReceived() : response only processed if in->type != 0x01 ?!
     uint16_t seq;              // 0x06  // in udpHandler::sendToken(), left zero; but in sendTrackedPacket(), set to 'sendSeq' !
     uint32_t sentid;           // 0x08  // in udpHandler::sendToken(), set to 'myId'
     uint32_t rcvdid;           // 0x0c  // in udpHandler::sendToken(), set to 'remoteId'
     BYTE     unuseda[3];       // 0x10
     uint16_t code;             // 0x13  : 0x0130 for "token removal / request"
#      define ICOM_TOKEN_PACKET_CODE_S2C   0x0130 // value used in all(?) Token packets from client to server ?
     uint16_t magic; /*ex:res*/ // 0x15  : 1="token removal", 2="token request/update/renew" ?
#      define ICOM_TOKEN_PACKET_MAGIC_REMOVAL  0x01
#      define ICOM_TOKEN_PACKET_MAGIC_REQUEST  0x02
#      define ICOM_TOKEN_PACKET_MAGIC_RESPONSE 0x05 // response to a 'Token request' ?
     // Seen somewhere in the Qt-stuff: " connect(tokenTimer, &QTimer::timeout, this, std::bind(&udpHandler::sendToken, this, 0x05));"
     // Oh dear. Too obfuscated for non-Qt-experts.
     BYTE     innerseq;         // 0x17  // in udpHandler::sendToken(), set to authSeq++
     BYTE     unusedb;          // 0x18
     BYTE     unusedc;          // 0x19
     uint16_t tokrequest;       // 0x1a  // in udpHandler::dataReceived(), copied from in->tokrequest, filled in in sendToken()
     uint32_t token;            // 0x1c  // in udpHandler::dataReceived(), copied from in->token, filled in in sendToken()
     BYTE     unusedd[7];       // 0x20
     uint16_t commoncap;        // 0x27
     BYTE     unuseddd[2];      // 0x29
     BYTE     identa;           // 0x2b
     uint32_t identb;           // 0x2c
     uint32_t response;         // 0x30  // possibly a SIGNED int32, -1 = 0xFFFFFFFF = "Radio rejected token renewal" ?
#      define ICOM_TOKEN_PACKET_RESPONSE_ACCEPTED 0x00000000 // "Radio accepted token renewal" ?
#      define ICOM_TOKEN_PACKET_RESPONSE_REJECTED 0xFFFFFFFF // "Radio rejected token renewal" ?

     BYTE     unusede[12];      // 0x34
    } s;
  BYTE packet[ICOM_PKT_TOKEN_SIZE];
} T_IcomTokenPacket;


// 0x50 length login status packet
typedef union tIcomStatusPacket
{
  struct
   {                            // Byte-Offset into UDP payload:
     uint32_t len;              // 0x00
     uint16_t type;             // 0x04
#      define ICOM_STATUS_PACKET_TYPE_IGNORE_ME 0x01 // in udpHandler::dataReceived() : response only processed if in->type != 0x01 ?!
     uint16_t seq;              // 0x06
     uint32_t sentid;           // 0x08
     uint32_t rcvdid;           // 0x0c
     BYTE     unuseda[3];       // 0x10
     uint16_t code;             // 0x13
     uint16_t res;              // 0x15
     BYTE     innerseq;         // 0x17
     BYTE     unusedb;          // 0x18
     BYTE     unusedc;          // 0x19
     uint16_t tokrequest;       // 0x1a
     uint32_t token;            // 0x1c
     BYTE     unusedd[6];       // 0x20
     uint16_t unknown;          // 0x26
     BYTE     unusede;          // 0x28
     BYTE     unusedf[2];       // 0x29
     BYTE     identa;           // 0x2b
     uint32_t identb;           // 0x2c
     uint32_t error;            // 0x30
     BYTE     unusedg[12];      // 0x34
     BYTE     disc;             // 0x40
     BYTE     unusedh;          // 0x41
     uint16_t civport;          // 0x42 // Sent bigendian
     uint16_t unusedi;          // 0x44 // Sent bigendian
     uint16_t audioport;        // 0x46 // Sent bigendian
     BYTE     unusedj[7];       // 0x49
    } s;
  BYTE packet[ICOM_PKT_STATUS_SIZE];
} T_IcomStatusPacket;

// 0x60 length login status packet
typedef union tIcomLoginResponsePacket
{
  struct
   {                            // Byte-Offset into UDP payload:
     uint32_t len;              // 0x00
     uint16_t type;             // 0x04
#      define ICOM_LOGIN_RESPONSE_TYPE_IGNORE_ME 0x01 // haven't got a clue about what this '0x01' REALLY means !
     uint16_t seq;              // 0x06
     uint32_t sentid;           // 0x08
     uint32_t rcvdid;           // 0x0c
     BYTE     unuseda[3];       // 0x10
     uint16_t code;             // 0x13
     uint16_t res;              // 0x15
     BYTE     innerseq;         // 0x17
     BYTE     unusedb;          // 0x18
     BYTE     unusedc;          // 0x19
     uint16_t tokrequest;       // 0x1a
     uint32_t token;            // 0x1c
     uint16_t authstartid;      // 0x20
     BYTE     unusedd[14];      // 0x22
     uint32_t error;            // 0x30
#      define ICOM_LOGIN_RESPONSE_ERROR_NO_ERROR                 0x00000000
#      define ICOM_LOGIN_RESPONSE_ERROR_INVALID_USER_OR_PASSWORD 0xfeffffff
     BYTE     unusede[12];      // 0x34
     BYTE     connection[16];   // 0x40
     BYTE     unusedf[16];      // 0x50
    } s;
  BYTE packet[ICOM_PKT_LOGIN_RESPONSE_SIZE];
} T_IcomLoginResponsePacket;


// 0x80 length login packet     (41)
typedef union tIcomLoginPacket
{
  struct
   {                            // Byte-Offset into UDP payload:
     uint32_t len;              // 0x00
     uint16_t type;             // 0x04
     uint16_t seq;              // 0x06
     uint32_t sentid;           // 0x08
     uint32_t rcvdid;           // 0x0c
     BYTE     unuseda[3];       // 0x10
     uint16_t code;             // 0x13
     uint16_t res;              // 0x15
     BYTE     innerseq;         // 0x17
     BYTE     unusedaa;         // 0x18;
     BYTE     unusedb;          // 0x19
     uint16_t tokrequest;       // 0x1a
     uint32_t token;            // 0x1c
     BYTE     unusedc[32];      // 0x20
     BYTE     username[16];     // 0x40
     BYTE     password[16];     // 0x50
     char     name[16];         // 0x60
     BYTE     unusedf[16];      // 0x70
   } s;
  BYTE packet[ICOM_PKT_LOGIN_SIZE];
} T_IcomLoginPacket;

// length 0x90 = #144 : "conninfo and stream request packet"
typedef union tIcomConnInfoPacket
{
  struct
   {                            // Byte-Offset into UDP payload:
     uint32_t len;              // 0x00
     uint16_t type;             // 0x04
#      define ICOM_CONNINFO_TYPE_IGNORE_ME 0x01 // <- not sure what this REALLY means
     uint16_t seq;              // 0x06
     uint32_t sentid;           // 0x08
     uint32_t rcvdid;           // 0x0c
     BYTE     unuseda[3];       // 0x10
     uint16_t code;             // 0x13
     uint16_t res;              // 0x15
     BYTE     innerseq;         // 0x17
     BYTE     unusedaa;         // 0x18
     BYTE     unusedb;          // 0x19
     uint16_t tokrequest;       // 0x1a
     uint32_t token;            // 0x1c
     uint16_t authstartid;      // 0x20
     BYTE     unusedd[5];       // 0x22
     uint32_t commoncap;        // 0x27
     BYTE     identa;           // 0x2b
     uint32_t identb;           // 0x2c
     BYTE     unusedf[16];      // 0x30
     char     name[16];         // 0x40
     char     unusedg[16];      // 0x50
     union { // This contains differences between the send/receive packet
      struct { // Receive
        uint32_t busy;          // 0x60
        char     computer[16];  // 0x64
        BYTE     unusedi[16];   // 0x74
        uint32_t ipaddress;     // 0x84
        BYTE     unusedj[8];    // 0x78
       }rx;
      struct { // Send
        char     username[16];  // 0x60
        BYTE     rxenable;      // 0x70
        BYTE     txenable;      // 0x71
        BYTE     rxcodec;       // 0x72
        BYTE     txcodec;       // 0x73
        uint32_t rxsample;      // 0x74
        uint32_t txsample;      // 0x78
        uint32_t civport;       // 0x7c
        uint32_t audioport;     // 0x80
        uint32_t txbuffer;      // 0x84
        BYTE     convert;       // 0x88
        BYTE     unusedl[7];    // 0x89
       }tx;
     }u;
   } s;
  BYTE packet[ICOM_PKT_CONNINFO_SIZE];
} T_IcomConnInfoPacket;


// 0xA8 length capabilities packet
typedef union tIcomCapabilitiesPacket
{
  struct
   {                            // Byte-Offset into UDP payload:
     uint32_t len;              // 0x00
     uint16_t type;             // 0x04
#      define ICOM_CAPABILITIES_TYPE_IGNORE_ME 0x01 // <- not sure what this REALLY means
     uint16_t seq;              // 0x06
     uint32_t sentid;           // 0x08
     uint32_t rcvdid;           // 0x0c
     BYTE     unuseda[3];       // 0x10
     uint16_t code;             // 0x13
     uint16_t res;              // 0x15
     BYTE     innerseq;         // 0x17
     BYTE     unusedb;          // 0x18
     BYTE     unusedc;          // 0x19
     uint16_t tokrequest;       // 0x1a
     uint32_t token;            // 0x1c
     BYTE     unusedd[33];      // 0x20
     BYTE     capa;             // 0x41
     BYTE     unusede[7];       // 0x42
     uint16_t commoncap;        // 0x49
     BYTE     unused;           // 0x4b
     BYTE     macaddress[6];    // 0x4c
     char     name[32];         // 0x52 // "Name" ? What name ? Whose name, mine, yours ? Difference to the one in "ConnInfo" ?
     char     audio[32];        // 0x72
     uint16_t conntype;         // 0x92
     BYTE     civ_addr;         // 0x94 // ex: just "civ". Too short for me. In fact, the CI-V Address (0xA4 for IC-705)
     uint16_t rxsample;         // 0x95
     uint16_t txsample;         // 0x97
     BYTE     enablea;          // 0x99
     BYTE     enableb;          // 0x9a
     BYTE     enablec;          // 0x9b
     uint32_t baudrate;         // 0x9c
     uint16_t capf;             // 0xa0
     BYTE     unusedi;          // 0xa2
     uint16_t capg;             // 0xa3
     BYTE     unusedj[3];       // 0xa5
    } s;
  BYTE packet[ICOM_PKT_CAPABILITIES_SIZE];
} T_IcomCapabilitiesPacket;

#pragma pack(pop)

typedef union uIcomPackets // union with POINTERS TO(!) any kind of 'Icom Packet'
{ T_IcomControlPacket         *control;
  T_IcomWatchdogPacket        *watchdog;
  T_IcomPingPacket            *ping;
  T_IcomCiVPacketHeader       *civ;  // <- "immediately" followed by the CI-V PAYLOAD, in the same datagram !
  T_IcomOpenClosePacket       *open_close;
  T_IcomAudioPacket           *audio;
  T_IcomRetransmitRangePacket *retransmit;
  T_IcomTokenPacket           *token;
  T_IcomStatusPacket          *status;
  T_IcomLoginPacket           *login;
  T_IcomLoginResponsePacket   *login_resp;
  T_IcomConnInfoPacket        *conninfo;
  T_IcomCapabilitiesPacket    *capabilities;
  BYTE                        *packet;
} U_IcomPackets;



//---------------------------------------------------------------------------

// Structures used in DL4YHF's "Remote CW Keyer" for communication

// between client (without direct connection to a real radio)

//     and server (usually connected to his 'local' radio via USB/serial/Ci-V):

//---------------------------------------------------------------------------


typedef struct t_RigCtrl_VfoReport // struct to report the radio's VFO frequency, etc. NOT ICOM-SPECIFIC.

{ // This packet travels "both ways" via TCP/IP, from client to server and server to client,
  // where it is passed on to RigControl.c (which either passes on the command
  // to the connected radio [server side: -> RigCtrl_SetVFOFrequency(), etc],
  // or just stores the data for the display in the GUI [client side]).
  BYTE bStructSize; // [0] set to sizeof(struct..) to allow future extensions
                    // at the end of the struct. With this, the receiver
                    // knows which members of the "new" component are valid.
  // If bStructSize exceeds EIGHT, the payload after CWNET_CMD_FREQ_REPORT
  // may contain any of the "Rig Control Parameters" identified by the following number:
  BYTE bRigControlParameterNumber; // RIGCTRL_PN_UNKNOWN(0), RIGCTRL_PN_FREQUENCY(1), ...
  BYTE bTrxStatus;  // [2] Bit combination of the following flags:
#      define CWNET_TRX_STATUS_TRANSMITTING  0x01 // H=transmitting, L=receiving
#      define CWNET_TRX_STATUS_SQUELCH_MUTED 0x02 // H=audio muted,  L=not muted ("squelch open")
  signed char i8SMeterLevel_dB; // [3] "decibel over S0", signed 8 bit, may be "slightly negative".
                                // i8SMeterLevel_dB = +127 = "S9 plus 73 dB" .. that's enough.
                                // 0x80 = -128 means "no valid S-meter report".
  BYTE   b4AlignmentDummy[4]; // [4..7] future reserve, and to align the 64-bit members to 8-byte offsets
  double dblVfoFrequency;     // [8..15]  current "VFO frequency" in Hertz(!)
  double dblParameterValue;   // [16..23] value of the parameter identified by bRigControlParameterNumber.
} T_RigCtrl_VfoReport;


typedef struct t_RigCtrl_ParamReport_Integer // short struct to report an 'int' parameter, identified by 'parameter number'

{ // Like similar structs defined further below, this packet travels "both ways"
  // via TCP/IP, from client to server and server to client,
  // where it is passed on to RigControl.c .
  // NO NEED FOR AN EXTRA BYTE WITH THE 'STRUCT SIZE' HERE, to keep it as short as possible!
  BYTE bRigControlParameterNumber; // byte [0] : RIGCTRL_PN_xyz
  BYTE bSubindex;             // [1] future reserve for "array-like" parameters or similar
  BYTE b2AlignmentDummy[2];   // [2..3] future reserve / dummy for 4-byte alignment for the following 32-bit parameter:
  long i32ParameterValue;     // [4..7] : 32-bit signed value of the parameter identified by bRigControlParameterNumber
} T_RigCtrl_ParamReport_Integer;


typedef struct t_RigCtrl_ParamReport_Double // struct to report a 'double' parameter, identified by 'parameter number'

{ BYTE bStructSize;         // [0] struct size in bytes
  BYTE bRigControlParameterNumber; // [1] RIGCTRL_PN_UNKNOWN(0), RIGCTRL_PN_FREQUENCY(1), ...
  BYTE bSubindex;           // [2] future reserve for "array-like" parameters or similar
  BYTE b5AlignmentDummy[5]; // [3..7] future reserve / dummy for 8-byte alignment for the following 64-bit parameter:
  double dblParameterValue; // [8..11] value of the parameter identified by bRigControlParameterNumber.
} T_RigCtrl_ParamReport_Double;


typedef struct t_RigCtrl_ParamReport_String // very short struct to report a 'string' parameter, identified by 'parameter number'

{ BYTE bStructSize;         // [0] USED struct size (depends on the actual string length; 255 characters is the MAXIMUM)
  BYTE bRigControlParameterNumber; // byte[1] : RIGCTRL_PN_xyz
  char sz255ParameterValue[256]; // byte [2..bStructSize-1] : zero-terminated "C" string, 8 bits per character
} T_RigCtrl_ParamReport_String;



typedef struct t_RigCtrl_MultiFunctionMeterReport // short struct to report most that an IC-7300 shows on the "Mult-Function Meter" screen

{ BYTE bStructSize;  // [0] USED struct size (depends on the actual string length; 255 characters is the MAXIMUM)
  signed char i8PowerMeterLevel_pcnt; // ICOM-like: They measure output power in PERCENTS, not WATTS !

  signed char i8SWRMeterValue_x10;    // dimensionless "SWR" value, multiplied by 10 to send it as an 8-bit integer
  signed char i8ALCMeterLevel_pcnt;   // 0 ... 100 % of whatever the MAXIMUM may be (don't ask me; ask Icom)
  signed char i8CompMeterLevel_dB;    // compression meter level (measured); Icom measures from 0 to 30 dB. We send it "in dB", not "0241 = 30 dB".
  BYTE u8SupplyVoltage_100mV; // supply voltage in 100-mV-units
  BYTE u8DrainCurrent_100mA;  // drain current in 100-mA-units
  BYTE u8PATemperature_degC;  // power amplifier temperature in degrees CELSIUS
       // Note: The Icom engineers have forgotten to implement a CI-V command
       //       to query the POWER AMPLIFIER TEMPERATURE.
       //       At least an IC-7300 can only display that important parameter LOCALLY !
} T_RigCtrl_MultiFunctionMeterReport;


/*-------------- global variables ---------------------------------------*/

extern const T_RigCtrl_RadioInfo RigCtrl_RadioInfo_CIV[];
extern const T_RigCtrl_ParamInfo RigCtrl_ParameterInfo[];
extern const T_SL_TokenList RigCtrl_OpModes[];
extern const T_SL_TokenList RigCtrl_BandNames[];
   // The above 'constant tables' are exposed via header
   // so the main application (e.g. Spectrum Lab) can use it
   // in a selection list for LAN- or WLAN-capable radios .

extern BOOL RigCtrl_fUpdateTrafficLog; // flag signalling "call TDebugForm::UpdateTrafficLog() a.s.a.p" (or whoever DISPLAYS the traffic log on-screen, and CLEARS that flag when 'done')
extern long RigCtrl_i32MessageNrForTrafficLog;
extern int  RigCtrl_iTrafficMonitorDisplayOptions; // 0=RIGCTRL_TRAFFIC_DISPLAY_OPTION_NORMAL; 1=RIGCTRL_TRAFFIC_DISPLAY_OPTION_SHOW_UDP_PAYLOAD = 1, etc, bitwise combineable:
#      define RIGCTRL_TRAFFIC_DISPLAY_OPTION_DISABLED                0 // if NONE of the following bits are set, RigCtrl_AddToTrafficLog() shows NOTHING..
#      define RIGCTRL_TRAFFIC_DISPLAY_OPTION_ENABLE                  1 // when cleared, RigCtrl_AddToTrafficLog() won't log "normal traffic" (only report "problems")
#      define RIGCTRL_TRAFFIC_DISPLAY_OPTION_SHOW_UDP_PAYLOAD        2
#      define RIGCTRL_TRAFFIC_DISPLAY_OPTION_PAUSE_ON_FULL_FIFO      4
#      define RIGCTRL_TRAFFIC_DISPLAY_OPTION_PAUSE_ON_ERROR          8
#      define RIGCTRL_TRAFFIC_DISPLAY_OPTION_OMIT_PING_IDLE_AND_CO  16

extern int  RigCtrl_iTrafficMonitorExtraColumns;   // 0=RIGCTRL_TRAFFIC_EXTRA_COLUMN_NONE; etc, bitwise combineable:
#      define RIGCTRL_TRAFFIC_EXTRA_COLUMN_NONE           0
#      define RIGCTRL_TRAFFIC_EXTRA_COLUMN_TIME           1 // show "time of day" (in UTC) ?
#      define RIGCTRL_TRAFFIC_EXTRA_COLUMN_TX_RX          2 // show "TX"/"RX" indicator ?
#      define RIGCTRL_TRAFFIC_EXTRA_COLUMN_PAYLOAD_LENGTH 4 // show payload length ("len") ?
#      define RIGCTRL_TRAFFIC_EXTRA_COLUMN_COMMENTS       8 // show "machine-generated comments, parameter names, etc" ?

extern BOOL RigCtrl_fTrafficMonitorPaused; // FALSE=not "automatically paused on a trigger event", TRUE="something happened, PLEASE TAKE A LOOK.."
       // '--> Reason for "pausing" / "stopping" : e.g. RIGCTRL_TRAFFIC_DISPLAY_OPTION_PAUSE_ON_FULL_FIFO / .. STOP_ON_ERROR / ...

extern int  RigCtrl_ITU_Region; // <- INITIALIZED global variable for simplicity, default = 1 :
       // ITU region 1 = Europe, Africa, FSU, Mongolia, Middle East, Iraq
       // ITU region 2 = North- and South America, Greenland, some east pacific islands
       // ITU region 3 = Far East, non-FSU Asia, Iran, Australia, Oceania


/*-------------- prototypes ----------------------------------------------*/
#ifndef CPROT  // For peaceful co-existence of "C" and "C++" ....
// Note on "extern": do not trust the lousy example from Borland !
//   >extern "c" = wrong    (not recognized by Borland C++ V4.0)
//   >extern "C" = correct  (though the help system says something different)
#ifdef __cplusplus
 #define CPROT extern "C"
#else
 #define CPROT
#endif  /* ! "cplusplus" */
#endif

// External functions (used by RigControl.c, but not implemented in RigControl.c) :
// ex: CPROT int _stdcall IO_SendToCOMPort( // "transmit"-function for 'COM' ports (RS-232, USB Virtual COM, etc)
//        struct tRigCtrlInstance *pRC, // [in] address of the 'Rig Control' instance
//        int  iRigCtrlPort, // [in] RIGCTRL_PORT_RADIO / RIGCTRL_PORT_CLIENT_1 / 2 ?
//        BYTE *source, int n_bytes, int fWaitForCompletion,
//        int *piErrorCode, int *piNumMsWaited );


//--- Common functions (not specific to CI-V or whatever may be added later) ----
      // Note: Radio address and protocol are filled in by the caller
      //       AFTER RigCtrl_Init(). Avoid a bloated API, and don't
      //       use 'getter' and 'setter' functions for each and everything.
      //       The same applies to many other members of T_RigCtrlInstance.
CPROT void  RigCtrl_Init( // initialize everything for a "single instance"
               T_RigCtrlInstance *pRC,    // [out] instance data
               int iRadioControlProtocol, // [in] e.g. RIGCTRL_PROTOCOL_NONE, RIGCTRL_PROTOCOL_ICOM_CI_V,
                                          //      or   RIGCTRL_PROTOCOL_HAMLIB_RIGCTLD (using TCP/IP, not "COM")
               int iRadioDefaultAddress,  // [in] e.g. RIGCTRL_DEF_ADDR_NO_REMOTE_CTRL, RIGCTRL_DEF_ADDR_AUTO_DETECT, RIGCTRL_DEF_ADDR_IC_7300, etc^10 ...
               T_CFIFO *pRadioPortRxFifo, // [in] thread-safe circular FIFO to RECEIVE data from the radio
               T_CFIFO *pRadioPortTxFifo, // [in] thread-safe circular FIFO to TRANSMIT data to the radio
               int nClientPorts,          // [in] 0..RIGCTRL_MAX_CLIENT_PORTS
               int *piClientOptions); // [in] array, one per client
CPROT void  RigCtrl_Close( T_RigCtrlInstance *pRC );     // closes handles, sockets, frees resources
CPROT void  RigCtrl_ReConnect( T_RigCtrlInstance *pRC ); // restart connection, called e.g. when clicking the "Apply"-button on SL's 'Radio Control' panel
CPROT void  RigCtrl_SetClientOptions( T_RigCtrlInstance *pRC, int iRigCtrlPort, int iClientOptions );
CPROT void  RigCtrl_StartReading( T_RigCtrlInstance *pRC ); // begins "communicating" actively
CPROT void  RigCtrl_Handler( T_RigCtrlInstance *pRC ); // periodically called every 50(?) milliseconds
#           define C_RIGCTRL_HANDLER_CALLING_INTERVAL_MS 50 // calling interval for RigCtrl_Handler() in milliseconds
CPROT int   RigCtrl_EnumerateRequiredPNs( int iZeroBasedIndex ); // used by RigControl.c to poll all *mandatory* parameters
CPROT void  RigCtrl_ProcessRxData( T_RigCtrlInstance *pRC, int iRigCtrlPort );
CPROT void  RigCtrl_SetLocalPTTFlag( T_RigCtrlInstance *pRC, BOOL fTransmitting);
CPROT void  RigCtrl_TurnPowerOnOrOff( T_RigCtrlInstance *pRC,BOOL fPowerOn );

CPROT BOOL  RigCtrl_SetVFOFrequency( T_RigCtrlInstance *pRC, double dblFreqHz );
CPROT BOOL  RigCtrl_SwitchToBand( T_RigCtrlInstance *pRC, DWORD dwNewBand/* RIGCTL_BAND_xyz(only ONE bit)*/ );
CPROT BOOL  RigCtrl_SwitchToFreqMemEntry( T_RigCtrlInstance *pRC, T_RigCtrlFreqMemEntry *pFreqMemEntry );
CPROT DWORD RigCtrl_GetAvailableBands( T_RigCtrlInstance *pRC ); // -> bitwise combination of RIGCTL_BAND_xyz
CPROT DWORD RigCtrl_FrequencyToBand( double dblVfoFrequency ); // -> RIGCTL_BAND_xyz

CPROT void  RigCtrl_OnVFOReportFromNetwork( T_RigCtrlInstance *pRC, T_RigCtrl_VfoReport *pVfoReport);
CPROT void  RigCtrl_OnParamReport_Integer( T_RigCtrlInstance *pRC, T_RigCtrl_ParamReport_Integer *pReport);
CPROT void  RigCtrl_OnParamReport_Double(T_RigCtrlInstance *pRC,T_RigCtrl_ParamReport_Double *pReport);
CPROT void  RigCtrl_OnParamReport_String(T_RigCtrlInstance *pRC,T_RigCtrl_ParamReport_String *pReport);
CPROT void  RigCtrl_OnMultiFunctionMeterReportFromNetwork(T_RigCtrlInstance *pRC, T_RigCtrl_MultiFunctionMeterReport *pMeterReport);
CPROT void  RigCtrl_OnBandStackingRegisterReportFromNetwork( T_RigCtrlInstance *pRC, const char *pszSource );


CPROT BOOL  RigCtrl_SendCmdToReadUnifiedPN( T_RigCtrlInstance *pRC, int iUnifiedPN );
CPROT BOOL  RigCtrl_SetOperatingMode( T_RigCtrlInstance *pRC, int iOpMode ); // .. and SEND TO RADIO ..
CPROT BOOL  RigCtrl_SetScopeRefLevel_dB( T_RigCtrlInstance *pRC, double dblNewValue ); // .. and SEND TO RADIO ..
CPROT BOOL  RigCtrl_SetScopeSpan_Hz( T_RigCtrlInstance *pRC, double dblNewDisplayWidth_Hz ); // .. and SEND TO RADIO ..

CPROT void  RigCtrl_UpdateBasebandFrequencyConversionParams( T_RigCtrlInstance *pRC ); // -> DemodBasebandFreqOffset, ..Factor
CPROT T_RigCtrl_RadioInfo *RigCtrl_GetRadioInfoByDefaultAddress( int iRadioCtrlProtocol, int iDefaultAddress );
CPROT const char* RigCtrl_OperatingModeToString( int iOpMode ); // e.g. "USB", "LSB", "CW", ...
CPROT const char* RigCtrl_DefaultAddressToString( int iRadioCtrlProtocol, int iDefaultAddress );
CPROT const char* RigCtrl_RigControlPortNrToString( int iRigCtrlPort );
CPROT const char* RigCtrl_ControlConnStateToString( T_RigCtrlInstance *pRC ); // more verbose than RigCtrl_ConnStateToString(), but only for the "CONTROL"-UDP-port
CPROT const char* RigCtrl_ParameterPollingStateToString(int iParameterPollingState);
CPROT const char* RigCtrl_GetRadioControlPortAsString(void);  // -> e.g. "COM5", if that's the port "talking to the radio" .
               // Since RighControl.c doesn't know anything about serial ports,
               // RigCtrl_GetRadioControlPortAsString() must be provided by the application.
CPROT int   RigCtrl_FreqMemEntryToString(T_RigCtrlFreqMemEntry *pFreqMemEntry,
                            int iArrayIndex, char *pszDest, int iMaxDestLen );
CPROT int   RigCtrl_StringToFreqMemEntry(T_RigCtrlFreqMemEntry *pFreqMemEntry,
                            int *piArrayIndex, const char *pszSource );



   // API to access parameters (stored in the instance data) via 'Unified Parameter Number'
   // (to eliminate any IcomKenwoodYeasu-specific addressing of settings/parameter/readout data)
CPROT int     RigCtrl_EnumerateUnifiedParameters( int iPreviousUnifiedPN );
CPROT T_RigCtrl_ParamInfo *RigCtrl_GetInfoForUnifiedParameterNumber(int iUnifiedPN);
CPROT const char* RigCtrl_UnifiedParameterNumberToString(int iUnifiedPN);
CPROT int*    RigCtrl_GetIntValuePtrForUnifiedParameterNumber( T_RigCtrlInstance *pRC, int iUnifiedPN );
CPROT double* RigCtrl_GetDblValuePtrForUnifiedParameterNumber( T_RigCtrlInstance *pRC, int iUnifiedPN );
CPROT BOOL    RigCtrl_HaveValidParamValue(T_RigCtrlInstance *pRC, T_RigCtrl_ParamInfo *pPI);
CPROT BOOL    RigCtrl_GetParamValueAString(T_RigCtrlInstance *pRC, T_RigCtrl_ParamInfo *pPI, char *pszDest, int iMaxLen );
CPROT BOOL    RigCtrl_SetParamValue_Int( T_RigCtrlInstance *pRC, T_RigCtrl_ParamInfo *pPI, long i32Value );
CPROT BOOL    RigCtrl_SetParamValue_Double(T_RigCtrlInstance *pRC,T_RigCtrl_ParamInfo *pPI,double dblValue);
CPROT long    RigCtrl_GetParamByPN_Int( T_RigCtrlInstance *pRC, int iUnifiedPN );
CPROT double  RigCtrl_GetParamByPN_Double( T_RigCtrlInstance *pRC, int iUnifiedPN );
CPROT void    RigCtrl_SetParamByPN_Int( T_RigCtrlInstance *pRC, int iUnifiedPN, long i32Value );
CPROT void    RigCtrl_SetParamByPN_Double( T_RigCtrlInstance *pRC, int iUnifiedPN, double dblValue );
CPROT BOOL    RigCtrl_AddParameterNumberToTxQueue(T_RigCtrlInstance *pRC, int iUnifiedPN); // .. for a later transmission via CwNet.c !


CPROT T_RigCtrl_Spectrum *RigCtrl_AppendSpectrumToFIFO( T_RigCtrlInstance *pRC,
                   double dblBinWidth_Hz, double dblFmin_Hz, int nFrequencyBins);
CPROT int     RigCtrl_GetNumSpectraAvailableInFIFO( T_RigCtrlInstance *pRC, int iTailIndex );
CPROT T_RigCtrl_Spectrum *RigCtrl_GetSpectrumFromFIFO( T_RigCtrlInstance *pRC, int *piTailIndex );
#      define RIGCTRL_GET_OLDEST_ENTRY -1 // dummy FIFO index (iTailIndex) for "the oldest available entry"
#      define RIGCTRL_GET_LATEST_ENTRY -2 // dummy FIFO index (iTailIndex) for "the most recent entry"

CPROT BOOL  RigCtrl_GetDisplayableRadioFrequencyRange( T_RigCtrlInstance *pRC,
        double *pdblRadioFreqMin, double *pdblRadioFreqMax, // [out] : RADIO  frequencies
        BOOL   *pfFixedEdgeMode );  // [out,optional: TRUE= "fixed edges", FALSE="centered on VFO"

CPROT void RigCtrl_ClearTrafficLog( T_RigCtrlInstance *pRC );
CPROT void RigCtrl_AddToTrafficLog( T_RigCtrlInstance *pRC, int iRigControlPort,
              BYTE *pbMessage, int iMsgLength, int iMsgType, const char *pszComment );
CPROT int  RigCtrl_GetNumEntriesInTrafficLog( T_RigCtrlInstance *pRC, int iTailIndex );
CPROT BOOL RigCtrl_ReadTrafficLog( T_RigCtrlInstance *pRC, int *piTrafficTailIndex,
                char *pszDest, int iWantedCharsPerLine, int *piMsgType, DWORD *pdwColor );
CPROT void RigCtrl_GetHeadlineForTrafficLog(int iMsgType, char *pszDest, int iWantedCharsPerLine);

CPROT void RigCtrl_EnterErrorHistory( const char *pszFormat,  // located "somewhere in the application" !
                                       ... );  // optional argument list (printf-style)
CPROT void RigCtrl_LimitDouble( double *pdblValue, double dblMin, double dblMax );


//--- CI-V specific functions -------------------------------------------------
CPROT int   RigCtrl_ParseCIV( T_RigCtrlInstance *pRC, int iRigCtrlPort, BYTE *pbMessage, int iMsgLength );


#endif //  RigControlH

/* EOF <RigControl.h> */
