//---------------------------------------------------------------------------
// File: C:\cbproj\Remote_CW_Keyer\SpecDisp.cpp
// Date: 2024-02-26
// Author: Wolfgang Buescher (DL4YHF)
// Purpose: Spectrum Display (waterfall and/or spectrum graph)
//          for the 'Remote CW Keyer' .
//---------------------------------------------------------------------------

#include "switches.h"  // project specific compiler switches ("options"),
                       // must be included before anything else !
#include "yhf_type.h"  // classic types like BYTE, WORD, DWORD, BOOL, ..
#include <vcl.h>       // Borland's stoneage Visual Component Library
#include <math.h>
#include <string.h>
#include <stdio.h>     // no "standard I/O" but "sprintf" used here

#ifdef __BORLANDC__
// prevent that bugger from bugging us with 'initialized data in header, cannot pre-compile..." :
# pragma hdrstop
#endif

#include "RigControl.h" // T_RigCtrlInstance "delivers" the spectra to display
          // (Icom call them "waveform data" but that's too ambiguous)
          // in a small circular FIFO with T_RigCtrl_Spectrum objects,
          // each of them with up to <RIGCTRL_MAX_FREQ_BINS_PER_SPECTRUM>.
#include "SpecDisp.h"   // header for THIS module ('Spectrum/Waterfall display')

   // Note the lack of dependencies from the 'main form' and 'Keyer GUI' here,
   // i.e. *not* #including Keyer_Main.h and Keyer_GUI.h in this module.
   // Try to keep it that way, to ease migrating to e.g. Qt or wxWidgets.


//---------------------------------------------------------------------------
// Variables
//---------------------------------------------------------------------------




//---------------------------------------------------------------------------
// Implementation of functions
//---------------------------------------------------------------------------

float Voltage_to_dBfs( float fltNormalizedVoltage/*0..1*/ )
{
  if( fltNormalizedVoltage >= 1e-10 )
   {  return 20.0 * log10( fltNormalizedVoltage ); // ->  -200 to 0 dB "over" full scale
   }
  else
   {  return -200.0; // dB "over" full scale
   }
} // end Voltage_to_dBfs()

//---------------------------------------------------------------------------
static void LimitInteger( int *piValue, int iMin, int iMax )
{ if( *piValue > iMax )
   {  *piValue = iMax;
   }
  if( *piValue < iMin )
   {  *piValue = iMin;
   }
} // end LimitInteger()

//---------------------------------------------------------------------------
void GetStepAndSubstepForScale( double dblMinStep, double *pdblLargeStep, double *pdblSubStep)
  // Retrieves a "nice" stepwidth for drawing a scale, with a given MINIMUM stepwidth:
  //  [in] dblMinStep = MINIMUM stepwidth, dictated by the displayed bandwidth
  //                       and the number of pixels along the frequency scale.
  //                    Example: IC-7300 spectrum display set to "+/-25k"
  //                              = 50 kHz displayed bandwidth,
  //                              at least 80 pixels per "step" (labelled tick),
  //                              1200 pixels wide 'drawing canvas' for the scale,
  //   -> dblMinStep = 80 min_pixels_per_step * 50 kHz / 1200 pixels = 3333 [Hz].
  //

{ double step, substep, fact_pow10 = 1.0;

  while( dblMinStep > 1000.0/*Hz*/ )
   { fact_pow10 *= 10.0;
     dblMinStep *=  0.1;
   }

  if     ( dblMinStep <=  1.0 ) {  step =  1.0;   substep=0.25; }
  else if( dblMinStep <=  2.0 ) {  step =  2.0;   substep=0.5;  } // not 2.5 - no decimals in the labels !
  else if( dblMinStep <=  5.0 ) {  step =  5.0;   substep=1.0;  }
  else if( dblMinStep <= 10.0 ) {  step = 10.0;   substep=2.5;  }
  else if( dblMinStep <= 25.0 ) {  step = 25.0;   substep=5.0;  }
  else if( dblMinStep <= 50.0 ) {  step = 50.0;   substep=10.0; }
  else if( dblMinStep <= 100.0) {  step = 100.0;  substep=25.0; }
  else if( dblMinStep <= 250.0) {  step = 250.0;  substep=50.0; }
  else if( dblMinStep <= 500.0) {  step = 500.0;  substep=100.0; }
  else                          {  step = 1000.0; substep=100.0; }
  step    *= fact_pow10;
  substep *= fact_pow10;

  if( pdblLargeStep != NULL )
   { *pdblLargeStep = step;
   }
  if( pdblSubStep != NULL )
   { *pdblSubStep = substep;
   }
}

//---------------------------------------------------------------------------
void SpecDisp_InitControl( T_SpecDispControl *pDispCtrl, int iFreqScaleHeight )
{
  memset( pDispCtrl, 0, sizeof(T_SpecDispControl) );
  pDispCtrl->fltSpectrumAmplMin_dB = 0.0;   // 0..127 dB on the "Icom scale" ..
  pDispCtrl->fltSpectrumAmplMax_dB = 100.0; // .. but this seemed to be a realistic maximum

  pDispCtrl->iFreqScaleHeight = iFreqScaleHeight;
  pDispCtrl->iWFBrightness_Percent = pDispCtrl->iWFContrast_Percent = 50;
  pDispCtrl->iWFColorPalette = WF_PALETTE_SUNRISE; // author's favourite
  pDispCtrl->iColourScheme = COLOUR_SCHEME_DEFAULT;
  // Configure some of the 'markers' in various parts of the GUI (spectrum displays):
  pDispCtrl->marker[SPECDISP_MARKER_VFO].iGuiControl=GUI_CONTROL_SPECTRUM;

} // end SpecDisp_InitControl()

//---------------------------------------------------------------------------
BOOL SpecDisp_PrepareDrawing(
                 Graphics::TBitmap *pbmpDest,  // [in,out] Borland-VCL-style "TBitmap"
                 T_SpecDispControl *pDispCtrl, // [in,out] layout info, frequency range, etc
                 T_RigCtrlInstance *pRigCtrl)  // [in] "Rig Control" instance
{
  // 2024-03-01: Crash-landed in "__InitExceptBlockLDTC()" [wtf..?] shortly after
  //     entering SpecDisp_UpdateFreqScale(), thus the following pointer-checks:
  if( pbmpDest==NULL )
   { return FALSE; // VCL-style TBitmap not properly initialized !
   }
  if( pDispCtrl==NULL )
   { return FALSE;  // missing 'Display Control' instance !
   }
  if( pRigCtrl==NULL )
   { return FALSE;  // no valid RigControl instance !
   }
#ifdef __BORLANDC__  // here the implementation for Borland/Embarcadero C++Builder/VCL :
  if( pbmpDest->PixelFormat!=pf32bit )  // <- crashed HERE when single-stepping
   { return FALSE; // Not compatible with our RGB-quad-mapper, T_RGBColor !
   }
  pDispCtrl->Canvas = pbmpDest->Canvas;
  if( pDispCtrl->Canvas==NULL )
   { return FALSE; // oops.. no TCanvas (~~~ graphic device context ?) for this TBitmap ?!
   }
  // Since we already have the VCL-TCanvas, prepare 'Brush'-, 'Pen'-, 'Font'-,
  // and some other colours, depending on the GUI's colour scheme:
  pDispCtrl->Canvas->Brush->Style = bsSolid;
  switch( pDispCtrl->iColourScheme )
   { case COLOUR_SCHEME_DEFAULT : // black text on white background
     default:
        pDispCtrl->Canvas->Brush->Color  = clWhite;
        pDispCtrl->Canvas->Pen->Color    = clBlack;
        pDispCtrl->Canvas->Font->Color   = clBlack;
        pDispCtrl->dwBackgndColor = RGB( 0xFF, 0xFF, 0xFF );
        pDispCtrl->dwGridColor    = RGB( 0x60, 0x60, 0x60 );
        pDispCtrl->dwCurveColor   = RGB( 0x00, 0x00, 0xFF );
        break;
     case COLOUR_SCHEME_DARK    : // white text on black background
        pDispCtrl->Canvas->Brush->Color  = clBlack;
        pDispCtrl->Canvas->Pen->Color    = clWhite;
        pDispCtrl->Canvas->Font->Color   = clWhite;
        pDispCtrl->dwBackgndColor = RGB( 0x00, 0x00, 0x00 );
        pDispCtrl->dwGridColor    = RGB( 0x60, 0x60, 0x60 );
        pDispCtrl->dwCurveColor   = RGB( 0xC0, 0xC0, 0xFF );
        break;
   } // end switch( pDispCtrl->iColourScheme )

  if( !RigCtrl_GetDisplayableRadioFrequencyRange( pRigCtrl,
            &pDispCtrl->fmin_from_RigCtrl, &pDispCtrl->fmax_from_RigCtrl,
            &pDispCtrl->fixedEdgeMode ) )
   { return FALSE;  // failed to retrieve the spectrum / waterfall display range !
   }
  pDispCtrl->iCanvasWidth  = pbmpDest->Width;
  pDispCtrl->iCanvasHeight = pbmpDest->Height;
  return TRUE;
#else
#  error "Sorry, this module hasn't been ported to other targets yet."
#endif
} // end SpecDisp_PrepareDrawing()

//---------------------------------------------------------------------------
void SpecDisp_FillRect(
      T_SpecDispControl *pDispCtrl, // [in,out] layout info, config data, etc
      int x1, int y1, int x2, int y2 )
{
#ifdef __BORLANDC__  // here the implementation for Borland/Embarcadero C++Builder/VCL :
  if( pDispCtrl->Canvas != NULL )
   { pDispCtrl->Canvas->FillRect( TRect(x1,y1,x2+1/*!*/,y2+1/*!*/) );
   }
#else
#  error "Sorry, this module hasn't been ported to other targets yet."
#endif
} // end SpecDisp_FillRect()

//-------------------------------------------------------------------------
void SpecDisp_DrawLine(T_SpecDispControl *pDispCtrl, int x1, int y1, int x2, int y2)
{
#ifdef __BORLANDC__  // here the implementation for Borland/Embarcadero C++Builder/VCL :
  if( pDispCtrl->Canvas != NULL )
   { pDispCtrl->Canvas->MoveTo(x1,y1);
     pDispCtrl->Canvas->LineTo(x2,y2); // this does not include the end coord !
   }
#else
#  error "Sorry, this module hasn't been ported to other targets yet."
#endif
} // end SpecDisp_DrawLine()

//-------------------------------------------------------------------------
void SpecDisp_DrawText(T_SpecDispControl *pDispCtrl, int x, int y, char *pszText)
{
#ifdef __BORLANDC__
  if( pDispCtrl->Canvas != NULL )
   { pDispCtrl->Canvas->TextOut(x,y,pszText);
   }
#else
#  error "Sorry, this module hasn't been ported to other targets yet."
#endif
} // end SpecDisp_DrawText()

//---------------------------------------------------------------------------
static int FreqToScreenX( T_SpecDispControl *pDispCtrl, double freq_Hz )
{ // Converts a frequency [Hz] into a screen offset (client abscissa, "x"),
  // for the waterfall area (and the associated frequency scale).
  // Note: To avoid desaster with invalid coordinates, the result will be
  //       clipped "almost" inside the 'waterfall coordinate range' :
  //   Return value = -1 : "frequency is TOO LOW to be displayed"
  //                  0 .. pDispCtrl->iCanvasWidth - 1 : "frequency is ON-SCREEN"
  //                  pDispCtrl->iCanvasWidth : .. TOO HIGH to be displayed .
  double fmin = pDispCtrl->fmin_from_RigCtrl;
  double fmax = pDispCtrl->fmax_from_RigCtrl;
  double span = fmax - fmin;
  int x = 0;
  if( span > 0 )
   { x = (int)( ( (freq_Hz-fmin) * (double)pDispCtrl->iCanvasWidth ) / span );
   }
  if( x < -1 )
   {  x = -1;
   }
  if( x > pDispCtrl->iCanvasWidth )
   {  x = pDispCtrl->iCanvasWidth;
   }
  return x;
} // FreqToScreenX()

//---------------------------------------------------------------------------
static double ScreenXToFreq( T_SpecDispControl *pDispCtrl, int x )
{ // Inverse function to FreqToScreenX() .
  double fmin = pDispCtrl->fmin_from_RigCtrl;
  double fmax = pDispCtrl->fmax_from_RigCtrl;
  double span = fmax - fmin;
  if( pDispCtrl->iCanvasWidth > 0 )
   { // inverse to x = (int)( ( (freq_Hz-fmin) * (double)pDispCtrl->iCanvasWidth ) / span ) ,
     // solved for freq_Hz :
     return fmin + ((double)x * span) / (double)pDispCtrl->iCanvasWidth;
   }
  else
   { return 0.0;
   }
} // ScreenXToFreq()


//---------------------------------------------------------------------------
void SpecDisp_GetFreqScaleParams( T_SpecDispControl *pDispCtrl,
                  double *pdblStartFreq,
                  double *pdblLargeSteps_Hz, double *pdblSmallSteps_Hz,
                  char   *psz16FormatString )
{ // Select a suitable frequency step for the ticks .. must depend on the
  //  frequency span, AND the display width.
  double fmin = pDispCtrl->fmin_from_RigCtrl;
  double fmax = pDispCtrl->fmax_from_RigCtrl;
  double freq; // 'minimum spacing' between labels [Hz]

  if( pDispCtrl->iCanvasWidth > 0 )
   { freq = 80/*pixels*/ * (fmax - fmin) / pDispCtrl->iCanvasWidth;
     // '--> min value of "Hertz per tick", but often a very odd number
   }
  else
   { freq = 1000.0; // Hz
   }
  GetStepAndSubstepForScale( freq, pdblLargeSteps_Hz, pdblSmallSteps_Hz);
  // Initial test with an IC-7300 :
  //    fmin = 7031970 [Hz],  fmin = 7041970 [Hz], pDispCtrl->iCanvasWidth = 1366 [pixels]
  // -> fstep=     500 [Hz],  substep = 100 [Hz] .  That's 10 kHz / 500 Hz = 20 'labelled ticks' .
  freq  = fmin;
  freq -= fmod( freq, *pdblLargeSteps_Hz );
  *pdblStartFreq = freq;

  if( psz16FormatString != NULL )  // caller needs a FORMAT STRING for the labels..
   { strcpy( psz16FormatString, "%.5lf" );
     // ToDo: Let the number of decimals (fractional digits) depend on the
     //       'large stepwidth' !
   }

} // end SpecDisp_GetFreqScaleParams()


//-------------------------------------------------------------------------
void SpecDisp_GetTextExtent( T_SpecDispControl *pDispCtrl,
                             char *pszString, int *piWidth, int *piHeight)
  // Computes the width and height of the specified string of text .
{
#ifdef __BORLANDC__
  TSize textSize;
  textSize = pDispCtrl->Canvas->TextExtent( pszString );  // the "VCL" way of doing things..
  if(piWidth)  *piWidth = textSize.cx; // the name (TSize) has changed, but not the principle
  if(piHeight) *piHeight= textSize.cy;
#elif( defined __WIN32__ ) // here the implementation for "bare Win32" (no VCL):
  HDC  hdc = ?;  // how to retrieve the Win32 GDI "device context" for what used to be the TCanvas ?
  SIZE size;
  if(GetTextExtentPoint32( hdc, pszString, strlen(pszString), &size) )
   { if(piWidth) *piWidth=size.cx;
     if(piHeight)*piHeight=size.cy;
   }
  else // function failed..
   { if(piWidth) *piWidth=1;
     if(piHeight)*piHeight=1;
   }
#else
#  error "Sorry, this module hasn't been ported to other targets yet."
#endif
} // end GetTextExtent()



//---------------------------------------------------------------------------
void SpecDisp_UpdateWaterfall( // aka spectrogram ..
        Graphics::TBitmap *pbmpDest,  // [out] Borland-VCL-style "TBitmap"
        T_SpecDispControl *pDispCtrl, // [in] layout info, config data, etc
        T_RigCtrlInstance *pRigCtrl,  // [in] "Rig Control" instance
        int iScopeDisplayTailIndex,   // [in] index of the OLDEST spectrum to plot
        int nSpectra )                // [in] number of spectra (waterfall lines) to plot
{
  double freq, fstep, substep;
  int pixel_brightness, red, green, blue;
  int i,n,x,y,w,h,x1,y1,x2,y2;
  int iLine;
  BOOL first;
  T_RigCtrl_Spectrum *pSpectrum;
  float flt, fltFreqToBinIdx, fltPeak;
  int   iBinIdx, iPrevBinIdx;
  long  i32SpectrumIndex;
  T_RGBColor *pPixels;  // <- this only works with bitmap 'PixelFormat' = 'pf32bit' !

  if( ! SpecDisp_PrepareDrawing( pbmpDest, pDispCtrl, pRigCtrl ) )
   { return; // something wrong with any of the three inputs above
   }

  if( pDispCtrl->fClearWaterfall )
   {
     SpecDisp_FillRect( pDispCtrl, 0,0, pDispCtrl->iCanvasWidth,
                                        pDispCtrl->iCanvasHeight );
     pDispCtrl->fClearWaterfall = FALSE;
   }

  // How many lines of the waterfall must be updated ?
  if( nSpectra > 0 )
   {
     // Scroll the (old) waterfall image down by as many lines as required:
     if( nSpectra < pDispCtrl->iCanvasHeight )
      { TRect rctSource = TRect( 0/*left*/, 0/*top*/, pbmpDest->Width/*right*/,
                                 pbmpDest->Height-nSpectra/*bottom*/);
        TRect rctDest   = TRect( 0/*left*/, nSpectra/*top*/, pbmpDest->Width/*right*/,
                                 pbmpDest->Height/*bottom*/);
        pDispCtrl->Canvas->CopyRect( // <- wrapper for BitBlt() with paramter 'SRCCOPY' ?
                                 rctDest, pDispCtrl->Canvas/*Source*/, rctSource );
      }

     // Plot a few more spectrogram lines into the "waterfall"
     //   (begin with most recent on top, thus the name .. "like water, falling down")
     for(iLine=0; iLine < nSpectra; ++iLine )
      { int iTailIndex = iScopeDisplayTailIndex-(nSpectra-1-iLine);
        pSpectrum = RigCtrl_GetSpectrumFromFIFO( pRigCtrl, &iTailIndex );
        pPixels   = (T_RGBColor*)pbmpDest->ScanLine[iLine];
        if( (pSpectrum != NULL) && (pPixels!=NULL) )
         { // Since neither the width of the window, not the FFT size is fixed anymore,
           // the ratio between 'screen pixels' and 'FFT frequency bins' is arbitrary.
           //
           // Principle used below:
           // Running along the frequency axis, the program looks into
           //    a RANGE(!) of FFT bins(!) to get the color of a pixel.
           // There may be less FFT bins than pixels, there may be more,
           // and only if we're lucky there is a 'perfect match' (one pixel per bin).
           if( pSpectrum->dblBinWidth_Hz > 0.0 )
            { fltFreqToBinIdx = 1.0 / pSpectrum->dblBinWidth_Hz;
            }
           else // avoid div-by-zero..
            { fltFreqToBinIdx = 0.0;
            }
           iPrevBinIdx = -1;
           for(x = 0; x < pDispCtrl->iCanvasWidth; ++x)
            {
              // Convert the x-coord to a frequency (in Hertz):
              freq = ScreenXToFreq( pDispCtrl, x );
              // Convert the pixel's center frequency to an FFT frequency bin index:
              iBinIdx = (int)( 0.5 + (freq - pSpectrum->dblFmin_Hz) * fltFreqToBinIdx);
              if( iBinIdx>=0 && iBinIdx<pSpectrum->nBinsUsed )  // an IC-7300 delivered 475 frequency bins...
               {
                 if( iPrevBinIdx < 0 )
                  {  iPrevBinIdx = iBinIdx-1;  // 1st loop
                     if( iPrevBinIdx < 0 )
                      {  iPrevBinIdx = 0;
                      }
                  }
                 if( iPrevBinIdx < iBinIdx ) // Is this a new FFT-bin at all ?
                  { fltPeak = pSpectrum->fltMagnitudes_dB[++iPrevBinIdx];
                    while( iPrevBinIdx < iBinIdx ) // MORE THAN ONE BIN per pixel : get the peak..
                     { flt = pSpectrum->fltMagnitudes_dB[++iPrevBinIdx];
                       if( flt>fltPeak )
                        {  fltPeak = flt;
                        }
                     }
                    // Convert Icom's "magnitude in dB" into a pixel colour .
                    //  'iWFBrightness_Percent'  = trackbar position, ranging from 0 to 100 .
                    //  'fltPeak' = FFT bin amplitude (proportional to a voltage),
                    //         normalized to 0....1 for a full-swing sinewave w/o AGC .
                    //  'pixel_brightness' = 0..255 .
                    //  The factor below was empirically found to have a similar
                    //  brightness as on an IC-7300's own display, with both
                    //  'Contrast' and 'Brightness' set to mid-range (50 %).
                    // How to "apply" the contrast and brightness percentage
                    //     to the logarithmized voltages (here: in dB) ?
                    // - 'Contrast' is a kind of GAIN adjustment, i.e. multiplied with the dB value
                    // - 'Brightness' is an OFFSET adjustment, i.e. added or subtracted from the brightness
                    flt = (  fltPeak/* from Icom: 0..127 dB relative to a phantasy reference (*) */
                           + 2.0*(float)pDispCtrl->iWFBrightness_Percent/* adds +/-100 dB to the "phantasy reference" */
                           - 127.0
                          ) // now MULTIPLY with the 'contrast' control value:
                        * ( 0.5 + 0.1 * (float)pDispCtrl->iWFContrast_Percent );
                       // (*) More on Icom's "phantasy unit" in their "waveform data"
                       //     in RigControl.c : RigCtrl_ParseSpectrumData_CIV() .....    
                    pixel_brightness = 127/* mid range of 0..255*/ + (int)flt;
                    LimitInteger( &pixel_brightness, 0, 255 );
                    switch( pDispCtrl->iWFColorPalette )
                     { case WF_PALETTE_BLUE :
                       default :  // Blue waterfall display
                          red   = pixel_brightness;  // -> 0..255 (for each colour component)
                          green = pixel_brightness;
                          blue  = pixel_brightness * 2;
                          break;
                       case WF_PALETTE_GREEN  : // Green waterfall display
                          red   = pixel_brightness;
                          green = pixel_brightness * 2;
                          blue  = pixel_brightness;
                          break;
                       case WF_PALETTE_SUNRISE: // 'sunrise' palette from Spectrum Lab
                          red   = SunriseColors[pixel_brightness] & 0xFF;
                          green =(SunriseColors[pixel_brightness] >> 8) & 0xFF;
                          blue  =(SunriseColors[pixel_brightness] >> 16) & 0xFF;
                          break;
                       case WF_PALETTE_LINRAD : // 'Linrad'-like palette (rainbow-like)
                          red   = LinradColors[pixel_brightness] & 0xFF;
                          green =(LinradColors[pixel_brightness] >> 8) & 0xFF;
                          blue  =(LinradColors[pixel_brightness] >> 16) & 0xFF;
                          break;
                     }
                    LimitInteger( &red,   0, 255 ); // just for safety..
                    LimitInteger( &green, 0, 255 );
                    LimitInteger( &blue,  0, 255 );

                  } // end if( iBinIndex > iPrevBinIdx )
                 else  // not a new FFT frequency bin index: use the same R,G,B colour as for the left neighbour
                  {
                  }

               } // end if( iBinIdx>0 && iBinIdx<pSpectrum->nBinsUsed )
              else // pixel position is not covered by a valid FFT frequency bin:
               { red = green = blue = 0;
               }
              pPixels->b[RGB_BYTE_RED]   = red;
              pPixels->b[RGB_BYTE_GREEN] = green;
              pPixels->b[RGB_BYTE_BLUE]  = blue;
              pPixels->b[RGB_BYTE_ALPHA] = 0x00;
              ++pPixels;
            }  // end for x...
         }    // end if( pSpectrum != NULL ) && ...
      }      // end  for(iLine... )
     pDispCtrl->fUpdateWaterfall = FALSE; // "done" (waterfall display is up-to-date now)
   }       // end if( nSpectra > 0 )
} // end SpecDisp_UpdateWaterfall()

//---------------------------------------------------------------------------
void SpecDisp_UpdateSpectrum(
        Graphics::TBitmap *pbmpDest,  // [out] Borland-VCL-style "TBitmap"
        T_SpecDispControl *pDispCtrl, // [in] layout info, config data, etc
        T_RigCtrlInstance *pRigCtrl)  // [in] "Rig Control" instance (provides
             // spectrum data, VFO frequency, and maybe even RIT / XIT offsets)
{
  T_RigCtrl_Spectrum *pSpectrum;
  double freq, fmin, fmax, fstep, substep;
  float flt, fltFreqToBinIdx, fltPeak, fltScalingFactor;
  int   i, n, tw, th, thh, iBinIdx, iPrevBinIdx, x,y, prev_x, prev_y, x_covered;
  TPoint *pPoints, poly_points[4];
  int     nPoints = 0;
  char   szLabel[44], sz16FormatString[16+4];
  T_SpecDispMarker *pMarker;


  if( ! SpecDisp_PrepareDrawing( pbmpDest, pDispCtrl, pRigCtrl ) )
   { return; // something wrong with any of the three inputs above
   }
  fmin = pDispCtrl->fmin_from_RigCtrl; // <- updated in SpecDisp_PrepareDrawing()..
  fmax = pDispCtrl->fmax_from_RigCtrl; // .. so don't call RigCtrl_GetDisplayableRadioFrequencyRange() here


  pPoints = (TPoint*)malloc( pDispCtrl->iCanvasWidth * sizeof(POINT) );
  if(pPoints==NULL) // oops .. malloc() failed !
   { return;
   }
  // NO "LAZY RETURN" after this point !

  fltScalingFactor = (float)pDispCtrl->iCanvasHeight;
  flt = pDispCtrl->fltSpectrumAmplMax_dB - pDispCtrl->fltSpectrumAmplMin_dB;
  if( flt > 0.0 ) // avoid div-by-zero
   { fltScalingFactor /= flt;
   }

  // First clear the background, then optionally fill "marker backgrounds",
  //             then draw the grid, then draw the curve(s)...
  pDispCtrl->Canvas->Brush->Color = (TColor)pDispCtrl->dwBackgndColor;
  pDispCtrl->Canvas->Pen->Color   = (TColor)pDispCtrl->dwGridColor;
  SpecDisp_FillRect( pDispCtrl, 0,0, pDispCtrl->iCanvasWidth,
                                     pDispCtrl->iCanvasHeight );

  for(i=0; i<SPECDISP_NUM_MARKERS; ++i)
   { pMarker = &pDispCtrl->marker[i];
     if( pMarker->fVisible && (pMarker->iGuiControl==GUI_CONTROL_SPECTRUM) )
      { // Adjust the position of certain "special" markers here:
        switch(i)
         { case SPECDISP_MARKER_VFO: // area that allows dragging the VFO (frequency) ..
              x = FreqToScreenX( pDispCtrl, pRigCtrl->dblVfoFrequency );
              pMarker->x1 = x-5;
              pMarker->x2 = x+5;
              pMarker->y1 = 0;
              pMarker->y2 = pDispCtrl->iCanvasHeight-1;
              if( pDispCtrl->iColourScheme == COLOUR_SCHEME_DARK )
               { // white text on black background -> darker marker
                 if( pDispCtrl->fDraggingVfoFrequency )
                  { pMarker->color.dw = RGB( 0x00, 0xA0, 0x00 );
                  }
                 else
                  { pMarker->color.dw = RGB( 0x00, 0x80, 0x00 );
                  }
               }
              else // COLOUR_SCHEME_DEFAULT : black text on white background
               { if( pDispCtrl->fDraggingVfoFrequency )
                  { pMarker->color.dw = RGB( 0xC0, 0xFF, 0xC0 );
                  }
                 else
                  { pMarker->color.dw = RGB( 0xC0, 0xFF, 0xC0 );
                  }
               }
              break; // end case SPECDISP_MARKER_VFO
           default:
              break;
         }
        pDispCtrl->Canvas->Brush->Color = (TColor)pMarker->color.dw;
        SpecDisp_FillRect( pDispCtrl, pMarker->x1, pMarker->y1,
                                      pMarker->x2, pMarker->y2);
      }
   }
  pDispCtrl->Canvas->Brush->Color = (TColor)pDispCtrl->dwBackgndColor;


  // Prepare a sufficiently SMALL font for the grid :
  pDispCtrl->Canvas->Font->Name  = "Arial";
  pDispCtrl->Canvas->Font->Size  = 8;
  pDispCtrl->Canvas->Font->Color = (TColor)pDispCtrl->dwGridColor;
  th = pDispCtrl->Canvas->TextHeight( "Q" );
  thh= th / 2; // half text height for alignment


  // The HORIZONTAL grid lines mark 10 dB steps ..
  flt = 0.0;
  prev_y = pDispCtrl->iCanvasHeight;
  while( flt <= 100.0/*dB*/ )
   { y = (int)(( flt - pDispCtrl->fltSpectrumAmplMin_dB ) * fltScalingFactor);
     y = pDispCtrl->iCanvasHeight - 1 - y; // same as in the curve-plotting loop below..
     SpecDisp_DrawLine( pDispCtrl, 0, y, pDispCtrl->iCanvasWidth, y );
     if( (y<(prev_y-thh)) && (y>thh) && (y<(pDispCtrl->iCanvasHeight-thh) ) )
      { sprintf( szLabel, "%d dB", (int)flt );
        prev_y = y-thh;
        SpecDisp_DrawText( pDispCtrl, pDispCtrl->iCanvasWidth - 40, prev_y, szLabel );
      }
     flt += 10.0/*dB*/;
   }

  // The VERTICAL grid lines must match the 'large ticks' on the frequency scale,
  //   thus use a common subroutine to calculate their positions.
  //   Considered "more important" than the amplitude grid, thus drawn afterwards.
  SpecDisp_GetFreqScaleParams( pDispCtrl, &freq, &fstep, &substep, sz16FormatString );
  n = 0;
  x_covered = 0;
  while( ((n++)<100) && (freq<fmax) ) // vertical grid lines ...
   { x = FreqToScreenX( pDispCtrl, freq );
     SpecDisp_DrawLine( pDispCtrl, x, 0/*y1*/, x, pDispCtrl->iCanvasHeight/*y2*/ );
     sprintf( szLabel, sz16FormatString, (double)(freq*1e-6) ); // Megahertz !
     tw = pDispCtrl->Canvas->TextWidth( szLabel );
     if( (x+tw/2) > x_covered )
      { SpecDisp_DrawText( pDispCtrl, x-tw/2, 2/*y:top*/, szLabel ); // frequency [MHz]
        x_covered = x+tw/2;
      }
     freq += fstep;
   }

  // If the "VFO frequency" is within the displayed range, show it in a simlar
  // way as Icom does in e.g. IC-7300 :
  //  * The  "RX" frequency marker is GREEN, and starts with a small triangle
  //      pointing down from the upper end of the SPECTRUM display.
  //  * The  "TX" frequency marker is ORANGE, and when equal to the "RX" frequency,
  //      appears as a green/orange dashed line .
  //  * If the mouse pointer is close enough to the VFO frequency to "drag" it,
  //      the marker
  x = FreqToScreenX( pDispCtrl, pRigCtrl->dblVfoFrequency );
      // ,--------------------------------|_____________|
      // '--> 2024-07: When inspecting pRigCtrl->dblVfoFrequency in the IDE here, the value was ok.
      //               But after STEPPING INTO FreqToScreenX(), the argument 'double freq_Hz' was trashed.
      //               Fixed by doing a "build all" with Borland C++ Builder V6 .
      //               Reason: Caching of pre-compiled HEADERS doesn't work
      //                       reliably anymore, at least not under "modern Windows".
      //                       Remember this when a similar bug bites again.
  if( (x>=0) && (x<pDispCtrl->iCanvasWidth) )
   { // the "current" VFO frequency is within the displayed range,
     // but is it the TRANSMIT and/or RECEIVE frequency ? (care for this later) ..
     pDispCtrl->Canvas->Pen->Color = (TColor)0x00C0FF; // "clOrange" doesn't exist in the VCL,
            // so use a hex constant(0xBBGRR; the eight RED BITS are the LSByte)
     SpecDisp_DrawLine( pDispCtrl, x, 0/*y1*/, x, pDispCtrl->iCanvasHeight/*y2*/ );
     // > Use Polygon() to draw a closed, many-sided shape on the canvas,
     // > using the value of Pen. After drawing the complete shape,
     // > Polygon() fills the shape using the value of Brush.
     // > The Points parameter is an array of points that give the vertices of the polygon.
     // > Note: The Points_Size parameter is the index of the last point
     // >       in the array (one less than the total number of points).
     // > The first point is always connected to the last point .
     poly_points[0] = Point( x-4, 0 ); // upper left corner of the triangle
     poly_points[1] = Point( x+4, 0 ); // upper right corner of the triangle
     poly_points[2] = Point( x ,  4 ); // lower corner (tip pointing down)
     pDispCtrl->Canvas->Brush->Color = pDispCtrl->Canvas->Pen->Color;
     pDispCtrl->Canvas->Polygon( poly_points, 2/*!*/ ); // draw triangle as a filled polygon
   } // end if < pRigCtrl->dblVfoFrequency within the displayed frequency range > ?

  // Plot the MOST RECENT spectrum as a curve into the FOREGROUND (thus last)
  int iTailIndex = RIGCTRL_GET_LATEST_ENTRY;
  pSpectrum = RigCtrl_GetSpectrumFromFIFO( pRigCtrl, &iTailIndex );
  if( pSpectrum != NULL ) // ok, got a valid spectrum so PLOT IT:
   { // Since neither the width of the window, not the FFT size is fixed anymore,
     // the ratio between 'screen pixels' and 'FFT frequency bins' is arbitrary.
     // Use the same principle to transform pixel coordinate (x)
     // into frequency bin numbers as in SpecDisp_UpdateWaterfall() :
     if( pSpectrum->dblBinWidth_Hz > 0.0 )
      { fltFreqToBinIdx = 1.0 / pSpectrum->dblBinWidth_Hz;
      }
     else // avoid div-by-zero..
      { fltFreqToBinIdx = 0.0;
      }
     iPrevBinIdx = -1;
     fltPeak = pSpectrum->fltMagnitudes_dB[0];
     for(int x = 0; x < pDispCtrl->iCanvasWidth; ++x)
      {
        // Convert the x-coord to a frequency (in Hertz):
        freq = ScreenXToFreq( pDispCtrl, x );
        // Convert the pixel's center frequency to an FFT frequency bin index:
        iBinIdx = (int)( 0.5 + (freq - pSpectrum->dblFmin_Hz) * fltFreqToBinIdx);
        if( iBinIdx>=0 && iBinIdx<pSpectrum->nBinsUsed )  // an IC-7300 delivered 475 frequency bins...
         {
           if( iPrevBinIdx < 0 )
            {  iPrevBinIdx = iBinIdx-1;  // 1st loop
               if( iPrevBinIdx < 0 )
                {  iPrevBinIdx = 0;
                }
            }
           if( iPrevBinIdx < iBinIdx ) // Is this a new FFT-bin at all ?
            { fltPeak = pSpectrum->fltMagnitudes_dB[++iPrevBinIdx];
              while( iPrevBinIdx < iBinIdx ) // MORE THAN ONE BIN per pixel : get the peak..
               { flt = pSpectrum->fltMagnitudes_dB[++iPrevBinIdx];
                 if( flt>fltPeak )
                  {  fltPeak = flt;
                  }
               }
            }
           // Convert Icom's "magnitude in dB" into a vertical pixel offset.
           //
           flt = ( fltPeak - pDispCtrl->fltSpectrumAmplMin_dB ) * fltScalingFactor;
           if( flt < 0.0 )
            {  flt = 0.0;
            }
           if( flt >= pDispCtrl->iCanvasHeight )
            {  flt =  pDispCtrl->iCanvasHeight-1;
            }
           y = pDispCtrl->iCanvasHeight - 1 - (int)flt;
           pPoints[nPoints].x   = x;  // append another point to the polyline
           pPoints[nPoints++].y = y;
         } // end if < valid FFT bin index >
      }  // end for x...

     if( nPoints >= 2 ) // draw the entire "curve" in a single GDI function call
      { pDispCtrl->Canvas->Pen->Color = (TColor)pDispCtrl->dwCurveColor;
        pDispCtrl->Canvas->Polyline( pPoints, nPoints-1 );
        // ,----------------------------------|_______|
        // '--> funny but true: Not the NUMBER OF POINTS,
        //                      but the NUMBER OF POINTS MINUS ONE !
      }
   } // end if( pSpectrum != NULL ) && ...
  pDispCtrl->fUpdateSpectrum = FALSE; // "done" (spectrum graph is up-to-date now)
  free(pPoints);

} // end SpecDisp_UpdateSpectrum()

//---------------------------------------------------------------------------
void SpecDisp_UpdateFreqScale(
        Graphics::TBitmap *pbmpDest,  // [out] Borland-VCL-style "TBitmap"
        T_SpecDispControl *pDispCtrl, // [in] layout info, config data, etc
        T_RigCtrlInstance *pRigCtrl)  // [in] "Rig Control" instance (with audio passband parameters)
{
  double freq, fmin, fmax, fstep, substep;
  int pixel_brightness, red, green, blue;
  int i,n,x,y,x1,y1,x2,y2,tw,th,x_covered,iDecoderIndex;
  int iLine;
  BOOL first;
  float flt, fltFreqToBinIdx, fltPeak, fltPeak_dBfs;
  int   iBinIdx, iPrevBinIdx;
  char  szLabel[80];    // also used for the yellow info text
  char  sz16FormatString[16+4];

  if( ! SpecDisp_PrepareDrawing( pbmpDest, pDispCtrl, pRigCtrl ) )
   { return; // something wrong with any of the three inputs above
   }

  fmin = pDispCtrl->fmin_on_FreqScale = pDispCtrl->fmin_from_RigCtrl; // <- polled LATER..
  fmax = pDispCtrl->fmax_on_FreqScale = pDispCtrl->fmax_from_RigCtrl; // .. to check if re-draw required,
     // just in case the user/operator changed the displayed range ON THE RADIO .
     //  fmin_from_RigCtrl and fmax_from_RigCtrl have been updated
     //  in SpecDisp_PrepareDrawing() already .

  pDispCtrl->Canvas->Brush->Color = (TColor)pDispCtrl->dwBackgndColor;
  pDispCtrl->Canvas->Pen->Color   = (TColor)pDispCtrl->dwGridColor;
  SpecDisp_FillRect( pDispCtrl, 0,0, pDispCtrl->iCanvasWidth,
                                     pDispCtrl->iCanvasHeight );

  SpecDisp_GetFreqScaleParams( pDispCtrl, &freq, &fstep, &substep, sz16FormatString );

  x1 = 0;
  x2 = pDispCtrl->iCanvasWidth - 1;
  y1 = 0;
  y2 = pDispCtrl->iCanvasHeight / 2;
# define L_HORZ_CLEARING 35
  x_covered = x1+L_HORZ_CLEARING;
  while( freq < fmax ) // draw frequency markers (vertical lines + labels) ...
   { x = FreqToScreenX( pDispCtrl, freq );
     if( x>x1 && x<x2 )  // inside the frequency scale
      {
        SpecDisp_DrawLine(pDispCtrl, x, y1, x, y2);  // draw frequency marker ("large tick")
        if( (x>x_covered) && (x+L_HORZ_CLEARING)<x2 )  // enough clearance for a frequency label ?
         { sprintf( szLabel, "%d Hz", (int)(freq+0.5) );
           SpecDisp_GetTextExtent( pDispCtrl, szLabel, &tw, &th );  // typical with ("1000 Hz") : 54 pixels wide
           SpecDisp_DrawText( pDispCtrl, x-(tw/2), y2, szLabel );
           x_covered = x + tw + 3/*pixels*/;  // just in case a label is 'surprisingly long'
         }
      }
     freq += fstep;
   }
  // Similar for the 'sub-steps' (with smaller ticks and no labels) :
  freq  = fmin;
  freq -= fmod( freq, substep );
  y2 = pDispCtrl->iCanvasHeight / 4;
  while( freq < fmax ) // draw frequency markers (vertical lines + labels) ...
   { x = FreqToScreenX( pDispCtrl, freq );
     if( x>x1 && x<x2 )  // inside the frequency scale
      { SpecDisp_DrawLine( pDispCtrl, x, y1, x, y2 ); // draw "small tick"
      }
     freq += substep;
   }
  pDispCtrl->fUpdateFreqScale = FALSE; // "done" (frequency scale is up-to-date now)


} // end SpecDisp_UpdateFreqScale()

//---------------------------------------------------------------------------
BOOL SpecDisp_HandleMouseEvent(  // processes any mouse-related event in the spectrum, spectrogram, or frequency scale
        T_SpecDispControl *pDispCtrl, // [in] layout info, config data, etc
        T_RigCtrlInstance *pRigCtrl,  // [in,out] "Rig Control" instance
        int iGuiEvent,                // [in] GUI_EVENT_MOUSE_DOWN, GUI_EVENT_MOUSE_MOVE, GUI_EVENT_MOUSE_UP
        int iGuiControlIndex,         // [in] GUI_CONTROL_SPECTRUM, GUI_CONTROL_SPECTROGRAM, GUI_CONTROL_FREQ_SCALE
        int iButtonsAndShiftKeys,     // [in] GUI_KEY_FLAGS_NONE, GUI_KEY_FLAGS_LEFT_MOUSE_BUTTON, etc etc (bitwise combined)
        int iClientX, int iClientY)   // [in] client coordinate of the mouse pointer (0,0 = upper left pixel of the visible control)
{
  // Common treatment for ALL THREE PARTS of the screen, and ALL THREE MOUSE-RELATED EVENTS:
  double freq  = ScreenXToFreq( pDispCtrl, iClientX );
  double span;
  int    x_vfo = FreqToScreenX( pDispCtrl, pRigCtrl->dblVfoFrequency );
  BOOL   fMouseNearVFO = (iClientX >= (x_vfo-5) ) && (iClientX <= (x_vfo+5) );
  BOOL   fHandledEvent = TRUE;
  BOOL   fVisible;

  if( pDispCtrl->iCanvasWidth <= 0 ) // avoid div-by-zero anywhere below
   { return FALSE;
   }

  // Specific treatment for mouse down/move/up :
  switch( iGuiEvent )
   { case GUI_EVENT_MOUSE_DOWN:
        // Stuff used in SpecDisp_HandleMouseEvent() :
        pDispCtrl->dblClickedVfoFreq_Hz = freq; // remember this for later (e.g. ..MOUSE_UP)
        if( fMouseNearVFO )
         { pDispCtrl->iVFOEditTimer_ms = -1; // avoid interference from the edit field
           pDispCtrl->fDraggingVfoFrequency = TRUE;
           pDispCtrl->iDraggingOffsetX = iClientX - x_vfo;
           pDispCtrl->fmin_at_drag_start = pDispCtrl->fmin_from_RigCtrl;
           pDispCtrl->fmax_at_drag_start = pDispCtrl->fmax_from_RigCtrl;
           // Similar as when typing a new frequency into edit field Ed_VFO
           // (and setting pDispCtrl->iVFOEditTimer_ms > 0),
           // fDraggingVfoFrequency = TRUE also temporarily cuts the connection
           // between 'RigControl' and the frequency displayed on the GUI.
           // In both cases, the background colour of the Ed_VFO (display field)
           // indicates "the new frequency may not have arrived
           //             on the remotely controlled radio yet".
         }
        else
         { pDispCtrl->fDraggingVfoFrequency = FALSE;
         }
        break;

     case GUI_EVENT_MOUSE_MOVE:
        fVisible = ( fMouseNearVFO || pDispCtrl->fDraggingVfoFrequency );
        if( pDispCtrl->marker[SPECDISP_MARKER_VFO].fVisible != fVisible )
         { pDispCtrl->marker[SPECDISP_MARKER_VFO].fVisible = fVisible;
           pDispCtrl->fUpdateSpectrum = TRUE;
         }
        if( pDispCtrl->fDraggingVfoFrequency )
         { // Similar to freq = ScreenXToFreq( pDispCtrl, iClientX );
           // but here using pDispCtrl->fmin_at_drag_start, fmax_at_drag_start
           //     instead of pDispCtrl->fmin_from_RigCtrl,  fmin_from_RigCtrl:
           span = pDispCtrl->fmax_at_drag_start - pDispCtrl->fmin_at_drag_start;
           if( span > 0.0 )
            { freq = pDispCtrl->fmin_at_drag_start
                      + ((double)(iClientX-pDispCtrl->iDraggingOffsetX) * span)
                       / (double)pDispCtrl->iCanvasWidth;
              pDispCtrl->iVFOEditTimer_ms = -1; // avoid interference from the edit field (user isn't TYPING there anymore)
              SpecDisp_UpdateVFODisplay(freq);  // -> Ed_VFO (a visible control element in the GUI)
              RigCtrl_SetVFOFrequency( pRigCtrl, freq );
            }
         }
        break;
     case GUI_EVENT_MOUSE_UP  :
        if( pDispCtrl->fDraggingVfoFrequency || pDispCtrl->marker[SPECDISP_MARKER_VFO].fVisible )
         { pDispCtrl->fUpdateSpectrum = TRUE;
         }
        if( pDispCtrl->fDraggingVfoFrequency )
         {  pDispCtrl->iVFOEditTimer_ms = -1; // edit field (Ed_VFO) back in default state
            // (may report a new frequency from RigControl now)
         }
        else // "mouse-up" event, but not after dragging the VFO frequency ->
         { fHandledEvent = FALSE; // let e.g. Keyer_Main.cpp handle this event, and open a context menu
         }
        pDispCtrl->fDraggingVfoFrequency = FALSE;
        pDispCtrl->marker[SPECDISP_MARKER_VFO].fVisible = FALSE;
        break;
     default:
        break;
   }


  return fHandledEvent;


} // end SpecDisp_HandleMouseEvent()

