//---------------------------------------------------------------------------
// File: C:\cbproj\Remote_CW_Keyer\SpecDisp.c
// Date: 2024-02-26
// Author: Wolfgang Buescher (DL4YHF)
// Purpose: Spectrum Display (waterfall and/or spectrum graph)
//          for the 'Remote CW Keyer' .
//      Unlike the main module for the GUI,
//      *this* module shall not have any dependencies
//      on Borland's VCL, Microsoft's don't-know-what-dot-something-framework,
//      or Qt   -    just PLAIN OLD WINDOWS GDI functions !
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Constants
//---------------------------------------------------------------------------

#define SPECTRUM_MIN_DB -110.0
#define SPECTRUM_MAX_DB 0.0


extern unsigned long SunriseColors[256]; // imported from SunriseColors.c
extern unsigned long LinradColors[256];  // imported from LinradColors.c

// Visible controls on the spectrum/spectrogram/frequency scale :
#define GUI_CONTROL_NONE        0
#define GUI_CONTROL_SPECTRUM    1
#define GUI_CONTROL_SPECTROGRAM 2 // .. aka 'waterfall'
#define GUI_CONTROL_FREQ_SCALE  3

// Codes for various 'GUI event handlers' :
#define GUI_EVENT_NONE       0
#define GUI_EVENT_MOUSE_DOWN 1 // abstraction of Borland's "OnMouseDown" (i.e. button JUST PRESSED or "pen down")
#define GUI_EVENT_MOUSE_MOVE 2 // abstraction of Borland's "OnMouseMove" (i.e. mouse coord changed while button pressed)
#define GUI_EVENT_MOUSE_UP   3 // abstraction of Borland's "OnMouseUp"   (i.e. mouse button released or "pen up")

// Bitwise combineable flags for function argument 'iButtonsAndShiftKeys',
//   used in mouse-event-handlers as a thin abstraction layer :
#define GUI_KEY_FLAGS_NONE               0
#define GUI_KEY_FLAGS_LEFT_MOUSE_BUTTON  0x0001
#define GUI_KEY_FLAGS_RIGHT_MOUSE_BUTTON 0x0002
#define GUI_KEY_FLAGS_SHIFT              0x0004
#define GUI_KEY_FLAGS_CONTROL            0x0008

//---------------------------------------------------------------------------
// Data types
//---------------------------------------------------------------------------

typedef union
{ DWORD dw;
  BYTE b[4];
} T_RGBColor;
#  define RGB_BYTE_BLUE  0
#  define RGB_BYTE_GREEN 1
#  define RGB_BYTE_RED   2
#  define RGB_BYTE_ALPHA 3

typedef struct t_SpecDispMarker
{
  BOOL fVisible;
  int x1,y1,x2,y2;
  int iGuiControl; // GUI_CONTROL_NONE, GUI_CONTROL_SPECTRUM / SPECTROGRAM / FREQ_SCALE
  T_RGBColor color;
} T_SpecDispMarker;


typedef struct t_SpecDispControl
{
  // Configuration parameters
  int iFreqScaleHeight;
  int iWFBrightness_Percent, iWFContrast_Percent;
  int iWFColorPalette; // one of the following values:
#   define WF_PALETTE_BLUE    0
#   define WF_PALETTE_GREEN   1
#   define WF_PALETTE_SUNRISE 2
#   define WF_PALETTE_LINRAD  3
  float fltSpectrumAmplMin_dB, fltSpectrumAmplMax_dB; // 0..127 dB on the "Icom scale"

  int iColourScheme;  // colour scheme for most GUI elements (and the frequency scale),
                      // may contain one of the following:
#   define COLOUR_SCHEME_DEFAULT 0
#   define COLOUR_SCHEME_DARK    1

  // Flags indicating which of the three parts (Spectrum, Waterfall, Frequency Scale)
  //   needs to be updated on the display:
  BOOL fUpdateSpectrum, fUpdateWaterfall, fClearWaterfall, fUpdateFreqScale;

  // Internal values ... most of them updated in each call of SpecDisp_UpdateWaterfall() / SpecDisp_UpdateSpectrum()
  DWORD clButtonFace; // ex: clBtnFace; actually the VCL form's BACKGROUND COLOUR
  DWORD clWindowText; // ex: clWindowText;
  DWORD clWindowBackground; // ex: clWindow; per default, WHITE background in e.g. combos and edit fields
  DWORD clActiveCaption;    // ex: clActiveCaption;   "color of the active window's title bar"
  DWORD clInactiveCaption;  // ex: clInactiveCaption; "color of inactive windows' title bar"
  DWORD clButtonFont;       // ex: the same as clWindowText

  DWORD dwBackgndColor/*for SPECTRUM DISPLAYS*/, dwGridColor, dwCurveColor;

  Graphics::TCanvas *Canvas; // A TCanvas can be painted on. A TBitmap is only the "pixel storage".
  int    iCanvasWidth, iCanvasHeight;
  double fmin_from_RigCtrl, fmax_from_RigCtrl;  // <- updated from the most recent T_RigCtrl_Spectrum
  BOOL   fixedEdgeMode;   // .. also info delivered by RigControl.c
  double fmin_on_FreqScale, fmax_on_FreqScale;  // <- currently displayed on the FREQUENCY SCALE
  int    nNewWaterfallLines;

  // Stuff used in various GUI event handlers (to display spectrum, VFO, etc) :
  double dblDisplayedVfoFreq_Hz; // VFO frequency (in Hertz) currently displayed in the GUI
  int    iVFOEditTimer_ms;       // "countdown" in milliseconds, > 0 while USER IS EDITING Ed_VFO in any way
  BOOL   fDraggingVfoFrequency;  // DRAGGING the VFO frequency (via mouse or touchscreen)
                                 // also temporarily stops updating the VFO display from 'RigControl'.
  double fmin_at_drag_start, fmax_at_drag_start;
  int    iDraggingOffsetX; // horizontal offset, in pixels, between initial mouse pointer and the dragged object (e.g. VFO)
  double dblClickedVfoFreq_Hz;   // "clicked" VFO frequency (in Hertz)
# define SPECDISP_NUM_MARKERS 4
# define SPECDISP_MARKER_VFO  0
  T_SpecDispMarker marker[SPECDISP_NUM_MARKERS];
} T_SpecDispControl;



//---------------------------------------------------------------------------
// Variables
//---------------------------------------------------------------------------



//---------------------------------------------------------------------------
// Function prototypes (details only in the implementation, SpecDisp.c)
//---------------------------------------------------------------------------
#ifndef CPROT  // For peaceful co-existence of "C" and "C++" ....
# ifdef __cplusplus
#   define CPROT extern "C"
# else
#   define CPROT
# endif  // ! "cplusplus"
#endif  // ndef CPROT

CPROT float Voltage_to_dBfs( float fltNormalizedVoltage/*0..1*/ );
CPROT void GetStepAndSubstepForScale( double dblMinStep, double *pdblLargeStep, double *pdblSubStep);
CPROT void SpecDisp_InitControl( T_SpecDispControl *pDispCtrl, int iFreqScaleHeight );
CPROT void SpecDisp_UpdateVFODisplay( double dblVfoFrequency_Hz ); // -> Ed_VFO (in the GUI)
CPROT void SpecDisp_UpdateWaterfall(
                 Graphics::TBitmap *pbmpDest,   // [out] Borland-VCL-style "TBitmap"
                 T_SpecDispControl *pCtrl,
                 T_RigCtrlInstance *pRigCtrl,
                 int iScopeDisplayTailIndex, int nSpectra );
CPROT void SpecDisp_UpdateSpectrum(
                 Graphics::TBitmap *pbmpDest,   // [out] Borland-VCL-style "TBitmap"
                 T_SpecDispControl *pCtrl,
                 T_RigCtrlInstance *pRigCtrl);
CPROT void SpecDisp_UpdateFreqScale(
                 Graphics::TBitmap *pbmpDest,   // [out] Borland-VCL-style "TBitmap"
                 T_SpecDispControl *pCtrl,
                 T_RigCtrlInstance *pRigCtrl );
CPROT BOOL SpecDisp_HandleMouseEvent( T_SpecDispControl *pDispCtrl,
                                      T_RigCtrlInstance *pRigCtrl,
        int iGuiEvent, int iGuiControlIndex, int iButtonsAndShiftKeys,
        int iClientX,  int iClientY);

/* EOF < Remote_CW_Keyer/SpecDisp.h > */

