//---------------------------------------------------------------------------
// File: C:\cbproj\Remote_CW_Keyer\RigControl.c
// Authors: Wolfgang Buescher (DL4YHF)
//       (the stripped-down variant for DL4YHF's "Remote CW Keyer" [RCW]
//        does NOT support the ugly "RS-BA1 protocol") .
// Date:   2024-11-12
// Purpose:
//  Initially just a parser for Icon's CI-V protocol, used for CAT control
//  over serial ports (including 'virtual COM ports' for USB).
//  Unlike the mimosa known as "Hamlib" (used in WSJT-X), RigControl.c
//  should work with ANY setting in the IC-7300's Setup / CI-V menu :
//        "Setup" .. "Connectors" .. "CI-V" :
//          "CI-V USB Port" : "Unlink from [REMOTE]" or "Link to [REMOTE]"
//          "CI-V USB Echo Back" : Either ON or OFF (WSJT-X needed "ON").
//
// Literature:
//  [IC7300FM]: Icom IC-7300 "Full Manual", Section 19 : "Control Command"
//              (contains the full set of CI-V commands in this radio)
//               C:\datasheets\Rig_Manuals\IC-7300\IC-7300_Full_Manual.pdf
//  [IC7610CIV]: Icom IC-7610 "CI-V Reference Guide", #A7380-7EX (2017)
//  [IC-R8600CIV]: IC-R8600 (RX) "CI-V Reference..",  #A7375-2EX-3a (2018)
//  [CIV_REF_MAN_V3_2]: "CI-V Reference Manual ver 3.2" from 2002, saved as
//               C:\datasheets\Rig_Manuals\IC-706\Icom_CI-V_manual.pdf .
//  [G3NRW_CIV]: IC-7300 TechNote: CI-V Controls  The Big Picture, G3NRW
//  [DF4OR_CIV]: http://www.plicht.de/ekki/civ/civ-p33.html .
//
// Revision History (latest entry first):
//
//   2024-11-25: Replaced HUNDREDS of "char pointers" by "const char *"
//               because 'modern compilers' (e.g. "ISO C++11") are
//               very pedantic about char pointers, and refuse to
//               use STRING LITERALS as a "char *" !
//             After that, most of the non-VCL-modules could be compiled
//             with the old Borland C++Builder V6 as well as with
//             Embarcadero C++Builder V12 "Athens", "Community Edition".
//               Details about the migration process from BCB V6 to V12
//               in DL4YHF's C:\cbproj\YHF_Tools\StringLib.h .
//
//   2024-11 : Entries in T_RigCtrlInstance.BandStackingRegs[RIGCTRL_NUM_BAND_STACKING_REGS]
//             shall now be distributed from the RCW-server to all RCW-clients
//             via the one-and-only TCP/IP connection between client and server,
//             so that all users (clients AND the server's "sysop") have the
//             same list of available BANDS or even BAND-STACKING REGISTERS.
//             To do this in the background, without causing a burst of network
//             traffic in CwNet.c, added T_CwNetClient.iBandStackingRegsTxIndex .
//
//   2024-06 : Added a "Hamlib Net rigctld"-compatible SERVER, which shares the
//             data with RigControl.c (e.g. current frequency, mode, and who-
//             knows-what). To avoid crowding RigControl.c even more, the
//             Hamlib-"rigctld"-emulation is in Remote_CW_Keyer/HamlibServer.c .
//
//   2024-02 : Made a stripped-down variant for the "Remote CW Keyer":
//              C:\cbproj\Remote_CW_Keyer\RigControl.c ( + *.h ) .
//              Threw out the support for Icom's undocumented/proprietary
//              "RS-BA1" protocol (that's the thing with those THREE
//              UDP PORTS used when connected to the radio via LAN (IC-9700)
//              or WLAN (IC-705). If you need all these whistles and bells,
//              try to let WFVIEW run side-by-side with the Remote CW Keyer.
//   2018-12 : Created for Spectrum Lab by DL4YHF . Original location:
//              C:\cbproj\SpecLab\RigControl\RigControl.c .
//
//---------------------------------------------------------------------------

#include "switches.h" // project specific 'compilation switches'

#include <string.h>     // memset() isn't really a STRING function, but..
#include <stdio.h>      // snprintf() isn't "I/O", but declared in string.h
#include <math.h>       // fmodl() ..
#include "Timers.h"     // high-resolution stopwatches, etc
#include "Utilities.h"  // UTL_FormatDateAndTime(), UTL_GetCurrentUnixDateAndTime(), etc
#include "Inet_Tools.h" // INET_GetU16FromLE(), etc
#include "RigControl.h" // header file for THIS module

#if( SWI_USE_HAMLIB_SERVER ) // build an application with integrated "Hamlib Net rigctld"-compatible server ?
# include "HamlibServer.h"
  // HLSRV_EnumerateRequiredPNs() is called from RigControl.c during the init-
  // phase to have all parameters cached when anyone wants them, instead of
  // repeatedly reading them from the radio. RigControl.c (not HamlibServer.c)
  // is the module that keeps track of everything.
  // Parameters that are expected to change (like S-meter, or SWR during transmit)
  // will be polled at a moderate speed by RigControl.c anyway, thus there is
  // no need to issue a real 'READ'-command to the radio on behalf of any
  // remote Hamlib client. If we allow a remote (and insecure) Hamlib client
  // to CONTROL (set, change) anything inside the radio is another issue.
#endif // SWI_USE_HAMLIB_SERVER ?


// Internal defines (only used within THIS module, thus not in the header file)
#define RIGCTRL_CLIENT_ACTION_NONE          0 // for iClientAction : "do nothing"
#define RIGCTRL_CLIENT_ACTION_RETRANSMIT    1 // immediately retransmit command from client to radio
#define RIGCTRL_CLIENT_ACTION_ENTER_FIFO    2 // store command in FIFO, retransmit or execute later
#define RIGCTRL_CLIENT_ACTION_FAKE_RESPONSE 3 // send a "fake response" without forwarding anything to the radio

#define RIGCTRL_MAX_PINGS_LOGGED 5
#define RIGCTRL_MAX_IDLES_LOGGED 5

// Internal forward references (because C compilers are still stupid..)
void RigCtrl_ForgetScopeFrequencyRanges( T_RigCtrlInstance *pRC );
BOOL RigCtrl_SetVFOFrequency_Internal( T_RigCtrlInstance *pRC, double dblFreqHz );
BOOL RigCtrl_SetOperatingMode_Internal( T_RigCtrlInstance *pRC, int iOpMode );
static BOOL RigCtrl_SetScopeSpan_Internal( T_RigCtrlInstance *pRC );
void RigCtrl_SendReadCommandFromFIFO( T_RigCtrlInstance *pRC );
BOOL RigCtrl_RegisterCIVReponseForClient(T_RigCtrlInstance *pRC, int iRigCtrlClientPort, T_RigCtrl_MsgFilter *pFilter);
long RigCtrl_GetCombinedCmdAndSubcodeForCIV( BYTE *pbMessage, int iMsgLength );
T_RigCtrl_ParamInfo * RigCtrl_GetInfoForUnifiedParameterNumber(int iUnifiedPN);
T_RigCtrl_ParamInfo * RigCtrl_GetInfoForCIVCommandAndSubcode(long i32CombinedCmdAndSubcode);
T_RigCtrl_RadioInfo * RigCtrl_GetRadioInfoByDefaultAddress( int iRadioCtrlProtocol, int iDefaultAddress );
BYTE RigCtrl_IntToBCD2( int iValue );  // converts an integer between 0 and 99 into BCD (single byte)
void RigCtrl_IntToBCD_CIV( long i32Value, BYTE *pbDest, int nDigits );
void RigCtrl_IntToBCD_MSDigitFirst( long i32Value, BYTE *pbDest, int nDigits );

BOOL IcomPkt_SendToken(T_RigCtrlInstance *pRC, BYTE magic, char *pszComment);
BOOL IcomPkt_SendConnInfo( T_RigCtrlInstance *pRC, int iPortType );
void IcomPkt_PeriodicHandler( T_RigCtrlInstance *pRC, int iPortType );
int _stdcall RigCtrl_TxFunctionForCivViaUDP( T_RigCtrlInstance *pRC, int iRigCtrlPort, BYTE *source, int n_bytes,
                              int fWaitForCompletion, int *piErrorCode, int *piNumMsWaited);


//----------------------------------------------------------------------------
// Global variables for hardcore debugging / post-mortem crash analysis.
//        Useful when the debugger's call stack shows nothing but garbage,
//        e.g. after an exception at the infamous address 0xFEEEFEEE .
//        ( C++Builder's debugger could still inspect SIMPLE GLOBAL
//          VARIABLES after most kinds of exceptions, but of course nothing
//          stack-based because the CPU- or task-stack was usually trashed.)
//----------------------------------------------------------------------------
#if(SWI_HARDCORE_DEBUGGING) // (1) = hardcore-debugging, (0)=normal compilation
 int RigCtrl_iLastSourceLine = 0; // WATCH THIS after crashing with e.g. "0xFEEEFEEE"  ...
# define HERE_I_AM__RIGCTRL()  RigCtrl_iLastSourceLine=__LINE__
# if( 0 )
    If certain worker threads of the application don't seem to terminate
    "politely", watch the following (complete list not HERE but in Keyer_Main.cpp):
     * GUI_iLastSourceLine
     * DSW_iLastSourceLine
     * DSP_iLastSourceLine
     * Keyer_iLastSourceLine
     * CwNet_iLastSourceLine
     * HttpSrv_iLastSourceLine
     * RigCtrl_iLastSourceLine
#  endif // (0) .. kludge to inspect the above variables in Borland via mouse-over ...
  CPROT void CheckSystemHealth(const char *pszModuleName, int iSourceLine); // <- periodically called via macro CHECK_SYSTEM_HEALTH()
  static  const char *s_ModuleName = "RigCtrl";
# define CHECK_SYSTEM_HEALTH() CheckSystemHealth(s_ModuleName,__LINE__)
  BYTE *RigCtrl_pbSource = NULL;
  int   RigCtrl_iScopeBin= 0;
  float *RigCtrl_pfltDest= NULL;
#else
# define HERE_I_AM__RIGCTRL()  /* nuffink   */
# define CHECK_SYSTEM_HEALTH() /* hamwanich */
#endif // SWI_HARDCORE_DEBUGGING ?


//----------------------------------------------------------------------------
// Global variables ...
//----------------------------------------------------------------------------

BOOL RigCtrl_fUpdateTrafficLog = FALSE; // flag signalling "call TDebugForm::UpdateTrafficLog() a.s.a.p" (or whoever DISPLAYS the traffic log on-screen, and CLEARS that flag when 'done')
long RigCtrl_i32MessageNrForTrafficLog = 0;
int  RigCtrl_iTrafficMonitorDisplayOptions = RIGCTRL_TRAFFIC_DISPLAY_OPTION_DISABLED;  // RigCtrl_AddToTrafficLog() disabled unless the application sets at least RIGCTRL_TRAFFIC_DISPLAY_OPTION_ENABLE
int  RigCtrl_iTrafficMonitorExtraColumns = RIGCTRL_TRAFFIC_EXTRA_COLUMN_TIME
                                         | RIGCTRL_TRAFFIC_EXTRA_COLUMN_TX_RX
                                         | RIGCTRL_TRAFFIC_EXTRA_COLUMN_PAYLOAD_LENGTH
                                         | RIGCTRL_TRAFFIC_EXTRA_COLUMN_COMMENTS;
     // The two variables above will be set to UConfig.iTrafficMonitorDisplayOptions and UConfig.iTrafficMonitorExtraColumns; somewhere in SpecMain.cpp.
     // But the above INITIALISATION VALUES will be used as DEFAULTS in Config.cpp : CFG_Load() !
BOOL RigCtrl_fTrafficMonitorPaused = FALSE; // FALSE=not "automatically paused on a trigger event", TRUE="something happened, PLEASE TAKE A LOOK.."
     // '--> Reason for "pausing" / "stopping" : e.g. RIGCTRL_TRAFFIC_DISPLAY_OPTION_PAUSE_ON_FULL_FIFO / .. STOP_ON_ERROR / ...

int  RigCtrl_ITU_Region = 1; // <- INITIALIZED global variable for simplicity :
       // ITU region 1 = Europe, Africa, FSU, Mongolia, Middle East, Iraq
       // ITU region 2 = North- and South America, Greenland, some east pacific islands
       // ITU region 3 = Far East, non-FSU Asia, Iran, Australia, Oceania




//--------------------------------------------------------------------------
// Implementation of functions
//--------------------------------------------------------------------------


//--------------------------------------------------------------------------
void RigCtrl_Init( T_RigCtrlInstance *pRC,
        int iRadioControlProtocol, // [in] e.g. RIGCTRL_PROTOCOL_ICOM_CI_V
        int iRadioDefaultAddress,  // [in] e.g. RIGCTRL_DEF_ADDR_NO_REMOTE_CTRL, RIGCTRL_DEF_ADDR_AUTO_DETECT, RIGCTRL_DEF_ADDR_IC_7300, etc^10 ...
        T_CFIFO *pRadioPortRxFifo, // [in] thread-safe circular FIFO to RECEIVE data from the radio
        T_CFIFO *pRadioPortTxFifo, // [in] thread-safe circular FIFO to TRANSMIT data to the radio
        int nClientPorts,          // [in] 0..RIGCTRL_MAX_CLIENT_PORTS
        int *piClientOptions )     // [in] array, one per client (*)
  // Initializes everything for a "single instance" .
  //
  // (*) These "clients" were initially just "CI-V port duplicators",
  //     connected to extra "COM ports" (in most cases, "com0com" virtual NULL MODEM CABLES).
  //     Later renamed into "Client #1" and "Client #2", at least for the
  //     configuration dialog in Spectrum Lab. Each of those clients can have
  //     its individual configuration. Thus using an ARRAY here !
{
  int i, iClient, iRigCtrlPort, iClientOptions;
  T_RigCtrl_ParamInfo *pPI;

  if( nClientPorts > RIGCTRL_MAX_CLIENT_PORTS )
   {  nClientPorts = RIGCTRL_MAX_CLIENT_PORTS; // oops.. more than expected !?
   }

  memset( pRC, 0, sizeof( T_RigCtrlInstance ) );
  pRC->initState = initState_ColdStart;     // initState_ColdStart /  _Opened / _Closed ?
  pRC->iRadioCtrlProtocol = iRadioControlProtocol;
  pRC->iDefaultAddress  = iRadioDefaultAddress; // RIGCTRL_DEF_ADDR_NO_REMOTE_CTRL, RIGCTRL_DEF_ADDR_AUTO_DETECT, RIGCTRL_DEF_ADDR_IC_7300, etc^10 ...
  pRC->iRadioMasterAddr = 0xE0;   // Icom-slang: 0xE0 = "controller" (in this case, the PC controlling the radio)
  pRC->pRadioPortTxFifo = pRadioPortTxFifo; // thread-safe circular FIFO to TRANSMIT data to the radio
  pRC->pRadioPortRxFifo = pRadioPortRxFifo; // thread-safe circular FIFO to RECEIVE data from the radio

  pRC->u32RxAudioSamplesPerSecond = pRC->u32TxAudioSamplesPerSecond = 48000;
  pRC->dblUnixTimeOfFirstSpectrum = 0.0;

  for(iClient=0; iClient<nClientPorts; ++iClient)
   { iRigCtrlPort = iClient + RIGCTRL_PORT_CLIENT_1;
     iClientOptions = piClientOptions[iClient]; // individual settings for each client !
     pRC->PortInstance[iRigCtrlPort].iClientOptions = iClientOptions;
     pRC->PortInstance[iRigCtrlPort].fEmulateCIVEcho = (iClientOptions & RIGCTRL_CLIENT_OPTION_EMULATE_CIV_ECHO) != 0;
   }
  pRC->iRigControlFlags = RIGCTRL_FLAGS_NONE; // don't SEND but READ setttings from the remote rig
  pRC->iCapabilities    = RIGCTRL_CAPS_ALL;   // be optimistic about FUTURE rigs .. IC-9700 anyone ?
  pRC->dwAvailableBands = RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_6M | RIGCTRL_BAND_2M | RIGCTRL_BAND_70CM;  // <- that's an IC-705 (until we know better)
  pRC->iMaxResponseDelay_ms = 1500; // maximum delay for getting a response, in milliseconds
  pRC->dblVfoFrequency = RIGCTL_NOVALUE_DOUBLE;
  pRC->iOpMode = RIGCTRL_OPMODE_UNKNOWN; // later: USB, LSB, CW, ...
  pRC->iParameterPollingState = RIGCTRL_POLLSTATE_PASSIVE; // don't do anything until the application calls RigCtrl_StartReading()
  pRC->iResponseCountdown_ms = 0;
  pRC->iScopeSpeed = RIGCTRL_SCOPE_SPEED_FAST;

  // Set everything that we MAY(!) read from the radio via CI-V to "unknown" / "not read yet".
  // This way, the "report" in SL's debug-window / CAT traffic monitor
  //      can spit out THOSE PARAMETERS THAT HAVE REALLY BEEN ACCESSED,
  //      and omits unused entries (in RigCtrl_ParameterInfo[]) .
  pPI = (T_RigCtrl_ParamInfo*)RigCtrl_ParameterInfo;
  while( pPI->iUnifiedPN != 0 )
   { if( ( pPI->iByteOffsetIntoRigCtrlInstance > 0 )
       &&( pPI->iByteOffsetIntoRigCtrlInstance < sizeof(T_RigCtrlInstance) ) )
      { switch ( pPI->iRigCtrlDataType )
         { case RIGCTRL_DT_INT  :
              *(int*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance ) = RIGCTL_NOVALUE_INT;
              break;
           case RIGCTRL_DT_DOUBLE:
              *(double*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance ) = RIGCTL_NOVALUE_DOUBLE;
              break;
           default:
              break;
         } // end switch ( pPI->iRigCtrlDataType )
      }
     ++pPI;
   } // end while < all "unified parameters" >

  RigCtrl_ForgetScopeFrequencyRanges( pRC );
  pRC->initState = initState_Opened; // initState_ColdStart /  _Opened / _Closed ? here: 'Opened' when returning from RigCtrl_Init()

} // end RigCtrl_Init()


//-------------------------------------------------------------------- API --
void RigCtrl_Close( T_RigCtrlInstance *pRC ) // closes handles, sockets, frees resources
  // Called internally when necessary, or via SL's main menu: "Start/Stop" .. "Rig Control" .. "Start" .
{
  // Won't stop the threads running in IOaccess.cpp, but will stop functions called from there:
  //    (closing those threads, and closing "COM" ports isn't RigControl's part
  //     because RigControl neither opened those ports or created those threads) :
  if( pRC->initState == initState_Opened )
   {  pRC->initState = initState_Closed; // here: 'Closed' in
     // Give any RigControl-subroutines and "handlers" the chance to return,
     // before closing sockets, freeing memory, and making pointers invalid:
     Sleep( 50 ); // should be long enough to return from e.g RigCtrl_Handler() to IO_RadioCommThdFunc() .
   }

} // end RigCtrl_Close()


//-------------------------------------------------------------------- API --
void RigCtrl_ReConnect( T_RigCtrlInstance *pRC )
  // Politely closes an existing connection (if any),
  // clears the 'counters' that limit certain messages from the 'traffic log',
  // and finally tries to re-connect the rig.
  // Keeps all previous settings from RigCtrl_SetLANParams() INTACT !
  //
  // Called for example when clicking the "Apply"-button on SL's 'Radio Control' panel,
  //        or via SL's main menu: "Start/Stop" .. "Rig Control" .. "Start" .
{ char sz80Msg[84], *cp;
  int iMsgTypeForTrafficLog;

  RigCtrl_Close( pRC );
  if( pRC->initState == initState_Closed )
   {  pRC->initState =  initState_Opened; // allow e.g. RigCtrl_ProcessIcomUDP(), RigCtrl_Handler(), and RigCtrl_ProcessRxData() to "work"
   }

  strcpy( sz80Msg, "RigControl: Re-connect" );  // show what's going on ..
  iMsgTypeForTrafficLog = RIGCTRL_MSGTYPE_ERROR;
  RigCtrl_AddToTrafficLog( pRC, 0/*iRigCtrlPort*/, NULL, 0, iMsgTypeForTrafficLog, sz80Msg );
  RigCtrl_StartReading( pRC ); // -> RigCtrl_OpenSocket() -> RigCtrl_LaunchUDPThread() ...

} // end RigCtrl_ReConnect()


//-------------------------------------------------------------------- API --
void RigCtrl_SetClientOptions( T_RigCtrlInstance *pRC, int iRigCtrlPort,
         int iClientOptions ) // [in] bitwise combineable options like ..
         //   RIGCTRL_CLIENT_OPTION_REJECT_UNSOLICITED_MSGS ,
         //   RIGCTRL_CLIENT_OPTION_EMULATE_CIV_ECHO, ..(?)
{
  if( (iRigCtrlPort>=RIGCTRL_PORT_RADIO) && (iRigCtrlPort<=/*!*/RIGCTRL_MAX_CLIENT_PORTS) )
   { pRC->PortInstance[iRigCtrlPort].iClientOptions = iClientOptions;
     pRC->PortInstance[iRigCtrlPort].fEmulateCIVEcho = (iClientOptions & RIGCTRL_CLIENT_OPTION_EMULATE_CIV_ECHO) != 0;
   }
} // end RigCtrl_SetClientOptions()


//--------------------------------------------------------------------------
void RigCtrl_ForgetScopeFrequencyRanges( T_RigCtrlInstance *pRC )
{  int iRange, iEdge;
   pRC->iScopeNumFreqRanges = pRC->iScopeNumEdgesPerFreqRange = 0;
   for( iRange=0; iRange<RIGCTRL_MAX_FIXED_EDGE_SCOPE_FREQ_RANGES; ++iRange)
    { for( iEdge=0; iEdge<RIGCTRL_MAX_FIXED_EDGES_PER_FREQ_RANGE; ++iEdge)
       { pRC->ScopeFreqRange[iRange][iEdge].dblFmin_Hz = 0.0;
         pRC->ScopeFreqRange[iRange][iEdge].dblFmin_Hz = 0.0;
       }
    }
} // end RigCtrl_ForgetScopeFrequencyRanges()



//--------------------------------------------------------------------------
void RigCtrl_StartReading( T_RigCtrlInstance *pRC )
  // Prepares reading "all we need to know" about the remotely controlled rig,
  //  but doesn't really 'communicate' via serial port, LAN or WLAN yet.
  //  For Icom radios connected via LAN or WLAN, also tries to open the
  //  required ports (in this case UDP sockets), and prepares to log-in .
  // Called from the application (e.g. TConfigDlg::ApplyMyControls() ),
  //        or (if a simple SERIAL PORT is used for CI-V) IOaccess.cpp : IO_OpenComPorts(),
  //        or (if UDP is used for CI-V) called from "RigControl.c" itself,
  //   to begin the first PARAMETER POLLING CYCLE via CI-V .
  // Communication parameters like pRC->iRadioDeviceAddr, iRadioCtrlProtocol
  //        and  [removed] iInetProtocol ( INET_PROTOCOL_OFF/HTTP/TCP/UDP/RAW/COM_PORT)
  //   must have been set by the application prior to this.
  // The rest (communication) happens in RigCtrl_Handler(); details there.
{

  // 'Forget' various counters and values read during a previous session:
  pRC->iRadioPortTxFifoHead = pRC->iRadioPortTxFifoTail = 0;
  pRC->nCompleteSpectraReceived   = 0;
  pRC->dblUnixTimeOfFirstSpectrum = 0.0;
   
  pRC->iParameterPollingState    = RIGCTRL_POLLSTATE_START_RD;
  pRC->iParameterPollingSubState = pRC->iParameterPollingArrayIndex = 0; // don't leave anything to fate !
  pRC->iResponseCountdown_ms = 0;

} // end RigCtrl_StartReading()

//--------------------------------------------------------------------------
BOOL RigCtrl_SendAndLogMessage( T_RigCtrlInstance *pRC,
        int iRigCtrlPort, // [in] RIGCTRL_PORT_RADIO / RIGCTRL_PORT_CLIENT_1 / 2 ?
        BYTE *pbMessage, int iMsgLength, // [in] message, including pre- and postamble but not the UDP-CIV-HEADER (!)
        BOOL fWaitForCompletion, // [in] TRUE=yes (blocking call), FALSE=no .
        int iMsgType,     // [in] message type category and 'flags', only for the "traffic log"
                          //      example:
        char *pszComment) // [in] human readable info, short enough for the "traffic log"
  // Sends any CI-V message, and (if successful) counts and optional logs it.
{
  int iResult, iErrorCode, iNumMsWaited; // optional results from IO_SendToCOMPort()
  char sz255Msg[256];

  if( pRC->iRadioCtrlProtocol == RIGCTRL_PROTOCOL_ICOM_CI_V )
   { iMsgType |= RIGCTRL_MSGTYPE_FLAG_CIV; // <- for the TRAFFIC LOG: pbMessage points to pure CI-V payload, without overhead from any "transport layer" (and no "UDP-CIV-Header")
     pRC->pszCommentForTxFunction = pszComment; // <- added 2023-07 to avoid yet another parameter for the (UDP-) transmit function called from RigCtrl_SendAndLogMessage()
   }
  pRC->iMsgTypeForTxFunction = iMsgType;   // <- added 2023-07 so RigCtrl_TxFunctionForCivViaUDP() can pass 'iMsgType' on to RigCtrl_AddToTrafficLog(),
                                           //    especially for flags like RIGCTRL_MSGTYPE_FLAG_SUGGEST_NEW_HEADLINE_FOR_TRAFFIC_LOG .

  if( (iRigCtrlPort==RIGCTRL_PORT_RADIO) && (pRC->pRadioPortTxFifo!=NULL) )
   { iResult = CFIFO_Write( pRC->pRadioPortTxFifo, pbMessage, iMsgLength );
     if( iResult >= 0 ) // only a NEGATIVE result indicates a 'real' error. Zero = FALSE = "not finished yet"
      { // Modified 2023-08-15 : If the port is a COM port, RIGCTRL_TRAFFIC_DISPLAY_OPTION_SHOW_UDP_PAYLOAD
        // is meaningless . Break up the too complex 'if' condition further below.
        BOOL fShowInLog = ( RigCtrl_iTrafficMonitorDisplayOptions
                          & RIGCTRL_TRAFFIC_DISPLAY_OPTION_ENABLE) != 0;
        if( pRC->dwRejectMessages4Log & (1<<iMsgType) )
         { fShowInLog = FALSE;
         }
        if( pRC->dwRejectMessages4Log & RIGCTRL_MSGTYPE_FLAG_PERIODIC_POLL )
         { if( iMsgType & RIGCTRL_MSGTYPE_FLAG_PERIODIC_POLL )
            { fShowInLog = FALSE;
            }
         }
        if( RigCtrl_iTrafficMonitorDisplayOptions & RIGCTRL_TRAFFIC_DISPLAY_OPTION_SHOW_UDP_PAYLOAD )
         { //  With RIGCTRL_TRAFFIC_DISPLAY_OPTION_SHOW_UDP_PAYLOAD SET, RigCtrl_TxFunctionForCivViaUDP() will have
           //        already called RigCtrl_AddToTrafficLog(), for reasons explained THERE.
           // Hmmm... but DO WE USE UDP AT ALL ?
         } // end if .. RIGCTRL_TRAFFIC_DISPLAY_OPTION_SHOW_UDP_PAYLOAD ?
        if( fShowInLog )
         {  RigCtrl_AddToTrafficLog( pRC, iRigCtrlPort, pbMessage, iMsgLength, iMsgType | RIGCTRL_MSGTYPE_FLAG_TX, pszComment );
         }
        ++pRC->PortInstance[iRigCtrlPort].dwNumMessagesSent;
        pRC->PortInstance[iRigCtrlPort].fLastSendFailed = FALSE;
        return TRUE;
      }
     else // CFIFO_Write( pRC->pRadioPortTxFifo ) failed .. "tried to send more than possible" ?
      { if( !pRC->PortInstance[iRigCtrlPort].fLastSendFailed )
         {   pRC->PortInstance[iRigCtrlPort].fLastSendFailed = TRUE;
           snprintf( sz255Msg,250, "Failed to send to %s: code %d",
                RigCtrl_RigControlPortNrToString(iRigCtrlPort), iErrorCode );
           RigCtrl_AddToTrafficLog( pRC, iRigCtrlPort, NULL, 0, RIGCTRL_MSGTYPE_ERROR, sz255Msg );
         }
      }
   } // end if < something 'sendable' > ?
  return FALSE;

} // end RigCtrl_SendAndLogMessage()


//--------------------------------------------------------------------------
BOOL RigCtrl_SendShortCIVMessage( T_RigCtrlInstance *pRC,
        int iRigCtrlPort, // [in] RIGCTRL_PORT_RADIO / RIGCTRL_PORT_CLIENT_1 / 2 ?
        int iMsgType,     // [in] message type category and 'flags', only for the "traffic log"
        char *pszComment, // [in] human readable info, short enough for the "traffic log"
        int iCmd,         // [in] CI-V main command like 0x27 for "Scope waveform data"
        int iSubCmd,      // [in] optional sub-command.        -1 to omit.
        int iDataByte1,   // [in] optional 1st parameter byte. -1 to omit.
        int iDataByte2 )  // [in] optional 2nd parameter byte. -1 to omit.
  // Sends a 'short' CI-V message, usually a 'request' for something.
  // To send WITHOUT a "sub-command" (byte), use iSubCmd = -1.
  // To send WITHOUT a "data byte", use iDataByte1 = -1,  etc.
  // Returns TRUE when successfully sent, otherwise FALSE. KISS.
  // In Spectrum Lab, sent messages MAY be displayed by [v] Show CAT traffic
  //    with a YELLOW background under 'View/Windows' / 'CAT traffic monitor'.

{
  BYTE b16TxBuffer[16];
  BYTE *pbBufPtr = b16TxBuffer;
  int  iMsgLength;
  *pbBufPtr++ = 0xFE;  // fixed preamble code for CI-V protocol (term used by Icom .. NOT "Begin of Message" ! )
  *pbBufPtr++ = 0xFE;  // fixed preamble code for CI-V protocol
  *pbBufPtr++ = (BYTE)pRC->iRadioDeviceAddr;  // "to" transceiver's address
  *pbBufPtr++ = (BYTE)pRC->iRadioMasterAddr;  // "from" controller's address
  if( iCmd >= 0 )
   { *pbBufPtr++ = (BYTE)iCmd;  // e.g. 0x03 = CI-V command "Read operating frequency"
   }
  if( iSubCmd >= 0 )
   { *pbBufPtr++ = (BYTE)iSubCmd;
   }
  if( iDataByte1 >= 0 )
   { *pbBufPtr++ = (BYTE)iDataByte1;
   }
  if( iDataByte2 >= 0 )
   { *pbBufPtr++ = (BYTE)iDataByte2;
   }
  *pbBufPtr++ = 0xFD;  // CI-V code for "End Of Message" (term used by Icom .. NOT "postamble", even though 0xFE is called "preamble" ! )

  iMsgLength = pbBufPtr - b16TxBuffer;
  return RigCtrl_SendAndLogMessage( pRC, iRigCtrlPort, b16TxBuffer, iMsgLength,
          FALSE/*fWaitForCompletion*/, RIGCTRL_MSGTYPE_FLAG_CIV | iMsgType, pszComment );
} // end RigCtrl_SendShortCIVMessage()


//--------------------------------------------------------------------------
BOOL RigCtrl_SendMidSizeCIVMessage( T_RigCtrlInstance *pRC,
        int iRigCtrlPort, // [in] RIGCTRL_PORT_RADIO / RIGCTRL_PORT_CLIENT_1 / 2 ?
        int iMsgType,     // [in] message type category and 'flags', only for the "traffic log"
        char *pszComment, // [in] human readable info, short enough for the "traffic log"
        int iCmd,         // [in] CI-V main command like 0x27 for "Scope waveform data"
        int iSubCmd,      // [in] optional sub-command.        -1 to omit.
        int iDataByte1,   // [in] optional 1st parameter byte. -1 to omit.
        int iDataByte2,   // [in] optional 2nd parameter byte. -1 to omit.
        int iDataByte3,   // [in] optional 3rd parameter byte. -1 to omit.
        int iDataByte4 )  // [in] optional 4th parameter byte. -1 to omit.
  // Sends a 'mid-sized' CI-V message, for example a "Scope Reference Level Setting".
  // To send WITHOUT a "sub-command" (byte), use iSubCmd = -1.
  // To send WITHOUT a "data byte", use iDataByte1 = -1,  etc.
  // Returns TRUE when successfully sent, otherwise FALSE. KISS.
{
  BYTE b16TxBuffer[16];
  BYTE *pbBufPtr = b16TxBuffer;
  int  iMsgLength;
  *pbBufPtr++ = 0xFE;  // fixed preamble code for CI-V protocol (term used by Icom .. NOT "Begin of Message" ! )
  *pbBufPtr++ = 0xFE;  // fixed preamble code for CI-V protocol
  *pbBufPtr++ = (BYTE)pRC->iRadioDeviceAddr;  // "to" transceiver's address
  *pbBufPtr++ = (BYTE)pRC->iRadioMasterAddr;  // "from" controller's address
  if( iCmd >= 0 )
   { *pbBufPtr++ = (BYTE)iCmd;    // e.g. 0x27 = CI-V command "Scope Settings"
   }
  if( iSubCmd >= 0 )
   { *pbBufPtr++ = (BYTE)iSubCmd; // e.g. 0x19 = "Scope Reference level settings" (in an IC-9700)
   }
  if( iDataByte1 >= 0 )
   { *pbBufPtr++ = (BYTE)iDataByte1; // e.g. 0x00 = Main scope, 0x01 = Sub Scope (IC-9700 CI-V Manual page 24)
   }
  if( iDataByte2 >= 0 )
   { *pbBufPtr++ = (BYTE)iDataByte2; // e.g. 10- and 1-dB digit (in the usual BCD format)
   }
  if( iDataByte3 >= 0 )
   { *pbBufPtr++ = (BYTE)iDataByte3; // e.g. 0.1- and 0.01-dB digit (the latter often fixed to ZERO)
   }
  if( iDataByte4 >= 0 )
   { *pbBufPtr++ = (BYTE)iDataByte4; // e.g. 0x00 = POSITIVE,  0x01 = NEGATIVE (quite counter-intuitive..)
   }
  *pbBufPtr++ = 0xFD;  // CI-V code for "End Of Message" (term used by Icom .. NOT "postamble", even though 0xFE is called "preamble" ! )

  iMsgLength = pbBufPtr - b16TxBuffer;
  return RigCtrl_SendAndLogMessage( pRC, iRigCtrlPort, b16TxBuffer, iMsgLength,
          FALSE/*fWaitForCompletion*/, RIGCTRL_MSGTYPE_FLAG_CIV | iMsgType, pszComment );
} // end RigCtrl_SendMidSizeCIVMessage()

//--------------------------------------------------------------------------
BOOL RigCtrl_SendCommandToTurnRigOn( T_RigCtrlInstance *pRC,
        int iRigCtrlPort ) // [in] RIGCTRL_PORT_RADIO / RIGCTRL_PORT_CLIENT_1 / 2 ?
  // Sends a special CI-V message to turn a modern Icom radio on, via CI-V.
  // Returns TRUE when successfully sent, otherwise FALSE.
{
  BYTE b256TxBuffer[256];
  BYTE *pbBufPtr = b256TxBuffer;
  int  i, iMsgLength;

  // From one of the dozens of Icom's "CI-V Reference Guides" ("A7380-7EX"):
  // > When sending the power ON command (18 01), you need
  // > to repeatedly send "FE" before the standard format.
  // > The following is the approximate number of needed repetitions.
  // >  * 115200 bps: 150 "FE"s
  // >  *  57600 bps: 75 "FE"s
  // >  *  38400 bps: 50 "FE"s
  // >  *  19200 bps: 25 "FE"s
  // >  *   9600 bps: 13 "FE"s
  // >  *   4800 bps: 7 "FE"s
  for( i=0; i<150; ++i ) // keep it simple, assume we need the MAXIMUM number of "FE"s
   { *pbBufPtr++ = 0xFE;  // fixed preamble code for CI-V protocol, and obviously also Icom's "wake-up pattern"
   }
  *pbBufPtr++ = (BYTE)pRC->iRadioDeviceAddr;  // "to" transceiver's address
  *pbBufPtr++ = (BYTE)pRC->iRadioMasterAddr;  // "from" controller's address
  *pbBufPtr++ = 0x18; // 0x18 = CI-V command "Turn OFF / ON the transceiver"..
  *pbBufPtr++ = 0x01; // 0x01 = subcommand to turn *ON*, not off
  *pbBufPtr++ = 0xFD; // CI-V code for "End Of Message"
  iMsgLength = pbBufPtr - b256TxBuffer;
  return RigCtrl_SendAndLogMessage( pRC, iRigCtrlPort, b256TxBuffer, iMsgLength,
          FALSE/*fWaitForCompletion*/, RIGCTRL_MSGTYPE_FLAG_CIV | RIGCTRL_MSGTYPE_POWER_ON_OFF, "turn ON" );
} // end RigCtrl_SendCommandToTurnRigOn()


//--------------------------------------------------------------------------
void RigCtrl_AppendByteToRadioPortTxFifo( BYTE *pbBuffer, int *piHeadIndex, BYTE b )
{
  *piHeadIndex &= (RIGCTL_RADIO_PORT_TX_FIFO_SIZE-1); // safety first..
  pbBuffer[*piHeadIndex] = b;
  *piHeadIndex = (*piHeadIndex + 1) & (RIGCTL_RADIO_PORT_TX_FIFO_SIZE-1);
}

//--------------------------------------------------------------------------
BYTE RigCtrl_ReadByteFromRadioPortTxFifo( BYTE *pbBuffer, int *piTailIndex )
{
  BYTE b;
  *piTailIndex &= (RIGCTL_RADIO_PORT_TX_FIFO_SIZE-1); // safety first..
  b = pbBuffer[*piTailIndex];
  *piTailIndex = (*piTailIndex + 1) & (RIGCTL_RADIO_PORT_TX_FIFO_SIZE-1);
  return b;
}


//--------------------------------------------------------------------------
BOOL RigCtrl_AppendMessageToRadioPortTxFifo( T_RigCtrlInstance *pRC,
        int iRigCtrlClientPort, // [in] RIGCTRL_PORT_CLIENT_1 / 2 ?
        BYTE *pbMessage, int iMsgLength ) // [in] message including pre- and postamble
  // Called from RigCtrl_ParseCIV() after receiving certain messages
  // from any CLIENT that need to be forwarded to the RADIO PORT,
  // but cannot be sent immediately because the RADIO is already busy
  // from a previously transmitted command (with response is still pending).
  //
{
  int nBytesOccupied, nBytesFree;
  int iHeadIndex = pRC->iRadioPortTxFifoHead;
  int iTailIndex = pRC->iRadioPortTxFifoTail;
  BYTE *pbBuffer = pRC->bRadioPortTxFifo;

  if( iMsgLength <= 0 )  // don't trash the TX-FIFO with negative lengths
   { return FALSE;
   }

  nBytesOccupied = iHeadIndex - iTailIndex;
  if( nBytesOccupied < 0 ) // circular wrap, e.g. head=0, tail=2047 ->
   {  nBytesOccupied += RIGCTL_RADIO_PORT_TX_FIFO_SIZE;
   }
  // Remember, a classic circular 2048-byte-FIFO has a maximum USEABLE capacity
  //   of 2047(!) bytes, because HeadIndex==TailIndex means buffer EMPTY (not FULL).
  //   The buffer is completely full when (HeadIndex+1)==TailIndex. Thus:
  nBytesFree = RIGCTL_RADIO_PORT_TX_FIFO_SIZE - 1 - nBytesOccupied;

  // TWO extra bytes (before the message itself) are used to store the LENGTH.
  // ONE additional byte is used to store the CLIENT PORT NUMBER on which
  //     the message was originally received .
  // This way, we can store anything in the FIFO without knowing the structure.
  // It doesn't have to be CI-V !
  if( nBytesFree < (iMsgLength+3) ) // bad luck, running out of TX-FIFO-space !
   { return FALSE;
   }

  // Enter the MESSAGE LENGTH and CLIENT PORT in the FIFO (low byte first as on an Intel CPU)
  RigCtrl_AppendByteToRadioPortTxFifo( pbBuffer, &iHeadIndex, (BYTE)iMsgLength );
  RigCtrl_AppendByteToRadioPortTxFifo( pbBuffer, &iHeadIndex, (BYTE)(iMsgLength >> 8) );
  RigCtrl_AppendByteToRadioPortTxFifo( pbBuffer, &iHeadIndex, (BYTE)iRigCtrlClientPort );

  while( (iMsgLength--) > 0 )
   { RigCtrl_AppendByteToRadioPortTxFifo( pbBuffer, &iHeadIndex, *(pbMessage++) );
   }
  pRC->iRadioPortTxFifoHead = iHeadIndex; // another fragment is now "occupied"

  return TRUE;  // successfully appended the message to the TX-FIFO

} // end RigCtrl_AppendMessageToRadioPortTxFifo()

//--------------------------------------------------------------------------
BOOL RigCtrl_SendMessageFromRadioPortTxFifo( T_RigCtrlInstance *pRC,
        T_RigCtrl_MsgFilter *pFilter) // [out] command, subcommand, and possibly accepted message type
  // Transmits the next message (if any) that has previously been
  //           queued up in the *RADIO PORT's* transmit FIFO, TO the radio
  //   - see details in RigCtrl_AppendMessageToRadioPortTxFifo() .
  // To register the expected RESPONSE (from radio, forwarded to the client),
  //    RigCtrl_SendMessageFromRadioPortTxFifo() also prepares a 'filter'
  //    which the caller will pass on to RigCtrl_RegisterCIVReponseForClient(),
  //    just as if the message HAD BEEN immediately forwarded
  //    from client to radio in RigCtrl_ParseCIV() or similar.
{
  int nBytesOccupied, iMsgLength, iClientPort, i;
  int iHeadIndex = pRC->iRadioPortTxFifoHead;
  int iTailIndex = pRC->iRadioPortTxFifoTail;
  BYTE *pbBuffer = pRC->bRadioPortTxFifo;
  T_RigCtrl_PortInstance *pPort = &pRC->PortInstance[ RIGCTRL_PORT_RADIO ];
  BYTE *pbMessage = pPort->bTxMsg; // temporary 'linear' buffer for transmission
  T_RigCtrl_MsgFilter msg_filter;
  long i32CombinedCmdAndSubcode;

  nBytesOccupied = iHeadIndex - iTailIndex;
  if( nBytesOccupied < 0 ) // circular wrap, e.g. head=0, tail=2047 ->
   {  nBytesOccupied += RIGCTL_RADIO_PORT_TX_FIFO_SIZE;
   }
  if( nBytesOccupied >= 3 ) // need AT LEAST three bytes : 16 bit length + 8 bit client port number
   { // Read the MESSAGE LENGTH and CLIENT PORT in the FIFO (again, low byte first)
     iMsgLength  = RigCtrl_ReadByteFromRadioPortTxFifo( pbBuffer, &iTailIndex );
     iMsgLength |= ((int)RigCtrl_ReadByteFromRadioPortTxFifo( pbBuffer, &iHeadIndex) << 8);
     iClientPort = RigCtrl_ReadByteFromRadioPortTxFifo( pbBuffer, &iTailIndex );
     if( nBytesOccupied >= (3+iMsgLength) )  // still looks good .. queued-up message is COMPLETE
      {
        if( iMsgLength <= RIGCTRL_MAX_MESSAGE_LENGTH ) // must fit inside the "non-circular" TX buffer..
         { for(i=0; i<iMsgLength; ++i)
            { pbMessage[i] = RigCtrl_ReadByteFromRadioPortTxFifo( pbBuffer, &iTailIndex );
            }
         }
        pRC->iRadioPortTxFifoTail = // release transmitted message from the circular FIFO
          ( pRC->iRadioPortTxFifoTail + iMsgLength + 3 ) & (RIGCTL_RADIO_PORT_TX_FIFO_SIZE-1);

        // The following is similar to the code after
        //      'case RIGCTRL_CLIENT_ACTION_RETRANSMIT' in RigCtrl_ParseCIV(),
        //      incl. registering a "filter" for this message to forward its RESPONSE
        //      to the client (later, when a matching response arrived from the radio):
        msg_filter.iCmd = pbMessage[4];
        // any further bytes in any CI-V message are IGNORED for filtering:
        msg_filter.iSubCmd = msg_filter.iDataB1 = msg_filter.iDataB2 = -1;
        // Also let the filter (for the client) pass reponse "OK" and "NOT OK":
        msg_filter.iMsgTypes = RIGCTRL_MSGTYPE_OK | RIGCTRL_MSGTYPE_NOT_OK;
        i32CombinedCmdAndSubcode = RigCtrl_GetCombinedCmdAndSubcodeForCIV( pbMessage, iMsgLength );
        if( pRC->pRadioPortTxFifo != NULL )
         { RigCtrl_RegisterCIVReponseForClient( pRC, iClientPort, &msg_filter );
           // Mark the "radio port" as occupied, until the response
           //      for the forwarded command (sent below) arrives :
           pRC->i32ExpectedResponse_CmdAndSubcode = i32CombinedCmdAndSubcode;
           // ,---------------------'
           // '--> This is mainly important for the CI-V traffic monitor,
           //      to show e.g. > RX 006 FE FE 00 94 FD ; "OK" (for cmd 0x05)
           //      because Icom's response codes 0xFB ("Ok") and 0xFA ("Not Ok")
           //      don't carry any information about WHAT was "Ok" or "Not Ok".
           //      Thus we store this information in yet another state variable.
           //  But do we EXPECT a response from the radio at all ?
           //  How do we know it the CI-V message from the TX FIFO is a
           //      "set" or "read" command, and if it's a "set"-command,
           //      will the radio ALWAYS reply with "Ok" or "Not Ok" ?
           //  Note the confusing multitude of commands to SET THE FREQUENCY:
           //  Some of them (e.g. cmd 0x05) WILL be responded with 0xFB or 0xFA,
           //       others (maybe cmd 0x00) will NOT be confirmed / acknowledged
           //       by the radio.
           return RigCtrl_SendAndLogMessage( pRC, RIGCTRL_PORT_RADIO, pbMessage, iMsgLength,
                    TRUE/*wait for completion*/, RIGCTRL_MSGTYPE_OTHER, "fwd from FIFO" );
         } // end if < pRadioPortTxFifo != NULL > ?
      } // end if( nBytesOccupied >= (3+iMsgLength) )
   }
  // Arrived here ? The TX-FIFO is either completely empty, or the next message
  //        has not been completely assembled by the "writer" in another thread,
  //        or the address of the "transmit function" for this port is invalid.
  return TRUE;

} // end RigCtrl_SendMessageFromRadioPortTxFifo()




//--------------------------------------------------------------------------
BOOL RigCtrl_RegisterCIVReponseForClient( T_RigCtrlInstance *pRC,
        int iRigCtrlClientPort,       // [in] RIGCTRL_PORT_CLIENT_1 / 2 ?
        T_RigCtrl_MsgFilter *pFilter) // [in] command, subcommand, and possibly accepted message type
  // Called from RigCtrl_ParseCIV() when certain messages (commands)
  //        arrive on one of the *CLIENT PORTS* (iRigCtrlClientPort).
  // Because the client "thinks" he's directly talking to the radio,
  //         anything that the radio sends IN RESPONSE must also be
  //         forwarded (not echoed) to the client's port (iRigCtrlClientPort).
  // For CI-V, the response may be an unspecific "Ok" or "Error".
  //         To allow forwarding such responses to the client,
  //         the caller may also specify a CATEGORY of message types, for example
  //         RIGCTRL_MSGTYPE_OK | RIGCTRL_MSGTYPE_NOT_OK (bitwise combineable).
{
  T_RigCtrl_PortInstance *pPortInstance;
  T_RigCtrl_MsgFilter    *pMsgFilter;

  int i;
  // Example: Call after receiving "Power On"-command, 1st frame from "RS-BA1":
  //    pbMessage[0  1  2  3   4   5   6  7 ...
  //         Len Pre.  Addr   cmd sub  d1 d2
  // >    r1 007 FE FE 94 E0   18  01        client to server) -> pFilter->iCmd = 0x18
  // >    TX 007 FE FE 94 E0   18  01        server lab to IC-7300
  // >    RX 007 FE FE 94 E0   18  01        echo directly from IC-7300
  // >    RX 006 FE FE E0 94   FB            IC-7300 to server: "OK", must pass the new filter because it's the "expected response"
  // >    t1 006 FE FE E0 94   FB            response relayed from server to client

  if( (iRigCtrlClientPort<0) || (iRigCtrlClientPort>/*!*/RIGCTRL_MAX_CLIENT_PORTS) )
   { return FALSE;  // not a valid 'port identifier'
   }

  pPortInstance = &pRC->PortInstance[iRigCtrlClientPort];

  // First check if this message (command, subcommand, sub-sub-command) is already registered:
  for( i=0; i<RIGCTRL_MAX_REGISTERED_MESSAGES_PER_CLIENT; ++i )
   { pMsgFilter = &pPortInstance->RegisteredMessages[i];
     if( ( pMsgFilter->iCmd    == pFilter->iCmd )
      && ( pMsgFilter->iSubCmd == pFilter->iSubCmd )
      && ( pMsgFilter->iDataB1 == pFilter->iDataB1 )
      && ( pMsgFilter->iDataB2 == pFilter->iDataB2 ) )
      {
        pMsgFilter->iMsgTypes |= pFilter->iMsgTypes; // e.g. "OK" + "NOT OK"
        return TRUE;
      }
   }

  // Not registered yet, so use the FIRST unused entry :
  for( i=0; i<RIGCTRL_MAX_REGISTERED_MESSAGES_PER_CLIENT; ++i )
   { pMsgFilter = &pPortInstance->RegisteredMessages[i];
     if( (pMsgFilter->iCmd <= 0 ) && (pMsgFilter->iMsgTypes <= 0 ) )
      { *pMsgFilter = *pFilter;
        return TRUE;
      }
   }

  // Running out of space to "register" another message for this client..
  return FALSE;


} // end RigCtrl_RegisterCIVReponseForClient()


//--------------------------------------------------------------------------
void RigCtrl_CheckAndForwardResponseToClients( T_RigCtrlInstance *pRC,
        BYTE *pbMessage, int iMsgLength, // [in] message including pre- and postamble
        T_RigCtrl_MsgFilter *pFilter,    // [in] command, subcommand, and possibly accepted message type
        int iReceivedMessageType,        // [in] class of the received message for the log
        BOOL fIsUnsolicitedMsg,          // [in] TRUE for unsolicited message, e.g. VFO frequency + Waveform data
        char * pszComment )              // [in] optional comment for the log
{
  int iRigCtrlClientPort, iFilter;
  T_RigCtrl_PortInstance *pPortInstance;
  T_RigCtrl_MsgFilter *pMsgFilter;
  BOOL fMatch;

  for( iRigCtrlClientPort = RIGCTRL_PORT_CLIENT_1;
       iRigCtrlClientPort < (RIGCTRL_PORT_CLIENT_1+RIGCTRL_MAX_CLIENT_PORTS);
       iRigCtrlClientPort++ )
   { pPortInstance = &pRC->PortInstance[iRigCtrlClientPort];

     // Does THIS client tolerate 'unsolicited' messages (e.g. new VFO-frequecy, Scope data) ?
     // A well-written CI-V parser must ignore messages he doesn't understand,
     //    because CI-V has once been designed as a multi-master bus system.
     //    Not so with the mimosa known as "Hamlib" (used per default in WSJT-X).
     if( fIsUnsolicitedMsg && (pPortInstance->iClientOptions & RIGCTRL_CLIENT_OPTION_REJECT_UNSOLICITED_MSGS) )
      { // Don't forward this unsolitited message to the client.
        // This stupid Extrawurst was added to prevent WSJT-X to bail out
        // with the oh-so-familar "Ding-Dong ! Rig Control Error" :
        continue;  // continue with the next client . Do NOT forward to the mimosa .
      }

     // Does the received message match any filter registered by THIS client ?
     fMatch = FALSE;
     for( iFilter=0; iFilter<RIGCTRL_MAX_REGISTERED_MESSAGES_PER_CLIENT; ++iFilter )
      { pMsgFilter = &pPortInstance->RegisteredMessages[iFilter];
        if( ( ( pMsgFilter->iCmd    == pFilter->iCmd    ) || ( pMsgFilter->iCmd < 0) )
         && ( ( pMsgFilter->iSubCmd == pFilter->iSubCmd ) || ( pMsgFilter->iSubCmd<0) )
         && ( ( pMsgFilter->iDataB1 == pFilter->iDataB1 ) || ( pMsgFilter->iDataB1<0) )
         && ( ( pMsgFilter->iDataB2 == pFilter->iDataB2 ) || ( pMsgFilter->iDataB2<0) ) )
         { fMatch = TRUE;
           break;
         }
        if( pMsgFilter->iMsgTypes == pFilter->iMsgTypes )
         { fMatch = TRUE;
           break;
         }
        // If the message received on the "RADIO"-port was "OK" or "NOT OK",
        //    pMsgFilter->iCmd will not match (e.g. client was asking for 0x18 = "POWER ON"
        //     but radio answered with Cmd=0xFB = "ok"). To pass such "expected" messages,
        //     when expecting "OK" or "NOT OK", ANY of these will match:
        if(  ((pMsgFilter->iMsgTypes & (RIGCTRL_MSGTYPE_OK|RIGCTRL_MSGTYPE_NOT_OK)) != 0)
          && ((pFilter->iMsgTypes    & (RIGCTRL_MSGTYPE_OK|RIGCTRL_MSGTYPE_NOT_OK)) != 0) )
         { fMatch = TRUE;  // "OK" matches "NOT OK", and "NOT OK" also matches "OK" !
           break;
         }
      }

     if( fMatch )
      { RigCtrl_SendAndLogMessage( pRC, iRigCtrlClientPort, pbMessage, iMsgLength,
           FALSE/*don't wait for completion*/, iReceivedMessageType, pszComment );
      }
   } // end for < iRigCtrlClientPort >
} // end RigCtrl_CheckAndForwardResponseToClients()


//--------------------------------------------------------------------------
BOOL RigCtrl_AskForRadioID( T_RigCtrlInstance *pRC )
{
  BOOL fResult = FALSE;

  switch( pRC->iRadioCtrlProtocol )
   {  // which language does the radio speak ?
    case RIGCTRL_PROTOCOL_NONE :
    default:
         break;  // unknown protocol, can only send "direct" strings

    case RIGCTRL_PROTOCOL_ICOM_CI_V : // Icom's "CI-V" protocol ...
         fResult = RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                       RIGCTRL_MSGTYPE_RADIO_ID | RIGCTRL_MSGTYPE_FLAG_SUGGEST_NEW_HEADLINE_FOR_TRAFFIC_LOG, // iMsgType for the "traffic log" ...
                       "get radio ID",     // comment for the "traffic log"
                       0x19,  // iCmd: CI-V command "Read transceiver ID" (and a few others)
                       0x00,  // iSubCmd: "Read transceiver ID code"
                       -1, -1 );  // additional data bytes : None !
         // > Returns always the default CI-V adress, regardless of the actual
         // > setting of the CI-V adress. This function can be used by software
         // > to determine which radio is connected to the computer.
         // > Despite the functions name this command works also on receivers,
         // > not only transceivers.
         // If all works as planned, the transceiver also responds with cmd 0x1900 :
         pRC->i32ExpectedResponse_CmdAndSubcode = RCTL_COMBINE_CMD_AND_SUB(0x19, 0x00);  // only one known subcode, but.. !
         // An IC-7300 required almost a SECOND to send an answer ?!
         pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms; // if the radio doesn't respond after this time, continue..
         break;  // end case <ICOM CI-V protocol>

   } // end switch( pRC->iRadioCtrlProtocol )


  return fResult;

} // end RigCtrl_AskForRadioID()

//--------------------------------------------------------------------------
BOOL RigCtrl_AskForVFOFrequency( T_RigCtrlInstance *pRC )
{
  BOOL fResult = FALSE;

  switch( pRC->iRadioCtrlProtocol )
   {  // which language does the radio speak ?
    case RIGCTRL_PROTOCOL_NONE :
    default:
         break;  // unknown protocol, can only send "direct" strings

    case RIGCTRL_PROTOCOL_ICOM_CI_V : // Icom's "CI-V" protocol ...
         fResult = RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                       RIGCTRL_MSGTYPE_SET_FREQUENCY, // iMsgType (for traffic log)
                       "read VFO freq", // short comment for the "traffic log"
                       0x03,  // iCmd: CI-V command "Read operating frequency"
                       -1, -1, -1 );  // iSubCmd and iDataBytes : Omit !
         // If all works as planned, the transceiver responds with 0x03 :
         pRC->i32ExpectedResponse_CmdAndSubcode = RCTL_COMBINE_CMD_ONLY(0x03); // CI-V response "Read operating frequency"
         // [IC7610CIV] doesn't explain the exact format of this (and many other)
         // 'simple' queries. But http://www.plicht.de/ekki/civ/civ-p41.html says:
         // > This command is understood by practically all Icom radios.
         // That's why we use THIS simple query, not one of the fancy new
         // multi-byte commands like 0x1A (with its awfully long list of sub-commands).
         pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms; // if radio doesn't respond after this time, continue..
         break;  // end case <ICOM CI-V protocol>

   } // end switch( pRC->iRadioCtrlProtocol )

  return fResult;

} // end RigCtrl_AskForVFOFrequency()

//--------------------------------------------------------------------------
BOOL RigCtrl_AskForOperatingMode( T_RigCtrlInstance *pRC )
{
  BOOL fResult = FALSE;

  switch( pRC->iRadioCtrlProtocol )
   {  // which language does the radio speak ?
    case RIGCTRL_PROTOCOL_NONE :
    default:
         break;  // unknown protocol, can only send "direct" strings

    case RIGCTRL_PROTOCOL_ICOM_CI_V : // Icom's "CI-V" protocol ...
         fResult = RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                     RIGCTRL_MSGTYPE_OP_MODE, // iMsgType (for traffic log)
                       "read op. mode", // short comment for traffic log
                       0x04,  // iCmd: CI-V command "Read operating mode"
                       -1, -1, -1 );  // iSubCmd and iDataBytes : Omit !
         // If all works as planned, the transceiver responds with 0x04, no sub-command-byte(s) :
         pRC->i32ExpectedResponse_CmdAndSubcode = RCTL_COMBINE_CMD_ONLY(0x04); // CI-V response with "operating mode" (command only, no subcode)
         pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms;  // now waiting for the radio to REPORT ITS OPERATING MODE
         break;  // end case <ICOM CI-V protocol>

   } // end switch( pRC->iRadioCtrlProtocol )

  return fResult;

} // end RigCtrl_AskForOperatingMode()


//--------------------------------------------------------------------------
void RigCtrl_AskForParameter( T_RigCtrlInstance *pRC, T_RigCtrl_ParamInfo *pPI )
  // ... without waiting for a response, as the name should imply ..
  //     Like any other response from the radio, the response will be analysed
  //     later, in RigCtrl_ProcessRxData() -> RigCtrl_ParseCIV()
  // [in] iUnifiedPN : e.g.
{
  BOOL fResult = FALSE;
  BYTE b16TxBuffer[16];
  BYTE *pbBufPtr = b16TxBuffer;
  int  i, iMsgLength;
  char sz80Comment[84];
  char *pszDest=sz80Comment, *pszEndstop=sz80Comment+80;

  if( pPI != NULL )
   { if( pPI->pszToken != NULL )
      { SL_AppendPrintf( &pszDest, pszEndstop, "read %s", pPI->pszToken );
      }
     else
      { SL_AppendPrintf( &pszDest, pszEndstop, "read #%d", (int)pPI->iUnifiedPN );
      }
     switch( pRC->iRadioCtrlProtocol )
      { // which language does the radio speak ?
        case RIGCTRL_PROTOCOL_NONE :
        default:
           break;  // unknown protocol, can only send "direct" strings

        case RIGCTRL_PROTOCOL_ICOM_CI_V : // Icom's "CI-V" protocol ...
           if( (pPI->pbCIVReadCmd != NULL )  && (pPI->iCIVReadCmdLength>0) && (pPI->iCIVReadCmdLength<8) )
            { // Assemble a CI-V READ COMMAND based on the T_RigCtrl_ParamInfo :
              *pbBufPtr++ = 0xFE;  // fixed preamble code for CI-V protocol (term used by Icom .. NOT "Begin of Message" ! )
              *pbBufPtr++ = 0xFE;  // fixed preamble code for CI-V protocol
              *pbBufPtr++ = (BYTE)pRC->iRadioDeviceAddr;  // "to" transceiver's address
              *pbBufPtr++ = (BYTE)pRC->iRadioMasterAddr;  // "from" controller's address
              for( i=0; i<pPI->iCIVReadCmdLength;  )
               { *pbBufPtr++ = pPI->pbCIVReadCmd[i++];  // examples (bytes appended here):
                  // 0x03      = CI-V command "Read operating frequency",
                  // 0x04      = CI-V command "Read operating mode" (modulation),
                  // 0x0F      = CI-V command "Read Split Setting" / "DUP-" / "DUP+" / "RepeaterSimplex" (for IC-9700)
                  // 0x14 0x01 = Read AF level,
                  // 0x14 0x02 = Read RF gain, etc etc
               }
              switch( pPI->iUnifiedPN ) // Check for "very special cases" that cannot be described by T_RigCtrl_ParamInfo
               {
                 default: // nothing special (a "Cmd" and maybe a "Sub cmd" are enough to READ the parameter)
                    break;
               } // end switch( iUnifiedPN )
              *pbBufPtr++ = 0xFD;  // CI-V code for "End Of Message" (term used by Icom .. NOT "postamble", even though 0xFE is called "preamble" ! )
              iMsgLength = pbBufPtr - b16TxBuffer;
              fResult = RigCtrl_SendAndLogMessage( pRC, RIGCTRL_PORT_RADIO,
                            b16TxBuffer, iMsgLength,
                            FALSE/*fWaitForCompletion*/,
                            RIGCTRL_MSGTYPE_FLAG_CIV | RIGCTRL_MSGTYPE_OTHER,
                            sz80Comment );
              // If all works as planned, the transceiver should respond shortly:
              if( fResult ) // successfully sent the READ COMMAND -> now waiting for a response:
               { pRC->iExpectReadResponseForUnifiedPN = pPI->iUnifiedPN; // <- important for RigCtrl_ParseCIV()
                 pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms; // now waiting for the radio to REPORT THE PARAMETER VALUE
               }
            } // end if( pRC->pbCIVReadCmd != NULL )
           break;  // end case <ICOM CI-V protocol>

      } // end switch( pRC->iRadioCtrlProtocol )
   } // end if( pPI != NULL )

} // end RigCtrl_AskForParameter()


//--------------------------------------------------------------------------
BOOL RigCtrl_StartStopSpectrumData( T_RigCtrlInstance *pRC, BOOL fStart )
  // Tries to kick the periodic, unsolicited transmission of an IC-7300's
  // (or IC-7610's, or maybe IC-R8600's) broadband spectrum alive.
  // Icom decided to call this 'Scope waveform data' (command 0x27)
  // but this name is too ambiguous (since a "waveform" sounds like TIME DOMAIN),
  // in fact what we're trying to read here are the data which the LOCAL DISPLAY
  // (TFT screen inside the rig) uses for the spectrum graph + waterfall display.
  //      Actually, this feature was the motivation for the author
  //      to roll his own 'Rig Control' unit for Spectrum Lab,
  //      instead of using some 3rd party library like 'HamLib' or 'OmniRig' .
  // Called from the incredible state machine in RigCtrl_Handler(),
  //      somewhere in state RIGCTRL_POLLSTATE_CONFIGURE_SPECTRUM .
{
  BOOL fResult = FALSE;

  switch( pRC->iRadioCtrlProtocol )
   {  // which language does the radio speak ?
    case RIGCTRL_PROTOCOL_NONE :
    default:
         break;  // unknown protocol, can only send "direct" strings

    case RIGCTRL_PROTOCOL_ICOM_CI_V : // Icom's "CI-V" protocol ...
         fResult = RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                     RIGCTRL_MSGTYPE_C_SPECTRUM, // iMsgType (for traffic log)
                     fStart ? "start spectrum" : "stop spectrum", // comment for traffic log
                       0x27,  // iCmd   : "Scope waveform data"
                       0x11,  // iSubCmd: "Send/read the Scope wave data output setting" [IC7300FM page 19-7]
                      fStart, // 1st data byte : 1 = ON,  0 = off
                       -1 );  // 2nd data byte : omit
         // If all works as planned, the transceiver agrees as shown below:
         // > 12:01:35.5 TX 008 FE FE 94 E0 27 11 01 FD           (scope)  ; start spectrum
         // > 12:01:35.5 RX 006 FE FE E0 94 FB FD                 (OK)
         // (other transceivers may simply start transmitting SPECTRUM DATA,
         //  don't rely on the "OK"-response for the above command)
         pRC->i32ExpectedResponse_CmdAndSubcode = RCTL_COMBINE_CMD_ONLY(0xFB); // at least an IC-7300 responded with "OK"
         pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms; // now waiting for the radio to respond with "OK" or "NOT OK"...
         // A few milliseconds later, an IC-7300 began sending "Scope waveform data" periodically,
         //  message format shown in RigCtrl_ParseCIV(), case 0x27 .
         break;  // end case <ICOM CI-V protocol>

   } // end switch( pRC->iRadioCtrlProtocol )

  return fResult;
} // end RigCtrl_StartStopSpectrumData()


//--------------------------------------------------------------------------
BOOL RigCtrl_SetScopeSpeed( T_RigCtrlInstance *pRC, int iScopeSpeed )
{
  BOOL fResult = FALSE;
  BYTE bData;
  char *pszComment;

  switch( pRC->iRadioCtrlProtocol )
   {  // which language does the radio speak ?
    case RIGCTRL_PROTOCOL_NONE :
    default:
         break;  // unknown protocol, can only send "direct" strings

    case RIGCTRL_PROTOCOL_ICOM_CI_V : // Icom's "CI-V" protocol ...
         switch( iScopeSpeed )
          { case RIGCTRL_SCOPE_SPEED_FAST:
            default:
               bData = 0x00;
               pszComment = "scope:FAST";
               break;
            case RIGCTRL_SCOPE_SPEED_MID:
               bData = 0x01;
               pszComment = "scope:MID";
               break;
            case RIGCTRL_SCOPE_SPEED_SLOW:
               bData = 0x02;
               pszComment = "scope:SLOW";
               break;
          }
         fResult = RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                     RIGCTRL_MSGTYPE_C_SPECTRUM, // iMsgType (for traffic log)
                       pszComment, // comment for traffic log
                       0x27,  // iCmd   : "Scope waveform data"
                       0x1A,  // iSubCmd: "Send/read the Scope Sweep speed settings"
                       0x00,  // 1st data byte : 0x00 = MAIN scope, 0x01 = SUB scope
                     bData);  // 2nd data byte : "speed setting"
         pRC->i32ExpectedResponse_CmdAndSubcode = RCTL_COMBINE_CMD_ONLY(0xFB); // at least an IC-7300 responded with "OK"
         pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms; // now waiting for the radio to respond with "OK" or "NOT OK" after "Set Scope Speed"
         break;  // end case <ICOM CI-V protocol>

   } // end switch( pRC->iRadioCtrlProtocol )

  return fResult;
} // end RigCtrl_SetScopeSpeed()

//--------------------------------------------------------------------------
void RigCtrl_EnterState_Done( T_RigCtrlInstance *pRC )
  // Called from e.g. RigCtrl_Handler() .
  // Here, prints an info like the following into the 'Debug' tab:
  //  "RigControl: IC-9700 on COM5, 432250.0 kHz, CWN"
{
  char sz255Msg[255];

  pRC->iParameterPollingState = RIGCTRL_POLLSTATE_DONE;  // <- will get here ONCE after init, causing the GUI to switch from the "Debug" tab to the "TRX" tab.

  snprintf( sz255Msg, 255, "RigControl: %s on %s, %.1lf kHz, %s",
      RigCtrl_DefaultAddressToString(pRC->iRadioCtrlProtocol,pRC->iDefaultAddress),
      RigCtrl_GetRadioControlPortAsString(),  // -> e.g. "COM5", if that's the port "talking to the radio"
      pRC->dblVfoFrequency*1e-3,
      RigCtrl_OperatingModeToString( pRC->iOpMode ) );
  // Emit a single line with 'info' in the error history, with a summary about the remote rig:
  ShowError( ERROR_CLASS_INFO | SHOW_ERROR_TIMESTAMP, sz255Msg );

  // Show the same also in the 'rig control traffic log' ?
  if( RigCtrl_iTrafficMonitorDisplayOptions & RIGCTRL_TRAFFIC_DISPLAY_OPTION_ENABLE )
   { RigCtrl_AddToTrafficLog( pRC, RIGCTRL_PORT_RADIO, NULL/*pbMessage*/, 0/*iMsgLength*/,
                              RIGCTRL_MSGTYPE_OK, sz255Msg );
   }
} // end RigCtrl_EnterState_Done()


//--------------------------------------------------------------------------
void RigCtrl_FinishReadingBandStackingRegs( T_RigCtrlInstance *pRC )
  // Called from RigCtrl_Handler() or RigCtrl_ParseCIV() when all
  // 'band stacking registers' have been read from the radio,
  // or are not supported by the radio at all.
  // Details about the UGLY, RIG-SPECIFIC FORMAT in RigCtrl_ParseMemoryContent().
{
  // Notify the application about possibly 'new data' in pRC->BandStackingRegs[] :
  pRC->fBandStackingRegsModified = TRUE; // <- application may POLL and even CLEAR this flag
  pRC->fBandStackingRegsModifiedForCwNet = TRUE; // similar flag for CwNet.c,
          // to start sending band stacking registers from server to clients.

  pRC->iParameterPollingState = RIGCTRL_POLLSTATE_PARAMS_FOR_HAMLIB;
  pRC->iParameterPollingSubState = 0;
} // end RigCtrl_FinishReadingBandStackingRegs()


//--------------------------------------------------------------------------
void RigCtrl_Handler( T_RigCtrlInstance *pRC )   // the incredible state machine ..
  // Periodically called every few ten milliseconds (roughly C_RIGCTRL_HANDLER_CALLING_INTERVAL_MS).
  //
  // Callers (in Spectrum Lab)
  //  * If a SERIAL PORT is used : IOaccess.cpp  : IO_RadioCommThdFunc() .
  //  * If an INTERNET SOCKET (e.g. UDP) is used : RigCtrl_UDPThreadFunc().
  //
  // Received data may be passed in ANYTIME between two calls of this handler,
  //               via RigCtrl_ProcessRxData() [for 'COM' ports, UDP, etc].
  // Potential multithreading issues must be avoided by the CALLER(s).
  // Implemented like a state machine for portability
  //    (with a single-thread microcontroller firmware in mind) .
  //
  // BUGLIST (endless, only a few 'recent cases' shown here):
  //
  //  2023-07-01 : When connecting an IC-705 via USB (and without knowing
  //               which of the TWO freshly appeared 'COM ports' to use),
  //               RigCtrl_Handler() wasn't called from anywhere,
  //               and all the 'CAT Traffic Log' had to say was :
  //               "RigControl: Re-connect" . Not very user-friendly !
  //
{
  char szComment[ RIGCTRL_MAX_COMMENT_LENGTH+4 ], *pszComment = NULL;
  // ex: T_RigCtrl_IcomUdpPortInstance *pUdpPortInst = NULL;
  int i;
  T_RigCtrl_ParamInfo *pPI;
  BOOL fDone;

#if(0)  // removed: variables to read the 'Fixed Edge' frequencies for Icom's waterfall/spectrum display
  int  iRange, iEdge;
  BYTE b1, b2;
#endif

  if( pRC == NULL ) // oops.. instance deleted but some thread is still calling us !
   { return;
   }
  if( pRC->initState != initState_Opened ) // oops.. didn't call RigCtrl_Init() yet, or called RigCtrl_Close()
   { return;
   }

  // Process the RECEIVED DATA (including "unsolicited reports")
  //    even in "passive" mode (receive-only) mode:
  RigCtrl_ProcessRxData( pRC, RIGCTRL_PORT_RADIO );

  if( pRC->fReceiveOnlyMode ) // "passive" mode ?
   { return;
   }
  if( pRC->iRadioCtrlProtocol == RIGCTRL_PROTOCOL_NONE ) // Neither CI-V nor Hamlib/Rigctld ?
   { return;  // don't try to connect at all, and don't bug me with messages like
              // "Remote spectrum timeout", etc etc pp ...
   }



  // Arrived here ? The 'connect-to-radio' phase seems to have been completed,
  //   or we are 'connected' via a simple 'COM port' (caller=IO_RadioCommThdFunc),
  //         so the normal 'CI-V parameter-polling' state machine jumps in
  //         to exchange whatever we need to with the remote radio.
  //   Below is the ALMOST UNMODIFIED code from the days when there were only
  //   'serial interfaces' (CI-V on a single wire or USB with virtual COM port).

  if( pRC->iResponseCountdown_ms > 0 ) // "do not send anything at the moment" because we're waiting for a response ->
   { pRC->iResponseCountdown_ms -= C_RIGCTRL_HANDLER_CALLING_INTERVAL_MS; // <- e.g. 50 milliseconds
     if( pRC->iResponseCountdown_ms <= 0 )  // just timed out ...
      { pRC->iResponseCountdown_ms = 0;
        if( (pRC->i32ExpectedResponse_CmdAndSubcode >= 0 )
         && (pRC->iParameterPollingState != RIGCTRL_POLLSTATE_PASSIVE )
         && (pRC->iParameterPollingState != RIGCTRL_POLLSTATE_DONE ) )
         { sprintf( pRC->sz255LastError, "timed out waiting for Cmd 0x%02X, state=%d",
             (int)pRC->i32ExpectedResponse_CmdAndSubcode, (int)pRC->iParameterPollingState );
           if( RigCtrl_iTrafficMonitorDisplayOptions & RIGCTRL_TRAFFIC_DISPLAY_OPTION_PAUSE_ON_ERROR )
            { if( ! RigCtrl_fTrafficMonitorPaused )
               { RigCtrl_AddToTrafficLog( pRC, RIGCTRL_PORT_RADIO, NULL/*pbMessage*/, 0/*iMsgLength*/,
                   RIGCTRL_MSGTYPE_ERROR, "Traffic Log paused after RESPONSE TIMEOUT (as configured)" );
               }
              RigCtrl_fTrafficMonitorPaused = TRUE;
            } // end if < RIGCTRL_TRAFFIC_DISPLAY_OPTION_PAUSE_ON_ERROR .. >
           if( pRC->iParameterPollingState==RIGCTRL_POLLSTATE_READ_RADIO_ID )
            { // This was the FIRST query sent to the radio, and it didn't respond.
              // It may have been turned off right from the start of the program,
              // so instead of wasting more time (polling other parameters),
              // try to turn the radio ON via command:
              if( pRC->iCapabilities & RIGCTRL_CAPS_POWER_ON_OFF_VIA_CAT )
               { // the rig can be turned on and off cia CAT, so try to turn it on:
                 pRC->iParameterPollingState = RIGCTRL_POLLSTATE_TURN_RIG_ON;
               }
            }
           else // not in RIGCTRL_POLLSTATE_READ_RADIO_ID ->
            {
              ++pRC->numResponseTimeouts;
              if( pRC->numResponseTimeouts > 3 ) // too many response timeouts in a row ?
               { // The rig doesn't seem to respond AT ALL. It may be turned off..
                 if( pRC->iCapabilities & RIGCTRL_CAPS_POWER_ON_OFF_VIA_CAT )
                  { // so if we know the rig can be turned on and off cia CAT, try to turn it on again, and start all over
                    pRC->iParameterPollingState = RIGCTRL_POLLSTATE_TURN_RIG_ON; // try to turn the rig on, then start all over
                  }
               } // end if( pRC->numResponseTimeouts > 3 ) ?
            }   // end if < ! RIGCTRL_POLLSTATE_READ_RADIO_ID >
         }     // end if < EXPECTED a response, AND one of the 'active' parameter-polling-states >
        else // pRC->i32ExpectedResponse_CmdAndSubcode < 0 .. but we may expect some other response:
        if( pRC->iExpectReadResponseForUnifiedPN > 0 ) // <- added 2024-07, more 'universal' replacement for i32ExpectedResponse_CmdAndSubcode..
         { // just timed out waiting waiting for a response after RigCtrl_AskForParameter() :
           pPI = RigCtrl_GetInfoForUnifiedParameterNumber(pRC->iExpectReadResponseForUnifiedPN);
           if( pPI != NULL )
            { sprintf( pRC->sz255LastError, "timed out waiting for read response for %s, state=%d",
              (char*)pPI->pszToken, (int)pRC->iParameterPollingState );
            }
           if( RigCtrl_iTrafficMonitorDisplayOptions & RIGCTRL_TRAFFIC_DISPLAY_OPTION_PAUSE_ON_ERROR )
            { if( ! RigCtrl_fTrafficMonitorPaused )
               { RigCtrl_AddToTrafficLog( pRC, RIGCTRL_PORT_RADIO, NULL/*pbMessage*/, 0/*iMsgLength*/,
                   RIGCTRL_MSGTYPE_ERROR, "Traffic Log paused after RESPONSE TIMEOUT (as configured)" );
               }
              RigCtrl_fTrafficMonitorPaused = TRUE;
            } // end if < RIGCTRL_TRAFFIC_DISPLAY_OPTION_PAUSE_ON_ERROR .. >
         }   // end if < iExpectReadResponseForUnifiedPN .. >
      }     // end if < iResponseCountdown_ms just timed out >
   }
  else  // enough time has passed to send the next command..
   { switch( pRC->iParameterPollingState )
      { case RIGCTRL_POLLSTATE_START_RD:  // RigCtrl_StartReading() has been called..
           pRC->numResponseTimeouts = 0;
           // If the protocol is CI-V, start by reading the TRANSCEIVER ID.
           // We'll know if a radio supports a certain 'advanced' feature then,
           // much better than relying on the radio's "default ID" !
           RigCtrl_AskForRadioID( pRC );
           pRC->iParameterPollingState = RIGCTRL_POLLSTATE_READ_RADIO_ID;
           pRC->iParameterPollingSubState = 0; // here: enumerator for RigCtrl_EnumerateRequiredPNs()
           pRC->iNumBandStackingRegs = 0; // forget old 'Band Stacking Register' entries
           break;
        case RIGCTRL_POLLSTATE_READ_RADIO_ID: // waiting for a response with the radio's "default" ID, as a connection test and to check its type
           pRC->iParameterPollingState = RIGCTRL_POLLSTATE_PARAMS_FOR_RIGCTL;
           pRC->iParameterPollingSubState = 0; // here: enumerator for RigCtrl_EnumerateRequiredPNs()
           pRC->iNumBandStackingRegs = 0; // forget old 'Band Stacking Register' entries
           break;
        case RIGCTRL_POLLSTATE_PARAMS_FOR_RIGCTL: // (try to) read the current VFO frequency, mode, etc
           fDone = TRUE;
           while( (i=RigCtrl_EnumerateRequiredPNs( pRC->iParameterPollingSubState++)) > 0 )
            { pPI = RigCtrl_GetInfoForUnifiedParameterNumber(i/*iUnifiedPN*/);
              if( pPI != NULL )
               { if( !RigCtrl_HaveValidParamValue( pRC, pPI ) )
                  { RigCtrl_AskForParameter( pRC, pPI ); // <- sets pRC->iResponseCountdown_ms and pRC->iExpectReadResponseForUnifiedPN
                    fDone = FALSE;
                    break;
                  }
               }
            }
           if( fDone ) // finished polling all important parameters for RigControl.c itself :
            { pRC->iParameterPollingSubState = pRC->iParameterPollingArrayIndex = 0;
              if( pRC->iNumTxBands > 0 )  // successfully detected the NUMBER OF TRANSMIT BANDS..
               { pRC->iParameterPollingState = RIGCTRL_POLLSTATE_TX_BAND_EDGES;
               }
              else // could not determine pRC->iNumTxBands -> skip RIGCTRL_POLLSTATE_TX_BAND_EDGES
               { pRC->iParameterPollingState = RIGCTRL_POLLSTATE_BAND_STACKING_REGS;
               }
            }
           break;
        case RIGCTRL_POLLSTATE_TX_BAND_EDGES :  // Determine the edge frequency of all TRANSMIT bands:
           // [in] pRC->iNumTxBands : has already been read because RIGCTRL_PN_NUM_TX_BANDS is included in RigCtrl_EnumerateRequiredPNs().
           // [out] pRC->TxBandEdges[ 0 .. pRC->iNumTxBands-1 ] will be read from THE RIG .
           fDone = TRUE;
           if( (pRC->iParameterPollingSubState < pRC->iNumTxBands )
            && (pRC->iParameterPollingSubState < RIGCTRL_NUM_TX_BAND_EDGES ) )
            { // how to read an array of "transmit band edge frequencies" ?
              switch( pRC->iRadioCtrlProtocol )
               { case RIGCTRL_PROTOCOL_ICOM_CI_V : // using Icom's CI-V :
                    // The following is based on Icom's "IC-7300 Full Manual",
                    //  revision 6, document ID "A7292-4EX-6", chapter 19 :
                    // Cmd 0x1E, SubCmd 0x00 = "Read number of available TX frequency band" (what they mean is "bands")
                    //           has already been sent and answered.
                    // Cmd 0x1E, SubCmd 0x01 = "Read TX band edge frequencies" .
                    //   For the READ REQUEST, we supposed to send
                    //   cmd 0x1E followed by SubSmd 0x01 followed by the "Edge Number"
                    //   to retrieve the edge frequencies for the requested "Edge".
                    //   The READ RESPONSE contains "Edge Number", start frequency
                    //   (delimited by a funny separator, code 0x2D), and the
                    //   end frequency (terminated by the CI-V postamble, 0xFD).
                    //   Tried that on an IC-9700, with the following result:
                    // >  TX 007 FE FE 00 00 1E 00 FD      ; read NumTxBands
                    // >  RX 008 FE FE 00 A2 1E 00 03 FD   ; NumTxBands:3
                    // >  TX 008 FE FE 00 00 1E 01 01 FD   ; tx band edge[0] ?
                    // >  RX 013 FE FE 00 A2 1E 01 01 00 00 00 44 01 2D 00 00 00 46 01 FD ; 144 MHz .. 146 MHz
                    // >  TX 008 FE FE 00 00 1E 01 02 FD   ; tx band edge[1] ?
                    // >  RX 013 FE FE 00 A2 1E 01 02 00 00 00 30 04 2D 00 00 00 40 04 FD ; 430 MHz .. 440 MHz
                    // >  TX 008 FE FE 00 00 1E 01 03 FD   ; tx band edge[2] ?
                    // >  RX 013 FE FE 00 A2 1E 01 03 00 00 00 40 12 2D 00 00 00 00 13 FD ; 1240 MHz .. 1300 MHz
                    //                              '-- "Edge number"
                    //   Tried the same on an IC-7300, modified to transmit on 5.3 and 70 MHz:
                    // >  TX 007 FE FE 00 00 1E 00 FD      ; read NumTxBands
                    // >  RX 008 FE FE 00 94 1E 00 01 FD   ; :1 (surprise surprise, the IC-7300 is a "monoband rig" now...)
                    // >  TX 008 FE FE 00 00 1E 01 01 FD   ; tx band edge[0] ?
                    // >  RX 013 FE FE 00 94 1E 01 01 00 00 10 00 00 2D 00 00 80 74 00 FD ; 0.1 .. 74.8 MHz
                    //      -> Ok.. but not really helpful anymore,
                    //              only a SINGLE BAND from 100 kHz to 74.8 MHz,
                    //              sounds like a risk for the PA, ATU, etc.
                    //         Also, with a single band "covering anything"
                    //         from the radio's point of view, the APPLICATION
                    //         (e.g. Remote CW Keyer) needs a USER-DEFINED list
                    //         of frequency bands.
                    sprintf(szComment, "tx band edge[%d] ?", (int)pRC->iParameterPollingSubState );
                    if( RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                       RIGCTRL_MSGTYPE_RADIO_ID, // MsgType for the "traffic log" ...
                       szComment,  // comment for the "traffic log"
                       0x1E,  // iCmd: CI-V command "TX bands"
                       0x01,  // iSubCmd: "TX band edge frequencies"
                       RigCtrl_IntToBCD2(++pRC->iParameterPollingSubState), // EDGE NUMBER (Icom starts counting at ONE, not ZERO)
                       -1 ) )  // additional data bytes : None !
                     { // If all works as planned, the transceiver also responds with cmd 0x1E 0x01,
                       // the rest happens in RigCtrl_ParseCIV(), case 0x1E .
                       pRC->iExpectReadResponseForUnifiedPN = RIGCTRL_PN_TX_BAND_EDGES; // <- for RigCtrl_ParseCIV()
                       pRC->i32ExpectedResponse_CmdAndSubcode = RCTL_COMBINE_CMD_AND_SUB(0x1E,0x01);  // <- expected command and subcode
                       pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms; // now waiting for the radio to respond with "OK" or "NOT OK" after "Report another BAND EDGE"
                       fDone = FALSE;
                     }
                    break;
                 default: // any other protocol has not been implemented yet
                    break;
               } // end switch( pRC->iRadioCtrlProtocol )
            }   // end if < more TRANSMIT BANDS > ?
           if( fDone ) // finished polling all TRANSMIT BANDS and their edge frequencies :
            { // Notify the application about possibly 'new data'
              // in pRC->TxBandEdges[ 0 .. pRC->iNumTxBands-1 ] :
              pRC->fTxBandsModified = TRUE; // <- application may POLL and even CLEAR this flag
              pRC->iParameterPollingState = RIGCTRL_POLLSTATE_BAND_STACKING_REGS;
              pRC->iParameterPollingSubState = 0;
            }
           break; // end case RIGCTRL_POLLSTATE_TX_BAND_EDGES

        case RIGCTRL_POLLSTATE_BAND_STACKING_REGS: // (try to) read out the 'band stacking registers' (special VFO memories).
           fDone = TRUE;
           // [in]  pRC->iParameterPollingSubState   : here, Icom's "Frequency Band Code" for cmd 0x1A 0x01 <BandCode> <RegisterCode>
           // [in]  pRC->iParameterPollingArrayIndex : here, Icom's "Register code" (see details further below)
           // [out] pRC->BandStackingRegs[ pRC->iNumBandStackingRegs++ ] ( in RigCtrl_ParseCIV() )
           //
           // How many "stackable bands" are supported by the radio ?
           // This definitely cannot be the number of TX bands, because for
           // example, an IC-7300 modified for all-band-transmit only reported
           // A SINGLE TRANSMIT-BAND (thus pRC->iNumTxBands == 1).
           // The IC-9700 CI-V manual (document #A7508-3EX-4) shows command
           //  0x1A 0x01 followed by a TWO-BYTE DATA FIELD (with 2 * two-digit BCD)
           //  with a rig-specific "Frequency band code" first (1)
           //               and a "Register code", 01..03 next (2) .
           //               "Register code" 01 is the newest 'stacked entry',
           //                               03 is the oldest 'stacked entry'.
           // "Frequency band codes" for an IC-9700 :
           //              01=2m, 02=70cm, 03=23 cm.  No code for "GENE" at all.
           // "Frequency band codes" for an IC-7300 :
           //              01=160m, 02=80m, 03=40m, 04=30m, 05=20m, 06=17m,
           //              07=15m,  08=12m, 09=10m, 10=6m,  11="GENE" (all the rest).
           // So, to avoid even more 'hard-coded rig-specific wisdom' in this
           // module, simply try to read as many Band stacking registers,
           // beginning at "Frequency band code" 01, until the radio responds
           // with an error for "Register code" 01.
           // As in most array-like items polled here,
           //      pRC->iParameterPollingSubState is the 'iterator',
           //      here: running through all "Frequency band codes" a la Icom.
           // In addition, because each "stackable frequency band"
           //              may have up to THREE STACKED ENTRIES,
           //      pRC->iParameterPollingArrayIndex is used for what Icom calls
           //              "Register code", where code 01 (in BCD) means the
           //              MOST RECENT entry (displayed on the left side in
           //              Icom's own 'BAND STACKING REGISTER' display), etc.
           //              (in an IC-7300, a short touch on the MHz digits
           //               first opens a band list only, a longer touch
           //               immediately shows all three 'stack levels'.
           //               That's an incredibly useful feature, thus our
           //               simple "remote GUI" tries something similar.)
           // The RESPONSE (sent by the radio on CI-V command 0x1A 0x01)
           //      will be parsed in RigCtrl_ParseCIV() [in the long list of
           //      sub-commands after 0x1A].
           if( (pRC->iParameterPollingSubState <= (RIGCTRL_NUM_BAND_STACKING_REGS/3) )
            && (pRC->iNumBandStackingRegs < RIGCTRL_NUM_BAND_STACKING_REGS) )
            { // how to read an array of "transmit band edge frequencies" ?
              switch( pRC->iRadioCtrlProtocol )
               { case RIGCTRL_PROTOCOL_ICOM_CI_V : // using Icom's CI-V :
                    ++pRC->iParameterPollingArrayIndex; // here: 1..3 for Icom's three "stack levels"
                    if( pRC->iParameterPollingArrayIndex > 3 ) // finished all three "Register codes" (stack levels) of a band ?
                     {  pRC->iParameterPollingArrayIndex = 1;
                        ++pRC->iParameterPollingSubState;   // next "Frequency band code"
                     }
                    if( pRC->iParameterPollingSubState < 1 )
                     {  pRC->iParameterPollingSubState = 1; // lowest value for Icom's "Frequency band code"
                     }
                    // The following is based on Icom's "IC-970 CI-V Reference Guide",
                    //  revision 4, document ID "A7508-3EX-4", page 15, "Band stacking register".
                    // READ REQUEST: Cmd 0x1A, SubCmd 0x01, followed by TWO two-digit BCD bytes:
                    //   Data byte "(1)" (circled 'one' seems to be a byte index)
                    //               = "Frequency band code" (rig specific,
                    //                 for the IC-9700, 01=2m, 02=70cm, 03=23cm)
                    //   Date byte "(2)" = "Register code" [01, 02, 03, full stop].
                    //
                    //   The READ RESPONSE is terribly complex and again rig-
                    //   specific; more on that in RigCtrl_ParseMemoryContent() .
                    //   Tried on an IC-9700, with the following result:
                    // >  TX 007 FE FE 00 00 1A 01 01 01   ; give me the the stacked memory for "Band 1", 1st entry (top of the stack
                    // >  RX 013 FE FE 00 A2 1A 01 01 01 ..... ; an INCREDIBLY long response
                    //                "band code"---'  '---"Register code"
                    //                (good for nothing
                    //                 because it's rig-specific.
                    //                 We find out the actual BAND from the
                    //                 FREQUENCY - see RigCtrl_ParseMemoryContent() .)
                    //    Lacking a way to determine the NUMBER of "band codes" for band stacking registers,
                    //    simply increment the "band code" until the radio responds with "Not Ok".
                    //    Here for example an IC-7300, which reported ELEVEN "band codes"
                    //    before responding with "NotOK" when asked for the 12th "band code":
                    // >  TX 009 FE FE 00 00 1A 01 11 03 FD ; BandStackReg[11][3} ?
                    // >  RX 017 FE FE 00 94 1A 01 11 03 20 73 10 70 00 03 02 00 00 08 85 00 08 85 .. ; band stack
                    // >  TX 009 FE FE 00 00 1A 01 12 01 FD ; BandStackReg[12][1] ?
                    // >  RX 006 FE FE 00 94 FA FD          ; NotOK for 0x1A010000
                    //    Because the "NotOK" response (0xFA) does not send back
                    //    the command or even sub-command, RigCtrl_ParseCIV()
                    //    is aware of pRC->i32ExpectedResponse_CmdAndSubcode,
                    //    and -in THAT STATE- uses the "NotOK" reponse as a
                    //    marker for the end of the "stackable bands list".
                    sprintf(szComment, "BandStackReg[%d][%d] ?",
                         (int)pRC->iParameterPollingSubState, (int)pRC->iParameterPollingArrayIndex );
                    if( RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                       RIGCTRL_MSGTYPE_RADIO_ID, // MsgType for the "traffic log" ...
                       szComment,  // comment for the "traffic log"
                       0x1A,  // iCmd: CI-V command for "All Kind Of Stuff"
                       0x01,  // iSubCmd: "Band Stacking Register"
                       RigCtrl_IntToBCD2(pRC->iParameterPollingSubState), // "Frequency Band Code" (01 .. ??=
                       RigCtrl_IntToBCD2(pRC->iParameterPollingArrayIndex) ) )  // additional data bytes : None !
                     { // If all works as planned, the transceiver also responds with cmd 0x1A 0x01 ...,
                       // the rest happens in RigCtrl_ParseCIV(), case 0x1A -> RigCtrl_ParseMemoryContent() .
                       pRC->iExpectReadResponseForUnifiedPN = RIGCTRL_PN_BAND_STACKING_REGS; // <- for RigCtrl_ParseCIV()
                       pRC->i32ExpectedResponse_CmdAndSubcode = RCTL_COMBINE_CMD_AND_SUB(0x1A,0x01);  // <- expected command and subcode
                       pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms; // now waiting for the radio to respond with "OK" or "NOT OK" after "Report another BAND STACKING REGISTER"
                       fDone = FALSE;
                     }
                    break;
                 default: // any other protocol has not been implemented yet
                    break;
               } // end switch( pRC->iRadioCtrlProtocol )
            }   // end if < more BAND STACKING REGISTERS > ?
           if( fDone ) // finished polling all 'Band Stacking Registers' :
            { RigCtrl_FinishReadingBandStackingRegs(pRC);
              // '--> next "parameter polling state", etc
            }
           break; // end case RIGCTRL_POLLSTATE_BAND_STACKING_REGS

        case RIGCTRL_POLLSTATE_PARAMS_FOR_HAMLIB: // (try to) read less important parameters that HamlibServer.c may need
           fDone = TRUE;
           while( (i=HLSRV_EnumerateRequiredPNs( pRC->iParameterPollingSubState++)) > 0 )
            { pPI = RigCtrl_GetInfoForUnifiedParameterNumber(i/*iUnifiedPN*/);
              if( pPI != NULL )
               { if( !RigCtrl_HaveValidParamValue( pRC, pPI ) )
                  { RigCtrl_AskForParameter( pRC, pPI );
                    fDone = FALSE;
                    break;
                  }
               }
            }
           if( fDone ) // finished polling all important parameters for HamlibServer.c :
            {
              // ex: if( pRC->iCapabilities  & RIGCTRL_CAPS_SPECTRUM_VIA_CAT )
              // Forget it. Don't rely on hard-coded (built-in) table with
              //            rig-capabilities. It's THE USER'S choice to try, so:
              if( (pRC->iRigControlFlags & RIGCTRL_FLAG_WANT_SPECTRUM_DATA )
                ||(pRC->iCapabilities & RIGCTRL_CAPS_SPECTRUM_VIA_CAT ) )
               { pRC->iParameterPollingState = RIGCTRL_POLLSTATE_CONFIGURE_SPECTRUM; // .. to turn spectrum data OFF
                 pRC->iParameterPollingSubState = 0;
               }
              else // the radio most likely will not "understand" CI-V 0x27, so:
               { // ex: pRC->iParameterPollingState = RIGCTRL_POLLSTATE_DONE;
                 RigCtrl_EnterState_Done( pRC );
               }
            }
           break;
        case RIGCTRL_POLLSTATE_CONFIGURE_SPECTRUM: // (try to) activate transmission of "spectrum scope waveform data"
           if( (pRC->iRigControlFlags & RIGCTRL_FLAG_WANT_SPECTRUM_DATA )
             ||(pRC->iCapabilities & RIGCTRL_CAPS_SPECTRUM_VIA_CAT ) )
            { if( pRC->iRadioCtrlProtocol == RIGCTRL_PROTOCOL_ICOM_CI_V ) // only for "CI-V"..
               {
                 switch( pRC->iParameterPollingSubState++ )
                  { case 0  : // find out if the spectrum scope is in CENTER- or FIXED mode:
                       RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                            RIGCTRL_MSGTYPE_C_SPECTRUM,
                            "center/fixed ?", // comment for traffic log
                            0x27,  // iCmd   : "Scope waveform data"
                            0x14,  // iSubCmd: "Send/read the Scope Center mode or Fixed mode setting"
                            0x00,  // 1st data byte : 0x00 "Fixed" (what a misnomer.. not the FIXED MODE)
                            -1 );  // 2nd data byte : omit for the REQUEST (only used in the RESPONSE)
                       pRC->i32ExpectedResponse_CmdAndSubcode = RCTL_COMBINE_CMD_AND_SUB(0x27,0x14); // at least an IC-7300 responded with 0x27 0x14 ..
                       pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms; // now waiting for the radio to respond after asking for "Spectrum Center/Fixed Mode"
                       break;
                    case 1  : // find out if the spectrum scope's "span" setting
                       RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                            RIGCTRL_MSGTYPE_C_SPECTRUM,
                            "scope span ?", // comment for traffic log
                            0x27,  // iCmd   : "Scope waveform data"
                            0x15,  // iSubCmd: "read the span setting in the Center mode Scope"
                            0x00,  // 1st data byte : 0x00 "Fixed" (what a misnomer.. not the FIXED MODE)
                            -1 );  // 2nd data byte : omit for the REQUEST (only used in the RESPONSE)
                       pRC->i32ExpectedResponse_CmdAndSubcode = RCTL_COMBINE_CMD_AND_SUB(0x27,0x15); // at least an IC-7300 responded with 0x27 0x15 ..
                       pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms; // now waiting for the radio to respond after asking for "Spectrum Span"
                       break;
                    case 2  : // find out the "Edge number setting in the Fixed mode Scope"
                       RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                            RIGCTRL_MSGTYPE_C_SPECTRUM,
                            "edge nr ?", // comment for traffic log
                            0x27,  // iCmd   : "Scope waveform data"
                            0x16,  // iSubCmd: "Edge number setting[s] in the Fixed mode Scope"
                            0x00,  // 1st data byte : 0x00 "Fixed" (what a misnomer.. not the FIXED MODE)
                            -1 );  // 2nd data byte : omit for the REQUEST (only used in the RESPONSE)
                       pRC->i32ExpectedResponse_CmdAndSubcode = RCTL_COMBINE_CMD_AND_SUB(0x27,0x16);
                       pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms; // now waiting for the radio to respond after asking for "Spectrum Edge Number"
                       break;
                    // don't need the "Scope hold function status" yet (0x27 0x17) !
                    case 3  : // find out the "Send/read the Scope Attenuator setting" (IC-7851, n/a on IC-7300)
                       RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                            RIGCTRL_MSGTYPE_C_SPECTRUM,
                            "scope atten. ?", // comment for traffic log
                            0x27,  // iCmd   : "Scope waveform data"
                            0x18,  // iSubCmd: "Scope Attenuator setting"
                            0x00,  // 1st data byte : 0 = main scope, 1 = sub scope
                            -1 );  // 2nd data byte : omit for the REQUEST (only used in the RESPONSE)
                       pRC->i32ExpectedResponse_CmdAndSubcode = RCTL_COMBINE_CMD_AND_SUB(0x27,0x18);
                       pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms; // now waiting for the radio to respond after asking for "Scope Attenuator Setting"
                       // An IC-7300 (which doesn't have an extra attenuator for the SCOPE)
                       // responded with "Not Good" (NG, cmd 0xFA) .
                       break;
                    case 4  : // find out the "Scope Reference level setting"
                       RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                            RIGCTRL_MSGTYPE_C_SPECTRUM,
                            "ref lvl ?", // comment for traffic log
                            0x27,  // iCmd   : "Scope waveform data"
                            0x19,  // iSubCmd: "Scope Reference level settings"
                            0x00,  // 1st data byte : 0 = main scope, 1 = sub scope
                            -1 );  // 2nd data byte : omit for the REQUEST (only used in the RESPONSE)
                       pRC->i32ExpectedResponse_CmdAndSubcode = RCTL_COMBINE_CMD_AND_SUB(0x27,0x19);
                       pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms; // now waiting for the radio to respond after asking for "Spectrum Reference Level"
                       pRC->iParameterPollingArrayIndex = 0; // prepare reading "Fixed edge frequency settings" ...
                       pRC->iScopeNumFreqRanges        = RIGCTRL_MAX_FIXED_EDGE_SCOPE_FREQ_RANGES;
                       pRC->iScopeNumEdgesPerFreqRange = RIGCTRL_MAX_FIXED_EDGES_PER_FREQ_RANGE;
                       break;
#                 if(0)
                    case 5 :  // read all available "Fixed edge frequency settings"
                       if( pRC->iParameterPollingArrayIndex <
                             (  RIGCTRL_MAX_FIXED_EDGE_SCOPE_FREQ_RANGES
                              * RIGCTRL_MAX_FIXED_EDGES_PER_FREQ_RANGE ) )
                        { --pRC->iParameterPollingSubState; // undo pRC->iParameterPollingSubState++
                          iRange = 1 + (pRC->iParameterPollingArrayIndex / RIGCTRL_MAX_FIXED_EDGES_PER_FREQ_RANGE );
                          iEdge  = 1 + (pRC->iParameterPollingArrayIndex % RIGCTRL_MAX_FIXED_EDGES_PER_FREQ_RANGE );
                          if( (iRange<=pRC->iScopeNumFreqRanges) && (iEdge<=pRC->iScopeNumEdgesPerFreqRange) )
                           { // For example, the IC-R8600 only has ONE "range" and ONE "edge" in that range.
                             // This 'if' will limit the number of "questions" we send to the radio.
                             sprintf(szComment, "edge[%d][%d] ?", iRange, iEdge );
                             RigCtrl_IntToBCD_CIV( iRange, &b1, 2/*nDigits*/ );
                             RigCtrl_IntToBCD_CIV( iEdge,  &b2, 2/*nDigits*/ );
                             ++pRC->iParameterPollingArrayIndex; // next "Fixed edge frequency setting"
                             RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                               RIGCTRL_MSGTYPE_C_SPECTRUM,
                               szComment, // comment for traffic log
                               0x27,  // iCmd   : "Scope waveform data"
                               0x1E,  // iSubCmd: "read the Fixed edge frequency settings"
                                 b1,  // 1st data byte : "frequency range" (01..13, guess this is BCD, not hex)
                                 b2); // 2nd data byte : "edge number) (01..03, guess this is also meant as BCD)
                             pRC->i32ExpectedResponse_CmdAndSubcode = RCTL_COMBINE_CMD_AND_SUB(0x27,0x1E);
                             pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms; // now waiting for the radio to respond after asking for a "Fixed Edge Frequency"
                             // When deliberately exceeding the number of "scope frequency ranges"
                             //  ( 13 for an IC-7300 ), the radio responded with "Not Good" :
                             // > 23:23:00.0 TX 009 FE FE 94 E0 27 1E 14 01 FD (scope)    ; edge[14][1] ?
                             // > 23:23:00.1 RX 006 FE FE E0 94 FA FD          (Not OK)
                             // If this happens, RigCtrl_ParseCIV(case 0x271E) limits
                             //    pRC->iScopeNumFreqRanges and pRC->iScopeNumEdgesPerFreqRange
                             //  to what the radio *really* supports.
                             // This seems to be the only method to avoid hard-coding the number,
                             // because other transceivers (like the IC-7851) have LESS frequency ranges,
                             // and future devices (like HF/VHF/UHF/SHF radios) may have MORE .
                           } // end if < valid frequency-RANGE and -EDGE > ?
                          else
                           { // Will waste about 50 ms until the next attempt. So what ?
                           }
                        }
                       break;
#                 else  // skip reading all "Fixed edge frequency settings" :
                    case 5 :
                       break;
#                 endif // (0,1)
                    case 6  : // try to make the scope FASTER VIA SERIAL PORT
                        // (even when "Waterfall" set to FAST, an IC-7300
                        //  only delivered 4 new spectra per second)
                       RigCtrl_SetScopeSpeed( pRC, pRC->iScopeSpeed );
                       break;
                    case 7  : // start or stop periodic transmission of spectrum data (which Icom calls "waveform")
                       RigCtrl_StartStopSpectrumData( pRC, TRUE );
                       break;
                    default :
                       // ex: pRC->iParameterPollingState = RIGCTRL_POLLSTATE_DONE;
                       RigCtrl_EnterState_Done( pRC );
                       pRC->iParameterPollingSubState = 0;
                       break;
                  } // end switch( pRC->iParameterPollingSubState )
               }
              else // not CI-V ?
               { // reading spectrum data not supported here !
                 // ex: pRC->iParameterPollingState = RIGCTRL_POLLSTATE_DONE;
                 RigCtrl_EnterState_Done( pRC );
               }
            }
           else  // ! RIGCTRL_FLAG_WANT_SPECTRUM_DATA ...  so turn it off
            { RigCtrl_StartStopSpectrumData( pRC, FALSE );
              // ex: pRC->iParameterPollingState = RIGCTRL_POLLSTATE_DONE;
              RigCtrl_EnterState_Done( pRC );
             }
           break; // end case RIGCTRL_POLLSTATE_CONFIGURE_SPECTRUM
        case RIGCTRL_POLLSTATE_PASSIVE : // nothing else to do.. or not open for business
        case RIGCTRL_POLLSTATE_DONE    : // finished polling "all we need to know" about the radio
           if( pRC->iPostponedSetMessages & (1<<RIGCTRL_MSGTYPE_SET_FREQUENCY) )
            { RigCtrl_SetVFOFrequency_Internal( pRC, pRC->dblPostponedFrequency );
              pRC->dblVfoFrequency = pRC->dblPostponedFrequency;
              pRC->iPostponedSetMessages &= ~(1<<RIGCTRL_MSGTYPE_SET_FREQUENCY); // "done"
            }
           else if( pRC->iPostponedSetMessages & (1<<RIGCTRL_MSGTYPE_OP_MODE) )
            { RigCtrl_SetOperatingMode_Internal( pRC, pRC->iPostponedOpMode );
              pRC->iOpMode = pRC->iPostponedOpMode;
              pRC->iPostponedSetMessages &= ~(1<<RIGCTRL_MSGTYPE_OP_MODE); // "done"
            }
           else if( pRC->iPostponedSetMessages & (1<<RIGCTRL_MSGTYPE_C_SPECTRUM) )
            { RigCtrl_StartStopSpectrumData( pRC, (pRC->iRigControlFlags & RIGCTRL_FLAG_WANT_SPECTRUM_DATA/*2*/) != 0 );
              pRC->iPostponedSetMessages &= ~(1<<RIGCTRL_MSGTYPE_C_SPECTRUM); // "done"
            }
           else if( pRC->iPostponedSetMessages & (1<<RIGCTRL_MSGTYPE_SCOPE_SPAN) )
            { RigCtrl_SetScopeSpan_Internal( pRC );
              pRC->iPostponedSetMessages &= ~(1<<RIGCTRL_MSGTYPE_SCOPE_SPAN); // "done"
            }
           else if( pRC->iReadPNFifoHead != pRC->iReadPNFifoTail )  // other 'read'-command pending ?
            { RigCtrl_SendReadCommandFromFIFO( pRC );
            }
           else // nothing else to do ... time for some 'periodic polling'
           if( pRC->iParameterPollingState == RIGCTRL_POLLSTATE_DONE )
            { pRC->iScopeDataCountdown_ms -= 50; // countdown for 'Scope'-data ...
              if( pRC->iScopeDataCountdown_ms < 0 )
               { // Reception of scope waveform data has timed out -> kick it alive again.
                 // Since we're in state RIGCTRL_POLLSTATE_DONE here, there's
                 // nothing else to do, so :
                 if( pRC->iRigControlFlags & RIGCTRL_FLAG_WANT_SPECTRUM_DATA )
                  { // Mr Scope NOT alive but he SHOULD be active ...
                    strcpy( szComment, "Remote spectrum timeout, trying to kick it alive again" );
                    strcpy( pRC->sz255LastError, szComment );
                    RigCtrl_AddToTrafficLog( pRC, RIGCTRL_PORT_RADIO, NULL/*pbMessage*/, 0/*iMsgLength*/, RIGCTRL_MSGTYPE_ERROR, szComment );
                    if( RigCtrl_iTrafficMonitorDisplayOptions & RIGCTRL_TRAFFIC_DISPLAY_OPTION_PAUSE_ON_ERROR )
                     { if( ! RigCtrl_fTrafficMonitorPaused )
                        { RigCtrl_AddToTrafficLog( pRC, RIGCTRL_PORT_RADIO, NULL/*pbMessage*/, 0/*iMsgLength*/,
                             RIGCTRL_MSGTYPE_ERROR, "Traffic Log paused after SPECTRUM TIMEOUT (as configured)" );
                        }
                       RigCtrl_fTrafficMonitorPaused = TRUE;
                     } // end if < RIGCTRL_TRAFFIC_DISPLAY_OPTION_PAUSE_ON_ERROR .. >

                    // The trouble started when WSJT-X switched an IC-7300 to TRANSMIT,
                    // causing it to stop the periodic 'Scope Waveform' transmissions:
                    // >  RX 008 FE FE E0 94 1C 00 01 FD (other) ; TRXStatus:1 [WSJT-X to SpecLab]
                    // >  t2 008 FE FE E0 94 1C 00 01 FD (other) ; TRXStatus:1 [SpecLab to IC7300]
                    //   ... other messages, here irrelevant. A second later:
                    // >  Remote spectrum timeout, trying to kick it alive again .
                    // Of course a hopeless case during TX .. and we don't want
                    //    to reconfigure the IC-7300 to have it sending scope data
                    //    during TRANSMISSION (possible, but meaningless, because
                    //    the waterfall would still stop in the preferred FIXED mode).
                    // Cure: The CI-V parser must keep track of all messages
                    //       that may cause the radio to TRANSMIT .
                    //       While the radio TRANSMITS, the scope-data-timeout
                    //       must be IGNORED (no attempt to "kick it alive" again).

                    pRC->iParameterPollingState = RIGCTRL_POLLSTATE_CONFIGURE_SPECTRUM;
                    // Got here, for example, when closing Icom's "RS-BA1"
                    //  because the damned thing sends 27 11 00 (hex) to STOP spectrum data.
                    //  See details in RigCtrl_ParseCIV(), case 0x27 .
                  }
                 pRC->iScopeDataCountdown_ms = RIGCTRL_SPECTRUM_TIMEOUT_MS; // stay calm for a short while..
                 // but longer than the slowest "Sweep speed setting", and
                 // also longer than the time Icom's "RS-BA1" temporarily
                 // turns off the scope while reading dozens of parameters.
               } // end if < reception of "waveform data" timed out >
            } // end else < no "postponed" messages to send >
           break;
        case RIGCTRL_POLLSTATE_TURN_RIG_ON : // (try to) turn the rig on via special command or "wake-up pattern"
           if( pRC->iResponseCountdown_ms <= 0 ) // not busy sending anything else / not waiting for a response anymore ?
            { RigCtrl_TurnPowerOnOrOff( pRC, TRUE/*fPowerOn*/ );
              // '--> sends the 'turn on' command and switches to RIGCTRL_POLLSTATE_WAIT_TURN_ON
              // The message (Icom's command 0x18, sub-command 0x01 = "turn on")
              // is on it's way, and will hopefully be answered with "FB" = "Fine Business":
              // > 19:26:28.3 TX 09B FE FE FE FE FE .. ; turn ON
              // > 19:26:28.3 RX 006 FE FE E0 94 FB FD ; 'OK' for cmd 0x1801
            }
           break; // end case RIGCTRL_POLLSTATE_TURN_RIG_ON
        case RIGCTRL_POLLSTATE_WAIT_TURN_ON: // waiting for the rig to turn on, after RIGCTRL_POLLSTATE_TURN_RIG_ON ..
           if( pRC->iResponseCountdown_ms <= 0 ) // been waiting for long enough (transceiver boot-up time)
            { pRC->i32ExpectedResponse_CmdAndSubcode = -1; // not waiting for any response now ..
              pRC->iResponseCountdown_ms = 500; // .. only WAIT a bit longer before starting to read e.g. the "radio ID" in the next state
              pRC->iParameterPollingState = RIGCTRL_POLLSTATE_START_RD;
            }
           break; // end case RIGCTRL_POLLSTATE_WAIT_TURN_ON
        case RIGCTRL_POLLSTATE_TURN_RIG_OFF: // (try to) turn the rig on via special command (for modern Icom rigs)
           if( pRC->iResponseCountdown_ms <= 0 ) // not busy sending anything else / not waiting for a response anymore ?
            { RigCtrl_TurnPowerOnOrOff( pRC, FALSE/*!fPowerOn*/ );
              // '--> sends the 'turn off' command and switches to RIGCTRL_POLLSTATE_WAIT_TURN_ON
              // The message (Icom's command 0x18, sub-command 0x01 = "turn on")
              // is on it's way, and will hopefully be answered with "FB" = "Fine Business":
              // > 19:26:28.3 TX 09B FE FE FE FE FE .. ; turn ON
              // > 19:26:28.3 RX 006 FE FE E0 94 FB FD ; 'OK' for cmd 0x1801
            }
           break; // end case RIGCTRL_POLLSTATE_TURN_RIG_OFF
        case RIGCTRL_POLLSTATE_WAIT_TURN_OFF: // waiting for the rig to turn off, after RIGCTRL_POLLSTATE_TURN_RIG_OFF ..
           if( pRC->iResponseCountdown_ms <= 0 ) // been waiting for long enough (transceiver turn-off time)
            { pRC->i32ExpectedResponse_CmdAndSubcode = -1; // not waiting for a response anymore ..
              pRC->iResponseCountdown_ms = 500; // .. "do nothing" for at least 500 ms in RIGCTRL_POLLSTATE_TURNED_OFF
              pRC->iParameterPollingState = RIGCTRL_POLLSTATE_TURNED_OFF;
            }
           break; // end case RIGCTRL_POLLSTATE_WAIT_TURN_OFF
        case RIGCTRL_POLLSTATE_TURNED_OFF :  // guess the rig HAS TURNED ITSELF OFF (after RIGCTRL_POLLSTATE_TURN_RIG_OFF)
           break; // end case RIGCTRL_POLLSTATE_TURNED_OFF
        default:
           break;
      } // end switch( pRC->iParameterPollingState )
   }   // end if < enough time passed to send the next command >

  (void)pszComment; // "assigned a value that was never used" .. oh, shut up (it's better than reading a non-initialized variable)



} // end RigCtrl_Handler()


//----------------------------------------------------------------------------
int RigCtrl_EnumerateRequiredPNs( int iZeroBasedIndex )
  // Called from RigControl.c shortly after 'connecting' the radio,
  // to poll certain parameters that a REMOTE HAMLIB CLIENT may want, like...
  //  RIGCTRL_PN_RIT_FREQ,    RIGCTRL_PN_XIT_FREQ,
  //  RIGCTRL_PN_RIT_ENABLED, RIGCTRL_PN_XIT_ENABLED, end who-knows-what .
  // [in]     : iZeroBasedIndex = the "enumerator" (zero-based index as in C).
  //            Incremented by the caller from zero, stepping by one,
  //            until the RETURN VALUE from HLSRV_EnumerateRequiredPNs() is zero.
  // [return] : "Unified parameter number" for RigControl.c ,
  //            or ZERO when all 'required parameter numbers' are through.
{ switch( iZeroBasedIndex )
   { // Begin with parameters that ANY rig should support...
 //ex:case 0: return RIGCTRL_PN_TRANSCEIVER_ID; // first determine the "Transceiver ID" alias "Radio ID" alias "Default CI-V Address"
     case  0: return RIGCTRL_PN_FREQUENCY;
     case  1: return RIGCTRL_PN_TRANSMITTING;   // aka "PTT flag" (0=receive, 1=transmit)
     case  2: return RIGCTRL_PN_TRANSMIT_FREQ;
     case  3: return RIGCTRL_PN_OP_MODE;
     case  4: return RIGCTRL_PN_SPLIT_MODE;
     case  5: return RIGCTRL_PN_NUM_TX_BANDS;
     case  6: return RIGCTRL_PN_AUDIO_VOLUME_PERCENT;
     case  7: return RIGCTRL_PN_RF_GAIN_PERCENT;
     case  8: return RIGCTRL_PN_SQUELCH_LEVEL_PERCENT;
     case  9: return RIGCTRL_PN_SCOPE_REF_LEVEL;
     default: break;
     // Note: Besides the list of parameters to read in RigCtrl_EnumerateRequiredPNs(),
     //      there's a similar list of parameters to read in HLSRV_EnumerateRequiredPNs().
   }
  return 0;
} // RigCtrl_EnumerateRequiredPNs()

//--------------------------------------------------------------------------
void RigCtrl_SetLocalPTTFlag( T_RigCtrlInstance *pRC, BOOL fTransmitting)
  // Warning: This function may be called from a WORKER THREAD,
  //          e.g. KeyerThread.c : KeyerThread(), and thus it interrupts
  //          RigCtrl_Handler().  To make this sufficiently thread safe
  //          without a synchronisation object, and without waiting for anything,
  //          pRC->fTransmitting is ONLY WRITTEN HERE, and READ-ONLY elsewhere.
{
  if( pRC->fLocalPTTFlag != fTransmitting )
   { pRC->fLocalPTTFlag = fTransmitting; // <- good place for a breakpoint (*)
     // '--> Monitored in RigCtrl_Handler(), and if this flag differs
     //      from pRC->iTransmitting (RIGCTRL_PN_TRANSMITTING),
     //      and the radio's protocol permits, a command to switch between
     //      RX and TX will be sent.
     // What else to do on a STATE TRANSITION of the 'PTT flag' ?
     // (*) In the Remote CW Keyer application, the PTT flag was switched by...
     //    - on "manual PTT control", client side, when called from
     //      KeyerThread() when CwKeyer_GetDigitalInput(CwKeyer_Config.iManualPTTInput)
     //      indicated a transition on the 'manual PTT input' (microphone, footswitch, etc)
     //    - when NOT using manual PTT control:
   } // end if( pRC->fLocalPTTFlag != fTransmitting )

} // end RigCtrl_SetLocalPTTFlag()


//--------------------------------------------------------------------------
void RigCtrl_TurnPowerOnOrOff( T_RigCtrlInstance *pRC,BOOL fPowerOn )
{
  // Turning an Icom radio ON (when it's really turned off) requires
  // a special pattern, with the 0xFE preamble repeated MANY times :
  if( fPowerOn )
   { RigCtrl_SendCommandToTurnRigOn( pRC, RIGCTRL_PORT_RADIO );
     pRC->i32ExpectedResponse_CmdAndSubcode = RCTL_COMBINE_CMD_AND_SUB(0x18,0x01); // IC-7300 responded with the 0xFB-"OK" to the 'turn on' command (cmd 0x18 sub 0x01) !
     pRC->iResponseCountdown_ms = 500; // wait for the response ("FB") for the "turn on" command just sent
     pRC->iParameterPollingState = RIGCTRL_POLLSTATE_WAIT_TURN_ON; // now waiting for an "OK"-response after the command to "turn on"
   }
  else // Turning a radio OFF is a "normal CI-V" command (cmd 0x18 + sub-cmd 0x00)
   { RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
       RIGCTRL_MSGTYPE_FLAG_CIV | RIGCTRL_MSGTYPE_POWER_ON_OFF, "turn OFF",
       0x18,  // [in] CI-V main command
       0x00,  // [in] sub-command, here: 0x00 to turn OFF
       -1,    // [in] optional 1st parameter byte. -1 to omit.
       -1);   // [in] optional 2nd parameter byte. -1 to omit.
     pRC->i32ExpectedResponse_CmdAndSubcode = RCTL_COMBINE_CMD_AND_SUB(0x18,0x00); // IC-7300 responded with the 0xFB-"OK" to the 'turn off' command (cmd 0x18 sub 0x00) !
     pRC->iResponseCountdown_ms  = 500; // not waiting for a response, but 'busy' for at least 500 ms
     pRC->iParameterPollingState = RIGCTRL_POLLSTATE_TURNED_OFF;
   }
} // end RigCtrl_TurnPowerOnOrOff()


//--------------------------------------------------------------------------
BOOL RigCtrl_SetVFOFrequency_Internal( T_RigCtrlInstance *pRC,
                              double dblFreqHz ) // [in] freq in HERTZ
  // As most functions, returns TRUE when successful, otherwise FALSE.
  // Returns BEFORE the remote rig has actually 'digested' the new setting !
  //
{
  char sz16FreqAsText[20], szComment[ RIGCTRL_MAX_COMMENT_LENGTH+4 ];
  char *pszLowDigit;
  BYTE b80TxBuffer[84];
  BYTE b;
  int  iTxLength = 0;
  int  i,n;
  int  iRigCtrlPort = RIGCTRL_PORT_RADIO;
  BOOL fOk = FALSE;

  // convert frequency in Hertz into a string: 10 digits without fraction,
  // LEADING ZEROES, ok for up to 9 GHz ...
  sprintf(sz16FreqAsText, "%010.0lf", (double)dblFreqHz );
  pszLowDigit = sz16FreqAsText + (strlen(sz16FreqAsText)-1); // now points to least significant digit

  switch( pRC->iRadioCtrlProtocol )
   {  // which language does the radio speak ?
    case RIGCTRL_PROTOCOL_NONE :
    default:
         break;  // unknown protocol, can only send "direct" strings

    case RIGCTRL_PROTOCOL_ICOM_CI_V : // Icom's "CI-V" protocol ...
         b80TxBuffer[iTxLength++] = 0xFE; // fixed preamble code for CI-V protocol
         b80TxBuffer[iTxLength++] = 0xFE; // fixed preamble code for CI-V protocol
         b80TxBuffer[iTxLength++] = (BYTE)pRC->iRadioDeviceAddr;  // "to" transceiver's address (0x48 = IC706)
         b80TxBuffer[iTxLength++] = (BYTE)pRC->iRadioMasterAddr;  // "from" controller's address (often 0xE0)
         b80TxBuffer[iTxLength++] = 0x05; // CI-V command "Write operating frequency data"
         // Note: Older CI-V reference manuals [CIV_REF_MAN_V3_2] clearly mention that
         //       command 0x05 "writes operating frequency data into a displayed VFO
         //                     or memory channel", and that the radio responds
         //                     to THIS command with 0xFB = "Ok code"
         //                                  or with 0xFA = "NG code".
         //       Newer CI-V reference manuals don't specify for which commands
         //       the radio will respond with 0xFB / 0xFA, or not respond at all.
         //
         // If the frequency is below 100 MHz, may have to use 4 BCD-bytes only,
         // because some old radios like the IC-735 only support 4 bytes = 8 digits.
         // Modern radios (and VHF/UHF rigs) support 5 BCD data bytes = 10 digits.
         // Furthermore, some radios (like the IC-706) NEED these long numbers,
         // because otherwise the 100-MHz-place remains UNCHANGED !
         n = (  pRC->iDefaultAddress==RIGCTRL_DEF_ADDR_IC_725
             || pRC->iDefaultAddress==RIGCTRL_DEF_ADDR_IC_735
             || pRC->iDefaultAddress==RIGCTRL_DEF_ADDR_IC_751A ) ? 4 : 5;
         for(i=0; (i<n) && (pszLowDigit>=sz16FreqAsText); ++i)
          { b =  (BYTE)( (*pszLowDigit--)-'0');     // lowest digit, beginning at 1 Hz
            b |= (BYTE)(((*pszLowDigit--)-'0')<<4); // next digit, beginning at 10 Hz
            b80TxBuffer[iTxLength++] = b;        // append next two digits to BCD data area
          }
         b80TxBuffer[iTxLength++] = 0xFD;        // CI-V code for "End Of Message"
         break;  // end case <ICOM CI-V protocol>

   } // end switch( pRC->iRadioCtrlProtocol )

  if( iTxLength>0 )
   {
     sprintf( szComment, "set freq to %.1lf Hz", (double)dblFreqHz );
     fOk = RigCtrl_SendAndLogMessage( pRC, iRigCtrlPort, b80TxBuffer, iTxLength,
              FALSE/*don't wait for completion*/,
              RIGCTRL_MSGTYPE_SET_FREQUENCY | RIGCTRL_MSGTYPE_FLAG_TX, szComment );
     if( fOk )
      { pRC->i32ExpectedResponse_CmdAndSubcode = RCTL_COMBINE_CMD_ONLY(0x05); // the radio's CI-V response "Set frequency" (cmd 0x05) doesn't contain command or subcode, so REMEMBER what has been "set" this way
        pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms; // now waiting for the "OK"/"NOT OK" response for "Set frequency" (cmd 0x05)
      }

   } // end if < something 'sendable' > ?
  else
   { fOk = FALSE;
   }

  return fOk;
} // end RigCtrl_SetVFOFrequency_Internal()


//--------------------------------------------------------------------------
BOOL RigCtrl_SetVFOFrequency( T_RigCtrlInstance *pRC, /* API */
                              double dblFreqHz ) // [in] new radio freq in HERTZ
  // As most functions, returns TRUE when successful, otherwise FALSE.
  // Returns BEFORE the remote rig has actually 'digested' the new setting,
  //         thus TRUE only means "the new setting is on its way" !
  //
  // To retrieve the current 'VFO frequency', simply read pRC->dblVfoFrequency .
  // To convert an 'baseband frequency' (in the radio's demodulated output stream)
  //        into a 'radio frequency', see RigCtrl_BasebandToRadioFrequency().
  //
  // Shall work on the server side (directly controlling a radio via 'COM' port)
  //        and on the client side (via CwNet.c, which polls the most important
  //              parameters and sends them 'when modified' in CwNet_OnPoll() )
  //
  // Callers in DL4YHF's "Remote CW Keyer" :
  //  (1) TCustomForm::MouseWheelHandler() -> ... -> TKeyerMainForm::FormMouseWheel()
  //       -> Ed_VFOKeyDown() -> RigCtrl_SetVFOFrequency()
  //  (2) CB_BandClick() [retrieves a "Band Stacking Register" from the multi-column combo]
  //       -> RigCtrl_SwitchToFreqMemEntry() -> RigCtrl_SetVFOFrequency() [+ others]
{
  // This API function may be called by the application at any time, but it may
  // have to be postponed (internally) when still busy from an 'uninterruptable'
  //      sequence of commands in RigCtrl_Handler().
  // In ANY case, immediately copy the new "wanted" VFO frequency :
  pRC->dblVfoFrequency = dblFreqHz;
  if( pRC->iResponseCountdown_ms > 0 ) // still waiting for a response -> "do not send anything at the moment"..
   {  pRC->iPostponedSetMessages |= (1<<RIGCTRL_MSGTYPE_SET_FREQUENCY);
      pRC->dblPostponedFrequency = dblFreqHz;
      return TRUE;  // cannot send AT THE MOMENT, but data are (almost) "on the way" !
   }
  else // no need to postpone the "Set Frequency"-command, so send immediately:
   {  pRC->iPostponedSetMessages &= ~(1<<RIGCTRL_MSGTYPE_SET_FREQUENCY);
      return RigCtrl_SetVFOFrequency_Internal( pRC, dblFreqHz );
      // Note: Even though an IC-7300 sent the "OK" (acknowledge) much later,
      //       the VFO frequency could be modified many times per second.
      //       Thus do NOT reload pRC->iResponseCountdown_ms
      //       in this case, even though we expect a response for the CI-V
      //       command that has just been sent. From SL's "CAT traffic" log:
      // > 18:08:39.5: TX 007 FE FE 94 E0 19 00 FD               (radio ID)    ; get transceiver ID
      // > 18:08:40.4: RX 008 FE FE E0 94 19 00 94 FD            (radio ID)    ; rig=IC-7300
      // > 18:08:40.4: TX 006 FE FE 94 E0 03 FD                  (VFO freq)    ; read VFO frequency
      // > 18:08:41.4: RX 00B FE FE E0 94 03 50 00 55 03 00 FD   (VFO freq)    ; 3550050.0 Hz
      // > 18:08:41.5: TX 006 FE FE 94 E0 04 FD                  (op mode)     ; read operating mode
      // > 18:08:42.4: RX 008 FE FE E0 94 04 03 03 FD            (op mode)     ; CW, filter=3
      // > 18:08:59.5: TX 00B FE FE 94 E0 05 60 00 55 03 00 FD   (VFO freq)    ; 3550060.0 Hz
      // > 18:09:00.3: RX 006 FE FE E0 94 FB FD                  (OK)
      // > 18:09:01.7: TX 00B FE FE 94 E0 05 70 00 55 03 00 FD   (VFO freq)    ; 3550070.0 Hz (step #1)
      // > 18:09:01.8: TX 00B FE FE 94 E0 05 80 00 55 03 00 FD   (VFO freq)    ; 3550080.0 Hz (step #2)
      // > 18:09:01.9: TX 00B FE FE 94 E0 05 90 00 55 03 00 FD   (VFO freq)    ; 3550090.0 Hz (step #3)
      // > 18:09:02.4: RX 006 FE FE E0 94 FB FD                  (OK)          (ack for step #1)
      // > 18:09:02.4: RX 006 FE FE E0 94 FB FD                  (OK)          (ack for step #2)
      // > 18:09:02.4: RX 006 FE FE E0 94 FB FD                  (OK)          (ack for step #3)
      // -> At least modern Icom radios seem to have large buffers
      //    for RECEPTION and TRANSMISSION. So for these *simple* commands,
      //     it's unnecessary to wait for a response (which would have wasted
      //     2.4 minus 1.7 = 0.7 seconds in the example shown above) .
   }

} // end RigCtrl_SetVFOFrequency()


//--------------------------------------------------------------------------
CPROT BOOL RigCtrl_SwitchToBand( T_RigCtrlInstance *pRC,
             DWORD dwNewBand ) // [in] RIGCTL_BAND_xyz (only ONE bit per band)
  // Return value: TRUE = band-switching command was accepted and it on its way
  //                      "to the radio". Some time later, we know the actual
  //                      FREQUENCY (selected by the rig itself, e.g. from
  //                      Icom's "Band Stacking" registers). Then,
  //                      pRC->dblVfoFrequency will change automatically.
  //               FALSE = illegal band (not supported), or rejected by
  //                       local control . pRC->dblVfoFrequency will not change.
  //        Note: There is no extra member in T_RigCtrlInstance for the
  //              currently active band. If necessary, use
  //              RigCtrl_FrequencyToBand() to convert e.g.
  //              pRC->dblVfoFrequency or pRC->dblTxFrequency into one of the
  //              'band bitmasks' (RIGCTRL_BAND_...) .
  //
  // This function is important for radios without 'general coverage', e.g.
  //  IC-9700 (with RIGCTL_BAND_2M, RIGCTL_BAND_70CM, and RIGCTL_BAND_23CM),
  //  where you cannot simply increment the frequency in 1- or 10-MHz steps
  //  because those non-general-coverage radios won't accept frequencies
  //  BETWEEN BANDs.
{

  return FALSE;

} // end  RigCtrl_SwitchToBand()

//--------------------------------------------------------------------------
CPROT DWORD RigCtrl_GetAvailableBands( T_RigCtrlInstance *pRC )
   // Returns a bitwise combination of RIGCTL_BAND_xyz .
   //  [in] pRC->dwAvailableBands : initialized in RigCtrl_Init()
   //                               but restrived/revised later, depending on
   //                               what the protocol (and THE RADIO) permit.
   //  [in] pRC->TxBandEdges[], pRC->iNumTxBands : read from the rig via CI-V .
{ DWORD dwAvailableBands = pRC->dwAvailableBands;
        // '-> Usually copied from  RigCtrl_RadioInfo_CIV[], e.g.
        //     RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_6M | RIGCTRL_BAND_4M = 0x35FB .
        //  Note the absence of e.g. the 60-meter-band and 11 meter (CB) here !

  // *IF* the connected radio was kind enough to report its 'Band Edge Frequencies',
  // revise the bands really supported from those 'Band Edges'.
  // Spoiler alert: An IC-7300 that had been drilled up to operate on 60 m and 4 m
  //                reported A SINGLE TRANSMIT BAND, from 100 kHz to 74 MHz.
  //                That's nonsense of course .. it cannot, it must not,
  //                and we DON'T WANT IT TO, transmit on the entire range !
  //                (Remember this after you fried the PA's output transformer
  //                 and/or the internal ATU when trying to send with 100 watts
  //                 on 100 kHz).
  //                Thus, unless already contained in dwAvailableBands,
  //                additional bands are only added if they are BETWEEN
  //                SUPPORTED BANDS, e.g. add RIGCTRL_BAND_60M only if the
  //                rig also supports 80 and 40 meters.
  //

  return dwAvailableBands;
  // In DL4YHF's "Remote CW Keyer", this result (dwAvailableBands) will be
  //  listed in the 'Bands' selection drop-down list on the 'TRX' tab,
  //  possibly along with the edge frequencies for a particular band.
  //  See FillComboWithBandList() .


} // end RigCtrl_GetAvailableBands()

//--------------------------------------------------------------------------
static BOOL InRange( double dblVal, double dblMin, double dblMax )
{  return (dblVal>=dblMin) && (dblVal<=dblMax);
}


//--------------------------------------------------------------------------
DWORD RigCtrl_FrequencyToBand( double dblFrequency ) // -> RIGCTL_BAND_xyz
{
  if( InRange( dblFrequency, 135700.0, 137800.0 ) ) // LF "2200 meters" ?
   { return RIGCTRL_BAND_2200M;
   }
  if( InRange( dblFrequency, 472000.0, 479000.0 ) ) // MF "630 meters"
   { return RIGCTRL_BAND_600M;
   }
  if( InRange( dblFrequency, 1800000.0, 2000000.0 ) ) // MF "160 meters"
   { return RIGCTRL_BAND_160M;
   }
  if( InRange( dblFrequency, 3500000.0, 4000000.0 ) ) // "80 m"
   { return RIGCTRL_BAND_80M;
   }
  if( InRange( dblFrequency, 5351.5e3, 5366.5e3 ) ) // "60 m" (from WRC-15, worldwide)
   { return RIGCTRL_BAND_60M;
   }
  if( InRange( dblFrequency,  7.0e6,  7.3e6  ) ) // "40 m"
   { return RIGCTRL_BAND_40M;
   }
  if( InRange( dblFrequency, 10.0e6, 10.15e6 ) ) // "30 m"
   { return RIGCTRL_BAND_30M;
   }
  if( InRange( dblFrequency, 14.0e6, 14.35e6 ) ) // "20 m"
   { return RIGCTRL_BAND_20M;
   }
  if( InRange( dblFrequency, 18.068e6,18.168e6 ) ) // "17 m"
   { return RIGCTRL_BAND_17M;
   }
  if( InRange( dblFrequency, 21.0e6, 21.45e6 ) )  // "15 m"
   { return RIGCTRL_BAND_15M;
   }
  if( InRange( dblFrequency, 24.89e6, 24.99e6 ) ) // "12 m"
   { return RIGCTRL_BAND_12M;
   }
  if( InRange( dblFrequency, 26.565e6, 27.404e6 ) ) // "11 m" CB in DL (80 channels)
   { return RIGCTRL_BAND_11M;
   }
  if( InRange( dblFrequency, 28e6, 29.7e6 ) ) // "10 m"
   { return RIGCTRL_BAND_10M;
   }
  if( InRange( dblFrequency, 50e6, 54e6 ) ) // "6 m"   in most of ITU Region 1
   { return RIGCTRL_BAND_6M;
   }
  if( InRange( dblFrequency, 70e6, 70.5e6 ) ) // "4 m" in a few lucky countries (not DL)
   { return RIGCTRL_BAND_4M;
   }
  if( InRange( dblFrequency, 144e6, 146e6 ) ) // "2 m"   in ITU Region 1
   { return RIGCTRL_BAND_2M;
   }
  if( InRange( dblFrequency, 430e6, 440e6 ) ) // "70 cm" in ITU Region 1
   { return RIGCTRL_BAND_70CM;
   }
  if( InRange( dblFrequency, 1240e6, 1300e6 ) ) // "23 cm" in ITU Region 1
   { return RIGCTRL_BAND_23CM;
   }
  if( InRange( dblFrequency, 2300e6, 2450e6 ) ) // "13 cm" (note the GAP between 2310 and 2390 MHz)
   { return RIGCTRL_BAND_13CM;
   }
  if( InRange( dblFrequency, 3300e6, 3500e6 ) ) // "9 cm"
   { return RIGCTRL_BAND_9CM;
   }
  if( InRange( dblFrequency, 5650e6, 5850e6 ) ) // "6 cm"
   { return RIGCTRL_BAND_6CM;
   }
  if( InRange( dblFrequency, 10e9, 10.5e9 ) )   // "3 cm"
   { return RIGCTRL_BAND_3CM;
   }

  return 0;

} // end RigCtrl_FrequencyToBand()

//--------------------------------------------------------------------------
static void AddRange( double dblMin, double dblMax,
                      double *pdblCombinedMin, double *pdblCombinedMax )
{ if( (*pdblCombinedMin == RIGCTL_NOVALUE_DOUBLE ) || ( dblMin < *pdblCombinedMin) )
   {   *pdblCombinedMin = dblMin;
   }
  if( (*pdblCombinedMax == RIGCTL_NOVALUE_DOUBLE ) || ( dblMax > *pdblCombinedMax) )
   {   *pdblCombinedMax = dblMax;
   }
} // end AddRange()


//--------------------------------------------------------------------------
void RigCtrl_BandToFrequencyRange(
         DWORD dwBand, // [in] RIGCTRL_BAND_... (usually only ONE BIT set,
                       //                        but combinations are allowed)
         double *pdblFmin_Hz, double *pdblFmax_Hz ) // [out] band edge frequencies
  // If dwBand doesn't contain a valid 'band bit', the result
  // (*pdblFmin_Hz and *pdblFmax_Hz) will be set to RIGCTL_NOVALUE_DOUBLE.
{
  DWORD dwRegion = dwBand & RIGCTRL_BAND_ITU_MASK;
        // '--> not really supported here yet.
        //      Only a few amateur radio bands 'demonstrate the purpose' .

  *pdblFmin_Hz = *pdblFmax_Hz = RIGCTL_NOVALUE_DOUBLE;
  if( dwBand & RIGCTRL_BAND_2200M )
   { AddRange( 135700.0, 137800.0, pdblFmin_Hz, pdblFmax_Hz );
   }
  if( dwBand & RIGCTRL_BAND_600M )
   { AddRange( 472000.0, 479000.0, pdblFmin_Hz, pdblFmax_Hz );
   }
  if( dwBand & RIGCTRL_BAND_160M )
   { AddRange( 1800e3, 2000e3, pdblFmin_Hz, pdblFmax_Hz );
   }
  if( dwBand & RIGCTRL_BAND_80M )
   { switch( dwRegion )
      { case RIGCTRL_BAND_ITU_RGN_1: // Europe, Africa, FSU, Mongolia, Middle East, Iraq
        default:
           AddRange( 3500e3, 3800e3, pdblFmin_Hz, pdblFmax_Hz ); // "80 m" in ITU region 1
           break;
        case RIGCTRL_BAND_ITU_RGN_2: // N & S America, Greenland, some east pacific islands
           AddRange( 3500e3, 4000e3, pdblFmin_Hz, pdblFmax_Hz ); // "80 m" in ITU region 2
           break;
        case RIGCTRL_BAND_ITU_RGN_3: // Far East, non-FSU Asia, Iran, Australia, Oceania
           AddRange( 3500e3, 3900e3, pdblFmin_Hz, pdblFmax_Hz ); // "80 m" in ITU region 3
           break;
      }
   }
  if( dwBand & RIGCTRL_BAND_60M )
   { AddRange( 5351.5e3, 5366.5e3, pdblFmin_Hz, pdblFmax_Hz ); // "60 m" (from WRC-15, worldwide)
   }
  if( dwBand & RIGCTRL_BAND_40M )
   { switch( dwRegion )
      { case RIGCTRL_BAND_ITU_RGN_1: // Europe, Africa, FSU, Mongolia, Middle East, Iraq
        default:
           AddRange( 7.0e6,  7.2e6, pdblFmin_Hz, pdblFmax_Hz );  // "40 m" in ITU region 1
           break;
        case RIGCTRL_BAND_ITU_RGN_2: // N & S America, Greenland, some east pacific islands
           AddRange( 7.0e6,  7.3e6, pdblFmin_Hz, pdblFmax_Hz );  // "40 m" in ITU region 2
           break;
        case RIGCTRL_BAND_ITU_RGN_3: // Far East, non-FSU Asia, Iran, Australia, Oceania
           AddRange( 7.0e6,  7.2e6, pdblFmin_Hz, pdblFmax_Hz );  // "40 m" in ITU region 3
           break;
      }
   }
  if( dwBand & RIGCTRL_BAND_30M )
   { AddRange( 10.0e6, 10.15e6, pdblFmin_Hz, pdblFmax_Hz ); // "30 m" in all ITU regions ?
   }
  if( dwBand & RIGCTRL_BAND_20M )
   { AddRange( 14.0e6, 14.35e6, pdblFmin_Hz, pdblFmax_Hz ); // "20 m" in all ITU regions ?
   }
  if( dwBand & RIGCTRL_BAND_17M )
   { AddRange( 18.068e6,18.168e6, pdblFmin_Hz, pdblFmax_Hz ); // "17 m" in all ITU regions ?
   }
  if( dwBand & RIGCTRL_BAND_15M )
   { AddRange( 21.0e6, 21.45e6, pdblFmin_Hz, pdblFmax_Hz );  // "15 m" in all ITU regions ?
   }
  if( dwBand & RIGCTRL_BAND_12M )
   { AddRange( 24.89e6, 24.99e6, pdblFmin_Hz, pdblFmax_Hz ); // "12 m" in all ITU regions ?
   }
  if( dwBand & RIGCTRL_BAND_11M )
   { AddRange( 26.565e6, 27.405e6, pdblFmin_Hz, pdblFmax_Hz ); // "11 m" Citizens Band
   }
  if( dwBand & RIGCTRL_BAND_10M )
   { AddRange( 28e6, 29.7e6, pdblFmin_Hz, pdblFmax_Hz ); // "10 m" in all ITU regions ?
   }
  if( dwBand & RIGCTRL_BAND_6M )
   { AddRange( 50e6, 54e6, pdblFmin_Hz, pdblFmax_Hz ); // "6 m" in most of ITU Region 1 (except for poor old Germany)
   }
  if( dwBand & RIGCTRL_BAND_4M )
   { AddRange( 70e6, 70.5e6, pdblFmin_Hz, pdblFmax_Hz ); // "4 m" in a few lucky countries (not DL; only "temporarily tolerated")
   }
  if( dwBand & RIGCTRL_BAND_2M )
   { switch( dwRegion )
      { case RIGCTRL_BAND_ITU_RGN_1: // Europe, Africa, FSU, Mongolia, Middle East, Iraq
        default:
           AddRange( 144e6, 146e6, pdblFmin_Hz, pdblFmax_Hz ); // "2 m"   in ITU Region 1
           break;
        case RIGCTRL_BAND_ITU_RGN_2: // N & S America, Greenland, some east pacific islands
           AddRange( 144e6, 148e6, pdblFmin_Hz, pdblFmax_Hz ); // "2 m"   in ITU Region 2
           break;
        case RIGCTRL_BAND_ITU_RGN_3: // Far East, non-FSU Asia, Iran, Australia, Oceania
           AddRange( 144e6, 148e6, pdblFmin_Hz, pdblFmax_Hz ); // "2 m"   in ITU Region 3
           break;
      }
   }
  if( dwBand & RIGCTRL_BAND_70CM )
   { AddRange( 430e6, 440e6, pdblFmin_Hz, pdblFmax_Hz ); // "70 cm" in ITU Region 1
   }
  if( dwBand & RIGCTRL_BAND_23CM )
   { AddRange( 1240e6, 1300e6, pdblFmin_Hz, pdblFmax_Hz ); // "23 cm" in ITU Region 1
   }
  if( dwBand & RIGCTRL_BAND_13CM )
   { AddRange( 2300e6, 2450e6, pdblFmin_Hz, pdblFmax_Hz ); // "13 cm" (note the GAP between 2310 and 2390 MHz)
   }
  if( dwBand & RIGCTRL_BAND_9CM )
   { AddRange( 3300e6, 3500e6, pdblFmin_Hz, pdblFmax_Hz ); // "9 cm"
   }
  if( dwBand & RIGCTRL_BAND_6CM )
   { AddRange( 5650e6, 5850e6, pdblFmin_Hz, pdblFmax_Hz ); // "6 cm"
   }
  if( dwBand & RIGCTRL_BAND_3CM )
   { AddRange( 10e9, 10.5e9, pdblFmin_Hz, pdblFmax_Hz );   // "3 cm"
   }

} // end RigCtrl_BandToFrequencyRange()


//--------------------------------------------------------------------------
BOOL RigCtrl_SwitchToFreqMemEntry( T_RigCtrlInstance *pRC,
                                   T_RigCtrlFreqMemEntry *pFreqMemEntry )
  // Switches to e.g. a certain "Band Stacking Register" (read out earlier)
  // or a classic 'memory channel'. Both include at least the operating
  // frequency and the operating mode (USB,LSB,CW,..), and maybe an awful
  // lot more (split mode, filter bandwidth, ... ) .
  //
  // Shall work on the server side (directly controlling a radio via 'COM' port)
  //        and on the client side (which sends a few commands via TCP) .
  //
{ BOOL fOk = FALSE;

  // So far, no command has been found in the IC-9700 CI-V manual to
  //         SWITCH TO a certain band stacking register. There only seemed
  //         to be a command to "Send/read band stacking register contents".
  //         So send MULTIPLE COMMANDS to switch frequency, mode, bandwidth,
  //         and whatever else belongs to a band stacking register (future plan).

  if( pFreqMemEntry->RxTx[0].dblOperatingFreq_Hz > 0.0 )
   { fOk = RigCtrl_SetVFOFrequency( pRC, pFreqMemEntry->RxTx[0].dblOperatingFreq_Hz );
   }
  if( pFreqMemEntry->RxTx[0].iOpMode != RIGCTRL_OPMODE_UNKNOWN )
   { fOk |= RigCtrl_SetOperatingMode( pRC, pFreqMemEntry->RxTx[0].iOpMode );
     // ,-----------'
     // '--> Must not immediately send anything yet, because the command
     //      to switch the FREQUENCY must be acknowledged by the remotely
     //      controlled radio first (RIGCTRL_MSGTYPE_OK, in CI-V: unspecific "OK" = cmd 0xFB).
     //
   }

  return fOk;

} // end RigCtrl_SwitchToFreqMemEntry()


//--------------------------------------------------------------------------
BOOL RigCtrl_SendCmdToReadUnifiedPN( T_RigCtrlInstance *pRC, int iUnifiedPN )
  // Non-blocking;  doesn't wait for "arrival" of the response.
  // The return value only indicates if we could *SEND* the command .
  // The read value will LATER be stored in the equivalent member variable,
  // e.g. iUnifiedPN=RIGCTRL_PN_S_METER_LEVEL_DB  ->  pRC->iSMeterLevel_dB .
{
  // Like many other API functions, the actual transmission may have to be
  // postponed (internally) when still busy from an 'uninterruptable'
  //      sequence of commands in RigCtrl_Handler().
  // Since there is a FIFO for this anyway, use it :
  int iNewFifoHead = (pRC->iReadPNFifoHead + 1) % RIGCTRL_READ_PN_FIFO_SIZE;
  if( iNewFifoHead != pRC->iReadPNFifoTail )  // FIFO not completely full ?
   { pRC->iReadPNFifo[pRC->iReadPNFifoHead] = iUnifiedPN;
     pRC->iReadPNFifoHead = iNewFifoHead;
     return TRUE;  // command to "read something" queued up successfully
     // The rest happens in RigCtrl_Handler()
     //    -> RigCtrl_SendReadCommandFromFIFO() and RigCtrl_ParseCIV() .
   }
  else
   { return FALSE; // out of FIFO space to queue up another 'read'-command
     // (got here quite often, because RigCtrl_Handler() was often busy
     //  from receiving spectra from an IC-7300 / IC-9700, and barely had
     //  the chance to ask for new S-meter readings or similar.)
   }
} // end RigCtrl_SendCmdToReadUnifiedPN()

//--------------------------------------------------------------------------
void RigCtrl_SendReadCommandFromFIFO( T_RigCtrlInstance *pRC )
  // [in] pRC->iReadPNFifo[pRC->iReadPNFifoTail++]
  //
  // Only called from RigCtrl_Handler() when another parameter needs to be
  // polled (e.g. RIGCTRL_PN_S_METER_LEVEL_DB), and the current state
  // ALLOWS sending another frame to the remotely controlled rig.
  //
  // The to-be-read "unified parameter" has been requested by the application
  // via RigCtrl_SendCmdToReadUnifiedPN( T_RigCtrlInstance *pRC, int iUnifiedPN );
{ int iUnifiedPN = pRC->iReadPNFifo[pRC->iReadPNFifoTail];
  int iCmd=-1, iSubCmd =-1;
  char *pszComment = "read param";
  pRC->iReadPNFifoTail = (pRC->iReadPNFifoTail+1) % RIGCTRL_READ_PN_FIFO_SIZE;

  switch( pRC->iRadioCtrlProtocol )
   {  // which language does the radio speak ?
    case RIGCTRL_PROTOCOL_NONE :
    default:
         break;  // unknown protocol, can only send "direct" strings

    case RIGCTRL_PROTOCOL_ICOM_CI_V : // Icom's "CI-V" protocol ...
         // What's the 'command' and 'subcommand' in CI-V to *READ* this parameter ?
         switch( iUnifiedPN )
          { // Various "meter" readings via CI-V,  command 0x15 --->
            // [IC7300FM] page 19-3 : cmd 0x15 subcode 2 = Read S-meter level, etc
            case RIGCTRL_PN_SQUELCH_STATUS     : iCmd=0x15; iSubCmd=0x01; pszComment = "read Squelch"; break;
            case RIGCTRL_PN_S_METER_LEVEL_DB   : iCmd=0x15; iSubCmd=0x02; pszComment = "read SMeter";  break;
            case RIGCTRL_PN_POWER_METER_PERCENT: iCmd=0x15; iSubCmd=0x11; pszComment = "read RF Pwr";  break;
            case RIGCTRL_PN_SWR_METER_VALUE    : iCmd=0x15; iSubCmd=0x12; pszComment = "read SWR";     break;
            case RIGCTRL_PN_ALC_METER_LEVEL    : iCmd=0x15; iSubCmd=0x13; pszComment = "read ALC";     break;
            case RIGCTRL_PN_COMP_METER_LEVEL_DB: iCmd=0x15; iSubCmd=0x14; pszComment = "read Compr";   break;
            case RIGCTRL_PN_SUPPLY_VOLTAGE_mV  : iCmd=0x15; iSubCmd=0x15; pszComment = "read V_dc";    break;
            case RIGCTRL_PN_DRAIN_CURRENT_mA   : iCmd=0x15; iSubCmd=0x16; pszComment = "read I_drain"; break;
            default:  // anything else isn't supported here yet ... UNFORTUNATELY, INCLUDING THE PA TEMPERATURE
               break; // leave iCmd negative to ignore the read-request
          } // end switch( iUnifiedPN )
         if( iCmd>=0 )
          { if( RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                     RIGCTRL_MSGTYPE_OTHER | RIGCTRL_MSGTYPE_FLAG_PERIODIC_POLL, // message type (category) for the traffic log
                     pszComment, // short comment for the traffic log
                     iCmd, iSubCmd, -1, -1 ) )
             { // If all works as planned, the transceiver responds with the same 'cmd' byte:
               pRC->i32ExpectedResponse_CmdAndSubcode = RCTL_COMBINE_CMD_AND_SUB(iCmd,iSubCmd); // e.g. cmd=0x15 subcmd=0x12 -> i32ExpectedResponse_CmdAndSubcode = 0x1512....
               pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms; // now waiting for the radio to respond after asking for yet another parameter..
               // In SL's CAT traffic monitor, the full 'data exchange' looked like this ...
               // > 20:57:48.4 TX 007 FE FE 94 E0 15 02 FD        (other) ; read param
               // > 20:57:48.5 RX 007 FE FE 94 E0 15 02 FD        (echo)
               // > 20:57:48.5 RX 009 FE FE E0 94 15 02 01 23 FD  (other) ; SMeterLevel:123
               // ... and the data arrived a bit later in RigCtrl_ParseCIV(),
               //     which scaled the result ('123' in BCD) into 'dB over S0'
               //     and updated pRC->iSMeterLevel_dB, from where Spectrum Lab
               //     polled it, and distributed it to OpenWebRX-alike clients.
             }
            else // failed to send the "read"-command...
             { // better luck next time !
             }
          }
         break;  // end case <ICOM CI-V protocol>

   } // end switch( pRC->iRadioCtrlProtocol )

} // end RigCtrl_SendReadCommandFromFIFO()


//--------------------------------------------------------------------------
void RigCtrl_UpdateBasebandFrequencyConversionParams( T_RigCtrlInstance *pRC )  // -> DemodBasebandFreqOffset, ..Factor
  // Called, for example, from Spectrum Lab when drawing the filter passband
  // into the 'Remote Spectrum Display', received from IC-7300 and similar rigs.
  //    This function must know all the details about the relation between
  //    VFO frequency (aka 'dial frequency' seen on the radio's display)
  //    and the audio frequency.
  //  For USB, this is simple : Audio-frequency = Radio-Frequency - VFO-frequency;
  //                            thus [out] pRC->dblDemodBasebandFreqOffset = 0.0
  //                             and [out] pRC->dblDemodBasebandFreqFactor = +1.0
  //  For CW, it's not,
  //    because (example) : (a) Audio-frequency = Radio-Frequency - VFO-frequency + CW-pitch
  //                   or   (b) Audio-frequency = VFO-frequency - Radio-Frequency + CW-pitch
  //    depending on the radio receiving CW in "narrow USB" or "narrow LSB".
  //
  //
{
  double dblCwPitch_Hz;
  if( (pRC->iCWPitch_Hz != 0) && (pRC->iCWPitch_Hz != RIGCTL_NOVALUE_INT ) )
   { dblCwPitch_Hz = pRC->iCWPitch_Hz;
   }
  else // CW pitch (audio center frequency + CW sidetone) hasn't been read yet
   { dblCwPitch_Hz = 650.0; // author's favourite (when adjustable..)
   }

  switch( pRC->iOpMode & RIGCTRL_OPMODE_MASK_TO_STRIP_FLAGS )
   {
     case RIGCTRL_OPMODE_CW : // here, the "demodulator baseband frequency offset"
        if( pRC->iOpMode & RIGCTRL_OPMODE_REVERSE )
         { // In most rigs, "CW-REVERSE" is actually USB (no frequency inversion)
           // With the VFO tuned tuned to 3560 kHz, a signal on 3560.1 Hz
           // will be 100 Hz *above* the configurable 'CW pitch' frequency, thus:
           pRC->dblDemodBasebandFreqOffset = dblCwPitch_Hz;
           pRC->dblDemodBasebandFreqFactor = 1.0;
         }
        else // In most rigs, "CW" is actually LSB (with frequency inversion)
         { pRC->dblDemodBasebandFreqOffset = dblCwPitch_Hz;
           pRC->dblDemodBasebandFreqFactor = -1.0;
         }
        break; // end case RIGCTRL_OPMODE_CW
     default:  // for any other mode, assume "0 Hz baseband" == "VFO frequency",
               // without sideband inversion :
        pRC->dblDemodBasebandFreqOffset = 0.0;
        pRC->dblDemodBasebandFreqFactor = 1.0;
        break;
   } // end switch( pRC->iOpMode & RIGCTRL_OPMODE_MASK_TO_STRIP_FLAGS )

} // end RigCtrl_UpdateBasebandFrequencyConversionParams()


//--------------------------------------------------------------------------
double RigCtrl_BasebandToRadioFrequency( T_RigCtrlInstance *pRC, double dblBasebandFreq)
{
  RigCtrl_UpdateBasebandFrequencyConversionParams( pRC );
  return pRC->dblDemodBasebandFreqOffset // e.g. -650 Hz in "USB CW" with a 650 Hz 'CW pitch' !
       + pRC->dblDemodBasebandFreqFactor * dblBasebandFreq // e.g. +1.0 for "USB", -1.0 for "LSB".
       + pRC->dblVfoFrequency; // plus e.g. 3.560 MHz "VFO" frequency aka "dialing frequency"

} // end RigCtrl_BasebandToRadioFrequency()


//--------------------------------------------------------------------------
BOOL  RigCtrl_SetOperatingMode_Internal( T_RigCtrlInstance *pRC, int iOpMode )
  // Sends the new operating mode (e.g. RIGCTRL_OPMODE_USB) to the radio.
  // As most functions, returns TRUE when successful, otherwise FALSE.
  // Returns BEFORE the remote rig has actually 'digested' the new setting !
  // (thus returning TRUE only means "successfully SENT" but not necessarily
  //  "new setting was accepted"). Also applies to many other 'set' functions.
{
  int  iMode, iFilter;
  BOOL fResult = FALSE;
  char sz80Comment[84];
  char *pszDest=sz80Comment, *pszEndstop=sz80Comment+80;

  SL_AppendString( &pszDest, pszEndstop, "set mode and filter: " );

  switch( pRC->iRadioCtrlProtocol )
   {  // which language does the radio speak ?
     case RIGCTRL_PROTOCOL_NONE :
     default:
        break;  // unknown protocol, can only send "direct" strings

     case RIGCTRL_PROTOCOL_ICOM_CI_V : // Icom's "CI-V" protocol ...
        // Use CI-V command 0x06 because we want a response ("OK" or "NoGo"), and ..
        // > This command is understood by practically all Icom radios.
        iFilter = 0x01;  // Icom's filter codes : 1="wide", 2="medium", 3="narrow"
        // From DF4OR [DF4OR_CIV] :
        // > Due to historical reasons the filter settings are somewhat confusing.
        // > The original CI-V protocol initially had only two settings
        // > for filter width: normal and narrow. When newer rigs came out,
        // > supporting three settings (wide, normal, narrow) the codes
        // > for the filter width had to be extended.
        // > This leads to some confusion about which value sets which filter,
        // > further aggravated by erroneous documentation (e.g. IC-R75).
        // > And when looking at the DSP-filter rigs like IC-756Pro ff.,
        // > where the user can define and arrange each of the three filters
        // > individually, the settings of 'wide' or 'narrow' become meaningless.
        // > Here a filter value of 0x01 corresponds to filter 1,
        // > value 0x02 to filter 2 and 0x03 to filter 3,
        // > regardless of whether filter 1 is acutally wider than filter 2 or 3.
        // > Formerly only 0x01 and 0x02 were used for wide and narrow,
        // > medium/normal was introduced later and the data rearranged.
        // > So it can happen that 'narrow' on older rigs is programmed using 0x02,
        // > on other, newer rigs as 0x03.
        // DL4YHF: In the DEFAULT settings of an IC-7300, "Fil3" is the NARROWEST.
        switch( iOpMode & RIGCTRL_OPMODE_MASK_TO_STRIP_FLAGS )  // translate RigControl's "operating mode" into Icom's
         {
           case RIGCTRL_OPMODE_CW :
               iMode = 0x03;  // CI-V : 3 = "CW". Hopefully on ALL radios.
               SL_AppendString( &pszDest, pszEndstop, "CW" );
               if( iOpMode & RIGCTRL_OPMODE_REVERSE )
                { iMode = 0x07; // CI-V : 0x07 = "CW-R". At least on IC-7300.
                  SL_AppendChar( &pszDest, pszEndstop, 'R' );
                }
               if( iOpMode & RIGCTRL_OPMODE_NARROW )
                { iFilter = 0x02;  // some radios: 1="wide", 2="medium", 3="narrow"
                  SL_AppendChar( &pszDest, pszEndstop, 'N' );
                }
               if( iOpMode & RIGCTRL_OPMODE_VERY_NARROW )
                { iFilter = 0x03;  // some radios: 1="wide", 2="medium", 3="narrow"
                  SL_AppendChar( &pszDest, pszEndstop, 'N' );
                }
               break;
           case RIGCTRL_OPMODE_LSB  :
               iMode = 0x00;  // CI-V : 0 = "LSB"
               SL_AppendString( &pszDest, pszEndstop, "LSB" );
               break;
           case RIGCTRL_OPMODE_USB  :
               iMode = 0x01;  // CI-V : 1 = "USB"
               SL_AppendString( &pszDest, pszEndstop, "USB" );
               break;
           case RIGCTRL_OPMODE_AM   :
               iMode = 0x02;  // CI-V : 2 = "AM"
               SL_AppendString( &pszDest, pszEndstop, "AM" );
               break;
           case RIGCTRL_OPMODE_FM   :   // Fasten seat belt and BE PREPARED FOR TROUBLE !
               iMode = 0x05;  // CI-V : 5 = "FM" (at least for IC-9700)
               SL_AppendString( &pszDest, pszEndstop, "FM" );
               break; // end case RIGCTRL_OPMODE_FM
           case RIGCTRL_OPMODE_FM_WIDE :
               iMode = 0x06;     // CI-V : 6 = "FM wide" (at least for IC-706)
               SL_AppendString( &pszDest, pszEndstop, "FM-Wide" );
               break;
           case RIGCTRL_OPMODE_RTTY :
               iMode = 0x04;    // CI-V : 4 = "RTTY". But on IC-910, 9 is FM. Holy shit.
               // (these annoying subtleties are one reason to use HEX VALUES here,
               //  not macro constants which everyone would have to look up permanently..)
               SL_AppendString( &pszDest, pszEndstop, "RTTY" );
               if( iOpMode & RIGCTRL_OPMODE_REVERSE )
                { iMode = 0x08; // CI-V : 0x08 = "RTTY-R". At least on IC-7300..
                  //  not listed for older radios, in Icom's funny-titled
                  //  "CI-V Reference Manual -.-.-ver 3.2.-.-. -2002-"
                }
               break;
           case RIGCTRL_OPMODE_PSK  :
               iMode = 0x12;    // CI-V : 0x12 = "PSK". At least on IC-7300..
               SL_AppendString( &pszDest, pszEndstop, "PSK" );
               if( iOpMode & RIGCTRL_OPMODE_REVERSE )
                { iMode = 0x13; // CI-V : 0x13 = "PSK-R". At least on IC-7300
                }
               break;
           default:
               iMode = 0x03;  // CI-V : "CW"
               SL_AppendString( &pszDest, pszEndstop, "CW" );
               break;
          } // end switch( pRC->iOpMode )

         if( iMode >= 0 )  // successfully translated at least the MODE..
          { fResult = RigCtrl_SendShortCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                    RIGCTRL_MSGTYPE_OP_MODE,  // message type for traffic log
                       sz80Comment, // comment for traffic log
                       0x06,  // iCmd: CI-V command "Set mode and filter"
                         -1,  // iSubCmd and iDataByte1 : Omit !
                       iMode, // iDataByte1 : here "mode".
                    iFilter); // iDataByte2 : here "filter" (optionally)
            pRC->i32ExpectedResponse_CmdAndSubcode = RCTL_COMBINE_CMD_ONLY(0x06); // the radio's CI-V response "Set operating mode" doesn't contain command or subcode, so REMEMBER what has been "set" this way
            pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms; // now waiting for the "OK"/"NOT OK" response for "Set Operating Mode"
          }
         break;  // end case <ICOM CI-V protocol>

   } // end switch( pRC->iRadioCtrlProtocol )

  return fResult;

} // end RigCtrl_SetOperatingMode_Internal()

//--------------------------------------------------------------------------
BOOL RigCtrl_SetOperatingMode( T_RigCtrlInstance *pRC, int iOpMode ) /* API */
  // Sends the new operating mode (e.g. RIGCTRL_OPMODE_USB) to the radio.
  //  [in] iOpMode : RIGCTRL_OPMODE_CW, RIGCTRL_OPMODE_LSB, RIGCTRL_OPMODE_USB,
  //                 RIGCTRL_OPMODE_AM, RIGCTRL_OPMODE_FM,  RIGCTRL_OPMODE_FM_WIDE
  //                   ("basic modes" below RIGCTRL_OPMODE_MASK_TO_STRIP_FLAGS),
  //                    OPTIONALLY bitwise combined with flags like
  //                 RIGCTRL_OPMODE_REVERSE (for "CW Reverse"),
  //                 RIGCTRL_OPMODE_NARROW, RIGCTRL_OPMODE_VERY_NARROW
  //                 (the latter for compatibility with older rigs,
  //                  which don't support reading the "filter bandwidth").
  //
  // As most functions, returns TRUE when successful, otherwise FALSE.
  // Returns BEFORE the remote rig has actually 'digested' the new setting !
  // (thus returning TRUE only means "successfully SENT" but not necessarily
  //  "new setting was accepted"). Also applies to many other 'set' functions.
{
  // This function may be called by the application at any time.
  // The actual transmission may have to be postponed when still busy
  // from an 'uninterruptable' sequence of commands in RigCtrl_Handler().
  // In any case, immediately copy the new mode ...
  if( iOpMode != pRC->iOpMode )
   { pRC->iOpMode = iOpMode;
     // ... and queue this up for a later transmission to the remote server, e.g. via CwNet_OnPoll()
     RigCtrl_AddParameterNumberToTxQueue(pRC, RIGCTRL_PN_OP_MODE ); // here: called from RigCtrl_SetOperatingMode()
   }
  if( pRC->iResponseCountdown_ms > 0 ) // "do not send anything at the moment" because still waiting for the response for a PREVIOUSLY SENT "Set"-command !
   {  pRC->iPostponedSetMessages |= (1<<RIGCTRL_MSGTYPE_OP_MODE);
      pRC->iPostponedOpMode = iOpMode;
      return FALSE;  // cannot send AT THE MOMENT (still busy from another transaction)
   }
  else // no need to postpone this "Set"-command :
   {  pRC->iPostponedSetMessages &= ~(1<<RIGCTRL_MSGTYPE_OP_MODE);
      return RigCtrl_SetOperatingMode_Internal( pRC, iOpMode );
   }

} // end RigCtrl_SetOperatingMode()

//--------------------------------------------------------------------------
const char* RigCtrl_OperatingModeToString(int iOpMode)
  // First used in Spectrum Lab for interpreter function "rct.mode_string" .
  // Later also used in the Remote CW Keyer / RigCtrl_FreqMemEntryToString().
  // Inverse functions (string builder and parser) : RigCtrl_OperatingModeToString(), RigCtrl_StringToOperatingMode() .
{ static char sz10Result[16];
  char *pszBasicMode;

  // If it's one of the common modes, don't use the evil static buffer:
  switch( iOpMode & 15 )
   {
     case RIGCTRL_OPMODE_CW : pszBasicMode = "CW";
          if( iOpMode & RIGCTRL_OPMODE_NARROW )
           { pszBasicMode = "CWN";
           }
          if( iOpMode & RIGCTRL_OPMODE_VERY_NARROW )
           { pszBasicMode = "CWNN";
           }
          break;
     case RIGCTRL_OPMODE_LSB     : pszBasicMode = "LSB"; break;
     case RIGCTRL_OPMODE_USB     : pszBasicMode = "USB"; break;
     case RIGCTRL_OPMODE_AM      : pszBasicMode = "AM";  break;
     case RIGCTRL_OPMODE_FM      : pszBasicMode = "FM";  break;
     case RIGCTRL_OPMODE_FM_WIDE : pszBasicMode = "FMW"; break;
     case RIGCTRL_OPMODE_RTTY    : pszBasicMode = "RTTY"; break;
     case RIGCTRL_OPMODE_PSK     : pszBasicMode = "PSK"; break;
     default: pszBasicMode = "???"; break;
   } // end switch( pRC->iOpMode )
  if( iOpMode & RIGCTRL_OPMODE_REVERSE )
   { strcpy( sz10Result, pszBasicMode );
     strcat( sz10Result, "R" );
     return  sz10Result;
   }
  else
   { return pszBasicMode;
   }

} // end RigCtrl_OperatingModeToString()

//--------------------------------------------------------------------------
int RigCtrl_StringToOperatingMode(const char **ppszSource)
  // Added 2024-11-21 to exchange e.g. "Band Stacking Registers" (etc) in a
  // non-Icom-specific format between client and server (via TCP/IP).
  // Inverse functions (string builder and parser) : RigCtrl_OperatingModeToString(), RigCtrl_StringToOperatingMode() .
{ int iOpMode = RIGCTRL_OPMODE_UNKNOWN;
  const char *pszOldSource = *ppszSource;

  // First check if the pszSource begins with a known "basic mode" like "CW".
  // After that, check for suffixes like "R" (Reverse), "N" (Narrow), "NN" (Very Narrow).
  if( SL_SkipString( ppszSource, "CW" ) )
   { iOpMode = RIGCTRL_OPMODE_CW;
   }
  else if( SL_SkipString( ppszSource, "LSB" ) )
   { iOpMode = RIGCTRL_OPMODE_LSB;
   }
  else if( SL_SkipString( ppszSource, "USB" ) )
   { iOpMode = RIGCTRL_OPMODE_USB;
   }
  else if( SL_SkipString( ppszSource, "AM" ) )
   { iOpMode = RIGCTRL_OPMODE_AM;
   }
  else if( SL_SkipString( ppszSource, "FMW" ) )
   { iOpMode = RIGCTRL_OPMODE_FM_WIDE;
   }
  else if( SL_SkipString( ppszSource, "FM" ) )
   { iOpMode = RIGCTRL_OPMODE_FM;
   }
  else if( SL_SkipString( ppszSource, "RTTY" ) )
   { iOpMode = RIGCTRL_OPMODE_RTTY;
   }
  else if( SL_SkipString( ppszSource, "PSK" ) )
   { iOpMode = RIGCTRL_OPMODE_FM;
   }
  else if( SL_SkipString( ppszSource, "???" ) ) // <- yes, that's a TOKEN !
   { iOpMode = RIGCTRL_OPMODE_UNKNOWN;
   }
  // Now for the "suffix", ignoring the fact that certain combinations
  // are useless (e.g. "USBNNR" = Upper Side Band, Very Narrow, REVERSE) :
  if( SL_SkipString( ppszSource, "NN" ) )
   { iOpMode |= RIGCTRL_OPMODE_VERY_NARROW;
   }
  else if( SL_SkipString( ppszSource, "N" ) )
   { iOpMode |= RIGCTRL_OPMODE_NARROW;
   }
  if( SL_SkipString( ppszSource, "R" ) )
   { iOpMode |= RIGCTRL_OPMODE_REVERSE;
   }
  // Just in case the sender had the "glorious" idea to use a NUMERIC value:
  if( (*ppszSource==pszOldSource )&& (SL_IsDigit( **ppszSource ) ) )
   { iOpMode = SL_ParseInteger(ppszSource);
   }
  return iOpMode;
} // end RigCtrl_StringToOperatingMode()


//--------------------------------------------------------------------------
BOOL RigCtrl_SetScopeRefLevel_dB( T_RigCtrlInstance *pRC, double dblRefLevel_dB )
  // Sends a new 'spectrum reference level' to a modern Icom radio.
  // This is the parameter that you need to modify whenever switching
  // to a new 'span' (displayed bandwidth) in IC-7300, IC-9700, IC-705, etc,
  // which can be VERY annoying because the transceiver doesn't store
  // the reference levels for each of the supported 'spans'
  // (2.5 kHz, 5 kHz, 10 kHz, 25 kHz, 50 kHz, 100 kHz, 250 kHz, 500 kHz) !
  // In Icom's "IC-9700 CI-V REFERENCE GUIDE", published in June 2019,
  //      (locally saved as IC-9700_ENG_CI-V_1.pdf),
  // this is one of an awful lot of "sub commands" (sub parameters)
  // for command 27, "Send/read scope related parameters" [page 24],
  //  here: sub-command 19 = "Scope Reference level settings",
  //                         ranging from -20.0 to +20.0 dB in 0.5 dB steps.
{
  BOOL fResult;
  BYTE bBCD[2];        // 4-digit BCD in TWO BYTES
  BYTE bSign = 0x00;

  // Icom's "Spectrum REF level" ranges from -20 to +20 dB, so limit to that range:
  RigCtrl_LimitDouble( &dblRefLevel_dB, -20.0, 20.0 );
  pRC->dblScopeRefLevel_dB = dblRefLevel_dB;

  if( dblRefLevel_dB < 0 ) // in CI-V cmd 0x29.0x19, the SIGN *follows* after the BCD string :
   { bSign = 0x01;
     dblRefLevel_dB = -dblRefLevel_dB;
   }
  RigCtrl_IntToBCD_MSDigitFirst( (long)(dblRefLevel_dB*100.0/*!*/), bBCD, 4/*nDigits*/ );
  // ,--------------'
  // '--> MOST SIGNIFICANT DIGIT FIRST ! Example:
  //  dblRefLevel_dB = 12.50 -> bBCD[0] = 0x12,  bBCD[1] = 0x50 (!) .

  fResult = RigCtrl_SendMidSizeCIVMessage( pRC, RIGCTRL_PORT_RADIO,
                 RIGCTRL_MSGTYPE_C_SPECTRUM,  // message type for traffic log
                       "set scope REF level", // comment for traffic log
                       0x27,  // iCmd: CI-V command "Scope settings"
                       0x19,  // iSubCmd: 0x19 = "Scope Reference level settings"
                       0x00,  // iDataByte1 : 0x00 = "main scope", 0x01 = "sub scope"
                    bBCD[0],  // iDataByte2 : here the 10 dB digit and 1 dB digit (BCD)
                    bBCD[1],  // iDataByte3 : here the 0.1 dB digit and 0.01 dB digit (BCD)
                     bSign);  // iDataByte4 : 0x00 for "plus", 0x01 for "minus"
  // pRC->i32ExpectedResponse_CmdAndSubcode = ?; // expected CI-V response ?
  // Seen in the Remote CW Keyer's built-in CI-V message display:
  // > 19 21:12:39.4 TX 00B FE FE 00 00 27 19 00 4F 32 32 FD ; set scope REF level
  // > 20 21:12:39.5 RX 006 FE FE 00 A2 FA FD                ; NotOK for ...

  return fResult;


} // end RigCtrl_SetScopeRefLevel_dB()

//--------------------------------------------------------------------------
BOOL RigCtrl_SetScopeSpan_Hz( T_RigCtrlInstance *pRC, double dblScopeSpan_Hz )
  // Sends a new 'spectrum scope span' (diplayed bandwidth in Hertz) to a modern Icom radio.
  // Radio generations like IC-7300 and IC-9700 only supported the following "SPAN" settings:
  //      "+/-2.5 kHz"      "+/-5.0 kHz"    "+/-10.0 kHz"
  //      "+/-25.0 kHz"     "+/-50.0 kHz"   "+/-100.0 kHz"
  //      "+/-250.0 kHz"    "+/-500.0 kHz" (<- that's the highest setting, 500 kHz "Span" = 1 MHz displayed bandwidth.
{
  // This API function may be called by the application at any time, but it may
  // have to be postponed (internally) when still busy from an 'uninterruptable'
  //      sequence of commands in RigCtrl_Handler().
  // In ANY case, immediately copy the new "wanted" setting :
  pRC->dblScopeSpan_Hz = dblScopeSpan_Hz;
  if( pRC->iResponseCountdown_ms > 0 ) // still waiting for a response -> "do not send anything at the moment"..
   {  pRC->iPostponedSetMessages |= (1<<RIGCTRL_MSGTYPE_SCOPE_SPAN); // remember to send pRC->dblScopeSpan_Hz later
      return TRUE;  // cannot send AT THE MOMENT, but data are (almost) "on the way"
   }
  else // no need to postpone the "Set SCOPE SPAN"-command :
   {  pRC->iPostponedSetMessages &= ~(1<<RIGCTRL_MSGTYPE_SCOPE_SPAN);
      return RigCtrl_SetScopeSpan_Internal( pRC );
   }
} // end RigCtrl_SetScopeSpan_Hz()

//--------------------------------------------------------------------------
static BOOL RigCtrl_SetScopeSpan_Internal( T_RigCtrlInstance *pRC )
  // Internal subroutine for RigCtrl_SetScopeSpan_Hz() .. details there
{
  BYTE b16TxBuffer[16];
  BYTE *pbBufPtr = b16TxBuffer;
  int  iMsgLength;
  BOOL fResult = FALSE;

  switch( pRC->iRadioCtrlProtocol )
   {
     case RIGCTRL_PROTOCOL_ICOM_CI_V : // Icom's "CI-V" protocol ...
        // About Icom's "Scope span settings", from the IC-7610 CI-V Reference Guide (A7380-7EX-2) page 14:
        // > Scope span settings (in the Center mode and SCROLL-C mode Scope)
        // > Command: 27 15
        //   Message-Byte[0  1  2  3  4  5  6  7  8  9 10 11 12]
        //               FE FE E0 94 27 15 00 00 25 00 00 00 FD (scope) ; span=2500 Hz
        //                           |  |  |_______________|
        //                           |  |  :: '- 12-digit BCD-encoded frequency in HERZ
        //                        Cmd Sub  ::    (only A FEW values are allowed)
        //                                 ::.... 1-Hz-Digit (!), etc.
        //                                 :
        //                                 10-Hz-Digit in the upper 4 bits of this byte
        *pbBufPtr++ = 0xFE;  // fixed preamble code for CI-V protocol (term used by Icom .. NOT "Begin of Message" ! )
        *pbBufPtr++ = 0xFE;  // fixed preamble code for CI-V protocol
        *pbBufPtr++ = (BYTE)pRC->iRadioDeviceAddr;  // "to" transceiver's address
        *pbBufPtr++ = (BYTE)pRC->iRadioMasterAddr;  // "from" controller's address
        *pbBufPtr++ = 0x27;  // 0x27 = CI-V command "Scope Settings"
        *pbBufPtr++ = 0x15;  // 0x15 = "Scope span settings"
        *pbBufPtr++ = 0x00;  // 0x00 for "MAIN", 0x01 would be "SUB" if the radio was an IC-7610 (not used here)
        RigCtrl_IntToBCD_CIV( (long)pRC->dblScopeSpan_Hz, pbBufPtr/*pbDest*/, 10/*nDigits*/ );
        pbBufPtr   += 5;     // RigCtrl_IntToBCD_CIV() has appended 5 bytes for 10 digits
        *pbBufPtr++ = 0xFD;  // 0xFD = CI-V code for "End Of Message"
        fResult = RigCtrl_SendAndLogMessage( pRC, RIGCTRL_PORT_RADIO,
                 b16TxBuffer/*pbMessage*/, pbBufPtr-b16TxBuffer/*iMsgLength*/,
                 FALSE/*fWaitForCompletion?*/,
                 RIGCTRL_MSGTYPE_SCOPE_SPAN,  // message type for the traffic log
                 "set scope span" ); // comment for traffic log
        pRC->i32ExpectedResponse_CmdAndSubcode = RCTL_COMBINE_CMD_AND_SUB(0x27,0x15);
        pRC->iResponseCountdown_ms = pRC->iMaxResponseDelay_ms; // now waiting for the "OK"/"NOT OK" response for "Set Scope Span"
        break;
     default: // command not supported (with the current pRC->iRadioCtrlProtocol)
        break;
   } // end switch( pRC->iRadioCtrlProtocol )

  return fResult;


} // end RigCtrl_SetScopeSpan_Internal()


//--------------------------------------------------------------------------
BYTE RigCtrl_IntToBCD2( int iValue ) // converts an integer between 0 and 99 into TWO-DIGIT BCD (in a single byte)
  // This is often used in Icom's CI-V protol as an additional byte
  // after the command and sub-command, e.g. when reading 'array-like' data .
{
  if( iValue < 0 )
   {  iValue = -iValue;
   }
  return (BYTE)( (((iValue / 10) % 10) << 4) | (iValue % 10) );
  //             |__________tens___________|   |___ones____|
}


//--------------------------------------------------------------------------
void RigCtrl_IntToBCD_CIV( long i32Value, BYTE *pbDest, int nDigits )
  // More or less inverse to RigCtrl_ParseBCD_CIV_LSByteFirst(). Details there.
  // Hopefully the number of digits is always EVEN...
{
  int i, iDigit;
  for(i=0; i<nDigits; ++i) // begin with the LEAST SIGNIFICANT digit at i=0 (*)
   { iDigit = i32Value % 10;
     i32Value /= 10;
     if( !(i & 1) )       // 1st, 3rd, 5th digit ? calculate LOW nibble
      { *pbDest = (BYTE)iDigit;
      }
     else                 // 2nd, 4th, 6th digit ? emit the next byte
      { *pbDest |= ((BYTE)iDigit << 4);
        ++pbDest;
      }
     // (*) Unfortunately, Icom uses a bizarre mix of "most significant digit first"
     //      and "least significant digit first". For example, to set a new
     //      "Scope reference level" (command 0x27 0x19),
     //      the first byte contains TENS and ONES, the second byte 0.1 and 0.01 ...
   }
} // end RigCtrl_IntToBCD_CIV()

//--------------------------------------------------------------------------
void RigCtrl_IntToBCD_MSDigitFirst( long i32Value, BYTE *pbDest, int nDigits )
  // This is the SECOND (and incompatible) flavour of "BCD" in Icom's CI-V .
  // Used, for example, to set a new "Scope reference level" (command 0x27 0x19),
  // where for reasons not even the Icom engineers will know,
  // the MOST SIGNIFICANT DIGIT (e.g. TENS) is in bit 7..4 of the first BCD byte,
  // the next digit (e.g. the ONES) is in bit 3..0 of the first BCD byte,
  // and the LEAST SIGNIFICANT DIGIT follows last.   OH MY GOD.
  //
{
  int i, iDigit;
  int ai = (nDigits-1)/2;  // first array offset into pbDest[ai] : 0 for 1 or 2 digits, 1 for 3 or 4 digits, etc
  for(i=0; (ai>=0) && (i<nDigits); ++i) // begin with the LEAST SIGNIFICANT digit,
   { // but emit that digit into the LAST BYTE, lower nibble !
     // Example (FOUR-DIGIT BCD as in the "Scope reference level") :
     // i32Value = 1234, nDigits = 4 -> ai = (4-1) & 0x0E = 3
     // i = 0  :  least significant digit, emitted FIRST in this loop,
     //                                      into pbDest[1] bits 3..0,
     // i = 1  :  medium low  digit, emitted into pbDest[1] bits 7..4,
     // i = 2  :  medium high digit, emitted into pbDest[0] bits 3..0,
     // i = 3  :  most significant digit, "  into pbDest[0] bits 7..4 .
     // Example after the conversion: pbDest[0] = 0x12, pbDest[1] = 0x34 . 
     iDigit = i32Value % 10;
     i32Value /= 10;
     if( !(i & 1) )  // i=0,2,4, .. : set lower nibble, clear the upper nibble
      { pbDest[ai] = (BYTE)iDigit;
      }
     else            // i=1,3,5, .. : bitwise OR the upper nibble and switch to the next byte
      { pbDest[ai--] |= ((BYTE)iDigit<<4);
      }
   }
} // end RigCtrl_IntToBCD_MSDigitFirst()


//--------------------------------------------------------------------------
long RigCtrl_ParseBCD_CIV_LSByteFirst( BYTE *pbData, int nDigits )
  // Parses an integer from a BCD sequency with the *LEAST SIGNIFICANT* byte first,
  //        as OFTEN(!) used in Icom's CI-V protocol .
  // Most may actually be 10-digit fields, but padded with zeros.
  // NOT SUITABLE FOR PARSING THE 4-DIGIT BCD PARAMETER NUMBERS
  //              after command 0x17 0x5 - see RigCtrl_ParseBCD_MSDigitFirst()
  // Example:  cmd 0x28, sub-cmd 0x15,  "Scope span settings" :
  //         pbData[0]  [1]     [2]     [4]     [5]
  //         |<----->|
  //          _______________________________________
  //         |   :   |   :   |   :   |   :   |   :   |
  //         | 0 : 0 | X : X | X : X | 0 : 0 | 0 : 0 |  <-- nDigits = 10 (!)
  //         |___:___|___:___|___:___|___:___|___:___|
  //                  /|\ /|\ /|\ /|\ /|\ /|\ /|\ /|\   .
  //                   |   |   |   |   |   |   |   |
  //         (10   1)  1  100 100  10  10MHz ? 1 GHz ?
  //            Hz    kHz  Hz  kHz kHz     1 MHz ? 100 MHz ?
{
  long sum = 0;
  long factor = 1;
  int  i = 0;
  BYTE b;

  while( (nDigits--) > 0 )
    { b = pbData[i++]; // get next byte (with two digits, BCD)
      sum += factor * (long)(b & 0x0F);
      factor *= 10;
      if( (nDigits--) > 0 )
       {
         sum += factor * (long)((b>>4) & 0x0F);
         factor *= 10;
       }
    }
  return sum;

} // end RigCtrl_ParseBCD_CIV_LSByteFirst()

//--------------------------------------------------------------------------
long RigCtrl_ParseBCD_CIV_MSByteFirst( BYTE *pbData, int nDigits )
  // Parses an integer from a BCD sequency with the *MOST SIGNIFICANT* byte first,
  //        as SOMETIMES(!) used in Icom's CI-V protocol .
  //
  // Example:  cmd 0x27, sub-cmd 0x19,  "Scope reference level settings"
  //         pbData[0]  [1]     [2]     [4]
  //         |<----->|<----value---->|<-sign>| "I'm going slightly mad.." (Queen)
  //          _______ _______________ _______
  //         |   :   |   :   |   :   |       |
  //         | 0 : 0 | X : X | X : X | x : x |
  //         |___:___|___:___|___:___|_______|
  //          00=main /|\ /|\ /|\ /|\    '-----  "SIGN" as another two-digit-BCD !
  //          01=sub   |   |   |   |               00 = + (positive)
  //                  10-  1- 0.1- 0.01-           01 = - (negative)
  //                  dB- db-  dB- dB-
  //                 digit ..      digit (!)
{
  long sum = 0;
  int  i = 0;
  BYTE b;

  while( (nDigits--) > 0 )
    { b = pbData[i++]; // get next byte (with two digits, BCD)
      sum = (10*sum) + (long)((b>>4) & 0x0F);
      if( (nDigits--) > 0 )  // 2nd digit from the same source-byte ?
       { sum = (10*sum) + (long)(b & 0x0F);
       }
    }
  return sum;

} // end RigCtrl_ParseBCD_CIV_MSByteFirst()

//--------------------------------------------------------------------------
int RigCtrl_CountBytesUntilDelimiter( BYTE *pbData, BYTE *pbEndstop,
                BYTE bDelimiterA, // [in] e.g. 0x2D for Icom's "Separator"
                BYTE bDelimiterB) // [in] e.g. 0xFD for Icom's "Postamble"
  // Subroutine for a few CI-V command parsers, e.g. RigCtrl_ParseCIVFreqRange()
{
  BYTE *pbStart=pbData;
  while( (pbData < pbEndstop ) && (*pbData!=bDelimiterA) && (*pbData!=bDelimiterB) )
   { ++pbData;
   }
  return pbData-pbStart; // "C" can be beautifully simple if you use pointers carefully :)
} // end RigCtrl_CountBytesUntilDelimiter()

//--------------------------------------------------------------------------
BOOL RigCtrl_ParseCIVFreqRange( // Used e.g. by Icom's cmd 0x1E, subCmd 0x01 .
      BYTE *pbData, int nBytes, // [in] BCD string a la Icom CI-V .
                                //      The leading "edge number" has already
                                //      been parsed (and skipped) by the caller.
      T_RigCtrlFrequencyRange *pFreqRange ) // [out] pFreqRange->dblFmin_Hz, dblFmax_Hz.
{
  BYTE *pbEndstop = pbData + nBytes;
  // pbData on ENTRY --------------,   (Example from an IC-9700. Don't assume
  //                               |    anything about the NUMBER OF DIGITS..
  //                              \|/   things will be different in an IC-905)
  // pbMessage[0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18]
  // > RX 013 FE FE 00 A2 1E 01 01 00 00 00 44 01 2D 00 00 00 46 01 FD ; 144 MHz .. 146 MHz
  //       |  |___| |addr |  |  |  |____________| |  |____________|  |
  //       | Preamble     |  | "Edge"   |        Sep.    |          Postamble
  //   iMsgLength=19   cmd sub  |     StartFreq   |    EndFreq       |
  //                               |<--iLen=5-->| |  |<--iLen=5-->|  '----,
  //                                              '---------------,       |
  int iLen = RigCtrl_CountBytesUntilDelimiter( pbData, pbEndstop, 0x2D, 0xFD );
  if( iLen>1 ) // at least ONE byte BEFORE the separator (or postamble):
   { pFreqRange->dblFmin_Hz = RigCtrl_ParseBCD_CIV_LSByteFirst( pbData, 2*iLen/*->nDigits*/ );
     pbData += iLen; // <- should point to the 'Separator' (0x2D) now, not to the postamble (0xFD)
     if( (pbData<pbEndstop) && (*pbData==0x2D) ) // Got the expected separator ?
      { ++pbData;      // skip the separator (0x2D) between "Lower edge" and "Higher edge"
        iLen = RigCtrl_CountBytesUntilDelimiter( pbData, pbEndstop, 0x2D, 0xFD );
        if( iLen>1 ) // at least ONE byte with the "Higher edge" (max frequency) ?
         { pFreqRange->dblFmax_Hz = RigCtrl_ParseBCD_CIV_LSByteFirst( pbData, 2*iLen/*->nDigits*/ );
           return TRUE;   // successfully parsed a frequency range a la CI-V (e.g. "Band edge frequencies")
         }
      }
   }
  return FALSE;
} // end RigCtrl_ParseCIVFreqRange()

//--------------------------------------------------------------------------
int RigCtrl_IcomToRigCtrlOpmode( BYTE bIcomsOpMode )
{ // Don't expect the numeric value for "PSK" and "RTTY" to be the same in all rigs !
  switch( bIcomsOpMode )
   { case 0x00 : return RIGCTRL_OPMODE_LSB;
     case 0x01 : return RIGCTRL_OPMODE_USB;
     case 0x02 : return RIGCTRL_OPMODE_AM;
     case 0x03 : return RIGCTRL_OPMODE_CW;
     case 0x04 : return RIGCTRL_OPMODE_RTTY;
     case 0x05 : return RIGCTRL_OPMODE_FM;
     case 0x06 : return RIGCTRL_OPMODE_FM_WIDE; // used by IC-706 and possibly other VHF-capable radios
     case 0x07 : return RIGCTRL_OPMODE_CW   | RIGCTRL_OPMODE_REVERSE;
     case 0x08 : return RIGCTRL_OPMODE_RTTY | RIGCTRL_OPMODE_REVERSE; // <- seems to be the highest supported mode in an IC-7300
     case 0x09 : return RIGCTRL_OPMODE_UNKNOWN;
     case 0x12 : return RIGCTRL_OPMODE_PSK; // questinable.. snapped up somewhere
     case 0x13 : return RIGCTRL_OPMODE_PSK  | RIGCTRL_OPMODE_REVERSE; // questinable.. snapped up somewhere
     case 0x17 : return RIGCTRL_OPMODE_DIGITAL_VOICE; // seen in page 13 ("Digital Voice", presumably D-Star, not DMR)
     case 0x22 : return RIGCTRL_OPMODE_DIGITAL_DATA; // seen in page 13 (no idea what "DD" stands for yet.. 128 kbp "Digital Data" on 23 cm ?)
     default: return RIGCTRL_OPMODE_UNKNOWN;
   }
} // end RigCtrl_IcomToRigCtrlOpmode()


//--------------------------------------------------------------------------
void RigCtrl_InitStruct_FreqMemEntry( T_RigCtrlFreqMemEntry *pFreqMem )
{
  memset( pFreqMem, 0, sizeof(T_RigCtrlFreqMemEntry) );
  // Make sure the FLOATING POINT members are valid numbers:
  pFreqMem->RxTx[0].dblOperatingFreq_Hz = pFreqMem->RxTx[1].dblOperatingFreq_Hz = 0.0;
} // end RigCtrl_InitStruct_FreqMemEntry()



//--------------------------------------------------------------------------
BOOL RigCtrl_ParseMemoryContent( // Used for by Icom's commands
                  //  0x1A sub 0x00 ("memory contents") and
                  //  0x1A sub 0x01 ("band stacking register contents").
      BYTE *pbData, int nBytes, // [in] BCD string a la Icom CI-V .
      //     '---> Begins with the "Operating frequency setting" (5),
      //           because that's the first field common of what
      //           Icom calls "Memory Content",
      //  i.e. the response for commands 0x1A 0x00 and 0x1A 0x01 .
      // The IC-7300 "Full Manual" v6 "Section 19 CONTROL COMMAND" leaves a lot
      // of doubt about the format of the "band stacking register contents".
      // Only the IC-9700 "CI-V REFERENCE GUIDE" v4 (A7508-3EX-4) explained:
      //  > NOTE: When sending the contents, the codes, such as
      //  >       operating frequency and operating mode*, should be
      //  >       added after the frequency band code and the register
      //  >       code, as shown below.
      //  >  * See (5) to (51) on 'Memory content setting.' (p. 14)
      //       "Should be added" ? ? Come on Icom, is this a "to do list"
      //                             for your firmware developers ?
      //       "(5)" to "(51)" (decimal numbers in full circles) seem to be
      //       BYTE OFFSETS in the message, where "(5)" is the operating frequency.
      // Unfortunately, even that is already different between IC-7300 and IC-9700 !
      // Fasten seat belt and read on .. the author had a lot of "fun" with this.
      T_RigCtrlFreqMemEntry *pFreqMem ) // [out] an AWFUL lot of members in Icom's "Memory Content"
{
  BYTE *pbEndstop = pbData + nBytes;
  int  nDigitsForFrequency = 10; // <- at least for IC-9700 and IC-7300 (including a 1-GHz-digit)
  //
  // pbData on ENTRY -----------------,   ( Example from an IC-7300's first
  //                                  |     "Band Stacking Register",
  //                                 \|/    including request and response)
  // pbMessage[0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18]
  // > TX 009 FE FE 00 00 1A 01 01 01 FD  ; BandStackReg[1][1] ?
  // > RX 017 FE FE 00 94 1A 01 01 01 30 96 83 01 00 00 02 01 00 08 85 00 08 85 .. ; band stack mem
  //       |  |___| |addr |  |  |  |  |____________|  |  |  '--- this is where it gets bizarre and incompatible
  //       | Preamble     |  |  |  |  OperatingFreq   | Filter setting: 01..03="Fil1".."Fil3"
  //   iMsgLength=19   cmd sub  |  |  ||  ......| ||  Operating Mode: 00=LSB, etc... (see RigCtrl_IcomToRigCtrlOpmode() )
  //                            |  |  |1Hz      | |100MHz
  // "Frequency Band Code" (1)--'  |  10Hz   1MHz 1GHz
  // "Register Code"       (2)-----'  (0001839.639 MHz)
  RigCtrl_InitStruct_FreqMemEntry( pFreqMem );
  pFreqMem->RxTx[0].dblOperatingFreq_Hz = RigCtrl_ParseBCD_CIV_LSByteFirst( pbData, nDigitsForFrequency );
  pbData += (nDigitsForFrequency+1)/2; // <- should now point to the "Operating Mode" ...
  if( pbData < pbEndstop ) // more bytes following ?  Next: two-BCD-digit "Operating Mode"..
   { pFreqMem->RxTx[0].iOpMode = RigCtrl_IcomToRigCtrlOpmode( *(pbData++) );
   }
  if( pbData < pbEndstop ) // more bytes following ?  Next: two-BCD-digit "Filter Setting"..
   { // (In the IC-9700 "CI-V Reference", this field also belongs to the "Operating Mode".
     //  HERE, we only care for the "Filter Setting" if the (unified) OpMode is CW :
     if( (pFreqMem->RxTx[0].iOpMode & RIGCTRL_OPMODE_MASK_TO_STRIP_FLAGS) == RIGCTRL_OPMODE_CW )
      { pFreqMem->RxTx[0].iOpMode &= ~(RIGCTRL_OPMODE_NARROW | RIGCTRL_OPMODE_VERY_NARROW);
        switch( *pbData )
         { case 0x01:  // Icom's "Filter 1" is usually the WIDE filter -> neither "CWN" nor "CWNN"
              break;
           case 0x02:  // Icom's "Filter 2" is usually the "quite narrow" CW filter -> "CWN"
              pFreqMem->RxTx[0].iOpMode |= RIGCTRL_OPMODE_NARROW;
              break;
           case 0x03:  // Icom's "Filter 3" is usually the "very narrow" CW filter -> "CWNN"
              pFreqMem->RxTx[0].iOpMode |= RIGCTRL_OPMODE_VERY_NARROW;
              break;
           default:    // all other "Filter settings" (term used in A7508-3EX-4) are unknown
              break;
          }
      }
     ++pbData; // skip Icom's "Filter Setting" (which THEY SAY belongs to the "Operating Mode")
   }

  // This is where the compatibility between IC-7300 and IC-9700 ends !
  //     After the "Operating Mode" with "Filter Setting", the next part of a
  //    "Memory Content" in an IC-9700 is the "Data mode setting" ("(12)"),
  //                 but in an IC-7300, it's "Data mode and tone type settings".
  if( pbData < pbEndstop ) // more bytes following ?  Next: two-BCD-digit "Data mode setting"..
   { // .. which, in the IC-7300, they call "Data mode and tone type settings" - buaah !
     switch( *pbData & 0xF0 )
      { case 0x00: // upper nibble (tens) set to ZERO ? IC-7300 : NO "data mode" .
             pFreqMem->RxTx[0].fDataMode = FALSE;
             break;
        case 0x10: // upper nibble (tens) set to ONE ? IC-7300 : "Data Mode" .
             pFreqMem->RxTx[0].fDataMode = TRUE;
             break;
        default:  // oops.. ignore all the rest, we're out of business !
             return TRUE; // .. but at least we've got the FREQUENCY and MODE
      }
     switch( *pbData & 0x0F ) // lower nibble (ones) : "Tone Type Settings" in an IC-7300..
      { case 0: pFreqMem->RxTx[0].iToneType = RIGCTRL_TONE_TYPE_OFF;  break;
        case 1: pFreqMem->RxTx[0].iToneType = RIGCTRL_TONE_TYPE_TONE; break;
        case 2: pFreqMem->RxTx[0].iToneType = RIGCTRL_TONE_TYPE_TSQL; break;
        case 3: pFreqMem->RxTx[0].iToneType = RIGCTRL_TONE_TYPE_DTCS; break;
        default: break;
      }
     ++pbData; // skip the "Data mode setting" (IC-9700) /
               //  "Data mode and tone type settings"  (IC-7300 extravaganza)
   }

  // In the IC-9700, the above field was just the "Data mode setting",
  //  but they invented a new field that didn't exist in the IC-7300,
  //  thus breaking compatibility.
  // In the IC-9700, the next field is "(13)", "Duplex and Tone settings".
  // In the IC-7300, the "tone type settings" were part of the "Data mode" byte. Eeek.


  return TRUE;
} // end RigCtrl_ParseMemoryContent()


//---------------------------------------------------------------------------
void RigCtrl_ParseSpectrumData_CIV( T_RigCtrlInstance *pRC,
               BYTE *pbPayload, int iPayloadLength,
               char **ppszCmtDest, char *cpCmtEndstop ) // [out,optional] comment for the "traffic log"
  // Parses the fragment of a spectrum ("Scope waveform data"),
  //     as sent by IC-7300, IC-7610, IC-7851, IC-9700, IC-705, IC-R8600, etc .
  //  [in] pbPayload[0..iPayloadLength-1] : begins with the
  //                  main scope / sub scope indicator byte,
  //                  and ends at the byte before the CI-V postable.
  //
  //  [out] pRC->SpectrumFifo[pRC->iScopeFifoHeadIndex++]
  //               |_.nBinsUsed, dblBinWidth_Hz, dblFmin_Hz, fltMagnitudes_dB[]
  //
  // NOTE: When the spectrum scope on an IC-7300 showed "Scope Out of Range"
  //       because the VFO frequency was too far away from the fixed edge
  //       frequencies, the radio kept sending unsolicited 0x27 0x00 messages,
  //       but WITHOUT waveform data !
{
  int iMainOrSubScope = pbPayload[0];       // IC-7300 : always 0x00 = MAIN scope
  int iFragment = RigCtrl_ParseBCD_CIV_LSByteFirst( pbPayload+1, 2/*nDigits*/ );  // IC-7300 : 1..11 (decimal)
  int nFragments= RigCtrl_ParseBCD_CIV_LSByteFirst( pbPayload+2, 2/*nDigits*/ );  // IC-7300 : 11 fragments
  int nBinsInFragment;
  long i32Fcenter_Hz, i32HalfSpan_Hz, i32RangeStart_Hz, i32RangeEnd_Hz;
  BYTE *pbSrcBins;
  T_RigCtrl_Spectrum  *pSpectrum;


  // [IC7300FM] page 19-12 :
  // > The 1st data sends only the wave information ( (1) .. (6) )
  // >     without the waveform data (7).
  // > The 2nd or later data sends the minimum wave information ( (1) .. (3) )
  //       with waveform data (7).
  //
  //  "Command 27 00 : Outputs the waveform data to the controller" .
  //   Example from SL's 'CAT traffic monitor', from an IC-7300 :
  //
  //                                 Cmd   pbPayload ...
  //                 Len Pre.  Addr  |<->| [0  1  2  3  4  5  6  7  8  9 10 11 12 13 14]
  // >             0x016 FE FE E0 94 27 00 00 01 11 00 56 34 12 70 00 00 25 00 00 00 00 FD ("waveform info" in CENTER mode)
  //                    /|\          |  |  |           |070.123456MHz|<--Span=2500Hz->|  |
  //                     |<------------------------- Message, here: 22 bytes ----------->|
  //   pbMessage --------'           |  |  |<------- Payload, here: 15 bytes -------->|  |
  //                                 |  |  |  |  |  |  |  |  |                           |
  //           CI-V main command ----'  |  |  |  |  |  |  |  '- (7)...                   CI-V
  //           CI-V sub command --------'  |  |  |  |  |  '---- (6)                      Postamble
  // (1) 00=main scope, 01=sub scope ------'  |  |  |                                    |
  // (2) Division number (Current, aka NOW) --'  |  | <-  here: iFragment                |
  // (3)       "Division number (Maximum *) -----'  | <-  here: nFragments               |
  // (4)        "Center or Fixed mode data" --------' <-  here: iFixedMode               |
  //                 Len             |  |  |  |  |  | |<----from---->|<---- to ------>|  |
  // >             0x016 FE FE E0 94 27 00 00 01 11 01 00 00 00 70 00 00 00 50 70 00 00 FD ("waveform info" in FIXED mode)
  //                                                   0070.000000MHz  000070.500000MHz
  //
  // Fragments #2 to #10 (full length on IC-7300) and last fragment (#11, shorter on IC-7300):
  //     "minimum wave information":       |<----->|
  //     "waveform data"           :       |       |<-------------------------------------------------------------------------------------------------------------------------------------------------->|
  //                             pbPayload[0   1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 32 33 34 35 36 37 38 39 40 41 42 43 44 45 46 47 48 49 50 51 52]
  // >             0x03C FE FE E0 94 27 00 00 02 11 32 34 31 30 2D 0D 10 1D 30 30 1C 1A 1A 1B 31 2C 12 2C 30 1C 11 1D 16 31 34 2D 12 16 11 2B 08 2E 33 2D 1E 12 10 1E 29 07 1D 27 10 1D 1C 29 09 08 19 1C FD (fragments #2..16 of 17)
  // >             0x023 FE FE E0 94 27 00 00 11 11 14 10 0D 08 0C 1A 1E 2F 1B 0A 08 13 15 31 34 1C 1C 0F 07 16 0A 1D 1A 1D 0F FD (LAST fragment, #11 of 11 from IC-7300)                                              |__ CI-V postamble (no payload)
  //                                 |  |  |<------- Payload, last fragment : Only 0x23 - 5 = #30 byte = 30 frequency bins-->|
  //                                 |  |  |  |  |
  //           CI-V main command ----'  |  |  |  |
  //           CI-V sub command --------'  |  |  |
  //                                       |  |  |
  //                                       |  |  |
  // (1) 00=main scope, 01=sub scope ------'  |  |  <- IC-7300: always 0x00 = main scope
  // (2) Division number (Current, aka NOW) --'  |  <-  here: iFragment  = 2..11 (decimal)
  // (3)       "Division number (Maximum *) -----'  <-  here: nFragments = 11 (decimal)
  //     No "Center or Fixed mode data" info in fragments #2..11 (thus "min wave info")
  //
  //  (*) The number of "divisions" (fragments in subsequent messages)
  //      depends on the radio model. This parser shall work on ANY radio.
  // Quoted from IC-7300 "Full Manual" page 19-12:
  //  > When sent through the USB port, the data is
  //  > divided by 11 (*) and sent in sequential order.
  //  > The 1st data sends only the wave information ( (1) ~ (6) )
  //  >                 without the waveform data (7).
  //  > The 2nd or later data sends the minimum wave information
  //  >             ( (1) ~ (3) ) with waveform data (7).
  // (*) "divided by 11" : What they mean is "split into 11 fragments" .
  //     Don't take this number of FRAGMENTS per spectrum for granted:
  //   IC-7300, IC-R8600 : "the data is divided by 11" (over USB).
  //   IC-7610, IC-7851  : "the data is divided by 15" (over USB),
  //         but sometimes "all data is sent together" (over LAN).
  if( iMainOrSubScope != 0 )  // only for the MAIN spectrum scope ...
   { return;  // sub-scope not supported. A donated IC-7851 would change this :)
   }

  // Assemble the received fragment in SpectrumFifo[] ..
  pRC->iScopeFifoHeadIndex %= RIGCTRL_SPECTRUM_FIFO_SIZE; // safety first...
  pSpectrum = &pRC->SpectrumFifo[ pRC->iScopeFifoHeadIndex ];

  if( iFragment == 1 )  // first fragment (Icom: "1st data") ?
   { pRC->iScopeBinCounter = 0; // index into pRC->Spectrum[].
     pSpectrum->dblUnixTime= UTL_GetCurrentUnixDateAndTime(); // capture timestamp a.s.a.p.
     if( pbPayload[3] == 0 ) // 0 = center mode, 1=fixed mode
      { // (this info is only contained in fragment #1)
        pSpectrum->fFixedEdgeMode = FALSE; // from IC-7300/IC-9700 with spectrum scope in CENTER mode
        i32Fcenter_Hz  = RigCtrl_ParseBCD_CIV_LSByteFirst( pbPayload+4, 10/*nDigits*/ );
        i32HalfSpan_Hz = RigCtrl_ParseBCD_CIV_LSByteFirst( pbPayload+9, 10/*nDigits*/ );
        pSpectrum->dblCenterFreq_Hz = i32Fcenter_Hz;
        // Unfortunately, before the total number of frequency bins is known,
        // it's impossible to determine pSpectrum->dblBinWidth_Hz. Thus:
        pRC->ScopeFreqRangeFromFirstFragment.dblFmin_Hz = (double)(i32Fcenter_Hz-i32HalfSpan_Hz); // "radio frequency" stored in the 1st bin
        pRC->ScopeFreqRangeFromFirstFragment.dblFmax_Hz = (double)(i32Fcenter_Hz+i32HalfSpan_Hz); // "radio frequency" stored in the last bin
            // ... without knowing how many bins the radio will actually send !
        SL_AppendPrintf( ppszCmtDest, cpCmtEndstop, "%ld+-%ld kHz",
             (long)i32Fcenter_Hz/1000, (long)i32HalfSpan_Hz/1000 );
      }
     else // not "center" mode but "fixed" mode :
      { pSpectrum->fFixedEdgeMode = TRUE; // from IC-7300/IC-9700 with spectrum scope in FIXED EDGE mode
        i32RangeStart_Hz = RigCtrl_ParseBCD_CIV_LSByteFirst( pbPayload+4, 10/*nDigits*/ );
        i32RangeEnd_Hz   = RigCtrl_ParseBCD_CIV_LSByteFirst( pbPayload+9, 10/*nDigits*/ );
        pSpectrum->dblCenterFreq_Hz = 0.5 * ( i32RangeStart_Hz + i32RangeEnd_Hz );
        pRC->ScopeFreqRangeFromFirstFragment.dblFmin_Hz = (double)i32RangeStart_Hz;
        pRC->ScopeFreqRangeFromFirstFragment.dblFmax_Hz = (double)i32RangeEnd_Hz;
        SL_AppendPrintf( ppszCmtDest, cpCmtEndstop, "%ld..%ld kHz",
             (long)i32RangeStart_Hz/1000, (long)i32RangeEnd_Hz/1000 );
      }
     pRC->dblSpectrumCenterFreq_Hz = pSpectrum->dblCenterFreq_Hz;

     // The "first data" (first of N fragments) often contains "wave information",
     //       and usually no "waveform data" (spectrum frequency bins) .
     // BUT : An IC-7610 (which not only has USB, but also a LAN port)
     //       may send EVERYTHING in a single fragment, when using LAN !
     // To support both (USB + LAN), use the PAYLOAD LENGTH to know
     // if (as quoted for the IC-7300 further above) this FIRST FRAGMENT
     // really only contains "wave information" but no "waveform data" :
     nBinsInFragment = iPayloadLength - 15; // -> IC-7300 : iPayloadLength = 15, nBinsInFragment = 0, i.e. NO FREQUENCY BINS in the first fragment.
     pbSrcBins = pbPayload + 15;

   } // end if( iFragment == 1 )
  else if( iFragment>1 && (iFragment<=nFragments) )
   { // the "2nd or later data" (2nd to Nth fragment) sends the minimum
     // wave information, no info about the frequency range :
     nBinsInFragment = iPayloadLength - 3; // IC-7300 : 53 minus 3 bytes "minimum wave information"
         // -> max 50 bins per fragment (except the last, which only contained
         //    28 minus 3 = 25 frequency bins -> total = 9 * 50 + 25 = 475 bins.
     pbSrcBins = pbPayload + 3;
     SL_AppendPrintf( ppszCmtDest, cpCmtEndstop, "fragment %d/%d", iFragment, nFragments );
   } // end if( iFragment>1 && (iFragment<=nFragments) )


  while( ((nBinsInFragment--) > 0)  // anything to append in THIS fragment ?
     &&  (pRC->iScopeBinCounter<RIGCTRL_MAX_FREQ_BINS_PER_SPECTRUM) )
   {
#   if( SWI_HARDCORE_DEBUGGING )
     RigCtrl_pbSource = pbSrcBins;
     RigCtrl_iScopeBin= pRC->iScopeBinCounter;
     RigCtrl_pfltDest = pSpectrum->fltMagnitudes_dB;
#   endif
     pSpectrum->fltMagnitudes_dB[ pRC->iScopeBinCounter++ ] = (float)( *pbSrcBins++ );
     // '-- 2024-04-21 : read from address ZERO ? ?  (only after pause/resume via debugger)
     //     2024-08-05 : similar trouble again, this time RigCtrl_pbSource confirmed
     //                  that pbSrcBins was NULL. This happened after switching
     //                  to a new BAND STACKING REGISTER though the
     //                  Remote CW Keyer's "band" combo on the "TRX" tab.
   }

  pSpectrum->iAmplitudeUnit = SCALE_UNIT_dB; // pSpectrum->fltMagnitudes_dB[] still uses Icom's "phantasy" unit.
  pSpectrum->fltAmplOffset_dBm = -146.0; // offset to convert fltMagnitudes_dB into dBm, see below
  // The CI-V manual only mentions "data range 0..160" but doesn't explain
  //   the scaling (one step = one decibel ? Relative to WHAT ? Details below.)
  // Design goal : Value RETURNED from here in dBm, comparable with S-meter units.
  // > IARU Region 1 Technical Recommendation R.1 defines S9 for the HF bands
  // >  to be a receiver input power of -73 dBm. This is a level of 50 microvolts
  // >  at the receiver's antenna input assuming the input impedance of the
  // > receiver is 50 ohms.
  // > For VHF bands the recommendation defines S9 to be a receiver input power
  // >  of -93 dBm. This is the equivalent of 5 microvolts in 50 ohms.[6]
  // > The recommendation defines a difference of one S-unit corresponds
  // >   to a difference of 6 decibels (dB), equivalent to a voltage ratio
  // >   of two, or power ratio of four.
  // An IC-7300, fed by a "-20 dBm" signal showed ~~ "S9+43dB" at 3.560 MHz,
  //             (Rigol tracking generator)   and ~~ "S9+22dB" WITH ATTENUATOR.
  //             Scope Bandwidth = 500 kHz .
  // With the generator attenuated until the S-meter JUST showed "S0", "S1", "S2", ..,
  //          the peak bin (in pbSrcBins) delivered  :
  // S-meter   |  "Scope waveform data" (value in pbSrcBins)
  // ----------+---------------------------------------------------
  //   "S0"    :  20 without attenuator,   0 (flat line) with attenuator
  //                 (should be  -127 dBm  =  -22 dBuV in a 50-Ohm system)
  //
  //   "S1"    :  26 without attenuator,   still flat line with attenuator
  //                 (should be  -121 dBm  =  -14 dBuV )
  //     ...
  //   "S5"    :  48 (should be   -97 dBm  =  +10 dBuV )
  //   "S6"    :  54 (should be   -91 dBm  =  +16 dBuV )
  //   "S7"    :  61 (should be   -85 dBm  =  +22 dBuV )
  //   "S8"    :  66 (should be   -79 dBm  =  +28 dBuV )
  //   "S9"    :  73 without attenuator,  35 with attenuator (-> "below S3").
  //                 (should be   -73 dBm  =  +34 dBuV )
  //   "S9+20" :  110 (!?!) without "  ,  73 with attenuator (-> "S9" ).
  // -> Conclusion: pbSrcBins uses neither dBm nor dBuV but some 'phantasy dB",
  //                possibly  "dB relative to the own noise floor".
  //                DL4YHF initially subtracted 73+73 = 146, if (in SpecLab)
  //                the result shall be displayed in dBm ( dB "over" a milliwatt) .
  //                But with a dummyload on the RX antenna input,
  //                the spectrum read '-146 dBm' which is nonsense for an
  //                equivalent receiver bandwidth of 10.5 Hz (IC-7300 spanning
  //                5 kHz of "scope bandwidth" in 475 frequency bins) .
  //     From https://en.wikipedia.org/wiki/Minimum_detectable_signal :
  //        > To calculate the minimum detectable signal we first need to establish
  //        > the noise floor in the receiver by the following equation:
  //        > Noise_floor_dBm = 10*log10( kb * To * 1000 ) + NF + 10*log10(BW),
  //        >  where  kb = Boltzmann constant, 1.38 * 10^-23 Joules per Kelvin
  //        >         To = Temperature of the receiver system in Kelvin, e.g. 290 K
  //        >         NF = Receiver noise figure in dB, let's assume NF = 1.5 [dB]
  //        >         BW = Detection bandwidth in Hz, let's use 84 Hz (Icom spectrum scope).
  //        > Numerical example (modified for the IC-7300's spectrum scope) :
  //        > Noise_floor_dBm = 10*log10( kb * To * 1000 ) + NF + 10*log10(BW)
  //        >                 = 10*log10(1.38E-23 * 290 * 1000) + 1.5 + 10*log10(10.5)
  //        >                 = -162.3 dBm .
  //     Ouch. So we're about 162 - 146 = 16 dB "too high" near the noise floor.
  // The MAXIMUM ("data range 160")
  //                would then be 160 "dB" - 146 dBm = +14 dBm = "overload".
  //     A broadcaster on 7.215 MHz, indicated as circa "S9+15 dB" by the IC-7300
  //       reached '119' (non-scaled peak in pbSrcBins), indicated as
  //       was peaking at (119-146) = -27 dBm in spectrum lab (from 5 kHz wide spectrum scope).
  //       "S9 + 15db" should be -73 dBM + 15 dB = -58 dBm,
  //     so here, we're about 58 - 27 = 31 dB "too low" near the maximum level.
  //     WB, 2020-06-07 : Decided to give up at this point.
  //                      An exact "S9" signal from a generator gave
  //                      "-73 dBm" in SL's own waterfall (with "FFT output unit"
  //                      set to "dBm" in Spectrum Lab) with the preamp OFF,
  //                      so decided to keep pSpectrum->fltAmplOffset_dBm .
  //          Then tried different preamp/attenuator settings in the IC-7300 :
  //             No preamp, no attenuator : "S9" on the S-meter, -73 dBm in SL.
  //             Preamp "1" : "S9 + 5dB" on the s-meter,  -47 dBm in SL (nuts..),
  //             Preamp "2" : "S9 + 10dB" on the s-meter, -39 dBm in SL (nuts..),
  //             ATTENUATOR : "S3" on the Icom's s-meter, -118 dBm in SL (CRAZY!).
  //
  if( (iFragment == nFragments ) && (pRC->iScopeBinCounter>1) ) // all fragments complete -> "emit" the spectrum
   {
     // With an IC-7300, got here with pRC->iScopeBinCounter = 475 .
     pSpectrum->nBinsUsed = pRC->nSpectrumBinsUsed = pRC->iScopeBinCounter;
     pSpectrum->dblFmin_Hz = pRC->ScopeFreqRangeFromFirstFragment.dblFmin_Hz;
     pSpectrum->dblBinWidth_Hz = (pRC->ScopeFreqRangeFromFirstFragment.dblFmax_Hz
                                - pRC->ScopeFreqRangeFromFirstFragment.dblFmin_Hz)
                      / (double)( pRC->iScopeBinCounter - 1 );
     // Allow reading this entry in another thread now :
     pRC->iScopeFifoHeadIndex = (pRC->iScopeFifoHeadIndex + 1) % RIGCTRL_SPECTRUM_FIFO_SIZE;
     if( pRC->iScopeFifoHeadIndex == 0 )
      {  pRC->fScopeFifoFull = TRUE;   // important for RigCtrl_GetSpectrumFromFIFO() with RIGCTRL_GET_LATEST_ENTRY
      }
     pRC->iScopeBinCounter = 0; // index into pRC->Spectrum[].
     if( pRC->dblUnixTimeOfFirstSpectrum <= 0.0 )
      {  pRC->dblUnixTimeOfFirstSpectrum = pSpectrum->dblUnixTime;
      }
     ++pRC->nCompleteSpectraReceived;  // total spectrum counter, added 2018-12
     // to find out if the spectrum-update-rate "seen" in Spectrum Lab
     // was really so much lower than it appeared to be on the radio.
     // Test result: see DebugU1.cpp : TDebugForm::Btn_GetReportOnCommsClick().
   }
} // end RigCtrl_ParseSpectrumData_CIV()


//--------------------------------------------------------------------------
T_RigCtrl_Spectrum *RigCtrl_AppendSpectrumToFIFO( T_RigCtrlInstance *pRC,
      double dblBinWidth_Hz, // [in] 'width' of a single FFT bin in Hertz (=resolution
      double dblFmin_Hz,     // [in] "radio frequency" stored in the first frequency bin
      int    nFrequencyBins) // [in] number of 'frequency bins' .
  // Added 2024-03 to use the Spectrum FIFO also for the CLIENT SIDE
  //       in the 'Remote CW Keyer', which receives spectra via TCP/IP
  //       (see cbproj/Remote_CW_Keyer/CwNet.c : )
  // No whistles and bells here, no averaging, etc.
  // Note the fixed limit of RIGCTRL_MAX_FREQ_BINS_PER_SPECTRUM !
  //
  // Returns NULL if the RigControl instance hasn't been properly initialized.
  //
{
  T_RigCtrl_Spectrum *pSpectrum;

  if( (pRC==NULL) || (nFrequencyBins<1) || (nFrequencyBins>RIGCTRL_MAX_FREQ_BINS_PER_SPECTRUM) )
   { return NULL;
   }

  pRC->iScopeFifoHeadIndex %= RIGCTRL_SPECTRUM_FIFO_SIZE; // safety first...
  pSpectrum = &pRC->SpectrumFifo[ pRC->iScopeFifoHeadIndex++ ];
  pRC->iScopeFifoHeadIndex %= RIGCTRL_SPECTRUM_FIFO_SIZE;
  if( pRC->iScopeFifoHeadIndex == 0 ) // similar as in RigCtrl_ParseSpectrumData_CIV() [which ALSO fills the spectrum-FIFO] ..
   {  pRC->fScopeFifoFull = TRUE;   // important for RigCtrl_GetSpectrumFromFIFO() with RIGCTRL_GET_LATEST_ENTRY
   }



  // Fill out the fixed part of the T_RigCtrl_Spectrum, so the spectrum is
  // at least 'valid' when returning, and the caller only needs to populate the
  // bins ( pSpectrum->fltMagnitudes_dB[0..RIGCTRL_MAX_FREQ_BINS_PER_SPECTRUM-1] )
  pSpectrum->nBinsUsed      = nFrequencyBins;
  pSpectrum->dblBinWidth_Hz = dblBinWidth_Hz;
  pSpectrum->dblFmin_Hz     = dblFmin_Hz;
  pSpectrum->dblCenterFreq_Hz= dblFmin_Hz + 0.5 * dblBinWidth_Hz * nFrequencyBins;
  pSpectrum->fFixedEdgeMode = FALSE; // <- just a GUESS here.. the caller may overwrite this when known
  pSpectrum->dblUnixTime = UTL_GetCurrentUnixDateAndTime(); // capture timestamp a.s.a.p.
  pSpectrum->iAmplitudeUnit = SCALE_UNIT_dB; // pSpectrum->fltMagnitudes_dB[] still uses Icom's "phantasy" unit.
  pSpectrum->fltAmplOffset_dBm = -146.0; // offset to convert fltMagnitudes_dB into dBm, see below

  ++pRC->nCompleteSpectraReceived;   // total counter, displayed on the 'Debug' tab   

  return pSpectrum;
} // end RigCtrl_AppendSpectrumToFIFO()


//--------------------------------------------------------------------------
T_RigCtrl_Spectrum *RigCtrl_GetSpectrumFromFIFO( T_RigCtrlInstance *pRC, int *piTailIndex )
  // API used by Spectrum Lab to retrieve the address of another spectrum in the FIFO.
  // Avoids unnecessary copying. No whistles and bells here, no averaging, etc.
  //     Any reader (caller) should initialize HIS "tail index" (*piTailIndex)
  //     with *piTailIndex = RIGCTRL_GET_OLDEST_ENTRY to retrieve the oldest available entry in the FIFO,
  //       or *piTailIndex = RIGCTRL_GET_LATEST_ENTRY to retrieve the newest (most recent) entry.
  //
  // Returns NULL if no (new) spectrum is available .
  //
  // In Spectrum Lab, periodically called from void TSpectrumLab::ProcessSpectraFromRemoteRig(),
  //    which may have to 'automatically adjust' the DISPLAYED FREQUENCY RANGE
  //    to the radio-frequency-range contained in the returned T_RigCtrl_Spectrum.
  //    (what makes this incredibly complex is the fact that the Icom radio's
  //     *VFO FREQUENCY* may be outside the *DISPLAYED FREQUENCY RANGE*,
  //     especially with the "Scope" (spectrum display) in "Fix" mode ("fixed edge
  //     without automatic scolling", which has been implemented in more recent
  //     firmware versions).
  //     Solution / details in SpecMain.cpp : TSpectrumLab::ProcessSpectraFromRemoteRig() !
{
  T_RigCtrl_Spectrum *pSpectrum = NULL;

  if( *piTailIndex == RIGCTRL_GET_LATEST_ENTRY )  // since 2020-06, dummy for the MOST RECENT spectrum
   { if( pRC->fScopeFifoFull ) // ALL FIFO entries (except the one at iScopeFifoHeadIndex) are valid...
      { *piTailIndex = pRC->iScopeFifoHeadIndex - 1; // signed index wrap below..
        if( *piTailIndex < 0 )
         {  *piTailIndex = RIGCTRL_SPECTRUM_FIFO_SIZE-1;
         }
      }
     else // index hasn't wrapped around yet, so the oldest available info is at index zero
      { *piTailIndex = pRC->iScopeFifoHeadIndex - 1; // if negative ..
        if( *piTailIndex < 0 ) // .. then there's no valid spectrum at all
         {  *piTailIndex = 0;
         }
      }
   }
  else if( *piTailIndex == RIGCTRL_GET_OLDEST_ENTRY ) // not the MOST RECENT but the OLDEST entry ...
   { if( pRC->fScopeFifoFull ) // ALL FIFO entries (except the one at iScopeFifoHeadIndex) are valid...
      { *piTailIndex = pRC->iScopeFifoHeadIndex + 1; // circular wrap further below..
      }
     else // index hasn't wrapped around yet, so the oldest available entry is at index zero
      { *piTailIndex = 0;
      }
   }
  *piTailIndex %= RIGCTRL_SPECTRUM_FIFO_SIZE;

  if( *piTailIndex != pRC->iScopeFifoHeadIndex ) // something new available for THIS caller !
   { pSpectrum = &pRC->SpectrumFifo[ *piTailIndex ];
     *piTailIndex = (*piTailIndex + 1) % RIGCTRL_SPECTRUM_FIFO_SIZE;
   }

  return pSpectrum;
} // end RigCtrl_GetSpectrumFromFIFO()

//--------------------------------------------------------------------------
CPROT int RigCtrl_GetNumSpectraAvailableInFIFO( T_RigCtrlInstance *pRC, int iTailIndex )
  // Called from e.g. OpenWebRX_Server.c to check IN ADVANCE (before
  //  calling RigCtrl_GetSpectrumFromFIFO() ) HOW MANY spectra (or "waterfall lines")
  //  are 'still waiting to be sent' to a certain client.
{
  int nEntries = 0;
  switch( iTailIndex )
   { case RIGCTRL_GET_LATEST_ENTRY : // dummy for the "MOST RECENT spectrum"
     case RIGCTRL_GET_OLDEST_ENTRY : // dummy for the OLDEST entry ...
          if( pRC->fScopeFifoFull ) // ALL FIFO entries (except the one at iScopeFifoHeadIndex) are valid...
           { nEntries = 1;
           }
          else // "scope FIFO" not full, thus check the *HEAD INDEX* :
           { if( pRC->iScopeFifoHeadIndex > 0 )
              { nEntries = 1;
              }
           }
          break;
     default:  // caller passed in a FIFO TAIL index, so compare against the HEAD index:
          nEntries = (pRC->iScopeFifoHeadIndex - iTailIndex);
          if( nEntries < 0 )   // circular wrap between head and tail ..
           {  nEntries += RIGCTRL_SPECTRUM_FIFO_SIZE;
           }
          break;
   }
  return nEntries;
} // end RigCtrl_GetNumSpectraAvailableInFIFO()


//--------------------------------------------------------------------------
BOOL RigCtrl_GetDisplayableRadioFrequencyRange( T_RigCtrlInstance *pRC,
        double *pdblRadioFreqMin, double *pdblRadioFreqMax, // [out] : RADIO  frequencies [Hz]
        BOOL   *pfFixedEdgeMode )   // [out,optional: TRUE= "fixed edges", FALSE="centered on VFO"
  // Spectrum Lab shall be aware of Icom's "FIXED" / "CENTER" mode setting !
  //    - in FIXED-EDGE mode, the frequency scale (plus spectrum, etc)
  //      must not(!) scroll along with the VFO frequency, because
  //      the VFO frequency may even be OUTSIDE the scope's displayed
  //      frequency range (that's not a bug, but a feature of Icom's
  //      spectrum scope).
  //    - in CENTER mode, the frequency scale (plus spectrum, etc)
  //      must (or at least should) scroll along with the VFO frequency.
  //
  // Return value : TRUE = ok ("we know the frequency range"),
  //                FALSE= no info received from the radio yet, using defaults.
  //
  // That's why RigCtrl_GetDisplayableRadioFrequencyRange() optionally returns
  //      the spectrum scope's current "FIXED" / "CENTER" setting .
{
  T_RigCtrl_Spectrum *pSpectrum;
  int iTailIndex;
  double dblSpan_Hz;

  // When already running, use the most recent spectrum to find the range :
  iTailIndex = RIGCTRL_GET_LATEST_ENTRY;
  pSpectrum  = RigCtrl_GetSpectrumFromFIFO( pRC, &iTailIndex );
  if( pSpectrum != NULL )
   { if( pfFixedEdgeMode != NULL )
      { *pfFixedEdgeMode =  pSpectrum->fFixedEdgeMode; // TRUE=fixed frequencies, FALSE=spectrum centered on VFO
      }
     // Bugfix 2020-06-19 : Span_Hz = (nBinsUse MINUS ONE) * BinWidth_Hz  !
     //        If there are only TWO bins per spectrum, the total span
     //        (from the center of the first  to  the center of the last bin)
     //        is only ONE frequency bin width !
     // ex: dblSpan_Hz = (double)pSpectrum->nBinsUsed * pSpectrum->dblBinWidth_Hz;
     dblSpan_Hz = (double)(pSpectrum->nBinsUsed-1) * pSpectrum->dblBinWidth_Hz;
     //   After this modification, the "sp" edit field in SL's main window
     //   showed the correct span ( 1000.0 kHz, not 1.0021 MHz, for an IC-7300 ).
     if( dblSpan_Hz > 0.0 )
      { *pdblRadioFreqMin = pSpectrum->dblCenterFreq_Hz - 0.5 * dblSpan_Hz;
        *pdblRadioFreqMax = pSpectrum->dblCenterFreq_Hz + 0.5 * dblSpan_Hz;
        return TRUE;
      }
   } // end if( pSpectrum != NULL )

  // Called "early", or spectrum FIFO still empty for any other reason: 2nd guess !
  // If it's an Icom radio, it may have reported its "Span setting in the Center mode Scope"...
  if( pRC->dblScopeSpan_Hz > 0.0 ) // An IC-7300 with a 5 kHz wide display reported "2500 Hz",
   { // .. but that applies to BOTH SIDES of the "center"- (not always the "VFO"-) frequency !
     *pdblRadioFreqMin = pRC->dblVfoFrequency - pRC->dblScopeSpan_Hz;
     *pdblRadioFreqMax = pRC->dblVfoFrequency + pRC->dblScopeSpan_Hz;
     return TRUE;
   }

  // Arrived here ? The spectrum scope's frequency range is completely unknown,
  //  so return the WIDEST possible setting (*) :
  *pdblRadioFreqMin = pRC->dblVfoFrequency-500e3;  // 1 MHz total span from IC-7300
  *pdblRadioFreqMax = pRC->dblVfoFrequency+500e3;
     // (*) Returning FALSE from here makes Spectrum Lab use the last known
     //     frequency range, stored between two sessions in e.g.
     //     UConfig.SpDisplay[0].Chn[0].dblBasebandFreqMin/Max, added to
     //     UConfig.SDR_dblNCOfreq .
  return FALSE;

} // end RigCtrl_GetDisplayableRadioFrequencyRange()


//--------------------------------------------------------------------------
int RigCtrl_SubcodeFrom1A_to_UnifiedParameterNumber( T_RigCtrlInstance *pRC,
           int iSubcode )
{
  switch( pRC->iDefaultAddress )
   { case RIGCTRL_DEF_ADDR_IC_7300 :
        switch( iSubcode ) // IC-7300-specific !!
         {
         //case 0x06 :  // "DATA mode with filter width setting" (IC-7300)
              // There's a hard-coded parser for this in RigCtrl_ParseCIV() !
           case 0x07 :  // "IP+ Setting"       (IC-7300)
              return RIGCTRL_PN_IP_PLUS;  // in many rigs, "IP+ ON" == "Preamp OFF".
                 // But there's a different command / subcode for the PREAMP,
                 // so at least in THIS radio, "IP+" may be something more...
           default:
              break;
         }
        break;
     case RIGCTRL_DEF_ADDR_IC_7610 :
        switch( iSubcode ) // IC-7610-specific !!
         {
           case 0x07 :  // "NTP server access" (IC-7610)
              return RIGCTRL_PN_NTP_SERVER_ACCESS;
           case 0x08 :  // "NTP access result" (IC-7610)
              return RIGCTRL_PN_NTP_ACCESS_RESULT;
           case 0x09 :  // "AF Mute"           (IC-7610)
              return RIGCTRL_PN_AF_MUTE;
           default:
              break;
         }
        break;
     default:
        break;
   }

  // Arrived here ? Good luck digging out the manual, and finding out
  //  the meaning of the subcode directly after command 0x1A yourself !
  return RIGCTRL_PN_UNKNOWN;


} // end  RigCtrl_SubcodeFrom1A_to_UnifiedParameterNumber()


//--------------------------------------------------------------------------
int RigCtrl_SubcodeFrom1A05_to_UnifiedParameterNumber( T_RigCtrlInstance *pRC,
           int iSubcode )
  // Tries to convert A FEW(!) of those rig-specific 4-digit subcodes
  //       after Icom's command 0x1A 0x05 into a UNIFIED, and CONSTANT
  //       parameter number that is usable for a switch-case list in "C".
  //       It's a shame that not even an IC-7300 and IC-7610 are compatible
  //       in this context . They could have simply added NEW "subcodes"
  //       for the IC-7610 (for parameters that don't exist in an IC-7300),
  //       but the Icom developers didn't. Thanks for nothing.
  //  [in] pRC->iDefaultAddress : e.g. RIGCTRL_DEF_ADDR_IC_7300, .._7610, etc.
  //       Has been read via CI-V *from the radio itself*.
  //  [in] iSubcode = rig-specific number, parsed from a 4-digit BCD coded
  //       value after CI-V command 0x1A 0x05 .
  //  Return value : RIGCTRL_PN_CIV_USB_ECHO, etc
  //       (only those parameter that appeared important are supported here)
{
  switch( pRC->iDefaultAddress )
   { case RIGCTRL_DEF_ADDR_IC_7300 :
        switch( iSubcode ) // IC-7300-specific !!
         { case    1 : return RIGCTRL_PN_SSB_RX_HPF_LPF;
           case    2 : return RIGCTRL_PN_SSB_RX_BASS_LEVEL;
           case    3 : return RIGCTRL_PN_SSB_RX_TREBLE_LEVEL;
           // etc etc, the above were just A FEW examples.. now for the important ones, especially those used by "RS-BA1":
           case   52 : return RIGCTRL_PN_SSB_CW_SYNC_TUNING; // "SSB/CW synchronous tuning". In IC-7300, PN #0052. Used by RS-BA1.
           case   57 : return RIGCTRL_PN_CALIBRATION_MARKER; // "Calibration marker". In IC-7300, PN #0057. Used by RS-BA1.
           case   60 : return RIGCTRL_PN_AF_OUTPUT_LEVEL_TO_ACC; // .. or "to ACC or USB". In IC-7300, PN #0060
           case   61 : return RIGCTRL_PN_SQUELCH_FOR_AF_ON_ACC;  // "Squelch function for the AF signal output to ACC/USB"
           case   62 : return RIGCTRL_PN_BEEP_AND_SPEECH_ON_ACC; // "Beep and speech output setting to ACC/USB"
           case   63 : return RIGCTRL_PN_IF_OUTPUT_LEVEL_TO_ACC; // "IF signal output level to ACC/USB" (0..255 for 100%)
           case   64 : return RIGCTRL_PN_MOD_INPUT_LEVEL_FROM_ACC; // "MOD input level from ACC" (0..255 for 100%)
           case   65 : return RIGCTRL_PN_MOD_INPUT_LEVEL_FROM_USB; // "MOD input level from USB" (0..255 for 100%)
           case   66 : return RIGCTRL_PN_MOD_IN_CONN_FOR_DATA_OFF; // "MOD input connector during DATA OFF" (0=MIC, 1=ACC, 2=MIC/ACC, 3=USB)
           case   67 : return RIGCTRL_PN_MOD_IN_CONN_FOR_DATA; // "MOD input connector during DATA" (0=MIC, 1=ACC, 2=MIC/ACC, 3=USB)
           case   71 : return RIGCTRL_PN_CIV_TRANSCEIVE;
           case   72 : return RIGCTRL_PN_CIV_ADDR_USB_TO_REMOTE; // data: hex address
           case   73 : return RIGCTRL_PN_CIV_OUTPUT_FOR_ANT;
           case   74 : return RIGCTRL_PN_CIV_UNLINK_FROM_REMOTE; // data: 0=linked, 1=unlinked
           case   75 : return RIGCTRL_PN_CIV_USB_ECHO;
           default   : break;
         }
        break; // end case RIGCTRL_DEF_ADDR_IC_7300

     case RIGCTRL_DEF_ADDR_IC_7610 :
        switch( iSubcode ) // IC-7610-specific !!
         { case    1 : return RIGCTRL_PN_SSB_RX_HPF_LPF;
           case    2 : return RIGCTRL_PN_SSB_RX_BASS_LEVEL;
           case    3 : return RIGCTRL_PN_SSB_RX_TREBLE_LEVEL;
           // etc etc, the above were just A FEW examples.. now for the important ones:
           case  112 : return RIGCTRL_PN_CIV_TRANSCEIVE;
           case  113 : return RIGCTRL_PN_CIV_ADDR_USB_TO_REMOTE; // here: also "LAN to remote"
           case  114 : return RIGCTRL_PN_CIV_OUTPUT_FOR_ANT;
           case  115 : return RIGCTRL_PN_CIV_UNLINK_FROM_REMOTE; // data: 0=linked, 1=unlinked
           case  116 : return RIGCTRL_PN_CIV_USB_ECHO;
           default   : break;
         }
        break; // end case RIGCTRL_DEF_ADDR_IC_7610
   } // end switch( pRC->iDefaultAddress )

  // Arrived here ? Good luck digging out the manual, and finding out
  //  the meaning of the subcode for command 0x1A 0x05 yourself !
  return RIGCTRL_PN_UNKNOWN;
} // end RigCtrl_SubcodeFrom1A05_to_UnifiedParameterNumber()


//--------------------------------------------------------------------------
T_RigCtrl_ParamInfo * RigCtrl_GetInfoForUnifiedParameterNumber(int iUnifiedPN)
  // As with other function returning a POINTER, the caller must check
  // for a NULL-result, because returning NULL means "not found" .
{
  T_RigCtrl_ParamInfo *pPI = (T_RigCtrl_ParamInfo*)RigCtrl_ParameterInfo;

  while( pPI->iUnifiedPN != 0 )
   { if( pPI->iUnifiedPN == iUnifiedPN )
      { return pPI;
      }
     ++pPI;
   }

  return NULL;  // no luck ..
}

//--------------------------------------------------------------------------
T_RigCtrl_ParamInfo * RigCtrl_GetInfoForCIVCommandAndSubcode(
          long i32CombinedCmdAndSubcode) // [in] "combined" CI-V command + "sub-command"
               //      and possibly even more (e.g. for 0x1A 0x05 ....)
               //       - see RigCtrl_GetCombinedCmdAndSubcodeForCIV() !
  // Does almost the same as RigCtrl_GetInfoForUnifiedParameterNumber(),
  //  but retrieves the parameter info for a certain "Icom CI-V command + sub-command".
  // As with other function returning a POINTER, the caller must check
  // for a NULL-result, because returning NULL means "not found" .
{
  T_RigCtrl_ParamInfo *pPI = (T_RigCtrl_ParamInfo*)RigCtrl_ParameterInfo;
  BYTE bCmd     = ( i32CombinedCmdAndSubcode >> 24 ) & 0xFF;
  BYTE bSubCmd  = ( i32CombinedCmdAndSubcode >> 16 ) & 0xFF;
  BYTE bSubSub1 = ( i32CombinedCmdAndSubcode >> 8  ) & 0xFF; // Icom doesn't have a name for this..
  BYTE bSubSub2 = ( i32CombinedCmdAndSubcode >> 0  ) & 0xFF; // it's the stuff after e.g. 0x1A 0x05
  BOOL fMatch;

  while( pPI->iUnifiedPN != 0 )
   { if( (pPI->pbCIVReadCmd != NULL )  // parameter supported by Icom CI-V at all ?
      && (pPI->iCIVReadCmdLength > 0) )
      { if( bCmd == pPI->pbCIVReadCmd[0] ) // byte-sequence for RigCtrl_SendAndLogMessage() to READ
         { // the "command" (a la CI-V) matches, but what about the optional SUB-command ?
           fMatch = TRUE;
           // The length of a READ-COMMAND (aka "read request" from host to radio)
           // indicates the presence of an extra byte for what Icom calls "Sub cmd." :
           if( pPI->iCIVReadCmdLength > 1 )
            { fMatch &= (bSubCmd == pPI->pbCIVReadCmd[1]);
            }
           if( pPI->iCIVReadCmdLength > 2 )
            { fMatch &= (bSubSub1 == pPI->pbCIVReadCmd[2]);
              // ,-------|______|
              // '--> e.g. 0x01 for the IC-9700's "SET > Network > Network Control"
              //           (0x1A 0x05 0145 ... we don't care if this is BCD or HEX)
              //             |    |   |  |
              //             |    |   \__|___ up to two extra bytes (four BCD digits?)
              //             |    |__________ the FIRST of up to THREE BYTES
              //             |                that Icom lists in column "Sub cmd."
              //             '--------------- This is what Icom lists as "Cmd." .
              //                 It's more or less the "most significant byte" .
            }
           if( pPI->iCIVReadCmdLength > 3 )
            { fMatch &= (bSubSub2 == pPI->pbCIVReadCmd[3]);
              // ,-------|______|
              // '--> e.g. 0x45 for the IC-9700's "SET > Network > Network Control"
              //           (0x1A 0x05 0145)
            }
           if( fMatch )
            { return pPI;
            }
         }    // matching FIRST 'Cmd' byte ?
      }      // end if( (pPI->pbCIVReadCmd != NULL ) && ... )
     ++pPI;  // check the next parameter in RigCtrl_ParameterInfo[]
   }
  return NULL;  // no luck ..
} // GetInfoForCIVCommandAndSubcode()


//--------------------------------------------------------------------------
const char *RigCtrl_UnifiedParameterNumberToString(int iUnifiedPN)
{
  static char sz11Unknown[12];
  T_RigCtrl_ParamInfo *pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
  if( pPI != NULL )
   { return pPI->pszToken;
   }

  // The following isn't thread-safe but shouldn't be necessary at all:
  sprintf( sz11Unknown, "PN[%d]",(int)iUnifiedPN);
  return   sz11Unknown;

} // end RigCtrl_UnifiedParameterNumberToString()


//---------------------------------------------------------------------------
int * RigCtrl_GetIntValuePtrForUnifiedParameterNumber( T_RigCtrlInstance *pRC, int iUnifiedPN )
{
  T_RigCtrl_ParamInfo *pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
  if( pPI != NULL )
   { if( (pPI->iByteOffsetIntoRigCtrlInstance > 0 )
      && (pPI->iByteOffsetIntoRigCtrlInstance < sizeof(T_RigCtrlInstance) ) )
      { if( pPI->iRigCtrlDataType == RIGCTRL_DT_INT )
         { return (int*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance );
         }
      }
   }
  return NULL;  // no luck .. also here, the caller MUST check for NULL-pointers !
} // end RigCtrl_GetIntValuePtrForUnifiedParameterNumber()

//---------------------------------------------------------------------------
double* RigCtrl_GetDblValuePtrForUnifiedParameterNumber( T_RigCtrlInstance *pRC, int iUnifiedPN )
{
  T_RigCtrl_ParamInfo *pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
  if( pPI != NULL )
   { if( (pPI->iByteOffsetIntoRigCtrlInstance > 0 )
      && (pPI->iByteOffsetIntoRigCtrlInstance < sizeof(T_RigCtrlInstance) ) )
      { if( pPI->iRigCtrlDataType == RIGCTRL_DT_DOUBLE )
         { return (double*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance );
         }
      }
   }
  return NULL;  // no luck .. also here, the caller MUST check for NULL-pointers !
} // end RigCtrl_GetDblValuePtrForUnifiedParameterNumber()


//---------------------------------------------------------------------------
BOOL RigCtrl_SetParamValue_Int( T_RigCtrlInstance *pRC, T_RigCtrl_ParamInfo *pPI, long i32Value )
  // Only stores the new value in the T_RigCtrlInstance,
  //       but does automatically send a CI-V "Set"-command to the locally connected radio.
  //       See implementation of CwNet_ExecuteCmd() -> RigCtrl_OnParamReport_Integer() ...
  // Return value :  TRUE when "modified", FALSE = i32Value was still the same.
{
  BOOL fModified = FALSE;
  if( (pRC!=NULL) && (pPI!=NULL) )
   { if( (pPI->iByteOffsetIntoRigCtrlInstance > 0 )
      && (pPI->iByteOffsetIntoRigCtrlInstance < sizeof(T_RigCtrlInstance) ) )
      { switch( pPI->iRigCtrlDataType )
         { case RIGCTRL_DT_INT :
              fModified = (*(int*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance ) != i32Value);
              *(int*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance ) = i32Value;
              break;
           case RIGCTRL_DT_DOUBLE :
              fModified = (*(double*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance ) != (double)i32Value);
              *(double*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance ) = (double)i32Value;
              break;
           default:
              break;
         }
      }
   }
  return fModified;
} // end RigCtrl_SetParamValue_Int()

//---------------------------------------------------------------------------
BOOL RigCtrl_SetParamValue_Double( T_RigCtrlInstance *pRC, T_RigCtrl_ParamInfo *pPI, double dblValue )
  // Only stores the new value in the T_RigCtrlInstance,
  //       but does automatically send a CI-V "Set"-command to the locally connected radio.
  //       See implementation of CwNet_ExecuteCmd() -> RigCtrl_OnParamReport_Double() ...
  // Return value :  TRUE when "modified", FALSE = i32Value was still the same.
{
  BOOL fModified = FALSE;
  if( (pRC!=NULL) && (pPI!=NULL) )
   { if( (pPI->iByteOffsetIntoRigCtrlInstance > 0 )
      && (pPI->iByteOffsetIntoRigCtrlInstance < sizeof(T_RigCtrlInstance) ) )
      { switch( pPI->iRigCtrlDataType )
         { case RIGCTRL_DT_INT :
              fModified = (*(int*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance ) != (int)dblValue);
              *(int*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance ) = (int)dblValue;
              break;
           case RIGCTRL_DT_DOUBLE :
              fModified = (*(double*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance ) != dblValue);
              *(double*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance ) = dblValue;
              break;
           default:
              break;
         }
      }
   }
  return fModified;
} // end RigCtrl_SetParamValue_Double()

//---------------------------------------------------------------------------
BOOL RigCtrl_HaveValidParamValue(T_RigCtrlInstance *pRC, T_RigCtrl_ParamInfo *pPI)
{
  int   iValue;
  double dblValue;
  if( pPI != NULL )
   { switch ( pPI->iRigCtrlDataType )
      { case RIGCTRL_DT_INT  :
           iValue = *(int*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance );
           return ( iValue!=RIGCTL_NOVALUE_INT );
        case RIGCTRL_DT_DOUBLE:
           dblValue = *(double*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance );
           return ( dblValue!=RIGCTL_NOVALUE_DOUBLE );
        default:
           break;
      } // end switch ( pPI->iRigCtrlDataType )
   }
  return FALSE;
} // end RigCtrl_HaveValidParamValue()


//---------------------------------------------------------------------------
BOOL RigCtrl_GetParamValueAString(T_RigCtrlInstance *pRC, T_RigCtrl_ParamInfo *pPI, char *pszDest, int iMaxLen )
{
  int   iValue    = RIGCTL_NOVALUE_INT;
  double dblValue = RIGCTL_NOVALUE_DOUBLE;
  const char *pszEndstop = pszDest + iMaxLen - 1;
  const char *pszToken;
  if( pPI != NULL )
   { switch ( pPI->iRigCtrlDataType )
      { case RIGCTRL_DT_INT  :
           iValue = *(int*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance );
           break;
        case RIGCTRL_DT_DOUBLE:
           dblValue = *(double*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance );
           break;
        default:
           break;
      } // end switch ( pPI->iRigCtrlDataType )
     if( iValue!=RIGCTL_NOVALUE_INT )
      { // Got a parameter with a valid value -> look it up in the string table..
        if( pPI->pStringValueTable != NULL ) // .. IF there is a string table ..
         { if( (pszToken=SL_GetStringFromTokenList(pPI->pStringValueTable,iValue)) != NULL )
            { SL_AppendString( &pszDest, pszEndstop, pszToken );
              return TRUE;
            }
         }
        // Without a string table, convert into a decimal string ...
        //    ... unless it's one of the following, "very special" parameters:
        switch( pPI->iUnifiedPN )
         { case RIGCTRL_PN_TRANSCEIVER_ID :
              if( (pszToken=RigCtrl_DefaultAddressToString( pRC->iRadioCtrlProtocol, iValue )) != NULL )
               { SL_AppendPrintf( &pszDest, pszEndstop, "0x%02X (%s)",(int)iValue, pszToken );
                 return TRUE;
               }
              break;
           default:
              break;
         }
        SL_AppendPrintf( &pszDest, pszEndstop, "%d",(int)iValue );
        return TRUE;
      }
     else if( dblValue!=RIGCTL_NOVALUE_DOUBLE )
      { SL_AppendPrintf( &pszDest, pszEndstop, "%.1lf",(double)dblValue );
        return TRUE;
      }
   }
  return FALSE;
} // end RigCtrl_GetParamValueAString()


//---------------------------------------------------------------------------
long RigCtrl_GetParamByPN_Int( T_RigCtrlInstance *pRC, int iUnifiedPN )
{
  T_RigCtrl_ParamInfo *pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
  long i32Result;
  if( (pRC!=NULL) && (pPI!=NULL) )
   { if( (pPI->iByteOffsetIntoRigCtrlInstance > 0 )
      && (pPI->iByteOffsetIntoRigCtrlInstance < sizeof(T_RigCtrlInstance) ) )
      { if( pPI->iRigCtrlDataType == RIGCTRL_DT_DOUBLE )
         { return (int)(*( (double*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance ) ) );
         }
        if( pPI->iRigCtrlDataType == RIGCTRL_DT_INT )
         { i32Result = *( (int*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance ) );
           // A few parameters need special treatment "on read" :
           switch( iUnifiedPN )
            { case RIGCTRL_PN_TRANSMITTING: // -> 0=receiving, 1=transmitting ..
                 // We don't want to rely on CI-V command "cmd 0x17, subcommand 0x00"
                 // because older radios, and non-Icom radios, may not support
                 // polling the PTT (aka 'transmit / receive state') via CAT at all.
                 // Also, the Remote CW Keyer often uses traditional keying via
                 // modem control line without CAT or CI-V at all. In that case,
                 // the keyer thread calls RigCtrl_SetLocalPTTFlag() to inform
                 // the 'Rig Control' module about the current transmitter state.
                 // Thus, if the "local PTT flag" indicates TRANSMITTING, ignore
                 // the value read via pointer from pRC->iTransmitting (above):
                 if( (pRC->fLocalPTTFlag) || (i32Result==RIGCTL_NOVALUE_INT ) )
                  { // the "local PTT flag" indicates we're TRANSMITTING,
                    // or pRC->iTransmitting is not valid (not read/reported BY THE RADIO yet):
                    i32Result = pRC->fLocalPTTFlag;
                  }
                 break; // end case RIGCTRL_PN_TRANSMITTING (read-access for e.g. the Hamlib server)

              default: // no special treatment for this parameter ON READ :
                 break;
            }
           return i32Result;
         }
      }
   }
  return 0;
} // end RigCtrl_GetParamByPN_Int()

//---------------------------------------------------------------------------
void RigCtrl_SetParamByPN_Int( T_RigCtrlInstance *pRC, int iUnifiedPN, long i32Value)
{ RigCtrl_SetParamValue_Int( pRC,
    RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN), i32Value );
}


//---------------------------------------------------------------------------
double RigCtrl_GetParamByPN_Double( T_RigCtrlInstance *pRC, int iUnifiedPN )
{
  T_RigCtrl_ParamInfo *pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
  if( (pRC!=NULL) && (pPI!=NULL) )
   { if( (pPI->iByteOffsetIntoRigCtrlInstance > 0 )
      && (pPI->iByteOffsetIntoRigCtrlInstance < sizeof(T_RigCtrlInstance) ) )
      { if( pPI->iRigCtrlDataType == RIGCTRL_DT_DOUBLE )
         { return *( (double*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance ) );
         }
        if( pPI->iRigCtrlDataType == RIGCTRL_DT_INT )
         { return (double)(*( (int*)( (BYTE*)pRC + pPI->iByteOffsetIntoRigCtrlInstance ) ) );
         }
      }
   }
  return 0.0;
} // end RigCtrl_GetParamByPN_Double()

//---------------------------------------------------------------------------
void RigCtrl_SetParamByPN_Double( T_RigCtrlInstance *pRC, int iUnifiedPN, double dblValue)
{ RigCtrl_SetParamValue_Double( pRC,
    RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN), dblValue );
}

//---------------------------------------------------------------------------
BOOL RigCtrl_AddParameterNumberToTxQueue(T_RigCtrlInstance *pRC, int iUnifiedPN) // .. for transmission via CwNet.c !
  // Adds a new item to the queue for all 'Unified parameter numbers' that still
  // need to be sent from CLIENT to SERVER, e.g. using Remote_CW_Keyer/CwNet.c .
  //
  // [in] iUnifiedPN : e.g. RIGCTRL_PN_OP_MODE, after the operator (GUI user)
  //                   switched between CW, USB, LSB, or switched the bandwidth
  //                   between CW, CWN (narrow), CWNN (very narrow filter).
  // [out] pRC->iTxParameterQueue[ pRC->iTxParameterQueueHeadIndex++ ]
  //       ,---------------------------------------------'
  //       '--> circular FIFO index, runs from 0 to RIGCTRL_TX_PARAMETER_QUEUE_LENGTH-1
  // [in]  pRC->iTxParameterQueueTailIndex : Used by CwNet.c; incremented when
  //                   a new parameter (including parameter NUMBER and VALUE)
  //                   is on its way to the 'other side' via CwNet-TCP/IP-socket.
  //       As in a classic, lock free circular FIFO, ...
  //         * HeadIndex == TailIndex means the FIFO (here: "Queue") is EMPTY,
  //         * HeadIndex+1 (wrapped) == TailIndex means 'completely full' .
  // Details:
  //     If there is sufficient space in the outbound network buffer, CwNet_OnPoll()
  //     will compare the queue's HEAD index (iTxParameterQueueHeadIndex) with its own,
  //     "per-client" TAIL index, and if they are different, append another "SET xyz"-command,
  //     along with the CURRENT VALUE retrieved via 'unified parameter number' in this queue:
{ int iNewHeadIndex = ( pRC->iTxParameterQueueHeadIndex+1 ) % RIGCTRL_TX_PARAMETER_QUEUE_LENGTH;
  if( iNewHeadIndex == pRC->iTxParameterQueueTailIndex )
   { return FALSE; // error : the queue with to-be-sent parameter numbers is already full !
   }

  // ToDo: Check if the to-be-sent 'unified parameter number' is already in the queue.
  //       If it is, don't append it to the queue, because CwNet_OnPoll()
  //       can only send THE NEWEST VALUE, e.g. pRC->iOpMode for RIGCTRL_PN_OP_MODE.

  pRC->iTxParameterQueue[pRC->iTxParameterQueueHeadIndex] = iUnifiedPN;

  pRC->iTxParameterQueueHeadIndex = iNewHeadIndex; // make the new item available to "readers in other threads"
  return TRUE;

} // end RigCtrl_AddParameterNumberToTxQueue()


//---------------------------------------------------------------------------
int RigCtrl_EnumerateUnifiedParameters( int iPreviousUnifiedPN )
  // To retrieve a list of ALL "unified parameter numbers" (which are NOT contiguous),
  // begin with iPreviousUnifiedPN = 0.  Each new call will then provide
  // a new unified PN. When all PNs are through, returns iUnifiedPN = 0 .
{
  T_RigCtrl_ParamInfo *pPI = (T_RigCtrl_ParamInfo*)RigCtrl_ParameterInfo;
  if( iPreviousUnifiedPN <= 0 )
   {  return pPI->iUnifiedPN;
   }
  while( pPI->iUnifiedPN != 0 )
   { if( pPI->iUnifiedPN == iPreviousUnifiedPN )
      { return pPI[1].iUnifiedPN;
      }
     ++pPI;
   }
  // Arrived here ? Looks like all known "unified parameter numbers" are through.
  return 0;
} // end RigCtrl_EnumerateUnifiedParameters()


//--------------------------------------------------------------------------
long RigCtrl_GetCombinedCmdAndSubcodeForCIV( BYTE *pbMessage, int iMsgLength )
  // Format of the RETURN VALUE (32-bit integer)
  //    Bit 31..24  |  23..16     |  15..0
  //   -------------+-------------+--------------------
  //        "Cmd."  | "Sub Cmd.", | "Sub Cmd.",   <-------------,
  //                |   1st byte  | up to FOUR extra digits     |
  //     ,------------------------------------------------------'
  //     '-> These are the names you will find in Icom' "CI-V REFERENCE GUIDE",
  //         "Command Table". See examples for an IC-9700 quoted from Icom's
  //         document number A7508-3EX-4 further below.
{
  long i32Result = (long)pbMessage[4] << 24; // result if it's only a "Cmd" without "Sub cmd"
  long i32FourDigitSubindex; // e.g. the stuff after 0x1A 0x05 (four-digit BCD)

  if( iMsgLength <= 5 ) // preamble(2), addr (2), postamble(1) : that's FIVE BYTES, no command in here
   { return -2;         // will never match anything in pRC->i32ExpectedResponse_CmdAndSubcode
   }
  if( iMsgLength == 6 ) // preamble(2), addr (2), command byte (1), postamble(1) but no SUBCODE..
   { return i32Result;
     // In this case, it's obvious that the command has no sub-command (subcode),
     //     e.g. cmd = 0x05 = "get operating frequency"
   }
  else // iMsgLength > 6 : This is tricky. The byte after the command (pbMessage[4])
   { // may be a sub-command (subcode), but it MAY be a variable ranging from 0 to 255 or even more !
     // To find out if it's a TWO-BYTE command (with "Cmd" and "Sub cmd." in IC-7300 FM page 19-3),
     // branch by the first byte. Not bullet-proof yet, because what Icom decided
     // to call 'Sub cmd.' in the "Command table" of the IC-9700 CI-V manual,
     // isn't always ONE BYTE, but may be MULTIPLE BYTES. Examples from A7508-3EX-4:
     //   "Cmd."  |  "Sub cmd."  | "Data"       | "Description" / comment by DL4YHF
     // ----------+--------------+--------------+-----------------------------------
     //    0x04   |              | See p. 13    | Read the operating mode
     //    0x07   |   (none??)   |              | Select the VFO mode
     //    0x07   | 0x00         |              | Select VFO A (in satellite mode,...)
     //    0x07   | 0x01         |              | Select VFO B (not in satellite mode)
     //    0x07   | 0xA0         |              | Equalize VFO A and VFO B
     //    0x07   | 0xD0         |              | Select the main band
     //    0x07   | 0xD1         |              | Select the sub band
     //    0x1A   | 0x00         | See pp 14,15 | Send/read memory contents
     //    0x1A   | 0x01         | See p. 15    | Send/read band stacking registers
     //    0x1A   | 0x05 0001    | See p. 16    | SET > Tone Control/TBW > RX > ... HPF/LPF
     //    0x1A   | 0x05 0002    | 00  to 10    | SET > Tone Control/TBW > RX > ... Bass ..
     //    0x1A   | 0x05 0003    | 00  to 10    | SET > Tone Control/TBW > RX > ... Treble ..
     //    0x1A   | 0x05 ......  |              | Several HUNDREDS of other parameters
     //    0x1A   | 0x05 0350    | See p. 17    | SCOPE > Fixed Edges > 1200M > No.4
     //             \__________/
     // ,--------------'
     // '--> To find THE LENGTH of the "Sub cmd.", first look at the
     //      "Cmd." and the FIRST BYTE of the "Sub cmd." .
     //      Then, depending on that, decide what else belongs to the
     //      "Sub cmd." a la IC-9700 / document number A7508-3EX-4 .

     switch( pbMessage[4] ) // check for all known TWO-BYTE commands :
      { case 0x07 : // "Select" something (Subcmd:0=VFO A, 1=VFO B, 0xA0=Equalize A+B, 0xB0=Exchange A+B)
        case 0x0E : // Subcmd: 0=scan stop, 1=scan start, 2=prog scan, 3=F scan, .. 0xB1="set as select channel"
        case 0x10 : // Subcmd: 0=tuning step off, 1="send/read the 100 Hz tuning step" (?!?!)
        case 0x13 : // "Speech" with subcmds 0,1,2
        case 0x14 : // read/set various "Levels" (AF, RF, Squelch)
        case 0x15 : // read S-meter, squelch status, power meter, SWR meter, ...
        case 0x16 : // preamp, AGC, noise blanker, noise reduction, ...
        case 0x19 : // so far, only ONE subcode but Icom say it HAS a subcode
  //No: case 0x1A : // "Set/read" a horrible number of differently addressed items (with sub-sub-commands)
        case 0x1B : // repeater tone frequency, TSQL tone frequency, ...
        case 0x1C : // transceiver status, ATU, XFC, ...
        case 0x1E : // number of TX frequency bands, TX frequency band edge frequencies, ...
        case 0x21 : // RIT frequ, RIT enable, XIT
        case 0x25 : // "selected" and "unselected" VFO freq. // main/sub VFO freq. (IC-7610)
        case 0x26 : // "selected" and "unselected" operating mode, filter
        case 0x27 : // scope waveform data, scope settings, level, .....
           i32Result |= ( (unsigned long)pbMessage[5] << 16);
           return i32Result;

        case 0x1A : // "Set/read" a horrible number of differently addressed items (with sub-sub-commands)
           // This horrible monster sometimes has a SINGLE-BYTE "Sub cmd.",
           // but very often (with the first sub-cmd-byte = 0x05)
           // followed by a FOUR-DIGIT "sub-index" (no name given by Icom)
           // as shown in the example quoted from A7508-3EX-4 further above.
           i32Result |= ( (unsigned long)pbMessage[5] << 16);

           // Guesswork: 0x1A 0x00 to 0x1A 0x04 are "older" commands,
           // introduced before Icom realized they would run out of "Sub cmd."-values
           // .. thus we end up with the mess that SOME "0x1A"-commands
           //    have a single byte in sub-command, while others have THREE.
           if( ( pbMessage[5] >= 0x05 )
             &&( iMsgLength >= 9 ) ) // at preamble(2), addr (2), command byte (1),
            { // THREE-BYTE "Sub cmd." plus postamble(1) ... that's 2+2+1+3+1 = 9 bytes
              i32Result |= ( (unsigned long)pbMessage[6] << 8) | pbMessage[7]; // omg !
            }
           return i32Result;

        case 0x06 : // definitely has NO subcode ("Sub cmd") - not even in an IC-9700 !
        default   : // Guess it's a single-byte command WITH DATA(!)
           return i32Result;
      }
   }
} // end RigCtrl_GetCombinedCmdAndSubcodeForCIV()

//--------------------------------------------------------------------------
int RigCtrl_ParseCIV( T_RigCtrlInstance *pRC,
          int  iRigCtrlPort, // [in] RIGCTRL_PORT_RADIO / RIGCTRL_PORT_CLIENT_1 / 2 / .. ?
          BYTE *pbMessage, int iMsgLength ) // [in] message data, message length in bytes
  // Called upon COMPLETE reception of any CI-V message, via 'COM' (USB) or LAN/WLAN (UDP),
  //    including those messages received on CLIENT PORTS (see "Callers" further below).
  // ALSO CALLED ON RECEPTION OF CI-V MESSAGES ON ONE OF THE "DUPLICATOR"-PORTS !
  //    Most COMMANDS(!) received on duplicator ports will be answered
  //    immediately (if we know the answer, to avoid flooding the radio with
  //    requests from all over the place). Others (especially those we don't
  //    understand) will be passed on from here to the "radio port",
  //    and the response relayed back to the originating port.
  //
  // Returns RIGCTRL_MSGTYPE_xyz, e.g. RIGCTRL_MSGTYPE_FREQUENCY_REPORT when successfull,
  //   otherwise RIGCTRL_MSGTYPE_UNKNOWN (i.e. "not recognized") .
  // In Spectrum Lab, anything that arrives HERE MAY be displayed usually
  //    with a CYAN background under 'View/Windows' / 'CAT traffic monitor'.
  //    Certain messages, like error responses, will have a different colour.
  //
  // Callers:
  //   (a) IO_RadioCommThdFunc( for any kind of "COM" port connected to the RADIO)
  //          -> RigCtrl_ProcessRxData(with iRigCtrlPort==RIGCTRL_PORT_RADIO) -> RigCtrl_ParseCIV()
  //   (b) IO_SerialPortDuplicatorThread( for "external clients" via "COM" port, e.g. via "com0com" )
  //          -> RigCtrl_ProcessRxData(with iRigCtrlPort==RIGCTRL_PORT_RADIO) -> RigCtrl_ParseCIV()
{
  int i, iSubcode, iUnifiedPN, iNewSetting, iRange, iEdge;
  long i32CombinedCmdAndSubcode; // comparable with pRC->i32ExpectedResponse_CmdAndSubcode
  T_RigCtrl_ParamInfo *pPI;  // "parameter info" from a lookup table ..
  T_RigCtrl_ParamInfo *pParamInfoFromExpectedUnifiedPN = NULL; // " " " after RigCtrl_AskForUnifiedPN()

  double dblConvFactor = 1.0;
  int iMsgType = RIGCTRL_MSGTYPE_UNKNOWN; // for SL's CAT traffic monitor, to selectively reject certain message types in the list
  int iClientAction = RIGCTRL_CLIENT_ACTION_RETRANSMIT; // RIGCTRL_CLIENT_ACTION_NONE / RIGCTRL_CLIENT_ACTION_ENTER_FIFO ?
  BYTE b;
  long i32Value = 0;
  double d, fact;        // for conversion from BCD to floating point
  char  szComment[ 256 ]; // for the human-readable "traffic log" ...
  char  *pszCmtDest = szComment; // incremented destination for string building functions
  char  *cpCmtEndstop = szComment+sizeof(szComment)-1; // "endstop" for string building functions
  char  sz80Temp[84];
  const char *cpToken = "";
  char  *cpValueAsString = NULL;
  T_RigCtrl_PortInstance *pPort = &pRC->PortInstance[ iRigCtrlPort/*client-port*/ ];
  T_RigCtrl_RadioInfo *pRadioInfo;
  T_RigCtrl_MsgFilter msg_filter;
  T_RigCtrlFrequencyRange *pFreqRange;
  T_RigCtrlFreqMemEntry   *pMemEntry;

  BOOL fShowInLog = (RigCtrl_iTrafficMonitorDisplayOptions & RIGCTRL_TRAFFIC_DISPLAY_OPTION_ENABLE) != 0;
  BOOL fIsEcho = FALSE;
  BOOL fIsExpectedResponse = FALSE;   // may be difficult to find out !
  BOOL fIsUnsolicitedMsg = FALSE; // initial guess : it's a REPONSE, not an unsolicited REPORT (new VFO frequ, Spectrum scope data)

  szComment[0] = '\0';

  // The TWO FIRST BYTES (0xFE = "Preamble") have already been checked by the
  //  caller (e.g. IO_HandleRcvdDataFromRadio()), so no need to check again.
  // Within RigCtrl_ParseCIV(), don't care for pbMessage[2] = transceiver's address (e.g. 0x48 = IC706)
  // and  for pbMessage[3] = controller's address (often 0xE0), EXCEPT FOR THIS:
  // If the rig is configured for "CI-V Echo" *ON* (alias "CI-V USB Echo Back" in an IC-7300)
  //  anything we send out (as command) returns as an echo, like this:
  //            pbMessage[0 1  2  3  4  5  6  7  8  9  10]
  // > 01:26:22.2 TX 006 FE FE 94 E0 03 FD                 (VFO freq) ; read VFO freq
  // > 01:26:23.2 RX 006 FE FE 94 E0 03 FD                 (VFO freq) ; echo
  // > 01:26:23.2 RX 00B FE FE E0 94 03 00 70 35 05 00 FD  (VFO freq) ; 5357000.0 Hz
  //   ,-------------'-' |___| |  |  |
  //   |          ,--------'   |  |  |
  //   '--length  '--preamble  |  |  |
  //       "to address" -------'  |  '--- "cmd" / i32ExpectedResponse_CmdAndSubcode
  //       "from address" --------'
  // Unfortunately, "HamLib" (and thus WSJT-X and FTDX) doesn't work
  // without this bandwidth-wasting stupidity. HERE (in RigControl.c)
  // don't confuse the ECHO with the EXPECTED RESPONSE :
  if( ( pbMessage[2] == pRC->iRadioDeviceAddr )   // message sent TO THE RADIO ?
    &&( pbMessage[3] == pRC->iRadioMasterAddr ) ) // message sent FROM THE MASTER ?
   { // If it's not from one of the 3rd party "clients", this message is an ECHO:
     if( iRigCtrlPort == RIGCTRL_PORT_RADIO )
      { // That's the same as in our "TX"-message (see CAT traffic dump above), so..
        fIsEcho = TRUE;  // .. so it's an ECHO, not the real response message !
      }
     else // guess a 3rd party application (like RS-BA1) is talking to us.
      {   // Example : First message from RS-BA1 = FE FE 94 E0 18 01 FD (7 bytes)
      }
   }

  // Prepare a new filter that will later be passed to RigCtrl_RegisterCIVReponseForClient():
  msg_filter.iCmd = pbMessage[4];
  // per default, any further bytes in any CI-V message are IGNORED for filtering:
  msg_filter.iSubCmd = msg_filter.iDataB1 = msg_filter.iDataB2 = -1;
  // per default, let the filter (for the client) pass reponse "OK" and "NOT OK"
  //   (exceptions will be made in the monster case-list further below) :
  msg_filter.iMsgTypes = RIGCTRL_MSGTYPE_OK | RIGCTRL_MSGTYPE_NOT_OK;

  i32CombinedCmdAndSubcode = RigCtrl_GetCombinedCmdAndSubcodeForCIV( pbMessage, iMsgLength );
  pPI = RigCtrl_GetInfoForCIVCommandAndSubcode( i32CombinedCmdAndSubcode );
  // '--> May be NULL if a response message (CI-V) does not contain all required info, like ...
  //  * no "Cmd." and "Sub cmd." (<- those are the column titles in Icom's
  //       Command tables, e.g. for the IC-9700 in document nr A7508-3EX-4 )
  //  * messages with pbMessage[4] == 0xFA = "No-Good" aka "FAilure" aka "NotOK"
  //  * messages with pbMessage[4] == 0xFB = "OK" (success) aka "Fine Business" from the rig
  //
  //  For the above cases, T_RigCtrlInstance must remember what has been asked for,
  //   instead of extracting everything from the received message itself .
  if( (pPI != NULL) && (pPI->pszToken != NULL) )
   { cpToken = pPI->pszToken; // -> e.g. "NumTxBands" for cmd=0x1E, subcmd=0x00
   }

  if( ( ! fIsEcho ) && (iRigCtrlPort==RIGCTRL_PORT_RADIO) )
   { if ( i32CombinedCmdAndSubcode == pRC->i32ExpectedResponse_CmdAndSubcode )
      { // This MAY BE (!) what RigCtrl_Handler() has been "waiting for", but:
        //  Don't confuse the ECHO with the EXPECTED RESPONSE :
        if( ! fIsEcho )  // only if this is NOT an "echo" :
         { fIsExpectedResponse = TRUE; // ok for simple cases without the need for a matching SUB-COMMAND (more later)
           // Example: read S-Meter level;
           //  CI-V  :     |pre| |adr| cmd sub  data  postamble
           //   request  = FE FE 00 00  15 02         FD
           //   response = FE FE 00 A2  15 02   00 02 FD
           // ,-------------------------|___|
           // '--> combined into i32ExpectedResponse_CmdAndSubcode : 0x1502
         }
        // Note: May have to use 'i32CombinedCmdAndSubcode' AFTER the monster-case-list
        //       further below, as the NEW "expected response" when a command
        //       retrieved from any CLIENT PORT is forwarded to the radio !
      }
     else // i32CombinedCmdAndSubcode != pRC->i32ExpectedResponse_CmdAndSubcode ->
      { // For example, after sending 0x03 ("read operating frequency")
        // an IC-7300 answered with   0xFB ("OK", but without data).
      } // end else < i32CombinedCmdAndSubcode != pRC->i32ExpectedResponse_CmdAndSubcode >
        // In other cases, we may be waiting for the response after RigCtrl_AskForUnifiedPN() :
     if( pRC->iExpectReadResponseForUnifiedPN > 0 ) // <- added 2024-07, more 'universal' replacement for i32ExpectedResponse_CmdAndSubcode..
      { pParamInfoFromExpectedUnifiedPN = RigCtrl_GetInfoForUnifiedParameterNumber(pRC->iExpectReadResponseForUnifiedPN);
        if( pParamInfoFromExpectedUnifiedPN != NULL )
         { // If we're lucky, the READ RESPONSE begins with the same "Cmd" and possibly "Sub Cmd"
           // as the READ REQUEST. In that case, finish what RigCtrl_AskForUnifiedPN() had begun:
           if( (pParamInfoFromExpectedUnifiedPN->pbCIVReadCmd != NULL)
            && (pParamInfoFromExpectedUnifiedPN->iCIVReadCmdLength > 0) )
            { if( memcmp(pParamInfoFromExpectedUnifiedPN->pbCIVReadCmd,
                         pbMessage+4 /* "Cmd" and possibly "Sub Cmd" */,
                         pParamInfoFromExpectedUnifiedPN->iCIVReadCmdLength) == 0 )
               { // Bingo; the data that may be parsed further below
                 // are the 'expected response' for RigCtrl_AskForUnifiedPN() :
                 fIsExpectedResponse = TRUE;
               }
            }
         } // if( pParamInfoFromExpectedUnifiedPN != NULL )
      }   // end if < iExpectReadResponseForUnifiedPN .. >
   }     // end if( ( ! fIsEcho ) && (iRigCtrlPort==RIGCTRL_PORT_RADIO) )
  if( ! fIsEcho ) // don't waste time trying to parse CI-V "echoes" !
   { switch( pbMessage[4] )  // [in] what Icom calls 'command'. But it may be a response..
      {
        case 0x05: // Icom CI-V code "frequency control" / "Set operating frequency"
        case 0x00: // Transfer operating freq. data (unsolicited message when turning VFO knob)
        case 0x03: // "Read operating frequency" (response on request, also cmd 0x03)
                   // 0x03 was included in WSJT-X's "polling cycle" via COM port.
           fIsUnsolicitedMsg = (pbMessage[4] == 0x00);
           iUnifiedPN = RIGCTRL_PN_FREQUENCY;
           pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
           if( pPI != NULL )
            { cpToken = pPI->pszToken;
            }
           else
            { cpToken = "Split";
            }
           i = 5;
           if( iMsgLength >= 8 )  // 2 bytes preamble + 2 byte address + 1 byte command + at least a few digits + 1 byte end-of-message
            { d = 0.0; fact=1.0;
              while(i<(iMsgLength-1))
               { b = pbMessage[i++];
                 d += fact * (double)(b & 0x0F);
                 fact *= 10.0;
                 d += fact * (double)((b>>4) & 0x0F);
                 fact *= 10.0;
               }
              // Info from an IC-905 user (where the REMOTE WATERFALL DISPLAY didn't work as well) :
              // > I noted in the 905 CI-V programming manual the command 0x00 (Frequency)
              // > now is variable length with 2 extra bytes added when on 10GHz
              // > or higher for the 10's and 100's GHz digit places.
              //    -> Compared (a) C:\datasheets\Rig_Manuals\IC-905\IC-905_CI-V_Reference_Guide.pdf (anno 2023)
              //        against (b) C:\datasheets\Rig_Manuals\IC-705\IC-705_ENG_CI-V_V4b_2021_05.pdf ...
              //   (a) page 16 : 6 bytes with 12 BCD-digits, from 1-Hz to 100-GHz digit (the latter always zero)
              //   (b) page 17 : 5 bytes with 10 BCD-digits, from 1-Hz to 1-GHz digit (the latter always zero)
              SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s: %.1lf Hz", cpToken, (double)d );
              // Since THIS CODE runs in a worker thread,
              // the rest (updating the GUI) happens somewhere else.
              //
              if( ! ( pRC->iPostponedSetMessages & (1<<RIGCTRL_MSGTYPE_SET_FREQUENCY) ) )
               { // Only accept the radio's reported value if we're not going to SEND a NEW frequency soon:
                 if( d != pRC->dblVfoFrequency )
                  { ++pRC->iFrequencyModifiedByRadio_cnt;
                    // This counter is polled in a GUI-thread somewhere.
                  }
                 pRC->dblVfoFrequency = d;
               } // end if < no postponed SET-command for the frequency pending >
              else // the radio has reported its VFO frequency, but a POSTPONED(!)
               {   // request to set a NEW (different) frequency is still pending.
                 // In this case, do NOT modify pRC->dblVfoFrequency now.
                 // This may cause trouble for clients that will read the VFO
                 // frequency through other means (with CI-V, there are many roads leading to Rome).
               }
            }
           iMsgType = RIGCTRL_MSGTYPE_FREQUENCY_REPORT; // end case 0x00, 0x03, 0x05
           break;

        case 0x01: // "Send mode data" (transceive)
        case 0x04: // "Read operating mode" (response after RigCtrl_AskForOperatingMode(?) )
                   //   (see also: case 0x26, similar purpose, other syntax)
                   // Cmd 0x04 is included in WSJT-X's "polling cycle",
                   //      but WSJT-X doesn't try to CHANGE the mode.
           iMsgType = RIGCTRL_MSGTYPE_OP_MODE;
           iUnifiedPN = RIGCTRL_PN_OP_MODE;
           pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
           if( pPI != NULL )
            { cpToken = pPI->pszToken;
            }
           else
            { cpToken = "OpMode";
            }
           iNewSetting = pRC->iOpMode;
           // IC-7300, IC-7610 : "Operating Mode" with two data bytes.. IC-7300 used iMsgLength=8
           if( iMsgLength >= 7 )  // 2 bytes preamble + 2 byte address + 1 byte command + at least one 'mode' byte + 1 byte end-of-message
            { iNewSetting = RigCtrl_IcomToRigCtrlOpmode( pbMessage[5] );
              SL_AppendString( &pszCmtDest, cpCmtEndstop, RigCtrl_OperatingModeToString(iNewSetting) );
            }
           if( iMsgLength >= 8 )  // "Operating mode" byte followed by optional "Filter setting" ?
            { switch( pbMessage[6] )
               { case 1 : break;  // "FIL1" (normal filter bandwidth)
                 case 2 : iNewSetting |= RIGCTRL_OPMODE_NARROW;      break; // "FIL2" [IC7610CIV page 10]
                 case 3 : iNewSetting |= RIGCTRL_OPMODE_VERY_NARROW; break; // "FIL3"
                 default: break;
               }
              SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, ", filter=%d", pbMessage[6] );
            } // end if < radio also reported the current FILTER SETTING >
           // Only accept the radio's reported value if we're not going to SEND a NEW operating mode:
           if( ! ( pRC->iPostponedSetMessages & (1<<RIGCTRL_MSGTYPE_OP_MODE) ) )
            {
              if( iNewSetting != pRC->iOpMode )
               { pRC->iOpMode = iNewSetting;
               }
            }
           break; // end case 0x01 = "Send mode data", 0x04 = "Read operating mode"

        case 0x02: // "Read band edge frequencies"
           iMsgType = RIGCTRL_MSGTYPE_OTHER; // end case 0x02 = "Read band edge frequencies"
           break;

        case 0x06: // "Set operating mode"  (possibly "transceived", e.g. relayed)
           iMsgType = RIGCTRL_MSGTYPE_OTHER; // end case 0x06 = "Set operating mode"
           break;

        case 0x07: // "Select the VFO mode". Seen when "RS-BA1" took over control:
                   // pbMessage[0  1  2  3  4  5  6]
                   // > r1 007 FE FE 94 E0 07 00 FD
                   //                         |____ 00 = VFO A, 01 = VFO B, A0 = Equalize A and B, ...
           if( iMsgLength >= 7 )  // "Select VFO mode" followed by at least one DATA byte ?
            { SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "SelVFO:%02X", (int)pbMessage[5] ); // for the log..
            }
           else
            { SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "SelVFO ?" );
            }
           iMsgType = RIGCTRL_MSGTYPE_OTHER;
           break; // end case 0x07 = "Select VFO mode"

        case 0x08: // "Select the Memory mode"
           iMsgType = RIGCTRL_MSGTYPE_OTHER; // end case 0x08 = "Select memory mode"
           break;

        case 0x09: // "Memory write"
           iMsgType = RIGCTRL_MSGTYPE_OTHER; // end case 0x09 = "Memory write"
           break;

        case 0x0A: // "Memory copy to VFO"
           iMsgType = RIGCTRL_MSGTYPE_OTHER; // end case 0x0A = "Memory copy to VFO"
           break;

        case 0x0B: // "Memory clear"
           iMsgType = RIGCTRL_MSGTYPE_OTHER; // end case 0x0B = "Memory clear"
           break;

        case 0x0E: // "Scan"-related commands. Used quite early when launching "RS-BA1":
           iMsgType = RIGCTRL_MSGTYPE_OTHER;
           //      pbMessage[0  1  2  3  4  5  6]
           //      > r1 007 FE FE 94 E0 0E 00 FD (with iMsgLength=7)
           //           len                |
           //                              |_____ sub command (no data)
           //      pbMessage[0  1  2  3  4  5  6  7]
           //      > r1 008 FE FE 94 E0 0E 00 03 FD (with iMsgLength=8)
           //           len             |  |  |__ optional data, if iMsgLength > 7
           //                          cmd sub
           if( iMsgLength >= 7 ) // valid length (with or without data) ?
            { iSubcode = pbMessage[5]; //
              switch( iSubcode )  // "sub-command" after 0x0E ?
               { case 0x00 : cpToken = "Scan stop";       // all for the log..
                   iUnifiedPN = RIGCTRL_PN_SCAN_MODE;
                    break;
                 case 0x01 : cpToken = "P/M scan start"; break;
                 case 0x02 : cpToken = "P scan start";   break;
                 case 0x03 : cpToken = "F scan start";   break;
                 case 0x12 : cpToken = "FP scan start";  break;
                    // etc, etc, a lot has been omitted here
                 default:
                    sprintf( sz80Temp, "Scan[%02X]", (int)iSubcode );
                    cpToken = sz80Temp;
                    break;
               } // end switch < "sub-command" after 0x0E, *WITHOUT* data >
              pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
              if( iMsgLength >= 8 )  // subcode followed by DATA ? Must be a "set"-command..
               { i32Value = RigCtrl_ParseBCD_CIV_MSByteFirst( pbMessage+6, 2*(iMsgLength-7) );
                 SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s:%ld", cpToken, (long)i32Value ); // for the log..
               }
              else // only command and sub-command ..
               { SL_AppendString( &pszCmtDest, cpCmtEndstop, cpToken );
               }
            } // end if( iMsgLength >= 6 )
           break; // end case 0x0E = "scan" commands

        case 0x0F: // "Read Split setting"; Response data for IC-9700: 0x00="Split off", 0x01="Split", 0x11="DUP-", 0x12="DUP+", 0x13="DD Repeater Simplex (RPS)"
           iMsgType = RIGCTRL_MSGTYPE_OTHER; // end case 0x0F = "Read Split setting"
           // Received from RS-BA1 : > r1 006 FE FE 94 E0 0F FD      ; Split ?
           // Response from IC-7300: > RX 007 FE FE E0 94 0F 00 FD   ; Split:0
           //                        pbMessage[0  1  2  3  4  5  6]
           // Also included in WSJT-X's "polling cycle" .
           if( iMsgLength >= 6 )  // 0x0F with or without data.. (NO subcodes here)
            { iUnifiedPN = RIGCTRL_PN_SPLIT_MODE;
              pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
              if( pPI != NULL )
               { cpToken = pPI->pszToken;
               }
              else
               { cpToken = "Split";
               }
              if( iMsgLength >= 7 )  // 0x0F followed by DATA byte(s)..
               { i32Value = RigCtrl_ParseBCD_CIV_MSByteFirst( pbMessage+5,
                                       2*(iMsgLength/*7*/-6)/*nDigits*/ );
                 // Replace the Icom-specific BCD-values by the 'unified'
                 // symbolic values in RigControl.h :
                 switch( i32Value )
                  { case 0: i32Value = RIGCTRL_SPLIT_MODE_OFF;
                               cpValueAsString = "off";
                               break;
                    case 1: i32Value = RIGCTRL_SPLIT_MODE_ON;
                               cpValueAsString = "ON";
                               break;
                    case 11: i32Value = RIGCTRL_SPLIT_MODE_DUP_NEG;
                               cpValueAsString = "DUP-";
                               break;
                    case 12: i32Value = RIGCTRL_SPLIT_MODE_DUP_POS;
                               cpValueAsString = "DUP+";
                               break;
                    case 13: i32Value = RIGCTRL_SPLIT_MODE_REPEATER_SIMPLEX; // <- IC-9700 extravaganza ?
                               cpValueAsString = "RPTR";
                               break;
                    default: // just leave i32Value as-is
                               cpValueAsString = "???";
                               break;
                  } // end switch < code for 'Split Mode' a la ICOM (IC-9700, and maybe others) >
                 SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s:%s", cpToken, cpValueAsString );
                 RigCtrl_SetParamValue_Int( pRC, pPI, i32Value );
               }
              else  // 0x0F followed by subcode but NO DATA ...
               { SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s ?", cpToken ); // query, no "set" command
               }
            } // end if < sufficient length for COMMAND > ?
           break;  // end case 0x0F

        case 0x10: // "Send/read tuning step"
           iMsgType = RIGCTRL_MSGTYPE_OTHER; // end case 0x10 = "Send/read tuning step"
           if( iMsgLength >= 6 )  // 0x10 with or without data.. (NO subcodes here)
            { cpToken = "TuningStep";
              iUnifiedPN = RIGCTRL_PN_TUNING_STEP;
              pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
              if( iMsgLength >= 7 )  // 0x10 followed by DATA byte(s)..
               { i32Value = RigCtrl_ParseBCD_CIV_MSByteFirst( pbMessage+6,
                                       2*(iMsgLength/*7*/-6)/*nDigits*/ );
                 // 0=off, 1=100 Hz, 2=1kHz, ... 8=25kHz .
                 SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s:%ld", cpToken, (long)i32Value );
               }
              else  // 0x10 followed by subcode but NO DATA ...
               { SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s ?", cpToken ); // query, no "set" command
               }
            } // end if < sufficient length for COMMAND > ?
           break; // end case 0x10

        case 0x11: // "Send/read attenuator setting". One of dozens of parameters that "RS-BA1" wanted to know..
           iMsgType   = RIGCTRL_MSGTYPE_RF_GAIN;  // attenuator contributes to "gain" !
           iUnifiedPN = RIGCTRL_PN_ATTENUATOR_DB;
           pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
           if( pPI != NULL )
            { cpToken = pPI->pszToken;
              // From RS-BA1 to SpecLab (on a client port) :
              //           [0  1  2  3  4  5]
              //  > r1 006 FE FE 94 E0 11 FD
              // From an IC-7300 back to SpecLab (on the RADIO port):
              //           [0  1  2  3  4  5  6]
              //  > RX 007 FE FE E0 94 11 00 FD
              //                          \|___ data byte ("current setting")
              if( iMsgLength > 6 )  // 0x11 followed by at least ONE data byte..
               { i32Value = RigCtrl_ParseBCD_CIV_MSByteFirst( pbMessage+5,
                                       2*(iMsgLength/*7*/-6)/*nDigits*/ );
                 SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s:%ld", cpToken, (long)i32Value );
                 RigCtrl_SetParamValue_Int( pRC, pPI, i32Value );
               }
              else  // 0x11 followed by subcode but NO DATA ...
               { SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s ?", cpToken ); // just a query
               }
            } // end if( pPI != NULL )
           break; // end case 0x11 = "Send/read attenuator setting"

        case 0x12: // "Send/read antenna selection"
           iMsgType = RIGCTRL_MSGTYPE_OTHER; // end case 0x12 = "Send/read antenna selection"
           break; // end case 0x12

        case 0x13: // "Speech (speak) with voice synthesizer"
           iMsgType = RIGCTRL_MSGTYPE_OTHER; // end case 0x13 = "Speak" (voice synth)
           break;

        case 0x14: // "Send/read an awful lot of parameters" (AF level, RF gain, squelch, audio pitch, noise reduction, ....)
           iMsgType = RIGCTRL_MSGTYPE_OTHER; // end case 0x14 = "Send/read a lot of parameters"
           // Seen while RS-BA1 tortured the bus with lots of polling traffic :
           //  pbMessage[0  1  2  3  4  5  6  7  8]
           //  > r1 007 FE FE 94 E0 14 0B FD        (query, iMsgLength=7)
           //  > RX 009 FE FE E0 94 14 0B 01 53 FD  (response with data, iMsgLength=9)
           //                       |  |_sub-cmd
           //                      |__ cmd
           if( iMsgLength >= 7 )  // 0x14 followed by subcode, with or without DATA ...
            { iSubcode = pbMessage[5];
              dblConvFactor = 100.0 / 255.0;  // conversion factor for most "percentages", instead of the strange CI-V internal unit,
                  // but BEWARE: CI-V is incredibly counter intuitive !
                  // A few deterring examples:
                  //   * Cmd=0x14, SubCmd=0x01="AF level" : 0000=min=0 %, 0255=max=100 %
                  //   * Cmd=0x15, SubCmd=0x11="PO meter level": 0000=0%, 0143=50%, 0213=100% .. that's NON-LINEAR !
                  //   * Cmd=0x15, SubCmd=0x13=ALC meter level :0000=Min, 0120=Max !
                  //   * Cmd=0x15, SubCmd=0x14=Comp meter level:0000=0 dB, 0130=15 dB, 0241=30 dB.. that's NON-LINEAR !
                  //   * Cmd=0x15, SubCmd=0x15=Supply/PA Voltage: 00=0 V, 0013=10 V, 0241=16 V .. that's incredibly non-linear
                  //   * Cmd=0x15, SubCmd=0x16=Drain Current:   0000=0 A, 0097=10 A, 0146=15 A, 0241=25 A .. almost linear
                  //
              switch( iSubcode )  // decode "sub-command" after 0x14 for the log:
               { case 0x01 : iUnifiedPN = RIGCTRL_PN_AUDIO_VOLUME_PERCENT; break;
                 case 0x02 : iUnifiedPN = RIGCTRL_PN_RF_GAIN_PERCENT;      break;
                 case 0x03 : iUnifiedPN = RIGCTRL_PN_SQUELCH_LEVEL_PERCENT; break;
                 case 0x04 : break;
                 case 0x05 : break; // IC-7610 : "APF position", not used here
                 case 0x06 : iUnifiedPN = RIGCTRL_PN_NOISE_REDUCTION_PERCENT; break;
                 case 0x07 : iUnifiedPN = RIGCTRL_PN_PASSBAND_TUNING_POS1; break;
                 case 0x08 : iUnifiedPN = RIGCTRL_PN_PASSBAND_TUNING_POS2; break;
                 case 0x09 : iUnifiedPN = RIGCTRL_PN_CW_PITCH_HZ;
                    dblConvFactor = 900.0/*Hz*/ / 255.0; // here: in Hertz, not the strange CI-V internal unit
                    break;
                 case 0x0A : iUnifiedPN = RIGCTRL_PN_RF_POWER_SETTING_PERCENT; break;
                 case 0x0B : iUnifiedPN = RIGCTRL_PN_MIC_GAIN_PERCENT;     break;
                 case 0x0C : iUnifiedPN = RIGCTRL_PN_KEYER_SPEED_PERCENT;  break;
                 case 0x0D : iUnifiedPN = RIGCTRL_PN_NOTCH_POS_PERCENT;    break;
                 case 0x0E : iUnifiedPN = RIGCTRL_PN_COMP_SETTING_PERCENT; break;
                 case 0x0F : iUnifiedPN = RIGCTRL_PN_BREAK_IN_DELAY_PERCENT; break;
                 case 0x12 : iUnifiedPN = RIGCTRL_PN_NBL_SETTING_PERCENT;  break;
                 case 0x15 : iUnifiedPN = RIGCTRL_PN_MONITOR_GAIN_PERCENT; break;
                 case 0x16 : iUnifiedPN = RIGCTRL_PN_VOX_GAIN_PERCENT;     break;
                 case 0x17 : iUnifiedPN = RIGCTRL_PN_ANTI_VOX_GAIN_PERCENT; break;
                 case 0x18 : break;
                 case 0x19 : iUnifiedPN = RIGCTRL_PN_BRIGHTNESS_PERCENT;   break;
                 default   : break;
               } // end switch < "sub-command" after 0x14 >
              pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
              if( pPI != NULL )
               { cpToken = pPI->pszToken;
               }
              else
               { sprintf(sz80Temp, "0x14[%02X]", (int)iSubcode );
                 cpToken = sz80Temp;
               }
              if( iMsgLength >= 9 )  // 0x14+subcode followed by DATA bytes. Example:
               { // pbMessage[0  1  2  3  4  5  6  7  8]
                 // > RX 009 FE FE E0 94 15 12 00 80 FD ; SWRMeterLvl
                 //       |              |  |  |___|-- value in some phantasy unit,
                 //   iMsgLength=9      cmd subcode       usually 4 digit BCD, MSDigit first
                 i32Value = RigCtrl_ParseBCD_CIV_MSByteFirst( pbMessage+6,
                                       2*(iMsgLength/*9*/-7)/*nDigits*/ );
                 SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s:%ld", cpToken, (long)i32Value ); // "set" command with value in some phantasy unit
                 RigCtrl_SetParamValue_Double( pRC, pPI, (double)i32Value * dblConvFactor );
               }
              else  // 0x15 followed by subcode but NO DATA ...
               { SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s ?", cpToken ); // query, no "set" command
               }
            } // end if < sufficient length for COMMAND + SUB-COMMAND > ?
           break; // end case 0x14

        case 0x15: // "Send/read noise or S-meter squelch status", etc etc
           iMsgType = RIGCTRL_MSGTYPE_OTHER;
           //                                     pbMessage[0  1  2  3  4  5  6]
           // Seen when RS-BA1 took over control: > r1 007 FE FE 94 E0 15 01 FD
           //                                                           |  |_sub-cmd
           //                                                           |__ cmd
           // Note: The radio does NOT send any of these messages on its own
           //       - we must ASK FOR them, e.g. the current "S meter" value
           //         as Spectrum Lab periodically does (not very often) by
           //         calling RigCtrl_SendCmdToReadUnifiedPN(RIGCTRL_PN_S_METER_LEVEL_DB)
           //         which, "on the next occasion", would send cmd 0x15 subcmd 0x02.
           iMsgType = RIGCTRL_MSGTYPE_OTHER;
           if( iMsgLength >= 7 )  // 0x15 followed by subcode, with or without DATA ...
            { iSubcode = pbMessage[5];
              switch( iSubcode )  // decode "sub-command" after 0x15 for the log:
               { case 0x01 :  // "Read noise or S-meter squelch status"  (IC-7300)
                    iUnifiedPN = RIGCTRL_PN_SQUELCH_STATUS;
                    break;
                 case 0x02 : iUnifiedPN = RIGCTRL_PN_S_METER_LEVEL_DB;
                    // > 0000="S0" (0 dB), 0120="S9" (9*6 dB), 0241="S9+60dB"
                    break;
                 case 0x05 : // cpToken = "SquelchStat";
                    break;
                 case 0x11 : iUnifiedPN = RIGCTRL_PN_POWER_METER_PERCENT;
                    // Cmd=0x15, SubCmd=0x11="PO meter level": 0000=0%, 0143=50%, 0213=100% .. that's NON-LINEAR,
                    // so use the special scaling formula further below.
                    break;
                 case 0x12 : iUnifiedPN = RIGCTRL_PN_SWR_METER_VALUE; // dimensionless "SWR" value, range 1.0 ... 3.0
                    break;
                 case 0x13 : iUnifiedPN = RIGCTRL_PN_ALC_METER_LEVEL;
                    // Cmd=0x15, SubCmd=0x13=ALC meter level :0000=Min, 0120=Max !
                    dblConvFactor = (100.0 / 120.0/* strange value for "120 = max", whatever that means*/ );
                    break;
                 case 0x14 : iUnifiedPN = RIGCTRL_PN_COMP_METER_LEVEL_DB;
                    // Cmd=0x15, SubCmd=0x14=Comp meter level:0000=0 dB, 0130=15 dB, 0241=30 dB.. that's NON-LINEAR !
                    // here: convert into decibel, 0=no compression, 30 dB = maximum reading for IC-7300)
                    dblConvFactor = (30.0 / 241.0/* strange value for "30 dB compression" */ );
                    break;
                 case 0x15 : iUnifiedPN = RIGCTRL_PN_SUPPLY_VOLTAGE_mV;
                    // Cmd=0x15, SubCmd=0x15=Supply/PA Voltage: 00=0 V, 0013=10 V, 0241=16 V .. that's incredibly non-linear
                    dblConvFactor = (16000.0/*mV*/ / 241.0/* strange value for "16 Volts" */ );
                    break;
                 case 0x16 : iUnifiedPN = RIGCTRL_PN_DRAIN_CURRENT_mA;
                    // Cmd=0x15, SubCmd=0x16=Drain Current:   0000=0 A, 0097=10 A, 0146=15 A, 0241=25 A .. almost linear
                    // here: convert into milliamps,  not the strange CI-V internal unit (241 = 25 A)
                    dblConvFactor = (25000.0/*mA*/ / 241.0/* strange value for "25 Ampere" */ );
                    break;
                 default   : sprintf(sz80Temp, "Meter[%02X]", (int)iSubcode );
                    // cpToken = sz80Temp;
                    break;
               } // end switch < "sub-command" after 0x15 >
              pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
              if( pPI != NULL )
               { cpToken = pPI->pszToken;
               }
              else
               { sprintf(sz80Temp, "Meter[%02X]", (int)iSubcode );
                 cpToken = sz80Temp;
               }
              if( iMsgLength >= 9 )  // 0x15+subcode followed by DATA bytes. Example:
               { // pbMessage[0  1  2  3  4  5  6  7  8]
                 // > RX 009 FE FE E0 94 15 12 00 80 FD ; SWRMeterLvl
                 //       |              |  |  |___|-- value in some phantasy unit,
                 //   iMsgLength=9      cmd subcode       usually 4 digit BCD, MSDigit first
                 i32Value = RigCtrl_ParseBCD_CIV_MSByteFirst( pbMessage+6,
                                       2*(iMsgLength/*9*/-7)/*nDigits*/ );
                 SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s:%ld", cpToken, (long)i32Value ); // "set" command with the RAW value in some phantasy unit
                 d = (double)i32Value * dblConvFactor;
                 switch( iUnifiedPN ) // extravagant or non-linear conversion formula from RAW VALUE to PHYSICAL UNIT ?
                  {
                    case RIGCTRL_PN_S_METER_LEVEL_DB :  // S-Meter a la IC-7300
                       // > 0000="S0" (0 dB), 0120="S9" (9*6 dB), 0241="S9+60dB"
                       if( i32Value <= 120 ) // lower range.. "S0" (0 dB) to "S9" (54 dB)
                        { d = (double)i32Value * ( 54.0 / (120-0) );
                        }
                       else // higher range.. "S9" (54 dB) to "S9 + 60 dB" = 114 dB
                        { d = 54.0/*dB*/ + (double)(i32Value-120) * ( (114.0-54.0) / (double)(241-120) );
                        }
                       break; // end case RIGCTRL_PN_S_METER_LEVEL_DB

                    case RIGCTRL_PN_POWER_METER_PERCENT: // "PO meter level" a la IC-7300 : not in Watt but "percent of the maximum"
                       // response for Cmd=0x15, SubCmd=0x11=: 0000=0%, 0143=50%, 0213=100% ..
                       // that's NON-LINEAR, thus scale in several sections:
                       if( i32Value <= 143 ) // 1st range.. measured power level = 0 to 50 percent :
                        { d = 0.0 + (double)(i32Value-0 ) * ( (50.0-0.0) / (143-0) );
                          // With an IC-7300, output power set to "5 %" = 5 Watt for QRP in CW,
                          // got here with a MEASURED i32Value =  16 ('raw' value).
                          // Scaled result: 5.59 %  = 5.6 Watt . Ok, close enough.
                          // 
                        }
                       else // 2nd range.. measured power level = 50 to 100 percent :
                        { d = 50.0 + (double)(i32Value-143) * ( (100.0-50.0) / (213-143) );
                        }
                       break; // end case RIGCTRL_PN_POWER_METER_PERCENT

                    case RIGCTRL_PN_SWR_METER_VALUE: // SWR meter reading a la IC-7300 (typically "1.0" .. "3.0")
                       // > 0000=SWR1.0, 0048=SWR1.5, 0080=SWR2.0, 0120=SWR3.0
                       // That's a non-linear relation, thus the special treatment:
                       if( i32Value <= 48 ) // 1st range.. SWR = 1.0 to 1.5 :
                        { d = 1.0 + (double)(i32Value-0 ) * ( (1.5-1.0) / (48-0) );
                        }
                       else if( i32Value <= 80 ) // 2nd range.. SWR = 1.5 to 2.0 :
                        { d = 1.5 + (double)(i32Value-48) * ( (2.0-1.5) / (80-48) );
                        }
                       else // 3rd range.. SWR = 2.0 to 3.0,  maybe even more :
                        { d = 2.0 + (double)(i32Value-80) * ( (3.0-2.0) / (120-80) );
                        }
                       break; // end case RIGCTRL_PN_SWR_METER_VALUE

                    case RIGCTRL_PN_COMP_METER_LEVEL_DB: // COMPRESSION meter : neither linear nor logarithmic :
                       // [in] i32Value (RAW)   : 0000    0130   0241
                       // [out] d= value in dB  : 0 dB    15 dB  30 dB
                       if( i32Value <= 130 ) // lower range..
                        { d = (double)i32Value * ( (15.0-0.0) / (double)(130-0) );
                        }
                       else // i32Value > 13  ->  "Comp meter" range 15 .. 30 dB:
                        { d = 15.0/*dB*/ + (double)(i32Value-130) * ( (30.0-15.0) / (double)(241-130) );
                        }
                       break; // end case RIGCTRL_PN_COMP_METER_LEVEL_DB

                    case RIGCTRL_PN_SUPPLY_VOLTAGE_mV: // very non-linear ..
                       // [in] i32Value (RAW)   : 0000    0013   0241
                       // [out] d=PHYSICAL value: 0 Volt  10 V   16 V
                       //                         |       |_________|--> Scaling Factor = 1000*(16-10) / (241-13)
                       //                         |________|--> Scaling Factor = 1000*(10-0) / (13-0)
                       if( i32Value <= 13 ) // lower range..
                        { d = (double)i32Value * ( 1000.0 * (10.0-0.0) / (double)(13-0) );
                        }
                       else // i32Value > 13  ->  Vd range 10 V .. 16 V :
                        { d = 10.0e3/*mV!*/ + (double)(i32Value-13) * ( 1000.0 * (16.0-10.0) / (double)(241-13) );
                          // typically seen here: i32Value = 166 -> d = 14026 [mV],
                          // when the voltage on an IC-7300's DC power cord was 13.94 V .
                        }
                       break; // end case RIGCTRL_PN_SUPPLY_VOLTAGE_mV
                    default: // nothing special -> leave 'd' (physical value) as-is
                       break;
                  }
                 RigCtrl_SetParamValue_Double( pRC, pPI, d );
               }
              else  // 0x15 followed by subcode but NO DATA ...
               { SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s ?", cpToken ); // query, no "set" command
               }
            } // end if < sufficient length for COMMAND + SUB-COMMAND > ?
           break; // end case 0x15 = "Send/read noise or S-meter squelch status", etc etc

        case 0x16: // Preamp, AGC speed, Noise blanker, Audio peak filter, Noise reduction, auto notch, etc etc
           iMsgType = RIGCTRL_MSGTYPE_OTHER;
           // Seen when RS-BA1 took over control: > r1 007 FE FE 94 E0 16 43 FD
           //                           but also: > r1 008 FE FE 94 E0 16 46 00 FD
           if( iMsgLength >= 7 )  // 0x1B followed by subcode, with or without DATA ...
            { iSubcode = pbMessage[5];
              switch( iSubcode )  // decode "sub-command" after 0x15 for the log:
               { case 0x02 : iUnifiedPN = RIGCTRL_PN_PREAMP_SETTING; break;
                 case 0x12 : iUnifiedPN = RIGCTRL_PN_AGC_SPEED;      break;
                 case 0x22 : iUnifiedPN = RIGCTRL_PN_NOISE_BLANKER_ON; break;
                 case 0x40 : iUnifiedPN = RIGCTRL_PN_NOISE_REDUCT_ON; break;
                 case 0x41 : iUnifiedPN = RIGCTRL_PN_AUTO_NOTCH_ON;   break;
                 case 0x42 : iUnifiedPN = RIGCTRL_PN_RPTR_TONE_ON;   break;
                 case 0x43 : iUnifiedPN = RIGCTRL_PN_TONE_SQUELCH_ON; break;
                 case 0x44 : iUnifiedPN = RIGCTRL_PN_COMP_ON;        break;
                 case 0x45 : iUnifiedPN = RIGCTRL_PN_MONITOR_ON;     break;
                 case 0x46 : iUnifiedPN = RIGCTRL_PN_VOX_ON;         break;
                 case 0x47 : iUnifiedPN = RIGCTRL_PN_BK_IN_ON;       break;
                 case 0x48 : iUnifiedPN = RIGCTRL_PN_MANUAL_NOTCH_ON; break;
                 case 0x4F : iUnifiedPN = RIGCTRL_PN_TWIN_PEAK_FILTER_ON; break;
                 case 0x50 : iUnifiedPN = RIGCTRL_PN_DIAL_LOCK_ON;   break;
                 case 0x56 : iUnifiedPN = RIGCTRL_PN_DSP_SHARP_SOFT; break;
                 case 0x57 : iUnifiedPN = RIGCTRL_PN_MANUAL_NOTCH_WIDTH; break;
                 case 0x58 : iUnifiedPN = RIGCTRL_PN_SSB_TX_BANDWIDTH; break;
                 default   : break;
               } // end switch < "sub-command" after 0x16 >
              pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
              if( pPI != NULL )
               { cpToken = pPI->pszToken;
               }
              else
               { sprintf(sz80Temp, "0x16[%02X]", (int)iSubcode );
                 cpToken = sz80Temp;
               }
              if( iMsgLength >= 8 )  // 0x16+subcode followed by one or more DATA bytes. Example:
               { // pbMessage[0  1  2  3  4  5  6  7  8]
                 // > RX 009 FE FE E0 94 16 12 00 80 FD ; SWRMeterLvl
                 //       |              |  |  |___|-- value in some phantasy unit,
                 //   iMsgLength=9      cmd subcode       usually 4 digit BCD, MSDigit first
                 i32Value = RigCtrl_ParseBCD_CIV_MSByteFirst( pbMessage+6,
                                       2*(iMsgLength/*9*/-7)/*nDigits*/ );
                 SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s:%ld", cpToken, (long)i32Value ); // "set" command with value in some phantasy unit
                 RigCtrl_SetParamValue_Int( pRC, pPI, i32Value ); // no conversion factor required here
               }
              else  // 0x16 followed by subcode but NO DATA ...
               { SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s ?", cpToken ); // query, no "set" command
               }
            } // end if < sufficient length for COMMAND + SUB-COMMAND > ?
           break; // end case 0x16

        case 0x17: // Send CW messages. Also seen when "RS-BA1 took over control:
                   // > r1 007 FE FE 94 E0 17 FF FD
                   //                         |____ "FF stops sending CW messages"
           iMsgType = RIGCTRL_MSGTYPE_OTHER;
           SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "CWmsg:%02X", (int)pbMessage[5] ); // for the log..
           break; // end case 0x17

        case 0x18: // Turn transceiver on/off (includes "wake-up via USB")
           // This was the first message received from RS-BA1 :
           //  pbMessage[0  1  2  3  4  5  6]
           // >  r1 007 FE FE 94 E0 18 01 FD
           if( iMsgLength >= 7 )
            { iMsgType = RIGCTRL_MSGTYPE_POWER_ON_OFF;
              SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "Power:%s", (pbMessage[5]==0) ? "OFF" : "ON" );
            }
           break; // end case 0x18

        case 0x19: // Read the transceiver ID (beware, different meaning depending on length) ? ?
           //   | Example from SL's 'CAT traffic' display, from an IC-7300 :
           //   '------------------------------,
           //   Time           Len Pre.  Adddr |     Postamble
           // > 14:39:59.2: TX 007 FE FE 94 E0 19 00 FD  |      (request the radio ID)
           // > 14:40:00.1: RX 008 FE FE E0 94 19 00 94 FD      (response with radio ID)
           //                                     |___|--- 0x19 followed by 0x00 AND a 'data byte' !!
           //
           // [IC7610CIV] page 4 : "Cmd 0x19 : Sub Cmd = 0x00  Data = ??? : "Read the transceiver ID".
           if( iMsgLength >= 8 )  // 0x19 followed by TWO BYTES (!) before the postamble:
            { switch( pbMessage[5] )  // "sub-command" after 0x19 ?
               { case 0x00 :  // Guess what follows is the "DEFAULT CI-V address" alias "Transceiver ID code" ...
                    pRC->iDefaultAddress = (int)pbMessage[6]; // <- this may overwrite iDefaultAddress=RIGCTRL_DEF_ADDR_AUTO_DETECT (0)
                      //  '--> e.g. 148 = 0x94 = RIGCTRL_DEF_ADDR_IC_7300 .        
                    SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "rig=%s",
                        RigCtrl_DefaultAddressToString(
                          pRC->iRadioCtrlProtocol, pRC->iDefaultAddress) ); // for the log..
                    iMsgType = RIGCTRL_MSGTYPE_RADIO_ID;
                    pRadioInfo = RigCtrl_GetRadioInfoByDefaultAddress( // -> e.g. from RigCtrl_RadioInfo_CIV[]
                        pRC->iRadioCtrlProtocol, pRC->iDefaultAddress ); // for the 'capabilities'
                    if( pRadioInfo != NULL )
                     {  pRC->iCapabilities    = pRadioInfo->iCapabilities;
                        pRC->dwAvailableBands = pRadioInfo->dwBands; // <- may also change later, after RIGCTRL_POLLSTATE_TX_BAND_EDGES
                        //    '--> This is only ONE of several inputs for RigCtrl_GetAvailableBands() !
                        //         Details there.. the radio may be able to report
                        //         its own 'TX band edge frequencies'. If it does,
                        //         RigCtrl_GetAvailableBands() respects limitations
                        //         or additional bands like 60m and 4m.
                        //
                     }
                    else // radio-info not available .. but be optimistic, it may be some brand-new model !
                     {  pRC->iCapabilities = RIGCTRL_CAPS_ALL;
                     }
                    break;
                 default   :  // anything else is neither documented in [IC7610CIV] nor [DF4OR_CIV] ...
                    break;
               }
            }
           break; // end case 0x19

        case 0x1A: // send/read memory contents, band stacking registers, memory keyer contents, etc
           iMsgType = RIGCTRL_MSGTYPE_OTHER;
           if( iMsgLength >= 7 )  // 0x1A followed by at least ONE byte for the sub-command...
            { iSubcode = pbMessage[5];
              switch( iSubcode )  // "sub-command" after 0x1A ?
               { case 0x00 :  // response for "Send / read memory contents" ?
                    SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "mem[%d]", (int)pbMessage[6] ); // for the log..
                    break;
                 case 0x01 :  // response for "Send / read band stacking registers" ?
                    SL_AppendString( &pszCmtDest, cpCmtEndstop, "band stack mem" );
                    if( (iMsgLength > 13) // at least "operating frequency" and "operating mode"..
                     && (pRC->iParameterPollingState==RIGCTRL_POLLSTATE_BAND_STACKING_REGS)
                     && (pRC->iNumBandStackingRegs >=0 )
                     && (pRC->iNumBandStackingRegs < RIGCTRL_NUM_BAND_STACKING_REGS) )
                     { // Try to(!) parse the another 'band stacking register' :
                       // [out] pRC->BandStackingRegs[ pRC->iNumBandStackingRegs++ ]
                       // [in] :  --------------------------,
                       //                                   |
                       //                                  \|/
                       // pbMessage[0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18]
                       // > RX 017 FE FE 00 94 1A 01 01 01 30 96 83 01 00 00 02 01 00 08 85 00 08 85 .. ; band stack mem
                       //          |PRE| |ADR| CMD |  |  | |____________|  | |__etc, etc, etc, etc, etc, etc, etc ...
                       //                         Sub |  | 10-digit freq  OpMode
                       //         "Frequency band code" "Register code"   (see RigCtrl_IcomToRigCtrlOpmode)
                       //          (ignored here)        (ignored here)
                       pMemEntry = &pRC->BandStackingRegs[pRC->iNumBandStackingRegs];
                       if( RigCtrl_ParseMemoryContent( pbMessage+8, iMsgLength-8-1/*nBytes*/, pMemEntry) )
                        { SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, ": %.1lf kHz, %s",
                            (double)(pMemEntry->RxTx[0].dblOperatingFreq_Hz * 1e-3),
                            RigCtrl_OperatingModeToString( pMemEntry->RxTx[0].iOpMode) );
                          ++pRC->iNumBandStackingRegs;
                        }
                     }
                    break;
                 case 0x02 :  // "Send / read memory keyer contents"
                    SL_AppendString( &pszCmtDest, cpCmtEndstop, "keyer mem" );
                    break;
                 case 0x03 :  // "Send / read selected IF filter"
                    // Surprisingly, Cmd 0x1A sub 0x03 is included in WSJT-X's "polling cycle".
                    // Thus this parameter (but converted into Hz) now has its own PN :
                    iUnifiedPN = RIGCTRL_PN_FILTER_BANDWIDTH;
                    pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
                    if( pPI != NULL )
                     { cpToken = pPI->pszToken;
                     }
                    else
                     { cpToken = "FilterBW";
                     }
                    // The strange unit used by Icom / IC-7300 is specified as below:
                    // > Cmd 0x1A subcmd 0x03: Send/read the selected filter width.
                    // >   ( AM: 00=200 Hz to 49=10 kHz;
                    // >     other than AM modes:
                    // >          00=50 Hz to 31/40=2700 Hz/3600 Hz)
                    // Heavens, no. A CONVERSION FORMULA or simply a SIX-DIGIT BCD
                    // number would be asking too much. Hey Icom, you managed
                    // to send SIX DIGIT DECIMAL NUMBERS elsewhere, so you could
                    // even report the IF filter if the radio (IC-705) was in "WFM" !
                    //
                    // Example with an IC-7300 in USB, "FIL2" (seen in each WSJT-X polling cycle):
                    // pbMessage[0  1  2  3  4  5  6  7]
                    // > TX 007 FE FE 94 E0 1A 03 FD    ; FilterBW ?
                    // > RX 008 FE FE E0 94 1A 03 28 FD ; FilterBW = "28" .
                    if( iMsgLength >= 7 ) // valid SET-command ?
                     { i32Value = RigCtrl_ParseBCD_CIV_MSByteFirst( pbMessage+6, 2/*digits*/ );
                       // For details on the CRAZY ENCODING of the 2-digit BCD value
                       // (here: i32Value), see the IC-9700 CI-V Reference "V4",
                       // Icom document number "A7508-3EX-4", page 16.
                       //  Read it, try to understand, and CRY OUT LOUD !
                       // If you ask the radio for the "current IF filter width setting"
                       // in mode FM, the radio will even reply with an ERROR !
                       // For AM, the meaning of the 2-digit "Data" value is different
                       // than any other mode (SSB/CW/RTTY).
                       if( i32Value <= 9 ) // IF filter bandwidth codes 0 .. 9 :
                        { i32Value = 50/*Hz*/ + 50 * i32Value; // 50 Hz units for narrow-band SSB/CW/RTTY
                        }
                       else if( i32Value <= 40 ) // IF filter bandwidth codes 10 .. 40 (SSB/CW ONLY):
                        { i32Value = 600/*Hz*/ + 100 * (i32Value-10); // 100 Hz units for 600 Hz and above
                        }
                       else // stupid filter bandwidth codes above "40" only occur in AM (?)
                        { i32Value = 200/*Hz*/ + 200 * i32Value;  // "200 Hz .. 10 kHz in 200-Hz-steps" (buaaaah)
                        }
                       RigCtrl_SetParamValue_Int( pRC, pPI, i32Value );
                       SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s=%d Hz", cpToken, i32Value );
                     }
                    else
                     { SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s ?", cpToken );
                     }
                    break;
                 case 0x04 :  // "Send / read selected AGC time constant"
                    SL_AppendString( &pszCmtDest, cpCmtEndstop, "AGC speed" );
                    break;
                 case 0x05 :  // "Send / read an awful lot of RIG-SPECIFIC PARAMETERS"
                    // Typical QUERY (seen after starting RS-BA1 on client port #1):
                    //      pbMessage[0  1  2  3  4  5  6  7  8]
                    //      > r1 009 FE FE 94 E0 1A 05 00 60 FD
                    //           |_|                   |___| '-- postamble
                    //           iMsgLength=9            '------ 4-digit "subcode"
                    if( iMsgLength >= 9 ) // valid QUERY or SET-command ?
                     { iSubcode = RigCtrl_ParseBCD_CIV_MSByteFirst( pbMessage+6, 4/*digits*/ );
                       // Actually, the list of subcodes addressed with this
                       // 4-digit BCD extends over MULTIPLE PAGES in the IC-7610's CI-V manual.
                       // Very unfortunately, the meaning of those 4-digit "subcodes"
                       // is extremely rig specific.
                       // Not even an IC-7300 and IC-7610 are compatible. Thus:
                       iUnifiedPN = RigCtrl_SubcodeFrom1A05_to_UnifiedParameterNumber(pRC, iSubcode);
                       if( iUnifiedPN != RIGCTRL_PN_UNKNOWN )
                        { cpToken = RigCtrl_UnifiedParameterNumberToString(iUnifiedPN);
                        }
                       else // must be one of the "less important" parameters..
                        {   // .. at least RigControl.c doesn't know what it is.
                          sprintf( sz80Temp, "Cfg[%04d]", (int)iSubcode );
                          cpToken = sz80Temp;
                        }
                       if( iMsgLength > 9 ) // long enough for a "SET"-command.
                        { i32Value = RigCtrl_ParseBCD_CIV_MSByteFirst( pbMessage+8,
                                       2*(iMsgLength/*11*/-9)/*nDigits*/ );
                          // Typical RESPONSE (from an IC-7300, here: subcode 0060
                          //   -> iUnifiedPN = RIGCTRL_PN_AF_OUTPUT_LEVEL_TO_ACC =
                          // "AF output level to ACC/USB", iMsgLength=11, 2 byte data = 4 BCD digit
                          //      pbMessage[0  1  2  3  4  5  6  7  8  9 10]
                          //      > RX 00B FE FE E0 94 1A 05 00 60 01 30 FD
                          //      _________________________________|___|
                          //     | nDigits = 2 * (11-9) = 4, i.e. BCD-data with
                          //   a 4-digit "value" (here: volume ranging from 000 to 0255 = 100%)
                          SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s=%d", cpToken, (long)i32Value );
                        }
                       else // message too short for a "set"-command, so guess it's a QUERY ("get"):
                        { SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s ?", cpToken );
                        }
                       // Seen when launching "RS-BA1", connected via SpecLab to an IC-7300:
                       //    Cfg[0052]=0 ( "Function > Memo Pad Quantity" : 5 channels )
                       //    Cfg[0057]=0 ( "Function > Quick RIT/deltaTX Clear" :  OFF )
                       //    Cfg[0071]=1 ( "Connectors > Phones > Level" ? )
                       //    Cfg[0066]=3 ( IC-7300: "MOD input connector during DATA" : 3=USB.
                       //                  Totally different on IC-7610 :
                       //                  0066 would be "Screen Capture Keyboard".. EEEK !!)
                       switch( iUnifiedPN )
                        { case  RIGCTRL_PN_CIV_USB_ECHO : // 0x1A 0x05 Cfg[0075] : "Echo back setting for CI-V operation from USB"
                             // [G3NRW_CIV]
                             // > Note that the description of the "Set echo back setting"
                             // >  in the Command Table in Section 19 of the
                             // >  IC-7300 Full Manual is misleading. The
                             // >  parameter 00 sets the control to OFF
                             // >  and the parameter 01 sets it to ON .
                             // Here, each client may have prefer its own setting.
                             // Most clients will not know about how to configure
                             // the CI-V "echo back setting" via CI-V itself,
                             // thus in Spectrum Lab, the initial setting of
                             //  T_RigCtrlInstance.PortInstance[x].fEmulateCIVEcho
                             // can be configured via checkmark in the setup dialog.
                             if( iRigCtrlPort != RIGCTRL_PORT_RADIO ) // received on one of the "CLIENT"-ports ?
                              { iClientAction = RIGCTRL_CLIENT_ACTION_FAKE_RESPONSE; // "do NOT forward this to the real radio" !
                              }
                             break;
                          default: // ignore hundreds(!) of other "subcodes" after 0x1A 0x05
                             break;
                        } // end switch( iUnifiedPN )
                     } // end < message long enough for cmd 0x1A 0x05 >
                    break; // end case "Send / read an awful lot of RIG-SPECIFIC PARAMETERS"
                 case 0x06 :  // (IC-7300) : "Data mode with filter width settings"
                    // Example with an IC-7300 in USB, "FIL2" (seen in each WSJT-X polling cycle):
                    // pbMessage[0  1  2  3  4  5  6  7]
                    // > TX 007 FE FE 94 E0 1A 06 FD          ; Data/Filter ?
                    // > RX RX 009 FE FE E0 94 1A 06 00 00 FD ; Data mode OFF, no filter info.
                    //                                |  |__ only when DATA ON: 01=Fil1, 02=Fil2, 03=Fil3
                    //                                0=Data mode off, 1=Data on.
                    iUnifiedPN = RIGCTRL_PN_DATA_MODE;
                    pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
                    if( pPI != NULL )
                     { cpToken = pPI->pszToken;
                     }
                    else
                     { cpToken = "DataMode";
                     }
                    if( iMsgLength >= 8 ) // valid SET-command or GET-response ?
                     { i32Value = RigCtrl_ParseBCD_CIV_MSByteFirst( pbMessage+7, 2/*digits*/ );
                       RigCtrl_SetParamValue_Int( pRC, pPI, i32Value );
                       SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s=%d", cpToken, i32Value );
                     }
                    else
                     { SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s ?", cpToken );
                     }
                    break;
                 case 0x07 :  // "NTP server access" (IC-7610) / "IP+ Setting" (IC-7300)
                 case 0x08 :  // "NTP access result" (IC-7610) / not used in IC-7300
                 case 0x09 :  // "AF Mute"           (IC-7610) / not used in IC-7300
                    iUnifiedPN = RigCtrl_SubcodeFrom1A_to_UnifiedParameterNumber(pRC, iSubcode);
                    pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
                    if( pPI != NULL )  // anything known about this sub-command ? Do we care ?
                     { cpToken = pPI->pszToken;
                       if( iMsgLength > 7 ) // valid SET-command or GET-response ?
                        { // Here just one of the RESPONSES from an IC-7300,
                          // after being asked by RS-BA1 for whatever-it-was :
                          //          [0  1  2  3  4  5  6  7]
                          // > RX 008 FE FE E0 94 1A 07 00 FD
                          i32Value = RigCtrl_ParseBCD_CIV_MSByteFirst( pbMessage+6,
                                         2*(iMsgLength/*8*/-7)/*nDigits*/ );
                          RigCtrl_SetParamValue_Int( pRC, pPI, i32Value );
                          SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s=%d", cpToken, i32Value );
                        }
                       else // Just a QUERY .. one of hundreds seen from RS-BA1 :
                        { //          [0  1  2  3  4  5  6]
                          // > r1 007 FE FE 94 E0 1A 07 FD (other)
                          SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s ?", cpToken );
                        }
                     }
                    else // don't care what it is .. DL4YHF was sick of hard-coding RIG-SPECIFIC details here
                     {
                     }
                    break;
                 default:
                    break;
               } // end switch < "sub-command" after 0x1A >
            } // end if < sufficient message length for this command ? >
           break; // end case 0x1A

        case 0x1B: // repeater tone frequency, TSQL tone frequency
           iMsgType = RIGCTRL_MSGTYPE_OTHER;
           if( iMsgLength >= 7 )  // 0x1B followed by subcode, with or without DATA ...
            { iSubcode = pbMessage[5];
              switch( iSubcode )  // decode "sub-command" after 0x1B for the log:
               { case 0x00 : iUnifiedPN = RIGCTRL_PN_REPEATER_TONE_FREQ; break;
                 case 0x01 : iUnifiedPN = RIGCTRL_PN_TONE_SQUELCH_FREQ;  break;
                 default   : break;
               } // end switch < "sub-command" after 0x1B >
              pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
              if( pPI != NULL )
               { cpToken = pPI->pszToken;
               }
              else
               { sprintf(sz80Temp, "0x1B[%02X]", (int)iSubcode );
                 cpToken = sz80Temp;
               }
              if( iMsgLength >= 10 )  // 0x1B+subcode followed by DATA bytes. Example:
               { // pbMessage[0  1  2  3  4  5  6  7  8  9]
                 // > RX 00A FE FE E0 94 1B 01 00 08 85 FD ; TSQL Tone
                 //       |              |  |  |______|-- value in 0.1 Hz units,
                 //   iMsgLength=10     cmd subcode       6 digit BCD, MSDigit first
                 d = 0.1 * (double)RigCtrl_ParseBCD_CIV_MSByteFirst( pbMessage+6,
                                       2*(iMsgLength/*10*/-7)/*nDigits*/ );
                 SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s:%.1lf Hz", cpToken, (double)d ); // "set" command with value in Hz
                 RigCtrl_SetParamValue_Double( pRC, pPI, d );
               }
              else  // 0x1B followed by subcode but NO DATA ...
               { // Seen when starting "RS-BA1" : > r1 007 FE FE 94 E0 1B 00 FD
                 SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s ?", cpToken ); // query, no "set" command
               }
            } // end if < sufficient length for COMMAND + SUB-COMMAND > ?
           break; // end case 0x1B

        case 0x1C: // transceiver status (RX/TX), antenna tuner, XFC, read TX frequency, etc .
           iMsgType = RIGCTRL_MSGTYPE_OTHER;
           // Received from RS-BA1 : > r1 007 FE FE 94 E0 1C 04 FD  (with iMsgLen=7, only a QUERY)
           //                                                |__ "CIV_ANT" ?
           if( iMsgLength >= 7 )  // 0x1C followed by at least <??> bytes...
            { iSubcode = pbMessage[5];
              switch( iSubcode )  // "sub-command" after 0x1C ?
               { case 0x00 : iUnifiedPN = RIGCTRL_PN_TRANSMITTING;
                    // Used by WSJT-X (Hamlib) for RX/TX switching. Must keep track of this
                    // regardless of who sent the command, because the IC-7300's spectrum scope
                    // will stop sending 'waveform data' while transmitting (in MOST cases).
                    break;
                 case 0x01 : iUnifiedPN = RIGCTRL_PN_ATU_ENABLED; break;
                 case 0x02 : iUnifiedPN = RIGCTRL_PN_XFC_PRESSED; break;
                 case 0x03 : iUnifiedPN = RIGCTRL_PN_TRANSMIT_FREQ; break;
                 case 0x04 : iUnifiedPN = RIGCTRL_PN_CIV_OUTPUT_FOR_ANT; // "CI-V output for ANT"
                    break;
                 default   :
                    break;
               } // end switch < "sub-command" after 0x1C >
              pPI = RigCtrl_GetInfoForUnifiedParameterNumber(iUnifiedPN);
              if( pPI != NULL )
               { cpToken = pPI->pszToken;
               }
              else
               { sprintf(sz80Temp, "TrxSt[%02X]", (int)iSubcode );
                 cpToken = sz80Temp;
               }
              if( iMsgLength >= 8 )  // 0x1C+subcode followed by one or more DATA bytes. Example:
               { // pbMessage[0  1  2  3  4  5  6  7]
                 // > RX 008 FE FE E0 94 1C 04 00 FD
                 //       |  'pre' to fm |  |  '-- value, here only two digits,
                 //   iMsgLength=8      cmd subcode       guess this is BCD, MSDigit first
                 i32Value = RigCtrl_ParseBCD_CIV_LSByteFirst( pbMessage+6,
                                       2*(iMsgLength/*8*/-7)/*nDigits*/ );
                   // ,-----------------------------------'
                   // '--> total message length minus SEVEN bytes
                   //      for preamble(2), addresses(2), com+sub(1), postamble(1).
                   // Other example (with "TX_freq" from an IC-9700) :
                   //   LEAST significant byte (with tens and ones) first
                   //    '--------------,           ,----  1-GHz-digit for IC-9700;
                   //                   |           |      thus expect more BYTES for e.g. IC-905 ! 
                   // FE FE 00 A2 1C 03 10 00 05 44 01 FD
                   //              |  | |____________| |
                   //            cmd sub     |         postamble
                   //                     12-7 = 5 BYTES = TEN BCD-digits,
                   //                decoded into 0144050010 Hz = 144.05001 MHz.
                 SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s:%ld", cpToken, (long)i32Value ); // "set" command with value
                 RigCtrl_SetParamValue_Int( pRC, pPI, i32Value );
               }
              else  // 0x1C followed by subcode but NO DATA ...
               { SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s ?", cpToken ); // query, no "set" command
               }
            } // end if < sufficient message length for this command ? >
           break; // end case 0x1C

        case 0x1E: // read number of TX frequency bands, TX band edge frequencies, etc
           iMsgType = RIGCTRL_MSGTYPE_OTHER;
           if( iMsgLength >= 7 )  // 0x1E followed by at least <??> bytes...
            {
              if( iMsgLength >= 8 )  // 0x1E+subcode followed by one or more DATA bytes. Example:
               { // pbMessage[0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18]
                 // > RX 013 FE FE 00 A2 1E 01 01 00 00 00 44 01 2D 00 00 00 46 01 FD ; 144 MHz .. 146 MHz
                 //       |  |___| |addr |  |  |  |____________| |  |____________|
                 //       | Preamble     |  | "Edge"   |        Sep.    |
                 //   iMsgLength=19   cmd sub  |     StartFreq        EndFreq
                 // ,--------------------------'
                 i32Value = RigCtrl_ParseBCD_CIV_MSByteFirst( pbMessage+6, 2/*nDigits*/ );
                 SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s=%ld", cpToken, (long)i32Value );
               }
              else  // 0x1E followed by a subcode but NO DATA ... possibly a query (?)
               { SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s ?", cpToken );
               }
              iSubcode = pbMessage[5];
              switch( iSubcode )  // "sub-command" after 0x1E ?
               { case 0x00 : // "Read number of available TX frequency band"[s??] IC-9700_ENG_CI-V_4.pdf page 13
                    pPI = RigCtrl_GetInfoForUnifiedParameterNumber(RIGCTRL_PN_NUM_TX_BANDS);
                    if( pPI != NULL )
                     { cpToken = pPI->pszToken;
                       RigCtrl_SetParamValue_Int( pRC, pPI, i32Value );
                       UTL_LimitInteger( &pRC->iNumTxBands, 0, RIGCTRL_NUM_TX_BAND_EDGES );
                     }
                    break;
                 case 0x01 : // "Read TX band edge frequencies" IC-9700_ENG_CI-V_4.pdf page 13
                    if( (i32Value>=1) && (i32Value<=RIGCTRL_NUM_TX_BAND_EDGES) ) // valid "edge number" ?
                     {
                       // This kind-of-array-index (Icom calls it "Edge Number", 01 .. 30 in BCD)
                       // isn't the sub-command but part of the DATA FIELD !
                       T_RigCtrlFrequencyRange *pFreqRange = &pRC->TxBandEdges[i32Value];
                       // From the IC-9700 CI-V Reference Guide Rev. 4, document "A7508-3EX-4", page 13:
                       // > Band edge frequency settings
                       // pbMessage[6       7  8  9 10 11  12  13 14 15 16 17  18]
                       //          Edge#    |___________|  2D  |____________|  FD
                       //          01..30   Lower edge,BCD  |  Upper edge, BCD  |
                       //                   LSByte first    |  LSByte first     |
                       //                            "Separator (fixed)"     Postamble
                       RigCtrl_ParseCIVFreqRange( pbMessage+7, iMsgLength-7-1/*nBytes*/, pFreqRange );
                       // Flag 'fIsExpectedResponse' has already been set above,
                       // so the state machine in RigCtrl_Handler(),
                       //         case RIGCTRL_POLLSTATE_TX_BAND_EDGES,
                       // will read the NEXT 'TX band edge frequencies' very soon.
                       cpToken = "TxBand";
                       SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%.1lf .. %.1lf MHz", pFreqRange->dblFmin_Hz*1e-6, pFreqRange->dblFmax_Hz*1e-6 );
                     } // end if < parameter-info Non-NULL for RIGCTRL_PN_TX_BAND_EDGES >
                    break;
                 case 0x02 : // "Read number of user-set TX frequency band"[s??]
                    cpToken = "User TX bands";
                    break;
                 case 0x03 : // "Set the user-set TX band edge frequencies"
                    cpToken = "User band edges";
                    // Format as for cmd 1E sub 01 "TX band edge frequencies"
                    break;
                 default   : sprintf(sz80Temp, "TxBands[%02X]", (int)iSubcode );
                    cpToken = sz80Temp;
                    break;
               } // end switch < "sub-command" after 0x1E >
            } // end if < sufficient message length for this command ? >
           break; // end case 0x1E

        case 0x21: // RIT frequency setting, delta TX setting (on/off)
           iMsgType = RIGCTRL_MSGTYPE_RIT_XIT;
           break; // end case 0x21

        case 0x25: // "Send/read main or sub-band frequency" (IC-7610)
                   // "Send/read the selected or unselected VFO frequency" (IC-7300)
           iMsgType = RIGCTRL_MSGTYPE_FREQUENCY_REPORT; // here, in RigCtrl_ParseCIV(), it's a frequency REPORT, not a "SET"-command
           //                        pbMessage[0  1  2  3  4  5  6  7  8  9 10 11]
           // Received from RS-BA1 : > r1 007 FE FE 94 E0 25 00 FD
           // Response from IC-7300: > RX 00C FE FE E0 94 25 00 00 30 35 05 00 FD
           //                              |              |   | |____________|-- data, BCD, 10 decimal digits, LEAST significant digits first !
           //                          iMsgLength=12     cmd  0=selected VFO,
           //                                                 1=unselected VFO.
           if( iMsgLength >= 7 )  // 0x25 + subcode ("sel/unsel") with or without data..
            { iSubcode = pbMessage[5];
              switch( iSubcode )  // "sub-command" after 0x25 ?
               { case 0x00 : cpToken = "SelVFOFreq";   break;
                 case 0x01 : cpToken = "UnselVFOFreq"; break;
                 default   : cpToken = "SomeOtherFreq"; break;
               } // end switch < "sub-command" after 0x1B >
              if( iMsgLength >= 8 )  // 0x25+subcode followed by one or more DATA bytes
               { i32Value = RigCtrl_ParseBCD_CIV_LSByteFirst( pbMessage+6,
                                       2*(iMsgLength/*12*/-7)/*nDigits*/ );
                 SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s:%ld", cpToken, (long)i32Value ); // "set" command with value
               }
              else  // 0x16 followed by subcode but NO DATA ...
               { SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s ?", cpToken ); // query, no "set" command
               }
            } // end if < sufficient message length for this command ? >
           break; // end case 0x25

        case 0x26: // Send/read operating mode and filter setting
           //                        pbMessage[0  1  2  3  4  5  6  7  8  9]
           // Received from RS-BA1 : > r1 007 FE FE 94 E0 26 01 FD
           // Response from IC-7300: > RX 00A FE FE E0 94 26 01 03 00 02 FD
           //                              |              |   |  |  |  |
           //                              |              |   |  |  |  |__ filter (1..3)
           //                              |              |   |  |  |__ DataMode, 0=off, 1=on
           //                              |              |   |  |__ OpMode, 3=CW
           //                          iMsgLength=10     cmd  0=selected VFO,
           //                                                 1=unselected VFO.
           if( iMsgLength >= 7 )  // 0x26 followed by at least ONE byte before the postamble:
            { iMsgType = RIGCTRL_MSGTYPE_OP_MODE;
              iSubcode = pbMessage[5];
              switch( iSubcode )  // "sub-command" after 0x25 ?
               { case 0x00 : cpToken = "SelVFO";   break;
                 case 0x01 : cpToken = "UnselVFO"; break;
                 default   : cpToken = "UnknVFO";  break;
               } // end switch < "sub-command" after 0x1B >
              if( iMsgLength >= 8 ) // 0x26 + VFO-Nr. + at least ONE data byte ?
               { iNewSetting = RigCtrl_IcomToRigCtrlOpmode( pbMessage[6] );
                 SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%s:%s", cpToken, RigCtrl_OperatingModeToString(iNewSetting) );
                 if( iMsgLength >= 9 )  // "Operating mode" byte followed by optional "Filter setting" ?
                  { switch( pbMessage[7] )
                     { case 1 : break;  // "FIL1" (normal filter bandwidth)
                       case 2 : iNewSetting |= RIGCTRL_OPMODE_NARROW;      break; // "FIL2" [IC7610CIV page 10]
                       case 3 : iNewSetting |= RIGCTRL_OPMODE_VERY_NARROW; break; // "FIL3"
                       default: break;
                     }
                    SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, ",FIL%d", pbMessage[6] );
                  } // end if < radio also reported the current FILTER SETTING >
                 if( iMsgLength >= 10 )  // "Filter setting" followed by "Data Mode"-flag ?
                  { SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, ",Data=%d", (int)pbMessage[7] );
                  } // end if < radio also reported the current FILTER SETTING >
                 // Only accept the radio's reported value if we're not going to SEND a NEW operating mode:
                 if( ! ( pRC->iPostponedSetMessages & (1<<RIGCTRL_MSGTYPE_OP_MODE) ) )
                  {
                    if( iNewSetting != pRC->iOpMode )
                     { pRC->iOpMode = iNewSetting;
                     }
                  }
               }
              else // not a "set"-command but a "get"-query :
               { SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%sMode ?", cpToken );
               }
            } // end if < sufficient message length for this command ? >
           break; // end case 0x26

        case 0x27: // Read "Scope" waveform data (spectrum display):
           if( iMsgLength >= 7 )  // 0x27 followed by at least ONE byte before the postamble:
            { // (RS-BA1 periodically sent > r1 007 FE FE 94 E0 27 1C FD,
              //       i.e. NO DATA FIELD but a "query". So beware.. )
              //
              if( iRigCtrlPort==RIGCTRL_PORT_RADIO )
               { if( pRC->i32ExpectedResponse_CmdAndSubcode == (0x2700 | pbMessage[5]) )
                  { // RigCtrl_Handler() has been waiting for THIS in case RIGCTRL_POLLSTATE_CONFIGURE_SPECTRUM ...
                    pRC->iResponseCountdown_ms = 0;   // don't wait any longer !
                    pRC->i32ExpectedResponse_CmdAndSubcode = -1;   // avoid misinterpreting the next unspecific "OK" or "NOT OK".
                    // (only RigCtrl_Handler() decides what to ask NEXT)
                  } // end if < 0x27.. was the EXPECTED response for something >
               } // end if < message received FROM THE RADIO (not from any of the clients) > ?
              iMsgType = RIGCTRL_MSGTYPE_C_SPECTRUM; // initial guess: spectrum CONTROL
              iSubcode = pbMessage[5];
              switch( iSubcode )  // "sub-command" after 0x27 ?
               {  case 0x00 :  // Guess what follows is a block of 'Scope waveform data', in this case: part of a SPECTRUM.
                    iMsgType = RIGCTRL_MSGTYPE_SPECTRUM; // not spectrum-CONTROL but -DATA !
                    fIsUnsolicitedMsg = TRUE;
                    pRC->iScopeDataCountdown_ms = RIGCTRL_SPECTRUM_TIMEOUT_MS; // reload 'Scope waveform' timer (for timeout monitoring)
                    // Do we want to receive spectra at all (configurable) ?
                    if( pRC->iRigControlFlags & RIGCTRL_FLAG_WANT_SPECTRUM_DATA )
                     { // Yes...
                       RigCtrl_ParseSpectrumData_CIV( pRC, pbMessage+6/*payload*/, iMsgLength-7/*payloadLength*/,
                                                      &pszCmtDest, cpCmtEndstop );
                       // See ASCII graphics in RigCtrl_ParseSpectrumData_CIV()
                       // about what is considered the 'payload' for a SPECTRUM !
                     } // end if < ..ControlFlags & RIGCTRL_FLAG_WANT_SPECTRUM_DATA >
                    else // option RIGCTRL_FLAG_WANT_SPECTRUM_DATA *not* set :
                     { // Turn off the unsolicited transmission of scope data.
                       // This reduces the 'bus load', so other messages
                       // (like VFO frequency control) have a smaller latency.
                       // Also prevents flooding the CAT traffic monitor .
                       pRC->iPostponedSetMessages |= (1<<RIGCTRL_MSGTYPE_C_SPECTRUM);
                       // The rest happens in RigCtrl_Handler(): As soon as
                       // there's an empty slot (in the heavy bus traffic),
                       // it will call RigCtrl_StartStopSpectrumData( pRC, TRUE ).
                     }
                    break;

                 case 0x10:  // "Send/read the Scope ON/OFF status (00=OFF, 01=ON)"
                    // If the scope (on the Icom's own display) is OFF,
                    // the radio also won't send "waveform data" via CI-V !
                    break;

                 case 0x11:  // "Send/read the Scope wave data output setting (00=OFF, 01=ON)"
                    // (should get here after RigCtrl_StartStopSpectrumData() )
                    // When "RS-BA1" connected through client port #1, it sent:
                    // > 21:22:57.3 r1 008 FE FE 94 E0 27 11 00 FD
                    //                                       |__ 0x00 = off, 0x01 = on
                    // No byte to distinguish between "main scope" and "sub scope" here !
                    if( iMsgLength < 8 ) // look mum, no DATA .. must be a QUERY .
                     { SL_AppendString( &pszCmtDest, cpCmtEndstop, "Scope data on ?" );
                     }
                    else
                     { if( pbMessage[6] == 0x00 )
                        { SL_AppendString( &pszCmtDest, cpCmtEndstop, "Scope data off" );
                          pRC->iScopeMode = RIGCTRL_SCOPE_MODE_OFF;
                          // When terminating RS-BA1 (connected to an IC-7300 through SL's client port),
                          //       it STOPPED the periodic transmission of scope waveform data :
                          // > 20:08:53.1 r1 008 FE FE 94 E0 27 11 00 FD (scope) ; Scope data off (from client)
                          // > 20:08:53.1 TX 008 FE FE 94 E0 27 11 00 FD (scope) ; Scope data off (forwarding to radio)
                          // > 20:08:53.2 RX 006 FE FE E0 94 FB FD       (OK)    ; response from radio
                          // > 20:08:53.2 t1 006 FE FE E0 94 FB FD       (OK)    ; forwarding response to client
                          // The TIMEOUT MONITOR for spectrum data in RigCtrl_Handler()
                          // will automatically try to kick the scope alive again.
                        }
                       else
                        { SL_AppendString( &pszCmtDest, cpCmtEndstop, "Scope data on" );
                          if( pRC->iScopeMode == RIGCTRL_SCOPE_MODE_OFF )
                           { // cannot be "OFF" not, so assume "fixed frequency" mode:
                             pRC->iScopeMode = RIGCTRL_SCOPE_MODE_FIXED;
                           }
                          // Note: RIGCTRL_SCOPE_MODE_CENTER / FIXED set further below
                        }
                     }
                    break;

                 case 0x12:  // "Send/read the Scope MAIN/SUB setting (00=Fixed to MAIN)"
                    break;

                 case 0x13:  // "Send/read the Scope Single/Dual setting (00=Fixed to Single)"
                    break;

                 case 0x14:  // "Send/read the Scope Center mode or Fixed mode setting"
                    // pbMessage[6] : 0x00 = main scope,  0x01 = sub scope (IC-7851) ; IC-7300 : "fixed" 0x00
                    // pbMessage[7] : 0x00 = center mode, 0x01 = fixed mode
                    // (looks like in contrast to Yaesu, Icom cares for compatibility .. in most cases)
                    if( iMsgLength < 9 ) // again, no DATA .. must be a QUERY .
                     { SL_AppendString( &pszCmtDest, cpCmtEndstop, "Scope C/F ?" );
                     }
                    else
                     {
                       if( pbMessage[6] == 0x00 ) // only care for the "main scope" here...
                        { if( pbMessage[7] == 0x00 )
                           { pRC->iScopeMode = RIGCTRL_SCOPE_MODE_CENTER;
                             SL_AppendString( &pszCmtDest, cpCmtEndstop, "Scope:Center" );
                           }
                          else
                           { pRC->iScopeMode = RIGCTRL_SCOPE_MODE_FIXED;
                             SL_AppendString( &pszCmtDest, cpCmtEndstop, "Scope:Fixed" );
                           }
                        }
                     }
                    break;

                 case 0x15:  // "Send/read the span setting in the Center mode Scope"
                    //        pbMesssage[0  1  2  3  4  5  6  7]
                    // RS-BA1: > r1 008 FE FE 94 E0 27 15 00 FD (scope) ; span ?
                    //               |                    |
                    //           iMsgLength               0=main scope, 1=sub scope
                    if( iMsgLength <= 8 ) // no DATA .. must be a QUERY .
                     { SL_AppendString( &pszCmtDest, cpCmtEndstop, "Scope span ?" );
                     }
                    else  // iMsgLength > 8 -> not a query ("get") but a SET-command :
                     { // Response from IC-7300 :
                       //     pbMesssage[0  1  2  3  4  5  6  7  8  9 10 11 12]
                       //      > RX 00D FE FE E0 94 27 15 00 00 25 00 00 00 FD (scope) ; span=2500 Hz
                       //            |                       |____________|
                       //        iMsgLength=13               BCD value, HERE: LEAST significant digits first !
                       if( pbMessage[6] == 0x00 ) // only care for the "main scope" here...
                        { pRC->dblScopeSpan_Hz = RigCtrl_ParseBCD_CIV_LSByteFirst( pbMessage+7,
                               /* number of digits: */ 2*(iMsgLength/*13*/-8) );
                          SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "ScopeSpan=%ld Hz", (long)pRC->dblScopeSpan_Hz );
                        }
                     }
                    break;

                 case 0x16:  // "Edge number setting in the Fixed mode Scope"
                    if( iMsgLength <= 8 ) // no DATA .. must be a QUERY .
                     { SL_AppendString( &pszCmtDest, cpCmtEndstop, "Scope edge# ?" );
                     }
                    else
                     { if( pbMessage[6] == 0x00 ) // only care for the "main scope" here...
                        { pRC->iScopeEdgeNumber = pbMessage[7]; // range : 1..3 (per BAND.. what a strange concept)
                          SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "ScopeEdge=%d", pRC->iScopeEdgeNumber );
                        }
                     }
                    break;

                 case 0x17:  // "Send/read the Scope hold function status"
                    if( iMsgLength <= 8 ) // no DATA .. must be a QUERY .
                     { SL_AppendString( &pszCmtDest, cpCmtEndstop, "Scope hold ?" );
                     }
                    else
                     { if( pbMessage[6] == 0x00 ) // only care for the "main scope" here...
                        { pRC->iScopeHoldFlag = pbMessage[7]; // 0=hold OFF, 1=hold ON
                          SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "ScopeHold=%d", pRC->iScopeHoldFlag );
                        }
                     }
                    break;

                 case 0x18:  // "Send/read the Scope Attenuator setting" (IC-7851, n/a on IC-7300)
                    if( iMsgLength <= 8 ) // no DATA .. must be a QUERY .
                     { SL_AppendString( &pszCmtDest, cpCmtEndstop, "Scope att ?" );
                     }
                    else
                     { if( pbMessage[6] == 0x00 ) // only care for the "main scope" here...
                        { pRC->dblScopeAttenuator_dB = RigCtrl_ParseBCD_CIV_LSByteFirst( pbMessage+7,
                               /* number of digits: */ 2*(iMsgLength-8) );
                          SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "ScopeAtt=%.1lf dB", pRC->dblScopeAttenuator_dB );
                        }
                     }
                    break;

                 case 0x19:  // "Send/read the Scope Reference level setting"
                    if( iMsgLength <= 8 ) // no DATA .. must be a QUERY from someone else.
                     { SL_AppendString( &pszCmtDest, cpCmtEndstop, "Scope ref ?" );
                     }
                    else
                     { // Response seen from an IC-7300 :
                       // pbMessage[0  1  2  3  4  5  6  7  8  9 10]
                       // > RX 00B FE FE E0 94 27 19 00 10 50 00 FD   (scope)    ; ref=+10.5 dB
                       if( pbMessage[6] == 0x00 ) // only care for the "main scope" here...
                        { pRC->dblScopeRefLevel_dB = 0.01f * (float)RigCtrl_ParseBCD_CIV_MSByteFirst( pbMessage+7, 4/*digits*/ );
                          if( pbMessage[9] == 0x01 )  // 0x00 = plus, 0x01 = minus
                           { pRC->dblScopeRefLevel_dB = -pRC->dblScopeRefLevel_dB;
                           }
                          SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "ScopeRef=%.1lf dB", pRC->dblScopeRefLevel_dB );
                        }
                     }
                    break;

                 case 0x1A: // "Send/read the Sweep speed setting"
                    if( iMsgLength <= 8 ) // no DATA .. must be a QUERY .
                     { SL_AppendString( &pszCmtDest, cpCmtEndstop, "Scope speed ?" );
                       // Seen from RS-BA1 :
                       //  > r1 008 FE FE 94 E0 27 1A 00 FD (scope)    ; Scope speed ?
                     }
                    else
                     { // Response seen from an IC-7300 :
                       // pbMessage[0  1  2  3  4  5  6  7  8  9 10]
                       //  > RX 009 FE FE E0 94 27 1A 00 00 FD  (scope) ; Speed:FAST
                       if( pbMessage[6] == 0x00 ) // only care for the "main scope" here...
                        { switch( pbMessage[7] )  // 0x00 = fast, 0x01 = mid, 0x02 = slow
                           { case 0x00 : pRC->iScopeSpeed = RIGCTRL_SCOPE_SPEED_FAST;
                                  SL_AppendString( &pszCmtDest, cpCmtEndstop, "Scope:FAST" );
                                  break;
                             case 0x01 : pRC->iScopeSpeed = RIGCTRL_SCOPE_SPEED_MID;
                                  SL_AppendString( &pszCmtDest, cpCmtEndstop, "Scope:MID" );
                                  break;
                             case 0x02 : pRC->iScopeSpeed = RIGCTRL_SCOPE_SPEED_SLOW;
                                  SL_AppendString( &pszCmtDest, cpCmtEndstop, "Scope:SLOW" );
                                  break;
                             default   : // ???
                                  break;
                           }
                        }
                     }
                    break;

                 case 0x1B: // "Send/read the Scope indication during TX in the Center mode" (0=off, 1=on)
                    if( iMsgLength <= 7/*!*/ ) // no DATA .. must be a QUERY .
                     { // Seen from RS-BA1 : > r1 007 FE FE 94 E0 27 1B FD
                       SL_AppendString( &pszCmtDest, cpCmtEndstop, "Scope on TX ?" );
                     }
                    else
                     { // Response seen from an IC-7300 :
                       // pbMessage[0  1  2  3  4  5  6  7  8  9 10]
                       //  > RX 008 FE FE E0 94 27 1B 00 FD
                       //                             |
                       //                          Here, this isn't the "main/sub"-scope-flag
                       //                          but the OFF/ON status !
                       SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "Scope on TX:%s", (pbMessage[6]==0) ? "OFF":"ON" );
                     }
                    break;

                 case 0x1C: // "Send/read scope center frequency setting in the Center mode"
                    // 00=Filter center, 01=Carrier point center, 02=Carrier point center (Abs. Freq.))
                    // Seen shortly after launching RS-BA1 :
                    //          [0  1  2  3  4  5  6  7]
                    // > TX 007 FE FE 94 E0 27 1C FD    (scope) ; CenterMode ?
                    // > RX 008 FE FE E0 94 27 1C 01 FD (scope) ; CenterMode=1
                    if( iMsgLength <= 7 ) // no DATA .. must be a QUERY .
                     { SL_AppendString( &pszCmtDest, cpCmtEndstop, "CenterMode ?" );
                     }
                    else      // message WITH data ..
                     { pRC->iScopeCenterMode = pbMessage[7]; // 0=RIGCTRL_SCOPE_CM_FILTER_CENTER / 1=CARRIER_POINT / 2=CARRIER_ABS (whatever that means)
                       SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "CenterMode=%d", pRC->iScopeCenterMode );
                     }
                    break;

                 case 0x1D: // "Send/read the Scope 'Video Band Width' setting" (narrow/wide)
                    // Seen shortly after launching RS-BA1 :
                    //          [0  1  2  3  4  5  6  7  8]
                    // > TX 008 FE FE 94 E0 27 1D 00 FD    (scope) ; CenterMode ?
                    // > RX 009 FE FE E0 94 27 1D 00 00 FD (scope) ; CenterMode=1
                    //                             |  |__ 0=narrow, 1=wide
                    //                             0=main scope, 1=sub scope
                    if( iMsgLength <= 7 ) // no DATA .. must be a QUERY .
                     { SL_AppendString( &pszCmtDest, cpCmtEndstop, "VideoBW ?" );
                     }
                    else      // message WITH data ..
                     {
                       if( pbMessage[6] == 0x00 ) // only care for the "main scope" here...
                        { pRC->iScopeVideoBW = pbMessage[7]; // 0=RIGCTRL_SCOPE_VBW_NARROR, 1=.._WIDE
                          SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "VideoBW=%d", pRC->iScopeVideoBW );
                        }
                     }
                    break;

                 case 0x1E: // "Send/read the Scope Fixed edge frequencies"
                    // There may be up to 12 selectable frequency ranges.
                    // Each range has THREE "edge numbers" (already in pRC->iScopeEdgeNumber).
                    // The command (read-response) contains RANGE, EDGE NUMBER,
                    //     and the two BCD-encoded edge frequencies (lower, higher).
                    // Which of the 12 frequency ranges is currenty in use, depends on the VFO frequency.
                    // To avoid having to read this repeatedly, only read the
                    // entire "table" ONCE, and store it in an array.
                    // Format :   (1) (2) (3) (4) (5) (6) (7) (8) (9) (10)(11)(12)
                    //           |X:X|0:X|X:X|X:X|X:X|X:X|0:0|X:X|X:X|X:X|X:X|0:0|
                    // range nr ---'   |   |   |   |   |   | | |   |   |   |   |
                    // edge nr (1..3)--' |<--- lower edge -->|<-- higher edge -->|
                    //                     |   |   |   |   | | |   |   |   |   |
                    //                    10+1 .........1GHZ |10+1 ...........1GHz
                    //                     Hz            100M| Hz              100MHz
                    // (expect those ZEROs in the 100 MHz- and 1 GHz to be used
                    //  sooner or later .. thinking about an IC-9700 with V/U/SHF)
                    // Icom's RS-BA1 reads out ALL these frequencies when starting:
                    // > r1 009 FE FE 94 E0 27 1E 01 01 FD  (scope)    ; 16..0 kHz
                    iRange = RigCtrl_ParseBCD_CIV_LSByteFirst( pbMessage+6, 2/*digits*/ );
                    iEdge  = RigCtrl_ParseBCD_CIV_LSByteFirst( pbMessage+7, 2/*digits*/ );
                    if( iMsgLength <= 9/*!*/ ) // no DATA, only RANGE- and EDGE-number .. must be a QUERY .
                     { SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "rng[%d],edg[%d] ?",(int)iRange, (int)iEdge );
                     }
                    else
                     {
                       iRange = RigCtrl_ParseBCD_CIV_LSByteFirst( pbMessage+6, 2/*digits*/ );
                       iEdge  = RigCtrl_ParseBCD_CIV_LSByteFirst( pbMessage+7, 2/*digits*/ );
                       if( (iRange>=1) && (iRange<=RIGCTRL_MAX_FIXED_EDGE_SCOPE_FREQ_RANGES)
                         &&(iEdge >=1) && (iEdge <=RIGCTRL_MAX_FIXED_EDGES_PER_FREQ_RANGE) )
                        { pRC->ScopeFreqRange[iRange-1][iEdge-1].dblFmin_Hz = RigCtrl_ParseBCD_CIV_LSByteFirst( pbMessage+8, 10/*digits*/ );
                          pRC->ScopeFreqRange[iRange-1][iEdge-1].dblFmax_Hz = RigCtrl_ParseBCD_CIV_LSByteFirst( pbMessage+13,10/*digits*/ );
                          SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "%ld..%ld kHz",
                             (long)(1e-3 * pRC->ScopeFreqRange[iRange-1][iEdge-1].dblFmin_Hz),
                             (long)(1e-3 * pRC->ScopeFreqRange[iRange-1][iEdge-1].dblFmax_Hz) );
                        }
                     }
                    break; // end case 0x27 0x1E ("Scope Fixed edge frequencies")

                 default : // purpose unknown ... future extensions ?
                    break;
               } // end switch < byte after 0x27 >
            } // end if < valid message length >
           break; // end case 0x27

        case 0x28: // Voice TX Memory (0=stop, 1 to 8 = memory #)
           iMsgType = RIGCTRL_MSGTYPE_OTHER;
           break; // end case 0x28

        case 0x29: // "directly specify main or sub band" (IC-7610)
           iMsgType = RIGCTRL_MSGTYPE_OTHER;
           break; // end case 0x29

        case 0xFA: // "NG" ("No-Good" or "FAilure" aka "NotOK"), indicates an error
           iMsgType = RIGCTRL_MSGTYPE_NOT_OK;
           if( iRigCtrlPort==RIGCTRL_PORT_RADIO ) // only care for "OK" / "Not OK" *FROM THE RADIO*...
            {
              pRC->iResponseCountdown_ms = 0; // don't wait any longer (there was a response, even though a BAD one)
              // Will usually get here when asking for something not supported by the radio.
              // Example: Asking an IC-7300 for the current SCOPE ATTENUATOR LEVEL :
              // > 21:50:22.7 TX 008 FE FE 94 E0 27 18 00 FD   (scope)  ; atten. ?
              // > 21:50:22.9 RX 006 FE FE E0 94 FA FD         (Not OK)
              if( pRC->i32ExpectedResponse_CmdAndSubcode >= 0 )
               { SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "Rejected cmd 0x%X", (int)pRC->i32ExpectedResponse_CmdAndSubcode );
                 // That's just the default comment. Human-friendly output below. Thus ...
                 pszCmtDest = szComment; // reset the "comment-building pointer" again
               }
              switch( pRC->i32ExpectedResponse_CmdAndSubcode ) // we know the answer ("NG"), but what was the question ?
               { case RCTL_COMBINE_CMD_AND_SUB(0x27,0x14) : // have asked for the spectrum scope's "Center/Fixed" setting
                    SL_AppendString( &pszCmtDest, cpCmtEndstop, "no scope-C/F" );
                    break;
                 case RCTL_COMBINE_CMD_AND_SUB(0x27,0x15) : // have asked for the spectrum scope's "Frequency Span"
                    SL_AppendString( &pszCmtDest, cpCmtEndstop, "no scope-span" );
                    break;
                 case RCTL_COMBINE_CMD_AND_SUB(0x27,0x16) : // have asked for the spectrum scope's "Edge Number"
                    SL_AppendString( &pszCmtDest, cpCmtEndstop, "no scope edges" );
                    break;
                 case RCTL_COMBINE_CMD_AND_SUB(0x27,0x17) : // have asked for the "Scope hold function status"
                    SL_AppendString( &pszCmtDest, cpCmtEndstop, "no scope-hold" );
                    break;
                 case RCTL_COMBINE_CMD_AND_SUB(0x27,0x18) : // have asked for the spectrum scope's "Attenuator Level",
                    // which for example the IC-7300 doesn't support.
                    SL_AppendString( &pszCmtDest, cpCmtEndstop, "no scope-att" );
                    break;
                 case RCTL_COMBINE_CMD_AND_SUB(0x27,0x1E) : // have asked for one of the scope's "frequency ranges":
                    // When deliberately exceeding the number of "scope frequency ranges"
                    //  ( > 13 for an IC-7300 ), the radio responded with "Not Good" :
                    // > 23:23:00.0 TX 009 FE FE 94 E0 27 1E 14 01 FD (scope)    ; edge[14][1] ?
                    // > 23:23:00.1 RX 006 FE FE E0 94 FA FD          (Not OK)
                    // If this happens, limit pRC->iScopeNumFreqRanges
                    //   and/or pRC->iScopeNumEdgesPerFreqRange
                    //   to what the radio *really* supports. But WHAT caused the "Not OK" ?
                    break; // end case "Not OK" for a SCOPE FREQUENCY RANGE (0x271E)
                 case RCTL_COMBINE_CMD_AND_SUB(0x1A,0x01) : // have asked for another "Band Stacking Registers":
                    // Obviously, what Icom calls "Frequency band codes" (rig-specific)
                    // is invalid, so we have exceeded the range of valid "Frequency band codes".
                    // More on that in RigCtrl_Handler(), case RIGCTRL_POLLSTATE_BAND_STACKING_REGS .
                    if( pRC->iParameterPollingState == RIGCTRL_POLLSTATE_BAND_STACKING_REGS )
                     { RigCtrl_FinishReadingBandStackingRegs( pRC );
                       SL_AppendString( &pszCmtDest, cpCmtEndstop, "End of Band Stacking Registers" );
                     }
                    break;
                 default:
                    SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "NotOK for 0x%X", (int)pRC->i32ExpectedResponse_CmdAndSubcode );
                    break;
              } // end switch( pRC->i32ExpectedResponse_CmdAndSubcode )
              pRC->i32ExpectedResponse_CmdAndSubcode = -1;  // after "Not OK" from the radio,
              // we're not expecting to receive the "wanted" parameter value or whatever.
            } // end if < "Not OK" received FROM THE RADIO (not on one of the client ports) >
           break; // end case 0xFA < receive "No-Good" / "Not OK">

        case 0xFB: // "OK" (success, "Fine Business") from the rig
           iMsgType = RIGCTRL_MSGTYPE_OK;
           // We will never know (exactly) which commands will be answered
           // with the SAME command-byte (in the response as in the request),
           // and which request will be answered with this simple "OK" (especially not when a CLIENT had
           // sent the request): The client may know more about CI-V than RigControl itself,
           // especially if a new CI-V command will be answered with "OK"
           // or the value of a read-out parameter, or whatever.
           // Solution: Similar as already used for response "No-Good" (above):
           if( iRigCtrlPort==RIGCTRL_PORT_RADIO ) // only care for "OK" / "Not OK" *FROM THE RADIO*...
            { pRC->iResponseCountdown_ms = 0; // don't wait any longer (the 0xFB-"OK" was a response)
              fIsExpectedResponse = TRUE;  // unspecific "OK" treated like an "expected response" !
              iMsgType |= RIGCTRL_MSGTYPE_FLAG_EXPECTED; // flag for the coloured live CI-V display
              if( pRC->i32ExpectedResponse_CmdAndSubcode >= 0 )
               { SL_AppendPrintf( &pszCmtDest, cpCmtEndstop, "'OK' for cmd 0x%X", (int)pRC->i32ExpectedResponse_CmdAndSubcode );
               }
              else
               { SL_AppendString( &pszCmtDest, cpCmtEndstop, "'OK'" );
               }
              pszCmtDest = szComment; // reset the "comment-building pointer" again
              if( (pRC->iParameterPollingState == RIGCTRL_POLLSTATE_TURN_RIG_ON )
                ||(pRC->iParameterPollingState == RIGCTRL_POLLSTATE_WAIT_TURN_ON ) )
               { fIsExpectedResponse = FALSE; // this WAS the expected response" for "turn on",
                      // '--> but don't use the default processing further below,
                      //      which would CLEAR pRC->iResponseCountdown_ms, etc.
                 pRC->i32ExpectedResponse_CmdAndSubcode = -1; // avoid misinterpreting the NEXT unspecific "OK" or "NOT OK".
                 pRC->iResponseCountdown_ms = 5000; // stay in RIGCTRL_POLLSTATE_WAIT_TURN_ON until the radio has booted up completely.
                 // Without this, an IC-7300 ignored the first few read-requests
                 // in RIGCTRL_POLLSTATE_PARAMS_FOR_RIGCTL :
                 // > 19:45:08.2 TX 09B FE FE FE FE FE FE FE .. ; turn ON
                 // > 19:45:08.3 RX 006 FE FE E0 94 FB FD       ; OK, turned on
                 // > 19:45:09.0 TX 007 FE FE 00 E0 19 00 FD    ; read RadioID
                 // > 19:45:11.0 TX 006 FE FE 00 E0 03 FD       ; read VFOfreq
                 // > 19:45:12.9 TX 007 FE FE 00 E0 1C 00 FD ; read Transmitting
                 // > 19:45:13.0 RX 008 FE FE E0 94 1C 00 00 FD ; Transmitting:0
                 //          '--> That's 13-8 = FIVE SECONDS after the "OK, turned on"
                 //               Obviously, the "FB" response for the "turn on"-command
                 //               was not sent by the main CPU, but from some kind
                 //               of 'standby controller' or whatever can communicate
                 //               via CI-V (and USB, or LAN, or WLAN) while the
                 //               main CPU is asleep, or busy from "booting" .

               }    // end if( pRC->iParameterPollingState == RIGCTRL_POLLSTATE_WAIT_TURN_ON )
            }      // end if < "FB" received on iRigCtrlPort==RIGCTRL_PORT_RADIO > ?
           break; // end case 0xFB ("Fine Business", "OK")

        case 0xFC: // Old CI-V manuals mention 0xFC ("FCH") as the 'Jammer code',
           // for example Icom's "CI-V Reference Manual ver 3.2" from 2002 [CIV_REF_MAN_V3_2]:
           // > The jammer code, FCH, prevents a message collision among radios and controller.
           // > During message transmission, a radio which is transmitting receives
           // > a transmitted message from itself to detect a message collision.
           // > If a message collision with another radio is detected, the radio
           // > halts message transmission, and checks that no other messages
           // > are transmitted on the CI-V bus line.
           // > When no other message is transmitted, the radio transmits the
           // > jammer code, 5 times as below.
           // >    ,----,----,----,----,----,
           // >    | FC | FC | FC | FC | FC |
           // >    '----'----'----'----'----'
           break; // end case 0xFC


        default:  // anything else is ignored here
           iMsgType = RIGCTRL_MSGTYPE_UNKNOWN; // this parser couldn't process the message !
           break;
      } // end switch < command code, pbMessage[4] >

     if( fIsExpectedResponse ) // "default handler" if this was the "expected response" :
      { // RigCtrl_Handler() has been waiting for THIS message..
        if( iRigCtrlPort==RIGCTRL_PORT_RADIO )
         { pRC->iResponseCountdown_ms = 0;   // don't wait any longer !
           pRC->i32ExpectedResponse_CmdAndSubcode = -1; // avoid misinterpreting the NEXT unspecific "OK" or "NOT OK".
           // (only RigCtrl_Handler() decides what to ask NEXT)
         }
      } // end if < parsed the EXPECTED response >

   } // end if( ! fIsEcho )
  else // received message is an ECHO, not a response, not a command :
   { iMsgType = RIGCTRL_MSGTYPE_ECHO; // for filtering via pRC->dwRejectMessages4Log
   }

  //--------------------------------------------------------------------------
  // At this point, finished parsing the RECEIVED message, and extracted some
  //                info (like message type, expected response type, etc).
  // Next: Decide what to do on reception, e.g. forward to another port, etc.
  //--------------------------------------------------------------------------

  //       ,----> bitwise combination of flags like
  //       |            ( 1 << RIGCTRL_MSGTYPE_ECHO     /*1*/ ),
  //       |            ( 1 << RIGCTRL_MSGTYPE_RADIO_ID /*2*/ ),
  //       |            ( 1 << RIGCTRL_MSGTYPE_FREQUENCY_REPORT /*3*/ ), etc.
  //       |   Customizeable via 'Settings' .. 'Rig Control Options' .
  if( pRC->dwRejectMessages4Log & (1<<(iMsgType & 31)) )
   { // ,------------------------------|______|
     // '--> this may be iMsgType = 32 = RIGCTRL_MSGTYPE_OK      (the 0xFB thing)
     //               or iMsgType = 64 = RIGCTRL_MSGTYPE_NOT_OK  (the 0xFA thing)
     //      which shall NOT be rejected from the CAT traffic log :
     if( iMsgType & ( RIGCTRL_MSGTYPE_OK | RIGCTRL_MSGTYPE_NOT_OK ) )
      { // keep fShowInLog unchanged here (usually SET)
      }
     else // not the unspecific "OK" or "NOT OK" response :
      { fShowInLog = FALSE; // avoid flooding the CAT traffic log by certain message types, e.g. spectrum scope waveforms
      }
   }
  if( pRC->dwRejectMessages4Log & RIGCTRL_MSGTYPE_FLAG_PERIODIC_POLL )
   { // Reject any "periodic" messages in display, e.g. S-meter polling ?
     if( pPI != NULL ) // got a 'parameter info' ?
      { switch( pPI->iUnifiedPN )  // is it one of the "periodically polled" parameters ?
         { case RIGCTRL_PN_SQUELCH_STATUS     :
           case RIGCTRL_PN_S_METER_LEVEL_DB   :
           case RIGCTRL_PN_POWER_METER_PERCENT:
           case RIGCTRL_PN_SWR_METER_VALUE    :
           case RIGCTRL_PN_ALC_METER_LEVEL    :
           case RIGCTRL_PN_COMP_METER_LEVEL_DB:
           case RIGCTRL_PN_SUPPLY_VOLTAGE_mV  :
           case RIGCTRL_PN_DRAIN_CURRENT_mA   :
              fShowInLog = FALSE;
           default:
              break;
         }
      }
   }
  if( fShowInLog )  // BEFORE retransmitting anything, show the *received* message in the log:
   { RigCtrl_AddToTrafficLog( pRC, iRigCtrlPort, pbMessage, iMsgLength,
          iMsgType | RIGCTRL_MSGTYPE_FLAG_RX | (fIsExpectedResponse ? RIGCTRL_MSGTYPE_FLAG_EXPECTED : 0 ), // turn 'em green ?
          szComment );
   }

  if( iRigCtrlPort == RIGCTRL_PORT_RADIO ) // response received on the "RADIO" port ?
   { // May have to "forward" this message to one or more of the clients.
     // But some clients (like 'Hamlib' used by WSJT) are irritated by ANYTHING
     //       THEY DO NOT EXPECT (-> WSJT-X : "Ploink... RIG CONTROL ERROR").
     if( ! fIsEcho )
      { RigCtrl_CheckAndForwardResponseToClients( pRC, pbMessage, iMsgLength,
                      &msg_filter, iMsgType, fIsUnsolicitedMsg, szComment );
      }
   }
  else // iRigCtrlPort != RIGCTRL_PORT_RADIO -> command from one of the CLIENT ports ?
   { // "Echo-back" (Icom slang) the RECEIVED message (on the port on which it was received) ?
     //   Some applications developed when CI-V was a simple SINGE-WIRE interface
     //   still expect their transmitted data echoed back, possibly for collision detection .
     if( pPort->fEmulateCIVEcho ) // <- this may have been turned on/off further above via CI-V itself !
      { RigCtrl_SendAndLogMessage( pRC, iRigCtrlPort, pbMessage, iMsgLength,
              FALSE/*don't wait for completion*/,
              RIGCTRL_MSGTYPE_ECHO | RIGCTRL_MSGTYPE_FLAG_TX, szComment );
      } // end if( pPort->fEmulateCIVEcho )

     // Forward a client's command to the RADIO port immediately,
     //  or store it in a FIFO (to retransmit it LATER), ignore it, or what else ?
     // Here's ONE example of forwarding CI-V messages, with WSJT-X connected to client #2 :
     // > r2 006 FE FE 94 E0 03 FD                 (VFO freq)                [WSJT-X to SpecLab] -> RIGCTRL_CLIENT_ACTION_RETRANSMIT
     // > TX 006 FE FE 94 E0 03 FD                 (VFO freq)                [SpecLab to IC7300]
     // > RX 00B FE FE E0 94 03 00 70 56 03 00 FD  (VFO freq) ; 3567000.0 Hz [IC7300 to SpecLab] -> RIGCTRL_CLIENT_ACTION_RETRANSMIT
     // > t2 00B FE FE E0 94 03 00 70 56 03 00 FD  (VFO freq) ; 3567000.0 Hz [SpecLab to WSJT-X]
     if(  (iClientAction==RIGCTRL_CLIENT_ACTION_RETRANSMIT )
       && (pRC->i32ExpectedResponse_CmdAndSubcode >= 0 ) ) // oops, cannot IMMEDIATELY retransmit command from client to radio !
      { iClientAction = RIGCTRL_CLIENT_ACTION_ENTER_FIFO; // .. so send whatever the client wants a few milliseconds LATER .
        // 2019-01-05 : Got here with pRC->i32ExpectedResponse_CmdAndSubcode = 0x0701,
        //     shortly after being "connected" by WSJT-X on a client port :
        // > r2 007 FE FE 94 E0 07 01 FD (other)    ; SelVFO:01 [WSJT-X to SpecLab]
        // > t2 007 FE FE 94 E0 07 01 FD (echo)     ; SelVFO:01 [immediate echo to make Hamlib happy]
        // > TX 007 FE FE 94 E0 07 01 FD (other)    ; SelVFO:01 [SpecLab to IC7300, i32ExpectedResponse_CmdAndSubcode now 0x0701]
        // > RX 007 FE FE 94 E0 07 01 FD (echo)                 [IC7300 to SpecLab, not the expected response yet]
        // > RX 006 FE FE E0 94 FB FD    (OK)                   [IC7300 to SpecLab, response to cmd 0x07]
        // > t2 006 FE FE E0 94 FB FD    (OK)                   [SpecLab to WSJT-X]
        // > r2 006 FE FE 94 E0 03 FD    (VFO freq)
        // Solution: See 'case 0xFB' (="OK") in RigCtrl_ParseCIV(), further above.
      } // end if < "wanted" to retransmit, but cannot because still waiting for a response on the RADIO port >
     switch( iClientAction )
      { case RIGCTRL_CLIENT_ACTION_NONE      : // "do nothing" (igore command from 3rd party client)
           break;
        case RIGCTRL_CLIENT_ACTION_RETRANSMIT: // immediately retransmit command from client to radio,
           // but also REGISTER THIS TYPE OF MESSAGE to forward the response to the client later:
           RigCtrl_RegisterCIVReponseForClient( pRC, iRigCtrlPort/*client-port*/, &msg_filter );
           if( pRC->pRadioPortTxFifo != NULL )
            { // Mark the "radio port" as occupied, until the response
              //      for the forwarded command (sent below) arrives :
              pRC->i32ExpectedResponse_CmdAndSubcode = i32CombinedCmdAndSubcode;
              RigCtrl_SendAndLogMessage( pRC, RIGCTRL_PORT_RADIO, pbMessage, iMsgLength,
                    TRUE/* wait for completion (*) ? */, iMsgType, szComment );
                    // (*) Tried WITH and WITHOUT "waiting for completion",
                    //     because WITH waiting, the worker thread for RigControl
                    //     was seriously slowed down when forwarding SPECTRUM DATA
                    //     to RS-BA1, *AFTER* RS-BA1 had disconnected from the
                    //     other end of the virtual Null-modem cable (com0com).
                    //     Since we RECEIVED these data over a serial port,
                    //     we should be able to RETRANSMIT *without* waiting.
                    //     But in the horrible "serial port API" in Windows,
                    //     things are not as simple as they should be,
                    //     and a thread can only wait for ONE outstanding
                    //     'FileWrite'-call (here: for a single CI-V message).
                    // With fWait=FALSE above, Spectrum Lab's waterfall wasn't
                    //     slowed down anymore, but the waterfall in RS-BA1
                    //     didn't work anymore, most likely because the
                    //     TRANSMITTED data (sent from here) were corrupted.
                    //   Got to emulate a nice simple reliable CIRCULAR TRANSMIT FIFO
                    //   (as in a microcontroller firmware) for Windoze one fine day !
            } // end if < address of the transmit-function available > ?
           break;
        case RIGCTRL_CLIENT_ACTION_ENTER_FIFO: // store command in FIFO, retransmit or execute later
           // NOT YET: RigCtrl_RegisterCIVReponseForClient( pRC, iRigCtrlPort/*client-port*/, &msg_filter );
           // NOT YET: RigCtrl_SendAndLogMessage( pRC, RIGCTRL_PORT_RADIO, ..) // forward to radio
           RigCtrl_AppendMessageToRadioPortTxFifo( pRC,
                iRigCtrlPort/*"transmit on behalf of THIS client"*/,
                pbMessage, iMsgLength );
           break;
        case RIGCTRL_CLIENT_ACTION_FAKE_RESPONSE: // immediately send a "fake response" to the client without forwarding anything to the radio
           RigCtrl_SendAndLogMessage( pRC, iRigCtrlPort/*client-port*/,
                    pPort->bTxMsg, pPort->iTxMsgLength,
                    TRUE/*wait for completion*/, iMsgType, szComment );
           break;
        default:
           break;
      } // end switch( iClientAction )
   }   // end else < command received on one of the CLIENT ports >


  return iMsgType;

} // end RigCtrl_ParseCIV()

//--------------------------------------------------------------------------
void RigCtrl_ProcessRxData(   // tries to de-packetize a 'serial bytestream'.
        // For CI-V, looks for the 0xFE 0xFE preamble, and the 0xFD postamble, etc..
        T_RigCtrlInstance *pRC,  // [in] 'Rig Control' instance (so far, only one, but who knows..)
        int  iRigCtrlPort )  // [in] RIGCTRL_PORT_RADIO / RIGCTRL_PORT_CLIENT_1 / 2 ?
{ // Called from the main threads ("GUI thread") after reception of a block of data.
  //        Initially only for data received from a serial port ("COM"),
  //        later (but not HERE) also used for CI-V embedded in UDP packets for LAN or WLAN.
  int iSource, nBytesReceived = 0;
  BYTE bTemp[1024], b;
  T_RigCtrl_PortInstance *pPortInstance;

  if( pRC == NULL ) // oops.. instance deleted but some thread is still calling us !
   { return;
   }
  if( pRC->initState != initState_Opened ) // oops.. didn't call RigCtrl_Init() yet, or called RigCtrl_Close()
   { return;
   }

  if( (iRigCtrlPort<0) || (iRigCtrlPort>/*!*/RIGCTRL_MAX_CLIENT_PORTS) )
   { // iRigCtrlPort = 0 = RIGCTRL_PORT_RADIO = communication port that "talks to the radio",
     // uses the same subroutine to process received data !
     return;  // not a valid 'port identifier' (not to be confused with a COM-port-number) !
   }

  pPortInstance = &pRC->PortInstance[iRigCtrlPort];

  // the keyer thread may have placed received bytes in a thread-safe FIFO.
  // Pass the content of that FIFO to RigControl, with a CI-V parser, etc:
  if( (iRigCtrlPort==RIGCTRL_PORT_RADIO) && (pRC->pRadioPortRxFifo!=NULL) )
   { nBytesReceived = CFIFO_Read( pRC->pRadioPortRxFifo, bTemp, sizeof(bTemp) );
   }

  // Convert the received data from the radio ("CI-V" or whatever protocol)
  //    into the an internal buffer, and parse it as soon as a "message" is complete:
  for(iSource=0; iSource<nBytesReceived; ++iSource)
   {
      b = bTemp[iSource];  // get the next BYTE (!) from the serial input stream
      // Collect all received bytes in an extra buffer until the end of the message is reached:
      if( pPortInstance->iRcvdMsgLength < RIGCTRL_MAX_MESSAGE_LENGTH )
       {  pPortInstance->bRcvdMsg[pPortInstance->iRcvdMsgLength++] = b;
       }
      else // couldn't append the new byte, count it as "garbage"
       {  ++pPortInstance->dwNumGarbageBytes;
       }

      // Check for end-of-message (depends on the protocol; radio-specific)
      switch( pRC->iRadioCtrlProtocol )
       {
        case RIGCTRL_PROTOCOL_NONE :   // "no" protocol, assume "normal text"..
        default :
             // in this case, "zero" bytes, carriage return, or new line are treated as end-markers:
             if( b==0 || b==10 || b==13 )
              {
                pPortInstance->iRcvdMsgLength = 0;   // begin receiving next frame
              }
             break;
        case RIGCTRL_PROTOCOL_ICOM_CI_V  :
             if( b==0xFD )  // received CI-V code for "End Of Message"
              {
               if( (pPortInstance->bRcvdMsg[0]==0xFE) && (pPortInstance->bRcvdMsg[1]==0xFE) )
                { // preamble code ok ?
                  if(  (pPortInstance->bRcvdMsg[2]==pRC->iRadioMasterAddr)
                     ||(pPortInstance->bRcvdMsg[2]==pRC->iRadioDeviceAddr) // listen to the "echo" too !
                     ||(pPortInstance->bRcvdMsg[2]==0x00) ) // directed to the "broadcast" address
                   { // Messages generated by a rig by turning the dial or by changing
                     // the mode are sent to this adress 0x00 (provided that "CI-V transceive"
                     // is set to ON. If OFF, no messages are generated by the rig).
                     // don't care for IO_b255MsgFromRadio[2] = transceiver's address (0x48 = IC706)
                     // and  for IO_b255MsgFromRadio[2] = controller's address (often 0xE0) .
                     // The "from"-address (=IO_b255MsgFromRadio[3] is ignored here .
                     ++pPortInstance->dwNumMessagesRcvd;
                     RigCtrl_ParseCIV( pRC, iRigCtrlPort, pPortInstance->bRcvdMsg, pPortInstance->iRcvdMsgLength );
                   } // end if < matching "address" >
                  else // pre- and postamble ok, but something wrong with the addresses
                   { pPortInstance->dwNumGarbageBytes += pPortInstance->iRcvdMsgLength;
                   }
                 } // end if < preamble code ok >
                pPortInstance->iRcvdMsgLength = 0;   // begin receiving next frame
              } // end if < received "end-of-message" code >
             else if( (b==0xFE) && (pPortInstance->bPrevRcvdByte==0xFE) )
              { // "Preamble code (fixed)" :
                // Expect the NEXT byte (at offset 2) to be the
                //        "Transceiver's default address"
                // or the "Controller's default address" .
                //      (more on those "default addresses", including a LIST, in RigCtrl_RadioInfo_CIV[] )
                //
                // The IC-7610 CI-V REFERENCE GUIDE, A7380-7EX-1 available from Icom,
                // clearly states that MORE THEN TWO of these 'preamble codes'
                // may be sent. Example:
                // > When sending the power ON command (18 01), you need
                // > to repeatedly send "FE" before the standard format.
                // (at 115200 bits/second, the receiver expects 150 (!) "FE"s)
                // When Icom's "RS-BA1" is at the other end of a com0com connection
                // (thinking it's talking to an Icom radio but in fact talking to us),
                // got here with a huge pile of 0xFE's before the first "real" message.
                if( pPortInstance->iRcvdMsgLength > 3 ) // oops.. something in between that was NOT 0xFE
                 {  pPortInstance->dwNumGarbageBytes += (pPortInstance->iRcvdMsgLength-3);
                 }
                pPortInstance->iRcvdMsgLength = 2; // discard anything before the "Preamble code"
              }

             break; // end case C_RADIO_CTRL_PROT_ICOM_CI_V
       } // end switch( pRC->iRadioCtrlProtocol )
      // For "decoder" of received messages (from radio, most likely CI-V protocol)

     pPortInstance->bPrevRcvdByte = b;

   }  // end for <all bytes from serial data block>


} // end RigCtrl_ProcessRxData()


//--------------------------------------------------------------------------
void RigCtrl_OnVFOReportFromNetwork( T_RigCtrlInstance *pRC,
        T_RigCtrl_VfoReport *pVfoReport) // [in] received network packet
  // Called from CwNet.c on reception of a packet with CWNET_CMD_FREQ_REPORT.
  // First used in the 'Remote CW Keyer', CLIENT SIDE (without a real radio),
  // where a 'Rig Control' instance only stores parameters for the display in
  // the GUI, but doesn't control a locally connected radio.
  // [out] pRC->dblVfoFrequency,
  //       pRC->iFrequencyModifiedByRadio_cnt,
  //       and maybe a few other parameters that would NORMALLY be updated
  //       on reception of an unsolicited CI-V message in RigCtrl_ProcessRxData().
  // Note: RigCtrl_OnVFOReportFromNetwork() is used by the Remote CW Keyer
  //       on the CLIENT as well as on the SERVER side. Only the SERVER side
  //       is locally connected to a radio (and passes on the new
  //       frequency).
  //       When running as a CLIENT (not controlling a 'local radio')
  //       the new VFO frequency is only stored in pRC->dblVfoFrequency
  //       (from where the GUI thread can poll it to update the display).
  //   The same note applies to similar handlers like
  //   RigCtrl_OnParamReport_Byte/Double/String(), implemented further below.
{
  if( pVfoReport != NULL )
   { if( pVfoReport->bStructSize >= 8 ) // Struct with the MINIMUM SIZE ?
      { if( (pVfoReport->dblVfoFrequency != pRC->dblVfoFrequency )
          &&(pVfoReport->dblVfoFrequency != RIGCTL_NOVALUE_DOUBLE ) )
         { pRC->dblVfoFrequency = pVfoReport->dblVfoFrequency;
           ++pRC->iFrequencyModifiedByRadio_cnt;
           // This counter is polled in a GUI-thread somewhere.
           // If this 'rig control instance' really controls a radio,
           //    pass the requested "new frequency" on to that radio:
           if( (pRC->initState == initState_Opened )
             &&(pRC->iDefaultAddress > RIGCTRL_DEF_ADDR_NO_REMOTE_CTRL) ) // not in "passive" mode ?
            { RigCtrl_SetVFOFrequency_Internal( pRC, pRC->dblVfoFrequency ); // send command to QSY to the real radio
            }
         }
        if( pVfoReport->i8SMeterLevel_dB != -128 )
         { pRC->iSMeterLevel_dB = pVfoReport->i8SMeterLevel_dB;
           // '--> polled and displayed in the GUI-thread
         }
      }
   }
} // end RigCtrl_OnVFOReportFromServer()

//--------------------------------------------------------------------------
CPROT void RigCtrl_OnMultiFunctionMeterReportFromNetwork(T_RigCtrlInstance *pRC,
               T_RigCtrl_MultiFunctionMeterReport *pMeterReport)
  // Called from CwNet.c on reception of a packet with CWNET_CMD_MULTI_FUNCTION_METER_REPORT.
  // First used in the 'Remote CW Keyer', CLIENT SIDE (without a real radio),
  // where a 'Rig Control' instance only stores parameters for the display in
  // the GUI, but doesn't control a locally connected radio.
  // [out] pRC->iPowerMeterLevel_pcnt, pRC->dblSWRMeterValue,
  //       pRC->iALCMeterLevel_pcnt,   pRC->iCompMeterLevel_dB,
  //       pRC->iSupplyVoltage_mV,     pRC->iDrainCurrent_mA,
  //       pRC->iPATemperature_degC .  ( if, one day, we know how to read this via CI-V)
{
  if( pMeterReport != NULL )
   { if( pMeterReport->bStructSize >= 8 ) // Struct with the MINIMUM SIZE (sizeof(T_RigCtrl_MultiFunctionMeterReport) in its oldest version)
      { pRC->iPowerMeterLevel_pcnt = pMeterReport->i8PowerMeterLevel_pcnt;
        pRC->dblSWRMeterValue      = 0.1 * (double)pMeterReport->i8SWRMeterValue_x10;
        pRC->iALCMeterLevel_pcnt   = pMeterReport->i8ALCMeterLevel_pcnt;
        pRC->iCompMeterLevel_dB    = pMeterReport->i8CompMeterLevel_dB;
        pRC->iSupplyVoltage_mV     = pMeterReport->u8SupplyVoltage_100mV * 100;
        pRC->iDrainCurrent_mA      = pMeterReport->u8DrainCurrent_100mA  * 100;
        pRC->iPATemperature_degC   = pMeterReport->u8PATemperature_degC;
      }
   }
} // end RigCtrl_OnMultiFunctionMeterReportFromNetwork()

//--------------------------------------------------------------------------
void RigCtrl_OnBandStackingRegisterReportFromNetwork( T_RigCtrlInstance *pRC, const char *pszSource )
  // Called from CwNet.c on reception of a packet with CWNET_CMD_BAND_STACKING_REGISTER.
  // First used in the 'Remote CW Keyer', CLIENT SIDE (without a real radio),
  // where a 'Rig Control' instance only stores parameters for the display in
  // the GUI (in this case, the combined "band selection combo" / "band-stacking registers"),
  // but doesn't control a locally connected radio.
  // [out] pRC->BandStackingRegs[RIGCTRL_NUM_BAND_STACKING_REGS],
  //       pRC->iNumBandStackingRegs (if the parsed ARRAY INDEX exceeds this)
  //       pRC->fBandStackingRegsModified  (flag possibly polled by the GUI to redraw itself if necessary)
{ T_RigCtrlFreqMemEntry freq_mem_entry;
  int iArrayIndex = 0;
  RigCtrl_InitStruct_FreqMemEntry( &freq_mem_entry );
  if( RigCtrl_StringToFreqMemEntry(&freq_mem_entry, &iArrayIndex, pszSource) > 0 )
   { if( (iArrayIndex>=0) && (iArrayIndex<RIGCTRL_NUM_BAND_STACKING_REGS) )
      { if( memcmp( &freq_mem_entry, &pRC->BandStackingRegs[iArrayIndex],
                    sizeof(freq_mem_entry) ) != 0 ) // MODIFIED ? Copy and inform the GUI..
         { pRC->BandStackingRegs[iArrayIndex] = freq_mem_entry;
           if( pRC->iNumBandStackingRegs <= iArrayIndex )
            {  pRC->iNumBandStackingRegs =  iArrayIndex+1;
            }
           pRC->fBandStackingRegsModified = TRUE;
         }
      }
   }
} // end RigCtrl_OnBandStackingRegisterReportFromNetwork()

//--------------------------------------------------------------------------
void RigCtrl_OnParamReport_Integer( T_RigCtrlInstance *pRC,
        T_RigCtrl_ParamReport_Integer *pReport) // [in] received network packet
  // Called from CwNet.c on reception of a packet with CWNET_CMD_PARAM_INTEGER.
  // For example, that may be a command to switch the 'operation mode' (RIGCTRL_PN_OP_MODE),
  //              which RigControl.c translates into an Icom CI-V message
  //              and send it to the server's "locally controlled radio".
  //              Note:
{ T_RigCtrl_ParamInfo *pPI = RigCtrl_GetInfoForUnifiedParameterNumber(pReport->bRigControlParameterNumber);
  BOOL fModified = FALSE;
  if( pPI != NULL )
   { fModified = RigCtrl_SetParamValue_Int( pRC, pPI, pReport->i32ParameterValue ); // only writes into pRC, doesn't cause CI-V command for the locally connected radio !

     // Only a few parameter numbers will be translated into e.g. a CI-V command
     // to "set" something inside the locally connected radio:
     if( fModified )
      { switch( pPI->iUnifiedPN )
         { case RIGCTRL_PN_OP_MODE: // switch the locally connected radio's "Op-Mode" on reception of a T_RigCtrl_ParamReport_Integer from CwNet.c :
                pRC->iPostponedSetMessages |= (1<<RIGCTRL_MSGTYPE_OP_MODE);
                pRC->iPostponedOpMode = pReport->i32ParameterValue;
                break;
           default:
                break;
         }
      }
   }
} // end RigCtrl_OnParamReport_Integer()


//--------------------------------------------------------------------------
void RigCtrl_OnParamReport_Double( T_RigCtrlInstance *pRC,
        T_RigCtrl_ParamReport_Double *pReport) // [in] received network packet
  // Called from CwNet.c on reception of a packet with CWNET_CMD_PARAM_DOUBLE.
{ T_RigCtrl_ParamInfo *pPI = RigCtrl_GetInfoForUnifiedParameterNumber(pReport->bRigControlParameterNumber);
  if( pPI != NULL )
   { RigCtrl_SetParamValue_Double( pRC, pPI, pReport->dblParameterValue );
   }
} // end RigCtrl_OnParamReport_Double()

//--------------------------------------------------------------------------
void RigCtrl_OnParamReport_String( T_RigCtrlInstance *pRC,
        T_RigCtrl_ParamReport_String *pReport) // [in] received network packet
  // Called from CwNet.c on reception of a packet with CWNET_CMD_PARAM_STRING.
{ T_RigCtrl_ParamInfo *pPI = RigCtrl_GetInfoForUnifiedParameterNumber(pReport->bRigControlParameterNumber);
  if( pPI != NULL )
   {
   }
} // end RigCtrl_OnParamReport_String()



//--------------------------------------------------------------------------
char* RigCtrl_MsgTypeToString(int iMsgType)
{ // Mainly used for dislay in Spectrum Lab's "CAT traffic log".
  // Keep the returned strings short (ideally 7 or 8 characters).
  static char sz10Result[16]; // If it's one of the common types, don't use this evil static buffer !

  switch( iMsgType & RIGCTRL_MSGTYPE_STRIP_FLAGS )
   {
     case RIGCTRL_MSGTYPE_INFO     :
        return "info";
     case RIGCTRL_MSGTYPE_ECHO     :
        return "echo";
     case RIGCTRL_MSGTYPE_RADIO_ID :
        return "radio ID";
     case RIGCTRL_MSGTYPE_FREQUENCY_REPORT: // "frequency report" (usually with the VFO frequency)
        return "VFO freq";
     case RIGCTRL_MSGTYPE_SET_FREQUENCY: // "SET frequency" (command from user to radio)
        return "set freq";
     case RIGCTRL_MSGTYPE_OP_MODE  : // "mode data" (often unsolicited, "on change")
        return "op mode";
     case RIGCTRL_MSGTYPE_RIT_XIT  : // RIT or XIT data (often unsolicited, "on change")
        return "RIT/XIT";
     case RIGCTRL_MSGTYPE_RF_GAIN  : // anything that affects the RF gain (preamp, attenuator, "AIP", ..)
        return "RF gain";
     case RIGCTRL_MSGTYPE_C_SPECTRUM: // "waveform data" (in fact, for Icom's spectrum scope)
        return "scope";              // "spectrum scope control" (not WAVEFORM DATA)
     case RIGCTRL_MSGTYPE_SPECTRUM : // "waveform data" (in fact, for Icom's spectrum scope)
        return "wform";              // short info because the hex field is long
     case RIGCTRL_MSGTYPE_POWER_ON_OFF:
        return "pwr ctl";
     case RIGCTRL_MSGTYPE_OTHER    : // acknowledge for 'set', etc pp.
        return "other";
     case RIGCTRL_MSGTYPE_OK       : // in CI-V : unspecific "OK", cmd 0xFB
        return "OK";
     case RIGCTRL_MSGTYPE_NOT_OK   : // in CI-V : unspecific "NG" / "Not Good", cmd 0xFA
        return "Not OK";
     case RIGCTRL_MSGTYPE_UDP_CONTROL:
        return "UDP Ctrl";
     case RIGCTRL_MSGTYPE_UDP_SERIAL :
        return "UDP Serial";
     case RIGCTRL_MSGTYPE_UDP_AUDIO:
        return "UDP Audio";
     case RIGCTRL_MSGTYPE_UNKNOWN  : // unknown CI-V or in fact UDP (used by Icom radios via LAN or WLAN)
        return "UNKNWN";
     default:
        sprintf( sz10Result, "?%d?", (int)iMsgType );
        return sz10Result;
   } // end switch( iMsgType )

} // end RigCtrl_MsgTypeToString()

//---------------------------------------------------------------------------
T_RigCtrl_RadioInfo *RigCtrl_GetRadioInfoByDefaultAddress( int iRadioCtrlProtocol, int iDefaultAddress )
{
  T_RigCtrl_RadioInfo *pRadioInfo = NULL;

  // For future extensions (Yaesu CAT protocol ?), there may be
  // different tables, not only the one for radios with Icom-compatible CI-V !
  switch( iRadioCtrlProtocol )
   {
     case RIGCTRL_PROTOCOL_NONE      : // if protocol not set yet, assume it's an Icom
     case RIGCTRL_PROTOCOL_ICOM_CI_V :
        pRadioInfo = (T_RigCtrl_RadioInfo *)RigCtrl_RadioInfo_CIV;
        break;
     case RIGCTRL_PROTOCOL_KENWOOD   : // Sorry folks... Kenwood not supported yet
     case RIGCTRL_PROTOCOL_YAESU_CAT : // Yeasu's "CAT" is a nightmare.. every radio is incompatible
         // (They don't give a shit for compatibility between Yasu radios :
         // > Unfortunately, there is no "master list" of Yaesu CAT commands.
         // > Most Yeasu models have unique command sets; in some cases these
         // > are similar to those of predecessors, in other cases
         // > they don't even use the same bit ordering.
         // > There are also ROM updates that introduce new commands or change
         // > the behaviour of existing commands. Yeasu's CAT documentation
         // > is rarely comprehensive or error free. (..)
         // > My current theory is that Yaesu engineers learned about computers
         // > and software from reading the backs of cereal boxes.
         //     Looks like Dave, AA6YQ, has also had a lot of fun with Yaesu.
         //     Or, as Simon HB9DRV writes in his 'Command Tester' user guide:
         // > Be aware that there are many errors in the Yaesu
         // > documentation, so the data returned may not agree with the
         // > handbook and the radios current settings.
         break;
     case RIGCTRL_PROTOCOL_AD9832_I2C  : // not supported HERE but somewhere else (in SpecLab)
         break;
   } // end  switch( iRadioCtrlProtocol )


  if( pRadioInfo != NULL )
   {
     // The table isn't sorted by anything, so use a linear search:
     while( pRadioInfo->pszName != NULL )  // NULL marks the end of the table
      { if( pRadioInfo->iDefaultAddress == iDefaultAddress )
         {  return pRadioInfo; // bingo, found what we're looking for
         }
        ++pRadioInfo; // check next table entry
      }
   }

  // Arrived here ? Out of luck...
  return NULL;
} // end  RigCtrl_GetRadioInfoByDefaultAddress()

//---------------------------------------------------------------------------
const char* RigCtrl_DefaultAddressToString( int iRadioCtrlProtocol, int iDefaultAddress )
{
  static char sz7Result[8];
  T_RigCtrl_RadioInfo *pRadioInfo = RigCtrl_GetRadioInfoByDefaultAddress(
                                          iRadioCtrlProtocol, iDefaultAddress );
  if( pRadioInfo != NULL )
   { return pRadioInfo->pszName;
   }
  else
   { sprintf( sz7Result, "<%02X>", iDefaultAddress );
     return sz7Result;
   }
} // end RigCtrl_DefaultAddressToString()

//---------------------------------------------------------------------------
const char* RigCtrl_RigControlPortNrToString( int iRigCtrlPort )
{ switch( iRigCtrlPort )
   { case RIGCTRL_PORT_RADIO    : return "Radio Port";
     case RIGCTRL_PORT_CLIENT_1 : return "Client Port #1";
     case RIGCTRL_PORT_CLIENT_2 : return "Client Port #2";
#ifdef    RIGCTRL_PORT_CLIENT_3
     case RIGCTRL_PORT_CLIENT_3 : return "Client Port #3";
#endif
     default : return "Invalid port";
   }
} // end RigCtrl_RigControlPortNrToString()

//---------------------------------------------------------------------------
const char *RigCtrl_ControlConnStateToString( T_RigCtrlInstance *pRC ) // more verbose than RigCtrl_ConnStateToString(), but mainly for the "CONTROL"-UDP-port
  // [in] pRC->iConnState    : RIGCTRL_CONNSTATE_OFF/CONNECTING/CONNECTED
  // [in] pRC->iConnSubstate : RIGCTRL_CONN_SUBSTATE_SEND_CONN_REQUEST, etc^10 ..
  // [return] e.g. "Connected to IC-705" (ideally).
  // Called from TDebugForm::Btn_GetReportOnCommsClick(),
  //    and from TSpectrumLab::MI_StartStopClick() to show the current state in SL's MAIN MENU.
{ static char sz80Result[84];
  char *cp = sz80Result, *pszEndstop = sz80Result+80;

  if( pRC==NULL )
   { return "Deleted";
   }
  switch( pRC->initState )
   { case initState_ColdStart : return "NoInit";
     case initState_Closed    : return "Stopped";  // e.g. RigCtrl_Close() called from the main menu, "Start/Stop".."Rig Control".."Stop" (thus this result)
     default : break;   // e.g. initState = 1 = "Opened" ...
   }

  // If a "real" serial port (RS-232 or single-wire CI-V adapter), or a VIRTUAL COM PORT on USB is used,
  // there's no such complex "Connection" state machine as for UDP (with the "three ports").
  // In this case, we consider ourselves "connected" as soon as we RECEIVED
  // the first properly formatted CI-V message from an Icom radio.
  SL_AppendString( &cp, pszEndstop, RigCtrl_ParameterPollingStateToString(pRC->iParameterPollingState) );
  // ,--------------------------------'
  // '--> e.g. "passive" /  "start reading" / "read VFO freq" / "read op-mode" / "start spectrum" /  "running"
  if( pRC->PortInstance[RIGCTRL_PORT_RADIO].dwNumMessagesRcvd > 0 )
   { SL_AppendPrintf( &cp, pszEndstop, ", rcvd %ld msgs",
            (long)pRC->PortInstance[RIGCTRL_PORT_RADIO].dwNumMessagesRcvd );
   }
  else
   { SL_AppendString( &cp, pszEndstop, ", no RX yet" );
   }
  return sz80Result; // end case INET_PROTOCOL_COM_PORT
} // end RigCtrl_ControlConnStateToString()

//---------------------------------------------------------------------------
const char *RigCtrl_ParameterPollingStateToString(int iParameterPollingState)
  // Added 2023-06 when, despite being successfully "connected" to all three
  // UDP ports in an IC-705, none of the essential parameters became valid.
  // Called from, and results displayed in, Spectrum Lab's "CAT traffic monitor".
{ static char sz15Result[16];
  switch( iParameterPollingState )
   { case RIGCTRL_POLLSTATE_PASSIVE             : return "passive";        // initial state: have neither READ nor WRITTEN any setting from/to the remote rig
     case RIGCTRL_POLLSTATE_START_RD            : return "start reading";  // begin reading all we need to know about/from the remote rig
     case RIGCTRL_POLLSTATE_READ_RADIO_ID       : return "read radio ID";  // (try to) read the radio's "default" ID, as a connection test and to check its type
     case RIGCTRL_POLLSTATE_PARAMS_FOR_RIGCTL   : return "read initial params";  // (try to) read the required initial parameters
     case RIGCTRL_POLLSTATE_PARAMS_FOR_HAMLIB   : return "read Hamlib params"; // (try to) read a few more params used by HamlibServer.c
     case RIGCTRL_POLLSTATE_CONFIGURE_SPECTRUM  : return "start spectrum"; // (try to) activate transmission of "spectrum scope waveform data"
     case RIGCTRL_POLLSTATE_DONE                : return "running";        // finished polling "all we need to know" about the radio; periodic polling is "running" (S-meter, VFO, Spectrum, ?)
     case RIGCTRL_POLLSTATE_TURN_RIG_ON         : return "turn rig on";    // (try to) turn the rig on via special command or "wake-up pattern"
     case RIGCTRL_POLLSTATE_WAIT_TURN_ON        : return "wait for turn-on"; // waiting for an "OK"-response after the command to "turn on"
     case RIGCTRL_POLLSTATE_TURN_RIG_OFF        : return "turn rig off";   // (try to) turn the rig on via special command (for modern Icom rigs)
     case RIGCTRL_POLLSTATE_WAIT_TURN_OFF       : return "wait for turn-off";
     case RIGCTRL_POLLSTATE_TURNED_OFF          : return "rig turned off";
     default : sprintf( sz15Result, "?%d?", (int)iParameterPollingState );
               return sz15Result;
   }
} // end RigCtrl_ParameterPollingStateToString()

//---------------------------------------------------------------------------
int RigCtrl_FreqMemEntryToString( T_RigCtrlFreqMemEntry *pFreqMemEntry,
       int iArrayIndex, // [in] zero-based array index for easy parsing (negative for "no array")
       char *pszDest, int iMaxDestLen ) // [out] key=value pairs, comma separated
  // Returns the length of the generated string (not including the trailing zero).
  // In the remote CW keyer, this function is also used to send
  // the "band-stacking registers" from server to all clients in a non-Icom-
  // specific, easily parsable(*), extendable format.
  // See also ("inverse function", the parser) : RigCtrl_StringToFreqMemEntry() .
  // (*) parsable, not parseable, because (English teachers will love this):
  // > "Verbs drop silent final -e except after c and g (e.g. movable but changeable)"
{ char *pszDestStart = pszDest;
  char *pszEndstop = pszDest+iMaxDestLen;

  // Most important, "mandatory" members of the T_RigCtrlFreqMemEntry entry first:
  //   * array index, memory channel name, operating frequency, operating mode,
  //   * ?
  if( iArrayIndex >= 0 )
   { SL_AppendPrintf( &pszDest, pszEndstop, "i=%d,", iArrayIndex );
   }
  if( pFreqMemEntry->c16MemoryName[0] != '\0' )
   { SL_AppendPrintf( &pszDest, pszEndstop, "name=%s,",pFreqMemEntry->c16MemoryName );
   }
  if( pFreqMemEntry->RxTx[0].dblOperatingFreq_Hz != 0.0 )
   { SL_AppendPrintf( &pszDest, pszEndstop, "freq=%lf,",pFreqMemEntry->RxTx[0].dblOperatingFreq_Hz );
   }
  if( (pFreqMemEntry->RxTx[1].dblOperatingFreq_Hz != 0.0)  // DIFFERENT and VALID *transmit* frequency ("split") ?
   && (pFreqMemEntry->RxTx[1].dblOperatingFreq_Hz != pFreqMemEntry->RxTx[0].dblOperatingFreq_Hz ) )
   { SL_AppendPrintf( &pszDest, pszEndstop, "rxfr=%lf,",pFreqMemEntry->RxTx[1].dblOperatingFreq_Hz );
   }
  if( pFreqMemEntry->RxTx[0].iOpMode != RIGCTRL_OPMODE_UNKNOWN )
   { SL_AppendPrintf( &pszDest, pszEndstop, "mode=%s,",RigCtrl_OperatingModeToString(pFreqMemEntry->RxTx[0].iOpMode) );
   }
  if( pFreqMemEntry->RxTx[0].fDataMode != 0 )
   { SL_AppendPrintf( &pszDest, pszEndstop, "data=%d,", pFreqMemEntry->RxTx[0].fDataMode );
   }
  if( pFreqMemEntry->RxTx[0].iDuplexMode != 0 )
   { SL_AppendPrintf( &pszDest, pszEndstop, "dupl=%d,", pFreqMemEntry->RxTx[0].iDuplexMode );
   }
  if( pFreqMemEntry->RxTx[0].iToneType != 0 )
   { SL_AppendPrintf( &pszDest, pszEndstop, "tone=%d,", pFreqMemEntry->RxTx[0].iToneType );
   }
  if( pFreqMemEntry->RxTx[0].iDigitalSquelch != 0 )
   { SL_AppendPrintf( &pszDest, pszEndstop, "dsql=%d,", pFreqMemEntry->RxTx[0].iDigitalSquelch );
   }
  if( pFreqMemEntry->RxTx[0].iRepeaterToneFreq_Hz != 0 )
   { SL_AppendPrintf( &pszDest, pszEndstop, "rptt=%d,", pFreqMemEntry->RxTx[0].iRepeaterToneFreq_Hz );
   }
  if( pFreqMemEntry->RxTx[0].iToneSquelchFreq_Hz != 0 )
   { SL_AppendPrintf( &pszDest, pszEndstop, "tsql=%d,", pFreqMemEntry->RxTx[0].iToneSquelchFreq_Hz );
   }
  if( pFreqMemEntry->RxTx[0].iDTCS_Code != 0 ) // Icom slang: "DTCS code settings", whatever that means
   { SL_AppendPrintf( &pszDest, pszEndstop, "dtcs=%d,", pFreqMemEntry->RxTx[0].iDTCS_Code );
   }
  if( pFreqMemEntry->RxTx[0].iDV_DigitalCodeSquelch != 0 ) // Icom slang: "DTCS code settings", whatever that means
   { SL_AppendPrintf( &pszDest, pszEndstop, "dvcs=%d,", pFreqMemEntry->RxTx[0].iDV_DigitalCodeSquelch );
   }
  if( pFreqMemEntry->RxTx[0].iDuplexOffsetFreq_Hz != 0 )
   { SL_AppendPrintf( &pszDest, pszEndstop, "offs=%d,", pFreqMemEntry->RxTx[0].iDuplexOffsetFreq_Hz );
   }
  if( pFreqMemEntry->RxTx[0].c8DestCall[0] != '\0' ) // aka "UR" (destination call sign for digital voice)
   { SL_AppendPrintf( &pszDest, pszEndstop, "dvUR=%s,", pFreqMemEntry->RxTx[0].c8DestCall );
   }
  if( pFreqMemEntry->RxTx[0].c8AccessRptrCall[0] != '\0' ) // "R1 (Access repeater) call sign
   { SL_AppendPrintf( &pszDest, pszEndstop, "dvR1=%s,", pFreqMemEntry->RxTx[0].c8AccessRptrCall );
   }
  if( pFreqMemEntry->RxTx[0].c8GatewayRptrCall[0] != '\0') // "R2 (Gateway/Link repeater) call sign
   { SL_AppendPrintf( &pszDest, pszEndstop, "dvR2=%s,", pFreqMemEntry->RxTx[0].c8GatewayRptrCall );
   }

  // Remove the LAST COMMA in the long list of key=value pairs assembled above:
  if( (pszDest > pszDestStart) && (pszDest[-1]==',') )
   { pszDest[-1] = '\0';
     --pszDest;
   }

  return pszDest - pszDestStart;
} // end RigCtrl_FreqMemEntryToString()

enum
{ FREQ_MEM_TOKEN_INDEX=0, // "i=" -> array index (not a part of the struct, but an extra function argument)
  FREQ_MEM_TOKEN_NAME,    // "name=" -> pFreqMemEntry->c16MemoryName
  FREQ_MEM_TOKEN_FREQ,    // "freq=" -> pFreqMemEntry->RxTx[0].dblOperatingFreq_Hz
  FREQ_MEM_TOKEN_RXFR,    // "rxfr=" -> pFreqMemEntry->RxTx[1].dblOperatingFreq_Hz
  FREQ_MEM_TOKEN_MODE,    // "mode=" -> pFreqMemEntry->RxTx[0].iOpMode
  FREQ_MEM_TOKEN_DATA,    // "data=" -> pFreqMemEntry->RxTx[0].iDataMode
  FREQ_MEM_TOKEN_DUPL,    // "dupl=" -> pFreqMemEntry->RxTx[0].iDuplexMode
  FREQ_MEM_TOKEN_TONE,    // "tone=" -> pFreqMemEntry->RxTx[0].iToneType
  FREQ_MEM_TOKEN_DSQL,    // "dsql=" -> pFreqMemEntry->RxTx[0].iDigitalSquelch
  FREQ_MEM_TOKEN_RPTT,    // "rptt=" -> pFreqMemEntry->RxTx[0].iRepeaterToneFreq_Hz
  FREQ_MEM_TOKEN_TSQL,    // "tsql=" -> pFreqMemEntry->RxTx[0].iToneSquelchFreq_Hz
  FREQ_MEM_TOKEN_DTCS,    // "dtcs=" -> pFreqMemEntry->RxTx[0].iDTCS_Code
  FREQ_MEM_TOKEN_DVCS,    // "dvcs=" -> pFreqMemEntry->RxTx[0].iDV_DigitalCodeSquelch
  FREQ_MEM_TOKEN_OFFS,    // "offs=" -> pFreqMemEntry->RxTx[0].iDuplexOffsetFreq_Hz
  FREQ_MEM_TOKEN_dvUR,    // "dvUR=" -> pFreqMemEntry->RxTx[0].c8DestCall        aka "UR" for digital voice
  FREQ_MEM_TOKEN_dvR1,    // "dvR1=" -> pFreqMemEntry->RxTx[0].c8AccessRptrCall  aka "R1" for digital voice
  FREQ_MEM_TOKEN_dvR2     // "dvR2=" -> pFreqMemEntry->RxTx[0].c8GatewayRptrCall aka "R2" for digital voice
} FreqMemTokens;
static const T_SL_TokenList FreqMemEntryTokens[] =
{ { "i=",     FREQ_MEM_TOKEN_INDEX }, // "i=" -> array index (not a part of the struct, but an extra function argument)
  { "name=",   FREQ_MEM_TOKEN_NAME }, // "name=" -> pFreqMemEntry->c16MemoryName
  { "freq=",   FREQ_MEM_TOKEN_FREQ }, // "freq=" -> pFreqMemEntry->RxTx[0].dblOperatingFreq_Hz
  { "rxfr=",   FREQ_MEM_TOKEN_RXFR }, // "rxfr=" -> pFreqMemEntry->RxTx[1].dblOperatingFreq_Hz
  { "mode=",   FREQ_MEM_TOKEN_MODE }, // "mode=" -> pFreqMemEntry->RxTx[0].iOpMode
  { "data=",   FREQ_MEM_TOKEN_DATA }, // "data=" -> pFreqMemEntry->RxTx[0].iDataMode
  { "dupl=",   FREQ_MEM_TOKEN_DUPL }, // "dupl=" -> pFreqMemEntry->RxTx[0].iDuplexMode                                     "
  { "tone=",   FREQ_MEM_TOKEN_TONE }, // "tone=" -> pFreqMemEntry->RxTx[0].iToneType
  { "dsql=",   FREQ_MEM_TOKEN_DSQL }, // "dsql=" -> pFreqMemEntry->RxTx[0].iDigitalSquelch
  { "rptt=",   FREQ_MEM_TOKEN_RPTT }, // "rptt=" -> pFreqMemEntry->RxTx[0].iRepeaterToneFreq_Hz
  { "tsql=",   FREQ_MEM_TOKEN_TSQL }, // "tsql=" -> pFreqMemEntry->RxTx[0].iToneSquelchFreq_Hz
  { "dtcs=",   FREQ_MEM_TOKEN_DTCS }, // "dtcs=" -> pFreqMemEntry->RxTx[0].iDTCS_Code
  { "dvcs=",   FREQ_MEM_TOKEN_DVCS }, // "dvcs=" -> pFreqMemEntry->RxTx[0].iDV_DigitalCodeSquelch
  { "offs=",   FREQ_MEM_TOKEN_OFFS }, // "offs=" -> pFreqMemEntry->RxTx[0].iDuplexOffsetFreq_Hz
  { "dvUR=",   FREQ_MEM_TOKEN_dvUR }, // "dvUR=" -> pFreqMemEntry->RxTx[0].c8DestCall        aka "UR" for digital voice
  { "dvR1=",   FREQ_MEM_TOKEN_dvR1 }, // "dvR1=" -> pFreqMemEntry->RxTx[0].c8AccessRptrCall  aka "R1" for digital voice
  { "dvR2=",   FREQ_MEM_TOKEN_dvR2 }, // "dvR2=" -> pFreqMemEntry->RxTx[0].c8GatewayRptrCall aka "R2" for digital voice
  { NULL, 0 } // "all zeros" mark the end of the list
};


//---------------------------------------------------------------------------
CPROT int RigCtrl_StringToFreqMemEntry(T_RigCtrlFreqMemEntry *pFreqMemEntry,
                int *piArrayIndex, // [out,optional] ARRAY INDEX (NEGATIVE if not specified in the string)
                const char *pszSource )
  // When successful ("valid syntax"), returns the number of characters
  // parsed from pszSource. This allows THE CALLER to continue parsing
  // after the "recognized" part of the source string.
  // Note: Members in T_RigCtrlFreqMemEntry for which no key (name) has
  //       been found in the list of key=value pairs will NOT be modified.
{ const char *pszSourceStart = pszSource;
  const char *pszKeyStart;
  int iToken, iValue;

  if( piArrayIndex != NULL )
   { *piArrayIndex = -1;   // didn't find "i=.." in the key=value pairs yet
   }

  while( *pszSource != '\0' ) // continue until the end of the string..
   {                          // .. or until the first unrecognized key (name)
     pszKeyStart = pszSource;
     iToken = SL_SkipOneOfNStrings( &pszSource, FreqMemEntryTokens );
     if( iToken < 0 )
      { break;
      }
     switch( iToken )
      { case FREQ_MEM_TOKEN_INDEX : // "i=<array index>" -> *piArrayIndex (if non-NULL)
             iValue = SL_ParseInteger( &pszSource );
             if( piArrayIndex != NULL )
              { *piArrayIndex = iValue;
              }
             break;
        case FREQ_MEM_TOKEN_NAME : // "name=" -> pFreqMemEntry->c16MemoryName
             SL_CopyStringUntilDelimiter( &pszSource, pFreqMemEntry->c16MemoryName, 16/*iMaxLen*/, ",\r\n"/*pszDelimiters*/ );
             break;
        case FREQ_MEM_TOKEN_FREQ : // "freq=" -> pFreqMemEntry->RxTx[0].dblOperatingFreq_Hz
             pFreqMemEntry->RxTx[0].dblOperatingFreq_Hz = SL_ParseFloat( &pszSource, '.'/*cDecimalSeparator*/ );
             break;
        case FREQ_MEM_TOKEN_RXFR : // "rxfr=" -> pFreqMemEntry->RxTx[1].dblOperatingFreq_Hz
             pFreqMemEntry->RxTx[1].dblOperatingFreq_Hz = SL_ParseFloat( &pszSource, '.'/*cDecimalSeparator*/ );
             break;
        case FREQ_MEM_TOKEN_MODE : // "mode=" -> pFreqMemEntry->RxTx[0].iOpMode, but "as a string" to avoid magic numbers
             // inverse to RigCtrl_OperatingModeToString(pFreqMemEntry->RxTx[0].iOpMode) :
             pFreqMemEntry->RxTx[0].iOpMode = RigCtrl_StringToOperatingMode( &pszSource );
             break;
        case FREQ_MEM_TOKEN_DATA : // "data=" -> pFreqMemEntry->RxTx[0].iDataMode (just a boolean, 0=false, 1=true, at least for Icom)
             pFreqMemEntry->RxTx[0].fDataMode = SL_ParseInteger( &pszSource );
             break;
        case FREQ_MEM_TOKEN_DUPL : // "dupl=" -> pFreqMemEntry->RxTx[0].iDuplexMode
             pFreqMemEntry->RxTx[0].iDuplexMode = SL_ParseInteger( &pszSource );
             break;
        case FREQ_MEM_TOKEN_TONE : // "tone=" -> pFreqMemEntry->RxTx[0].iToneType
             pFreqMemEntry->RxTx[0].iToneType = SL_ParseInteger( &pszSource );
             break;
        case FREQ_MEM_TOKEN_DSQL : // "dsql=" -> pFreqMemEntry->RxTx[0].iDigitalSquelch
             pFreqMemEntry->RxTx[0].iDigitalSquelch = SL_ParseInteger( &pszSource );
             break;
        case FREQ_MEM_TOKEN_RPTT : // "rptt=" -> pFreqMemEntry->RxTx[0].iRepeaterToneFreq_Hz
             pFreqMemEntry->RxTx[0].iRepeaterToneFreq_Hz = SL_ParseInteger( &pszSource );
             break;
        case FREQ_MEM_TOKEN_TSQL : // "tsql=" -> pFreqMemEntry->RxTx[0].iToneSquelchFreq_Hz
             pFreqMemEntry->RxTx[0].iToneSquelchFreq_Hz = SL_ParseInteger( &pszSource );
             break;
        case FREQ_MEM_TOKEN_DTCS : // "dtcs=" -> pFreqMemEntry->RxTx[0].iDTCS_Code
             pFreqMemEntry->RxTx[0].iDTCS_Code = SL_ParseInteger( &pszSource );
             break;
        case FREQ_MEM_TOKEN_DVCS : // "dvcs=" -> pFreqMemEntry->RxTx[0].iDV_DigitalCodeSquelch
             pFreqMemEntry->RxTx[0].iDV_DigitalCodeSquelch = SL_ParseInteger( &pszSource );
             break;
        case FREQ_MEM_TOKEN_OFFS : // "offs=" -> pFreqMemEntry->RxTx[0].iDuplexOffsetFreq_Hz
             pFreqMemEntry->RxTx[0].iDuplexOffsetFreq_Hz = SL_ParseInteger( &pszSource );
             break;
        case FREQ_MEM_TOKEN_dvUR : // "dvUR=" -> pFreqMemEntry->RxTx[0].c8DestCall aka "UR" for digital voice
             SL_CopyStringUntilDelimiter( &pszSource, pFreqMemEntry->RxTx[0].c8DestCall, 8/*iMaxLen*/, ",\r\n"/*pszDelimiters*/ );
             break;
        case FREQ_MEM_TOKEN_dvR1 : // "dvR1=" -> pFreqMemEntry->RxTx[0].c8AccessRptrCall  aka "R1" for digital voice
             SL_CopyStringUntilDelimiter( &pszSource, pFreqMemEntry->RxTx[0].c8AccessRptrCall, 8/*iMaxLen*/, ",\r\n"/*pszDelimiters*/ );
             break;
        case FREQ_MEM_TOKEN_dvR2 : // "dvR2=" -> pFreqMemEntry->RxTx[0].c8GatewayRptrCall aka "R2" for digital voice
             SL_CopyStringUntilDelimiter( &pszSource, pFreqMemEntry->RxTx[0].c8GatewayRptrCall, 8/*iMaxLen*/, ",\r\n"/*pszDelimiters*/ );
             break;
        default:
             break;
      }
     SL_SkipSpaces( &pszSource ); // SPACES outside a key=value pair have no syntactic meaning.
               // They are accepted for "pretty printed" text files.
               // The string builder (RigCtrl_FreqMemEntryToString) will not emit spaces.
     if( ! SL_SkipChar( &pszSource, ',' ) ) // no comma ? End of the key=value string
      { break;
      }
    } // end while( *pszSource != '\0' )
   return pszSource-pszSourceStart;
} // end RigCtrl_StringToFreqMemEntry()

//---------------------------------------------------------------------------
// Built-in "traffic log" (circular FIFO for the display in e.g. DebugU1.cpp)
//    Initially used for CI-V via serial port, but in the meantime,
//    can optionally store the first few bytes in any kind of UDP packets,
//    packetized analog samples received from external A/D converters, etc.
//    (Why that ? Certain Icom radios send / receive audio via UDP too,
//     thus module RigControl.c must be able to exchange audio streams
//     with Spectrum Lab anyway. Thus, to develop microcontroller firmware
//     driving external A/D or D/A converters, the former "CAT Traffic
//     Monitor" shall be able to display other kinds of 'packets', too.
//
//  -> file:///C:/cbproj/SpecLab/html/troubles.htm#CAT_traffic_monitor
//
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
CPROT void RigCtrl_ClearTrafficLog( T_RigCtrlInstance *pRC )
{
  pRC->iTrafficLogHeadIndex = 0; // circular FIFO head index (post-incremented, wraps from RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES-1 to zero)
  pRC->fTrafficLogFull  = FALSE; // TRUE when all <RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES> entries are IN USE.
  // (but in most cases, the circular head index 'iTrafficLogHeadIndex' keeps incrementing,
  //  unless RigCtrl_iTrafficMonitorDisplayOptions.RIGCTRL_TRAFFIC_DISPLAY_OPTION_PAUSE_ON_FULL_FIFO is set.)


} // end RigCtrl_ClearTrafficLog()

//---------------------------------------------------------------------------
void RigCtrl_AddToTrafficLog( T_RigCtrlInstance *pRC, // for 'decoded CI-V' messages (*)
         int  iRigCtrlPort, // [in] RIGCTRL_PORT_RADIO / RIGCTRL_PORT_CLIENT_1 / 2 / .. ?
         BYTE *pbMessage, int iMsgLength, // [in] CI-V message as text, including pre- and postamble,
                            //      or just an 'info' about an important internal state change,
                            //      in complete plain text, with a few parameters.
         int  iMsgType,     // [in] e.g. RIGCTRL_MSGTYPE_RADIO_ID | RIGCTRL_MSGTYPE_NO_OK, etc;
                            //      plus some bitwise combineable flags like RIGCTRL_MSGTYPE_FLAG_UDP .
         const char *pszComment)  // [in] optional comment (*not* a format string)
  // (*) Since 2021-09, RigCtrl_AddToTrafficLog() is also used to show
  //     a few Icom-specific UDP-"control" messages (with UDP payload only).
  //     The COLOUR used for the display (in a Rich Edit control) is selected
  //     later, in RigCtrl_ReadTrafficLog(), depending on "Message Type" flags like
  //       RIGCTRL_MSGTYPE_ERROR, RIGCTRL_MSGTYPE_FLAG_TX, RIGCTRL_MSGTYPE_FLAG_RX, ...
  //     Examples (copied and pasted from the RichEdit control into HTML *as text*)
  //     are in the Spectrum Lab manual [SpecLab_CAT_Monitor] .
  //
  // Non-CI-V-entries (but UDP) are flagged by RIGCTRL_MSGTYPE_CONTROL_VIA_UDP in iMsgType,
  // and are formatted slightly different than CI-V in RigCtrl_ReadTrafficLog().
{
  T_RigCtrlTrafficLogEntry *pNewEntry;
  double dblUnixTime = UTL_GetCurrentUnixDateAndTime();
  int iHeadIndex = pRC->iTrafficLogHeadIndex % RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES;
  int i,n;

  // Since 2023-07, not only the DISPLAY may be paused, but also the acquisition
  //  of TRAFFIC LOG DATA. Controllable via 'More'-button -> menu in DebugU1.cpp:
  if( RigCtrl_fTrafficMonitorPaused )  // one of the 'configurable triggers' has PAUSED the traffic log ->
   { return;  // avoid those "important captured data" from being overwritten,
     // while the user scrolls back and forth, and SWITCHES THE DISPLAY MODE,
     // which is only possible if the 'important' entries are still in pRC->TrafficLog[] !
   } // end if( RigCtrl_fTrafficMonitorPaused )

  pNewEntry = &pRC->TrafficLog[iHeadIndex];
  memset( pNewEntry, 0, sizeof(T_RigCtrlTrafficLogEntry) );  // provides zero-terminated strings
  if( iMsgLength > 0 )  // Only increment the 'message number' when logging REAL PACKETS,
   { ++RigCtrl_i32MessageNrForTrafficLog; // .. because only 'real packets' are NUMBERED.
     // Info strings like "Opening 'Serial' port: Hostname 'IC-705' resolved into IP 192.168.178.28"
     // are no PACKETS TRAVELLING ON THE NETWORK, and will not have a number in the first column
     // when converted into text in RigCtrl_ReadTrafficLog() .
     // Despite that, if a log entry WITH packet data (i.e. T_RigCtrlTrafficLogEntry.iMsgLength > 0)
     //  is followed by a couple of 'info lines WITHOUT packet data',
     //  those subsequent lines will have the same T_RigCtrlTrafficLogEntry.i32MessageNr .
   }
  pNewEntry->i32MessageNr = RigCtrl_i32MessageNrForTrafficLog; // unique "message identifier",
    // used in the GUI (DebugU1.cpp) when the operator clicked on a certain line
    // in the RichText control. The "Message Number" (or "No.", as in Wireshark)
    // can be used to retrieve more info than currently visible on the screen,
    // as long as the circular FIFO (

  pNewEntry->dblUnixTime = dblUnixTime;
  n = iMsgLength; // number of "raw message bytes" that can be stored in a log-entry
  if( n>80 )
   {  n=80;
   }
  if( pbMessage != NULL )  // may be NULL for special messages like "Spectrum timeout", etc
   { for( i=0; i<n; ++i )
      { pNewEntry->b80Message[i] = pbMessage[i];  // copy up to 80 bytes of the message data
      }
   }
  pNewEntry->iMsgLength   = iMsgLength;
  pNewEntry->iMsgType     = iMsgType;  // RIGCTRL_MSGTYPE_FLAG_UDP, etc ..
  pNewEntry->iRigCtrlPort = iRigCtrlPort;
  if( pszComment != NULL )
   { SL_strncpy( pNewEntry->szComment, pszComment, RIGCTRL_MAX_COMMENT_LENGTH );
   }

  ++iHeadIndex;
  if( RigCtrl_iTrafficMonitorDisplayOptions & RIGCTRL_TRAFFIC_DISPLAY_OPTION_PAUSE_ON_FULL_FIFO )
   { if( iHeadIndex == (RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES-1) )
      { RigCtrl_fTrafficMonitorPaused = TRUE;
        pRC->fTrafficLogFull = TRUE; // all <RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES> are valid now !
        // In the VERY LAST FIFO entry, leave a note that we stopped because the FIFO is "full":
        pNewEntry = &pRC->TrafficLog[iHeadIndex];
        memset( pNewEntry, 0, sizeof(T_RigCtrlTrafficLogEntry) );  // provides zero-terminated strings
        pNewEntry->i32MessageNr = RigCtrl_i32MessageNrForTrafficLog; // unique "message identifier",
        pNewEntry->dblUnixTime = dblUnixTime;
        SL_strncpy( pNewEntry->szComment, "traffic monitor paused on 'full FIFO' (as configured)", RIGCTRL_MAX_COMMENT_LENGTH );
        pRC->iTrafficLogHeadIndex = 0;
        RigCtrl_fUpdateTrafficLog = TRUE; // now "ready for being UPDATED on-screen"
        return;
      }
   } // RIGCTRL_TRAFFIC_DISPLAY_OPTION_PAUSE_ON_FULL_FIFO ?
  if( iHeadIndex >= RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES )
   {  iHeadIndex = 0;
      pRC->fTrafficLogFull = TRUE; // all <RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES> are valid now !
      // Because there may be an unknown number of READERS,
      // T_RigCtrlInstance contains no FIFO *TAIL* INDEX itself !
      // An example for a "traffic log display" (reader) is in
      // C:\cbproj\SpecLab\DebugU1.cpp .
   }
  pRC->iTrafficLogHeadIndex = iHeadIndex;
  RigCtrl_fUpdateTrafficLog = TRUE; // now "ready for being UPDATED on-screen"


} // end RigCtrl_AddToTrafficLog()


//---------------------------------------------------------------------------
int  RigCtrl_GetNumEntriesInTrafficLog( T_RigCtrlInstance *pRC,
             int iTailIndex )     // [in] FIFO tail index
{
  int iHeadIndex = pRC->iTrafficLogHeadIndex % RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES;
  int iOldestEntry, nEntries;
  if( pRC->fTrafficLogFull )
   {  // ALL entries, pRC->TrafficLog[0..RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES-1], are valid.
      iOldestEntry = (iHeadIndex+1) % RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES;
   }
  else // only entries in pRC->TrafficLog[0..iHeadIndex-1] are valid
   {  // Reading would BEGIN at index zero .
      // No more entries are available at index iHeadIndex .
      iOldestEntry = 0;
   }
  if( iTailIndex < 0 ) // caller would read the FIRST entry (when available)
   {  iTailIndex = iOldestEntry;
   }
  else
   {  iTailIndex %= RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES;
   }
  nEntries = iHeadIndex - iTailIndex;
  if( nEntries < 0 ) // circular wrap..
   {  nEntries += RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES;
   }
  return nEntries;
} // end RigCtrl_GetNumEntriesInTrafficLog(


//---------------------------------------------------------------------------
BOOL RigCtrl_ReadTrafficLog( // .. as 'human readable text' with colour highlighting
             T_RigCtrlInstance *pRC,   // [in] rig control instance (just in case we support more than one)
             int  *piTrafficTailIndex, // [in,out] FIFO tail index (incremented in each call)
             char *pszDest,            // [out] human readable text, capacity at least <iWantedCharsPerLine+4> bytes
             int  iWantedCharsPerLine, // [in] typically 80 characters per line, but maybe more (see TDebugForm::UpdateCATTrafficLog() )
             int  *piMsgType,          // [out,optional] RIGCTRL_MSGTYPE_FREQUENCY_REPORT (protocol independent)
             DWORD *pdwColor )         // [out,optional] colour, from MSBIT to LSBIT: BLUE, GREEN, RED (8 bit each)
  // Returns TRUE as long as more entries are available (and the TAIL INDEX is incremented),
  //     or  FALSE when no more entries are available in the CAT message FIFO.
  //
  // Since 2023-07, RigCtrl_ReadTrafficLog(!) may decide HOW TO DISPLAY packets,
  //    depending on RIGCTRL_TRAFFIC_DISPLAY_OPTION_SHOW_UDP_PAYLOAD  .
  //    This was easy to achieve because if RIGCTRL_MSGTYPE_FLAG_UDP is set in iMsgType,
  //    pbMessage points to THE FIRST BYTE of the *UDP* payload;
  //    not just the CI-V payload (see RigCtrl_SendCIVDataViaUDP() -> RigCtrl_AddToTrafficLog() ).
  //
{
  int iHeadIndex = pRC->iTrafficLogHeadIndex % RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES;
  int iTailIndex = *piTrafficTailIndex;
  int iOldestEntry, i, iType2, n, iSlen, nCharsForMsgTypeAndComment;
  T_RigCtrlTrafficLogEntry *pEntry;
  char *cp = pszDest;
  char *pszEndstop = pszDest + iWantedCharsPerLine;
  DWORD dwColor;
  BOOL  fHavePacketData; // TRUE : got a log entry with "packet data";  FALSE : "just an INFO or missing data"
  BYTE  *pbSource;

  *cp = '\0';

  if( pRC->fTrafficLogFull )
   {   // ALL entries, pRC->TrafficLog[0..RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES-1], are valid.
       // Reading must BEGIN at iHeadIndex+1 because that's THE OLDEST entry
       //   (==the next to be overwritten in RigCtrl_AddToTrafficLog() .
       // When the tail index reaches iHeadIndex, no more entries are available.
       iOldestEntry = (iHeadIndex+1) % RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES;
   }
  else // only entries in pRC->TrafficLog[0..iHeadIndex-1] are valid
   {   // Reading must BEGIN at index zero .
       // No more entries are available at index iHeadIndex .
       iOldestEntry = 0;
   }
  if( iTailIndex < 0 ) // caller wants to read the FIRST, aka OLDEST entry (when available)
   {  iTailIndex = iOldestEntry;
   }
  else
   {  iTailIndex %= RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES;
   }
  if( iHeadIndex == iTailIndex )
   { return FALSE;
   }
  else
   { pEntry = &pRC->TrafficLog[iTailIndex];
     *piTrafficTailIndex = (iTailIndex+1) % RIGCTRL_MAX_TRAFFIC_LOG_ENTRIES;
     iType2 = pEntry->iMsgType & RIGCTRL_MSGTYPE_STRIP_FLAGS;

     // Which background colour (that's the easy part...) ?
     if( pEntry->iMsgType & RIGCTRL_MSGTYPE_ERROR ) // e.g. a response-reception-TIMEOUT
      { dwColor = 0x7F00FF; // PINK(ish) .. from MSBIT to LSBIT: BLUE, GREEN, RED (8 bit each)
      }
     else if( pEntry->iMsgType & RIGCTRL_MSGTYPE_FLAG_TX ) // this message was SENT (in CI-V speak, from "controller" to "radio")
      { dwColor = 0x00FFFF; // YELLOW .. from MSBIT to LSBIT: BLUE, GREEN, RED (8 bit each)
      }
     else if( pEntry->iMsgType & RIGCTRL_MSGTYPE_FLAG_RX ) // this message contains 'real DATA' (UDP or CI-V), and it was RECEIVED
      { // RECEIVED message : light cyan background (unless it's a "NoGo"-response, or some other "special case" :
        if( pEntry->iMsgType & RIGCTRL_MSGTYPE_FLAG_UDP )
         { // Added 2021-09 : not only the normal CI-V payload but UDP-payload :
           // If the message type was recognized, the "meaning" of this message
           // is in pEntry->szComment originating from RigCtrl_ProcessIcomUDP().
           // Show "CI-V less" messages in YELLOW .
           // Otherwise (UDP Control packet from an Icom radio that we DON'T understand),
           // show the message in ORANGE colour. At least, that was the plan...
           if( pEntry->szComment[0] != '\0' ) // ok, RigCtrl_ProcessIcomUDP() had recognized the received type ->
            { dwColor = 0xC0FF7F; // B,G,R for "aqua"(?) = RECEIVED UDP (not yet reduced to "pure CI-V" for the 'Traffic Display")
            }
           else // not recognized ?
            { dwColor = 0x007FFF; // orange = received UDP, but content not recognized
            }
         }   // RIGCTRL_MSGTYPE_CONTROL_VIA_UDP ?
        else // not an entry with UDP "payload" but most likely a CI-V message:
         {
           dwColor = 0xFFFF00; // 100 % BLUE  + 100 % GREEN + 0 % RED (8 bit each) : lightcyan = RECEIVED CiV (not UDP)
           if( pEntry->iMsgType & RIGCTRL_MSGTYPE_FLAG_EXPECTED )
            { dwColor = 0x7FFF7F; // more green-ish than Cyan : "received the EXPECTED response"
            }
         }
      }
     else // none of the above, no "real DATA packet" (neither RX- nor TX-data), so maybe...
      { // .. a 'state transition', like the ones emitted from RigCtrl_EnterState_Done(),
        // with only the bit 'RIGCTRL_MSGTYPE_OK' set (or not even that) :
        dwColor = 0xC0C0C0; // B,G,R for "lightgray"
      }

     switch( iType2 )
      { case RIGCTRL_MSGTYPE_FREQUENCY_REPORT: // "frequency report" (from radio to user)
        case RIGCTRL_MSGTYPE_SET_FREQUENCY: // command to SET a new VFO frequency (from user to radio)
        case RIGCTRL_MSGTYPE_OP_MODE  : // "mode data" (often unsolicited, "on change")
        case RIGCTRL_MSGTYPE_RIT_XIT  : // RIT or XIT data (often unsolicited, "on change")
        case RIGCTRL_MSGTYPE_RF_GAIN  : // anything that affects the RF gain (preamp, attenuator, "AIP", ..)
        case RIGCTRL_MSGTYPE_C_SPECTRUM: // configure transmission of spectrum data
        case RIGCTRL_MSGTYPE_SPECTRUM : // "waveform data" (in fact, for Icom's spectrum scope)
        case RIGCTRL_MSGTYPE_POWER_ON_OFF: // power on/off (special case, needs long preamble)
        case RIGCTRL_MSGTYPE_OTHER    : // acknowledge for 'set', etc pp.
        case RIGCTRL_MSGTYPE_OK       : // in CI-V : unspecific "OK", cmd 0xFB
             break;
        case RIGCTRL_MSGTYPE_NOT_OK   : // in CI-V : unspecific "NG" / "Not Good", cmd 0xFA
             dwColor = 0xFF00FF; // colour format, from MSBIT to LSBIT: BLUE, GREEN, RED (8 bit each)
             break;
        case RIGCTRL_MSGTYPE_INFO     : // just an INFO, e.g. about an important state transition, NO PACKET DATA.
             dwColor = 0xC0C0C0; // colour format, from MSBIT to LSBIT: BLUE, GREEN, RED (8 bit each)
             break;
        default:
             break;
      }

     // Begin each line in the "traffic log" with the 'Number'
     //   ("No." as in Wireshark, with leading spaces, and RIGHT-ALIGNED.
     //    Four digits should be sufficient, because 'the trouble usually starts early'):
     if( (pEntry->iMsgLength > 0) && ( iType2 != RIGCTRL_MSGTYPE_INFO) ) // traffic log entry with REAL PACKET DATA ?
      { sprintf( cp, "%4ld", (long)( pEntry->i32MessageNr % 10000 ) ); // Show their "No." in the first column.
        fHavePacketData = TRUE;
      }
     else // traffic log entry WITHOUT real packet data -> leave several columns away further below !
      { fHavePacketData = FALSE;
      }
     SL_PadSpaces( &cp, pszEndstop, 5/*iFixedFieldLength*/ );

     // Next (optional) column: "time of day" in in UTC .. also makes sense for RIGCTRL_MSGTYPE_INFO (without PACKET DATA)
     if( RigCtrl_iTrafficMonitorExtraColumns & RIGCTRL_TRAFFIC_EXTRA_COLUMN_TIME )
      { UTL_FormatDateAndTime( "hh:mm:ss.s", pEntry->dblUnixTime, cp );
        SL_SkipToEndOfString( (const char**)&cp );
        // ,------------------|____________|
        // '--> Why do we need this bloody cast, Mr. Pedantic Compiler ?
        //      "cp" points to a char, so "&cp" is a pointer to a pointer to a char.
        //      SL_SkipToEndOfString() expects a "const char **", which means
        //      it promises not to modify the CHAR(s). So what's so "suspicious"
        //      about the original code ( SL_SkipToEndOfString( &cp ); ) ?

        SL_AppendChar( &cp, pszEndstop, ' ' );  // TEN characters plus one column-separating SPACE
      }

     if( fHavePacketData ) // most columns that follow are only displayed if there are 'packet data' ...
      {

        // Next (optional) column: show "TX" / "RX" (transmit/receive flag; useful if this cannot be indicated by COLOURS)
        if( RigCtrl_iTrafficMonitorExtraColumns & RIGCTRL_TRAFFIC_EXTRA_COLUMN_TX_RX )
         { switch( pEntry->iRigCtrlPort ) // message on the RADIO PORT or one of the "duplicator" ports ?
            { case RIGCTRL_PORT_RADIO :   // message on the port that connects TO THE RADIO...
                 if( pEntry->iMsgType & RIGCTRL_MSGTYPE_FLAG_TX )
                  { strcpy( cp, "TX " );
                  }
                 else
                  { strcpy( cp, "RX " );
                  }
                 break;
              default :
                 if( pEntry->iMsgType & RIGCTRL_MSGTYPE_FLAG_TX )
                  { sprintf( cp, "t%d ", (int)pEntry->iRigCtrlPort );  // "t1".."t4" : transmitted to the 1st .. 4th "CLIENT PORT"
                  }
                 else
                  { sprintf( cp, "r%d ", (int)pEntry->iRigCtrlPort );  // "r1".."r4" : received from the 1st .. 4th "CLIENT PORT"
                  }
                 break;
            } // end switch( iRigCtrlPort )
           SL_PadSpaces( &cp, pszEndstop, 3/*iFixedFieldLength*/ ); // TWO characters plus one column-separating SPACE
         } // end if < .. RIGCTRL_TRAFFIC_EXTRA_COLUMN_TX_RX > ?

        // Next (optional) column: show PAYLOAD LENGTH (three hex digits without prefix) ?
        if( RigCtrl_iTrafficMonitorExtraColumns & RIGCTRL_TRAFFIC_EXTRA_COLUMN_PAYLOAD_LENGTH )
         { sprintf( cp, "%03X", (int)pEntry->iMsgLength ); // length (3 digits hex, should be ok for up to 4095 bytes from LAN)
           SL_PadSpaces( &cp, pszEndstop, 4/*iFixedFieldLength*/ ); // THREE hex digits plus one column-separating SPACE
         } // end if < .. RIGCTRL_TRAFFIC_EXTRA_COLUMN_PAYLOAD_LENGTH > ?


        // Convert the (mostly binary) entry into human-readable text,
        //   unless the "traffic log entry" contains no data but only a TEXT MESSAGE:
        switch( pEntry->iRigCtrlPort ) // message on the RADIO PORT or one of the "duplicator" ports ?
         { case RIGCTRL_PORT_RADIO :   // message on the port that connects TO THE RADIO...
              // leave dwColor as-is (not "weaker than usual".. see below)
              break;
           default :
              dwColor &= 0xBFBFBF; // make the three colour components a bit (3) darker to indicate a "duplicator client port"
              break;
         } // end switch( iRigCtrlPort )
        n = pEntry->iMsgLength; // number of "raw message bytes" that can be stored in a log-entry
        if( n>80 )
         {  n=80;
         }
        nCharsForMsgTypeAndComment = 0;
        if(  ( pEntry->szComment[0] != '\0' )
          && ( ( RigCtrl_iTrafficMonitorExtraColumns & RIGCTRL_TRAFFIC_EXTRA_COLUMN_COMMENTS ) // show "machine-generated comments, parameter names, etc" ? ...
             ||( ! fHavePacketData ) ) // without "packet data" to show, all we can show is the "comment" or "info line" !
          )
         { nCharsForMsgTypeAndComment += strlen(pEntry->szComment) + 2/*strlen("; ")*/;
         }
        if( nCharsForMsgTypeAndComment == 0 )
         { nCharsForMsgTypeAndComment = 32;  // use the 'old, fixed default'
         }

        for( i=0; (i<n) && (cp<(pszDest+iWantedCharsPerLine-nCharsForMsgTypeAndComment-5)); ++i )
         { sprintf( cp, "%02X ", (int)pEntry->b80Message[i] ); // show data (CI-V or UDP) as 2-digit HEX
           cp += strlen(cp);
         }
        if( i<n ) // couldn't dump everything (e.g. with waveform data) ?
         { *cp++ = '.';
           *cp++ = '.'; // marker for omitted hex bytes
           *cp++ = ' ';
         }
        while( cp<(pszDest+iWantedCharsPerLine-nCharsForMsgTypeAndComment) ) // pad spaces for short messages, incl. VFO frequency
         { *cp++ = ' '; // (for rare long messages, columns may be misaligned but that's ok)
         }
        *cp = '\0';
        iSlen = strlen(pEntry->szComment);
        if( (iSlen>0) && ( (cp+iSlen)<(pszDest+iWantedCharsPerLine) ) )
         {
           // if there's enough space, pad up to the next column
           while( (cp<(pszDest+iWantedCharsPerLine-iSlen-2) )
               && (cp<(pszDest+iWantedCharsPerLine-16) )  )
            { *cp++ = ' ';
            }
           sprintf( cp, "; %s", pEntry->szComment );
           cp += strlen(cp);
         } // end if < append the COMMENT > ?
      } // end if( fHavePacketData )
     else // "coloured text" only, no message data:
      { sprintf( cp, " %s", pEntry->szComment );
        cp += strlen(cp);
      } // end else < !fHavePacketData >
     while( cp<(pszDest+iWantedCharsPerLine) )  // for the coloured background, pad even more spaces..
      { *cp++ = ' '; // formerly 80 chars per line, now depends on width of the window
      }
     *cp = '\0';

     if( piMsgType != NULL )
      { *piMsgType = pEntry->iMsgType; // [out,optional] RIGCTRL_MSGTYPE_xyz (protocol independent)
      }
     if( pdwColor != NULL )
      { *pdwColor = dwColor;
      }
     return TRUE;
   }
} // end RigCtrl_ReadTrafficLog()


//---------------------------------------------------------------------------
void RigCtrl_GetHeadlineForTrafficLog(int iMsgType, char *pszDest, int iWantedCharsPerLine)
  // Generates a 'headline' to simplify interpreting the (hex) message payload
  // in "Icom UDP" packets, any maybe other message types.
  // Occasionally called from TDebugForm::UpdateCATTrafficLog() (!) .
  // The FORMAT must match the strings assembled by RigCtrl_ReadTrafficLog().
  //
  //  [in] iMsgType : bit combination including flags like  ..
  //       RIGCTRL_MSGTYPE_FLAG_UDP : UDP (datagram, usually Icom RS-BA compatible)
  //  [out] pszDest : capactiy should be iWantedCharsPerLine + 4 (!) chars or more.
{
  int byte_index;
  char *cp         = pszDest;
  char *pszEndstop = pszDest + iWantedCharsPerLine; // <- points to the STRING TERMINATOR
  int  nCharsRemaining;


  // First (non-optional) column: "No. " (with FOUR digits plus a column-separating SPACE) ..
  SL_AppendString( &cp, pszEndstop, "; Nr " );

  // Next (optional) column: "time of day" in in UTC ..
  if( RigCtrl_iTrafficMonitorExtraColumns & RIGCTRL_TRAFFIC_EXTRA_COLUMN_TIME )
   { SL_AppendString( &cp, pszEndstop, "hh:mm:ss.s " );
   }

  // Next (optional) column:  header for the "TX" / "RX" indicator
  if( RigCtrl_iTrafficMonitorExtraColumns & RIGCTRL_TRAFFIC_EXTRA_COLUMN_TX_RX )
   { SL_AppendString( &cp, pszEndstop, "tr " );
   }

  // Next (optional) column: packet length (3 digits hex)
  if( RigCtrl_iTrafficMonitorExtraColumns & RIGCTRL_TRAFFIC_EXTRA_COLUMN_PAYLOAD_LENGTH )
   { SL_AppendString( &cp, pszEndstop, "len " );
   } // end if < .. RIGCTRL_TRAFFIC_EXTRA_COLUMN_PAYLOAD_LENGTH > ?

  // Next (NON-OPTIONAL) column : "legend for the hex dump"
  if( iMsgType & RIGCTRL_MSGTYPE_FLAG_UDP ) // line in the traffic log with *UDP payload* :
   {
     // Example (line with UDP hex payload emitted by RigCtrl_ReadTrafficLog() ) :
     //                                 10 00 00 00 03 00 00 00 D7 E9 00 00 00 00 00 00
     SL_AppendString( &cp, pszEndstop, "|-IcomLen-| |typ| |seq| |-sent_id-| |-rcvd_id-|" );
     nCharsRemaining = pszEndstop - cp;
     if( nCharsRemaining > 3 )  // beginning at this character-index,
      { // just add HEXADECIMAL BYTE INDICES, as used in the RS-BA-compatible
        // structure definitions in RigControl.h :
        cp = pszDest + strlen(pszDest);
        byte_index = 0x10;
        while( (cp+2)<pszEndstop )
         { sprintf( cp, " %02X", (int)(byte_index++) );
           SL_SkipToEndOfString( (const char**)&cp );
           // ,------------------|____________|
           // '--> stupid cast .. ?
         }
      }
   }
  else if(iMsgType & RIGCTRL_MSGTYPE_FLAG_CIV ) // only CI-V, without overhead from any "transport layer"
   { //               byte-index----------> 00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F 10 11 12 13 14 15 16 17 ...
     //               3-digit LENGTH--->010 FE FE A4 E0 27 .. .. .. .. .. .. .. FD
     SL_AppendString( &cp, pszEndstop, "CiV |pre| to fm parameters..  postamble=FD   " );
   }
  else
   { SL_AppendString( &cp, pszEndstop, "len generic hex data, type unknown" );
   }
  // Pad with spaces for the wanted length (just in case the headline has a coloured background) :
  cp = pszDest + SL_strnlen( pszDest, iWantedCharsPerLine );
  while(cp<pszEndstop)
   { *cp++ = ' ';
   }
  *pszEndstop = '\0';  // <- one of the reasons why the capacity of pszDest must exceed iWantedCharsPerLine (by at least ONE BYTE)

} // end RigCtrl_GetHeadlineForTrafficLog()

//---------------------------------------------------------------------------
void RigCtrl_LimitDouble( double *pdblValue, double dblMin, double dblMax )
{ if( *pdblValue < dblMin )
   {  *pdblValue = dblMin;
   }
  if( *pdblValue > dblMax )
   {  *pdblValue = dblMax;
   }
} // end RigCtrl_LimitDouble()


//---------------------------------------------------------------------------
// Lookup tables and similar ...
//---------------------------------------------------------------------------

const T_RigCtrl_RadioInfo RigCtrl_RadioInfo_CIV[] =
{ // { iDefaultAddress,            pszName,        iCapabilities (if known),
  //      dwBands (bitwise combined, 32 bit parameter, see RIGCTRL_BAND_xyz)
  { RIGCTRL_DEF_ADDR_IC_271      , "IC-271"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_2M },
  { RIGCTRL_DEF_ADDR_IC_275      , "IC-275"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_2M },
  { RIGCTRL_DEF_ADDR_IC_471      , "IC-471"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_70CM },
  { RIGCTRL_DEF_ADDR_IC_475      , "IC-475"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_70CM },
  { RIGCTRL_DEF_ADDR_IC_575      , "IC-575"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_10M | RIGCTRL_BAND_6M },
  { RIGCTRL_DEF_ADDR_IC_703      , "IC-703"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_6M },
  { RIGCTRL_DEF_ADDR_IC_706      , "IC-706"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_6M | RIGCTRL_BAND_2M },
  { RIGCTRL_DEF_ADDR_IC_706MkII  , "IC-706MkII"  , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_6M | RIGCTRL_BAND_2M },
  { RIGCTRL_DEF_ADDR_IC_706MkIIG , "IC-706MkIIG" , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_6M | RIGCTRL_BAND_2M | RIGCTRL_BAND_70CM },
  { RIGCTRL_DEF_ADDR_IC_707      , "IC-707"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE },
  { RIGCTRL_DEF_ADDR_IC_718      , "IC-718"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE },
  { RIGCTRL_DEF_ADDR_IC_725      , "IC-725"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE },
  { RIGCTRL_DEF_ADDR_IC_726      , "IC-726"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_6M },
  { RIGCTRL_DEF_ADDR_IC_728      , "IC-728"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE },
  { RIGCTRL_DEF_ADDR_IC_729      , "IC-729"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_6M },
  { RIGCTRL_DEF_ADDR_IC_735      , "IC-735"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE },
  { RIGCTRL_DEF_ADDR_IC_736      , "IC-736"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_6M },
  { RIGCTRL_DEF_ADDR_IC_737      , "IC-737"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE },
  { RIGCTRL_DEF_ADDR_IC_738      , "IC-738"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE },
  { RIGCTRL_DEF_ADDR_IC_746      , "IC-746"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE },
  { RIGCTRL_DEF_ADDR_IC_751A     , "IC-751A"     , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE },
  { RIGCTRL_DEF_ADDR_IC_756      , "IC-756"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_6M },
  { RIGCTRL_DEF_ADDR_IC_756Pro   , "IC-756Pro"   , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_6M },
  { RIGCTRL_DEF_ADDR_IC_756ProII , "IC-756ProII" , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_6M },
  { RIGCTRL_DEF_ADDR_IC_756ProIII, "IC-756ProIII", 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_6M },
  { RIGCTRL_DEF_ADDR_IC_761      , "IC-761"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_6M | RIGCTRL_BAND_2M | RIGCTRL_BAND_70CM },
  { RIGCTRL_DEF_ADDR_IC_765      , "IC-765"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE },
  { RIGCTRL_DEF_ADDR_IC_775      , "IC-775"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE },
  { RIGCTRL_DEF_ADDR_IC_78       , "IC-78"       , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE },
  { RIGCTRL_DEF_ADDR_IC_7000     , "IC-7000"     , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_6M | RIGCTRL_BAND_2M | RIGCTRL_BAND_70CM },
  { RIGCTRL_DEF_ADDR_IC_7100     , "IC-7100"     , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_6M | RIGCTRL_BAND_2M | RIGCTRL_BAND_70CM },
  { RIGCTRL_DEF_ADDR_IC_7200     , "IC-7200"     , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_6M },
  { RIGCTRL_DEF_ADDR_IC_7300     , "IC-7300"  , RIGCTRL_CAPS_SPECTRUM_VIA_CAT
          | RIGCTRL_CAPS_POWER_ON_OFF_VIA_CAT | RIGCTRL_CAPS_HAS_NO_IQ_OUTPUT,
          RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_6M /* | RIGCTRL_BAND_4M */ },
          // ,----------------------------------------------|_____________|
          // '--> 70 MHz and 5 MHz transmit only after hardware modification.
          //      The "all band TX mod" (removal of diode D422) along with the
          //          "all band RX mod" (removal of diode D416) caused the
          //          radio to report only A SINGLE TRANSMIT BAND (0.1 to 74 MHz).
          //      That will cause RigCtrl_GetAvailableBands() to report
          //      the additional bands (RIGCTRL_BAND_4M, RIGCTRL_BAND_6M,
          //                            and even RIGCTRL_BAND_11M).
          //      Similar may apply to other 'wide-band modifyable' radios.
  { RIGCTRL_DEF_ADDR_IC_7400     , "IC-7400"     , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_6M | RIGCTRL_BAND_2M },
  { RIGCTRL_DEF_ADDR_IC_7600     , "IC-7600"     , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_6M },
  { RIGCTRL_DEF_ADDR_IC_7610     , "IC-7610"  , RIGCTRL_CAPS_SPECTRUM_VIA_CAT
          | RIGCTRL_CAPS_POWER_ON_OFF_VIA_CAT | RIGCTRL_CAPS_HAS_NO_IQ_OUTPUT,
          RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_6M }, // not sure about 4m
  { RIGCTRL_DEF_ADDR_IC_7700     , "IC-7700"     , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_6M },
  { RIGCTRL_DEF_ADDR_IC_7800     , "IC-7800"     , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_6M },
  { RIGCTRL_DEF_ADDR_IC_781      , "IC-781"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_ALL_SHORTWAVE },
  { RIGCTRL_DEF_ADDR_IC_820      , "IC-820"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_2M | RIGCTRL_BAND_70CM },
  { RIGCTRL_DEF_ADDR_IC_821      , "IC-821"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_2M | RIGCTRL_BAND_70CM },
  { RIGCTRL_DEF_ADDR_IC_910      , "IC-910"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_2M | RIGCTRL_BAND_70CM | RIGCTRL_BAND_23CM },
  { RIGCTRL_DEF_ADDR_IC_970      , "IC-970"      , 0 /*unknown capabilities*/,
          RIGCTRL_BAND_2M | RIGCTRL_BAND_70CM | RIGCTRL_BAND_23CM }, // 23cm was optional
  { RIGCTRL_DEF_ADDR_IC_9700     , "IC-9700"  , RIGCTRL_CAPS_SPECTRUM_VIA_CAT
          | RIGCTRL_CAPS_POWER_ON_OFF_VIA_CAT | RIGCTRL_CAPS_HAS_NO_IQ_OUTPUT,
          RIGCTRL_BAND_2M | RIGCTRL_BAND_70CM | RIGCTRL_BAND_23CM }, // 23cm always supported
  { RIGCTRL_DEF_ADDR_IC_705      , "IC-705"   , RIGCTRL_CAPS_SPECTRUM_VIA_CAT
          | RIGCTRL_CAPS_POWER_ON_OFF_VIA_CAT | RIGCTRL_CAPS_HAS_NO_IQ_OUTPUT,
         RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_2M | RIGCTRL_BAND_70CM }, // a gem !
  { RIGCTRL_DEF_ADDR_IC_905      , "IC-905"   , RIGCTRL_CAPS_SPECTRUM_VIA_CAT,
         RIGCTRL_BAND_2M | RIGCTRL_BAND_70CM | RIGCTRL_BAND_23CM
       | RIGCTRL_BAND_13CM   | RIGCTRL_BAND_6CM  | RIGCTRL_BAND_3CM }, // 3cm optional
  { RIGCTRL_DEF_ADDR_IC_1271     , "IC-1271"     , 0 /*unknown capabilities*/,
         RIGCTRL_BAND_23CM },
  { RIGCTRL_DEF_ADDR_IC_1275     , "IC-1275"     , 0 /*unknown capabilities*/,
         RIGCTRL_BAND_23CM },
  { RIGCTRL_DEF_ADDR_IC_R10      , "IC-R10"      , 0 /*unknown capabilities*/,
         RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_2M | RIGCTRL_BAND_70CM }, // wideband 'scanner'
  { RIGCTRL_DEF_ADDR_IC_R20      , "IC-R20"      , 0 /*unknown capabilities*/,
         RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_2M | RIGCTRL_BAND_70CM }, // not sure..
  { RIGCTRL_DEF_ADDR_IC_R71      , "IC-R71"      , 0 /*unknown capabilities*/,
         RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_2M | RIGCTRL_BAND_70CM }, // not sure..
  { RIGCTRL_DEF_ADDR_IC_R72      , "IC-R72"      , 0 /*unknown capabilities*/,
         RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_2M | RIGCTRL_BAND_70CM }, // not sure..
  { RIGCTRL_DEF_ADDR_IC_R75      , "IC-R75"      , 0 /*unknown capabilities*/,
         RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_2M | RIGCTRL_BAND_70CM }, // not sure..
  { RIGCTRL_DEF_ADDR_IC_R7000    , "IC-R7000"    , 0 /*unknown capabilities*/,
         RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_2M | RIGCTRL_BAND_70CM }, // not sure..
  { RIGCTRL_DEF_ADDR_IC_R7100    , "IC-R7100"    , 0 /*unknown capabilities*/,
         RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_2M | RIGCTRL_BAND_70CM }, // not sure..
  { RIGCTRL_DEF_ADDR_IC_R8500    , "IC-R8500"    , 0 /*unknown capabilities*/,
         RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_2M | RIGCTRL_BAND_70CM }, // not sure..
  { RIGCTRL_DEF_ADDR_IC_R8600    , "IC-R8600" , RIGCTRL_CAPS_SPECTRUM_VIA_CAT,
         RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_2M | RIGCTRL_BAND_70CM }, // not sure..
  { RIGCTRL_DEF_ADDR_IC_R9000    , "IC-R9000"    , 0 /*unknown capabilities*/,
         RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_2M | RIGCTRL_BAND_70CM }, // not sure..
  { RIGCTRL_DEF_ADDR_IC_R9500    , "IC-R9500"    , 0 /*unknown capabilities*/,
         RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_2M | RIGCTRL_BAND_70CM }, // not sure..
  { RIGCTRL_DEF_ADDR_IC_RX7      , "IC-RX7"      , 0 /*unknown capabilities*/,
         RIGCTRL_BAND_ALL_SHORTWAVE | RIGCTRL_BAND_2M | RIGCTRL_BAND_70CM }, // not sure..
  { 0, NULL, 0, 0 } // <- and entry with "all zeroes" marks the end of the table
}; // end RigCtrl_RadioInfo_CIV[]

const T_SL_TokenList RigCtrl_BandNames[] =
{ // pszKeyword, iTokenValue :
  { "2200 m", RIGCTRL_BAND_2200M}, // note the space between number and SI unit [m]
  { "600 m",  RIGCTRL_BAND_600M },
  { "160 m",  RIGCTRL_BAND_160M },
  { "80 m",   RIGCTRL_BAND_80M  },
  { "60 m",   RIGCTRL_BAND_60M  },
  { "40 m",   RIGCTRL_BAND_40M  },
  { "30 m",   RIGCTRL_BAND_30M  },
  { "20 m",   RIGCTRL_BAND_20M  },
  { "17 m",   RIGCTRL_BAND_17M  },
  { "15 m",   RIGCTRL_BAND_15M  },
  { "12 m",   RIGCTRL_BAND_12M  },
  { "11 m",   RIGCTRL_BAND_11M  }, // no ham band but a 'Sporadic E indicator'
  { "10 m",   RIGCTRL_BAND_10M  },
  { "HF generic", RIGCTRL_BAND_R_SW},
  { "6 m",    RIGCTRL_BAND_6M   },
  { "4 m",    RIGCTRL_BAND_4M   },
  { "2 m",    RIGCTRL_BAND_2M   },
  { "VHF GEN",RIGCTRL_BAND_R_VHF},
  { "70 cm",  RIGCTRL_BAND_70CM },
  { "23 cm",  RIGCTRL_BAND_23CM },
  { "13 cm",  RIGCTRL_BAND_13CM },
  { "9 cm" ,  RIGCTRL_BAND_9CM  },
  { "6 cm" ,  RIGCTRL_BAND_6CM  },
  { "3 cm" ,  RIGCTRL_BAND_3CM  },
  { NULL, 0 } // "all zeros" mark the end of the list
}; // end RigCtrl_BandNames[]


const T_SL_TokenList RigCtrl_OpModes[] =
{ { "CW",  RIGCTRL_OPMODE_CW },
  { "CWN", RIGCTRL_OPMODE_CW | RIGCTRL_OPMODE_NARROW },
  { "CWNN",RIGCTRL_OPMODE_CW | RIGCTRL_OPMODE_VERY_NARROW },
  { "LSB", RIGCTRL_OPMODE_LSB},
  { "USB", RIGCTRL_OPMODE_USB},
  { "AM",  RIGCTRL_OPMODE_AM },
  { "FM",  RIGCTRL_OPMODE_FM },
  { "WFM", RIGCTRL_OPMODE_FM_WIDE },
  { "RTTY",RIGCTRL_OPMODE_RTTY},

  { "CWR",  RIGCTRL_OPMODE_CW | RIGCTRL_OPMODE_REVERSE },
  { "CWRN", RIGCTRL_OPMODE_CW | RIGCTRL_OPMODE_REVERSE | RIGCTRL_OPMODE_NARROW },

  { NULL, 0 } // "all zeros" mark the end of the list
}; // end RigCtrl_OpModes[]

const T_SL_TokenList RigCtrl_SplitModes[] =
{ // In certain Icom transceivers, "Split Mode" isn't simply ON or OFF !
  // Instead, the response for read-command 0x0F indicates
  // "Split Off" (0x00), "Split ON" (0x01), "DUP-" (0x11), "DUP+" (0x12),
  // or [IC-9700] "DD Repeater Simplex mode (RPS)" (0x13).
  // Most of the strings below are similar to the display in modern Icom rigs.
  { "off",   RIGCTRL_SPLIT_MODE_OFF }, // neither "Split" not "Duplex for Repeater operation"
  { "SPLIT", RIGCTRL_SPLIT_MODE_ON  }, // classic "Split" mode, usually a few kHz offset between RX and TX
  { "DUP-",  RIGCTRL_SPLIT_MODE_DUP_NEG }, // "DUP- operation" (IC-9700)
  { "DUP+",  RIGCTRL_SPLIT_MODE_DUP_POS }, // "DUP+ operation" (IC-9700)
  { "RPS",   RIGCTRL_SPLIT_MODE_REPEATER_SIMPLEX }, // <- just an IC-9700 extravaganza ? "DD Repeater Simplex mode (RPS)"

  { NULL, 0 } // "all zeros" mark the end of the list
}; // end RigCtrl_SplitModes[]




const T_RigCtrl_ParamInfo RigCtrl_ParameterInfo[] =
{ // iUnifiedPN,                   pszToken,              pszUnit,  pStringValueTable,
  //  iByteOffsetIntoRigCtrlInstance  , iRigCtrlDataType, iAccessFlags,
  //       pbCIVReadCmd,           iCIVReadCmdLength,     iCIVDataType :

 { RIGCTRL_PN_TRANSCEIVER_ID,      "RadioID",             "",       NULL,
      // "Transceiver ID" alias "Radio ID" alias "Default CI-C Address" (holy crap..)
      offsetof( T_RigCtrlInstance,iDefaultAddress), RIGCTRL_DT_INT, RIGCTRL_AF_READONLY,
      (const BYTE*)"\x19\x00",     2,                     CIV_DTYPE_BCD2 },
 { RIGCTRL_PN_FREQUENCY,           "VFOfreq",             "Hz",     NULL,
      offsetof( T_RigCtrlInstance, dblVfoFrequency), RIGCTRL_DT_DOUBLE, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x03",         1,                     CIV_DTYPE_BCD10 },
 { RIGCTRL_PN_OP_MODE,             "OpMode",              "",       RigCtrl_OpModes,
      offsetof( T_RigCtrlInstance, iOpMode),      RIGCTRL_DT_INT,  RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x04",         1,                     CIV_DTYPE_BCD4/*with "Mode" and "Filter"*/  },
 { RIGCTRL_PN_FILTER_BANDWIDTH,    "FilterBW",            "Hz",     NULL,
      offsetof( T_RigCtrlInstance, iFilterBW_Hz), RIGCTRL_DT_INT,  RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x1A\x03",     2,                     CIV_DTYPE_BCD2/*lousy CODES, not FREQUENCIES*/ },
           // In mode *FM* (Frequency Modulation), an IC-9700 was not happy with "Cmd 1A Subcmd 03":
           // > TX 007 FE FE 00 00 1A 03 FD        ; read FilterBW
           // > RX 006 FE FE 00 A2 FA FD    ; NotOK for 0xFFFFFFFF
           // In mode *CW* (Morse code with "FIL2", an IC-9700 happily responded to "Cmd 1A Subcmd 03":
           // > TX 007 FE FE 00 00 1A 03 FD        ; read FilterBW
           // > RX 008 FE FE 00 A2 1A 03 09 FD   ; FilterBW=815 Hz
           // Expect more "fun" like this (special cases all over the place,
           // even the ENCODING of the stupid TWO-DIGIT BCD VALUE
           // depends on iOpMode (aka modulation type): See IC-9700 CI-V Reference "V4",
           // Icom document number "A7508-3EX-4", page 16. Read, dig it, and CRY OUT LOUD !
           //

  // iUnifiedPN,                   pszToken,              pszUnit,  pStringValueTable,
  //  iByteOffsetIntoRigCtrlInstance  , iRigCtrlDataType, iAccessFlags,
  //       pbCIVReadCmd,           iCIVReadCmdLength,     iCIVDataType :
 { RIGCTRL_PN_DATA_MODE,           "DataMode",            "",       NULL,
      offsetof( T_RigCtrlInstance, iDataMode),     RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_TRANSMITTING,        "Transmitting",        "",       NULL,
      offsetof( T_RigCtrlInstance, iTransmitting), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x1C\x00",     2,                     CIV_DTYPE_BCD2  },
 { RIGCTRL_PN_ATU_ENABLED,         "ATU_enabled",         "",       NULL,
      offsetof( T_RigCtrlInstance, iATU_enabled ), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_XFC_PRESSED,         "XFC_pressed",         "",       NULL,
      offsetof( T_RigCtrlInstance, iXFC_pressed ), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x1C\x02",     2,                     CIV_DTYPE_BCD2  },
 { RIGCTRL_PN_TRANSMIT_FREQ,       "TX_freq",             "Hz",     NULL,
      offsetof( T_RigCtrlInstance, dblTxFrequency), RIGCTRL_DT_DOUBLE, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x1C\x03",     2,                     CIV_DTYPE_BCD10 },
 { RIGCTRL_PN_REPEATER_TONE_FREQ,  "RptrTone",            "Hz",     NULL,
      offsetof( T_RigCtrlInstance, dblRepeaterTone_Hz),RIGCTRL_DT_DOUBLE, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_TONE_SQUELCH_FREQ,   "ToneSquelch",         "Hz",     NULL,
      offsetof( T_RigCtrlInstance, dblToneSquelch_Hz), RIGCTRL_DT_DOUBLE, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },


 // --- SPECTRUM SCOPE related (with CI-V, exchanged via command 0x27) ------>
 { RIGCTRL_PN_SCOPE_MODE         , "ScopeMode",           "",       NULL,
      offsetof( T_RigCtrlInstance, iScopeMode),    RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_SCOPE_SPAN         , "ScopeSpan",           "Hz",     NULL,
      offsetof( T_RigCtrlInstance, dblScopeSpan_Hz), RIGCTRL_DT_DOUBLE, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_SCOPE_ATTEN        , "ScopeAtt",            "dB",     NULL,
      offsetof( T_RigCtrlInstance, dblScopeAttenuator_dB), RIGCTRL_DT_DOUBLE, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_SCOPE_REF_LEVEL    , "ScopeRefLevel",       "dB",     NULL,
      offsetof( T_RigCtrlInstance, dblScopeRefLevel_dB), RIGCTRL_DT_DOUBLE, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x27\x19\x00"/*main scope*/, 3,       CIV_DTYPE_BCD4 | CIV_DTYPE_SUFFIX_BCD2_SIGN },
 { RIGCTRL_PN_SCOPE_EDGE_NR      , "ScopeEdge",           "",       NULL,
      offsetof( T_RigCtrlInstance, iScopeEdgeNumber), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_SCOPE_HOLD_FLAG    , "ScopeHold",           "",       NULL,
      offsetof( T_RigCtrlInstance, iScopeHoldFlag), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_SCOPE_SPEED        , "ScopeSpeed",          "",       NULL,
      offsetof( T_RigCtrlInstance, iScopeSpeed), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },

 //--- Various "poti" settings (with CI-V, exchanged via command 0x14) ------>
 { RIGCTRL_PN_AUDIO_VOLUME_PERCENT, "AudioVolume",        "%",      NULL,
      offsetof( T_RigCtrlInstance, iAudioVolume_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x14\x01",     2,                     CIV_DTYPE_BCD4  },
 { RIGCTRL_PN_RF_GAIN_PERCENT    , "RFGain",              "%",      NULL,
      offsetof( T_RigCtrlInstance, iRFGain_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x14\x02",     2,                     CIV_DTYPE_BCD4  },
 { RIGCTRL_PN_SQUELCH_LEVEL_PERCENT, "SquelchLevel",      "%",      NULL,
      offsetof( T_RigCtrlInstance, iSquelchLevel_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x14\x03",     2,                     CIV_DTYPE_BCD4  },
 { RIGCTRL_PN_NOISE_REDUCTION_PERCENT, "NoiseReduction",  "%",      NULL,
      offsetof( T_RigCtrlInstance, iNoiseReduction_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x14\x06",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_PASSBAND_TUNING_POS1, "PassbandT1",         "%",      NULL, // "inner ring" of "TWIN PBT", 50 % = center
      offsetof( T_RigCtrlInstance, iPassbandTuningPos1), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x14\x07",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_PASSBAND_TUNING_POS2, "PassbandT2",         "%",      NULL, // "outer ring" of "TWIN PBT", 50 % = center
      offsetof( T_RigCtrlInstance, iPassbandTuningPos2), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x14\x08",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_CW_PITCH_HZ        , "CW_Pitch",            "Hz",     NULL,
      offsetof( T_RigCtrlInstance, iCWPitch_Hz), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x14\x09",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_RF_POWER_SETTING_PERCENT, "PowerSet",       "%",      NULL, // "setting" to avoid confusion with the POWER METER MEASUREMENT !
      offsetof( T_RigCtrlInstance, iRFPowerSetting_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x14\x0A",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_MIC_GAIN_PERCENT   , "MicGain",             "%",      NULL,
      offsetof( T_RigCtrlInstance, iMicGain_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x14\x0B",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_KEYER_SPEED_PERCENT, "KeyerSpeed",          "%",      NULL, // 0 = often 6 WPM, 100 % = 48 WPM (at least for Icom)
      offsetof( T_RigCtrlInstance, iKeyerSpeed_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x14\x0C",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_NOTCH_POS_PERCENT  , "NotchPos",            "%",      NULL,
      offsetof( T_RigCtrlInstance, iNotchPos_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x14\x0D",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_COMP_SETTING_PERCENT, "CompSet",            "%",      NULL, // audio compressor SETTING (not a "measurement")
      offsetof( T_RigCtrlInstance, iCompSetting_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x14\x0E",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_BREAK_IN_DELAY_PERCENT, "BkInDelay",        "%",      NULL,
      offsetof( T_RigCtrlInstance, iBreakInDelay_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x14\x0F",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_NBL_SETTING_PERCENT, "NoiseBlanker",        "%",      NULL,
      offsetof( T_RigCtrlInstance, iNoiseBlankerSetting_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x14\x12",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_MONITOR_GAIN_PERCENT, "MonitorGain",        "%",      NULL,
      offsetof( T_RigCtrlInstance, iMonitorGain_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x14\x15",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_VOX_GAIN_PERCENT   , "VoxGain",             "%",      NULL,
      offsetof( T_RigCtrlInstance, iVoxGain_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x14\x16",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_ANTI_VOX_GAIN_PERCENT, "AntiVoxGain",       "%",      NULL,
      offsetof( T_RigCtrlInstance, iAntiVoxGain_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x14\x17",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_BRIGHTNESS_PERCENT , "Brightness",          "%",      NULL,
      offsetof( T_RigCtrlInstance, iBrightness_Percent), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x14\x19",     2,                     CIV_DTYPE_NONE  },
 //------ Various "meter" readings and related statuses --------------------->
 { RIGCTRL_PN_SQUELCH_STATUS     , "SquelchStatus",       "",       NULL,
      offsetof( T_RigCtrlInstance, iSquelchStatus), RIGCTRL_DT_INT, RIGCTRL_AF_READONLY,
      (const BYTE*)"\x15\x01",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_S_METER_LEVEL_DB   , "SMeterLevel",         "dB",     NULL, // here: "decibel over S0",  not the strange CI-V internal unit
      offsetof( T_RigCtrlInstance, iSMeterLevel_dB), RIGCTRL_DT_INT, RIGCTRL_AF_READONLY,
      (const BYTE*)"\x15\x01",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_POWER_METER_PERCENT, "PowerMeter",          "%",      NULL,
      offsetof( T_RigCtrlInstance, iPowerMeterLevel_pcnt), RIGCTRL_DT_INT, RIGCTRL_AF_READONLY,
      (const BYTE*)"\x15\x11",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_SWR_METER_VALUE    , "SWRMeter",            "",       NULL, // dimensionless, usual value range: 1.0 (perfect) to 3.0 (poor)
      offsetof( T_RigCtrlInstance, dblSWRMeterValue),    RIGCTRL_DT_DOUBLE, RIGCTRL_AF_READONLY,
      (const BYTE*)"\x15\x12",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_ALC_METER_LEVEL    , "ALCMeter",            "%",      NULL,
      offsetof( T_RigCtrlInstance, iALCMeterLevel_pcnt), RIGCTRL_DT_INT, RIGCTRL_AF_READONLY,
      (const BYTE*)"\x15\x13",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_COMP_METER_LEVEL_DB, "CompMeter",           "dB",     NULL, // decibel, 0=no compression, 30 dB = maximum reading for IC-7300)
      offsetof( T_RigCtrlInstance, iCompMeterLevel_dB), RIGCTRL_DT_INT, RIGCTRL_AF_READONLY,
      (const BYTE*)"\x15\x14",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_SUPPLY_VOLTAGE_mV  , "SupplyVoltage",       "mV",     NULL,
      offsetof( T_RigCtrlInstance, iSupplyVoltage_mV), RIGCTRL_DT_INT, RIGCTRL_AF_READONLY,
      (const BYTE*)"\x15\x15", /* "Vd" */  2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_DRAIN_CURRENT_mA   , "DrainCurrent",        "mA",     NULL,
      offsetof( T_RigCtrlInstance, iDrainCurrent_mA), RIGCTRL_DT_INT, RIGCTRL_AF_READONLY,
      (const BYTE*)"\x15\x16", /* "Id" */  2,                     CIV_DTYPE_NONE  },

 //-- extended Preamp, AGC speed, Noise blanker, Audio peak filter,
 //   noise reduction, auto notch, etc. With CI-V, via command 0x16) -----
 { RIGCTRL_PN_PREAMP_SETTING,      "PreampSet",           "",       NULL,
      offsetof( T_RigCtrlInstance, iPreampSet), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x16\x02",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_AGC_SPEED,           "AGCSpeed",            "",       NULL,
      offsetof( T_RigCtrlInstance, iAGCSpeed), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x16\x12",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_NOISE_BLANKER_ON,    "NoiseBlankerON",      "",       NULL,
      offsetof( T_RigCtrlInstance, iNoiseBlankerON),RIGCTRL_DT_INT,RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x16\x22",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_NOISE_REDUCT_ON,     "NoiseReductON",       "",       NULL,
      offsetof( T_RigCtrlInstance, iNoiseReductON),RIGCTRL_DT_INT,RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x16\x40",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_AUTO_NOTCH_ON,       "AutoNotchON",         "",       NULL,
      offsetof( T_RigCtrlInstance, iAutoNotchON), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x16\x41",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_RPTR_TONE_ON,        "RptrToneON",          "",       NULL,
      offsetof( T_RigCtrlInstance, iRptrToneON), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x16\x42",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_TONE_SQUELCH_ON,     "ToneSquelchON",       "",       NULL,
      offsetof( T_RigCtrlInstance, iToneSquelchON), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x16\x43",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_COMP_ON,             "CompressON",          "",       NULL,
      offsetof( T_RigCtrlInstance, iCompressON), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x16\x44",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_MONITOR_ON,          "MonitorON",           "",       NULL,
      offsetof( T_RigCtrlInstance, iMonitorON), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x16\x45",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_VOX_ON,              "VoxON",               "",       NULL,
      offsetof( T_RigCtrlInstance, iVoxON), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x16\x46",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_BK_IN_ON,            "BreakInON",           "",       NULL,
      offsetof( T_RigCtrlInstance, iBreakInON), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x16\x47",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_MANUAL_NOTCH_ON,     "ManualNotchON",       "",       NULL,
      offsetof( T_RigCtrlInstance, iManualNotchON), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x16\x48",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_TWIN_PEAK_FILTER_ON, "TwinPeakFilterON",    "",       NULL,
      offsetof( T_RigCtrlInstance, iTwinPeakFilterON),RIGCTRL_DT_INT,RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x16\x4F",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_DIAL_LOCK_ON,        "DialLockON",          "",       NULL,
      offsetof( T_RigCtrlInstance, iDialLockON), RIGCTRL_DT_INT,RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x16\x50",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_DSP_SHARP_SOFT,      "DSP_SharpSoft",       "",       NULL,
      offsetof( T_RigCtrlInstance, iDSP_SharpSoft),RIGCTRL_DT_INT,RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x16\x56",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_MANUAL_NOTCH_WIDTH,  "ManualNotchWidth",    "",       NULL,
      offsetof( T_RigCtrlInstance, iManualNotchWidth),RIGCTRL_DT_INT,RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x16\x57",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_SSB_TX_BANDWIDTH,    "SSB_Tx_BW",           "",       NULL,
      offsetof( T_RigCtrlInstance, iSSB_Tx_BW),RIGCTRL_DT_INT,RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x16\x58",     2,                     CIV_DTYPE_NONE  },

 //-- "TX Band Edge Frequencies", CI-V cmd 0x1E with various sub-commands -->
 { RIGCTRL_PN_NUM_TX_BANDS,       "NumTxBands",           "",       NULL,
      offsetof( T_RigCtrlInstance, iNumTxBands),RIGCTRL_DT_INT,RIGCTRL_AF_READONLY,
      (const BYTE*)"\x1E\x00",     2,                     CIV_DTYPE_BCD2  },
 { RIGCTRL_PN_TX_BAND_EDGES,      "TxBandEdges",          "",       NULL,
      offsetof( T_RigCtrlInstance, TxBandEdges),RIGCTRL_DT_OTHER,RIGCTRL_AF_READONLY,
      (const BYTE*)"\x1E\x01",     2/*plus X*/, CIV_DTYPE_BAND_EDGE_FREQUENCIES  },


 //-- "Special" settings, audio equalizer, connector settings, etc ---------->
 { RIGCTRL_PN_SSB_RX_HPF_LPF,      "SSB_RX_HPF_LPF",      "",       NULL,
      offsetof( T_RigCtrlInstance, iSSB_RX_HPF_LPF), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_SSB_RX_BASS_LEVEL,   "SSB_RX_BASS_LEVEL",   "",       NULL,
      offsetof( T_RigCtrlInstance, iSSB_RX_BASS_LEVEL), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_SSB_RX_TREBLE_LEVEL, "SSB_RX_TREBLE_LEVEL", "",       NULL,
      offsetof( T_RigCtrlInstance, iSSB_RX_TREBLE_LEVEL), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_SSB_CW_SYNC_TUNING,  "SSB_CW_sync_tuning",  "",       NULL,
      offsetof( T_RigCtrlInstance, iSSB_CW_sync_tuning), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_CALIBRATION_MARKER,  "calib_marker",        "",       NULL,
      offsetof( T_RigCtrlInstance, iCalibMarker), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_AF_OUTPUT_LEVEL_TO_ACC, "AF2ACC_level",     "%",      NULL,
      offsetof( T_RigCtrlInstance,    iAF2ACC_level), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_SQUELCH_FOR_AF_ON_ACC, "squelch_for_ACC",   "%",      NULL,
      offsetof( T_RigCtrlInstance, iSquelch_for_ACC), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_BEEP_AND_SPEECH_ON_ACC, "beep_on_ACC",      "",       NULL,
      offsetof( T_RigCtrlInstance, iBeep_on_ACC), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_IF_OUTPUT_LEVEL_TO_ACC, "IF_level_on_ACC",  "%",      NULL,
      offsetof( T_RigCtrlInstance, iIF_level_on_ACC), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_MOD_INPUT_LEVEL_FROM_ACC, "MOD_level_ACC",  "%",      NULL,
      offsetof( T_RigCtrlInstance, iMOD_level_ACC), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_MOD_INPUT_LEVEL_FROM_USB, "MOD_level_USB",  "%",      NULL,
      offsetof( T_RigCtrlInstance, iMOD_level_USB), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_MOD_IN_CONN_FOR_DATA_OFF, "MOD_conn_NoData","",       NULL,
      offsetof( T_RigCtrlInstance, iMOD_conn_NoData), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_MOD_IN_CONN_FOR_DATA, "MOD_conn_Data",      "",       NULL,
      offsetof( T_RigCtrlInstance, iMOD_conn_Data), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_CIV_TRANSCEIVE,      "CIV_transceive",      "",       NULL,
      offsetof( T_RigCtrlInstance, iCIV_transceive), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_CIV_ADDR_USB_TO_REMOTE, "CIV_addr_USB_Rem", "",       NULL,
      offsetof( T_RigCtrlInstance,    iCIV_addr_USB_Rem), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_CIV_OUTPUT_FOR_ANT,  "CIV_for_ANT",         "",       NULL,
      offsetof( T_RigCtrlInstance, iCIV_for_ANT), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_CIV_UNLINK_FROM_REMOTE,"CIV_unlink_Rem",    "",       NULL,
      offsetof( T_RigCtrlInstance, iCIV_unlink_Remote), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_CIV_USB_ECHO,        "CIV_USB_echo",        "",       NULL,
      offsetof( T_RigCtrlInstance, iCIV_USB_echo), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_IP_PLUS,             "IP_Plus",             "",       NULL,
      offsetof( T_RigCtrlInstance, iIP_Plus),      RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_NTP_SERVER_ACCESS,   "NTP_access",          "",       NULL,
      offsetof( T_RigCtrlInstance, iNTP_Server_Access), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_NTP_ACCESS_RESULT,   "NTP_result",          "",       NULL,
      offsetof( T_RigCtrlInstance, iNTP_Access_Result), RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_AF_MUTE,             "AF_mute",             "",       NULL,
      offsetof( T_RigCtrlInstance, iAF_mute),      RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 //-- Even more settings (in CI-V, controlled via command 0x0E..0x10) --
 { RIGCTRL_PN_SCAN_MODE,           "ScanMode",            "",       NULL,
      offsetof( T_RigCtrlInstance, iScanMode),     RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE ,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_SPLIT_MODE,          "SplitMode",           "",       RigCtrl_SplitModes,
      offsetof( T_RigCtrlInstance, iSplitMode),    RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x0F",         1,                     CIV_DTYPE_BCD1  },
 { RIGCTRL_PN_TUNING_STEP,         "TuningStep",          "Hz",     NULL,
      offsetof( T_RigCtrlInstance, iTuningStep_Hz),RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_ATTENUATOR_DB,       "Attenuator",          "dB",     NULL,
      offsetof( T_RigCtrlInstance, iAttenuator_dB),RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },

 { RIGCTRL_PN_RIT_FREQ, "RIT_freq",  "Hz", NULL, // 'Receiver Incremental Tuning' in Hertz. IC-7300: Cmd 0x21, Sub 0x00
      offsetof( T_RigCtrlInstance, iRIT_freq_Hz),  RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x21\x00",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_XIT_FREQ, "XIT_freq",  "Hz", NULL, // 'Transmitter Incremental Tuning' in Hertz. IC-7300: no extra VALUE for "delta TX" (same variable as the "RIT frequency in Hertz") !
      offsetof( T_RigCtrlInstance, iXIT_freq_Hz),  RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)NULL,           0,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_RIT_ENABLED, "RIT_ON",  "", NULL,  // 'Receiver Incremental Tuning' ON/OFF. IC-7300: Cmd 0x21, Sub 0x01
      offsetof( T_RigCtrlInstance, iRIT_ON ),  RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x21\x01",     2,                     CIV_DTYPE_NONE  },
 { RIGCTRL_PN_XIT_ENABLED, "XIT_ON",  "", NULL,  // 'Transmitter Incremental Tuning' ON/OFF. IC-7300: Cmd 0x21, Sub 0x02.
                                           // IC-9700: No such SUB-COMMAND for Cmd 0x21 ! !
      offsetof( T_RigCtrlInstance, iXIT_ON ),  RIGCTRL_DT_INT, RIGCTRL_AF_READWRITE,
      (const BYTE*)"\x21\x02",     2,                     CIV_DTYPE_NONE  },

 // As usual, the end of the table is marked by an entry with "all zeroes" :
  // iUnifiedPN,                   pszToken,              pszUnit,  pStringValueTable,
  //  iByteOffsetIntoRigCtrlInstance  , iRigCtrlDataType, iAccessFlags,
  //       pbCIVReadCmd,           CIVReadCmdLength,        iCIVDataType :
 {   0,                             NULL,                 NULL,     NULL,
       0,                               0,                0,
       (const BYTE*)NULL,          0,                       0 }

}; // end RigCtrl_ParameterInfo[]


/* EOF < C:\cbproj\Remote_CW_Keyer\RigControl.c > */


